/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <math.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <snmp.h>
#include <mibsup.h>
#include <message.h>
#include <mac.h>

#include "dstate.h"
#include "dstatc.h"
#include "dstat.h"


IMPORT BOOLEAN BeholderStrictRMON;


#define INDEXSIZE      		1
#define IFTINDEXSIZE     	2
#define LENGTHINDEXSIZE     2
#define TYPEINDEXSIZE     	2


static MIB_LOCAL    *etherStats = NULL;


static BOOLEAN RmonNext (SNMP_OBJECT *obj, MIB_LOCAL **local, WORD idlen, WORD indexsize);



BOOLEAN DnpapEtherStatsMInit(VOID)
{
    MessageConfig(DNPAP_STATISTICS_ERROR, "DnpapEtherStat");
    return TRUE;
}


BOOLEAN RmonNext (SNMP_OBJECT *obj, MIB_LOCAL **local, WORD idlen, WORD indexsize)
{
DNPAP_ETHER_STATS *data;

    if (*local == NULL || indexsize < 1)
        return FALSE;

    switch (indexsize)
    {            
    case IFTINDEXSIZE:
        if (obj->IdLen >= idlen + indexsize)
        {
    	data = (DNPAP_ETHER_STATS*)(*local)->Data;
        if (data->Status == SNMP_VALID && obj->Id[idlen+1] < data->Buckets)
        {
            obj->Id[idlen] = (*local)->Index;
            obj->Id[idlen+1]++;
            obj->IdLen = idlen + IFTINDEXSIZE;
            break;
        }
        *local = (*local)->Next;
        }
        while (*local != NULL && 
               (((DNPAP_ETHER_STATS*)(*local)->Data)->Status != SNMP_VALID ||
                ((DNPAP_ETHER_STATS*)(*local)->Data)->Buckets == 0))
            *local = (*local)->Next;
        if (*local == NULL)
            return FALSE;
        obj->Id[idlen] = (*local)->Index;
        obj->Id[idlen+1] = 1;
        obj->IdLen = idlen + IFTINDEXSIZE;
        break;
    case INDEXSIZE:
        if (obj->IdLen >= idlen + indexsize)
        {
	        /* only go to next collector if requested index was higher or equal */
            if (obj->Id[idlen] >= (*local)->Index)
                *local = (*local)->Next;
            if (*local == NULL)
                return FALSE;
        }
        obj->Id[idlen] = (*local)->Index;
        obj->IdLen = idlen + INDEXSIZE;
        break;
    }
    return TRUE;
}

WORD DnpapEtherStatsIndex (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;
    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        Obj->Syntax.LngInt = local->Index;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsDataSource (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_IFACE    *iface;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {           
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, STAT_BOOLEAN_DATASOURCE))
			return SNMP_NOSUCHNAME;
        memcpy (Obj->Syntax.BufInt, data->Source, data->SourceLen * sizeof (LONG));
        Obj->SyntaxLen = data->SourceLen;
        return SNMP_NOERROR;
    case SNMP_PDU_SET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_UNDERCREATION)
            return SNMP_READONLY;
	    if (data->SourceLen != Obj->SyntaxLen ||
	        memcmp(data->Source, Obj->Syntax.BufInt, (data->SourceLen-1)*sizeof(data->Source[0])))
	        return SNMP_BADVALUE;
	    if ((iface = MacIfaceGet((WORD) Obj->Syntax.BufInt[Obj->SyntaxLen-1])) == NULL)
	        return SNMP_BADVALUE;
        if (iface->type != MAC_TYPE_ETHERNET_CSMACD && iface->type != MAC_TYPE_88023_CSMACD)
        	return SNMP_BADVALUE;
	    memcpy(data->Source, Obj->Syntax.BufInt, Obj->SyntaxLen*sizeof(data->Source[0]));
	    data->SourceLen = Obj->SyntaxLen;
	    data->Iface = iface;
		BooleanSetTrue(data->ObjectSet, STAT_BOOLEAN_DATASOURCE);
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsDropEvents (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.LostPkts + stat.DiscardedPkts - (data->BaseStat.LostPkts + data->BaseStat.DiscardedPkts);
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsOctets (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        Obj->Syntax.LngUns = data->Octets;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        Obj->Syntax.LngUns = data->Pkts;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}
                                                
WORD DnpapEtherStatsBroadcastPkts (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.BroadcastPkts - data->BaseStat.BroadcastPkts;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsMulticastPkts (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.MulticastPkts - data->BaseStat.MulticastPkts;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsCRCAlignErrors (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.CRCAlignErrors - data->BaseStat.CRCAlignErrors;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsUndersizePkts (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.UndersizePkts - data->BaseStat.UndersizePkts;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsOversizePkts (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.OversizePkts - data->BaseStat.OversizePkts;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsFragments (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.Fragments - data->BaseStat.Fragments;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsJabbers (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.Jabbers - data->BaseStat.Jabbers;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

#ifdef REALLY_IFMIB_SPECIFIC
WORD DnpapEtherStatsInPkts (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.inPkts - data->BaseStat.inPkts;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}
                                                
WORD DnpapEtherStatsInErrors (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.inErrors - data->BaseStat.inErrors;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}
                                                
WORD DnpapEtherStatsOutPkts (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.outPkts - data->BaseStat.outPkts;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}
                                                
WORD DnpapEtherStatsOutErrors (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.outErrors - data->BaseStat.outErrors;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}
#endif

WORD DnpapEtherStatsCollisions (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
MAC_STAT     stat;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_VALID)
        {
            Obj->Syntax.LngUns = 0UL;
            return SNMP_NOERROR;
        }
        MacStatistics(data->Iface, &stat);
        Obj->Syntax.LngUns = stat.Collisions - data->BaseStat.Collisions;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts64 (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        Obj->Syntax.LngUns = data->Pkts64Octets;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts65to127 (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        Obj->Syntax.LngUns = data->Pkts65to127Octets;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts128to255 (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        Obj->Syntax.LngUns = data->Pkts128to255Octets;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts256to511 (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        Obj->Syntax.LngUns = data->Pkts256to511Octets;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts512to1023 (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        Obj->Syntax.LngUns = data->Pkts512to1023Octets;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts1024to1518 (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        Obj->Syntax.LngUns = data->Pkts1024to1518Octets;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsUtilInterval (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
        case SNMP_PDU_NEXT:
            if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
                return SNMP_NOSUCHNAME;
        case SNMP_PDU_GET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            Obj->Syntax.LngUns = data->UtilInterval;
            return SNMP_NOERROR;
        case SNMP_PDU_SET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            if (data->Status != SNMP_UNDERCREATION)
                return SNMP_READONLY;              
            if (Obj->Syntax.LngUns < 1)
            	return SNMP_BADVALUE;
            data->UtilInterval = Obj->Syntax.LngUns;
            return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsUtilization (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        Obj->Syntax.LngInt = data->Utilization;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsIFTBuckets (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
ULONG *buckets1, *buckets2;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
        case SNMP_PDU_NEXT:
            if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
                return SNMP_NOSUCHNAME;
        case SNMP_PDU_GET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            Obj->Syntax.LngInt = data->IFTBuckets;
            return SNMP_NOERROR;
        case SNMP_PDU_SET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            if (data->Status != SNMP_UNDERCREATION)
                return SNMP_READONLY;              
            if (Obj->Syntax.LngInt < 1)
            	return SNMP_BADVALUE;
            if ((buckets1 = ReNewBuckets(data->IFTArrayPkts, Obj->Syntax.LngInt)) == NULL)
            	return SNMP_BADVALUE;
            if ((buckets2 = ReNewBuckets(data->IFTArrayOctets, Obj->Syntax.LngInt)) == NULL)
            	return SNMP_BADVALUE;
            data->IFTBuckets = Obj->Syntax.LngInt;
            data->IFTArrayPkts = buckets1;
            data->IFTArrayOctets = buckets2;
            return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsIFTSize (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
        case SNMP_PDU_NEXT:
            if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
                return SNMP_NOSUCHNAME;
        case SNMP_PDU_GET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            Obj->Syntax.LngInt = data->IFTSize;
            return SNMP_NOERROR;
        case SNMP_PDU_SET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            if (data->Status != SNMP_UNDERCREATION)
                return SNMP_READONLY;              
            if (Obj->Syntax.LngInt < 1)
            	return SNMP_BADVALUE;
            data->IFTSize = Obj->Syntax.LngInt;
            return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsLengthBuckets (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
ULONG *buckets1, *buckets2;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
        case SNMP_PDU_NEXT:
            if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
                return SNMP_NOSUCHNAME;
        case SNMP_PDU_GET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            Obj->Syntax.LngInt = data->LengthBuckets;
            return SNMP_NOERROR;
        case SNMP_PDU_SET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            if (data->Status != SNMP_UNDERCREATION)
                return SNMP_READONLY;              
            if (Obj->Syntax.LngInt < 1 || Obj->Syntax.LngInt > MAC_FRAME_ETHERNET_CSMACD)
            	return SNMP_BADVALUE;
            if ((buckets1 = ReNewBuckets(data->LengthArrayPkts, Obj->Syntax.LngInt)) == NULL)
            	return SNMP_BADVALUE;
            if ((buckets2 = ReNewBuckets(data->LengthArrayOctet, Obj->Syntax.LngInt)) == NULL)
            	return SNMP_BADVALUE;
            data->LengthBuckets = Obj->Syntax.LngInt;
            data->LengthSize = ((FLOAT) MAC_FRAME_ETHERNET_CSMACD)/Obj->Syntax.LngInt + 0.999999;
            data->LengthArrayPkts = buckets1;
            data->LengthArrayOctets = buckets2;
            return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsLengthSize (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;
ULONG *buckets1, *buckets2;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
        case SNMP_PDU_NEXT:
            if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
                return SNMP_NOSUCHNAME;
        case SNMP_PDU_GET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            Obj->Syntax.LngInt = data->LengthSize;
            return SNMP_NOERROR;
        case SNMP_PDU_SET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            if (data->Status != SNMP_UNDERCREATION)
                return SNMP_READONLY;              
            if (Obj->Syntax.LngInt < 1 || Obj->Syntax.LngInt > MAC_FRAME_ETHERNET_CSMACD)
            	return SNMP_BADVALUE;
            buckets = ((FLOAT) MAC_FRAME_ETHERNET_CSMACD)/Obj->Syntax.LngInt + 0.999999;
            if ((buckets1 = ReNewBuckets(data->LengthArrayPkts, Obj->Syntax.LngInt)) == NULL)
            	return SNMP_BADVALUE;
            if ((buckets2 = ReNewBuckets(data->LengthArrayOctet, Obj->Syntax.LngInt)) == NULL)
            	return SNMP_BADVALUE;
            data->LengthSize = Obj->Syntax.LngInt;
            data->LengthBuckets = buckets;
            data->LengthArrayPkts = buckets1;
            data->LengthArrayOctets = buckets2;
            return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsTypeBuckets (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
        case SNMP_PDU_NEXT:
            if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
                return SNMP_NOSUCHNAME;
        case SNMP_PDU_GET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            Obj->Syntax.LngInt = data->TypeBuckets;
            return SNMP_NOERROR;
        case SNMP_PDU_SET:
            return SNMP_READONLY;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsOwner (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        data = (DNPAP_ETHER_STATS*) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, STAT_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
        memcpy (Obj->Syntax.BufChr, data->Owner, data->OwnerLen);
        Obj->SyntaxLen = data->OwnerLen;
        return SNMP_NOERROR;
    case SNMP_PDU_SET:
        data = (DNPAP_ETHER_STATS*) local->Data;
        if (data->Status != SNMP_UNDERCREATION)
            return SNMP_READONLY;
        memcpy (data->Owner, Obj->Syntax.BufChr, Obj->SyntaxLen);
        data->OwnerLen = Obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, STAT_BOOLEAN_OWNER);
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsStatus (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE);

    switch (Obj->Request)
    {
    case SNMP_PDU_GET:
        if (local == NULL)
        {
            Obj->Syntax.LngInt = SNMP_INVALID;
            return SNMP_NOSUCHNAME;
        }
        data = (DNPAP_ETHER_STATS*) local->Data;
        Obj->Syntax.LngInt = data->Status;
        return SNMP_NOERROR;
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, INDEXSIZE) == FALSE)
        {
            Obj->Syntax.LngInt = SNMP_INVALID;
            return SNMP_NOSUCHNAME;
        }
        data = (DNPAP_ETHER_STATS*) local->Data;
        Obj->Syntax.LngInt = data->Status;
        return SNMP_NOERROR;
    case SNMP_PDU_SET:
        if (local == NULL)
        {
            switch (Obj->Syntax.LngInt)
            {
            case SNMP_CREATEREQUEST:
                if ((local = MibInsert(Obj, &etherStats, IdLen, INDEXSIZE)) == NULL)
                    return SNMP_GENERROR;
                if ((local->Data = DnpapMalloc(sizeof(DNPAP_ETHER_STATS))) == NULL)
                    return SNMP_GENERROR;
                data = (DNPAP_ETHER_STATS*) local->Data;
                memset(data, 0, sizeof(DNPAP_ETHER_STATS));
                if (DnpapStatCInit(data) == TRUE)
                {
                    data->Status = SNMP_UNDERCREATION;
                    DnpapMessage(DMC_MESSAGE, DSTAT_CREATE, "detherstat: collector %ld created", local->Index);
                    return SNMP_NOERROR;
                }
                DnpapFree(local->Data);
                MibRemove (Obj, &etherStats, IdLen, INDEXSIZE);
                return SNMP_GENERROR;
            default:
                return SNMP_NOSUCHNAME;
            }
        }
        data = (DNPAP_ETHER_STATS*) local->Data;
        switch (data->Status)
        {
        case SNMP_UNDERCREATION:
            switch (Obj->Syntax.LngInt)
            {
            case SNMP_VALID:
                if (DnpapStatCStart(data) == TRUE)
                {
                    data->Status = SNMP_VALID;
                    DnpapMessage(DMC_MESSAGE, DSTAT_ACTIVE, "detherstat: collector %ld active", local->Index);
                    return SNMP_NOERROR;
                }
                else
                    return SNMP_BADVALUE;
            case SNMP_INVALID:
                if (DnpapStatCStop(data) == TRUE)
                {
                    DnpapFree(local->Data);
                    DnpapMessage(DMC_MESSAGE, DSTAT_DESTROY, "detherstat: collector %ld destroyed", local->Index);
                    if (MibRemove (Obj, &etherStats, IdLen, INDEXSIZE) == TRUE)
                        return SNMP_NOERROR;
                }
                return SNMP_GENERROR;
            default:
                return SNMP_BADVALUE;
            }
        case SNMP_VALID:
            switch (Obj->Syntax.LngInt)
            {
            case SNMP_INVALID:
                if (DnpapStatCStop(data) == TRUE)
                {
                    DnpapFree(local->Data);
                    DnpapMessage(DMC_MESSAGE, DSTAT_DESTROY, "detherstat: collector %ld destroyed", local->Index);
                    if (MibRemove (Obj, &etherStats, IdLen, INDEXSIZE) == TRUE)
                        return SNMP_NOERROR;
                }
                return SNMP_GENERROR;
            case SNMP_VALID:
                return SNMP_NOERROR;
            default:
                return SNMP_BADVALUE;
            }
        }
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsIFTIndex (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;

    if ((local = MibRmon(Obj, etherStats, IdLen, IFTINDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, IFTINDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        Obj->Syntax.LngInt = local->Index;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsIFTBucket (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;

    if ((local = MibRmon(Obj, etherStats, IdLen, IFTINDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, IFTINDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        Obj->Syntax.LngInt = Obj->Id[IdLen+1];
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsIFTPkts (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, IFTINDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
        case SNMP_PDU_NEXT:
            if (RmonNext (Obj, &local, IdLen, IFTINDEXSIZE) == FALSE)
                return SNMP_NOSUCHNAME;
        case SNMP_PDU_GET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            Obj->Syntax.LngInt = data->IFTArrayPkts[Obj->Id[IdLen+1]-1];
            return SNMP_NOERROR;
        case SNMP_PDU_SET:
            return SNMP_READONLY;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsIFTOctets (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, IFTINDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
        case SNMP_PDU_NEXT:
            if (RmonNext (Obj, &local, IdLen, IFTINDEXSIZE) == FALSE)
                return SNMP_NOSUCHNAME;
        case SNMP_PDU_GET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            Obj->Syntax.LngInt = data->IFTArrayOctets[Obj->Id[IdLen+1]-1];
            return SNMP_NOERROR;
        case SNMP_PDU_SET:
            return SNMP_READONLY;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsLengthIndex (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;

    if ((local = MibRmon(Obj, etherStats, IdLen, LENGTHINDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, LENGTHINDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        Obj->Syntax.LngInt = local->Index;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsLengthBucket (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;

    if ((local = MibRmon(Obj, etherStats, IdLen, LENGTHINDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, LENGTHINDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        Obj->Syntax.LngInt = Obj->Id[IdLen+1];
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsLengthPkts (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, LENGTHINDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
        case SNMP_PDU_NEXT:
            if (RmonNext (Obj, &local, IdLen, LENGTHINDEXSIZE) == FALSE)
                return SNMP_NOSUCHNAME;
        case SNMP_PDU_GET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            Obj->Syntax.LngInt = data->IFTArrayOctets[Obj->Id[IdLen+1]-1];
            return SNMP_NOERROR;
        case SNMP_PDU_SET:
            return SNMP_READONLY;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsLengthOctets (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, LENGTHINDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
        case SNMP_PDU_NEXT:
            if (RmonNext (Obj, &local, IdLen, LENGTHINDEXSIZE) == FALSE)
                return SNMP_NOSUCHNAME;
        case SNMP_PDU_GET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            Obj->Syntax.LngInt = data->LengthArrayOctets[Obj->Id[IdLen+1]-1];
            return SNMP_NOERROR;
        case SNMP_PDU_SET:
            return SNMP_READONLY;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsTypeIndex (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;

    if ((local = MibRmon(Obj, etherStats, IdLen, TYPEINDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, TYPEINDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        Obj->Syntax.LngInt = local->Index;
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsTypeBucket (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;

    if ((local = MibRmon(Obj, etherStats, IdLen, TYPEINDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
    case SNMP_PDU_NEXT:
        if (RmonNext (Obj, &local, IdLen, TYPEINDEXSIZE) == FALSE)
            return SNMP_NOSUCHNAME;
    case SNMP_PDU_GET:
        Obj->Syntax.LngInt = Obj->Id[IdLen+1];
        return SNMP_NOERROR;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsTypePkts (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, TYPEINDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
        case SNMP_PDU_NEXT:
            if (RmonNext (Obj, &local, IdLen, TYPEINDEXSIZE) == FALSE)
                return SNMP_NOSUCHNAME;
        case SNMP_PDU_GET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            Obj->Syntax.LngInt = data->TypeArrayOctets[Obj->Id[IdLen+1]-1];
            return SNMP_NOERROR;
        case SNMP_PDU_SET:
            return SNMP_READONLY;
    }
    return SNMP_GENERROR;
}

WORD DnpapEtherStatsTypeOctets (SNMP_OBJECT *Obj, WORD IdLen)
{
MIB_LOCAL    *local = NULL;
DNPAP_ETHER_STATS  *data;

    if ((local = MibRmon(Obj, etherStats, IdLen, TYPEINDEXSIZE)) == NULL)
        return SNMP_NOSUCHNAME;

    switch (Obj->Request)
    {
        case SNMP_PDU_NEXT:
            if (RmonNext (Obj, &local, IdLen, TYPEINDEXSIZE) == FALSE)
                return SNMP_NOSUCHNAME;
        case SNMP_PDU_GET:
            data = (DNPAP_ETHER_STATS*) local->Data;
            Obj->Syntax.LngInt = data->TypeArrayOctets[Obj->Id[IdLen+1]-1];
            return SNMP_NOERROR;
        case SNMP_PDU_SET:
            return SNMP_READONLY;
    }
    return SNMP_GENERROR;
}
