/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>   
#include <memory.h>   
#include <dmemory.h>   
#include <message.h>   
#include <stdio.h>   
#include <string.h>   
#include <chain.h>   
#include <udp.h>

#ifdef OS2
	#include <pipe.h>
#endif

#include <snmp.h>
#include <sysmib.h>
#include <mibrqs.h>
#include <mibsup.h>
#include <config.h>

#include "agent.h"





static BOOLEAN      RcveUdp(UDP_DESCR *descr, CHAIN *chain, UDP_HDR *udpHdr, IP_HDR *ipHdr);

#ifdef OS2
static BOOLEAN      RcvePipe(PIPE_DESCR *descr, USHORT pipe, BYTE *data, WORD length);
#endif

static BOOLEAN      Request(BYTE *rqs, UINT rqsLen, BYTE **rsp, UINT* rspLen, DWORD addr);
static AGENT_COMM   *CommFind(BYTE *comm, WORD commLen);
static BOOLEAN      HostInComm(BYTE *comm, WORD commLen, DWORD addr);



static BYTE *trapComm = NULL;
static USHORT trapCommLen;
static BOOLEAN trapAuthenTraps = TRUE;


static AGENT_STAT agentStat;


static UDP_DESCR agentUdp =
{
    RcveUdp,
    161,
    UDP_ADDR_ANY,
    UDP_PORT_ANY,
    UDP_ADDR_ANY,
    0
};
	   

#ifdef OS2
static PIPE_DESCR agentPipe =
{
    RcvePipe,
    0
};
#endif


static AGENT_COMM *agentCommList = 0;



BOOLEAN AgentInit(LONG port)
{
    BYTE            name[255];
    BYTE            *p;
    USHORT          n;
    DWORD           addr;
    AGENT_COMM      *comm;
    AGENT_HOST      *host;
    WORD            i,j;
    static BOOLEAN  init = FALSE;
    
    if (!init)
    {
		MessageConfig(AGENT_ERROR, "Agent");
		
        i = 1;
        sprintf(name, "agent.community.%d", i);
        while (ConfigGetBytes(name, &p, &n))
        {
            comm = DnpapMalloc(sizeof(AGENT_COMM));
            if (comm!=0)
            {    
                comm->index     = i;
                memcpy(comm->comm, p, n);
                comm->commLen   = n;
                comm->hostList  = 0;
                comm->next      = agentCommList;
                agentCommList   = comm;

                j = 1;
                sprintf(name, "agent.host.%d.%d", i, j);
                while (ConfigGetIPAddr(name, &addr))
                {
                    host = DnpapMalloc(sizeof(AGENT_HOST));
                    if (host!=0)
                    {
                        host->addr      = addr;
                        host->next      = comm->hostList;
                        comm->hostList  = host;
                    }

                    j++;
                    sprintf(name, "agent.host.%d.%d", i, j);
                }
            }
            
            i++;
            sprintf(name, "agent.community.%d", i);
        }
        
		if (ConfigGetBytes("agent.trap.community", &trapComm, &trapCommLen) == FALSE)
		{
			DnpapMessage(DMC_WARNING, AGENT_ERROR, "agent: authentication-failure traps will not be sent");
			trapAuthenTraps = FALSE;
		}
		else				
		{
			ConfigGetBoolean("agent.trap.enable", &trapAuthenTraps);
		}
						   
		if (port == 0)
		{
			port = agentUdp.locPort;
			ConfigGetLong("agent.local.port", &port);
		}

        agentUdp.locPort = port;
			
        if (UdpRegister(&agentUdp))
        {
#ifdef OS2
            init = PipeRegister(&agentPipe);
#else
			init = TRUE;			
#endif
        }
    }

    return init;
}


AGENT_STAT *AgentStatistics(void)
{
    return &agentStat;
}


BOOLEAN AgentGetAuthenTraps(void)
{
    return trapAuthenTraps;
}


VOID AgentSetAuthenTraps(BOOLEAN enable)
{
    trapAuthenTraps = enable;
}


BOOLEAN AgentSendTrap (BYTE *comm, WORD commLen, ULONG systime, WORD gen,
                       WORD spec, SNMP_OBJECT *list, WORD listLen)
{
    BYTE            buffer[AGENT_BUFFER_SIZE];
    BYTE            *trapBuffer;
    UINT            trapLen;
    SNMP_PDU        pdu;
    UDP_HDR         udpHdr;
    IP_HDR          ipHdr;
    CHAIN           new;
    AGENT_COMM      *descr;
    AGENT_HOST      *host;
    DWORD           addr;
    BOOLEAN         success = FALSE;
    SNMP_OBJECT     object = {SNMP_PDU_GET, {1,3,6,1,2,1,1,2,0}, 9};

    if (MibRequest(&object) == SNMP_NOERROR)
    {
        pdu.Type            = SNMP_PDU_TRAP;
        pdu.Trap.General    = gen;
        pdu.Trap.Specific   = spec;
        pdu.Trap.Time       = systime;
        pdu.Trap.IdLen      = object.SyntaxLen;
        memcpy(pdu.Trap.Id, object.Syntax.BufInt,
            object.SyntaxLen * sizeof(object.Syntax.BufInt[0]));

        descr=CommFind(comm, commLen);
        if (descr!=0)
        {
            success = TRUE;
            for (host=descr->hostList; host!=0; host=host->next)
            {
                addr = IpRouteFindSrc(host->addr);
                
                pdu.Trap.IpAddress = IpH2NDWord(addr);
                
                udpHdr.src  = UDP_PORT_ANY;
                udpHdr.dst  = 162;

                ipHdr.dst   = host->addr;
                ipHdr.src   = addr;
                ipHdr.iol   = 0;
                ipHdr.tos   = 0;
                
                trapBuffer  = buffer;
                trapLen     = sizeof(buffer);
                if (SnmpEnc(&trapBuffer,
                    &trapLen,
                    &pdu,
                    comm,
                    commLen,
                    list,
                    listLen))
                {
                    ChainAlloc(&new, buffer, sizeof(buffer),
                                            trapLen, trapBuffer-buffer, 0);
                    success = success && UdpSend(&new, &udpHdr, &ipHdr, 0);
                }
            }
        }
    }
    return success;
}

static BOOLEAN RcveUdp(UDP_DESCR *descr, CHAIN *chain, UDP_HDR *udpHdr, IP_HDR *ipHdr)
{
    WORD    port;
    DWORD   addr;
    BYTE    buffer[AGENT_BUFFER_SIZE];
    CHAIN   new;
    BYTE    *rqsBuffer;
    WORD    rqsLen;
    CHAIN   *rqsChain;
    BYTE    *rspBuffer;
    UINT    rspLen;
    CHAIN   *rspChain;
    BOOLEAN success = FALSE;
    

    rqsChain    = chain;
    rqsLen      = ChainLength(rqsChain);
    rqsBuffer   = ChainPop(&rqsChain, rqsLen);

    if (rqsBuffer!=0)
    {
        rspChain    = &new;
        rspBuffer   = buffer;
        rspLen      = sizeof(buffer);

        if (Request(rqsBuffer, rqsLen, &rspBuffer, &rspLen, ipHdr->src))
        {
            ChainAlloc(rspChain, buffer, sizeof(buffer),
                                    rspLen, rspBuffer-buffer, 0);
            
            port        = udpHdr->src;
            udpHdr->src = udpHdr->dst;
            udpHdr->dst = port;

            addr        = ipHdr->src;
            ipHdr->src  = ipHdr->dst;
            ipHdr->dst  = addr;

            success = UdpSend(rspChain, udpHdr, ipHdr, descr);

        }
        if (rqsChain != chain)
            ChainFree(rqsChain);
    }
    
    
    if (success)
        agentStat.datagrams++;
    else
        agentStat.errors++;

    return success;
}


#ifdef OS2
static BOOLEAN RcvePipe(PIPE_DESCR *descr, USHORT pipe, BYTE *data, WORD length)
{
    BYTE    buffer[AGENT_BUFFER_SIZE];
    BYTE    *rspBuffer;
    WORD    rspLen;
    BOOLEAN success = FALSE;

    rspBuffer = buffer;
    rspLen    = sizeof(buffer);

    /* set source IP address to 0 for pipes */
    if (Request(data, length, &rspBuffer, &rspLen, 0))
    {
        success = PipeSend(pipe, rspBuffer, rspLen);
    }
    
    if (success)
        agentStat.datagrams++;
    else
        agentStat.errors++;

    return success;
}
#endif




BOOLEAN Request(BYTE *rqsBuffer, UINT rqsLen, BYTE **rspBuffer, UINT* rspLen, DWORD addr)
{
    SNMP_PDU    pdu;
	SNMP_REQUEST *rqs = &pdu.Request;
    BYTE        comm[AGENT_COMMUNITY_SIZE];
    UINT        commLen;
    SNMP_OBJECT list[AGENT_LIST_SIZE];
    UINT        listLen;
    WORD        status, index;
	BYTE		address[4];
	WORD		perm;

    if (!SnmpDec(rqsBuffer,
                rqsLen,
                &pdu,
                comm,
                AGENT_COMMUNITY_SIZE,
                &commLen,
                list,
                AGENT_LIST_SIZE,
                &listLen))
    {    
		rqs->Type = SNMP_PDU_RESPONSE;
		rqs->ErrorStatus = snmpErrStatus;
		rqs->ErrorIndex = snmpErrIndex;

		if (!SnmpErrEnc(rqs))
        { 
            DnpapMessage(DMC_ERROR, AGENT_ERROR, "agent: SNMP error encode failed");
            return FALSE;  
        }
 
   		memcpy(*rspBuffer, rqsBuffer, rqsLen);
		*rspLen = rqsLen;
		return TRUE;
    }
    
    if (pdu.Type != SNMP_PDU_GET  &&
        pdu.Type != SNMP_PDU_NEXT &&
        pdu.Type != SNMP_PDU_SET)
    {       
        DnpapMessage(DMC_ERROR, AGENT_ERROR, "agent: unknown pdu");
        return FALSE;
    }
    
    if ((perm = MibProfil(comm, commLen)) == 0)
	{
        DnpapMessage(DMC_WARNING, AGENT_ERROR, "agent: unknown community name");
		agentStat.InBadCommunityNames++;
        return FALSE;
	}
	
#ifdef OS2 
    if (addr != 0 && !HostInComm(comm, commLen, addr))
#else
    if (!HostInComm(comm, commLen, addr))
#endif
    {
		dword2mem(address, IpH2NDWord(addr));
        DnpapMessage(DMC_ERROR, AGENT_ERROR, "agent: authentication failure community \"%.*s\", host %d.%d.%d.%d", 
					 commLen, comm, address[0], address[1], address[2], address[3]);
		if (trapAuthenTraps == TRUE && trapComm != NULL &&
			AgentSendTrap(trapComm, trapCommLen, SysTime(), 
							SNMP_TRAP_AUTFAILURE, 0, NULL, 0) == FALSE)
	        DnpapMessage(DMC_ERROR, AGENT_ERROR, 
							"agent: authentication-failure trap send failed");
        return FALSE;
    }
        
	switch (pdu.Type)
	{
	case SNMP_PDU_NEXT:
	case SNMP_PDU_GET:
		if ((perm & MIB_READ) == 0)
		{
			agentStat.InBadCommunityUses++;
        	return FALSE;
		}
		break;
	case SNMP_PDU_SET:
		if ((perm & MIB_WRITE) == 0)
		{
			agentStat.InBadCommunityUses++;
        	return FALSE;
		}
		break;
	}

	status = MibRequest(listLen, list, &index);

	if (status != SNMP_NOERROR)
	{
		rqs->Type = SNMP_PDU_RESPONSE;
		rqs->ErrorStatus = status;
		rqs->ErrorIndex = index;

		if (!SnmpErrEnc(rqs))
        {
            DnpapMessage(DMC_ERROR, AGENT_ERROR, "agent: SNMP error encode failed, error %d", snmpErrStatus);
            return FALSE;  
        }

   		memcpy(*rspBuffer, rqsBuffer, rqsLen);
		*rspLen = rqsLen;
		return TRUE;
	}

	switch (pdu.Type)
	{
		case SNMP_PDU_GET:
		case SNMP_PDU_NEXT:
			agentStat.InTotalReqVars += listLen;
			break;
		case SNMP_PDU_SET:
			agentStat.InTotalSetVars += listLen;
			break;
	}

    rqs->Type = SNMP_PDU_RESPONSE;
    rqs->ErrorStatus = SNMP_NOERROR;
    rqs->ErrorIndex  = 0;
    
    if (!SnmpEnc(rspBuffer,
                rspLen,
                &pdu,
                comm,
                commLen,
                list,
                listLen))
    {
		rqs->Type = SNMP_PDU_RESPONSE;
		rqs->ErrorStatus = snmpErrStatus;
		rqs->ErrorIndex = snmpErrIndex;

		if (!SnmpErrEnc(rqs))
        { 
            DnpapMessage(DMC_ERROR, AGENT_ERROR, "agent: SNMP error encode failed");
            return FALSE;  
        }
 
   		memcpy(*rspBuffer, rqsBuffer, rqsLen);
		*rspLen = rqsLen;
		return TRUE;
    }
    
    return TRUE;
}


AGENT_COMM *AgentCommunity(int index)
{
    AGENT_COMM *p;

    for (p=agentCommList; p!=0; p=p->next)
    {
        if (p->index == index)
            break;
    }
    return p;
}



static AGENT_COMM *CommFind(BYTE *comm, WORD commLen)
{
    AGENT_COMM *p;

    for (p=agentCommList; p!=0; p=p->next)
    {
        if (p->commLen == commLen &&
            memcmp(p->comm, comm, commLen)==0)
            break;
    }
    return p;
}


static BOOLEAN HostInComm(BYTE *comm, WORD commLen, DWORD addr)
{
    AGENT_COMM  *p; 
    AGENT_HOST  *h;
    
    
    p=CommFind(comm, commLen);
    if (p==0)
        return FALSE;
     
    if (p->hostList==0) 
        return TRUE;             
        
    for (h=p->hostList;h!=0;h=h->next)
    {
        if (h->addr==addr)
            break;
    }
    
    return (h!=0);
}    
