/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <stdlib.h>
#include <limits.h>
#include <memory.h>
#include <dnpap.h>
#include <timer.h>
#include <config.h>
#include <message.h>
#include <mac.h>
#include <hash.h>
#include <dmath.h>
#include <dtime.h>
#include <sysmib.h>
#include <protocol.h>
#include <channel.h>

#include "discd.h"
#include "disce.h"
#include "discconc.h"


IMPORT BOOLEAN BeholderStrictRMON;


static LONG DiscConnMaxNrConns = 20000;

static LONG DiscConnTimeout = 10000;


static BOOLEAN DiscConnCallback(MAC_COLL *collector, PROT_PKT *pkt);
static VOID DiscConnHandlePkt(DISCCONN_CONTROL *discconncontrol, int type, PROT_PKT *pkt);

static BOOLEAN DiscConnUpdateTables(DISCCONN_CONTROL *discconncontrol, INT type, BYTE *src, INT srclen, BYTE *dst, INT dstlen, LWORD size, ULONG now);
static VOID DiscConnRemoveEntry(DISCCONN_CONTROL *discconncontrol, DISCCONN *oldsrcdst);

static BOOLEAN DiscConnAddConnList(INT type, DISCCONN **sdlexlist, DISCCONN **dslexlist, DISCCONN *host);
static BOOLEAN DiscConnRemoveConnList(DISCCONN **sdlexlist, DISCCONN **dslexlist, DISCCONN *host);

static DISCCONN* DiscConnAddLRUList(DISCCONN_CONTROL *discconncontrol, DISCCONN *srcdst);
static BOOLEAN DiscConnUpdateLRUList(DISCCONN_CONTROL *discconncontrol, DISCCONN *srcdst);
static BOOLEAN DiscConnRemoveLRUList(DISCCONN_CONTROL *discconncontrol, DISCCONN *srcdst);

static VOID LRUTimerCallback(TIMER_DESCR *timer, ULONG now, VOID *param);

static VOID DelConns(DISCCONN_CONTROL *discconncontrol, DISCCONN **sdlexlist, DISCCONN **dslexlist);

static INT ConnLen(INT type);

static int connrcmp(BYTE *srcdst1, BYTE *srcdst2, WORD addrlen, WORD len);



BOOLEAN DiscConnConfigInit(VOID)
{
	ConfigGetLong("beholder.discconn.maxnrconns", &DiscConnMaxNrConns);

	if (DiscConnMaxNrConns < 2)
    {
        DnpapMessage(DMC_WARNING, DISC_MAX, "discconncontrol: beholder.discconn.maxnrconns < 2, setting it to 2");
        DiscConnMaxNrConns = 2;
    }

	ConfigGetLong("beholder.discconn.timeout", &DiscConnTimeout);
    
	if (DiscConnTimeout < 2)
    {
        DnpapMessage(DMC_WARNING, DISC_MAX, "discconncontrol: beholder.discconn.timeout < 2, setting it to 2");
        DiscConnTimeout = 2;
    }

	return TRUE;
}


BOOLEAN DiscConnCInit(DISCCONN_CONTROL *discconncontrol)
{
LONG source[] = {1,3,6,1,2,1,2,2,1,1,1};

    memcpy(discconncontrol->Source, source, sizeof(source));
    discconncontrol->SourceLen = sizeof(source)/sizeof(source[0]);
    discconncontrol->Channel = FALSE;
    discconncontrol->ChannelCallback = NULL;
    discconncontrol->AddressLen = 0;
    discconncontrol->EthTableSize = 0;
    discconncontrol->IPTableSize = 0;
    discconncontrol->UDPTableSize = 0;
    discconncontrol->TCPTableSize = 0;
    discconncontrol->TableSize = 0;
    discconncontrol->TurnOnTime = 0;
    discconncontrol->LastDeleteTime = 0;
    discconncontrol->Owner[0] = '\0';
    discconncontrol->OwnerLen = 0;
    discconncontrol->Status = SNMP_INVALID;

    if ((discconncontrol->Iface =
        MacIfaceGet((WORD) discconncontrol->Source[discconncontrol->SourceLen-1])) == NULL)
    {
        DnpapMessage(DMC_ERROR, DISC_NETINIT, "discconncontrol: network initialisation failed");
        return (FALSE);
    }
	
    discconncontrol->Table = NULL;

    discconncontrol->EthSrcDstList = NULL;
    discconncontrol->IPSrcDstList = NULL;
    discconncontrol->UDPSrcDstList = NULL;
    discconncontrol->TCPSrcDstList = NULL;
    discconncontrol->EthDstSrcList = NULL;
    discconncontrol->IPDstSrcList = NULL;
    discconncontrol->UDPDstSrcList = NULL;
    discconncontrol->TCPDstSrcList = NULL;

    discconncontrol->LRUList = NULL;
    discconncontrol->LRULast = NULL;

    discconncontrol->LRUTimer = NULL;
    discconncontrol->LastLRUEntry = NULL;
    
	BooleanSetAllTrue(discconncontrol->ObjectSet);
	BooleanSetFalse(discconncontrol->ObjectSet, DISC_BOOLEAN_DATASOURCE);
	BooleanSetFalse(discconncontrol->ObjectSet, DISC_BOOLEAN_OWNER);

	return TRUE;
}


BOOLEAN DiscConnCStart(DISCCONN_CONTROL *discconncontrol)
{
	if (BeholderStrictRMON && !BooleanCheckAllTrue(discconncontrol->ObjectSet))
		return FALSE;

    if ((discconncontrol->Table = NewHash(DiscConnMaxNrConns/2+11, NULL)) == NULL)
    {
        DnpapMessage(DMC_ERROR, DISC_HASHERR, "discconncontrol: can not create hashtable");
        return FALSE;
    }

    if ((discconncontrol->LRUTimer = TimerRegister(LRUTimerCallback, discconncontrol, DiscConnTimeout, TIMER_FOREVER, TIMER_TYPE_RECOVER)) == NULL)
    {
        DnpapMessage(DMC_ERROR, DISC_TIMER, "discconncontrol: can not set LRU timer");
        return FALSE;
    }

	if (discconncontrol->Channel == TRUE)
    {
        discconncontrol->ChannelCallback = ChannelRegisterSink(discconncontrol->Source[discconncontrol->SourceLen - 1],
                                                               DiscConnChannelPkt, DiscConnDeRegister, discconncontrol);
    }
    else
	{
        discconncontrol->Coll.Rcve       = DiscConnCallback;
        discconncontrol->Coll.specific   = discconncontrol;
        if (!MacCollRegister(&(discconncontrol->Coll)))
        {
            DnpapMessage(DMC_ERROR, DISC_NETERR, "discconncontrol: network initialisation failed");
            TimerRemove(discconncontrol->LRUTimer);
            discconncontrol->LRUTimer = NULL;
            return FALSE;
        }
    }

    discconncontrol->AddressLen = 0;
	discconncontrol->TurnOnTime = TimeUsec();
	
    return TRUE;
}


BOOLEAN DiscConnCStop(DISCCONN_CONTROL *discconncontrol)
{
    TimerRemove(discconncontrol->LRUTimer);
	discconncontrol->LRUTimer = NULL;
        
    if (discconncontrol->Channel == TRUE)
	{
        ChannelDeRegisterSink(discconncontrol->Source[discconncontrol->SourceLen - 1],
                              discconncontrol->ChannelCallback);
		discconncontrol->ChannelCallback = NULL;
	}
    else
        MacCollRemove(&(discconncontrol->Coll));

    DelHash(discconncontrol->Table);
	discconncontrol->Table = NULL;

    DelConns(discconncontrol, &discconncontrol->EthSrcDstList, &discconncontrol->EthDstSrcList);
    DelConns(discconncontrol, &discconncontrol->IPSrcDstList, &discconncontrol->IPDstSrcList);
    DelConns(discconncontrol, &discconncontrol->UDPSrcDstList, &discconncontrol->UDPDstSrcList);
    DelConns(discconncontrol, &discconncontrol->TCPSrcDstList, &discconncontrol->TCPDstSrcList);

    discconncontrol->EthSrcDstList = NULL;
    discconncontrol->IPSrcDstList = NULL;
    discconncontrol->UDPSrcDstList = NULL;
    discconncontrol->TCPSrcDstList = NULL;
    discconncontrol->EthDstSrcList = NULL;
    discconncontrol->IPDstSrcList = NULL;
    discconncontrol->UDPDstSrcList = NULL;
    discconncontrol->TCPDstSrcList = NULL;

    discconncontrol->LRUList = NULL;
    discconncontrol->LRULast = NULL;

    discconncontrol->LRUTimer = NULL;
    discconncontrol->LastLRUEntry = NULL;
    
	discconncontrol->TableSize = 0;
    discconncontrol->EthTableSize = 0;
    discconncontrol->IPTableSize = 0;
    discconncontrol->UDPTableSize = 0;
    discconncontrol->TCPTableSize = 0;
	
    discconncontrol->TurnOnTime = 0;
    discconncontrol->LastDeleteTime = 0;
	
    return TRUE;
}


BOOLEAN DiscConnCallback(MAC_COLL *collector, PROT_PKT *pkt)
{
DISCCONN_CONTROL *discconncontrol = collector->specific;
PROT_OBJ Interface = {0, {1,2}};

    if (discconncontrol->Status == SNMP_VALID &&
        ProtGetField(pkt,&Interface) == TRUE &&
        Interface.Syntax.LngInt == discconncontrol->Source[discconncontrol->SourceLen-1])
	    DiscConnHandlePkt(discconncontrol, discconncontrol->Iface->type, pkt);
		
	return TRUE;
}


VOID DiscConnChannelPkt(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *sink)
{
	DISCCONN_CONTROL *discconncontrol = (DISCCONN_CONTROL *) sink;

	if (discconncontrol->Status == SNMP_VALID && discconncontrol->Channel == TRUE &&
		discconncontrol->Source[discconncontrol->SourceLen - 1] == channelIndex)
		DiscConnHandlePkt(discconncontrol, type, pkt);
}


VOID DiscConnHandlePkt(DISCCONN_CONTROL *discconncontrol, int type, PROT_PKT *pkt)
{
PROT_OBJ EthDst = {1, {2,1}};
PROT_OBJ EthSrc = {1, {2,2}};
PROT_OBJ EthSize = {0, {1,4}};
PROT_OBJ EthTime = {0, {1,6}};
PROT_OBJ IPSize = {2, {3,4}};
PROT_OBJ IPOffset = {2, {3,7}};
PROT_OBJ IPSrc = {2, {3,11}};
PROT_OBJ IPDst = {2, {3,12}};
PROT_OBJ UDPSrc = {3, {5,1}};
PROT_OBJ UDPDst = {3, {5,2}};
PROT_OBJ UDPSize = {3, {5,3}};
PROT_OBJ TCPSrc = {3, {4,1}};
PROT_OBJ TCPDst = {3, {4,2}};
PROT_OBJ TCPSize = {3, {4,5}};
ULONG now;
BYTE src[DISC_SIZE_NODE], dst[DISC_SIZE_NODE];
LWORD size;            

	if (type != MAC_TYPE_ETHERNET_CSMACD && type != MAC_TYPE_88023_CSMACD)
		return;

    if (ProtGetField(pkt,&EthSize) == TRUE &&
        ProtGetField(pkt,&EthTime) == TRUE &&
        ProtGetField(pkt,&EthSrc) == TRUE &&
        ProtGetField(pkt,&EthDst) == TRUE)
	{	
        size = EthSize.Syntax.LngUns + 4L;
    	now = EthTime.Syntax.LngUns - discconncontrol->TurnOnTime;
            
        if (EthSrc.SyntaxLen != DISC_SIZE_ETHADDR || EthDst.SyntaxLen != DISC_SIZE_ETHADDR)
        {
            DnpapMessage(DMC_ERROR, DISC_INVPACKET, "discconncontrol: invalid packet");
            return;
        }
        memcpy(src, EthSrc.Syntax.BufChr, DISC_SIZE_ETHADDR);
        memcpy(dst, EthDst.Syntax.BufChr, DISC_SIZE_ETHADDR);
        DiscConnUpdateTables(discconncontrol, DISC_TYPE_ETH, src, DISC_SIZE_ETHADDR, dst, DISC_SIZE_ETHADDR, size, now);

        if (ProtGetField(pkt,&IPSize) == TRUE &&
            ProtGetField(pkt,&IPSrc) == TRUE &&
            ProtGetField(pkt,&IPDst) == TRUE)
        {
            size = IPSize.Syntax.LngUns;
            NETLVALUEPTR2BYTES(src+DISC_SIZE_ETHADDR, IPSrc.Syntax.BufChr);
            NETLVALUEPTR2BYTES(dst+DISC_SIZE_ETHADDR, IPDst.Syntax.BufChr);
            DiscConnUpdateTables(discconncontrol, DISC_TYPE_IP, src, DISC_SIZE_TOIPADDR, dst, DISC_SIZE_TOIPADDR, size, now);

            if (ProtGetField(pkt,&IPOffset) == TRUE &&
                IPOffset.Syntax.LngUns == 0)
            {
                if (ProtGetField(pkt,&UDPSize) == TRUE &&
                    ProtGetField(pkt,&UDPSrc) == TRUE &&
                    ProtGetField(pkt,&UDPDst) == TRUE)
                {
                    size = UDPSize.Syntax.LngUns;
                    VALUE2BYTES(src+DISC_SIZE_TOIPADDR, UDPSrc.Syntax.LngInt);
                    VALUE2BYTES(dst+DISC_SIZE_TOIPADDR, UDPDst.Syntax.LngInt);
                    DiscConnUpdateTables(discconncontrol, DISC_TYPE_UDP, src, DISC_SIZE_TOUDPADDR, dst, DISC_SIZE_TOUDPADDR, size, now);
                }
                else
                if (ProtGetField(pkt,&TCPSize) == TRUE &&
                    ProtGetField(pkt,&TCPSrc) == TRUE &&
                    ProtGetField(pkt,&TCPDst) == TRUE)
                {
                    size = TCPSize.Syntax.LngUns;
                    VALUE2BYTES(src+DISC_SIZE_TOIPADDR, TCPSrc.Syntax.LngInt);
                    VALUE2BYTES(dst+DISC_SIZE_TOIPADDR, TCPDst.Syntax.LngInt);
                    DiscConnUpdateTables(discconncontrol, DISC_TYPE_TCP, src, DISC_SIZE_TOTCPADDR, dst, DISC_SIZE_TOTCPADDR, size, now);
                }
            }
        }
	}

    return;
}


VOID DiscConnDeRegister(ULONG channelIndex, VOID *sink)
{
DISCCONN_CONTROL *data = (DISCCONN_CONTROL *) sink;

    if (data->Channel == TRUE && data->Source[data->SourceLen - 1] == channelIndex)
    {
        ChannelDeRegisterSink(data->Source[data->SourceLen - 1],
                              data->ChannelCallback);
		data->ChannelCallback = NULL;
    }
}


BOOLEAN DiscConnUpdateTables(DISCCONN_CONTROL *discconncontrol, INT type, BYTE *src, INT srclen, BYTE *dst, INT dstlen, LWORD size, ULONG now)
{
BYTE srcdstaddr[DISC_SIZE_CONN];
DISCCONN *srcdst, *oldsrcdst;
                                  
    memcpy(srcdstaddr, src, srclen);
    memcpy(srcdstaddr+srclen, dst, dstlen);
    if ((srcdst = HashSearch(discconncontrol->Table, srcdstaddr, srclen+dstlen)) == NULL)
    {
        /*  first try to add the new srcdst  */
        if ((srcdst = DnpapMalloc(sizeof(DISCCONN))) != NULL)
        {
            memset(srcdst, 0, sizeof(DISCCONN));
            memcpy(srcdst->SrcDst, srcdstaddr, (srcdst->SrcDstLen = srclen+dstlen));
            srcdst->Type = (BYTE)type;
            if (HashAdd(discconncontrol->Table, srcdst->SrcDst, srclen+dstlen, srcdst) == NULL)
            {
                DnpapMessage(DMC_WARNING, DISC_NADD, "discconncontrol: connection could not be added to the hash table");
                DnpapFree(srcdst);
                srcdst = NULL;
            }
            else
            {
                switch (type)
                {
                case DISC_TYPE_ETH:
                    DiscConnAddConnList(type, &discconncontrol->EthSrcDstList, &discconncontrol->EthDstSrcList, srcdst);
                    discconncontrol->EthTableSize++;
                    break;
                case DISC_TYPE_IP:
                    DiscConnAddConnList(type, &discconncontrol->IPSrcDstList, &discconncontrol->IPDstSrcList, srcdst);
                    discconncontrol->IPTableSize++;
                    break;
                case DISC_TYPE_UDP:
                    DiscConnAddConnList(type, &discconncontrol->UDPSrcDstList, &discconncontrol->UDPDstSrcList, srcdst);
                    discconncontrol->UDPTableSize++;
                    break;
                case DISC_TYPE_TCP:
                    DiscConnAddConnList(type, &discconncontrol->TCPSrcDstList, &discconncontrol->TCPDstSrcList, srcdst);
                    discconncontrol->TCPTableSize++;
                    break;
                }

                oldsrcdst = DiscConnAddLRUList(discconncontrol, srcdst);
                discconncontrol->TableSize++;

                if (discconncontrol->TableSize > (LONG)DiscConnMaxNrConns)
                    DiscConnRemoveEntry(discconncontrol, oldsrcdst);
            }
        }
    }

    if (srcdst != NULL)
    {
        srcdst->Pkts++;
        srcdst->Octets += size;
 
        srcdst->LRUTime = now;

        DiscConnUpdateLRUList(discconncontrol, srcdst);
    }

    return TRUE;
}

                    
VOID DiscConnRemoveEntry(DISCCONN_CONTROL *discconncontrol, DISCCONN *oldsrcdst)
{
    DiscConnRemoveLRUList(discconncontrol, oldsrcdst);
    
    switch (oldsrcdst->Type)
    {
    case DISC_TYPE_ETH:
        DiscConnRemoveConnList(&discconncontrol->EthSrcDstList, &discconncontrol->EthDstSrcList, oldsrcdst);
        discconncontrol->EthTableSize--;
        break;
    case DISC_TYPE_IP:
        DiscConnRemoveConnList(&discconncontrol->IPSrcDstList, &discconncontrol->IPDstSrcList, oldsrcdst);
        discconncontrol->IPTableSize--;
        break;
    case DISC_TYPE_UDP:
        DiscConnRemoveConnList(&discconncontrol->UDPSrcDstList, &discconncontrol->UDPDstSrcList, oldsrcdst);
        discconncontrol->UDPTableSize--;
        break;
    case DISC_TYPE_TCP:
        DiscConnRemoveConnList(&discconncontrol->TCPSrcDstList, &discconncontrol->TCPDstSrcList, oldsrcdst);
        discconncontrol->TCPTableSize--;
        break;
    }
    
    HashRemove(discconncontrol->Table, oldsrcdst->SrcDst, oldsrcdst->SrcDstLen);
    DnpapFree(oldsrcdst);
    discconncontrol->LastDeleteTime = SysTime();
    discconncontrol->TableSize--;
}


BOOLEAN DiscConnAddConnList(INT type, DISCCONN **sdlexlist, DISCCONN **dslexlist, DISCCONN *srcdst)
{
DISCCONN *p, *q;
WORD addrlen;

    addrlen = ConnLen(type);

    if (*sdlexlist == NULL && *dslexlist == NULL)
    {
        *sdlexlist = srcdst;
        srcdst->SrcDstPrev = NULL;
        srcdst->SrcDstNext = NULL;
        
        *dslexlist = srcdst;
        srcdst->DstSrcPrev = NULL;
        srcdst->DstSrcNext = NULL;
        
        return TRUE;
    }
            
    /*  update SrcDstList  */
    for (p = *sdlexlist, q = NULL; p != NULL; q = p, p = p->SrcDstNext)
    {
        if (memcmp(p->SrcDst, srcdst->SrcDst, 2*addrlen) >= 0)
            break;
    }

    if (q != NULL)
        q->SrcDstNext = srcdst;
    else
        *sdlexlist = srcdst;

    if (p != NULL)
        p->SrcDstPrev = srcdst;

    srcdst->SrcDstPrev = q;
    srcdst->SrcDstNext = p;


    /*  update DstSrcList  */
    for (p = *dslexlist, q = NULL; p != NULL; q = p, p = p->DstSrcNext)
    {
        if (connrcmp(p->SrcDst, srcdst->SrcDst, addrlen, 2*addrlen) >= 0)
            break;
    }

    if (q != NULL)
        q->DstSrcNext = srcdst;
    else
        *dslexlist = srcdst;

    if (p != NULL)
        p->DstSrcPrev = srcdst;

    srcdst->DstSrcPrev = q;
    srcdst->DstSrcNext = p;
        

    return TRUE;
}


DISCCONN* DiscConnAddLRUList(DISCCONN_CONTROL *discconncontrol, DISCCONN *srcdst)
{
    if (discconncontrol->LRUList == NULL)
    {
        discconncontrol->LRUList = srcdst;
        discconncontrol->LRULast = srcdst;
        srcdst->LRUPrev = NULL;
        srcdst->LRUNext = NULL;
        
        return NULL;
    }

    srcdst->LRUNext = discconncontrol->LRUList;
    srcdst->LRUPrev = NULL;
    discconncontrol->LRUList->LRUPrev = srcdst;
    discconncontrol->LRUList = srcdst;

    return discconncontrol->LRULast;
}


BOOLEAN DiscConnUpdateLRUList(DISCCONN_CONTROL *discconncontrol, DISCCONN *srcdst)
{
    if (discconncontrol->TableSize > 1 && discconncontrol->LRUList != srcdst)
    {          
        DiscConnRemoveLRUList(discconncontrol, srcdst);
        DiscConnAddLRUList(discconncontrol, srcdst);
    }
    
    return TRUE;
}


BOOLEAN DiscConnRemoveConnList(DISCCONN **sdlexlist, DISCCONN **dslexlist, DISCCONN *srcdst)
{
    if (srcdst->SrcDstPrev != NULL)
        srcdst->SrcDstPrev->SrcDstNext = srcdst->SrcDstNext;
    else
        *sdlexlist = srcdst->SrcDstNext;
    if (srcdst->SrcDstNext != NULL)
        srcdst->SrcDstNext->SrcDstPrev = srcdst->SrcDstPrev;

    if (srcdst->DstSrcPrev != NULL)
        srcdst->DstSrcPrev->DstSrcNext = srcdst->DstSrcNext;
    else
        *dslexlist = srcdst->DstSrcNext;
    if (srcdst->DstSrcNext != NULL)
        srcdst->DstSrcNext->DstSrcPrev = srcdst->DstSrcPrev;

    return TRUE;
}


BOOLEAN DiscConnRemoveLRUList(DISCCONN_CONTROL *discconncontrol, DISCCONN *srcdst)
{
    if (srcdst->LRUPrev != NULL)
        srcdst->LRUPrev->LRUNext = srcdst->LRUNext;
    else
        discconncontrol->LRUList = srcdst->LRUNext;
    if (srcdst->LRUNext != NULL)
        srcdst->LRUNext->LRUPrev = srcdst->LRUPrev;
    else                              
        discconncontrol->LRULast = srcdst->LRUPrev;

    srcdst->LRUPrev = NULL;
    srcdst->LRUNext = NULL;
    
    return TRUE;
}


VOID LRUTimerCallback(TIMER_DESCR *timer, ULONG now, VOID *param)
{
DISCCONN_CONTROL *discconncontrol = (DISCCONN_CONTROL *) param;
                                            
    if (discconncontrol->LRULast != NULL && discconncontrol->LRULast == discconncontrol->LastLRUEntry)
        DiscConnRemoveEntry(discconncontrol, discconncontrol->LRULast);
    discconncontrol->LastLRUEntry = discconncontrol->LRULast;
}


VOID DelConns(DISCCONN_CONTROL *discconncontrol, DISCCONN **sdlexlist, DISCCONN **dslexlist)
{
DISCCONN *srcdst1, *srcdst2;

    for (srcdst2 = *sdlexlist; srcdst2 != NULL; srcdst2 = srcdst1)
    {
        srcdst1 = srcdst2->SrcDstNext;
        DnpapFree(srcdst2);
        discconncontrol->TableSize--;
    }
    *sdlexlist = NULL;
    *dslexlist = NULL;
}


DISCCONN* DiscConnSearchNext(DISCCONN_CONTROL *discconncontrol, INT type, BYTE *srcdst, WORD len, BOOLEAN SDorder)
{
DISCCONN* p, *lexlist = NULL;
BYTE srcdst2[DISC_SIZE_CONN];
WORD addrlen;

	if (SDorder == TRUE)
	{
        switch (type)
        {
        case DISC_TYPE_ETH:
	        if (len == 2*DISC_SIZE_ETHADDR &&
	            (p = HashSearch(discconncontrol->Table, srcdst, 2*DISC_SIZE_ETHADDR)) != NULL)
	            return p->SrcDstNext;
            lexlist = discconncontrol->EthSrcDstList;
            break;
        case DISC_TYPE_IP:
	        if (len == 2*DISC_SIZE_TOIPADDR &&
	            (p = HashSearch(discconncontrol->Table, srcdst, 2*DISC_SIZE_TOIPADDR)) != NULL)
	            return p->SrcDstNext;
            lexlist = discconncontrol->IPSrcDstList;
            break;
        case DISC_TYPE_UDP:
	        if (len == 2*DISC_SIZE_TOUDPADDR &&
	            (p = HashSearch(discconncontrol->Table, srcdst, 2*DISC_SIZE_TOUDPADDR)) != NULL)
	            return p->SrcDstNext;
            lexlist = discconncontrol->UDPSrcDstList;
            break;
        case DISC_TYPE_TCP:
	        if (len == 2*DISC_SIZE_TOTCPADDR &&
	            (p = HashSearch(discconncontrol->Table, srcdst, 2*DISC_SIZE_TOTCPADDR)) != NULL)
	            return p->SrcDstNext;
            lexlist = discconncontrol->TCPSrcDstList;
            break;
        }

        for (p = lexlist; p != NULL; p = p->SrcDstNext)
	    {
	        if (memcmp(p->SrcDst, srcdst, len) >= 0)
	            break;
	    }
	    return p;
    }
    else
    {   
        addrlen = ConnLen(type);
        
        switch (type)
        {
        case DISC_TYPE_ETH:
            lexlist = discconncontrol->EthDstSrcList;
            break;
        case DISC_TYPE_IP:
            lexlist = discconncontrol->IPDstSrcList;
            break;
        case DISC_TYPE_UDP:
            lexlist = discconncontrol->UDPDstSrcList;
            break;
        case DISC_TYPE_TCP:
            lexlist = discconncontrol->TCPDstSrcList;
            break;
        }

        /*  srcdst contains destination and source in that order, so switch them  */
	    memcpy(srcdst2, srcdst+addrlen, addrlen);
	    memcpy(srcdst2+addrlen, srcdst, addrlen);
	    if (len == 2U*addrlen)
		{
	        if ((p = HashSearch(discconncontrol->Table, srcdst2, 2*addrlen)) != NULL)
	            return p->DstSrcNext;
		}

	    for (p = lexlist; p != NULL; p = p->DstSrcNext)
	    {
	        if (connrcmp(p->SrcDst, srcdst2, addrlen, len) >= 0)
	            break;
	    }
        return p;
	}
}


DISCCONN* DiscConnSearchHash(DISCCONN_CONTROL *discconncontrol, INT type, SNMP_OBJECT *obj, WORD idlen, BOOLEAN SDorder)
{
WORD addrlen;
BYTE srcdst[DISC_SIZE_CONN], srcdst2[DISC_SIZE_CONN];
                 
    addrlen = ConnLen(type);

	if (SDorder == TRUE)
	{
        Id2Conn(type, obj->Id, idlen+1, obj->IdLen, srcdst);
	    return HashSearch(discconncontrol->Table, srcdst, 2*addrlen);
    }
    else
    {
        Id2Conn(type, obj->Id, idlen+1, obj->IdLen, srcdst);
        /*  srcdst contains destination and source in that order, so switch them  */
	    memcpy(srcdst2, srcdst+addrlen, addrlen);
	    memcpy(srcdst2+addrlen, srcdst, addrlen);
	    return HashSearch(discconncontrol->Table, srcdst2, 2*addrlen);
    }
}


INT Conn2Id(INT type, BYTE *conn, INT len, LONG *id, INT match)
{
INT i, addrstart, addrend, coffset, ioffset;
             
    coffset = ConnLen(type);
    if (type == DISC_TYPE_UDP || type == DISC_TYPE_TCP)
        ioffset = coffset-1;
    else
        ioffset = coffset;
    
    switch (type)
    {
    case DISC_TYPE_UDP:
    case DISC_TYPE_TCP:
        addrstart = ConnLen(DISC_TYPE_IP);
        addrend = ConnLen(DISC_TYPE_UDP);
        if (addrstart < len)
            id[match+addrstart] = 256*(LONG)conn[addrstart] + conn[addrstart+1];
        if (coffset+addrstart < len)
            id[match+ioffset+addrstart] = 256*(LONG)conn[coffset+addrstart] + conn[coffset+addrstart+1];
    case DISC_TYPE_IP:
        addrstart = ConnLen(DISC_TYPE_ETH);
        addrend = ConnLen(DISC_TYPE_IP);
        for (i = addrstart; i < IntMin(len, addrend); i++)
            id[match+i] = conn[i];
        for (i = addrstart; i < IntMin(len-coffset, addrend); i++)
            id[ioffset+match+i] = conn[coffset+i];
    case DISC_TYPE_ETH:
        addrstart = 0;
        addrend = ConnLen(DISC_TYPE_ETH);
        for (i = addrstart; i < IntMin(len, addrend); i++)
            id[match+i] = conn[i];
        for (i = addrstart; i < IntMin(len-coffset, addrend); i++)
            id[ioffset+match+i] = conn[coffset+i];
        break;
    default:
        i = 0;
        break;
    }
    
    if (i > 0)
    {
        if (type == DISC_TYPE_UDP || type == DISC_TYPE_TCP)
            if (len < coffset)
                return match+len-1;
            else
                return match+len-2;
        else
            return match+len;
    }
    
    return i;
}


INT Id2Conn(INT type, LONG *id, INT match, INT len, BYTE *conn)
{
INT i, addrstart, addrend, sublen, coffset, ioffset, connlen;
         
    if ((sublen = len-match) < 0)
        return 0;
	
	connlen = ConnLen(type);
	if (sublen > 2*connlen)
		sublen = 2*connlen;
		
    coffset = connlen;
    if (type == DISC_TYPE_UDP || type == DISC_TYPE_TCP)
        ioffset = coffset-1;
    else
        ioffset = coffset;
    
    switch (type)
    {
    case DISC_TYPE_UDP:
    case DISC_TYPE_TCP:
        addrstart = ConnLen(DISC_TYPE_IP);
        addrend = ConnLen(DISC_TYPE_UDP);
        if (ioffset+addrstart < sublen)
        {
            conn[coffset+addrstart] = (BYTE)(id[ioffset+match+addrstart] >> 8);
            conn[coffset+addrstart+1] = (BYTE)(id[ioffset+match+addrstart] & 0x00ff);
        }
        if (addrstart < sublen)
        {
            conn[addrstart] = (BYTE)(id[match+addrstart] >> 8);
            conn[addrstart+1] = (BYTE)(id[match+addrstart] & 0x00ff);
        }
    case DISC_TYPE_IP:
        addrstart = ConnLen(DISC_TYPE_ETH);
        addrend = ConnLen(DISC_TYPE_IP);
        for (i = addrstart; i < IntMin(sublen-ioffset, addrend); i++)
            conn[coffset+i] = (BYTE)id[ioffset+match+i];
        for (i = addrstart; i < IntMin(sublen, addrend); i++)
            conn[i] = (BYTE)id[match+i];
    case DISC_TYPE_ETH:
        addrstart = 0;
        addrend = ConnLen(DISC_TYPE_ETH);
        for (i = addrstart; i < IntMin(sublen-ioffset, addrend); i++)
            conn[coffset+i] = (BYTE)id[ioffset+match+i];
        for (i = addrstart; i < IntMin(sublen, addrend); i++)
            conn[i] = (BYTE)id[match+i];
        break;
    default:
        i = 0;
        break;
    }
    
    if (i > 0)
    {
        if (type == DISC_TYPE_UDP || type == DISC_TYPE_TCP)
            if (sublen < ioffset)
                return sublen+1;
            else
                return sublen+2;
        else
            return sublen;
    }
    
    return i;
}


INT ConnLen(INT type)
{
    switch (type)
    {
    case DISC_TYPE_ETH:
        return DISC_SIZE_ETHADDR;
    case DISC_TYPE_IP:
        return DISC_SIZE_TOIPADDR;
    case DISC_TYPE_UDP:
        return DISC_SIZE_TOUDPADDR;
    case DISC_TYPE_TCP:
        return DISC_SIZE_TOTCPADDR;
    default:
        return 0;
    }
}


int connrcmp(BYTE *srcdst1, BYTE *srcdst2, WORD addrlen, WORD len)
{
INT i1, i2;

	if (len > 2*addrlen)
		len = 2*addrlen;

	for (i1 = addrlen, i2 = addrlen; len > 0; i1++, i2++, len--)
	{
		if (i1 >= 2*addrlen)
			i1 = i2 = 0;
		
		if (srcdst1[i1] < srcdst2[i2])
			return -1;
		else
		if (srcdst1[i1] > srcdst2[i2])
			return 1;
	}

	return 0;
}
