/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#ifndef _DISCCONC_H
#define _DISCCONC_H


#include <dnpap.h>
#include <boolset.h>
#include <timer.h>
#include <snmp.h>
#include <hash.h>
#include <protocol.h>
#include <channel.h>

#include "discd.h"


typedef struct _DISCCONN
{
    BYTE        SrcDst[DISC_SIZE_CONN];
    WORD        SrcDstLen;

    BYTE        Type;

    ULONG       LRUTime;

    ULONG       Pkts;
    ULONG       Octets;

    struct _DISCCONN  *SrcDstPrev;
    struct _DISCCONN  *SrcDstNext;
    
    struct _DISCCONN  *DstSrcPrev;
    struct _DISCCONN  *DstSrcNext;

    struct _DISCCONN  *LRUPrev;
    struct _DISCCONN  *LRUNext;
    
} DISCCONN;


typedef struct _DISCCONN_CONTROL
{
    ULONG           Source[SNMP_SIZE_BUFINT];
    WORD            SourceLen;
    BOOLEAN         Channel;
    SinkCallback   *ChannelCallback;
	WORD			AddressLen;
    LONG            EthTableSize;
    LONG            IPTableSize;
    LONG            UDPTableSize;
    LONG            TCPTableSize;
    LONG            TableSize;
    ULONG           TurnOnTime;
    ULONG           LastDeleteTime;
    BYTE            Owner[SNMP_SIZE_BUFCHR];
    WORD            OwnerLen;
    LONG            Status;
    
    MAC_COLL        Coll;
    MAC_IFACE      *Iface;
    
	HashTable  	   *Table;
	
	DISCCONN	   *EthSrcDstList;
	DISCCONN	   *EthDstSrcList;
	DISCCONN	   *IPSrcDstList;
	DISCCONN	   *IPDstSrcList;
	DISCCONN	   *UDPSrcDstList;
	DISCCONN	   *UDPDstSrcList;
	DISCCONN	   *TCPSrcDstList;
	DISCCONN	   *TCPDstSrcList;

    DISCCONN       *LRUList;
    DISCCONN       *LRULast;
    
    TIMER_DESCR    *LRUTimer;
    DISCCONN       *LastLRUEntry;
    
    BooleanSet      ObjectSet;

} DISCCONN_CONTROL;



BOOLEAN DiscConnConfigInit(VOID);

BOOLEAN DiscConnCInit(DISCCONN_CONTROL *discconncontrol);
BOOLEAN DiscConnCStart(DISCCONN_CONTROL *discconncontrol);
BOOLEAN DiscConnCStop(DISCCONN_CONTROL *discconncontrol);

VOID DiscConnChannelPkt(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *sink);
VOID DiscConnDeRegister(ULONG channelIndex, VOID *sink);

DISCCONN* DiscConnSearchNext(DISCCONN_CONTROL *discconncontrol, INT type, BYTE *srcdst, WORD len, BOOLEAN SDorder);
DISCCONN* DiscConnSearchHash(DISCCONN_CONTROL *discconncontrol, INT type, SNMP_OBJECT *obj, WORD idlen, BOOLEAN SDorder);

INT Conn2Id(INT type, BYTE *conn, INT len, LONG *id, INT match);
INT Id2Conn(INT type, LONG *id, INT match, INT len, BYTE *conn);


#endif

