/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#ifndef _DISCNODC_H
#define _DISCNODC_H


#include <dnpap.h>
#include <boolset.h>
#include <timer.h>
#include <snmp.h>
#include <hash.h>
#include <protocol.h>
#include <channel.h>

#include "discd.h"


typedef struct _DISCNODE
{
    BYTE        Address[DISC_SIZE_NODE];
    WORD        AddressLen;

    BYTE        Type;

    ULONG       LRUTime;

    ULONG       InOctets;
    ULONG       OutOctets;
    ULONG       InPkts;
    ULONG       OutPkts;

    struct _DISCNODE  *LexPrev;
    struct _DISCNODE  *LexNext;

    struct _DISCNODE  *LRUPrev;
    struct _DISCNODE  *LRUNext;
    
    struct _DISCNODE  *TimePrev;
    struct _DISCNODE  *TimeNext;

    struct _DISCNODE **TimeListPtr;

} DISCNODE;


typedef struct _DISCNODE_CONTROL
{
    ULONG           Source[SNMP_SIZE_BUFINT];
    WORD            SourceLen;
    BOOLEAN         Channel;
    SinkCallback   *ChannelCallback;
	WORD			AddressLen;
    LONG            EthTableSize;
    LONG            IPTableSize;
    LONG            UDPTableSize;
    LONG            TCPTableSize;
    LONG            TableSize;
    ULONG           TurnOnTime;
    ULONG           LastDeleteTime;
    BYTE            Owner[SNMP_SIZE_BUFCHR];
    WORD            OwnerLen;
    LONG            Status;
    
    MAC_COLL        Coll;
    MAC_IFACE      *Iface;
    
    HashTable	   *Table;
    
    DISCNODE       *EthLexList;
    DISCNODE   	   *IPLexList;
    DISCNODE  	   *UDPLexList;
    DISCNODE	   *TCPLexList;
    
    DISCNODE       *LRUList;
    DISCNODE       *LRULast;
    
    DISCNODE	   *EthTimeList;
    DISCNODE	   *EthTimeLast;
    DISCNODE	  **EthTimeTable;
    BOOLEAN			EthTimeListUpdated;
    
    DISCNODE	   *IPTimeList;
    DISCNODE	   *IPTimeLast;
    DISCNODE	  **IPTimeTable;
    BOOLEAN			IPTimeListUpdated;
    
    DISCNODE	   *UDPTimeList;
    DISCNODE	   *UDPTimeLast;
    DISCNODE	  **UDPTimeTable;
    BOOLEAN			UDPTimeListUpdated;
    
    DISCNODE	   *TCPTimeList;
    DISCNODE	   *TCPTimeLast;
    DISCNODE	  **TCPTimeTable;
    BOOLEAN			TCPTimeListUpdated;
    
    TIMER_DESCR    *LRUTimer;
    DISCNODE       *LastLRUEntry;
    
    BooleanSet      ObjectSet;

} DISCNODE_CONTROL;



BOOLEAN DiscNodeConfigInit(VOID);

BOOLEAN DiscNodeCInit(DISCNODE_CONTROL *disccontrol);
BOOLEAN DiscNodeCStart(DISCNODE_CONTROL *disccontrol);
BOOLEAN DiscNodeCStop(DISCNODE_CONTROL *disccontrol);
 
VOID DiscNodeChannelPkt(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *sink);
VOID DiscNodeDeRegister(ULONG channelIndex, VOID *sink);

DISCNODE *DiscNodeSearchNext(DISCNODE_CONTROL *discnodecontrol, INT type, DISCNODE *disc, WORD len);
DISCNODE *DiscNodeSearchHash(DISCNODE_CONTROL *discnodecontrol, INT type, SNMP_OBJECT *obj, WORD idlen);
DISCNODE* DiscNodeTimeSearch(DISCNODE_CONTROL *discnodecontrol, INT type, SNMP_OBJECT *obj, WORD idlen);
LONG DiscNodeSearchCreationOrder(DISCNODE_CONTROL *discnodecontrol, INT type, DISCNODE *disc);

INT Node2Id(INT type, BYTE *node, INT len, LONG *id, INT match);
INT Id2Node(INT type, LONG *id, INT match, INT len, BYTE *node);


#endif
