/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <message.h>
#include <mibsup.h>
#include <mac.h>

#include "discd.h"
#include "disce.h"
#include "discnodc.h"
#include "discnod.h"

IMPORT BOOLEAN BeholderStrictRMON;

static MIB_LOCAL *DiscNodeControlInstance = NULL;

static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, INT type, WORD indexsize, DISCNODE ** elem);
static DISCNODE *DiscNodeGetFirstNext(DISCNODE_CONTROL * discnodecontrol, INT type, WORD indexsize, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first);

static LONG TableSize(DISCNODE_CONTROL * discnodecontrol, INT type);

static INT DiscTableType(SNMP_OBJECT * obj, WORD idlen);
static WORD DiscTableLen(INT type, INT addrtype, WORD * offset, WORD * sublen);

BOOLEAN DiscNodeMInit(VOID)
{
	MessageConfig(DISCNODE_ERROR, "DiscNode");
	return DiscNodeConfigInit();
}

#define CONTROL_DATA    ((DISCNODE_CONTROL *) (*local)->Data)

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, INT type, WORD indexsize, DISCNODE ** elem)
{
	BOOLEAN first = FALSE;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case DISC_ETHNODEINDEXSIZE:
	case DISC_IPNODEINDEXSIZE:
	case DISC_UDPNODEINDEXSIZE:
		/* case DISC_TCPNODEINDEXSIZE:  */
		while (*local != NULL && (*elem = DiscNodeGetFirstNext(CONTROL_DATA, type, indexsize, obj, idlen, first)) == NULL)
		{
			first = TRUE;
			*local = (*local)->Next;
		}
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		break;
	case DISC_ETHTIMEINDEXSIZE:
		/* case DISC_IPTIMEINDEXSIZE:  */
		/* case DISC_UDPTIMEINDEXSIZE:  */
		/* case DISC_TCPTIMEINDEXSIZE:  */
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			if (obj->Id[idlen + 1] < TableSize(CONTROL_DATA, type))
			{
				obj->Id[idlen] = (*local)->Index;
				obj->Id[idlen + 1]++;
				obj->IdLen = idlen + DISC_ETHTIMEINDEXSIZE;
				*elem = DiscNodeTimeSearch(CONTROL_DATA, type, obj, idlen);
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL && TableSize(CONTROL_DATA, type) == 0)
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen + 1] = 1;
		obj->IdLen = idlen + DISC_ETHTIMEINDEXSIZE;
		*elem = DiscNodeTimeSearch(CONTROL_DATA, type, obj, idlen);
		break;
	case CONTROLINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + CONTROLINDEXSIZE;
		break;
	}
	return TRUE;
}

DISCNODE *DiscNodeGetFirstNext(DISCNODE_CONTROL * discnodecontrol, INT type, WORD indexsize, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first)
{
	DISCNODE disc, *p, *lexlist = NULL;
	INT len, addrlen = 0;

	if (discnodecontrol->TableSize == 0)
		return NULL;

	switch (type)
	{
	case DISC_TYPE_ETH:
		lexlist = discnodecontrol->EthLexList;
		addrlen = DISC_SIZE_ETHADDR;
		break;
	case DISC_TYPE_IP:
		lexlist = discnodecontrol->IPLexList;
		addrlen = DISC_SIZE_TOIPADDR;
		break;
	case DISC_TYPE_UDP:
		lexlist = discnodecontrol->UDPLexList;
		addrlen = DISC_SIZE_TOUDPADDR;
		break;
	case DISC_TYPE_TCP:
		lexlist = discnodecontrol->TCPLexList;
		addrlen = DISC_SIZE_TOTCPADDR;
		break;
	default:
		DnpapMessage(DMC_FATAL, DISC_TYPE, "discnodecontrol: unknown type");
		DnpapExit(DISC_TYPE);
	}

	if (first == TRUE)
	{
		p = lexlist;
		obj->IdLen = Node2Id(type, p->Address, addrlen, obj->Id, idlen + 1);
		return p;
	}

	len = Id2Node(type, obj->Id, idlen + 1, obj->IdLen, disc.Address);

	p = DiscNodeSearchNext(discnodecontrol, type, &disc, len);

	if (p != NULL)
		obj->IdLen = Node2Id(type, p->Address, addrlen, obj->Id, idlen + 1);

	return p;
}

WORD DiscNodeControlIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, 0, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeControlDataSource(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE_CONTROL *data;
	MAC_IFACE *iface;
	LONG IfaceOid[] =
	{1, 3, 6, 1, 2, 1, 2, 2, 1, 1, 1};
	WORD IfaceOidLen = sizeof(IfaceOid) / sizeof(LONG);
	LONG ChannelOid[] =
	{1, 3, 6, 1, 2, 1, 16, 7, 2, 1, 1, 1};
	WORD ChannelOidLen = sizeof(ChannelOid) / sizeof(LONG);

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, 0, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DISCNODE_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, DISC_BOOLEAN_DATASOURCE))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufInt, data->Source, data->SourceLen * sizeof(data->Source[0]));
		Obj->SyntaxLen = data->SourceLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DISCNODE_CONTROL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (IfaceOidLen == Obj->SyntaxLen &&
			memcmp(IfaceOid, Obj->Syntax.BufInt, (IfaceOidLen - 1) * sizeof(IfaceOid[0])) == 0)
		{
			data->Channel = FALSE;
			if ((iface = MacIfaceGet((WORD) Obj->Syntax.BufInt[Obj->SyntaxLen - 1])) == NULL)
				return SNMP_BADVALUE;
			if (iface->type != MAC_TYPE_ETHERNET_CSMACD && iface->type != MAC_TYPE_88023_CSMACD)
				return SNMP_BADVALUE;
			memcpy(data->Source, Obj->Syntax.BufInt, Obj->SyntaxLen * sizeof(data->Source[0]));
			data->SourceLen = Obj->SyntaxLen;
			data->Iface = iface;
			BooleanSetTrue(data->ObjectSet, DISC_BOOLEAN_DATASOURCE);
			return SNMP_NOERROR;
		}
		else if (ChannelOidLen == Obj->SyntaxLen &&
				 memcmp(ChannelOid, Obj->Syntax.BufInt, (ChannelOidLen - 1) * sizeof(ChannelOid[0])) == 0)
		{
			data->Channel = TRUE;
			memcpy(data->Source, Obj->Syntax.BufInt, Obj->SyntaxLen * sizeof(data->Source[0]));
			data->SourceLen = Obj->SyntaxLen;
			BooleanSetTrue(data->ObjectSet, DISC_BOOLEAN_DATASOURCE);
			return SNMP_NOERROR;
		}
		return SNMP_BADVALUE;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeControlEthTableSize(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE_CONTROL *data;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, 0, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DISCNODE_CONTROL *) local->Data;
		Obj->Syntax.LngInt = data->EthTableSize;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeControlIPTableSize(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE_CONTROL *data;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, 0, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DISCNODE_CONTROL *) local->Data;
		Obj->Syntax.LngInt = data->IPTableSize;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeControlUDPTableSize(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE_CONTROL *data;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, 0, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DISCNODE_CONTROL *) local->Data;
		Obj->Syntax.LngInt = data->UDPTableSize;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeControlTCPTableSize(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE_CONTROL *data;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, 0, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DISCNODE_CONTROL *) local->Data;
		Obj->Syntax.LngInt = data->TCPTableSize;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeControlTableSize(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE_CONTROL *data;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, 0, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DISCNODE_CONTROL *) local->Data;
		Obj->Syntax.LngInt = data->TableSize;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeControlLastDeleteTime(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE_CONTROL *data;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, 0, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DISCNODE_CONTROL *) local->Data;
		Obj->Syntax.LngUns = data->LastDeleteTime;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeControlOwner(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE_CONTROL *data;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, 0, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DISCNODE_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, DISC_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		Obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DISCNODE_CONTROL *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, Obj->Syntax.BufChr, Obj->SyntaxLen);
		data->OwnerLen = Obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, DISC_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeControlStatus(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE_CONTROL *data;

	local = MibRmon(Obj, DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE);

	switch (Obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (DISCNODE_CONTROL *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, 0, CONTROLINDEXSIZE, NULL) == FALSE)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (DISCNODE_CONTROL *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(Obj, &DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(DISCNODE_CONTROL))) == NULL)
				{
					MibRemove(Obj, &DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (DISCNODE_CONTROL *) local->Data;
				if (DiscNodeCInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, DISC_CREATE, "discnode: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(Obj, &DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (DISCNODE_CONTROL *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (DiscNodeCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, DISC_START, "discnode: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (DiscNodeCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, DISC_DESTROY, "discnode: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				return SNMP_NOERROR;
			case SNMP_INVALID:
				if (DiscNodeCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, DISC_DESTROY, "discnode: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &DiscNodeControlInstance, IdLen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (DiscNodeCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeEthAddress(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD i, len, offset, sublen;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, DISC_TYPE_ETH, &offset, &sublen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeSearchHash(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		for (i = 0; i < sublen; i++)
			Obj->Syntax.BufChr[i] = (BYTE) data->Address[offset + i];
		Obj->SyntaxLen = sublen;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeIPAddress(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD i, len, offset, sublen;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, DISC_TYPE_IP, &offset, &sublen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeSearchHash(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		for (i = 0; i < sublen; i++)
			Obj->Syntax.BufChr[i] = (BYTE) data->Address[offset + i];
		Obj->SyntaxLen = sublen;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeUDPAddress(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len, offset, sublen;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, DISC_TYPE_UDP, &offset, &sublen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeSearchHash(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = 256 * (LONG) data->Address[offset] + data->Address[offset + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeTCPAddress(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len, offset, sublen;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, DISC_TYPE_TCP, &offset, &sublen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeSearchHash(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = 256 * (LONG) data->Address[offset] + data->Address[offset + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeTimeEthAddress(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD i, len, offset, sublen;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, DISC_TYPE_ETH, &offset, &sublen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeTimeSearch(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		for (i = 0; i < sublen; i++)
			Obj->Syntax.BufChr[i] = (BYTE) data->Address[offset + i];
		Obj->SyntaxLen = sublen;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeTimeIPAddress(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD i, len, offset, sublen;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, DISC_TYPE_IP, &offset, &sublen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeTimeSearch(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		for (i = 0; i < sublen; i++)
			Obj->Syntax.BufChr[i] = (BYTE) data->Address[offset + i];
		Obj->SyntaxLen = sublen;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeTimeUDPAddress(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len, offset, sublen;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, DISC_TYPE_UDP, &offset, &sublen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeTimeSearch(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = 256 * (LONG) data->Address[offset] + data->Address[offset + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeTimeTCPAddress(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len, offset, sublen;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, DISC_TYPE_TCP, &offset, &sublen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeTimeSearch(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = 256 * (LONG) data->Address[offset] + data->Address[offset + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeCreationOrder(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeSearchHash(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = DiscNodeSearchCreationOrder(local->Data, type, data);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeSearchHash(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeInPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeSearchHash(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->InPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeOutPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeSearchHash(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeInOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeSearchHash(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->InOctets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeOutOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeSearchHash(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutOctets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeLRUTime(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeSearchHash(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->LRUTime;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeTimeCreationOrder(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeTimeSearch(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = Obj->Id[IdLen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeTimeIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeTimeSearch(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeTimeInPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeTimeSearch(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->InPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeTimeOutPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeTimeSearch(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeTimeInOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeTimeSearch(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->InOctets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeTimeOutOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeTimeSearch(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutOctets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DiscNodeTimeLRUTime(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DISCNODE *data = NULL;
	INT type;
	WORD len;

	if ((type = DiscTableType(Obj, IdLen)) == 0)
		return SNMP_NOSUCHNAME;

	if ((len = DiscTableLen(type, type, NULL, NULL)) == 0)
		return SNMP_NOSUCHNAME;

	if ((local = MibRmon(Obj, DiscNodeControlInstance, IdLen, len)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, type, len, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DiscNodeTimeSearch(local->Data, type, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->LRUTime;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

BOOLEAN DiscNodeRegisterAll(ULONG channelIndex)
{
	MIB_LOCAL *local;
	DISCNODE_CONTROL *discnodecontrol;

	for (local = DiscNodeControlInstance; local != NULL; local = local->Next)
	{
		discnodecontrol = (DISCNODE_CONTROL *) local->Data;
		if (discnodecontrol->Channel == TRUE &&
			discnodecontrol->Source[discnodecontrol->SourceLen - 1] == channelIndex &&
			discnodecontrol->ChannelCallback == NULL)
		{
			discnodecontrol->ChannelCallback = ChannelRegisterSink(discnodecontrol->Source[discnodecontrol->SourceLen - 1],
																   DiscNodeChannelPkt, DiscNodeDeRegister, discnodecontrol);
		}
	}

	return TRUE;
}

LONG TableSize(DISCNODE_CONTROL * discnodecontrol, INT type)
{
	switch (type)
	{
	case DISC_TYPE_ETHTIME:
		return discnodecontrol->EthTableSize;
	case DISC_TYPE_IPTIME:
		return discnodecontrol->IPTableSize;
	case DISC_TYPE_UDPTIME:
		return discnodecontrol->UDPTableSize;
	case DISC_TYPE_TCPTIME:
		return discnodecontrol->TCPTableSize;
	default:
		DnpapMessage(DMC_FATAL, DISC_TYPE, "discnodecontrol: unknown type");
		DnpapExit(DISC_TYPE);
	}

	return 0;
}

INT DiscTableType(SNMP_OBJECT * obj, WORD idlen)
{
	LONG disctable[] =
	{BTNG_MIB, 16};
	INT disctablelen = sizeof(disctable) / sizeof(disctable[0]);

	if (obj->Id[disctablelen] == DISC_TABLE_ETHNODE)
		return DISC_TYPE_ETH;
	if (obj->Id[disctablelen] == DISC_TABLE_IPNODE)
		return DISC_TYPE_IP;
	if (obj->Id[disctablelen] == DISC_TABLE_UDPNODE)
		return DISC_TYPE_UDP;
	if (obj->Id[disctablelen] == DISC_TABLE_TCPNODE)
		return DISC_TYPE_TCP;
	if (obj->Id[disctablelen] == DISC_TABLE_ETHTIME)
		return DISC_TYPE_ETHTIME;
	if (obj->Id[disctablelen] == DISC_TABLE_IPTIME)
		return DISC_TYPE_IPTIME;
	if (obj->Id[disctablelen] == DISC_TABLE_UDPTIME)
		return DISC_TYPE_UDPTIME;
	if (obj->Id[disctablelen] == DISC_TABLE_TCPTIME)
		return DISC_TYPE_TCPTIME;

	return 0;
}

WORD DiscTableLen(INT type, INT addrtype, WORD * offset, WORD * sublen)
{
	LONG disctable[] =
	{BTNG_MIB, 16};
	INT disctablelen = sizeof(disctable) / sizeof(disctable[0]);

	switch (addrtype)
	{
	case DISC_TYPE_ETH:
	case DISC_TYPE_ETHTIME:
		offset != NULL ? *offset = 0 : 0;
		sublen != NULL ? *sublen = DISC_SIZE_ETHADDR : 0;
		break;
	case DISC_TYPE_IP:
	case DISC_TYPE_IPTIME:
		offset != NULL ? *offset = DISC_SIZE_ETHADDR : 0;
		sublen != NULL ? *sublen = DISC_SIZE_IPADDR : 0;
		break;
	case DISC_TYPE_UDP:
	case DISC_TYPE_UDPTIME:
		offset != NULL ? *offset = DISC_SIZE_TOIPADDR : 0;
		sublen != NULL ? *sublen = DISC_SIZE_UDPADDR : 0;
		break;
	case DISC_TYPE_TCP:
	case DISC_TYPE_TCPTIME:
		offset != NULL ? *offset = DISC_SIZE_TOIPADDR : 0;
		sublen != NULL ? *sublen = DISC_SIZE_TCPADDR : 0;
		break;
	default:
		DnpapMessage(DMC_FATAL, DISC_TYPE, "discnodecontrol: unknown type");
		DnpapExit(DISC_TYPE);
	}

	switch (type)
	{
	case DISC_TYPE_ETH:
		return DISC_ETHNODEINDEXSIZE;
	case DISC_TYPE_IP:
		return DISC_IPNODEINDEXSIZE;
	case DISC_TYPE_UDP:
		return DISC_UDPNODEINDEXSIZE;
	case DISC_TYPE_TCP:
		return DISC_TCPNODEINDEXSIZE;
	case DISC_TYPE_ETHTIME:
		return DISC_ETHTIMEINDEXSIZE;
	case DISC_TYPE_IPTIME:
		return DISC_IPTIMEINDEXSIZE;
	case DISC_TYPE_UDPTIME:
		return DISC_UDPTIMEINDEXSIZE;
	case DISC_TYPE_TCPTIME:
		return DISC_TCPTIMEINDEXSIZE;
	default:
		DnpapMessage(DMC_FATAL, DISC_TYPE, "discnodecontrol: unknown type");
		DnpapExit(DISC_TYPE);
	}

	return 0;
}
