/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#ifndef _DHOSTC_H
#define _DHOSTC_H


#include <dnpap.h>
#include <boolset.h>
#include <snmp.h>
#include <hash.h>
#include <channel.h>

/*  dhostx.h exposes the DNPAP_HOST structure  */
#include "dhostx.h"


#define HOST_BOOLEAN_DATASOURCE		1
#define HOST_BOOLEAN_OWNER			2


typedef struct _DNPAP_HOST_CONTROL
{
    ULONG           Index;
    ULONG           Source[SNMP_SIZE_BUFINT];
    WORD            SourceLen;
    BOOLEAN         Channel;
    SinkCallback   *ChannelCallback;
	WORD			AddressLen;
    LONG            TableSize;
    ULONG           LastDeleteTime;
    BYTE            Owner[SNMP_SIZE_BUFCHR];
    WORD            OwnerLen;
    LONG            Status;

    /*  DNPAP extensions  */
    LONG            NrBuckets;
    ULONG           Interval;


    MAC_COLL        Coll;
    MAC_IFACE       *Iface;
    
    HashTable	    *Table;
    
    DNPAP_HOST	    *LexList;
    
    DNPAP_HOST	    *TimeList;
    DNPAP_HOST	    *TimeLast;
    DNPAP_HOST	    **TimeTable;
    BOOLEAN			TimeListUpdated;
    
    DNPAP_HOST      *LRUList;
    DNPAP_HOST      *LRULast;

    DnpapTopNCallback  *TopNs;
    
	BooleanSet       ObjectSet;
    
} DNPAP_HOST_CONTROL;


struct DnpapTopNCallback_
{   
    DnpapHostTopNCallback Notify;
    VOID *TopN;             
    DnpapTopNCallback *Prev;
    DnpapTopNCallback *Next;
};


BOOLEAN DnpapHostConfigInit(VOID);

BOOLEAN DnpapHostCInit(DNPAP_HOST_CONTROL *hostcontrol);
BOOLEAN DnpapHostCStart(DNPAP_HOST_CONTROL *hostcontrol);
BOOLEAN DnpapHostCStop(DNPAP_HOST_CONTROL *hostcontrol);

DnpapTopNCallback *DnpapHostAddTopN(DNPAP_HOST_CONTROL *hostcontrol, DnpapHostTopNCallback notify, VOID *topn);
BOOLEAN DnpapHostRemoveTopN(DNPAP_HOST_CONTROL *hostcontrol, DnpapTopNCallback *callback);

VOID DnpapHostChannelPkt(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *sink);
VOID DnpapHostDeRegister(ULONG channelIndex, VOID *sink);

DNPAP_HOST* DnpapHostSearchNext(DNPAP_HOST_CONTROL *hostcontrol, DNPAP_HOST *host, WORD len);
DNPAP_HOST *DnpapHostBucketSearchNext(DNPAP_HOST_CONTROL * hostcontrol, DNPAP_HOST * host, WORD len, LONG *bucket);
DNPAP_HOST* DnpapHostSearchHash(DNPAP_HOST_CONTROL *hostcontrol, SNMP_OBJECT *obj, WORD idlen);
DNPAP_HOST* DnpapHostTimeSearch(DNPAP_HOST_CONTROL *hostcontrol, SNMP_OBJECT *obj, WORD idlen);
LONG  DnpapHostSearchCreationOrder(DNPAP_HOST_CONTROL *hostcontrol, DNPAP_HOST *host);


#endif
