/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>
#include <dmath.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <message.h>
#include <mibsup.h>
#include <mac.h>

#include "dhoste.h"
#include "dhostc.h"
#include "dhost.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define CONTROLINDEXSIZE   		1
#define HOSTINDEXSIZE			7		/* not the real size, only used as switch value */
#define TIMEINDEXSIZE      		2
#define HOSTBUCKETINDEXSIZE     8		/* not the real size, only used as switch value */
#define TIMEBUCKETINDEXSIZE     3

static MIB_LOCAL *HostControlInstance = NULL;

static MIB_LOCAL *DnpapHostMibRmon(SNMP_OBJECT * obj, MIB_LOCAL * local, WORD idlen, WORD extra);
static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, DNPAP_HOST ** elem);

static DNPAP_HOST *HostGetFirstNext(DNPAP_HOST_CONTROL * hostcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first);
static DNPAP_HOST *HostBucketGetFirstNext(DNPAP_HOST_CONTROL * hostcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first);
static DNPAP_HOST *HostTimeBucketGetFirstNext(DNPAP_HOST_CONTROL * hostcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first);

BOOLEAN DnpapHostMInit(VOID)
{
	MessageConfig(DNPAP_HOST_ERROR, "DnpapHost");
	return DnpapHostConfigInit();
}

MIB_LOCAL *DnpapHostMibRmon(SNMP_OBJECT * obj, MIB_LOCAL * local, WORD idlen, WORD extra)
{
	WORD addrlen;

	if (local == NULL)
		return NULL;
	if (obj->Request != SNMP_PDU_NEXT)
	{
		if (obj->IdLen == idlen)
			return NULL;
		/* first search the right table */
		while (local != NULL && local->Index < obj->Id[idlen])
			local = local->Next;
		if (local == NULL)
			return NULL;
		addrlen = ((DNPAP_HOST_CONTROL *) local->Data)->AddressLen;
		if (local->Index != obj->Id[idlen] || addrlen == 0)
			return NULL;
		if (obj->IdLen != idlen + 1 + 1 + addrlen + extra ||
			obj->Id[idlen + 1] != addrlen)
			return NULL;
		return local;
	}
	if (obj->IdLen == idlen)			/* first instance */
		return local;
	while (local != NULL && local->Index < obj->Id[idlen])
		local = local->Next;
	if (local == NULL || local->Index < obj->Id[idlen])
		return NULL;
	return local;
}

#define CONTROL_DATA    ((DNPAP_HOST_CONTROL *) (*local)->Data)

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, DNPAP_HOST ** elem)
{
	BOOLEAN first = FALSE;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case HOSTBUCKETINDEXSIZE:
		while (*local != NULL && (*elem = HostBucketGetFirstNext(CONTROL_DATA, obj, idlen, first)) == NULL)
		{
			first = TRUE;
			*local = (*local)->Next;
		}
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		break;
	case HOSTINDEXSIZE:
		while (*local != NULL && (*elem = HostGetFirstNext(CONTROL_DATA, obj, idlen, first)) == NULL)
		{
			first = TRUE;
			*local = (*local)->Next;
		}
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		break;
	case TIMEBUCKETINDEXSIZE:
		while (*local != NULL && (*elem = HostTimeBucketGetFirstNext(CONTROL_DATA, obj, idlen, first)) == NULL)
		{
			first = TRUE;
			*local = (*local)->Next;
		}
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		break;
	case TIMEINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			if (obj->Id[idlen + 1] < CONTROL_DATA->TableSize)
			{
				obj->Id[idlen] = (*local)->Index;
				obj->Id[idlen + 1]++;
				obj->IdLen = idlen + TIMEINDEXSIZE;
				*elem = DnpapHostTimeSearch(CONTROL_DATA, obj, idlen);
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL && CONTROL_DATA->TableSize == 0)
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen + 1] = 1;
		obj->IdLen = idlen + TIMEINDEXSIZE;
		*elem = DnpapHostTimeSearch(CONTROL_DATA, obj, idlen);
		break;
	case CONTROLINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + CONTROLINDEXSIZE;
		break;
	}
	return TRUE;
}

DNPAP_HOST *HostGetFirstNext(DNPAP_HOST_CONTROL * hostcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first)
{
	DNPAP_HOST host, *p;
	INT i, len;
	WORD addrlen = hostcontrol->AddressLen;

	if (hostcontrol->TableSize == 0)
		return NULL;

	if (first == FALSE)
	{
		if (obj->IdLen > idlen + 1)
		{
			if (obj->Id[idlen + 1] > addrlen)
				return NULL;
			if (obj->Id[idlen + 1] < addrlen)
				first = TRUE;
		}
		else
			first = TRUE;
	}

	if (first == TRUE)
	{
		p = hostcontrol->LexList;
		for (i = 0; i < addrlen; i++)
			obj->Id[idlen + 2 + i] = (LONG) p->Address[i];
		obj->Id[idlen + 1] = addrlen;
		obj->IdLen = idlen + 1 + 1 + addrlen;
		return p;
	}

	len = obj->IdLen - idlen - 2;
	if (len < 0)
		len = 0;
	else if (len > addrlen)
		len = addrlen;

	for (i = 0; i < len; i++)
		host.Address[i] = (BYTE) obj->Id[idlen + 2 + i];

	p = DnpapHostSearchNext(hostcontrol, &host, len);

	if (p != NULL)
	{
		for (i = 0; i < addrlen; i++)
			obj->Id[idlen + 2 + i] = (LONG) p->Address[i];
		obj->Id[idlen + 1] = addrlen;
		obj->IdLen = idlen + 1 + 1 + addrlen;
	}
	return p;
}

DNPAP_HOST *HostBucketGetFirstNext(DNPAP_HOST_CONTROL * hostcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first)
{
	DNPAP_HOST host, *p;
	INT i, len;
	LONG bucket;
	WORD addrlen = hostcontrol->AddressLen;

	if (hostcontrol->TableSize == 0 || hostcontrol->NrBuckets == 0)
		return NULL;

	if (first == FALSE)
	{
		if (obj->IdLen > idlen + 1)
		{
			if (obj->Id[idlen + 1] > addrlen)
				return NULL;
			if (obj->Id[idlen + 1] < addrlen)
				first = TRUE;
		}
		else
			first = TRUE;
	}

	if (first == TRUE)
	{
		p = hostcontrol->LexList;
		for (i = 0; i < addrlen; i++)
			obj->Id[idlen + 2 + i] = (LONG) p->Address[i];
		obj->Id[idlen + 2 + addrlen] = 1;
		obj->Id[idlen + 1] = addrlen;
		obj->IdLen = idlen + 1 + 1 + addrlen + 1;
		return p;
	}

	len = obj->IdLen - idlen - 2;
	if (len < 0)
		len = 0;
	else if (len > addrlen + 1)
		len = addrlen + 1;

	for (i = 0; i < IntMin(len, addrlen); i++)
		host.Address[i] = (BYTE) obj->Id[idlen + 2 + i];

	bucket = obj->Id[idlen + 2 + addrlen];
	p = DnpapHostBucketSearchNext(hostcontrol, &host, len, &bucket);

	if (p != NULL)
	{
		for (i = 0; i < addrlen; i++)
			obj->Id[idlen + 2 + i] = (LONG) p->Address[i];
		obj->Id[idlen + 1] = addrlen;
		obj->Id[idlen + 2 + addrlen] = bucket;
		obj->IdLen = idlen + 1 + 1 + addrlen + 1;
	}

	return p;
}

DNPAP_HOST *HostTimeBucketGetFirstNext(DNPAP_HOST_CONTROL * hostcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first)
{
	INT len;

	if (hostcontrol->TableSize == 0 || hostcontrol->NrBuckets == 0)
		return NULL;

	if (first == TRUE)
	{
		obj->Id[idlen + 1] = 1;
		obj->Id[idlen + 1 + 1] = 1;
		obj->IdLen = idlen + TIMEBUCKETINDEXSIZE;
		return DnpapHostTimeSearch(hostcontrol, obj, idlen);
	}

	len = obj->IdLen - idlen - 1;
	if (len < 0)
		len = 0;
	else if (len > 2)
		len = 2;

	switch (len)
	{
	case 0:
		obj->Id[idlen + 1] = 1;
	case 1:
		obj->Id[idlen + 1 + 1] = 1;
		break;
	case 2:
		if (obj->Id[idlen + 1 + 1] < hostcontrol->NrBuckets)
			obj->Id[idlen + 1 + 1]++;
		else
		{
			if (obj->Id[idlen + 1] < hostcontrol->TableSize)
				obj->Id[idlen + 1]++;
			else
				return NULL;
			obj->Id[idlen + 1 + 1] = 1;
		}
		break;
	}

	obj->IdLen = idlen + TIMEBUCKETINDEXSIZE;

	return DnpapHostTimeSearch(hostcontrol, obj, idlen);
}

WORD DnpapHostControlIndex(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostControlDataSource(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST_CONTROL *data;
	MAC_IFACE *iface;
	LONG IfaceOid[] =
	{1, 3, 6, 1, 2, 1, 2, 2, 1, 1, 1};
	WORD IfaceOidLen = sizeof(IfaceOid) / sizeof(LONG);
	LONG ChannelOid[] =
	{1, 3, 6, 1, 4, 1, 464, 2, 7, 2, 1, 1, 1};
	WORD ChannelOidLen = sizeof(ChannelOid) / sizeof(LONG);

	if ((local = MibRmon(obj, HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_HOST_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, HOST_BOOLEAN_DATASOURCE))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufInt, data->Source, data->SourceLen * sizeof(data->Source[0]));
		obj->SyntaxLen = data->SourceLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_HOST_CONTROL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (IfaceOidLen == obj->SyntaxLen &&
			memcmp(IfaceOid, obj->Syntax.BufInt, (IfaceOidLen - 1) * sizeof(IfaceOid[0])) == 0)
		{
			data->Channel = FALSE;
			if ((iface = MacIfaceGet((WORD) obj->Syntax.BufInt[obj->SyntaxLen - 1])) == NULL)
				return SNMP_BADVALUE;
			if (iface->type != MAC_TYPE_ETHERNET_CSMACD && iface->type != MAC_TYPE_88023_CSMACD)
				return SNMP_BADVALUE;
			memcpy(data->Source, obj->Syntax.BufInt, obj->SyntaxLen * sizeof(data->Source[0]));
			data->SourceLen = obj->SyntaxLen;
			data->Iface = iface;
			BooleanSetTrue(data->ObjectSet, HOST_BOOLEAN_DATASOURCE);
			return SNMP_NOERROR;
		}
		else if (ChannelOidLen == obj->SyntaxLen &&
				 memcmp(ChannelOid, obj->Syntax.BufInt, (ChannelOidLen - 1) * sizeof(ChannelOid[0])) == 0)
		{
			data->Channel = TRUE;
			memcpy(data->Source, obj->Syntax.BufInt, obj->SyntaxLen * sizeof(data->Source[0]));
			data->SourceLen = obj->SyntaxLen;
			BooleanSetTrue(data->ObjectSet, HOST_BOOLEAN_DATASOURCE);
			return SNMP_NOERROR;
		}
		return SNMP_BADVALUE;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostControlTableSize(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST_CONTROL *data;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_HOST_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->TableSize;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostControlLastDeleteTime(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST_CONTROL *data;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_HOST_CONTROL *) local->Data;
		obj->Syntax.LngUns = data->LastDeleteTime;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostControlBuckets(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST_CONTROL *data;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_HOST_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->NrBuckets;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_HOST_CONTROL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt < 0)
			return SNMP_BADVALUE;
		data->NrBuckets = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostControlInterval(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST_CONTROL *data;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_HOST_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Interval;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_HOST_CONTROL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt < 1)
			return SNMP_BADVALUE;
		data->Interval = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostControlOwner(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST_CONTROL *data;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_HOST_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, HOST_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_HOST_CONTROL *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, obj->Syntax.BufChr, obj->SyntaxLen);
		data->OwnerLen = obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, HOST_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostControlStatus(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST_CONTROL *data;

	local = MibRmon(obj, HostControlInstance, IdLen, CONTROLINDEXSIZE);

	switch (obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (DNPAP_HOST_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (DNPAP_HOST_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(obj, &HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(DNPAP_HOST_CONTROL))) == NULL)
				{
					MibRemove(obj, &HostControlInstance, IdLen, CONTROLINDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (DNPAP_HOST_CONTROL *) local->Data;
				if (DnpapHostCInit(data) == TRUE)
				{
					data->Index = local->Index;
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, DHOST_CREATE, "dhost: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(obj, &HostControlInstance, IdLen, CONTROLINDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (DNPAP_HOST_CONTROL *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (DnpapHostCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, DHOST_START, "dhost: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (DnpapHostCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, DHOST_DESTROY, "dhost: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &HostControlInstance, IdLen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				return SNMP_NOERROR;
			case SNMP_INVALID:
				if (DnpapHostCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, DHOST_DESTROY, "dhost: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &HostControlInstance, IdLen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (DnpapHostCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostAddress(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Address, obj->SyntaxLen = ((DNPAP_HOST_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostCreationOrder(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = DnpapHostSearchCreationOrder(local->Data, data);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIndex(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostInPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostInOctets(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InOctets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutOctets(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutOctets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutErrors(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutErrors;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutBroadcastPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBroadcastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutMulticastPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutMulticastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

#ifdef NOT_HOST_GENERIC
WORD DnpapHostInTCPPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InTCPPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutTCPPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutTCPPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostInOtherPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InOtherPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutOtherPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutOtherPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
#endif

WORD DnpapHostInPkts64(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InBucket1Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostInPkts65to127(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InBucket2Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostInPkts128to255(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InBucket3Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostInPkts256to511(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InBucket4Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostInPkts512to1023(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InBucket5Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostInPkts1024to1518(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InBucket6Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutPkts64(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBucket1Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutPkts65to127(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBucket2Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutPkts128to255(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBucket3Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutPkts256to511(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBucket4Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutPkts512to1023(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBucket5Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostOutPkts1024to1518(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 0)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBucket6Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTAddress(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Address, obj->SyntaxLen = ((DNPAP_HOST_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTInterval(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = obj->Id[IdLen + 1 + ((DNPAP_HOST_CONTROL *) local->Data)->AddressLen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTCreationOrder(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = DnpapHostSearchCreationOrder(local->Data, data);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTIndex(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTIATPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->IATBucketsPkts[obj->Id[IdLen + 1 + ((DNPAP_HOST_CONTROL *) local->Data)->AddressLen + 1] - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTISTPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->ISTBucketsPkts[obj->Id[IdLen + 1 + ((DNPAP_HOST_CONTROL *) local->Data)->AddressLen + 1] - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTIATOctets(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->IATBucketsOctets[obj->Id[IdLen + 1 + ((DNPAP_HOST_CONTROL *) local->Data)->AddressLen + 1] - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTISTOctets(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = DnpapHostMibRmon(obj, HostControlInstance, IdLen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, HOSTBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostSearchHash(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->ISTBucketsOctets[obj->Id[IdLen + 1 + ((DNPAP_HOST_CONTROL *) local->Data)->AddressLen + 1] - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeAddress(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Address, obj->SyntaxLen = ((DNPAP_HOST_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeCreationOrder(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = obj->Id[IdLen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeIndex(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeInPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeInOctets(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InOctets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutOctets(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutOctets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutErrors(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutErrors;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutBroadcastPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBroadcastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutMulticastPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutMulticastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

#ifdef NOT_HOST_GENERIC
WORD DnpapHostTimeInTCPPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InTCPPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutTCPPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutTCPPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeInOtherPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InOtherPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutOtherPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutOtherPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
#endif

WORD DnpapHostTimeInPkts64(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InBucket1Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeInPkts65to127(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InBucket2Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeInPkts128to255(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InBucket3Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeInPkts256to511(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InBucket4Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeInPkts512to1023(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InBucket5Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeInPkts1024to1518(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->InBucket6Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutPkts64(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBucket1Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutPkts65to127(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBucket2Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutPkts128to255(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBucket3Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutPkts256to511(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBucket4Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutPkts512to1023(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBucket5Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostTimeOutPkts1024to1518(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OutBucket6Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTTimeAddress(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEBUCKETINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Address, obj->SyntaxLen = ((DNPAP_HOST_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTTimeInterval(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEBUCKETINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = obj->Id[IdLen + 1 + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTTimeCreationOrder(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEBUCKETINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = obj->Id[IdLen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTTimeIndex(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEBUCKETINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTTimeIATPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEBUCKETINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->IATBucketsPkts[obj->Id[IdLen + 1 + 1] - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTTimeISTPkts(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEBUCKETINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->ISTBucketsPkts[obj->Id[IdLen + 1 + 1] - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTTimeIATOctets(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEBUCKETINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->IATBucketsOctets[obj->Id[IdLen + 1 + 1] - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapHostIFTTimeISTOctets(SNMP_OBJECT * obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_HOST *data = NULL;

	if ((local = MibRmon(obj, HostControlInstance, IdLen, TIMEBUCKETINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, IdLen, TIMEBUCKETINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapHostTimeSearch(local->Data, obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->ISTBucketsOctets[obj->Id[IdLen + 1 + 1] - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

DnpapTopNCallback *DnpapHostRegisterTopN(ULONG hostindex, DnpapHostTopNCallback notify, VOID * topn)
{
	MIB_LOCAL *local;
	DNPAP_HOST_CONTROL *data;
	DnpapTopNCallback *callback;

	for (local = HostControlInstance; local != NULL; local = local->Next)
	{
		data = (DNPAP_HOST_CONTROL *) local->Data;
		if (data->Index < hostindex)
			continue;
		if (data->Index != hostindex)
			return NULL;
		if ((callback = DnpapHostAddTopN(data, notify, topn)) != NULL)
			return callback;
		break;
	}

	return NULL;
}

BOOLEAN DnpapHostDeRegisterTopN(ULONG hostindex, DnpapTopNCallback * topn)
{
	MIB_LOCAL *local;
	DNPAP_HOST_CONTROL *data;

	for (local = HostControlInstance; local != NULL; local = local->Next)
	{
		data = (DNPAP_HOST_CONTROL *) local->Data;
		if (data->Index < hostindex)
			continue;
		if (data->Index != hostindex)
			return FALSE;
		if (DnpapHostRemoveTopN(data, topn) == TRUE)
			return TRUE;
		break;
	}

	return FALSE;
}

BOOLEAN DnpapHostRegisterAll(ULONG hostindex)
{
	MIB_LOCAL *local;
	DNPAP_HOST_CONTROL *data;

	for (local = HostControlInstance; local != NULL; local = local->Next)
	{
		data = (DNPAP_HOST_CONTROL *) local->Data;
		if (data->Channel == TRUE && data->Source[data->SourceLen - 1] == hostindex &&
			data->ChannelCallback == NULL)
		{
			data->ChannelCallback = ChannelRegisterSink(data->Source[data->SourceLen - 1],
														DnpapHostChannelPkt, DnpapHostDeRegister, data);
		}
	}

	return TRUE;
}
