/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dnpap.h>
#include <config.h>
#include <mac.h>
#include <snmp.h>
#include <mibrqs.h>
#include <mibsup.h>
#include <message.h>

#include "infoe.h"
#include "infoc.h"
#include "info.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define INDEXSIZE       1

static MIB_LOCAL *beholderInfo = NULL;

static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize);

BOOLEAN InfoMInit(VOID)
{
	LONG i, nriface;
	CHAR s[81];
	BOOLEAN v;

	MAC_IFACE *iface;

	LONG infosource[] =
	{1, 3, 6, 1, 2, 1, 2, 2, 1, 1, 1};
	CHAR infoowner[] = "monitorBeholderInfo";

#define STATUSINDEX     (BTNG_MIBLEN + 4)
	SNMP_OBJECT infoobj[] =
	{
		{SNMP_PDU_SET,
		 {BTNG_MIB, 1, 1, 1, 10, 1}, (BTNG_MIBLEN + 5), SNMP_INTEGER},
		{SNMP_PDU_SET,
		 {BTNG_MIB, 1, 1, 1, 9, 1}, (BTNG_MIBLEN + 5), SNMP_DISPLAYSTR},
		{SNMP_PDU_SET,
		 {BTNG_MIB, 1, 1, 1, 2, 1}, (BTNG_MIBLEN + 5), SNMP_OBJECTID}
	};
	WORD dummy;

	MessageConfig(INFO_ERROR, "BeholderInfo");

	nriface = (LONG) MacIfaceCount();
	for (i = 1; i <= nriface; i++)
	{
		sprintf(s, "beholder.info.iface.%ld", i);
		if (ConfigGetBoolean(s, &v) == TRUE)
		{
			iface = MacIfaceGet((WORD) i);
			if (iface != NULL && v == TRUE &&
				(iface->type == MAC_TYPE_ETHERNET_CSMACD ||
				 iface->type == MAC_TYPE_88023_CSMACD))
			{
				infoobj[0].Id[STATUSINDEX] = i;
				infoobj[1].Id[STATUSINDEX] = i;
				infoobj[2].Id[STATUSINDEX] = i;

				infoobj[0].Request = SNMP_PDU_SET;
				infoobj[0].Syntax.LngInt = SNMP_INVALID;
				MibRequest(1, &infoobj[0], &dummy);
				infoobj[0].Request = SNMP_PDU_SET;
				infoobj[0].Syntax.LngInt = SNMP_CREATEREQUEST;
				if (MibRequest(1, &infoobj[0], &dummy) != SNMP_NOERROR)
				{
					DnpapMessage(DMC_WARNING, INFO_INFOINIT, "beholderInfo: couldn't create beholderInfo collector %ld", i);
					continue;
				}
				infoobj[1].Request = SNMP_PDU_SET;
				strncpy(infoobj[1].Syntax.BufChr, infoowner, infoobj[1].SyntaxLen = strlen(infoowner));
				infoobj[2].Request = SNMP_PDU_SET;
				memcpy(infoobj[2].Syntax.BufInt, infosource, sizeof(infosource));
				infoobj[2].Syntax.BufInt[sizeof(infosource) / sizeof(infosource[0]) - 1] = i;
				infoobj[2].SyntaxLen = sizeof(infosource) / sizeof(infosource[0]);
				if (MibRequest(2, &infoobj[1], &dummy) != SNMP_NOERROR)
				{
					DnpapMessage(DMC_WARNING, INFO_INFOINIT, "beholderInfo: couldn't set owner of beholderInfo collector %ld", i);
					infoobj[0].Request = SNMP_PDU_SET;
					infoobj[0].Syntax.LngInt = SNMP_INVALID;
					MibRequest(1, &infoobj[0], &dummy);
					continue;
				}
				infoobj[0].Request = SNMP_PDU_SET;
				infoobj[0].Syntax.LngInt = SNMP_VALID;
				if (MibRequest(1, &infoobj[0], &dummy) != SNMP_NOERROR)
				{
					DnpapMessage(DMC_WARNING, INFO_INFOINIT, "beholderInfo: couldn't set beholderInfo collector %ld to valid", i);
					infoobj[0].Request = SNMP_PDU_SET;
					infoobj[0].Syntax.LngInt = SNMP_INVALID;
					MibRequest(1, &infoobj[0], &dummy);
					continue;
				}
			}
		}
	}

	return TRUE;
}

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize)
{
	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->IdLen == idlen || (*local)->Index > obj->Id[idlen] ||
		(*local = (*local)->Next) != NULL)
	{
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + indexsize;
		return TRUE;
	}

	return FALSE;
}

WORD InfoIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(Obj, beholderInfo, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;
	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD InfoDataSource(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BEHOLDER_INFO *data;
	MAC_IFACE *iface;

	if ((local = MibRmon(Obj, beholderInfo, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BEHOLDER_INFO *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, INFO_BOOLEAN_DATASOURCE))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufInt, data->Source, data->SourceLen * sizeof(LONG));
		Obj->SyntaxLen = data->SourceLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (BEHOLDER_INFO *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (data->SourceLen != Obj->SyntaxLen ||
			memcmp(data->Source, Obj->Syntax.BufInt, (data->SourceLen - 1) * sizeof(data->Source[0])))
			return SNMP_BADVALUE;
		if ((iface = MacIfaceGet((WORD) Obj->Syntax.BufInt[Obj->SyntaxLen - 1])) == NULL)
			return SNMP_BADVALUE;
		if (iface->type != MAC_TYPE_ETHERNET_CSMACD && iface->type != MAC_TYPE_88023_CSMACD)
			return SNMP_BADVALUE;
		memcpy(data->Source, Obj->Syntax.BufInt, Obj->SyntaxLen * sizeof(data->Source[0]));
		data->SourceLen = Obj->SyntaxLen;
		data->Iface = iface;
		BooleanSetTrue(data->ObjectSet, INFO_BOOLEAN_DATASOURCE);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD InfoLostPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BEHOLDER_INFO *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, beholderInfo, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BEHOLDER_INFO *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.LostPkts - data->BaseStat.LostPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD InfoPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BEHOLDER_INFO *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, beholderInfo, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BEHOLDER_INFO *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.Pkts - data->BaseStat.Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD InfoDiscardedPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BEHOLDER_INFO *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, beholderInfo, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BEHOLDER_INFO *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.DiscardedPkts - data->BaseStat.DiscardedPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD InfoTruncatedPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BEHOLDER_INFO *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, beholderInfo, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BEHOLDER_INFO *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.TruncatedPkts - data->BaseStat.TruncatedPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD InfoBufferedPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BEHOLDER_INFO *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, beholderInfo, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BEHOLDER_INFO *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.BufferedPkts /* - data->BaseStat.BufferedPkts */ ;	/* DMW */
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD InfoStackedPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BEHOLDER_INFO *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, beholderInfo, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BEHOLDER_INFO *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.StackedPkts - data->BaseStat.StackedPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD InfoOwner(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BEHOLDER_INFO *data;

	if ((local = MibRmon(Obj, beholderInfo, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BEHOLDER_INFO *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, INFO_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		Obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (BEHOLDER_INFO *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, Obj->Syntax.BufChr, Obj->SyntaxLen);
		data->OwnerLen = Obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, INFO_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD InfoStatus(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BEHOLDER_INFO *data;

	local = MibRmon(Obj, beholderInfo, IdLen, INDEXSIZE);

	switch (Obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (BEHOLDER_INFO *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (BEHOLDER_INFO *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(Obj, &beholderInfo, IdLen, INDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(BEHOLDER_INFO))) == NULL)
				{
					MibRemove(Obj, &beholderInfo, IdLen, INDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (BEHOLDER_INFO *) local->Data;
				memset(data, 0, sizeof(BEHOLDER_INFO));
				if (InfoInit(data) == TRUE)
				{
					((BEHOLDER_INFO *) local->Data)->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, INFO_CREATE, "beholderInfo: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(Obj, &beholderInfo, IdLen, INDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (BEHOLDER_INFO *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (InfoStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, INFO_ACTIVE, "beholderInfo: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				else
					return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (InfoStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, INFO_DESTROY, "beholderInfo: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &beholderInfo, IdLen, INDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_INVALID:
				if (InfoStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, INFO_DESTROY, "beholderInfo: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &beholderInfo, IdLen, INDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (InfoStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			case SNMP_VALID:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
