/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */

#ifndef _DMATRIX_C_H
#define _DMATRIX_C_H


#include <dnpap.h>
#include <boolset.h>
#include <snmp.h>
#include <channel.h>

#include "hash.h"


#define MATRIX_BOOLEAN_DATASOURCE		1
#define MATRIX_BOOLEAN_OWNER			2


typedef struct _ConnStats
{
    ULONG  Pkts;
    ULONG  Octets;
    ULONG  Errors;

} ConnStats;


typedef struct _ConnInfo
{               
    struct _Host * ConnHost;
    
    ConnStats *Stats;

    struct _ConnInfo *LexNext;
    struct _ConnInfo *LexPrev;

    struct _ConnInfo *TopNext;
    struct _ConnInfo *TopPrev;

    struct _LRUEntry *Entry;
    
} ConnInfo;


typedef struct _Host
{
    BYTE            Addr[HOST_SIZE_ADDR];
    INT             NumberConnect;

    ConnInfo        *TopList;
    ConnInfo        *LexList;
    ConnInfo        *TopLast;

    struct _Host  *Next;
    struct _Host  *Prev;

} Host;


typedef struct _LRUEntry
{
    ConnInfo          *ConnSrc;
    ConnInfo          *ConnDst;

    struct _LRUEntry  *LRUPrev;
    struct _LRUEntry  *LRUNext;

} LRUEntry;

	   
typedef struct _DNPAP_MATRIX_CONTROL
{
    ULONG           Source[SNMP_SIZE_BUFINT];
    WORD            SourceLen;
    BOOLEAN         Channel;
    SinkCallback   *ChannelCallback;
	WORD			AddressLen;
    LONG            TableSize;
    ULONG           LastDeleteTime;
    BYTE            Owner[SNMP_SIZE_BUFCHR];
    WORD            OwnerLen;
    LONG            Status;

    MAC_COLL        Coll;
    MAC_IFACE       *Iface;
    
    HashTable       *SrcTable;
    HashTable       *DstTable;
	
    LRUEntry        *LRUList;
    LRUEntry        *LRULast;

    Host            *SrcList;
    Host            *DstList;
    
	BooleanSet       ObjectSet;
    
} DNPAP_MATRIX_CONTROL;



BOOLEAN DnpapMatrixConfigInit(VOID);

BOOLEAN DnpapMatrixCInit(DNPAP_MATRIX_CONTROL *matrixcontrol);
BOOLEAN DnpapMatrixCStart(DNPAP_MATRIX_CONTROL *matrixcontrol);
BOOLEAN DnpapMatrixCStop(DNPAP_MATRIX_CONTROL *matrixcontrol);

VOID DnpapMatrixChannelPkt(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *sink);
VOID DnpapMatrixDeRegister(ULONG channelIndex, VOID *sink);

ConnInfo *DnpapMatrixSearchHostObj(DNPAP_MATRIX_CONTROL * matrixcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN source, Host ** host, INT * rank);
ConnInfo *DnpapMatrixSearchTopObj(DNPAP_MATRIX_CONTROL * matrixcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN source, Host ** host, INT * rank);

Host *DnpapMatrixSearchHost(DNPAP_MATRIX_CONTROL *matrixcontrol, BYTE *addr, WORD len, BOOLEAN source);

ConnInfo *DMatrixListSearchAddr(Host *host, BYTE *addr, WORD len);
ConnInfo *DMatrixListSearchTop(Host *host, INT rank);
INT DMatrixListSearchRank(Host * host, BYTE * addr, ConnInfo ** conn, WORD addrlen);


#endif
