/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */

#include <dnpap.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <message.h>
#include <memory.h>
#include <mibsup.h>
#include <mac.h>

#include "dmatrixd.h"
#include "dmatrixe.h"
#include "dmatrixc.h"
#include "dmatrix.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define CONTROLINDEXSIZE   1
#define SRCDSTINDEXSIZE	  13			/* not the real size, only used as switch value */
#define RANKINDEXSIZE	   8

static MIB_LOCAL *MatrixControlInstance = NULL;

static MIB_LOCAL *DnpapMatrixMibRmon(SNMP_OBJECT * obj, MIB_LOCAL * local, WORD idlen);
static MIB_LOCAL *DnpapMatrixRankMibRmon(SNMP_OBJECT * obj, MIB_LOCAL * local, WORD idlen);
static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, ConnInfo ** elem, BOOLEAN SDorder, Host ** source, Host ** dest, INT * rank);
static ConnInfo *SrcDstGetFirstNext(DNPAP_MATRIX_CONTROL * matrixcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first, BOOLEAN SDorder, Host ** source, Host ** dest, INT * rank);
static ConnInfo *GetTopFirstNext(DNPAP_MATRIX_CONTROL * matrixcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first, BOOLEAN SDorder, Host ** source, Host ** dest, INT * rank);

BOOLEAN DnpapMatrixMInit(VOID)
{
	MessageConfig(DNPAP_MATRIX_ERROR, "DnpapMatrix");
	return DnpapMatrixConfigInit();
}

MIB_LOCAL *DnpapMatrixMibRmon(SNMP_OBJECT * obj, MIB_LOCAL * local, WORD idlen)
{
	WORD addrlen;

	if (local == NULL)
		return NULL;
	if (obj->Request != SNMP_PDU_NEXT)
	{
		if (obj->IdLen == idlen)
			return NULL;
		/* first search the right table */
		while (local != NULL && local->Index < obj->Id[idlen])
			local = local->Next;
		if (local == NULL)
			return NULL;
		addrlen = ((DNPAP_MATRIX_CONTROL *) local->Data)->AddressLen;
		if (local->Index != obj->Id[idlen] || addrlen == 0)
			return NULL;
		if (obj->IdLen != idlen + 1 + 2 + 2 * addrlen ||
			obj->Id[idlen + 1] != addrlen || obj->Id[idlen + 1 + addrlen + 1] != addrlen)
			return NULL;
		return local;
	}
	if (obj->IdLen == idlen)			/* first instance */
		return local;
	while (local != NULL && local->Index < obj->Id[idlen])
		local = local->Next;
	if (local == NULL || local->Index < obj->Id[idlen])
		return NULL;
	return local;
}

MIB_LOCAL *DnpapMatrixRankMibRmon(SNMP_OBJECT * obj, MIB_LOCAL * local, WORD idlen)
{
	WORD addrlen;

	if (local == NULL)
		return NULL;
	if (obj->Request != SNMP_PDU_NEXT)
	{
		/* first search the right table */
		while (local != NULL && local->Index < obj->Id[idlen])
			local = local->Next;
		if (local == NULL)
			return NULL;
		addrlen = ((DNPAP_MATRIX_CONTROL *) local->Data)->AddressLen;
		if (local->Index != obj->Id[idlen] || addrlen == 0)
			return NULL;
		if (obj->IdLen != idlen + 1 + 1 + addrlen + 1 ||
			obj->Id[idlen + 1] != addrlen)
			return NULL;
		return local;
	}
	if (obj->IdLen == idlen)			/* first instance */
		return local;
	while (local != NULL && local->Index < obj->Id[idlen])
		local = local->Next;
	if (local == NULL || local->Index < obj->Id[idlen])
		return NULL;
	return local;
}

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, ConnInfo ** elem, BOOLEAN SDorder, Host ** source, Host ** dest, INT * rank)
{
	BOOLEAN first = FALSE;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case SRCDSTINDEXSIZE:
		while (*local != NULL && (*elem = SrcDstGetFirstNext((DNPAP_MATRIX_CONTROL *) (*local)->Data, obj, idlen, first, SDorder, source, dest, rank)) == NULL)
		{
			first = TRUE;
			*local = (*local)->Next;
		}
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		break;
	case RANKINDEXSIZE:
		while (*local != NULL && (*elem = GetTopFirstNext((DNPAP_MATRIX_CONTROL *) (*local)->Data, obj, idlen, first, SDorder, source, dest, rank)) == NULL)
		{
			first = TRUE;
			*local = (*local)->Next;
		}
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		break;
	case CONTROLINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + CONTROLINDEXSIZE;
		break;
	}
	return TRUE;
}

WORD DnpapMatrixControlIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = MibRmon(obj, MatrixControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL, FALSE, &source, &dest, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixControlDataSource(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_MATRIX_CONTROL *data;
	MAC_IFACE *iface;
	LONG IfaceOid[] =
	{1, 3, 6, 1, 2, 1, 2, 2, 1, 1, 1};
	WORD IfaceOidLen = sizeof(IfaceOid) / sizeof(LONG);
	LONG ChannelOid[] =
	{1, 3, 6, 1, 4, 1, 464, 2, 7, 2, 1, 1, 1};
	WORD ChannelOidLen = sizeof(ChannelOid) / sizeof(LONG);
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = MibRmon(obj, MatrixControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL, FALSE, &source, &dest, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_MATRIX_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, MATRIX_BOOLEAN_DATASOURCE))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufInt, data->Source, data->SourceLen * sizeof(data->Source[0]));
		obj->SyntaxLen = data->SourceLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_MATRIX_CONTROL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (IfaceOidLen == obj->SyntaxLen &&
			memcmp(IfaceOid, obj->Syntax.BufInt, (IfaceOidLen - 1) * sizeof(IfaceOid[0])) == 0)
		{
			data->Channel = FALSE;
			if ((iface = MacIfaceGet((WORD) obj->Syntax.BufInt[obj->SyntaxLen - 1])) == NULL)
				return SNMP_BADVALUE;
			if (iface->type != MAC_TYPE_ETHERNET_CSMACD && iface->type != MAC_TYPE_88023_CSMACD)
				return SNMP_BADVALUE;
			memcpy(data->Source, obj->Syntax.BufInt, obj->SyntaxLen * sizeof(data->Source[0]));
			data->SourceLen = obj->SyntaxLen;
			data->Iface = iface;
			BooleanSetTrue(data->ObjectSet, MATRIX_BOOLEAN_DATASOURCE);
			return SNMP_NOERROR;
		}
		else if (ChannelOidLen == obj->SyntaxLen &&
				 memcmp(ChannelOid, obj->Syntax.BufInt, (ChannelOidLen - 1) * sizeof(ChannelOid[0])) == 0)
		{
			data->Channel = TRUE;
			memcpy(data->Source, obj->Syntax.BufInt, obj->SyntaxLen * sizeof(data->Source[0]));
			data->SourceLen = obj->SyntaxLen;
			BooleanSetTrue(data->ObjectSet, MATRIX_BOOLEAN_DATASOURCE);
			return SNMP_NOERROR;
		}
		return SNMP_BADVALUE;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixControlTableSize(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_MATRIX_CONTROL *data;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = MibRmon(obj, MatrixControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL, FALSE, &source, &dest, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_MATRIX_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->TableSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixControlLastDelTime(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_MATRIX_CONTROL *data;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = MibRmon(obj, MatrixControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL, FALSE, &source, &dest, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_MATRIX_CONTROL *) local->Data;
		obj->Syntax.LngUns = data->LastDeleteTime;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixControlOwner(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_MATRIX_CONTROL *data;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = MibRmon(obj, MatrixControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL, FALSE, &source, &dest, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_MATRIX_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, MATRIX_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_MATRIX_CONTROL *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, obj->Syntax.BufChr, obj->SyntaxLen);
		data->OwnerLen = obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, MATRIX_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixControlStatus(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_MATRIX_CONTROL *data;
	Host *source = NULL;
	Host *dest = NULL;

	local = MibRmon(obj, MatrixControlInstance, idlen, CONTROLINDEXSIZE);

	switch (obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (DNPAP_MATRIX_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL, FALSE, &source, &dest, NULL) == FALSE)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (DNPAP_MATRIX_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(obj, &MatrixControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(DNPAP_MATRIX_CONTROL))) == NULL)
				{
					MibRemove(obj, &MatrixControlInstance, idlen, CONTROLINDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (DNPAP_MATRIX_CONTROL *) local->Data;
				memset(data, 0, sizeof(DNPAP_MATRIX_CONTROL));
				if (DnpapMatrixCInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, DMATRIX_CREATE, "dmatrix: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(obj, &MatrixControlInstance, idlen, CONTROLINDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (DNPAP_MATRIX_CONTROL *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (DnpapMatrixCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, DMATRIX_START, "dmatrix: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (DnpapMatrixCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, DMATRIX_DESTROY, "dmatrix: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &MatrixControlInstance, idlen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				return SNMP_NOERROR;
			case SNMP_INVALID:
				if (DnpapMatrixCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, DMATRIX_DESTROY, "dmatrix: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &MatrixControlInstance, idlen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (DnpapMatrixCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixSDSourceAddress(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, source->Addr, obj->SyntaxLen = ((DNPAP_MATRIX_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixSDDestAddress(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->ConnHost->Addr, obj->SyntaxLen = ((DNPAP_MATRIX_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixSDIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixSDRanking(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = rank;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixSDPkts(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Stats->Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixSDOctets(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Stats->Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixSDErrors(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Stats->Errors;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixSDConnections(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = source->NumberConnect;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixDSSourceAddress(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->ConnHost->Addr, obj->SyntaxLen = ((DNPAP_MATRIX_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixDSDestAddress(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, dest->Addr, obj->SyntaxLen = ((DNPAP_MATRIX_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixDSIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixDSRanking(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = rank;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixDSPkts(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Stats->Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixDSOctets(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Stats->Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixDSErrors(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Stats->Errors;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapMatrixDSConnections(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchHostObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = dest->NumberConnect;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixSRankTopNSource(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, source->Addr, obj->SyntaxLen = ((DNPAP_MATRIX_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixSRankTopNDest(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->ConnHost->Addr, obj->SyntaxLen = ((DNPAP_MATRIX_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixSRankTopNIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixSRankTopNRanking(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = rank;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixSRankTopNPkts(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Stats->Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixSRankTopNOctets(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Stats->Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixSRankTopNErrors(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Stats->Errors;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixSRankTopNConnections(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, TRUE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, TRUE, &source, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = source->NumberConnect;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixDRankTopNSource(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->ConnHost->Addr, obj->SyntaxLen = ((DNPAP_MATRIX_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixDRankTopNDest(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, dest->Addr, obj->SyntaxLen = ((DNPAP_MATRIX_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixDRankTopNIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixDRankTopNRanking(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = rank;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixDRankTopNPkts(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Stats->Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixDRankTopNOctets(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Stats->Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixDRankTopNErrors(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Stats->Errors;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DMatrixDRankTopNConnections(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ConnInfo *data = NULL;
	INT rank = 0;
	Host *source = NULL;
	Host *dest = NULL;

	if ((local = DnpapMatrixRankMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, RANKINDEXSIZE, &data, FALSE, &source, &dest, &rank) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = DnpapMatrixSearchTopObj(local->Data, obj, idlen, FALSE, &dest, &rank)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = dest->NumberConnect;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

BOOLEAN DnpapMatrixRegisterAll(ULONG channelIndex)
{
	MIB_LOCAL *local;
	DNPAP_MATRIX_CONTROL *data;

	for (local = MatrixControlInstance; local != NULL; local = local->Next)
	{
		data = (DNPAP_MATRIX_CONTROL *) local->Data;
		if (data->Channel == TRUE && data->Source[data->SourceLen - 1] == channelIndex &&
			data->ChannelCallback == NULL)
		{
			data->ChannelCallback = ChannelRegisterSink(data->Source[data->SourceLen - 1],
														DnpapMatrixChannelPkt, DnpapMatrixDeRegister, data);
		}
	}

	return TRUE;
}

ConnInfo *SrcDstGetFirstNext(DNPAP_MATRIX_CONTROL * matrixcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first, BOOLEAN SDorder, Host ** source, Host ** dest, INT * rank)
{
	BYTE srcdst[2 * HOST_SIZE_ADDR];
	BYTE *src, *dst;
	INT i, len;
	ConnInfo *conn = NULL, *conn2 = NULL;
	BOOLEAN nextconn, firstconn;
	WORD addrlen = matrixcontrol->AddressLen;

	*source = *dest = NULL;
	*rank = 0;

	if (matrixcontrol->TableSize == 0)
		return NULL;

	nextconn = FALSE;
	firstconn = FALSE;

	if (first == FALSE)
	{
		if (obj->IdLen > idlen + 1)
		{
			if (obj->Id[idlen + 1] > addrlen)
				return NULL;
			if (obj->Id[idlen + 1] < addrlen)
				first = TRUE;
		}
		else
			first = TRUE;

		if (obj->IdLen > idlen + 1 + addrlen + 1)
		{
			if (obj->Id[idlen + 1 + addrlen + 1] > addrlen)
				nextconn = TRUE;
			if (obj->Id[idlen + 1 + addrlen + 1] < addrlen)
				firstconn = TRUE;
		}
	}

	if (first == TRUE)
	{
		if (SDorder == TRUE)
		{
			*source = matrixcontrol->SrcList;
			conn = (*source)->LexList;
			if (conn == NULL)			/* however, there should always be one */
				return NULL;
			*rank = DMatrixListSearchRank(*source, conn->ConnHost->Addr, &conn2, addrlen);
			if (*rank == 0)				/* however, there should always be one */
				return NULL;

			for (i = 0; i < addrlen; i++)
			{
				obj->Id[idlen + 2 + i] = (LONG) (*source)->Addr[i];
				obj->Id[idlen + 3 + i + addrlen] = (LONG) conn->ConnHost->Addr[i];
			}
			obj->Id[idlen + 1] = addrlen;
			obj->Id[idlen + 1 + addrlen + 1] = addrlen;
			obj->IdLen = idlen + 1 + 2 + 2 * addrlen;
		}
		else
		{
			*dest = matrixcontrol->DstList;
			conn = (*dest)->LexList;
			if (conn == NULL)			/* however, there should always be one */
				return NULL;
			*rank = DMatrixListSearchRank(*dest, conn->ConnHost->Addr, &conn2, addrlen);
			if (*rank == 0)				/* however, there should always be one */
				return NULL;

			for (i = 0; i < addrlen; i++)
			{
				obj->Id[idlen + 2 + i] = (LONG) (*dest)->Addr[i];
				obj->Id[idlen + 3 + i + addrlen] = (LONG) conn->ConnHost->Addr[i];
			}
			obj->Id[idlen + 1] = addrlen;
			obj->Id[idlen + 1 + addrlen + 1] = addrlen;
			obj->IdLen = idlen + 1 + 2 + 2 * addrlen;
		}

		return conn;
	}

	len = obj->IdLen - idlen - 2;
	if (len < 0)
		len = 0;
	if (len > addrlen)
		len--;
	if (len > 2 * addrlen)
		len = 2 * addrlen;

	/* fill in srcdst in both orders: SDorder or reverse  */
	for (i = 0; i < len; i++)
		srcdst[i] = (BYTE) obj->Id[idlen + 2 + (i < addrlen ? 0 : 1) + i];

	if (firstconn == TRUE)
		len = addrlen;					/* force finding of first connection */
	if (nextconn == TRUE)
	{
		len = addrlen;
		/* force finding of next connection */
		for (i = addrlen - 1; i >= 0; i--)
			if (++srcdst[i] != 0)
				break;
		if (i < 0 && srcdst[0] == 0)
			return NULL;
	}

	if (SDorder == TRUE)
	{
		src = srcdst;
		dst = srcdst + addrlen;

		if (len == 0)
		{
			*source = matrixcontrol->SrcList;
			conn = (*source)->LexList;
		}
		else if (len > 0 && len < addrlen)
		{
			*source = DnpapMatrixSearchHost(matrixcontrol, src, len, TRUE);
			if (*source == NULL)
				return NULL;
			conn = (*source)->LexList;
		}
		else if (len >= addrlen && len <= 2 * addrlen)
		{
			*source = HashSearch(matrixcontrol->SrcTable, src, addrlen);
			if (*source == NULL)
			{
				if ((*source = DnpapMatrixSearchHost(matrixcontrol, src, addrlen, TRUE)) == NULL)
					return NULL;
				conn = (*source)->LexList;
			}
			else
			{
				conn = DMatrixListSearchAddr(*source, dst, len - addrlen);
				if (len == 2 * addrlen)
				{
					if (conn != NULL && memcmp(conn->ConnHost->Addr, dst, addrlen) == 0)
						conn = conn->LexNext;
					if (conn == NULL)
					{
						*source = (*source)->Next;
						if (*source != NULL)
							conn = (*source)->LexList;
					}
				}
			}
		}

		if (conn != NULL)
		{
			for (i = 0; i < addrlen; i++)
			{
				obj->Id[idlen + 2 + i] = (*source)->Addr[i];
				obj->Id[idlen + 3 + i + addrlen] = conn->ConnHost->Addr[i];
			}

			obj->Id[idlen + 1] = addrlen;
			obj->Id[idlen + 1 + addrlen + 1] = addrlen;
			obj->IdLen = idlen + 1 + 2 + 2 * addrlen;

			*rank = DMatrixListSearchRank(*source, conn->ConnHost->Addr, &conn2, addrlen);
		}
	}
	else
	{
		src = srcdst + addrlen;
		dst = srcdst;

		if (len == 0)
		{
			*dest = matrixcontrol->DstList;
			conn = (*dest)->LexList;
		}
		else if (len > 0 && len < addrlen)
		{
			*dest = DnpapMatrixSearchHost(matrixcontrol, dst, len, FALSE);
			if (*dest == NULL)
				return NULL;
			conn = (*dest)->LexList;
		}
		else if (len >= addrlen && len <= 2 * addrlen)
		{
			*dest = HashSearch(matrixcontrol->DstTable, dst, addrlen);
			if (*dest == NULL)
			{
				if ((*dest = DnpapMatrixSearchHost(matrixcontrol, dst, addrlen, FALSE)) == NULL)
					return NULL;
				conn = (*dest)->LexList;
			}
			else
			{
				conn = DMatrixListSearchAddr(*dest, src, len - addrlen);
				if (len == 2 * addrlen)
				{
					if (conn != NULL && memcmp(conn->ConnHost->Addr, src, addrlen) == 0)
						conn = conn->LexNext;
					if (conn == NULL)
					{
						*dest = (*dest)->Next;
						if (*dest != NULL)

							conn = (*dest)->LexList;
					}
				}
			}
		}

		if (conn != NULL)
		{
			for (i = 0; i < addrlen; i++)
			{
				obj->Id[idlen + 2 + i] = (*dest)->Addr[i];
				obj->Id[idlen + 3 + i + addrlen] = conn->ConnHost->Addr[i];
			}

			obj->Id[idlen + 1] = addrlen;
			obj->Id[idlen + 1 + addrlen + 1] = addrlen;
			obj->IdLen = idlen + 1 + 2 + 2 * addrlen;

			*rank = DMatrixListSearchRank(*dest, conn->ConnHost->Addr, &conn2, addrlen);
		}
	}

	return conn;
}

ConnInfo *GetTopFirstNext(DNPAP_MATRIX_CONTROL * matrixcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first, BOOLEAN SDorder, Host ** source, Host ** dest, INT * rank)
{
	BYTE addr[HOST_SIZE_ADDR];
	INT i, len;
	ConnInfo *conn = NULL;
	WORD addrlen = matrixcontrol->AddressLen;

	*source = *dest = NULL;
	*rank = 0;

	if (matrixcontrol->TableSize == 0)
		return NULL;

	if (first == FALSE)
	{
		if (obj->IdLen > idlen + 1)
		{
			if (obj->Id[idlen + 1] > addrlen)
				return NULL;
			if (obj->Id[idlen + 1] < addrlen)
				first = TRUE;
		}
		else
			first = TRUE;
	}

	if (first == TRUE)
	{
		if (SDorder == TRUE)
		{
			*source = matrixcontrol->SrcList;
			*rank = 1;
			conn = (*source)->TopList;
			if (conn == NULL)
				return NULL;

			for (i = 0; i < addrlen; i++)
				obj->Id[idlen + 2 + i] = (LONG) (*source)->Addr[i];
			obj->Id[idlen + 1] = addrlen;
			obj->Id[idlen + 2 + addrlen] = *rank;
			obj->IdLen = idlen + 1 + 1 + addrlen + 1;
		}
		else
		{
			*dest = matrixcontrol->DstList;
			*rank = 1;
			conn = (*dest)->TopList;
			if (conn == NULL)
				return NULL;

			for (i = 0; i < addrlen; i++)
				obj->Id[idlen + 2 + i] = (LONG) (*dest)->Addr[i];
			obj->Id[idlen + 1] = addrlen;
			obj->Id[idlen + 2 + addrlen] = *rank;
			obj->IdLen = idlen + 1 + 1 + addrlen + 1;
		}

		return conn;
	}

	len = obj->IdLen - idlen - 2;

	if (len < 0)
		len = 0;
	else if (len > addrlen + 1)
		len = addrlen + 1;

	for (i = 0; i < (len > addrlen ? addrlen : len); i++)
		addr[i] = (BYTE) obj->Id[idlen + 2 + i];

	if (SDorder == TRUE)
	{
		if (len == 0)
		{
			*source = matrixcontrol->SrcList;
			*rank = 1;
			conn = (*source)->TopList;
		}
		else if (len > 0 && len < addrlen)
		{
			*source = DnpapMatrixSearchHost(matrixcontrol, addr, len, TRUE);
			if (*source == NULL)
				return NULL;
			*rank = 1;
			conn = (*source)->TopList;
		}
		else if (len == addrlen)
		{
			*source = HashSearch(matrixcontrol->SrcTable, addr, addrlen);
			if (*source == NULL)
			{
				if ((*source = DnpapMatrixSearchHost(matrixcontrol, addr, addrlen, TRUE)) == NULL)
					return NULL;
			}
			*rank = 1;
			conn = (*source)->TopList;
		}
		else if (len == addrlen + 1)
		{
			*source = HashSearch(matrixcontrol->SrcTable, addr, addrlen);
			if (*source == NULL)
			{
				if ((*source = DnpapMatrixSearchHost(matrixcontrol, addr, addrlen, TRUE)) == NULL)
					return NULL;
				*rank = 1;
				conn = (*source)->TopList;
			}
			else
			{
				*rank = obj->Id[idlen + 2 + addrlen];
				conn = DMatrixListSearchTop(*source, *rank);
				if (conn != NULL)
				{
					conn = conn->TopNext;
					(*rank)++;
				}
				if (conn == NULL)
				{
					*source = (*source)->Next;
					if (*source != NULL)
					{
						conn = (*source)->TopList;
						*rank = 1;
					}
				}
			}
		}

		if (conn != NULL)
		{
			for (i = 0; i < addrlen; i++)
				obj->Id[idlen + 2 + i] = (LONG) (*source)->Addr[i];
			obj->Id[idlen + 2 + addrlen] = *rank;

			obj->Id[idlen + 1] = addrlen;
			obj->IdLen = idlen + 1 + 1 + addrlen + 1;
		}
	}
	else
	{
		if (len == 0)
		{
			*dest = matrixcontrol->DstList;
			*rank = 1;
			conn = (*dest)->TopList;
		}
		else if (len > 0 && len < addrlen)
		{
			*dest = DnpapMatrixSearchHost(matrixcontrol, addr, len, FALSE);
			if (*dest == NULL)
				return NULL;
			*rank = 1;
			conn = (*dest)->TopList;
		}
		else if (len == addrlen)
		{
			*dest = HashSearch(matrixcontrol->DstTable, addr, addrlen);
			if (*dest == NULL)
			{
				if ((*dest = DnpapMatrixSearchHost(matrixcontrol, addr, addrlen, FALSE)) == NULL)
					return NULL;
			}
			*rank = 1;
			conn = (*dest)->TopList;
		}
		else if (len == addrlen + 1)
		{
			*dest = HashSearch(matrixcontrol->DstTable, addr, addrlen);
			if (*dest == NULL)
			{
				if ((*dest = DnpapMatrixSearchHost(matrixcontrol, addr, addrlen, FALSE)) == NULL)
					return NULL;
				*rank = 1;
				conn = (*dest)->TopList;
			}
			else
			{
				*rank = obj->Id[idlen + 2 + addrlen];
				conn = DMatrixListSearchTop(*dest, *rank);
				if (conn != NULL)
				{
					conn = conn->TopNext;
					(*rank)++;
				}
				if (conn == NULL)
				{
					*dest = (*dest)->Next;
					if (*dest != NULL)
					{
						conn = (*dest)->TopList;
						*rank = 1;
					}
				}
			}
		}

		if (conn != NULL)
		{
			for (i = 0; i < addrlen; i++)
				obj->Id[idlen + 2 + i] = (LONG) (*dest)->Addr[i];
			obj->Id[idlen + 2 + addrlen] = *rank;

			obj->Id[idlen + 1] = addrlen;
			obj->IdLen = idlen + 1 + 1 + addrlen + 1;
		}
	}

	return conn;
}
