/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <math.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <snmp.h>
#include <mibsup.h>
#include <message.h>
#include <mac.h>

#include "dstate.h"
#include "dstatc.h"
#include "dstat.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define INDEXSIZE      		1
#define IFTINDEXSIZE     	2
#define LENGTHINDEXSIZE     3			/* not the real size, only needed for switch */
#define TYPEINDEXSIZE     	4			/* not the real size, only needed for switch */

static MIB_LOCAL *etherStats = NULL;

static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, TypeEntry ** entry);

BOOLEAN DnpapEtherStatsMInit(VOID)
{
	MessageConfig(DNPAP_STATISTICS_ERROR, "DnpapEtherStat");
	return TRUE;
}

#define CONTROL_DATA    ((DNPAP_ETHER_STATS *) (*local)->Data)

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, TypeEntry ** entry)
{
	DNPAP_ETHER_STATS *data;
	BOOLEAN first = FALSE;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case TYPEINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + 2)
		{
			data = CONTROL_DATA;
			if (data->Status == SNMP_VALID &&
				data->TypeBuckets > 0 && obj->Id[idlen + 1] < data->TypeLast->Type)
			{
				if ((*entry = DnpapNextTypeEntry(data, (WORD) obj->Id[idlen + 1])) == NULL)
					return FALSE;
				obj->Id[idlen] = (*local)->Index;
				obj->Id[idlen + 1] = (*entry)->Type;
				obj->IdLen = idlen + 2;
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL &&
			   (CONTROL_DATA->Status != SNMP_VALID ||
				CONTROL_DATA->TypeBuckets == 0))
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen + 1] = ((DNPAP_ETHER_STATS *) (*local)->Data)->TypeList->Type;
		obj->IdLen = idlen + 2;
		break;
	case LENGTHINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + 2)
		{
			data = CONTROL_DATA;
			if (data->Status == SNMP_VALID && obj->Id[idlen + 1] < data->LengthBuckets)
			{
				obj->Id[idlen] = (*local)->Index;
				obj->Id[idlen + 1]++;
				obj->IdLen = idlen + 2;
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL &&
			   (CONTROL_DATA->Status != SNMP_VALID ||
				CONTROL_DATA->LengthBuckets == 0))
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen + 1] = 1;
		obj->IdLen = idlen + 2;
		break;
	case IFTINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			data = CONTROL_DATA;
			if (data->Status == SNMP_VALID && obj->Id[idlen + 1] < data->IFTBuckets)
			{
				obj->Id[idlen] = (*local)->Index;
				obj->Id[idlen + 1]++;
				obj->IdLen = idlen + IFTINDEXSIZE;
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL &&
			   (CONTROL_DATA->Status != SNMP_VALID ||
				CONTROL_DATA->IFTBuckets == 0))
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen + 1] = 1;
		obj->IdLen = idlen + IFTINDEXSIZE;
		break;
	case INDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + INDEXSIZE;
		break;
	}
	return TRUE;
}

WORD DnpapEtherStatsIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;
	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsDataSource(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_IFACE *iface;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, STAT_BOOLEAN_DATASOURCE))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufInt, data->Source, data->SourceLen * sizeof(LONG));
		Obj->SyntaxLen = data->SourceLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (data->SourceLen != Obj->SyntaxLen ||
			memcmp(data->Source, Obj->Syntax.BufInt, (data->SourceLen - 1) * sizeof(data->Source[0])))
			return SNMP_BADVALUE;
		if ((iface = MacIfaceGet((WORD) Obj->Syntax.BufInt[Obj->SyntaxLen - 1])) == NULL)
			return SNMP_BADVALUE;
		if (iface->type != MAC_TYPE_ETHERNET_CSMACD && iface->type != MAC_TYPE_88023_CSMACD)
			return SNMP_BADVALUE;
		memcpy(data->Source, Obj->Syntax.BufInt, Obj->SyntaxLen * sizeof(data->Source[0]));
		data->SourceLen = Obj->SyntaxLen;
		data->Iface = iface;
		BooleanSetTrue(data->ObjectSet, STAT_BOOLEAN_DATASOURCE);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsDropEvents(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.LostPkts + stat.DiscardedPkts - (data->BaseStat.LostPkts + data->BaseStat.DiscardedPkts);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsBroadcastPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.BroadcastPkts - data->BaseStat.BroadcastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsMulticastPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.MulticastPkts - data->BaseStat.MulticastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsCRCAlignErrors(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.CRCAlignErrors - data->BaseStat.CRCAlignErrors;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsUndersizePkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.UndersizePkts - data->BaseStat.UndersizePkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsOversizePkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.OversizePkts - data->BaseStat.OversizePkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsFragments(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.Fragments - data->BaseStat.Fragments;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsJabbers(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.Jabbers - data->BaseStat.Jabbers;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

#ifdef REALLY_IFMIB_SPECIFIC
WORD DnpapEtherStatsInPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.inPkts - data->BaseStat.inPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsInErrors(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.inErrors - data->BaseStat.inErrors;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsOutPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.outPkts - data->BaseStat.outPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsOutErrors(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.outErrors - data->BaseStat.outErrors;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
#endif

WORD DnpapEtherStatsCollisions(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.Collisions - data->BaseStat.Collisions;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts64(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts64Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts65to127(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts65to127Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts128to255(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts128to255Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts256to511(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts256to511Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts512to1023(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts512to1023Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsPkts1024to1518(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts1024to1518Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsUtilInterval(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->UtilInterval;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->Syntax.LngUns < 1)
			return SNMP_BADVALUE;
		data->UtilInterval = Obj->Syntax.LngUns;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsUtilization(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngInt = data->Utilization;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsIFTBuckets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	ULONG *buckets1, *buckets2;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngInt = data->IFTBuckets;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->Syntax.LngInt < 1)
			return SNMP_BADVALUE;
		if ((buckets1 = ReNewBuckets(data->IFTArrayPkts, Obj->Syntax.LngInt)) == NULL)
			return SNMP_BADVALUE;
		if ((buckets2 = ReNewBuckets(data->IFTArrayOctets, Obj->Syntax.LngInt)) == NULL)
			return SNMP_BADVALUE;
		data->IFTBuckets = Obj->Syntax.LngInt;
		data->IFTArrayPkts = buckets1;
		data->IFTArrayOctets = buckets2;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsIFTSize(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngInt = data->IFTSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->Syntax.LngInt < 1)
			return SNMP_BADVALUE;
		data->IFTSize = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsLengthBuckets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	ULONG *buckets1, *buckets2;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngInt = data->LengthBuckets;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->Syntax.LngInt < 1 || Obj->Syntax.LngInt > MAC_FRAME_ETHERNET_CSMACD)
			return SNMP_BADVALUE;
		if ((buckets1 = ReNewBuckets(data->LengthArrayPkts, Obj->Syntax.LngInt)) == NULL)
			return SNMP_BADVALUE;
		if ((buckets2 = ReNewBuckets(data->LengthArrayOctets, Obj->Syntax.LngInt)) == NULL)
			return SNMP_BADVALUE;
		data->LengthBuckets = Obj->Syntax.LngInt;
		data->LengthSize = ceil(((FLOAT) MAC_FRAME_ETHERNET_CSMACD) / Obj->Syntax.LngInt);
		data->LengthArrayPkts = buckets1;
		data->LengthArrayOctets = buckets2;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsLengthSize(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	INT buckets;
	ULONG *buckets1, *buckets2;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngInt = data->LengthSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->Syntax.LngInt < 1 || Obj->Syntax.LngInt > MAC_FRAME_ETHERNET_CSMACD)
			return SNMP_BADVALUE;
		buckets = ceil(((FLOAT) MAC_FRAME_ETHERNET_CSMACD) / Obj->Syntax.LngInt);
		if ((buckets1 = ReNewBuckets(data->LengthArrayPkts, buckets)) == NULL)
			return SNMP_BADVALUE;
		if ((buckets2 = ReNewBuckets(data->LengthArrayOctets, buckets)) == NULL)
			return SNMP_BADVALUE;
		data->LengthSize = Obj->Syntax.LngInt;
		data->LengthBuckets = buckets;
		data->LengthArrayPkts = buckets1;
		data->LengthArrayOctets = buckets2;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsTypeBuckets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngInt = data->TypeBuckets;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsOwner(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, STAT_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		Obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (DNPAP_ETHER_STATS *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, Obj->Syntax.BufChr, Obj->SyntaxLen);
		data->OwnerLen = Obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, STAT_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsStatus(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE);

	switch (Obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE, NULL) == FALSE)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (DNPAP_ETHER_STATS *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(Obj, &etherStats, IdLen, INDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(DNPAP_ETHER_STATS))) == NULL)
				{
					MibRemove(Obj, &etherStats, IdLen, INDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (DNPAP_ETHER_STATS *) local->Data;
				memset(data, 0, sizeof(DNPAP_ETHER_STATS));
				if (DnpapStatCInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, DSTAT_CREATE, "detherstat: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(Obj, &etherStats, IdLen, INDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (DNPAP_ETHER_STATS *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (DnpapStatCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, DSTAT_ACTIVE, "detherstat: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				else
					return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (DnpapStatCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, DSTAT_DESTROY, "detherstat: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &etherStats, IdLen, INDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_INVALID:
				if (DnpapStatCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, DSTAT_DESTROY, "detherstat: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &etherStats, IdLen, INDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (DnpapStatCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			case SNMP_VALID:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsIFTIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, IFTINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, IFTINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID ||
			Obj->Id[IdLen + 1] < 1 || Obj->Id[IdLen + 1] > data->IFTBuckets)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsIFTBucket(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, IFTINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, IFTINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID ||
			Obj->Id[IdLen + 1] < 1 || Obj->Id[IdLen + 1] > data->IFTBuckets)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = Obj->Id[IdLen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsIFTPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, IFTINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, IFTINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID ||
			Obj->Id[IdLen + 1] < 1 || Obj->Id[IdLen + 1] > data->IFTBuckets)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->IFTArrayPkts[Obj->Id[IdLen + 1] - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsIFTOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, IFTINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, IFTINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID ||
			Obj->Id[IdLen + 1] < 1 || Obj->Id[IdLen + 1] > data->IFTBuckets)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->IFTArrayOctets[Obj->Id[IdLen + 1] - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsLengthIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, LENGTHINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, LENGTHINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID ||
			Obj->Id[IdLen + 1] < 1 || Obj->Id[IdLen + 1] > data->LengthBuckets)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsLengthBucket(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, LENGTHINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, LENGTHINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID ||
			Obj->Id[IdLen + 1] < 1 || Obj->Id[IdLen + 1] > data->LengthBuckets)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = Obj->Id[IdLen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsLengthPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, LENGTHINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, LENGTHINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID ||
			Obj->Id[IdLen + 1] < 1 || Obj->Id[IdLen + 1] > data->LengthBuckets)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->LengthArrayPkts[Obj->Id[IdLen + 1] - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsLengthOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, LENGTHINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, LENGTHINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID ||
			Obj->Id[IdLen + 1] < 1 || Obj->Id[IdLen + 1] > data->LengthBuckets)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->LengthArrayOctets[Obj->Id[IdLen + 1] - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsTypeIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	TypeEntry *entry = NULL;

	if ((local = MibRmon(Obj, etherStats, IdLen, TYPEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TYPEINDEXSIZE, &entry) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (entry == NULL && (entry = DnpapTypeEntrySearch(data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsTypeType(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	TypeEntry *entry = NULL;

	if ((local = MibRmon(Obj, etherStats, IdLen, TYPEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TYPEINDEXSIZE, &entry) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (entry == NULL && (entry = DnpapTypeEntrySearch(data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = Obj->Id[IdLen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsTypePkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	TypeEntry *entry = NULL;

	if ((local = MibRmon(Obj, etherStats, IdLen, TYPEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TYPEINDEXSIZE, &entry) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (entry == NULL && (entry = DnpapTypeEntrySearch(data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = entry->Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD DnpapEtherStatsTypeOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	DNPAP_ETHER_STATS *data;
	TypeEntry *entry = NULL;

	if ((local = MibRmon(Obj, etherStats, IdLen, TYPEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TYPEINDEXSIZE, &entry) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (DNPAP_ETHER_STATS *) local->Data;
		if (entry == NULL && (entry = DnpapTypeEntrySearch(data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = entry->Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
