/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>
#include <maxmem.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <message.h>
#include <mibsup.h>
#include <dmath.h>

#include "statbufd.h"
#include "statbufe.h"
#include "statbufc.h"
#include "statbuf.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define CONTROLINDEXSIZE    1
#define ENTRYINDEXSIZE      2

extern SHORT StatBufMaxNrEntries;

static MIB_LOCAL *StatBufControlInstance = NULL;

static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, StatBuf ** elem);

BOOLEAN StatBufMInit(VOID)
{
	MessageConfig(STATBUF_ERROR, "StatBuffer");
	return StatBufConfigInit();
}

#define CONTROL_DATA    ((StatBufControl*)(*local)->Data)

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, StatBuf ** elem)
{
	StatBufControl *data;
	BOOLEAN first = FALSE;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case ENTRYINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			data = CONTROL_DATA;
			if ((data->BufferWrapped == TRUE && obj->Id[idlen + 1] < data->GrantedSize) ||
				(data->BufferWrapped == FALSE && obj->Id[idlen + 1] < data->BufferPos))
			{							/* BufferWrapped == FALSE and BufferPos == 0 if Status != VALID  */
				obj->Id[idlen] = (*local)->Index;
				obj->Id[idlen + 1]++;
				obj->IdLen = idlen + ENTRYINDEXSIZE;
				*elem = StatBufSearch(CONTROL_DATA, obj, idlen);
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL &&
			   (CONTROL_DATA->GrantedSize == 0 ||
				(CONTROL_DATA->BufferWrapped == FALSE &&
				 CONTROL_DATA->BufferPos == 0)))
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen + 1] = 1;
		obj->IdLen = idlen + ENTRYINDEXSIZE;
		*elem = StatBufSearch(CONTROL_DATA, obj, idlen);
		break;
	case CONTROLINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + CONTROLINDEXSIZE;
		break;
	}
	return TRUE;
}

WORD StatBufControlIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufControlVariable(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBufControl *data = NULL;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (StatBufControl *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, STATBUF_BOOLEAN_VARIABLE))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufInt, data->Variable, (obj->SyntaxLen = data->VariableLen) * sizeof(data->Variable[0]));
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (StatBufControl *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		memcpy(data->Variable, obj->Syntax.BufInt, (data->VariableLen = obj->SyntaxLen) * sizeof(data->Variable[0]));
		BooleanSetTrue(data->ObjectSet, STATBUF_BOOLEAN_VARIABLE);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufControlRequestedSize(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBufControl *data;
	StatBuf *buffer;
	LONG i, ni, j;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (StatBufControl *) local->Data;
		obj->Syntax.LngInt = data->RequestedSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (StatBufControl *) local->Data;
		if (obj->Syntax.LngInt < 1)
			return SNMP_BADVALUE;
		data->GrantedSize = LongMin(MemoryGetAvail(), MemoryGetMaxChunk()) / sizeof(StatBuf);
		if (data->GrantedSize > LongMin(obj->Syntax.LngInt, StatBufMaxNrEntries))
			data->GrantedSize = LongMin(obj->Syntax.LngInt, StatBufMaxNrEntries);
		if (data->GrantedSize == data->BufferAllocated)
		{
			data->RequestedSize = obj->Syntax.LngInt;
			return SNMP_NOERROR;
		}
		if ((buffer = NewStatBuf(data->GrantedSize)) == NULL)
		{
			DelStatBuf(data->Buffer);
			data->Buffer = NULL;
			data->BufferPos = 0;
			data->GrantedSize = 0;
			data->BufferAllocated = 0;
			data->BufferWrapped = FALSE;
		}
		else
		{
			if (data->BufferWrapped == TRUE)
				ni = LongMin(data->GrantedSize, data->BufferAllocated);
			else
				ni = LongMin(data->GrantedSize, data->BufferPos);
			for (i = 0; i < ni; i++)
			{
				j = data->BufferPos - ni + i;
				if (j < 0)
					j = j + data->BufferAllocated;
				memcpy(buffer + i, data->Buffer + j, sizeof(StatBuf));
			}
			DelStatBuf(data->Buffer);
			data->Buffer = buffer;
			data->BufferPos = ni;
			data->BufferAllocated = data->GrantedSize;
			data->BufferWrapped = FALSE;
		}
		data->RequestedSize = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufControlGrantedSize(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBufControl *data;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (StatBufControl *) local->Data;
		obj->Syntax.LngInt = data->GrantedSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufControlInterval(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBufControl *data = NULL;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (StatBufControl *) local->Data;
		obj->Syntax.LngInt = data->Interval;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (StatBufControl *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt <= 0 || obj->Syntax.LngInt > 2000000L)
			return SNMP_BADVALUE;
		data->Interval = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufControlSampleType(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBufControl *data = NULL;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (StatBufControl *) local->Data;
		obj->Syntax.LngInt = data->SampleType;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (StatBufControl *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt != STATBUF_ABSOLUTE && obj->Syntax.LngInt != STATBUF_DELTA)
			return SNMP_BADVALUE;
		data->SampleType = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufControlFullStatus(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBufControl *data;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (StatBufControl *) local->Data;
		obj->Syntax.LngInt = data->FullStatus;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufControlFullAction(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBufControl *data = NULL;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (StatBufControl *) local->Data;
		obj->Syntax.LngInt = data->FullAction;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (StatBufControl *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt != STATBUF_LOCK && obj->Syntax.LngInt != STATBUF_WRAP)
			return SNMP_BADVALUE;
		data->FullAction = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufControlAverage(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBufControl *data;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (StatBufControl *) local->Data;
		obj->Syntax.LngInt = data->Average;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufControlDeviation(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBufControl *data;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (StatBufControl *) local->Data;
		obj->Syntax.LngInt = data->Deviation;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufControlOwner(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBufControl *data = NULL;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (StatBufControl *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, STATBUF_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Owner, obj->SyntaxLen = data->OwnerLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (StatBufControl *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, obj->Syntax.BufChr, data->OwnerLen = obj->SyntaxLen);
		BooleanSetTrue(data->ObjectSet, STATBUF_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufControlStatus(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBufControl *data;

	local = MibRmon(obj, StatBufControlInstance, idlen, 1);

	switch (obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (StatBufControl *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1, NULL) == FALSE)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (StatBufControl *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(obj, &StatBufControlInstance, idlen, 1)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = NewStatBufControl(1)) == NULL)
				{
					MibRemove(obj, &StatBufControlInstance, idlen, 1);
					return SNMP_GENERROR;
				}
				data = (StatBufControl *) local->Data;
				if (StatBufCInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, STATBUF_CREATE, "statbufcontrol: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DelStatBufControl(local->Data);
				MibRemove(obj, &StatBufControlInstance, idlen, 1);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (StatBufControl *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (StatBufCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, STATBUF_ACTIVE, "statbufcontrol: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (StatBufCStop(data) == TRUE)
				{
					DelStatBufControl(local->Data);
					DnpapMessage(DMC_MESSAGE, STATBUF_DESTROY, "statbufcontrol: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &StatBufControlInstance, idlen, 1) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_INVALID:
				if (StatBufCStop(data) == TRUE)
				{
					DelStatBufControl(local->Data);
					DnpapMessage(DMC_MESSAGE, STATBUF_DESTROY, "statbufcontrol: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &StatBufControlInstance, idlen, 1) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (StatBufCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			case SNMP_VALID:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufLogIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBuf *entry = NULL;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, ENTRYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, ENTRYINDEXSIZE, &entry) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (entry == NULL && (entry = StatBufSearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = obj->Id[idlen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBuf *entry = NULL;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, ENTRYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, ENTRYINDEXSIZE, &entry) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (entry == NULL && (entry = StatBufSearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufTime(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBuf *entry = NULL;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, ENTRYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, ENTRYINDEXSIZE, &entry) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (entry == NULL && (entry = StatBufSearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = entry->Time;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD StatBufValue(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	StatBuf *entry = NULL;

	if ((local = MibRmon(obj, StatBufControlInstance, idlen, ENTRYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, ENTRYINDEXSIZE, &entry) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (entry == NULL && (entry = StatBufSearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = entry->Value;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
