/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <agent.h>
#include <snmp.h>
#include <dnpap.h>
#include <config.h>
#include <message.h>
#include <dmath.h>
#include <dstring.h>

#include "mibsup.h"
#include "mibrqs.h"


MIB_RMON Mib =  {
                    NULL,
                    0,
                    NULL,
                    0,
                };


static MIB_ROOT *MibRoot;


static VOID MibExit(VOID);
static INT MibCmpProfil(MIB_COMMUNITY *Prf1, MIB_COMMUNITY *Prf2);
static INT MibCmpComm(BYTE *comm1, WORD comm1len, BYTE *comm2, WORD comm2len);
static INT MibCmpObject(MIB_ELEMENT *Obj1, MIB_ELEMENT *Obj2);
static INT MibCmpObjId(ULONG *ObjId, WORD ObjIdLen1, ULONG *ObjId2, WORD ObjIdLen2);
static INT MibObjectFind(ULONG *Id, INT IdLen, INT *cmp);
static BOOLEAN MibObjectInsert(MIB_OBJECT *Object);
static BOOLEAN MibObjectDelete(MIB_OBJECT *Object);

static WORD Request (SNMP_OBJECT *Obj, INT *lastindex);


BOOLEAN MibInit (MIB_ELEMENT *mib, WORD mibsize)
{
INT commnr, i, n;
CHAR configname[81];
CHAR *access;
AGENT_COMM *comm;
WORD support;
MIB_COMMUNITY *mibprf, *p;

	MessageConfig(MIB_ERROR, "Mib");

	mibprf = NULL;
	commnr = 0;
	sprintf(configname, "mib.access.%d", n = commnr+1);
	while (ConfigGetString(configname, &access) == TRUE)
	{
		if ((comm = AgentCommunity(n)) == NULL)
		{
			DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: no community %d defined in agent", n);
			break;
		}

		if (stricmp(access, "read-only") == 0)
			support = MIB_READ;
		else
		if (stricmp(access, "write-only") == 0)
			support = MIB_WRITE;
		else
		if (stricmp(access, "read-write") == 0)
			support = MIB_READ | MIB_WRITE;
		else
		{
			DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: invalid access specified");
			break;
		}

		if ((p = DnpapRealloc(mibprf, n*sizeof(MIB_COMMUNITY))) == NULL)
		{
			DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: out of community name space");
			break;
		}

		mibprf = p;
		memcpy(mibprf[commnr].Comm, comm->comm, comm->commLen);
		mibprf[commnr].CommLen = comm->commLen;
		mibprf[commnr].Support = support;
		commnr++;

		sprintf(configname, "mib.access.%d", n = commnr+1);
	}

	if (commnr == 0)
	{
		DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: no external access possible");
		return FALSE;
	}

	Mib.Prf = mibprf;
    Mib.PrfSze = commnr;

    Mib.Obj = mib;
    Mib.ObjSze = mibsize;

    qsort (Mib.Prf, Mib.PrfSze, sizeof(MIB_COMMUNITY), MibCmpProfil);
    qsort (Mib.Obj, Mib.ObjSze, sizeof(MIB_ELEMENT), MibCmpObject);

	DnpapAtExit(MibExit);

    MibRoot = (MIB_ROOT *)DnpapMalloc(sizeof(MIB_ROOT));
    MibRoot->Size = 64;
    MibRoot->Count = 0;
    MibRoot->Table = (MIB_OBJECT **)DnpapMalloc(MibRoot->Size * sizeof(MIB_OBJECT *));

    for (i=0; i<mibsize; i++)
    {
        if (!MibRegister(mib[i].Id,
                         mib[i].IdLen,
                         mib[i].Rqs,
                         mib[i].Type,
                         mib[i].Support,
                         0))
		{
			DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: register failed");
			break;
		}
    }

    return TRUE;
}


VOID MibExit(VOID)
{
}


MIB_OBJECT *MibRegister(LONG         *Id,
                        WORD         IdLen,
                        MIB_CALLBACK Rqs,
                        WORD         Type,
                        WORD         Support,
                        VOID         *Param)
{
    MIB_OBJECT *object;

    object = DnpapMalloc(sizeof(MIB_OBJECT));
    if (object == 0)
        return 0;
    object->Id = DnpapMalloc(IdLen * sizeof(LONG));
    if (object->Id == 0)
        return 0;
    memcpy(object->Id, Id, IdLen * sizeof(LONG));
    object->IdLen = IdLen;
    object->Rqs = Rqs;
    object->Type = Type;
    object->Support = Support;
	object->Param = Param;

    if (!MibObjectInsert(object))
    {
        DnpapFree(object->Id);
        DnpapFree(object);
        return 0;
    }

    return object;
}


VOID MibDeRegister(MIB_OBJECT *object)
{
    MibObjectDelete(object);
    DnpapFree(object->Id);
    DnpapFree(object);
}


WORD MibProfil (BYTE *comm, WORD commLen)
{
    INT first, last, index, cmp;

    first = 0;
    last = Mib.PrfSze-1;
	cmp = 1;

	if (last < 0)
		return 0;

    while (first < last)
    {
        index = (first + last)/2;
        cmp = MibCmpComm(Mib.Prf[index].Comm, Mib.Prf[index].CommLen, 
							comm, commLen);
		switch (cmp)
		{
		case -1:
            first = index + 1;
			break;
		case 0:
			first = last = index;
    		return Mib.Prf[index].Support;
		case 1:
            last = index;
			break;
		}
    }

	index = first;

    cmp = MibCmpComm(Mib.Prf[index].Comm, Mib.Prf[index].CommLen, 
						comm, commLen);

    return cmp == 0 ? Mib.Prf[index].Support : 0;
}


/* MibRequest allows for different requests per varbind list */

WORD MibRequest (UINT listLen, SNMP_OBJECT *list, WORD *errindex)
{
WORD status, index, errstatus;
SNMP_OBJECT backup[AGENT_LIST_SIZE];
static INT lastindex[AGENT_LIST_SIZE];

    for (index = 0, status = SNMP_NOERROR;
         index < listLen && status == SNMP_NOERROR;
         index++)
    {
		if (list[index].Request != SNMP_PDU_SET)
			backup[index] = list[index];
		else					/* prevent backup later on */
			backup[index].Request = SNMP_PDU_SET;
		/* object request type in list is set correctly */
        status = Request(list + index, lastindex + index);
    }
	errstatus = status;
	*errindex = index;

	if (errstatus == SNMP_NOERROR)
	{
		for (index = 0, status = SNMP_NOERROR;
			 index < listLen && status == SNMP_NOERROR;
			 index++)
		{
			if (list[index].Request == SNMP_PDU_SET)
			{
				list[index].Request = SNMP_PDU_COMMIT;
				status = Request(list + index, lastindex + index);
			}
		}
		errstatus = status;
		*errindex = index;
	}

	if (errstatus != SNMP_NOERROR)
	{
		for (index = 0, status = SNMP_NOERROR;
			 index < *errindex && status == SNMP_NOERROR;
			 index++)
		{
			if (list[index].Request == SNMP_PDU_SET ||
				list[index].Request == SNMP_PDU_COMMIT)
			{
				list[index].Request = SNMP_PDU_UNDO;
				status = Request(list + index, lastindex + index);
			}
			else
			{
				list[index] = backup[index];
			}
		}
	}

	return errstatus;
}


WORD Request (SNMP_OBJECT *Obj, INT *lastindex)
{
    INT index, cmp;
	WORD error;

	if (MibRoot == NULL)
		return SNMP_GENERROR;

    error = SNMP_NOSUCHNAME;

    cmp = MibCmpObjId (MibRoot->Table[*lastindex]->Id, MibRoot->Table[*lastindex]->IdLen, Obj->Id, Obj->IdLen);

	if (cmp == -2 && Obj->Request == SNMP_PDU_NEXT && 
		++(*lastindex) < MibRoot->Count)
    	cmp = MibCmpObjId (MibRoot->Table[*lastindex]->Id, MibRoot->Table[*lastindex]->IdLen, Obj->Id, Obj->IdLen);

	if (cmp == -1)
		index = *lastindex;
	else
    	index = *lastindex = MibObjectFind(Obj->Id, Obj->IdLen, &cmp);

    switch (Obj->Request)
    {
	case SNMP_PDU_GET:
		if (cmp != -1 || MibRoot->Table[index]->Rqs == NULL)
			return error;
		if (!(MibRoot->Table[index]->Support & MIB_READ))
			return error;
		error = MibRoot->Table[index]->Rqs(Obj, MibRoot->Table[index]->IdLen, MibRoot->Table[index]->Param);
		if (error == SNMP_NOERROR)
			Obj->Type = MibRoot->Table[index]->Type;
		return error;
	case SNMP_PDU_SET:
		if (cmp != -1 || MibRoot->Table[index]->Rqs == NULL)
			return error;
		if (!(MibRoot->Table[index]->Support & MIB_WRITE))
			return error;
		if (MibRoot->Table[index]->Type != Obj->Type)
			return SNMP_BADVALUE;
		if (Obj->Type == SNMP_OBJECTID && Obj->SyntaxLen > SNMP_SIZE_BUFINT)
			return SNMP_BADVALUE;
		if ((Obj->Type == SNMP_OCTETSTR ||
			 Obj->Type == SNMP_DISPLAYSTR ||
			 Obj->Type == SNMP_OPAQUE) &&
			(Obj->SyntaxLen > SNMP_SIZE_BUFCHR))
			return SNMP_BADVALUE;
		error = MibRoot->Table[index]->Rqs(Obj, MibRoot->Table[index]->IdLen, MibRoot->Table[index]->Param);
		return error;
	case SNMP_PDU_NEXT:
		switch (cmp)
		{
		case -2:
			return error;
		case -1:
		case 0:
			if (MibRoot->Table[index]->Support & MIB_READ &&
				MibRoot->Table[index]->Rqs != NULL)
			{
				error = MibRoot->Table[index]->Rqs(Obj, MibRoot->Table[index]->IdLen, MibRoot->Table[index]->Param);
				if (error == SNMP_NOERROR)
				{
					Obj->Type = MibRoot->Table[index]->Type;
					break;
				}
			}
			index++;
			break;
		}
		while (error == SNMP_NOSUCHNAME && index < MibRoot->Count)
		{
			memcpy(Obj->Id, MibRoot->Table[index]->Id,
				   MibRoot->Table[index]->IdLen * sizeof(LONG));
			Obj->IdLen = MibRoot->Table[index]->IdLen;
			if (MibRoot->Table[index]->Support & MIB_READ &&
				MibRoot->Table[index]->Rqs != NULL)
			{
				error = MibRoot->Table[index]->Rqs(Obj, MibRoot->Table[index]->IdLen, MibRoot->Table[index]->Param);
				if (error == SNMP_NOERROR)
				{
					Obj->Type = MibRoot->Table[index]->Type;
					break;
				}
			}
			index++;
		}
		return error;
	case SNMP_PDU_COMMIT:
	case SNMP_PDU_UNDO:
		if (cmp != -1 || MibRoot->Table[index]->Rqs == NULL)
			return error;
		error = MibRoot->Table[index]->Rqs(Obj, MibRoot->Table[index]->IdLen, MibRoot->Table[index]->Param);
		return error;
    }

	return SNMP_GENERROR;
}


BOOLEAN MibSimple (SNMP_OBJECT *Obj, WORD IdLen)
{
    if (Obj->Request != SNMP_PDU_NEXT)
    {
        if (Obj->IdLen == IdLen + 1 && Obj->Id[IdLen] == 0)
            return TRUE;
        return FALSE;
    }
    if (Obj->IdLen == IdLen)
    {
        Obj->Id[IdLen] = 0;
        Obj->IdLen++;
        return TRUE;
    }
    return FALSE;
}


MIB_LOCAL * MibRmon (SNMP_OBJECT *Obj, MIB_LOCAL *Local, WORD IdLen, WORD IdSize)
{
#ifndef MARC
    if (Local == NULL)
        return NULL;
    if (Obj->Request != SNMP_PDU_NEXT)
    {
        if (Obj->IdLen != IdLen + IdSize)
			return NULL;
        while (Local != NULL && Local->Index < Obj->Id[IdLen])
            Local = Local->Next;
		if (Local == NULL || Local->Index != Obj->Id[IdLen])
            return NULL;
		return Local;
    }
    if (Obj->IdLen == IdLen)        /* first instance */
        return Local;
    while (Local != NULL && Local->Index < Obj->Id[IdLen])
        Local = Local->Next;
	if (Local == NULL || Local->Index < Obj->Id[IdLen])
        return NULL;
    return Local;
#else
    if (Local == NULL)
        return NULL;
    if (Obj->Request != SNMP_PDU_NEXT)
    {
        if (Obj->IdLen != IdLen + IdSize)
            return NULL;
        while (Local->Index < Obj->Id[IdLen] && Local->Next != NULL)
            Local = Local->Next;
        if (Local->Index == Obj->Id[IdLen])
            return Local;
        return NULL;
    }
    if (Obj->IdLen > IdLen + IdSize)        /* bad index */
        return NULL;
    if (Obj->IdLen == IdLen)                /* first instance */
        return Local;
    while (Local->Index < Obj->Id[IdLen] && Local->Next != NULL)
        Local = Local->Next;
    if (Local->Index >= Obj->Id[IdLen])
        return Local;
    return NULL;
#endif
}


MIB_LOCAL * MibInsert (SNMP_OBJECT *Obj, MIB_LOCAL **local, WORD IdLen, WORD IdSize)
{
    MIB_LOCAL * Junk;
    MIB_LOCAL * Local = *local;

    if (Obj->IdLen != IdLen + IdSize || Obj->Id[IdLen] < 1)
        return NULL;
    if ((Junk = DnpapMalloc (sizeof(MIB_LOCAL))) == NULL)
        return NULL;
    Junk->Index = Obj->Id[IdLen];

    if (Local == NULL || Local->Index > Junk->Index)
    {
        Junk->Next = Local;
        *local = Junk;
        return Junk;
    }
    while (Local->Next != NULL && Local->Next->Index <= Junk->Index)
        Local = Local->Next;
    if (Local->Index == Junk->Index)
    {
        DnpapFree(Junk);
        return Local;
    }
    Junk->Next = Local->Next;
    Local->Next = Junk;

    return Junk;
}


BOOLEAN MibRemove (SNMP_OBJECT *Obj, MIB_LOCAL **local, WORD IdLen, WORD IdSize)
{
    MIB_LOCAL * Junk;
    MIB_LOCAL * Local = *local;

    if (Local == NULL)
        return FALSE;
    if (Obj->IdLen != IdLen + IdSize)
        return FALSE;
    if (Local->Index == Obj->Id[IdLen])
    {
        Junk = Local->Next;
        DnpapFree(Local);
        *local = Junk;
        return TRUE;
    }
    while (Local->Next != NULL && Local->Next->Index < Obj->Id[IdLen])
        Local = Local->Next;
    if (Local->Next->Index == Obj->Id[IdLen])
    {
        Junk = Local->Next->Next;
        DnpapFree(Local->Next);
        Local->Next = Junk;
        return TRUE;
    }
    return FALSE;
}


INT MibCmpProfil (MIB_COMMUNITY *Prf1, MIB_COMMUNITY *Prf2)
{
	return MibCmpComm(Prf1->Comm, Prf1->CommLen, Prf2->Comm, Prf2->CommLen);
}


INT MibCmpComm(BYTE *comm1, WORD comm1len, BYTE *comm2, WORD comm2len)
{
INT result;

    if ((result = memcmp(comm1, comm2, min(comm1len, comm2len))) == 0)
	{
		if (comm1len < comm2len)
			return -1;
		else
		if (comm1len > comm2len)
			return 1;
		return 0;
	}
	return result < 0 ? -1 : 1;
}


INT MibCmpObject (MIB_ELEMENT *Obj1, MIB_ELEMENT *Obj2)
{
    return MibCmpObjId (Obj1->Id, Obj1->IdLen, Obj2->Id, Obj2->IdLen);
}


INT MibCmpObjId (ULONG *ObjId1, WORD ObjIdLen1, ULONG *ObjId2, WORD ObjIdLen2)
{
    while (ObjIdLen1 > 0 && ObjIdLen2 > 0 && *ObjId1 == *ObjId2)
    {
        ObjIdLen1--;
        ObjIdLen2--;
        ObjId1++;
        ObjId2++;
    }
    if (!ObjIdLen1 && !ObjIdLen2)
        return 0;
    if (!ObjIdLen1)
        return -1;
    if (!ObjIdLen2)
        return 1;
    if (*ObjId1 < *ObjId2)
        return -2;
    else
        return 2;
}


INT MibObjectFind(ULONG *Id, INT IdLen, INT *cmp)
{
    INT first, last, index;

    first = 0;
    last = MibRoot->Count - 1;
    *cmp = 2;

	if (last < 0)
		return 0;

    while (first < last)
    {
        index = (first + last)/2;
        *cmp = MibCmpObjId(MibRoot->Table[index]->Id,
							MibRoot->Table[index]->IdLen, Id, IdLen);
        switch (*cmp)
        {
            case -2:
                first = index+1;
                break;
            case -1:
            case 0:
                first = last = index;
    			return index;
            case 1:
            case 2:
                last = index;
                break;
        }
    }

	index = first;

    *cmp = MibCmpObjId(MibRoot->Table[index]->Id,
						MibRoot->Table[index]->IdLen, Id, IdLen);
	
    return index;
}


BOOLEAN MibObjectInsert(MIB_OBJECT *Object)
{
    INT index, cmp, i;

    if (MibRoot->Count >= MibRoot->Size)
    {
        int n;
        MIB_OBJECT **p;

        n = MibRoot->Size + 100;
        p = (MIB_OBJECT **)DnpapRealloc(MibRoot->Table, n*sizeof(MIB_OBJECT *));
        if (p == 0)
            return FALSE;
        MibRoot->Table = p;
        MibRoot->Size = n;
    }

    MibRoot->Table[MibRoot->Count] = Object;
	MibRoot->Count++;

    index = MibObjectFind(Object->Id, Object->IdLen, &cmp);
    if (cmp == 0 && index != MibRoot->Count-1)
	{
		MibRoot->Count--;
        return FALSE;
	}

    for (i = MibRoot->Count-1; i > index; i--)
        MibRoot->Table[i] = MibRoot->Table[i-1];
    MibRoot->Table[index] = Object;

    return TRUE;
}


BOOLEAN MibObjectDelete(MIB_OBJECT *Object)
{
    INT index, cmp, i;

    index = MibObjectFind(Object->Id, Object->IdLen, &cmp);
    if (cmp == 0)
        return FALSE;

    MibRoot->Count--;
    for (i = index; i < MibRoot->Count; i++)
        MibRoot->Table[i] = MibRoot->Table[i + 1];

    return TRUE;
}
