/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <agent.h>
#include <snmp.h>
#include <dnpap.h>
#include <config.h>
#include <message.h>
#include <dmath.h>
#include <dstring.h>

#include "mibsup.h"
#include "mibrqs.h"



MIB_RMON Mib =  {
                    NULL,
                    0,
                    NULL,
                    0,
                };


static MIB_ROOT *MibRoot;
  

static VOID MibExit(VOID);
static INT MibCmpProfil(MIB_COMMUNITY *Prf1, MIB_COMMUNITY *Prf2);
static INT MibCmpObject(MIB_ELEMENT *Obj1, MIB_ELEMENT *Obj2);
static INT MibCmpObjId(ULONG *ObjId, WORD ObjIdLen1, ULONG *ObjId2, WORD ObjIdLen2);
static INT MibObjectFind(ULONG *Id, INT IdLen, INT *Cmp);
static BOOLEAN MibObjectInsert(MIB_OBJECT *Object);
static BOOLEAN MibObjectDelete(MIB_OBJECT *Object);


BOOLEAN MibInit (MIB_ELEMENT *Obj, WORD ObjSze)
{				  							 
INT commnr, i;
CHAR configname[81];
CHAR *access;	  
AGENT_COMM *comm;
WORD support;
MIB_COMMUNITY *mibprf;
MIB_COMMUNITY *MibPrf = NULL;
											  
	MessageConfig(MIB_ERROR, "Mib");
	
	commnr = 1;
	sprintf(configname, "mib.access.%d", commnr);
	while (ConfigGetString(configname, &access) == TRUE)
	{		     
		if ((comm = AgentCommunity(commnr)) == NULL)
		{
			DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: no community %d defined in agent", commnr);
			break;
		}
		
		if (stricmp(access, "read-only") == 0)
			support = MIB_READ;
		else
		if (stricmp(access, "write-only") == 0)
			support = MIB_WRITE;
		else
		if (stricmp(access, "read-write") == 0)
			support = MIB_READ | MIB_WRITE;
		else
		{
			DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: invalid access specified");
			break;
		}
		
		if ((mibprf = DnpapRealloc(MibPrf, commnr*sizeof(MIB_COMMUNITY))) == NULL)
		{
			DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: out of community name space");
			break;
		}
		
		MibPrf = mibprf;
		memcpy(MibPrf[commnr-1].Comm, comm->comm, comm->commLen);
		MibPrf[commnr-1].CommLen = comm->commLen;
		MibPrf[commnr-1].Support = support;
		
		commnr++;
		sprintf(configname, "mib.access.%d", commnr);
	}
	
	if (--commnr == 0)
	{
		DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: no external access possible");
		return FALSE;
	}
	
	Mib.Prf = MibPrf;
    Mib.PrfSze = commnr;
	
    Mib.Obj = Obj;
    Mib.ObjSze = ObjSze;

    qsort (Mib.Prf, Mib.PrfSze, sizeof(MIB_COMMUNITY), MibCmpProfil);
    qsort (Mib.Obj, Mib.ObjSze, sizeof(MIB_ELEMENT), MibCmpObject);
	
	DnpapAtExit(MibExit);
      
    
    MibRoot = (MIB_ROOT *)DnpapMalloc(sizeof(MIB_ROOT)); 
    MibRoot->Size = 64;
    MibRoot->Count = 0;
    MibRoot->Table = (MIB_OBJECT **)DnpapMalloc(MibRoot->Size * sizeof(MIB_OBJECT *)); 
                   
    for (i=0; i<ObjSze; i++)   
    {
        if (!MibRegister(Obj[i].Id,
                         Obj[i].IdLen,
                         Obj[i].Rqs,
                         Obj[i].Type,
                         Obj[i].Support,
                         0))   
		{
			DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: register failed");
			break;
		}
    }
    
    return TRUE;
}


VOID MibExit(VOID)
{
}
                       

MIB_OBJECT *MibRegister(LONG         *Id,
                        WORD         IdLen,
                        MIB_CALLBACK Rqs,
                        WORD         Type,
                        WORD         Support,
                        VOID         *Param)
{    
    MIB_OBJECT *Object;

    Object = DnpapMalloc(sizeof(MIB_OBJECT));  
    if (Object == 0)
        return 0;
    Object->Id = DnpapMalloc(IdLen * sizeof(LONG));
    if (Object->Id == 0)
        return 0;   
    memcpy(Object->Id, Id, IdLen * sizeof(LONG));  
    Object->IdLen = IdLen;
    Object->Rqs = Rqs;
    Object->Type = Type;
    Object->Support = Support;
	Object->Param = Param;
    
    if (!MibObjectInsert(Object))  
    {
        DnpapFree(Object->Id);
        DnpapFree(Object);
        return 0;     
    }
    return Object;
}
   
                    
VOID MibDeRegister(MIB_OBJECT *Object)
{
    MibObjectDelete(Object);
    DnpapFree(Object->Id);
    DnpapFree(Object);  
}

        
WORD MibProfil (BYTE *Community, WORD CommunityLen)
{
    INT Fst, Lst, Ind;

    Fst = 0;
    Lst = Mib.PrfSze-1;

    while (Fst < Lst)
    {
        Ind = (Fst + Lst)/2;
        if (memcmp(Mib.Prf[Ind].Comm, Community, Mib.Prf[Ind].CommLen) < 0)
            Fst = Ind + 1;
        else
            Lst = Ind;
    }
    while (Fst < Mib.PrfSze && !memcmp(Mib.Prf[Fst].Comm, Community, Mib.Prf[Fst].CommLen))
    {
        if (Mib.Prf[Fst].CommLen == CommunityLen)
            return Mib.Prf[Fst].Support;
        Fst++;
    }
    return 0;
}


WORD MibRequest (UINT listLen, SNMP_OBJECT *list, WORD *errindex)
{
WORD status, index, errstatus;

    for (index = 0, status = SNMP_NOERROR;
         index < listLen && status == SNMP_NOERROR;
         index++)
    {
        list[index].Request = pdu.Type;
        status = Request(list + index);
    }
	errstatus = status;
	*errindex = index;

	if (pdu.Type == SNMP_PDU_SET)
	{
		if (errstatus == SNMP_NOERROR)
		{
			for (index = 0, status = SNMP_NOERROR;
				 index < listLen && status == SNMP_NOERROR;
				 index++)
			{
				list[index].Request = SNMP_PDU_COMMIT;
				status = Request(list + index);
			}
			errstatus = status;
			*errindex = index;
		}

		if (errstatus != SNMP_NOERROR)
		{
			for (index = 0, status = SNMP_NOERROR;
				 index < listLen && status == SNMP_NOERROR;
				 index++)
			{
				list[index].Request = SNMP_PDU_UNDO;
				status = Request(list + index);
			}
		}
	}
}


WORD Request (SNMP_OBJECT *Obj)
{
	static INT LastIndex = 0;
	static WORD LastRequest = SNMP_PDU_GET;
    INT Index, Cmp, Error;
	SNMP_OBJECT SaveObj;

    Error = SNMP_NOSUCHNAME;
    if (MibCmpObjId (MibRoot->Table[LastIndex]->Id, MibRoot->Table[LastIndex]->IdLen, Obj->Id, Obj->IdLen)!=-1)
    	Index = LastIndex = MibObjectFind(Obj->Id, Obj->IdLen, &Cmp);

    switch (Obj->Request)
    {
	case SNMP_PDU_GET:
		if (!(MibRoot->Table[Index]->Support & MIB_READ))
			return Error;
		if (MibRoot->Table[Index]->Rqs == NULL)
			return Error;
		if (!(MibCmpObjId (MibRoot->Table[Index]->Id, MibRoot->Table[Index]->IdLen, Obj->Id, Obj->IdLen)==-1))
			return Error;
		Error = MibRoot->Table[Index]->Rqs(Obj, MibRoot->Table[Index]->IdLen, MibRoot->Table[Index]->Param);
		if (Error == SNMP_NOERROR)
			Obj->Type = MibRoot->Table[Index]->Type;
		return Error;
	case SNMP_PDU_SET:
		if (!(MibRoot->Table[Index]->Support & MIB_WRITE))
			return Error;
		if (MibRoot->Table[Index]->Rqs == NULL)
			return Error;
		if (!(MibCmpObjId (MibRoot->Table[Index]->Id, MibRoot->Table[Index]->IdLen, Obj->Id, Obj->IdLen)==-1))
			return Error;
		if (MibRoot->Table[Index]->Type != Obj->Type)
			return SNMP_BADVALUE;
		if (Obj->Type == SNMP_OBJECTID && Obj->SyntaxLen > SNMP_SIZE_BUFINT)
			return SNMP_BADVALUE;
		if ((Obj->Type == SNMP_OCTETSTR ||
			 Obj->Type == SNMP_DISPLAYSTR ||
			 Obj->Type == SNMP_OPAQUE) && 
			(Obj->SyntaxLen > SNMP_SIZE_BUFCHR))
			return SNMP_BADVALUE;
		Error = MibRoot->Table[Index]->Rqs(Obj, MibRoot->Table[Index]->IdLen, MibRoot->Table[Index]->Param);
		return Error;
	case SNMP_PDU_NEXT:
		switch (MibCmpObjId(MibRoot->Table[Index]->Id, MibRoot->Table[Index]->IdLen, 
							Obj->Id, Obj->IdLen))
		{
		case -2:
			return Error;
		case -1:
		case 0:
			if (MibRoot->Table[Index]->Rqs != NULL && MibRoot->Table[Index]->Support & MIB_READ)
			{
				Error = MibRoot->Table[Index]->Rqs(Obj, MibRoot->Table[Index]->IdLen, MibRoot->Table[Index]->Param);
				if (Error == SNMP_NOERROR)
					Obj->Type = MibRoot->Table[Index]->Type;
			}
			Index++;
			break;
		}
		if (Error == SNMP_NOSUCHNAME)
			memcpy(SaveObj.Id, Obj->Id, 
				   (SaveObj.IdLen = Obj->IdLen) * sizeof(Obj->Id[0]));
		while (Index < MibRoot->Count && Error == SNMP_NOSUCHNAME)
		{
			memcpy(Obj->Id, MibRoot->Table[Index]->Id,
				   MibRoot->Table[Index]->IdLen * sizeof(LONG));
			Obj->IdLen = MibRoot->Table[Index]->IdLen;
			if (MibRoot->Table[Index]->Rqs != NULL && MibRoot->Table[Index]->Support & MIB_READ)
			{
				Error = MibRoot->Table[Index]->Rqs(Obj, MibRoot->Table[Index]->IdLen, MibRoot->Table[Index]->Param);
				if (Error == SNMP_NOERROR)
					Obj->Type = MibRoot->Table[Index]->Type;
			}
			Index++;
		}
		if (Error == SNMP_NOSUCHNAME)
			memcpy(Obj->Id, SaveObj.Id, 
				   (Obj->IdLen = SaveObj.IdLen) * sizeof(Obj->Id[0]));
		return Error;
	case SNMP_PDU_COMMIT:
	case SNMP_PDU_UNDO:
		if (MibRoot->Table[Index]->Rqs == NULL)
			return Error;
		Error = MibRoot->Table[Index]->Rqs(Obj, MibRoot->Table[Index]->IdLen, MibRoot->Table[Index]->Param);
		return Error;
    }
	
	return SNMP_GENERROR;
}


BOOLEAN MibSimple (SNMP_OBJECT *Obj, WORD IdLen)
{
    if (Obj->Request != SNMP_PDU_NEXT)
    {
        if (Obj->IdLen == IdLen + 1 && Obj->Id[IdLen] == 0)
            return TRUE;
        return FALSE;
    }
    if (Obj->IdLen == IdLen)
    {
        Obj->Id[IdLen] = 0;
        Obj->IdLen++;
        return TRUE;
    }
    return FALSE;
}


MIB_LOCAL * MibRmon (SNMP_OBJECT *Obj, MIB_LOCAL *Local, WORD IdLen, WORD IdSize)
{
#ifndef MARC
    if (Local == NULL)
        return NULL;
    if (Obj->Request != SNMP_PDU_NEXT)
    {
        if (Obj->IdLen != IdLen + IdSize)
			return NULL;
        while (Local != NULL && Local->Index < Obj->Id[IdLen])
            Local = Local->Next;
		if (Local == NULL || Local->Index != Obj->Id[IdLen])
            return NULL;
		return Local;
    }
    if (Obj->IdLen == IdLen)        /* first instance */
        return Local;
    while (Local != NULL && Local->Index < Obj->Id[IdLen])
        Local = Local->Next;
	if (Local == NULL || Local->Index < Obj->Id[IdLen])
        return NULL;
    return Local;
#else
    if (Local == NULL)
        return NULL;
    if (Obj->Request != SNMP_PDU_NEXT)
    {
        if (Obj->IdLen != IdLen + IdSize)
            return NULL;
        while (Local->Index < Obj->Id[IdLen] && Local->Next != NULL)
            Local = Local->Next;
        if (Local->Index == Obj->Id[IdLen])
            return Local;
        return NULL;
    }
    if (Obj->IdLen > IdLen + IdSize)        /* bad index */
        return NULL;
    if (Obj->IdLen == IdLen)                /* first instance */
        return Local;
    while (Local->Index < Obj->Id[IdLen] && Local->Next != NULL)
        Local = Local->Next;
    if (Local->Index >= Obj->Id[IdLen])
        return Local;
    return NULL;
#endif
}


MIB_LOCAL * MibInsert
    (SNMP_OBJECT *Obj, MIB_LOCAL **local, WORD IdLen, WORD IdSize)
{
    MIB_LOCAL * Junk;
    MIB_LOCAL * Local = *local;

    if (Obj->IdLen != IdLen + IdSize || Obj->Id[IdLen] < 1)
        return NULL;
    if ((Junk = DnpapMalloc (sizeof(MIB_LOCAL))) == NULL)
        return NULL;
    Junk->Index = Obj->Id[IdLen];
    
    if (Local == NULL || Local->Index > Junk->Index)
    {
        Junk->Next = Local;
        *local = Junk;
        return Junk;
    }
    while (Local->Next != NULL && Local->Next->Index <= Junk->Index)
        Local = Local->Next;
    if (Local->Index == Junk->Index)
    {
        DnpapFree(Junk);
        return Local;
    }
    Junk->Next = Local->Next;
    Local->Next = Junk;

    return Junk;
}


BOOLEAN MibRemove (SNMP_OBJECT *Obj, MIB_LOCAL **local, WORD IdLen, WORD IdSize)
{
    MIB_LOCAL * Junk;
    MIB_LOCAL * Local = *local;

    if (Local == NULL)
        return FALSE;
    if (Obj->IdLen != IdLen + IdSize)
        return FALSE;
    if (Local->Index == Obj->Id[IdLen])
    {
        Junk = Local->Next;
        DnpapFree(Local);
        *local = Junk;
        return TRUE;
    }
    while (Local->Next != NULL && Local->Next->Index < Obj->Id[IdLen])
        Local = Local->Next;
    if (Local->Next->Index == Obj->Id[IdLen])
    {
        Junk = Local->Next->Next;
        DnpapFree(Local->Next);
        Local->Next = Junk;
        return TRUE;
    }
    return FALSE;
}


INT MibCmpProfil (MIB_COMMUNITY *Prf1, MIB_COMMUNITY *Prf2)
{
    return memcmp(Prf1->Comm, Prf2->Comm, SNMP_SIZE_COMM); 
}


INT MibCmpObject (MIB_ELEMENT *Obj1, MIB_ELEMENT *Obj2)
{
    return MibCmpObjId (Obj1->Id, Obj1->IdLen, Obj2->Id, Obj2->IdLen);
}


INT MibCmpObjId (ULONG *ObjId1, WORD ObjIdLen1, ULONG *ObjId2, WORD ObjIdLen2)
{
    while (ObjIdLen1 > 0 && ObjIdLen2 > 0 && *ObjId1 == *ObjId2)
    {
        ObjIdLen1--;
        ObjIdLen2--;
        ObjId1++;
        ObjId2++;
    }
    if (!ObjIdLen1 && !ObjIdLen2)
        return 0;
    if (!ObjIdLen1)
        return -1;
    if (!ObjIdLen2)
        return 1;
    if (*ObjId1 < *ObjId2)
        return -2;
    else
        return 2;
}


INT MibObjectFind(ULONG *Id, INT IdLen, INT *Cmp)
{
    INT First, Last, Index;

    First = 0;
    Last = MibRoot->Count - 1;    
    *Cmp = 2;

    while (First < Last)
    {
        Index = (First + Last)/2;  
        *Cmp = MibCmpObjId(MibRoot->Table[Index]->Id, MibRoot->Table[Index]->IdLen, 
						   Id, IdLen);
        switch (*Cmp)
        {
            case -2:
                First = Index+1;
                break;
            case -1:
            case 0:
                First = Last = Index;
                break;
            case 1:
            case 2:
                Last = Index;
                break;
        }
    }
    return First;
}

                    
BOOLEAN MibObjectInsert(MIB_OBJECT *Object)
{                 
    INT Index, Cmp, i;

    if (MibRoot->Count >= MibRoot->Size)   
    { 
        MIB_OBJECT **p;
        int n;
                   
        n = MibRoot->Size * 2;
        p = (MIB_OBJECT **)DnpapRealloc(MibRoot->Table, n * sizeof(MIB_OBJECT *));   
        if (p == 0) 
            return FALSE;  
        MibRoot->Size = n;    
        MibRoot->Table = p;
    }
                         
    MibRoot->Table[MibRoot->Count] = Object;
	MibRoot->Count++;
    Index = MibObjectFind(Object->Id, Object->IdLen, &Cmp);
    if (Cmp == 0 && Index != MibRoot->Count-1)
	{
		MibRoot->Count--;
        return FALSE;  
	}
    for (i = MibRoot->Count-1; i > Index; i--)
        MibRoot->Table[i] = MibRoot->Table[i-1];  
       
    MibRoot->Table[Index] = Object;
    return TRUE;   
}
    
    
BOOLEAN MibObjectDelete(MIB_OBJECT *Object)
{                 
    INT Index, Cmp, i;

    Index = MibObjectFind(Object->Id, Object->IdLen, &Cmp);
    if (Cmp == 0)
        return FALSE;  
                         
    MibRoot->Count--;
    for (i = Index; i < MibRoot->Count; i++)
        MibRoot->Table[i] = MibRoot->Table[i + 1];  
       
    return TRUE;   
}
