/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>
#include <config.h>
#include <mibsup.h>
#include <mibrqs.h>
#include <agent.h>
#include <trap.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <timer.h>
#include <sysmib.h>

#include "neuro_d.h"
#include "neuro_e.h"
#include "neuro_k.h"
#include "neuro_c.h"
#include "neuro.h"

IMPORT BOOLEAN BeholderStrictRMON;

static CHAR MODULE[] = "neuro";

static MIB_LOCAL *NeuroDataInstance = NULL;

static WORD TotalNrNetworks = 0, RunningNrNetworks = 0;

static BOOLEAN NeuroMibNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize);
static BOOLEAN NeuroGetFirstNext(NeuroData * data, SNMP_OBJECT * obj, WORD idlen, WORD indexsize, BOOLEAN first);

static WORD NeuroSetId(SNMP_OBJECT * obj, WORD idc, WORD indexsize, LONG c, LONG l, LONG n, LONG i);

static WORD MibName2ID(LONG * id, CHAR * name, UINT * n);
static WORD MibID2String(const LONG * id, UINT nr, CHAR * name, INT maxlen);

static BOOLEAN NeuroMib2Neuron(NeuroData * net, InOutData * mibinout, LONG n, InOutBatch * batch, InOutput * inout);
static BOOLEAN NeuroNeuron2Mib(InOutput * inout, LONG n, InOutData * iodata, FLOAT * mibinout);

static FLOAT NeuroMean(FLOAT value[], LONG nrvalues);
static FLOAT NeuroHarmonic(FLOAT value[], LONG nrvalues);

static VOID NeuroMCalcCallback(TIMER_DESCR * timer, ULONG now, VOID * param);

static BOOLEAN NeuroInitOutput(LONG n, InOutput * inout);

static BOOLEAN NeuroAllocBatch(NeuroData * data, LONG size, BOOLEAN reset);

static BOOLEAN NeuroCheckThresholds(NeuroData * data);

/*  config data  */
#define MAXCONFIGNAME       81
#define MAXDATASRCNAME	    81

static SHORT NrNetworks = 0;

BOOLEAN NeuroMInit(VOID)
{
	SHORT NrLayers, NrNeurons, DefaultNrLayers, DataSourceSteps, DataSourceAverageMethod;
	BOOLEAN DataSourceTrap, DataSourceThresholdCheck;
	DOUBLE LearningRate, Momentum, DefaultLearningRate, DefaultMomentum, DataSourceValue, DataSourceThresholdValue;
	SNMP_OBJECT obj = {SNMP_PDU_SET, {NMIB_BASE}, NMIB_BASELEN, SNMP_INTEGER, SNMP_VALID};
	WORD dummy;
	CHAR configname[MAXCONFIGNAME], *DataSourceName;
	SHORT c, l, n, i;

	MessageConfig(NEURO_ERROFFSET, "Neuro");

	if (NeuroErrorInit() == FALSE)
		FATAL(MODULE, NEURO_ERRINIT);

	if (ConfigGetShort("beholder.neuro.networks", &NrNetworks) == TRUE)
	{
		ConfigGetShort("beholder.neuro.Layers", &DefaultNrLayers);
		ConfigGetDouble("beholder.neuro.rate", &DefaultLearningRate);
		ConfigGetDouble("beholder.neuro.momentum", &DefaultMomentum);

		for (c = 1; c <= NrNetworks; c++)
		{
			/* set Status to SNMP_CREATEREQUEST  */
			obj.Type = SNMP_INTEGER;
			obj.Syntax.LngInt = SNMP_CREATEREQUEST;
			NeuroSetId(&obj, NMIB_STATUS, 1, c, 0, 0, 0);
			if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
				WARNING(MODULE, NEURO_MCREATE);

			/* set Owner to MODULE  */
			obj.Type = SNMP_OCTETSTR;
			memcpy(obj.Syntax.BufChr, MODULE, sizeof(MODULE));
			obj.SyntaxLen = sizeof(MODULE);
			NeuroSetId(&obj, NMIB_OWNER, 1, c, 0, 0, 0);
			if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
				WARNING(MODULE, NEURO_MOWNER);

			/* set NrLayers to config value or default  */
			obj.Type = SNMP_INTEGER;
			sprintf(configname, "beholder.neuro.layers.%d", c);
			if (ConfigGetShort(configname, &NrLayers) == TRUE)
				obj.Syntax.LngInt = NrLayers;
			else
				obj.Syntax.LngInt = DefaultNrLayers;
			NeuroSetId(&obj, NMIB_NRLAYERS, 1, c, 0, 0, 0);
			if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
				WARNING(MODULE, NEURO_MNRLAYERS);

			/* set LearningRate to config value or default  */
			obj.Type = SNMP_INTEGER;
			sprintf(configname, "beholder.neuro.rate.%d", c);
			if (ConfigGetDouble(configname, &LearningRate) == TRUE)
				obj.Syntax.LngInt = (LONG) (FLOAT_SCALE * LearningRate);
			else
				obj.Syntax.LngInt = (LONG) (FLOAT_SCALE * DefaultLearningRate);
			NeuroSetId(&obj, NMIB_LRATE, 1, c, 0, 0, 0);
			if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
				WARNING(MODULE, NEURO_MLRATE);

			/* set Momentum to config value or default  */
			obj.Type = SNMP_INTEGER;
			sprintf(configname, "beholder.neuro.momentum.%d", c);
			if (ConfigGetDouble(configname, &Momentum) == TRUE)
				obj.Syntax.LngInt = (LONG) (FLOAT_SCALE * Momentum);
			else
				obj.Syntax.LngInt = (LONG) (FLOAT_SCALE * DefaultMomentum);
			NeuroSetId(&obj, NMIB_MOMENTUM, 1, c, 0, 0, 0);
			if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
				WARNING(MODULE, NEURO_MMOMENTUM);

			for (l = 1; l <= NrLayers; l++)
			{
				obj.Type = SNMP_INTEGER;
				sprintf(configname, "beholder.neuro.neurons.%d.%d", c, l);
				if (ConfigGetShort(configname, &NrNeurons) == TRUE)
					obj.Syntax.LngInt = NrNeurons;
				else
				{
					ERROR2(MODULE, NEURO_MNRNEURONS, c, l);
					return FALSE;
				}
				NeuroSetId(&obj, NMIB_NRNEURONS, 2, c, l, 0, 0);
				if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
					ERROR(MODULE, NEURO_MSETNEURONS);

				if (l == 1 || l == NrLayers)
				{
					for (n = 1; n <= NrNeurons; n++)
					{
						obj.Type = SNMP_OCTETSTR;
						sprintf(configname, "beholder.neuro.datasource.%d.%d.%d", c, l, n);
						if (ConfigGetString(configname, &DataSourceName) == TRUE)
							MibName2ID(obj.Syntax.BufInt, DataSourceName, &obj.SyntaxLen);
						else
						{
							ERROR3(MODULE, NEURO_MDATASRCNAME, c, l, n);
							return FALSE;
						}
						NeuroSetId(&obj, NMIB_DATASRC, 3, c, l, n, 0);
						if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
							ERROR(MODULE, NEURO_MSETDATASRC);

						obj.Type = SNMP_INTEGER;
						sprintf(configname, "beholder.neuro.datasource.max.%d.%d.%d", c, l, n);
						if (ConfigGetDouble(configname, &DataSourceValue) == TRUE)
							obj.Syntax.LngInt = (LONG) (FLOAT_SCALE * DataSourceValue);
						else
						{
							ERROR3(MODULE, NEURO_MDATASRCMAX, c, l, n);
							return FALSE;
						}
						NeuroSetId(&obj, NMIB_DATASRCMAX, 3, c, l, n, 0);
						if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
							ERROR(MODULE, NEURO_MSETDATASRCMAX);

						obj.Type = SNMP_INTEGER;
						sprintf(configname, "beholder.neuro.datasource.min.%d.%d.%d", c, l, n);
						if (ConfigGetDouble(configname, &DataSourceValue) == TRUE)
							obj.Syntax.LngInt = (LONG) (FLOAT_SCALE * DataSourceValue);
						else
						{
							ERROR3(MODULE, NEURO_MDATASRCMIN, c, l, n);
							return FALSE;
						}
						NeuroSetId(&obj, NMIB_DATASRCMIN, 3, c, l, n, 0);
						if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
							ERROR(MODULE, NEURO_MSETDATASRCMIN);

						obj.Type = SNMP_INTEGER;
						sprintf(configname, "beholder.neuro.datasource.method.%d.%d.%d", c, l, n);
						if (ConfigGetShort(configname, &DataSourceAverageMethod) == TRUE)
							obj.Syntax.LngInt = DataSourceAverageMethod;
						else
						{
							ERROR3(MODULE, NEURO_MDATASRCAVERAGEMETHOD, c, l, n);
							return FALSE;
						}
						NeuroSetId(&obj, NMIB_DATASRCAVERAGEMETHOD, 3, c, l, n, 0);
						if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
							ERROR(MODULE, NEURO_MSETDATASRCAVERAGEMETHOD);

						obj.Type = SNMP_INTEGER;
						sprintf(configname, "beholder.neuro.datasource.steps.%d.%d.%d", c, l, n);
						if (ConfigGetShort(configname, &DataSourceSteps) == TRUE)
							obj.Syntax.LngInt = DataSourceSteps;
						else
						{
							ERROR3(MODULE, NEURO_MDATASRCSTEPS, c, l, n);
							return FALSE;
						}
						NeuroSetId(&obj, NMIB_DATASRCSTEPS, 3, c, l, n, 0);
						if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
							ERROR(MODULE, NEURO_MSETDATASRCSTEPS);

						obj.Type = SNMP_INTEGER;
						sprintf(configname, "beholder.neuro.datasource.trap.%d.%d.%d", c, l, n);
						if (ConfigGetBoolean(configname, &DataSourceTrap) == TRUE)
							obj.Syntax.LngInt = DataSourceTrap;
						else
						{
							ERROR3(MODULE, NEURO_MDATASRCTRAP, c, l, n);
							return FALSE;
						}
						NeuroSetId(&obj, NMIB_DATASRCTRAP, 3, c, l, n, 0);
						if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
							ERROR(MODULE, NEURO_MSETDATASRCTRAP);

						obj.Type = SNMP_INTEGER;
						sprintf(configname, "beholder.neuro.datasource.check.%d.%d.%d", c, l, n);
						if (ConfigGetBoolean(configname, &DataSourceThresholdCheck) == TRUE)
							obj.Syntax.LngInt = DataSourceThresholdCheck;
						else
						{
							ERROR3(MODULE, NEURO_MDATASRCCHECK, c, l, n);
							return FALSE;
						}
						NeuroSetId(&obj, NMIB_DATASRCTHRESHOLDCHECK, 3, c, l, n, 0);
						if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
							ERROR(MODULE, NEURO_MSETDATASRCCHECK);

						for (i = 0; i < MAXNRTHRESHOLDVALUES; i++)
						{
							obj.Type = SNMP_INTEGER;
							sprintf(configname, "beholder.neuro.datasource.threshold.%d.%d.%d.%d", c, l, n, i);
							if (ConfigGetDouble(configname, &DataSourceThresholdValue) == TRUE)
								obj.Syntax.LngInt = (LONG) (FLOAT_SCALE * DataSourceThresholdValue);
							else
							{
								ERROR4(MODULE, NEURO_MDATASRCTHRESHOLDVALUE, c, l, n, i);
								return FALSE;
							}
							NeuroSetId(&obj, NMIB_DATASRCTHRESHOLDVALUE, 4, c, l, n, i);
							if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
								ERROR(MODULE, NEURO_MSETDATASRCTHRESHOLDVALUE);
						}
					}
				}
			}

			NeuroSetId(&obj, NMIB_STATUS, 1, c, 0, 0, 0);
			if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
				WARNING(MODULE, NEURO_MSTART);
		}
	}

	return TRUE;
}

BOOLEAN NeuroMibNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize)
{
	BOOLEAN first = FALSE;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	while (*local != NULL && NeuroGetFirstNext((NeuroData *) (*local)->Data, obj, idlen, indexsize, first) == FALSE)
	{
		first = TRUE;
		/* only go to next collector if requested index was higher or equal */
		if (obj->Id[idlen] >= (*local)->Index)
			*local = (*local)->Next;
	}

	if (*local == NULL)
		return FALSE;
	obj->Id[idlen] = (*local)->Index;
	obj->IdLen = idlen + indexsize;

	return TRUE;
}

BOOLEAN NeuroGetFirstNext(NeuroData * data, SNMP_OBJECT * obj, WORD idlen, WORD indexsize, BOOLEAN first)
{
	LONG layer, neuron, step;
	WORD i, len;

	if (first == TRUE)
	{
		/* for (i = obj->IdLen; i < idlen+indexsize; i++) */
		for (i = idlen; i < idlen + indexsize; i++)
			obj->Id[i] = 1;
		obj->IdLen = idlen + indexsize;
		return TRUE;
	}

	layer = neuron = step = 0;

	len = obj->IdLen - idlen;

	switch (len)
	{
	case NMIB_IOFFSET:
		step = obj->Id[idlen - 1 + NMIB_IOFFSET];
	case NMIB_NOFFSET:
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
	case NMIB_LOFFSET:
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
	}

	switch (indexsize)
	{
	case NMIB_IOFFSET:
		/* layer index must be 1 or NrLayers  */
		if ((layer == 1 || layer == data->NrLayers) &&
			(neuron > 0 && neuron <= data->NrNeurons[layer - 1]) &&
		 step < (layer == 1 ? NeuroThresholds(&data->InputData[neuron - 1]) : NeuroThresholds(&data->TargetData[neuron - 1])))
		{
			step++;
			break;
		}
	case NMIB_NOFFSET:
		/* layer index must be 1 or NrLayers  */
		if ((layer == 1 || layer == data->NrLayers) &&
			(neuron > 0 && neuron < data->NrNeurons[layer - 1]))
		{
			neuron++;
			switch (indexsize)
			{
			case NMIB_IOFFSET:
				step = 1;
			}
			break;
		}
	case NMIB_LOFFSET:
		if (layer > 0 && layer < data->NrLayers)
		{
			/* layer index must be 1 or NrLayers in case   */
			/* neuron index or subindex must be specified  */
			if (indexsize == NMIB_IOFFSET || indexsize == NMIB_NOFFSET)
				layer = data->NrLayers;
			else
				layer++;
			switch (indexsize)
			{
			case NMIB_IOFFSET:
				step = 1;
			case NMIB_NOFFSET:
				neuron = 1;
			}
			break;
		}
	case NMIB_COFFSET:
		if (len == 0)
		{
			layer = neuron = step = 1;
			break;
		}
		return FALSE;
	}

	switch (indexsize)
	{
	case NMIB_IOFFSET:
		obj->Id[idlen - 1 + NMIB_IOFFSET] = step;
	case NMIB_NOFFSET:
		obj->Id[idlen - 1 + NMIB_NOFFSET] = neuron;
	case NMIB_LOFFSET:
		obj->Id[idlen - 1 + NMIB_LOFFSET] = layer;
	}

	return TRUE;
}

WORD NeuroNrNetworks(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	if (MibSimple(obj, idlen) == FALSE)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = TotalNrNetworks;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroRunningNrNetworks(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	if (MibSimple(obj, idlen) == FALSE)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = RunningNrNetworks;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroLayerIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 2)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 2) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroLayerIndexLayer(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 2)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 2) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = obj->Id[idlen - 1 + NMIB_LOFFSET];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroLayerNeuronIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroLayerNeuronIndexLayer(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = obj->Id[idlen - 1 + NMIB_LOFFSET];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroLayerNeuronIndexNeuron(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = obj->Id[idlen - 1 + NMIB_NOFFSET];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroLayerNeuronValueIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 4)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 4) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroLayerNeuronValueIndexLayer(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 4)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 4) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = obj->Id[idlen - 1 + NMIB_LOFFSET];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroLayerNeuronValueIndexNeuron(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 4)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 4) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = obj->Id[idlen - 1 + NMIB_NOFFSET];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroLayerNeuronValueIndexValue(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 4)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 4) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = obj->Id[idlen - 1 + NMIB_IOFFSET];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroNrLayers(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	InOutData *iodata = NULL;
	LONG l;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		obj->Syntax.LngInt = data->NrLayers;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt > MAXNRLAYERS)
			return SNMP_BADVALUE;
		if ((iodata = ReNewInOutData(data->TargetData,
									 data->NrLayers >= obj->Syntax.LngInt ? data->NrNeurons[obj->Syntax.LngInt] : 1,
									 data->NrNeurons[data->NrLayers])) == NULL)
			return SNMP_GENERROR;
		data->TargetData = iodata;
		for (l = data->NrLayers; l < obj->Syntax.LngInt; l++)
			data->NrNeurons[l] = 1;
		data->NrLayers = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroNrNeurons(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	LONG layer;
	InOutData *iodata = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 2)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 2) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		if (layer > data->NrLayers)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = data->NrNeurons[layer - 1];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		if (layer > data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (obj->Syntax.LngInt < 1 || obj->Syntax.LngInt > MAXNRNEURONS)
			return SNMP_BADVALUE;
		/* input or output layer?  */
		if (layer == 1)
		{
			if ((iodata = ReNewInOutData(data->InputData, obj->Syntax.LngInt, data->NrNeurons[layer - 1])) == NULL)
				return SNMP_GENERROR;
			data->InputData = iodata;
		}
		else if (layer == data->NrLayers)
		{
			if ((iodata = ReNewInOutData(data->TargetData, obj->Syntax.LngInt, data->NrNeurons[layer - 1])) == NULL)
				return SNMP_GENERROR;
			data->TargetData = iodata;
		}
		data->NrNeurons[layer - 1] = obj->Syntax.LngInt;
		if (NeuroAllocBatch(data, data->RequestedSize, TRUE) == FALSE)
			return SNMP_GENERROR;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroNetRefresh(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		obj->Syntax.LngUns = data->NetRefresh;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		if (data->Status == SNMP_VALID)	/* if running change timer  */
			if (NeuroCChangeTimer(data, obj->Syntax.LngUns) == FALSE)
				return SNMP_GENERROR;
		data->NetRefresh = obj->Syntax.LngUns;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroTrain(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		obj->Syntax.LngInt = data->Train;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		if (obj->Syntax.LngInt < 1 && obj->Syntax.LngInt > LAST_TRAINMETHOD)
			return SNMP_BADVALUE;
		if (obj->Syntax.LngInt == NEURO_BATCH)
		{
			data->BatchSize = 0;
			data->BatchIndex = 0;
		}
		data->Train = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroRequestedSize(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		obj->Syntax.LngUns = data->RequestedSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt < 1 || obj->Syntax.LngInt > MAXNRPATTERNS)
			return SNMP_BADVALUE;
		if (NeuroAllocBatch(data, obj->Syntax.LngInt, FALSE) == FALSE)
			return SNMP_GENERROR;
		data->RequestedSize = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroGrantedSize(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		obj->Syntax.LngInt = data->GrantedSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroBatchSize(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		obj->Syntax.LngInt = data->BatchSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroLearningRate(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		obj->Syntax.LngInt = (LONG) (FLOAT_SCALE * data->LearningRate);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		data->LearningRate = ((FLOAT) obj->Syntax.LngInt / FLOAT_SCALE);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroMomentum(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		obj->Syntax.LngInt = (LONG) (FLOAT_SCALE * data->Momentum);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		data->Momentum = ((FLOAT) obj->Syntax.LngInt / FLOAT_SCALE);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroError(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		obj->Syntax.LngInt = (LONG) (FLOAT_SCALE * data->Error);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroTotalSteps(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		obj->Syntax.LngUns = data->TotalSteps;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroTrainingSteps(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		obj->Syntax.LngUns = data->TrainSteps;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroDataSource(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	InOutData *datasource = NULL;
	LONG layer, neuron;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		memcpy(obj->Syntax.BufInt, datasource->Id,
			   (obj->SyntaxLen = datasource->IdLen) * sizeof(obj->Syntax.BufInt[0]));
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		if (obj->SyntaxLen < 2 ||
			(obj->SyntaxLen == 2 && (obj->Syntax.BufInt[0] != 0 || obj->Syntax.BufInt[1] != 0)))
			return SNMP_BADVALUE;
		memcpy(datasource->Id, obj->Syntax.BufInt,
			   (datasource->IdLen = obj->SyntaxLen) * sizeof(obj->Syntax.BufInt[0]));
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroDataSourceValue(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	InOutData *datasource = NULL;
	LONG layer, neuron;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		obj->Syntax.LngInt = (LONG) (FLOAT_SCALE * datasource->TheValue);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroDataSourceMax(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	InOutData *datasource = NULL;
	LONG layer, neuron;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		obj->Syntax.LngInt = datasource->MaxValue;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		datasource->MaxValue = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroDataSourceMin(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	InOutData *datasource = NULL;
	LONG layer, neuron;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		obj->Syntax.LngInt = datasource->MinValue;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		datasource->MinValue = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroDataSourceAverageMethod(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	InOutData *datasource = NULL;
	LONG layer, neuron;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		obj->Syntax.LngInt = datasource->AverageMethod;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		if (obj->Syntax.LngInt < 1 || obj->Syntax.LngInt > LAST_AVERAGEMETHOD)
			return SNMP_BADVALUE;
		datasource->AverageMethod = obj->Syntax.LngInt;
		if (datasource->AverageMethod == SIMPLE_DIFF || datasource->AverageMethod == SIMPLE_VALUE)
			datasource->AverageSteps = 1;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroDataSourceAverageSteps(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	InOutData *datasource = NULL;
	LONG layer, neuron;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		obj->Syntax.LngInt = datasource->AverageSteps;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		if (datasource->AverageMethod == SIMPLE_DIFF || datasource->AverageMethod == SIMPLE_VALUE)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt < 1 || obj->Syntax.LngInt > MAXNRSTEPS)
			return SNMP_BADVALUE;
		datasource->AverageSteps = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroDataSourceTrap(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	InOutData *datasource = NULL;
	LONG layer, neuron;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		obj->Syntax.LngInt = datasource->Trap == TRUE ? 1 : 2;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		if (obj->Syntax.LngInt != 1 && obj->Syntax.LngInt != 2)
			return SNMP_BADVALUE;
		datasource->Trap = obj->Syntax.LngInt == 1 ? TRUE : FALSE;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroDataSourceTrapComm(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	InOutData *datasource = NULL;
	LONG layer, neuron;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		memcpy(obj->Syntax.BufChr, datasource->TrapComm,
			   (obj->SyntaxLen = datasource->TrapCommLen) * sizeof(obj->Syntax.BufChr[0]));
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		memcpy(datasource->TrapComm, obj->Syntax.BufChr,
			   (datasource->TrapCommLen = obj->SyntaxLen) * sizeof(obj->Syntax.BufChr[0]));
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroDataSourceThresholdCheck(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	InOutData *datasource = NULL;
	LONG layer, neuron;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		obj->Syntax.LngInt = datasource->ThresholdCheck;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		if (obj->Syntax.LngInt < 1 || obj->Syntax.LngInt > LAST_THRESHOLDCHECK)
			return SNMP_BADVALUE;
		datasource->ThresholdCheck = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroDataSourceOutputValue(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	InOutData *datasource = NULL;
	LONG layer, neuron;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 3)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 3) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron < 1 || neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
		{
			if (data->Status != SNMP_VALID)
				obj->Syntax.LngInt = 0;
			else
			{
				datasource = &data->InputData[neuron - 1];
				obj->Syntax.LngInt = (LONG) (FLOAT_SCALE * datasource->TheValue);
			}
		}
		else
		{
			if (data->Status != SNMP_VALID)
				obj->Syntax.LngInt = 0;
			else
				obj->Syntax.LngInt = (LONG) (FLOAT_SCALE * data->OutputValue[neuron - 1]);
		}
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroDataSourceThresholdValue(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;
	InOutData *datasource = NULL;
	LONG layer, neuron;
	LONG index;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 4)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 4) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		index = obj->Id[idlen - 1 + NMIB_IOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		if (index > MAXNRTHRESHOLDVALUES)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = (LONG) (datasource->ThresholdValue[index - 1] * FLOAT_SCALE);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
		layer = obj->Id[idlen - 1 + NMIB_LOFFSET];
		neuron = obj->Id[idlen - 1 + NMIB_NOFFSET];
		index = obj->Id[idlen - 1 + NMIB_IOFFSET];
		if (layer != 1 && layer != data->NrLayers)
			return SNMP_NOSUCHNAME;
		if (neuron > data->NrNeurons[layer - 1])
			return SNMP_NOSUCHNAME;
		if (layer == 1)
			datasource = &data->InputData[neuron - 1];
		else
			datasource = &data->TargetData[neuron - 1];
		if (index > MAXNRTHRESHOLDVALUES)
			return SNMP_NOSUCHNAME;
		datasource->ThresholdValue[index - 1] = ((FLOAT) obj->Syntax.LngInt / FLOAT_SCALE);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroOwner(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;

	if ((local = MibRmon(obj, NeuroDataInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (NeuroData *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, NEURO_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Owner, obj->SyntaxLen = data->OwnerLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (NeuroData *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, obj->Syntax.BufChr, data->OwnerLen = obj->SyntaxLen);
		BooleanSetTrue(data->ObjectSet, NEURO_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD NeuroStatus(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	NeuroData *data = NULL;

	local = MibRmon(obj, NeuroDataInstance, idlen, 1);

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (NeuroMibNext(obj, &local, idlen, 1) == FALSE)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (NeuroData *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(obj, &NeuroDataInstance, idlen, 1)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = NewNeuroData()) == NULL)
				{
					MibRemove(obj, &NeuroDataInstance, idlen, 1);
					return SNMP_GENERROR;
				}
				data = (NeuroData *) local->Data;
				if (NeuroCInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					TotalNrNetworks++;
					MESSAGE(MODULE, NEURO_CREATE);
					return SNMP_NOERROR;
				}
				NeuroCStop(data);
				DelNeuroData(data);
				MibRemove(obj, &NeuroDataInstance, idlen, 1);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		else
		{
			data = (NeuroData *) local->Data;
			switch (data->Status)
			{
			case SNMP_UNDERCREATION:
				switch (obj->Syntax.LngInt)
				{
				case SNMP_VALID:
					if (NeuroCStart(data, NeuroMCalcCallback, local->Data, data->NetRefresh) == TRUE)
					{
						data->Status = SNMP_VALID;
						RunningNrNetworks++;
						MESSAGE(MODULE, NEURO_ACTIVE);
						return SNMP_NOERROR;
					}
					return SNMP_BADVALUE;
				case SNMP_INVALID:
					if (NeuroCStop(data) == TRUE)
					{
						DelNeuroData(data);
						if (MibRemove(obj, &NeuroDataInstance, idlen, 1) == TRUE)
						{
							TotalNrNetworks--;
							MESSAGE(MODULE, NEURO_DESTROY);
							return SNMP_NOERROR;
						}
					}
					return SNMP_GENERROR;
				case SNMP_UNDERCREATION:
					return SNMP_NOERROR;
				default:
					return SNMP_BADVALUE;
				}
			case SNMP_VALID:
				switch (obj->Syntax.LngInt)
				{
				case SNMP_INVALID:
					if (NeuroCStop(data) == TRUE)
					{
						DelNeuroData(data);
						if (MibRemove(obj, &NeuroDataInstance, idlen, 1) == TRUE)
						{
							RunningNrNetworks--;
							TotalNrNetworks--;
							MESSAGE(MODULE, NEURO_DESTROY);
							return SNMP_NOERROR;
						}
					}
					return SNMP_GENERROR;
				case SNMP_UNDERCREATION:
					if (NeuroCStop(data) == FALSE)
						return SNMP_GENERROR;
					if (NeuroCSubInit(data) == FALSE)
						return SNMP_GENERROR;
					data->Status = SNMP_UNDERCREATION;
					return SNMP_NOERROR;
				case SNMP_VALID:
					return SNMP_NOERROR;
				default:
					return SNMP_BADVALUE;
				}
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

VOID NeuroMCalcCallback(TIMER_DESCR * timer, ULONG now, VOID * param)
{
	NeuroData *net = param;
	FLOAT error;
	BOOLEAN update;

	if (net->In != NULL && net->Out != NULL && net->Target != NULL)
	{
		if (net->Train == NEURO_BATCH && net->BatchIndex == net->GrantedSize)
			net->BatchIndex = 0;

		if (NeuroMib2Neuron(net, net->InputData, net->NrNeurons[0], net->InputBatch, net->In) == FALSE)
		{
			ERROR(MODULE, NEURO_NOINPUT);
			return;
		}

		if (net->Train != NEURO_NOTRAIN)
		{
			if (NeuroMib2Neuron(net, net->TargetData, net->NrNeurons[net->NrLayers - 1], net->TargetBatch, net->Target) == FALSE)
			{
				ERROR(MODULE, NEURO_NOTARGET);
				return;
			}
			NeuronInOutputCpy(net->Out, net->Target, net->NrNeurons[net->NrLayers - 1]);
		}
		else
		{
			NeuroInitOutput(net->NrNeurons[net->NrLayers - 1], net->Out);
		}

		if (net->Train != NEURO_BATCH || net->BatchIndex + 1 == net->GrantedSize)
			update = TRUE;
		else
			update = FALSE;

		if (CalcNeuronNet(net->Net,
						  net->In, net->NrNeurons[0],
						  net->Out, net->NrNeurons[net->NrLayers - 1],
						  net->Train != NEURO_NOTRAIN, update,
						  net->LearningRate, net->Momentum) != NEURO_OK)
		{
			ERROR(MODULE, NEURO_CALCERR);
			return;
		}

		if (NeuroNeuron2Mib(net->Out, net->NrNeurons[net->NrLayers - 1], net->TargetData, net->OutputValue) == FALSE)
		{
			ERROR(MODULE, NEURO_NORESULT);
			return;
		}

		if (net->Train != NEURO_NOTRAIN)
		{
			if (CalcNeuronNetError(net->Net,
								   net->Target, net->NrNeurons[net->NrLayers - 1],
								   &error) != NEURO_OK)
			{
				ERROR(MODULE, NEURO_CALCERR);
				return;
			}

			net->BatchError += error;
			if (update == TRUE || net->BatchSize < net->GrantedSize)
			{
				net->Error = net->BatchError;
				if (net->BatchSize == net->GrantedSize)
					net->BatchError = 0;
			}

			net->TrainSteps++;
		}
		else
		{
			net->Error = 0;
		}

		if (NeuroCheckThresholds(net) == FALSE)
			ERROR(MODULE, NEURO_THRESHOLD);

		if (net->Train == NEURO_BATCH && net->BatchSize < net->GrantedSize)
			net->BatchSize++;

		if (net->Train == NEURO_BATCH && net->GrantedSize > 0)
			net->BatchIndex++;

		net->TotalSteps++;
	}
	else
	{
		FATAL(MODULE, NEURO_ALLOCATE);
	}

	return;
}

BOOLEAN NeuroMib2Neuron(NeuroData * net, InOutData * mibinout, LONG n, InOutBatch * batch, InOutput * inout)
{
	SNMP_OBJECT obj = {SNMP_PDU_GET, {NMIB_BASE}, NMIB_BASELEN, SNMP_INTEGER, SNMP_VALID};
	WORD dummy;
	FLOAT value, tmpval;
	LONG i, s;

	for (i = 0; i < n; i++)
	{
		if (net->Train == NEURO_RETRAIN && net->GrantedSize == 0)
		{
			/* just use already available data */
			inout[i].value = NeuronValue((mibinout[i].TheValue - mibinout[i].MinValue) / (mibinout[i].MaxValue - mibinout[i].MinValue));
		}
		else
		{
			if ((net->Train == NEURO_BATCH || net->Train == NEURO_RETRAIN) &&
				net->GrantedSize > 0 && net->BatchSize == net->GrantedSize)
			{
				mibinout[i].TheValue = batch[net->BatchIndex * n + i].TheValue;
			}
			else
			{
				memcpy(obj.Id, mibinout[i].Id, (obj.IdLen = mibinout[i].IdLen) * sizeof(mibinout[i].Id[0]));
				if (MibRequest(1, &obj, &dummy) != SNMP_NOERROR)
					return FALSE;

				inout[i].inoutnr = i;
				switch (obj.Type)
				{
				case SNMP_INTEGER:
					value = (FLOAT) obj.Syntax.LngInt;
					break;
				case SNMP_COUNTER:
				case SNMP_GAUGE:
				case SNMP_TIMETICKS:
					value = (FLOAT) obj.Syntax.LngUns;
					break;
				default:
					ERROR(MODULE, NEURO_WRONGTYPE);
					return FALSE;
				}

				if (mibinout[i].AverageMethod != SIMPLE_DIFF && mibinout[i].AverageMethod != SIMPLE_VALUE)
					for (s = 1; s < mibinout[i].AverageSteps; s++)
						mibinout[i].Value[s] = mibinout[i].Value[s - 1];

				tmpval = value;

				switch (mibinout[i].AverageMethod)
				{
				case SIMPLE_DIFF:
					value -= mibinout[i].PrevValue;
				case SIMPLE_VALUE:
					mibinout[i].Value[0] = value;
					mibinout[i].TheValue = value;
					break;
				case MEAN_DIFF:
					value -= mibinout[i].PrevValue;
				case MEAN_VALUE:
					mibinout[i].Value[0] = value;
					mibinout[i].TheValue = NeuroMean(mibinout[i].Value, mibinout[i].AverageSteps);
					break;
				case HARMONIC_DIFF:
					value -= mibinout[i].PrevValue;
				case HARMONIC_VALUE:
					mibinout[i].Value[0] = value;
					mibinout[i].TheValue = NeuroHarmonic(mibinout[i].Value, mibinout[i].AverageSteps);
					break;
				}

				mibinout[i].PrevValue = tmpval;

				if (net->Train == NEURO_BATCH && net->GrantedSize > 0)
					batch[net->BatchIndex * n + i].TheValue = mibinout[i].TheValue;
			}

			inout[i].value = NeuronValue((mibinout[i].TheValue - mibinout[i].MinValue) / (mibinout[i].MaxValue - mibinout[i].MinValue));
		}
	}

	return TRUE;
}

FLOAT NeuroMean(FLOAT value[], LONG nrvalues)
{
	FLOAT result = 0;
	LONG i;

	for (i = 0; i < nrvalues; i++)
		result += value[i];
	return result / nrvalues;
}

FLOAT NeuroHarmonic(FLOAT value[], LONG nrvalues)
{
	FLOAT result = 0;
	LONG i;

	for (i = 0; i < nrvalues; i++)
		if (value[i] != 0)
			result += 1.0 / value[i];
	return nrvalues / (result != 0 ? result : 1);
}

BOOLEAN NeuroNeuron2Mib(InOutput * inout, LONG n, InOutData * iodata, FLOAT * out)
{
	LONG i;

	for (i = 0; i < n; i++)
		out[inout[i].inoutnr] = NeuronRealValue(inout[i].value) * (iodata[i].MaxValue - iodata[i].MinValue) + iodata[i].MinValue;
	return TRUE;
}

BOOLEAN NeuroInitOutput(LONG n, InOutput * inout)
{
	LONG i;

	for (i = 0; i < n; i++)
		inout[i].inoutnr = i;
	return TRUE;
}

BOOLEAN NeuroCheckThresholds(NeuroData * data)
{
	InOutData *datasource;
	LONG l, n;
	FLOAT value;
	BOOLEAN trap;
	ULONG systime;

	systime = SysTime();

	for (l = 0; l < data->NrLayers; l += data->NrLayers - 1)
	{
		for (n = 0; n < data->NrNeurons[l]; n++)
		{
			if (l > 0)
			{
				datasource = &data->TargetData[n];
				if (data->Train != NEURO_NOTRAIN)
					value = datasource->TheValue;
				else
					value = data->OutputValue[n];
			}
			else
			{
				datasource = &data->InputData[n];
				value = datasource->TheValue;
			}

			trap = FALSE;

			switch (datasource->ThresholdCheck)
			{
			case LESS_THAN:
				if (value < datasource->ThresholdValue[0])
					trap = TRUE;
				break;
			case GREATER_THAN:
				if (value > datasource->ThresholdValue[0])
					trap = TRUE;
				break;
			case IN_RANGE:
				if (value > datasource->ThresholdValue[0] &&
					value < datasource->ThresholdValue[1])
					trap = TRUE;
				break;
			case OUT_RANGE:
				if (value < datasource->ThresholdValue[0] &&
					value > datasource->ThresholdValue[1])
					trap = TRUE;
				break;
			case EQUAL:
				if (value == datasource->ThresholdValue[0])
					trap = TRUE;
				break;
			case NOT_EQUAL:
				if (value != datasource->ThresholdValue[0])
					trap = TRUE;
				break;
			case NO_THRESHOLD:
			default:
				break;
			}

			if (trap == TRUE)
			{
				if (datasource->Trap)
					AgentSendTrap(datasource->TrapComm, datasource->TrapCommLen, systime,
								  SNMP_TRAP_ENTSPECIFIC, TRAP_SPECIFIC_NEUROTHRESHOLD, NULL, 0);
			}
		}
	}

	return TRUE;
}

BOOLEAN NeuroAllocBatch(NeuroData * data, LONG size, BOOLEAN reset)
{
	InOutBatch *input = NULL, *target = NULL;
	LONG dsize;

	data->BatchSize = 0;
	data->BatchIndex = 0;

	if (size <= 0)
	{
		FREE(data->InputBatch);
		FREE(data->TargetBatch);
		data->InputBatch = NULL;
		data->TargetBatch = NULL;
		data->GrantedSize = 0;
		return TRUE;
	}

	if ((dsize = size / 10) == 0)
		dsize = 1;

	if (data->GrantedSize > size || reset == TRUE)
	{
		FREE(data->InputBatch);
		FREE(data->TargetBatch);
		data->InputBatch = NULL;
		data->TargetBatch = NULL;
		data->GrantedSize = 0;
	}

	while (size > 0)
	{
		input = NewInOutBatch(size * data->NrNeurons[0]);
		target = NewInOutBatch(size * data->NrNeurons[data->NrLayers - 1]);
		if (input != NULL && target != NULL)
			break;
		else
		{
			FREE(input);
			FREE(target);
			size -= dsize;
		}
	}

	if (size > data->GrantedSize)
	{
		FREE(data->InputBatch);
		FREE(data->TargetBatch);
		data->InputBatch = input;
		data->TargetBatch = target;
		data->GrantedSize = size;
	}
	else
	{
		FREE(input);
		FREE(target);
	}

	return TRUE;
}

WORD NeuroSetId(SNMP_OBJECT * obj, WORD idc, WORD indexsize, LONG c, LONG l, LONG n, LONG i)
{
	LONG baseindex[] =
	{NMIB_BASE, 0, 0, 0};

	switch (idc)
	{
	case NMIB_NRNETWORKS:
		NEUROSETID(baseindex, NMIB_NRNETWORKSID);
		break;
	case NMIB_INDEX:
		NEUROSETID(baseindex, NMIB_INDEXID);
		break;
	case NMIB_NRLAYERS:
		NEUROSETID(baseindex, NMIB_NRLAYERSID);
		break;
	case NMIB_NETREFRESH:
		NEUROSETID(baseindex, NMIB_NETREFRESHID);
		break;
	case NMIB_TRAIN:
		NEUROSETID(baseindex, NMIB_TRAINID);
		break;
	case NMIB_REQSIZE:
		NEUROSETID(baseindex, NMIB_REQSIZEID);
		break;
	case NMIB_GRANTSIZE:
		NEUROSETID(baseindex, NMIB_GRANTSIZEID);
		break;
	case NMIB_BATCHSIZE:
		NEUROSETID(baseindex, NMIB_BATCHSIZEID);
		break;
	case NMIB_LRATE:
		NEUROSETID(baseindex, NMIB_LRATEID);
		break;
	case NMIB_MOMENTUM:
		NEUROSETID(baseindex, NMIB_MOMENTUMID);
		break;
	case NMIB_TOTALSTEPS:
		NEUROSETID(baseindex, NMIB_TOTALSTEPSID);
		break;
	case NMIB_TRAININGSTEPS:
		NEUROSETID(baseindex, NMIB_TRAININGSTEPSID);
		break;
	case NMIB_ERROR:
		NEUROSETID(baseindex, NMIB_ERRORID);
		break;
	case NMIB_OWNER:
		NEUROSETID(baseindex, NMIB_OWNERID);
		break;
	case NMIB_STATUS:
		NEUROSETID(baseindex, NMIB_STATUSID);
		break;
	case NMIB_NRNEURONS:
		NEUROSETID(baseindex, NMIB_NRNEURONSID);
		break;
	case NMIB_DATASRC:
		NEUROSETID(baseindex, NMIB_DATASRCID);
		break;
	case NMIB_DATASRCMAX:
		NEUROSETID(baseindex, NMIB_DATASRCMAXID);
		break;
	case NMIB_DATASRCMIN:
		NEUROSETID(baseindex, NMIB_DATASRCMINID);
		break;
	case NMIB_DATASRCAVERAGEMETHOD:
		NEUROSETID(baseindex, NMIB_DATASRCAVERAGEMETHODID);
		break;
	case NMIB_DATASRCSTEPS:
		NEUROSETID(baseindex, NMIB_DATASRCSTEPSID);
		break;
	case NMIB_DATASRCTRAP:
		NEUROSETID(baseindex, NMIB_DATASRCTRAPID);
		break;
	case NMIB_DATASRCTRAPCOMM:
		NEUROSETID(baseindex, NMIB_DATASRCTRAPCOMMID);
		break;
	case NMIB_DATASRCTHRESHOLDCHECK:
		NEUROSETID(baseindex, NMIB_DATASRCTHRESHOLDCHECKID);
		break;
	case NMIB_DATASRCVALUE:
		NEUROSETID(baseindex, NMIB_DATASRCVALUEID);
		break;
	case NMIB_OUTPUTVALUE:
		NEUROSETID(baseindex, NMIB_OUTPUTVALUEID);
		break;
	case NMIB_DATASRCTHRESHOLDVALUE:
		NEUROSETID(baseindex, NMIB_DATASRCTHRESHOLDVALUEID);
		break;
	}

	memcpy(obj->Id, baseindex, (obj->IdLen = sizeof(baseindex)) * sizeof(baseindex[0]));
	while (obj->Id[obj->IdLen - 1] == 0)
		obj->IdLen--;

	switch (indexsize)
	{
	case NMIB_IOFFSET:
		if (i > 0)
			obj->Id[obj->IdLen - 1 + NMIB_IOFFSET] = i;
	case NMIB_NOFFSET:
		if (n > 0)
			obj->Id[obj->IdLen - 1 + NMIB_NOFFSET] = n;
	case NMIB_LOFFSET:
		if (l > 0)
			obj->Id[obj->IdLen - 1 + NMIB_LOFFSET] = l;
	case NMIB_COFFSET:
		if (c > 0)
			obj->Id[obj->IdLen - 1 + NMIB_COFFSET] = c;
	default:
		break;
	}
	return SNMP_NOERROR;
}

WORD MibName2ID(LONG * id, CHAR * name, UINT * n)
{
	LONG tmp;
	CHAR *p;
	CHAR name2[10 * SNMP_SIZE_BUFINT];

	if (name == NULL)
		return SNMP_NOSUCHNAME;

	strcpy(name2, name);

	*n = 0;

	p = strtok(name2, ".");
	while (p != NULL)
	{
		if (*n < SNMP_SIZE_BUFINT && (tmp = atoi(p)) != 0)
			id[*n++] = tmp;
		else
			return SNMP_NOSUCHNAME;
		p = strtok(NULL, ".");
	}

	return SNMP_NOERROR;
}

WORD MibID2String(const LONG * id, UINT nr, CHAR * name, INT maxlen)
{
	INT ni, n;
	UINT i;
	CHAR tmp[20];						/* ample space for 1 number and a '.'  */

	ni = 0;
	name[0] = '\0';

	sprintf(tmp, "%ld%n", id[0], &n);	/* initialize name  */
	if (ni + n > maxlen)
		return 0;
	strcpy(name + ni, tmp);
	ni += n;

	for (i = 1; i < nr; i++)
	{
		sprintf(tmp, ".%ld%n", id[i], &n);
		if (ni + n > maxlen)
			return 0;
		strcpy(name + ni, tmp);
		ni += n;
	}

	return ni;
}
