/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>
#include <dmath.h>
#include <dmemory.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mibsup.h>
#include <message.h>

#include "reache.h"
#include "reachc.h"
#include "reach.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define CONTROLINDEXSIZE    1

#define FLOAT_SCALE         1000
#define DOUBLE_SCALE        1000000000L

static MIB_LOCAL *ReachInstance = NULL;

static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize);

BOOLEAN ReachMInit(VOID)
{
	MessageConfig(REACH_ERROR, "Reachability");
	return ReachConfigInit();
}

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize)
{
	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->IdLen == idlen || (*local)->Index > obj->Id[idlen] ||
		(*local = (*local)->Next) != NULL)
	{
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + indexsize;
		return TRUE;
	}

	return FALSE;
}

WORD ReachIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachIPAddress(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, REACH_BOOLEAN_ADDRESS))
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = mem2ulong(data->IPAddress);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Reach *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		ulong2mem(data->IPAddress, obj->Syntax.LngUns);
		BooleanSetTrue(data->ObjectSet, REACH_BOOLEAN_ADDRESS);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachDelay(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->Delay;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Reach *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt <= 0 || obj->Syntax.LngInt > 2000000L)
			return SNMP_BADVALUE;
		data->Delay = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachTimeout(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->Timeout;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Reach *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt <= 10 || obj->Syntax.LngInt > 2000000L)
			return SNMP_BADVALUE;
		data->Timeout = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachValueSteps(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->ValueSteps;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Reach *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt < 1 || obj->Syntax.LngInt > 100)
			return SNMP_BADVALUE;
		data->ValueSteps = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachICMPWeight(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = (LONG) (data->ICMPWeight * DOUBLE_SCALE);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Reach *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		data->ICMPWeight = ((DOUBLE) obj->Syntax.LngInt) / DOUBLE_SCALE;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachUDPWeight(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = (LONG) (data->UDPWeight * DOUBLE_SCALE);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Reach *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		data->UDPWeight = ((DOUBLE) obj->Syntax.LngInt) / DOUBLE_SCALE;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachRemaining(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->Remaining;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Reach *) local->Data;
		if (obj->Syntax.LngInt < 1)
			return SNMP_BADVALUE;
		data->Remaining = obj->Syntax.LngInt;
		if (data->Status == SNMP_VALID)
		{
			if (ReachCStop(data) == FALSE)
			{
				DnpapMessage(DMC_ERROR, REACH_STOP, "reach: collector %ld could not be stopped", local->Index);
				return SNMP_GENERROR;
			}
			if (ReachCStart(data) == FALSE)
			{
				DnpapMessage(DMC_ERROR, REACH_START, "reach: collector %ld could not be started", local->Index);
				return SNMP_BADVALUE;
			}
		}
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachICMPReceived(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngUns = data->ICMPReceived;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachUDPReceived(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngUns = data->UDPReceived;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachICMPDropped(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngUns = data->ICMPDropped;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachUDPDropped(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngUns = data->UDPDropped;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachICMPTime(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->ICMPTime;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachUDPTime(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->UDPTime;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachICMPTimeMin(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->ICMPTimeMin;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachUDPTimeMin(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->UDPTimeMin;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachICMPTimeMax(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->ICMPTimeMax;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachUDPTimeMax(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->UDPTimeMax;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachTotalICMPTime(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->TotalICMPTime;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachTotalUDPTime(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->TotalUDPTime;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachValue(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data;
	DOUBLE ICMPvalue, UDPvalue, subtotalICMPtime, subtotalUDPtime;
	LONG i, ICMPsteps, UDPsteps;

	if ((local = MibRmon(obj, ReachInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		/* do calculation here because the routines */
		/* in reachc.c may never be called          */
		ICMPsteps = LongMin(data->ICMPReceived, data->ValueSteps);
		UDPsteps = LongMin(data->UDPReceived, data->ValueSteps);
		/* determine the harmonic mean */
		for (i = 0, subtotalICMPtime = 0; i < ICMPsteps; i++)
			if (data->PingBuffer[i] != 0)
				subtotalICMPtime += (DOUBLE) 1.0 / data->PingBuffer[i];
		for (i = 0, subtotalUDPtime = 0; i < UDPsteps; i++)
			if (data->EchoBuffer[i] != 0)
				subtotalUDPtime += (DOUBLE) 1.0 / data->EchoBuffer[i];
		ICMPvalue = (ICMPsteps == 0 || subtotalICMPtime == 0)
			? (DOUBLE) 1.0
			: (DOUBLE) data->ICMPWeight * (ICMPsteps / subtotalICMPtime - data->ICMPTimeMin);
		UDPvalue = (UDPsteps == 0 || subtotalUDPtime == 0)
			? (DOUBLE) 1.0
			: (DOUBLE) data->UDPWeight * (UDPsteps / subtotalUDPtime - data->UDPTimeMin);
		if ((data->Value = (DOUBLE) 1.0 - ICMPvalue - UDPvalue) < 0)
			data->Value = (DOUBLE) 0.0;
		obj->Syntax.LngInt = (LONG) (data->Value * FLOAT_SCALE);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachOwner(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data = NULL;

	if ((local = MibRmon(obj, ReachInstance, idlen, 1)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Reach *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, REACH_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Owner, obj->SyntaxLen = data->OwnerLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Reach *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, obj->Syntax.BufChr, data->OwnerLen = obj->SyntaxLen);
		BooleanSetTrue(data->ObjectSet, REACH_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ReachStatus(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	Reach *data;

	local = MibRmon(obj, ReachInstance, idlen, 1);

	switch (obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, 1) == FALSE)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (Reach *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(obj, &ReachInstance, idlen, 1)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(Reach))) == NULL)
				{
					MibRemove(obj, &ReachInstance, idlen, 1);
					return SNMP_GENERROR;
				}
				data = (Reach *) local->Data;
				if (ReachCInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, REACH_CREATE, "reach: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(obj, &ReachInstance, idlen, 1);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (Reach *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (ReachCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, REACH_ACTIVE, "reach: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (ReachCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, REACH_DESTROY, "reach: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &ReachInstance, idlen, 1) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_INVALID:
				if (ReachCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, REACH_DESTROY, "reach: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &ReachInstance, idlen, 1) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (ReachCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			case SNMP_VALID:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
