/************************************************************************
** MODULE INFORMATION*
**********************
**     FILE     NAME:       neusinit.c
**     SYSTEM   NAME:       
**     ORIGINAL AUTHOR(S):  Paul Lemmers
**     VERSION  NUMBER:     
**     CREATION DATE:       1991/11/23
**
** DESCRIPTION: This file contains the handling of the Deveice driver
**              INIT function.
**              It runs with DS!=SS at ring 3.
**              
*************************************************************************
** CHANGES INFORMATION **
*************************
** REVISION:    $Revision:   1.8  $
** WORKFILE:    $Workfile:   neusinit.c  $
** LOGINFO:     $Log:   D:/CPROG/NDIS/NEUS/VCS/NEUSINIT.C_V  $
**              
**                 Rev 1.8   30 Mar 1993 09:39:28   PAUL
**              OS/2 version check added
**              
**                 Rev 1.7   06 Dec 1992 15:55:12   PAUL
**              Version number 0.1.1 for multiple loading of driver
**              
**                 Rev 1.6   20 Sep 1992 21:28:24   PAUL
**              DevHelp_SemHandle addition
**              
**                 Rev 1.5   15 Jul 1992 13:42:26   PAUL
**              Modifications for inactivity timer
**              
**                 Rev 1.4   12 Mar 1992 22:31:46   PAUL
**              High Resolution timer
**              
**                 Rev 1.3   05 Mar 1992 13:15:30   PAUL
**              New version string
**              
**                 Rev 1.2   24 Feb 1992 18:23:46   PAUL
**              Version number update
**              
**                 Rev 1.1   21 Feb 1992 22:25:52   PAUL
**              Version number update
**              
**                 Rev 1.0   16 Feb 1992 18:22:26   PAUL
**              Initial revision.
*************************************************************************/

#include <os2def.h>

#include <ndisdef.h>
#include <neus.h>
#include <devlib.h>
#include <devdos.h>
#include <cdev.h>
#include <protparm.h>

#include "neusglb.h"



/* A lot of Init data    */

char myname[] = {"NEUS0$"};
char hallo[] = {" driver v1.0.0\r\n"};
char ProtName[] = {"/DEV/PROTMAN$"};
char ModSign[] = {"NEUS0"};

#define MIN_OS2_MAJ   1
#define MIN_OS2_MIN   20

PARMD	myparms[] = {
	{ '\1', '\1', 16,
	"BINDINGS",	0L, 0L,   myBindings.ModName1, NULL},

	{ '\1', '\1',  0,
	"DRIVERNAME",	0L, 0L,   NULL, NULL},

	{ '\0', '\1',  4,
	"INTERFACEFLAGS", 0L, -1L, &(myPLD.pld_InterfF), NULL},

	{ '\0', '\1',  2,
	"MAXBLOCKS",	2L, 16L,  &MaxBlocks, NULL},

	{ '\0', '\1',  2,
	"MAXFILTERS",	2L, 16L,  &MaxFilters, NULL},

	{ '\0', '\1',  2,
	"MAXRECEIVES",	1L, 20L,  &MaxRx, NULL},

	{ '\0', '\1',  2,
	"MAXREQUESTS",	1L, 12L,  &MaxGenReq, NULL},

	{ '\0', '\1',  2,
	"MAXSEG",	1L, (long)(MAXSEG-1), &maxsel, NULL},

	{ '\0', '\1',  2,
	"MAXTRANSMITS",	1L, 12L,  &MaxTx, NULL}
	};
#define NPARMS (sizeof(myparms)/sizeof(*myparms))

char *ConfErr[] = {
"NEUS0%03dE MACSPY could not obtain Global Infoseg pointer\r\n",
#define ERR_INFOPTR	0
"NEUS0%03dE Well, well, 10 NEUS drivers loaded?\r\n",
#define ERR_10NEUS	1
"NEUS0%03dE Couldn't open %s\r\n",
#define ERR_NO_PROTMAN	2
"NEUS0%03dE IOCtl (GetInfo) on Protman failed %u and %x\r\n",
#define ERR_GET_PROTINF	3
"NEUS0%03dE NDIS level (%c.%c) of %s to low\r\n",
#define ERR_LOW_NDIS	4
"NEUS0%03dE %s found for multiple sections\r\n",
#define ERR_MUL_SECT	5
"NEUS0%03dE No section found in %s\r\n",
#define ERR_NO_SECT	6
"NEUS0%03dE IOCtl (RegisterModule) failed %u %x\r\n",
#define ERR_REG_FAIL	7
"NEUS0%03dE OS/2 1.20 or higher required\r\n",
#define ERR_OLD_OS      8
"NEUS0%03dE Not enough memory in DGROUP for Alloc\r\n",
#define ERR_OUT_MEM	9
"NEUS0%03dE Couldn't allocate %d GDT selectors\r\n"
#define ERR_ALLOC_GDT	10
"NEUS0%03dE OS/2 error code %u at location %d\r\n"
#define ERR_OS2		11
};

void   OutConfErr(USHORT erno, USHORT ArgCnt, ...);
int    CheckSystem(void);
int    WhatIsMyName(void);
USHORT ParseInfo(void);
int    Register(USHORT fh);
int    AllocVar(void);
void   InitVar(void);

/**************************************************************
** NAME:        Init
** SYNOPSIS:    void Init(INIT_R _far *req_p)
**              INIT_R _far *req_p
** DESCRIPTION: 
** RETURNS:     nothing
**************************************************************/
void	Init(INIT_R _far *req_p)
{
	USHORT	fh;

	/* clear _BSS segment in DGROUP */
	init_bss();
/*	SaveSel(InitSel); */

	/* save pointer to devhelp */
	DevHelp = req_p->rdata.sizes_and_ptr.dh_ptr;

	/* answer the request */
	req_p->rdata.nr_units = 0;

#if defined(DEBUG)
	dev_printf("parms: >%Fs<\r\n", req_p->rdata.argp);
#endif

	if (	CheckSystem() ||
			WhatIsMyName() ||
			((fh=ParseInfo()) == -1) ||
			Register(fh) )
		{
		req_p->rdata.sizes_and_ptr.sizes.code_size = 0;
		req_p->rdata.sizes_and_ptr.sizes.data_size = 0;
		}
	else
		{
		req_p->rdata.sizes_and_ptr.sizes.code_size = (UI) freeit;
		req_p->rdata.sizes_and_ptr.sizes.data_size = freemem;
		}

	req_p->rdata.argp = FNULL;
	req_p->hdr.status = STAT_DONE;
}

#define va_arg(ap,t) ((t _far *)(ap += sizeof(t)))[-1]
/**************************************************************
** NAME:        OutConfErr
** SYNOPSIS:    void OutConfErr(unsigned short erno,
**				KEYW_E _far *keywp, ...)
**              USHORT erno;   error number
**              USHORT ArgCnt; # of arguments
** DESCRIPTION: Function to write an error message about configuration
**              parameters to the screen
** RETURNS:     nothing
**
** I know! This function is nonsense. I must have been drunk when
** I wrote this! (Paul, 1993)
**
**************************************************************/
void	OutConfErr(USHORT erno, USHORT ArgCnt, ...)
{
	char _far *ap;

	ap = (char _far *)&ArgCnt + sizeof(ArgCnt);

	switch ( ArgCnt )
		{
		case 0:
			dev_printf(ConfErr[erno], erno);
			break;
		case 1:
			dev_printf(ConfErr[erno], erno,
				va_arg(ap, short));
			break;
		case 2:
			dev_printf(ConfErr[erno], erno,
				va_arg(ap, short), va_arg(ap, short));
			break;
		case 3:
			dev_printf(ConfErr[erno], erno,
				va_arg(ap, short), va_arg(ap, short),
				va_arg(ap, short));
			break;
		}
}

/**************************************************************
** NAME:        CheckSystem
** SYNOPSIS:    int CheckSystem(void)
** DESCRIPTION: Get global Info segment for time stamps and checks
**              the system for OS/2 version....
** RETURNS:     0 -->   no error
**              else    error code
**************************************************************/
int    CheckSystem(void)
{
	USHORT	usRet;
	SEL	selLoc;
	SEL	selGlb;
	PUSHORT psegTime;

	/* Get ring 0 global Infoseg pointer */
	if ( dev_GetDOSVar(1, &psegTime) )
		{
		OutConfErr(ERR_INFOPTR, 0);
		return(1);
		}
	else
		{
		/* poiter to SysINFOSeg ( GINFOSEG ) */
		pgisGlobTime = MAKEP(*psegTime, 0);
		}

	/* Get ring 3 global Infoseg pointer */
	if ( (usRet=DosGetInfoSeg(&selGlb, &selLoc)) == 0 )
		{
		PGINFOSEG pgisVerCheck;
		USHORT usLow, usHigh;

		pgisVerCheck = MAKEP(selGlb, 0);
		usLow = pgisVerCheck->uchMinorVersion;
		usHigh = pgisVerCheck->uchMajorVersion/10;
		if ( (usHigh < MIN_OS2_MAJ) ||
			   ((usHigh==MIN_OS2_MAJ) && (usLow<MIN_OS2_MIN)) )
			{
			OutConfErr(ERR_OLD_OS, 0);
			return(-1);
			}
		}
	else
		{
		OutConfErr(ERR_OS2, 2, usRet, 1);
		return(-1);
		}

	return(0);
}

/**************************************************************
** NAME:        WhatIsMyName
** SYNOPSIS:    int  WhatIsMyName(void)
** DESCRIPTION: Finds a device driver name for in the device header by
**              trying to open the name "MACSPY?$", where ? in [0-9].
**              If a driver is already present the the open succeeds and
**              when it is not present then the name is still free and
**              will be used for this incarnation of the driver.
**              This way we can use one executable for two or more
**              distinct drivers.
**              When a new name is found it is copied to the driver header
**              and a hello message is displayed.
** RETURNS:     0 -->   no error
**              else    error code
**************************************************************/
int  WhatIsMyName(void)
{
	int	i;
	int	retc;
	USHORT	fh;
	USHORT	usAction = 0;
	char	*hp;

	for ( i=0 ; i<10 ; i++ )
		{
		myname[4] = (char)('0' + i);
		if ( (retc=DosOpen(myname, &fh, &usAction, 0L,
				0, 1, 0xC2, 0L)) != 0 )
			{
			/* Yes, Open Error ==> this name is new!!! */
			break;
			}
		else
			{
			DosClose(fh);
			}
		}

	if ( retc == 0 )
		{
		OutConfErr(ERR_10NEUS, 0);
		return(1);
		}

	hp = (char *) 0xa; /* device name in device header */
	i = 0;
	while ( i<8 && myname[i] )
		{
		*hp++ = myname[i++]; /* copy device name */
		}
	dev_printf("Paul's %s%s", myname, hallo);

	return(0);
}

/**************************************************************
** NAME:        ParseInfo
** SYNOPSIS:    USHORT ParseInfo(void)
** DESCRIPTION: This function opens protman, tries to find its section
**              and parses that info if found.
**              It will return an error if there is no section with the
**              correct drivername,if there ais more then one such a
**              section or if there is an parameter error in the section.
** RETURNS:     -1      if error
**              else    file handle of PROTMAN
**************************************************************/
USHORT ParseInfo(void)
{
	int	retc;
	USHORT	fh;
	USHORT	usAction = 0;
	unsigned char *hp;
	unsigned char _far *fhp;
	MOD_E _far *ModP;

	if ( (retc=DosOpen(ProtName, &fh, &usAction, 0L, 0, 1, 0xC2, 0L)) != 0 )
		{
		OutConfErr(ERR_NO_PROTMAN, 1, ProtName);
		return(-1);
		}

	ProtReq.Opcode = GetProtocolManagerInfo;
	if (	(retc=DosDevIOCtl(0L, &ProtReq, 0x58, 0x81, fh)) ||
		(ProtReq.Status != 0) )
		{
		/* That was terribly wrong */
		OutConfErr(ERR_GET_PROTINF, 2, retc, ProtReq.Status);
		DosClose(fh);
		return(-1);
		}

	ProtVer[0] = (MBYTE) ProtReq.Word1;
	ProtVer[1] = (MBYTE) (ProtReq.Word1 >> 8);
	if ( ProtVer[0] < 1 )
		{
		OutConfErr(ERR_LOW_NDIS, 3, ProtVer[0]+'0', ProtVer[1]+'0',
			ProtName);
		return(-1);
		}

	ModP = (MOD_E _far *)ProtReq.Pointer1;

	retc = 0;
	while ( (long)ModP != 0L )
		{
		if ( IsMySection(ModP, myname) )
			{
			hp = mycct.cct_mod_name;
			if ( *hp )
				{
				OutConfErr(ERR_MUL_SECT, 1, myname);
				}
			else
				{
				fhp = ModP->ModName;
				while ( (*hp++ = *fhp++) != 0 )
					; /* copy module name */

				retc = DoSection(ModP, myparms, NPARMS);
				}
			}
		ModP = ModP->Next;
		}

/*	test_counter(); */

	if ( mycct.cct_mod_name[0] && (retc == 0))
		{
		return(fh);
		}
	else
		{
		if ( mycct.cct_mod_name[0] == 0 )
			OutConfErr(ERR_NO_SECT, 1, ProtName);

		DosClose(fh);
		return(-1);
		}
}

/**************************************************************
** NAME:        Register
** SYNOPSIS:    int  Register(USHORT fh)
**              USHORT fh;     file handle for protman
** DESCRIPTION: After parameter verification, this function can be called
**              to tell the Protocol Manager that we are alive and kicking.
**              But first all memory will be reorganised by a call to
**              AllocVar.
**              After registration a call to InitVar is made which initia-
**              lises a lot of memory to 0. Now most of the INIT time
**              variables are lost!!!!!!!!
** RETURNS:     0 -->   no error
**              else    error code
**************************************************************/
int  Register(USHORT fh)
{
	int	retc;

	if ( (retc=AllocVar()) != 0 )
		return(retc);

	/* allocate GDT selectors */
	if ( dev_AllocGDTSelector(awRing0Sels, maxsel+1) )
		{
		OutConfErr(ERR_ALLOC_GDT, 1, maxsel+1);
		return(2);
		}

	ProtReq.Opcode = RegisterModule;
	ProtReq.Pointer1 = (LPCHAR)&mycct;      /* selector OK? */
	ProtReq.Pointer2 = (LPCHAR)&myBindings; /* selector OK? */

	if (	(retc=DosDevIOCtl(0L, &ProtReq, 0x58, 0x81, fh)) ||
		(ProtReq.Status != 0) )
		{
		/* That was terribly wrong */
		OutConfErr(ERR_REG_FAIL, 2, retc, ProtReq.Status);
		retc = 1;
		}

	DosClose(fh);

	InitVar();

	return(retc);
}

/**************************************************************
** NAME:        AllocVar
** SYNOPSIS:    int AllocVar()
** DESCRIPTION: This function will only set pointers to memory
**              in the _DATA segment. Later, when we are sure that we no
**              longer need the Init time memory, we can call InitVar()
**              to clear and initialise this memory. Otherwise we would
**              lose our precious strings an variables to early.
** RETURNS:     0 -->   no error
**              else    error code
**************************************************************/
int	AllocVar()
{
	int	retc = 0;

	freemem = (unsigned short int) end_asmdata;

	/* padd to double word boundary */
	if ( freemem & 0x3 )
		{
		freemem &= 0xfffc;
		freemem += 4;
		}

	/* selector descriptors array */
	aSel = (PSEGD)freemem;
	freemem += (maxsel+1) * sizeof(*aSel);

	/* Allocate queues */
	GenReqFree.Mem = (PQEL)freemem;
	GenReqFree.q_size = MaxGenReq;
	freemem += MaxGenReq * sizeof(struct _NLEV);
	TxFree.Mem = (PQEL)freemem;
	TxFree.q_size = MaxTx;
	freemem +=  MaxTx * sizeof(struct _NLEV);
	RxFree.Mem = (PQEL)freemem;
	RxFree.q_size = MaxRx;
	freemem += MaxRx * SIZE_RX;

	if ( freemem > (UI)&_end_data )
		{
		OutConfErr(ERR_OUT_MEM, 0);
		return(-1);
		}

	return(retc);
}

/**************************************************************
** NAME:        InitVar
** SYNOPSIS:    void InitVar(void)
** DESCRIPTION: Initialises the "Dynamic memory" of the driver
**              to 0 and initialises the queues.
**              WARNING
**              All init variables except those in mspend.asm
**              (including error strings!!) will be lost!!!
** RETURNS:     nothing
**************************************************************/
void	InitVar(void)
{
	/*************************************************/
	/* O.K. once we are here we can start "reusing"  */
	/* the memory.                                   */
	/*************************************************/

	/* clear memory */
	dev_memset(end_asmdata, 0, freemem - (unsigned short int)end_asmdata);

}
