;************************************************************************
;* MODULE INFORMATION*
;**********************
;*     FILE     NAME:       neusmain.asm
;*     SYSTEM   NAME:       
;*     ORIGINAL AUTHOR(S):  Paul Lemmers
;*     VERSION  NUMBER:     
;*     CREATION DATE:       1991/11/12
;*
;* DESCRIPTION: Main file for NEUS device driver.
;*              The NEUS NDIS module supports any MAC driver under it and
;*              reports responses to a ring 3 application by means of IOCtl
;*              for futher analysis.
;*              
;************************************************************************
;* CHANGES INFORMATION **
;************************
;* REVISION:    $Revision:   1.10  $
;* WORKFILE:    $Workfile:   neusmain.asm  $
;* LOGINFO:     $Log:   D:/CPROG/NDIS/NEUS/VCS/NEUSMAIN.ASV  $
;*              
;*                 Rev 1.10   30 Mar 1993 09:40:06   PAUL
;*              Multisend removed; MASM 6.00B prototypes
;*              
;*                 Rev 1.9   06 Dec 1992 15:53:04   PAUL
;*              Changes for multiple loading of driver. Starting with
;*              an unknown release of OS/2, OS/2 inpects the Device
;*              Header for an already installed driver. When the loader
;*              sees a duplicate name it will try to Deinstall the
;*              previous driver with that name. If that fails it will
;*              not load the new driver. Therefore the name NEUS0$ in
;*              the header is modified to NEUSX$. During Init packet
;*              processing the real name is set (as it was done before).
;*              The driver recieves an IOCtl call Category 5h
;*              (Printer Control), Function 48h (PRT_ACTIVATEFONT).
;*              Why, what does that mean?
;*              
;*              
;*                 Rev 1.8   20 Sep 1992 21:28:24   PAUL
;*              DevHelp_SemHandle addition
;*              
;*                 Rev 1.7   15 Jul 1992 13:42:26   PAUL
;*              Modifications for inactivity timer
;*              
;*                 Rev 1.6   31 Mar 1992 09:06:38   PAUL
;*              Ahum....
;*              
;*                 Rev 1.5   12 Mar 1992 22:31:48   PAUL
;*              High Resolution timer
;*              
;*                 Rev 1.4   05 Mar 1992 13:15:40   PAUL
;*              MULTI_SEND hacked into this part of the driver
;*              
;*                 Rev 1.3   26 Feb 1992 18:10:58   PAUL
;*              StatusIndication without IndicationComplete in same
;*              interrupt fixed.
;*              
;*                 Rev 1.2   24 Feb 1992 18:22:46   PAUL
;*              RcvChkFreed optimizations and first simple length
;*              FrameFilters implemented.
;*              
;*                 Rev 1.1   21 Feb 1992 22:24:26   PAUL
;*              Updates for always pxNextEvent pointer in event
;*              records. (Most of them in Chkfreed and MallocEvent).
;*              Updates for LOST_FRAMES event generation in
;*              FlushBlockedQue.
;*              
;*              
;*                 Rev 1.0   16 Feb 1992 18:23:40   PAUL
;*              Initial revision.
;************************************************************************/
;
	INCLUDE neussdef.inc
	NAME    NEUS


	.XLIST
	.XCREF
	INCLUDE devhlp.inc
	INCLUDE neus.inc
	INCLUDE neusglb.inc
	SUBTITLE
	.CREF
	.LIST

public	__acrtused 		; avoids force in of startup
	__acrtused = 9876h	; value not easily matched in debuggers

SAV_IF	macro
	pushf
	cli
endm

REST_IF	macro
local a
	pop	ax		;; Pop IF to ax.
	and	ax,0200h	;; Test if interrupts were enabled.
	jz	a		;; Jump if they were not.
	sti			;; Else enable interrupts.
a:
endm

NDIS_INTRO macro
	push	bp
	mov	bp,sp
	push	si
	push	di
	push	ds
	mov	ax,[bp+MYDS]	;; Get my DS
	mov	ds,ax		;; load default DS
endm

MYCOPY macro
local a
	shr	cx,1		;; cx = Length/2
	rep movsw		;; copy frame words
	jnc	a
	movsb			;; copy last byte
a:
endm
	.XLIST
	SUBTITLE	'Device driver header and global data'
	.LIST
	PAGE
_DATA	SEGMENT
;
; header of the device
header	dd	-1		;Pointer to next header
	dw	88C0h		;Device attribute:
				;  15 = 1,    Character device
				;  11 = 1,    Open/Close support
				;  9-7 = 001, MS OS/2 driver
				;  6 = 1,     Accepts Generic IOCtl
	dw	codeOFFSET _a_strat ;Offset to strategy routine
	dw	0		;reserved
	db	'NEUSX$  '	;Name, patched at Init time
ProtCS	dw	0		;reserved
	dw	0		;reserved
	dw	0		;reserved
	dw	0		;reserved
; End of device driver header
;
;
ufGlb	dw	0		; Global flags

;padded to DWORD boundary (KEEP IT THAT WAY!)

DevHelp		USPF	0	; Pointer to DevHelp entry point
pgisGlobTime	LPBUF	0	; Pointer to Global Info segment for time
dherrno		dw	0	; internal DevHelp error number
open_cnt	SWORD	0	; Open/Close device counter
;
; (Must be DWORD boundary)
; The CCT and other tables
;
mycct	_CCT	<SIZE _CCT,2,,,0,0,2,,0FFh,,1,1,,SEG mycct,SystReq,myssc,0,0,myPLD>
myssc	dw	18			; Protocol Serv. Spec. CharTab
	db	16 dup (0)
	dw	0
myPLD	_PLD	<mycct,7,_ReqConf,_TrConf,_RecvLooka,_IndCompl,_RecvInd,_StatInd>

; Protocol manager entry etc.
ProtAddr	LPFUN	0	;
ProtReq		_ProtmanReqBlock <>
ProtVer		db	0, 0	; Prototcol Manager version

myBindings	BINDL	<1,>
;	db	(16 * ....) dup (0)	; second & other bindings
;	Watch it, DOUBLE WORD allignment lost
ticks	dw	0
;	Watch it, DOUBLE WORD allignment O.K. now

	PUBLIC	_InitSel
_InitSel dw	6 dup (0)	; Init time selectors

;
; MAC Module info
LoMod		MACMOD	<>

;
; Request and Transmit Queue Descriptors
;
BlockedQue	_QDESC	<0,0,0,0>
GenReqFree	_QDESC	<0,0,0,0>
TxFree		_QDESC	<0,0,0,0>

RxQue		_QDESC	<0,0,0,0>
RxFree		_QDESC	<0,0,0,0>
;

		dw	0	; Remove this when allignment OK
inact_timer	dw	0	; Inactivity timer decremented every tick,
				; going to 0 probably means memory freed
inact_user	dw	8	; User set inactivity timer
				; This is approxomately a 0.25 s idle timer.

	PUBLIC	TimeLo, TimeHi, TimeVeryHi
TimeLo		dw	0	; inverted counter register value
TimeHi		dw	0	; Our counter
TimeVeryHi 	dw	0	; Our overflow

usMaxLen	dw	0	; Maximum data copy length (0=all)
nsel		dw	0	; actual # of event selectors
maxsel		dw	8	; max # selectors for shared event segments
				; (NOT including Tx selector). Set at
				; protocol.ini parsing time.

awRing0Sels	dw	MAXSEG dup (0)
;		ALIGN	DWORD	; Don't do this because it will skip wrong!
				; BUG: PRMSM9108005  Online: Q75280

MaxRx		dw	8	; number of delayed Rx descriptors
MaxGenReq	dw	6	; max # of general requests
MaxTx		dw	6	; max # of TransmitChain requests
MaxFilters	dw	6	; max # of ReceiveChain requests
MaxBlocks	dw	8	; max # of blocks in NDIS buffer descriptors

usLastRecsAvail dw	0	; 

usLostFrames	dw	0	; Counter of lost frames
usLostStatus	dw	0	; Counter of lost StatusIndications

usCurrentSeg	dw	1	; index of current selector descriptor
aSel		PSEGD	0	; ptr to array of segment descriptors

usFreeOffs	dw	0	; OFFSET of free (not allocated) shared memory
usOldestSeg	dw	0	; segment index in aSel of oldest event record
usOldestOffs	dw	0	; OFFSET of oldest event record
pTxSeg		PTXSEG	0	; Ptr to Transmit segment (for usRecsAvail,
				;     semaphore, etc.)

;

myAdapt	db	"NEUS net monitor module",0
;
	EVEN
_end_asmdata	equ	$
	public	_end_asmdata

_DATA	ENDS

CONST	SEGMENT
_beg_const	equ	$
	public	_beg_const
CONST	ENDS


_BSS	SEGMENT
	public	__beg_bss
__beg_bss equ	$
_BSS	ENDS
	.XLIST
	SUBTITLE	'Required global code segment entry points'
	.LIST
	PAGE
_TEXT   SEGMENT

	ASSUME	CS: _TEXT
	EXTRN	_c_strat:NEAR
	EXTRN	QueueGet:near, QueueAdd:near, QueuePeek:near, RxAlloc:near

	PUBLIC	_a_strat, _MyTickHandler
	PUBLIC	__help

_a_strat PROC FAR
	nop			; Nop for int 3 patch
	call	_trapit		; Debug call for real driver.
__help:: push	es
	push	bx		; Push far pointer to request.

	call	_c_strat

	add	sp,4		;remove request pointer from stack

	ret	
_a_strat ENDP

_MyTickHandler	PROC	FAR
	pushf
	cld
	push	ax
	push	dx
	cli
	inc	[ticks]			; only for calibration checks
	call	_time_upd		; Update high resolution timer

	test	Word Ptr [ufGlb],GF_GIVE
	jz	$L004			; jump if not tracing
	test	Word Ptr[inact_timer],0FFFFh	; timer already expired?
	jz	$L004			; jump if already expired
	dec	Word Ptr[inact_timer]	; decrement inactivity timer
	jnz	$L004			; jump if still not expired

;	Yes, the inactivity timer expired call C function to handle it
;	but first check if we have to switch to Protected mode
	push	ds			; save for mode switches
	smsw	ax			; Get MSW (=CR0)
	push	ax			; Save original MSW (=CR0)
	rcr	ax,1			; Shift PE into Carry
	jc	$L003			; Jump if already in Protected mode
	mov	dl,DevHlp_RealToProt	; DEVHELP_REALTOPROT
	call	[DevHelp]

$L003:	call	_inact_expired

	pop	ax			; Retrieve original MSW (=CR0)
	rcr	ax,1			; Shift PE into CF
	jc	$L003A			; Jump if originally in protected mode
	mov	dl,DevHlp_ProtToReal	; DEVHELP_PROTTOREAL
	call	[DevHelp]
$L003A:	pop	ds			; Restore from mode switch.

$L004:	pop	dx
	pop	ax
	popf
	ret
_MyTickHandler	ENDP

InstallTick PROC NEAR C
	xor	ax,ax
	mov	[TimeLo],ax
	mov	[TimeHi],ax
	mov	[TimeVeryHi],ax
	cli
	call	init_8253
	call	_time_upd

	mov	ax,codeOFFSET _MyTickHandler ; Ptr to Handler
	mov	dx,DevHlp_SetTimer	; SetTimer
	call	[DevHelp]
	sti

	jc	$L005
	xor	ax,ax
$L005:	ret
InstallTick ENDP

DeInstallTick PROC NEAR C
	mov	ax,codeOFFSET _MyTickHandler
	mov	dx,DevHlp_ResetTimer
	call	[DevHelp]
	jc	$L007
	xor	ax,ax
$L007:	ret
DeInstallTick ENDP
	PAGE
;
; SystemRequest entry point for the CCT
; It checks the Opcode range and jumps to the corresponding
; request in SysTab.
; All system requests are SYNCHRONOUS!
;
	EVEN
SysTab	LABEL NEAR PTR WORD
	WORD	InitB	; InitiateBind
	WORD	SysNS	; Bind
	WORD	SysNS	; InitiatePrebind Not Supported
	WORD	SysNS	; InitiateUnbind Not Supported
	WORD	SysNS	; Unbind Not Supported

	PUBLIC	SystReq

SystReq PROC NEAR
	push	bp
	mov	bp,sp
	push	si
	push	di
	push	ds

	; parameter list
	MYDS	=	6
	Opc	=	8
	Param3	=	0Ah
	Param2	=	0Ch
	Param1	=	10h

	mov	si,[bp+Opc]		; Get Opcode.
	mov	ds,[bp+MYDS]		; Set my ring 0 DS.
	call	_trapit
	dec	si			; Opcode != 0 ?
	cmp	si,Unbind
	jnc	SysNS			; Jump if out of range.

	shl	si,1			; Change Opcode into jump index.
	jmp	WORD ptr cs:[si+SysTab]	; Jump to specified request.

SysNS::	mov	ax,NDIS_INVALID_FUNCTION

SysRet:	pop	ds
	pop	di
	pop	si
	pop	bp
	retf	0Eh

;
InitB::	test	Word Ptr [ufGlb],GF_MAC_OK
	mov	ax,NDIS_GENERAL_FAILURE	; We already have a MAC layer!
	jnz	$S001			; jump if already bound
	les	bx,[bp+Param2]		; CharTab pointer
	mov	ax,es:[bx.CCT.cct_modDS]

	push	ds			; my segment
	push	dataOFFSET mycct	; my offset
	push	ds
	push	dataOFFSET LoMod	; dest for his cct ptr...
	push	0			; reserved
	push	2			; Bind Opcode
	push	ax			; his DS (es:[bx+cct_modDS])
	call	es:[bx].CCT.cct_SysReq	; Bind to me, dear MAC (please...)
	or	ax,ax			; Test Bind returncode
	jnz	$S001			; jump if error
	or	Word Ptr [ufGlb],GF_MAC_OK ; Mark as bindable for next layers

	call	_CopyMAC

	xor	ax,ax			; returncode OK
$S001:	jmp	SysRet


SystReq ENDP

_CopyMAC PROC	NEAR

	les	bx,LoMod.pCCT		; Get his CCT pointer
	mov	di,dataOFFSET LoMod.usID ; set dest to usID in local copy
	lea	si,[bx.CCT.cct_modID]	; set src to module ID in CCT
	mov	bx,ds
	mov	dx,es
	mov	ds,dx
	mov	es,bx
	mov	cx,10			; length = 2 words and 4 pointers
	rep movsw			; perform blockmove
	sub	si,4			; reset ds:si to MAC UpperDispatch
	lds	si,[si]			; get UpperDispatch pointer
	add	si,4			; set src to GenReq pointer
	mov	cx,12			; copy 6 pointers
	rep movsw

	mov	ds,bx
	ret

_CopyMAC ENDP
	.XLIST
	SUBTITLE	'Protocol LOWER dispatch table functions'
	.LIST
	PAGE
	PUBLIC	_ReqConf,_TrConf,_RecvLooka,_IndCompl,_RecvInd,_StatInd

_ReqConf PROC FAR	; RequestConfirm
	ProtID	=   10h
	MACID	=   0Eh
	ReqHandle = 0Ch
	Status_	=   0Ah
	Request =    8h

	NDIS_INTRO

	cli
	test	ufGlb,GF_GIVE
	jz	$L001			; return if not bound to MAC

	mov	si,[bp+ReqHandle]	; load pointer to event descriptor
	mov	ax,[bp+Status_]		; get final result
	mov	WORD PTR [si._NLEV.usRet2],ax ; store final result
	xor	[si._NLEV.usFlags],REQ_QUEUED ; double release check
	jnz	$L001			; jump if not completed

	call	_GetStamp		; get time stamp
	mov	WORD PTR [si._NLEV.ulTime],ax
	mov	WORD PTR [si._NLEV.ulTime+2],dx ; store time stamp

	call	ChkFreed

	call	DirectEvent

$L001:	xor	ax,ax			; set NDIS OK returncode
	pop	ds
	pop	di
	pop	si
	leave
	sti
	ret	0Ch
_ReqConf ENDP

_TrConf PROC FAR	; TransmitConfirm
	ProtID	=   0Eh
	MACID	=   0Ch
	ReqHandle = 0Ah
	Status_	=    8h

	NDIS_INTRO

	cli
	test	ufGlb,GF_GIVE
	jz	$L001			; return if not bound to MAC

	mov	si,[bp+ReqHandle]	; load pointer to event descriptor
	mov	ax,[bp+Status_]		; get final result
	mov	WORD PTR [si._NLEV.usRet2],ax ; store final result
	xor	[si._NLEV.usFlags],REQ_QUEUED ; double release check
	jnz	$L001			; jump if not completed

	call	_GetStamp		; get time stamp
	mov	WORD PTR [si._NLEV.ulTime],ax
	mov	WORD PTR [si._NLEV.ulTime+2],dx ; store time stamp

	call	ChkFreed

	call	DirectEvent

$L001:	xor	ax,ax			; set NDIS OK returncode
	pop	ds
	pop	di
	pop	si
	leave
	sti
	ret	0Ah

_TrConf ENDP

;
; The intro of this function until the GetStamp call must take
; the same number of CPU cycles as _RecvInd
_RecvLooka PROC FAR	; ReceiveLookahead
	MACID      = 14h
	FrameSize  = 12h
	BytesAvail = 10h
	Buffer     = 0Ch
	Indicate   =  8h

	BytesCopied = 2
	CopyLen    = 4
	TDESC      = 4+SIZE TDBufDescr

	push	bp
	mov	bp,sp
	sub	sp,TDESC		; Make place for local vars
	push	si
	push	di
	push	ds
	mov	ax,[bp+MYDS]		; Get my DS
	mov	ds,ax			; load default DS

	cli
	test	ufGlb,GF_GIVE
	mov	ax,NDIS_FRAME_NOT_RECOGNIZED
	jz	$L010			; return if no shared memory

	test	ufGlb,GF_BLOCKED	; test if blocked
	jz	$L002			; jump if not blocking

$L000:	inc	usLostFrames		; remember that
$L000A:	mov	ax,NDIS_FRAME_REJECTED	; discard the frame
$L010:	pop	ds
	pop	di
	pop	si
	leave
	sti
	ret	10h

$L001:	add	sp,4			; clear timestamp from stack
	jmp	short $L000

$L002:	call	GetStamp
	push	ax
	push	dx

	les	si,[bp+Buffer]		; get frame pointer
	call	FrameFilter
	jnc	$L002B			; jump if frame is not discarded
	sub	sp,4			; remove timestamp from stack
	jmp	short $L000A		; and return

$L002B:	mov	[bp-CopyLen],ax		; save copy length for later

	call	IndChkFreed

	call	FlushRcvQue

	; determine length......
	mov	ax,[bp+FrameSize]
	or	ax,ax
	jnz	$L003

	; Aaaargh, We have no idea about the frame size. We would have
	; that idea if we were a real protocol stack. This is
	; explicitly allowed by NDIS. We don't want to cause any
	; protection violations, so lets be on the safe side and
	; hope that the ring 3 application knows what to do.......
	; Gna, Gna :-)
	mov	ax,[bp+BytesAvail]	; Try to copy at least something

$L003:	mov	cx,[bp-CopyLen]		; get filter length
	or	cx,cx			; is filter length all data?
	jz	$L003A
	cmp	ax,cx			; FrameSize < Filter length
	jb	$L003A			; jump if whole frame
	mov	ax,cx			; usLen is filter length
$L003A:	cmp	ax,[bp+BytesAvail]	; enough visible?
	ja	$L004			; jump if not all visible

; We can copy frame to event record with a block move
	mov	[bp-BytesCopied],ax		; save length
	call	MallocEvent
	jz	$L001
	sti

	mov	es:[di._EVENT.usEventType],EVENT_FRAME
	pop	dx
	pop	ax
	mov	WORD PTR es:[di._EVENT.ulTime],ax
	mov	WORD PTR es:[di._EVENT.ulTime+2],dx
	mov	ax,[bp+FrameSize]
	mov	es:[di._EVENT.usRet1],ax ; frame size
	mov	cx,[bp-BytesCopied]	; get filter length

	push	di			; save event record ptr
	push	ds
	lds	si,[bp+Buffer]		; load source pointer (frame)
	add	di,_EVENT.abData	; set dest pointer (event record)
	MYCOPY
	pop	ds
	jmp	short $L006

; Let TransferData do the copy
$L004:	mov	[bp-TDESC].TDBufDescr.TDArray.Data.Len,ax ; prepare descriptor
	call	MallocEvent
	jz	$L001
	sti

	mov	es:[di._EVENT.usEventType],EVENT_FRAME
	pop	dx
	pop	ax
	mov	WORD PTR es:[di._EVENT.ulTime],ax
	mov	WORD PTR es:[di._EVENT.ulTime+2],dx
	mov	ax,[bp+FrameSize]
	mov	es:[di._EVENT.usRet1],ax ; frame size

	; Prepare TDBufDescr
	mov	[bp-TDESC].TDBufDescr.TDDataCount,1
	mov	[bp-TDESC].TDBufDescr.TDArray.PtrType,2
	mov	[bp-TDESC].TDBufDescr.TDArray.ResByte,0
	mov	ax,_EVENT.abData
	add	ax,di			; set es:ax to frame destination
	mov	WORD PTR [bp-TDESC].TDBufDescr.TDArray.Data.DataPtr,ax
	mov	ax,es
	mov	WORD PTR [bp-TDESC].TDBufDescr.TDArray.Data.DataPtr+2,ax

	push	di			; save event-
	push	es			; record ptr

	lea	ax,[bp-BytesCopied]
	push	ss
	push	ax			; push pointer to BytesCopied
	push	0			; Whole frame
	push	ss
	lea	ax,[bp-TDESC]
	push	ax			; push ptr to TDBufDescr
	push	LoMod.usDS
	call	[LoMod.pTrData]

	pop	es
$L006:	pop	di			; restore event record pointer
	mov	ax,[bp-BytesCopied]
	mov	es:[di._EVENT.usRet2],ax ; # of copied bytes

	mov	ax,1
	call	IncAvail		; Signal this frame

	xor	ax,ax			; Set all OK returncode...
	jmp	$L010

_RecvLooka ENDP

_IndCompl PROC FAR	; IndicationComplete
	MACID = 8h

	NDIS_INTRO

	cli
	test	ufGlb,GF_GIVE
	jz	$L001			; return if not bound to MAC

	call	IndChkFreed

	call	FlushRcvQue

$L001:	xor	ax,ax			; Set all OK returncode...
	and	ufGlb,NOT GF_RCV_CHK	; Clear already checked flag
	pop	ds
	pop	di
	pop	si
	leave
	sti
	ret	4
_IndCompl ENDP

;
; The intro of this function until the GetStamp call must take
; the same number of CPU cycles as _RecvLooka
_RecvInd PROC FAR	; ReceiveChain indication
	MACID	   = 14h
	FrameSize  = 12h
	ReqHandle  = 10h
	pRxBufDescr = 0Ch
	Indicate   =  8h

	CopyLen    = 2

	push	bp
	mov	bp,sp
	sub	sp,CopyLen		; Make place for local vars
	push	si
	push	di
	push	ds
	mov	ax,[bp+MYDS]		; Get my DS
	mov	ds,ax			; load default DS

	cli
	test	ufGlb,GF_GIVE
	mov	ax,NDIS_FRAME_NOT_RECOGNIZED
	jz	$L010			; return if no shared memory

	test	ufGlb,GF_BLOCKED	; test if blocked
	jz	$L002			; jump if not blocking

$L000:	inc	usLostFrames		; remember that
$L000A:	mov	ax,NDIS_FRAME_REJECTED	; discard the frame
$L010:	pop	ds
	pop	di
	pop	si
	leave
	sti
	ret	10h

$L001:	add	sp,4			; clear timestamp from stack
	jmp	short $L000

$L002:	call	GetStamp
	push	ax
	push	dx

	les	si,[bp+pRxBufDescr]	; get buffer descriptor ptr
	les	si,es:[si.BlockD.DataPtr+2] ; get first frame pointer
	call	FrameFilter
	jnc	$L002B			; jump if frame is not discarded
	sub	sp,4			; remove timestamp from stack
	jmp	short $L000A		; and return

$L002B:	mov	[bp-CopyLen],ax		; save copy length for later

	call	IndChkFreed

$L002A:	call	RxAlloc			; allocate Rx record
	jnz	$L003			; jump if successful allocation
	call	FlushRcvQue		; else flush the Receive queue
	test	ufGlb,GF_BLOCKED	; test if blocked
	jz	$L002A			; if not blocking retry allocation
	jmp	$L001			; else discard...

;
; We have allocated an RxQue element (ds:si)
; Do NOT enable interrupts because any Queue Flusher may try working
; with incomplete queue elements......
$L003:	pop	dx
	pop	ax			; restore time stamp
	mov	WORD PTR [si._DELEV.ulTime],ax
	mov	WORD PTR [si._DELEV.ulTime+2],dx ; write time stamp
	mov	ax,[bp+ReqHandle]
	mov	[si._DELEV.usHandleMAC],ax
	mov	ax,[bp+FrameSize]
	mov	[si._DELEV.usFrameSize],ax
	mov	cx,[bp-CopyLen]		; get filter length
	or	cx,cx			; is filter length all data?
	jz	$L003A
	cmp	ax,cx			; FrameSize < Filter length
	jb	$L003A			; jump if whole frame
	mov	ax,cx			; usLen is filter length
$L003A:	mov	[si._DELEV.usLen],ax

	; COPY FRAME descriptor........
	mov	ax,ds
	mov	es,ax			; prepare es as destination segment
	mov	di,SIZE _DELEV		; OFFSET to RxBufDescr in event record
	add	di,si			; make di destination pointer
	mov	cx,MaxBlocks		; get our max block count
	lds	si,[bp+pRxBufDescr]	; get MAC's RxBufDescr pointer
	lodsw				; fetch MAC block count
	cmp	ax,cx			; compare block count with our max
	jle	$L004			; jump if actual block count <= max
	mov	ax,cx			; use our MaxBlocks
$L004:	mov	cx,ax			; ax = cx = block count
	add	cx,cx			;
	add	cx,ax			; cx = 3 * ax
	stosw				; store block count
	rep movsw			; copy blocks

	mov	ax,NDIS_WAIT_FOR_RELEASE
	jmp	$L010

_RecvInd ENDP

;
; Although not obvious from the specification, it may happen
; that the IndicationComplete call doesn't follow the
; StatusIndication within the same hardware interrupt! That
; is most certainly not the case with the StartReset and EndReset
; StatusIndications!!!! Since that means that Ring 3 may be scheduled
; we cannot use IndChkFreed!!!
_StatInd PROC FAR	; Status indication
	MACID    = 10h
	Param1   = 0Eh
	Indicate = 0Ah
	Opc      =  8

	NDIS_INTRO

	cli
	test	ufGlb,GF_GIVE
	jz	$L012			; return if not bound to MAC

	test	ufGlb,GF_BLOCKED	; test if blocked
	jnz	$L011			; jump if blocking

	call	GetStamp
	push	ax
	push	dx

	call	ChkFreed		; The real check!!

	call	FlushRcvQue

	mov	ax,6			; long + short
	call	MallocEvent
	jz	$L010

	sti

	mov	es:[di]._EVENT.usEventType,EVENT_STATUS
	pop	dx
	pop	ax			; restore time stamp
	mov	WORD PTR es:[di._EVENT.ulTime],ax
	mov	WORD PTR es:[di._EVENT.ulTime+2],dx ; write time stamp
	mov	ax,[bp+Opc]
	mov	es:[di]._EVENT.usRet1,ax
	mov	ax,[bp+Param1]
	mov	es:[di]._EVENT.usRet2,ax
	add	di,SIZE _EVENT		; set es:di to 
	push	ds
	lds	si,LoMod.pSSST		; fetch Service Specific Status ptr
	add	si,MSSS.msss_status
	cli				; no interrupts during status copy
	mov	cx,3
	rep movsw
	pop	ds			; restore ds to default data segment

	mov	ax,1
	call	IncAvail		; signal this record to ring 3

	jmp	short $L012

$L010:	add	sp,4			; clear timestamp from stack
$L011:	inc	usLostStatus		; remember losing status indication

$L012:	xor	ax,ax			; set OK return code
	pop	ds
	pop	di
	pop	si
	leave
	sti
	ret	0Ch
_StatInd ENDP

	.XLIST
	SUBTITLE	'Frame filter function'
	.LIST
	PAGE
;************************************************************************
;* NAME:        FrameFilter
;* SYNOPSIS:    FrameFilter(pFrame)
;*              bytes _far *pFrame
;* DESCRIPTION: Looks at the frame contents an decides if it must
;*              be copied or not.
;*              WARNING.
;*              Be extremely carefull, this function only has the
;*              pointer to the frame; NOT its length......
;*              Lets it improve that some day, when it does
;*              something more elaborate.
;* CALL REGISTERS:
;*              es:si      pFrame
;* UNMODIFIED REGISTERS:
;*              ds bp si di
;* RETURNS:     CY=1    (carry set)
;*                   frame must be discarded
;*              CY=0    (carry clear)
;*                   ax == 0      copy whole frame
;*                   else         copy MIN(ax,len) bytes
;************************************************************************
FrameFilter PROC NEAR

	mov	ax,usMaxLen		; Simple all frames limit
	clc				; Yes copy it
	ret

FrameFilter ENDP

	.XLIST
	SUBTITLE	'Calls to and from MAC entry points'
	.LIST
	PAGE
;************************************************************************
;* NAME:        do_Send
;* SYNOPSIS:    void do_Send(struct _TxBufD _far *pTx, struct _NLEV *pEvent)
;*              
;* DESCRIPTION: This function will call the MAC's TransmitChain entry
;*              point to send the frame through the adapter.
;*              It makes a transmit descriptor on the stack and brackets
;*              the TransmitChain with IndicationOff/IndicationOn.
;* CALL REGISTERS:
;* UNMODIFIED REGISTERS:
;*              ds bp si di
;* RETURNS:     nothing
;************************************************************************
do_Send	PROC NEAR C USES di si, pTx :FAR PTR _TxBufD, pEvent :PTR _NLEV
;	push	bp
;	mov	bp,sp
;	push	si
;	push	di

;	pTx	= 4
;	pEvent	= 8

	push	LoMod.usDS
	call	LoMod.pIndOff

	push	mycct.cct_modID		; NEUS module ID
	push	[pEvent]		; Request handle
	push	[(WORD PTR pTx)+2]	; push descriptor segment
	push	[WORD PTR pTx]		; Pointer to framebufferdescriptor
	push	LoMod.usDS		; MAC DS
	call	LoMod.pTxChain		; TransmitChain

	cli
	mov	bx,[pEvent]		; get pointer to event descriptor
	mov	[bx]._NLEV.usRet1,ax	; store return code
	cmp	ax,NDIS_REQUEST_QUEUED	; test if queued
	jne	$L001			; jump if not queued

	xor	[bx]._NLEV.usFlags,REQ_ACTIVE OR REQ_QUEUED
					; test if request has completed
	jnz	$L002			; jump if not yet confirmed

$L001:	mov	si,bx			; copy event pointer to si
	call	_GetStamp		; get time stamp
	mov	WORD PTR [si._NLEV.ulTime],ax
	mov	WORD PTR [si._NLEV.ulTime+2],dx ; store time stamp
	call	DirectEvent

$L002:	push	LoMod.usDS
	call	LoMod.pIndOn

;	pop	di
;	pop	si
;	pop	bp
	sti				; IndicationOn disables; we enable
	ret

do_Send	ENDP

;************************************************************************
;* NAME:        do_GenReq
;* SYNOPSIS:    void do_GenReq(struct _iocparm _far *pParm,
;*                                    struct _NLEV *pEvent)
;*              pParm        pointer to the parameters of ring3
;*              pEvent       pointer to allocated event descriptor
;* DESCRIPTION: Performs a GeneralRequest call to the MAC
;* CALL REGISTERS:
;* UNMODIFIED REGISTERS:
;*              ds bp si di
;* RETURNS:     ax = 0 -->   no error
;*              else    error code
;************************************************************************
do_GenReq PROC NEAR C USES di si, pParm :FAR PTR _iocparm, pEvent :PTR _NLEV
;	push	bp
;	mov	bp,sp
;	push	si
;	push	di

;	pParm	= 4
;	pEvent	= 8

	les	bx,[pParm]		; get user pointer
	push	mycct.cct_modID		; NEUS module ID
	push	[pEvent]		; Request handle
	push	es:[bx].IOCPARM.usParam1
	push	WORD PTR es:[bx].IOCPARM.ulParam2+2
	push	WORD PTR es:[bx].IOCPARM.ulParam2
	push	es:[bx].IOCPARM.usSub	; push Opcode
	push	LoMod.usDS		; MAC DS
	call	LoMod.pReq		; TransmitChain

	cli
	mov	bx,[pEvent]		; get pointer to event descriptor
	mov	[bx]._NLEV.usRet1,ax	; store return code
	cmp	ax,NDIS_REQUEST_QUEUED	; test if queued
	jne	$L001			; jump if not queued

	xor	[bx]._NLEV.usFlags,REQ_ACTIVE OR REQ_QUEUED
					; test if request has completed
	jnz	$L002			; jump if not yet confirmed

$L001:	mov	si,bx			; copy event pointer to si
	call	_GetStamp		; get time stamp
	mov	WORD PTR [si._NLEV.ulTime],ax
	mov	WORD PTR [si._NLEV.ulTime+2],dx ; store time stamp
	call	DirectEvent

$L002:	sti				; be polite and enable
;	pop	di
;	pop	si
;	pop	bp
	ret

do_GenReq ENDP
	.XLIST
	SUBTITLE	'Support functions'
	.LIST
	PAGE
;************************************************************************
;* NAME:        IncSelD
;* SYNOPSIS:    IncSelD.......
;*              
;* DESCRIPTION: Sets DX to next value for usCurrentSeg and set up BX
;*              to point to the corresponding segment descriptor
;* CALL REGISTERS:
;*              dx     usCurrentSeg
;*              bx     ptr to segment descriptor
;* UNMODIFIED REGISTERS:
;*              all except dx and bx...
;* RETURNS:     dx     new usCurrentSeg
;*              bx     new ptr to segment descriptor
;************************************************************************
IncSelD	PROC	NEAR
	inc	dx			; increment usCurrentSeg
	cmp	dx,nsel			; check for wrap in descriptor array
	jbe	$L0			; jump if it is not a wrap

	mov	dx,1			; set to first event segment descr.
	mov	bx,aSel			; get begin segment descriptor array
$L0:	add	bx,SIZE SEGD		; bump descriptor pointer
	ret
IncSelD	ENDP

;************************************************************************
;* NAME:        ChkFreed
;* SYNOPSIS:    void ChkFreed(void)
;* DESCRIPTION: Checks if the ring 3 application has freed memory in
;*              the shared segments.
;*              If so it will reclaim that memory and if the state was
;*              blocked, then it will try to flush the queues.
;*              IndChkFreed is an extra entry point for frame received
;*              and status indications. Since they may come in group
;*              for one hardware interrupt from the adapter and ended
;*              with an IndicationComplete, this saves some CPU cycles...
;* CALL REGISTERS: none
;*              Interrupts must be disabled
;* UNMODIFIED REGISTERS:
;*              ds bp si di
;* RETURNS:     nothing
;************************************************************************
ChkFreed PROC	NEAR C

	jmp	$L00

; Extra entry point for receive indications
IndChkFreed LABEL NEAR
	test	ufGlb,GF_RCV_CHK
	jnz	$L00A
	or	ufGlb,GF_RCV_CHK

$L00:	push	di			; save di

	les	di,pTxSeg		; get pointer to Tx segment
	mov	ax,es:[di].TXSEG.usRecsAvail ; get usRecsAvail
	mov	cx,usLastRecsAvail
	sub	cx,ax			; cx becomes # of freed records =
					;     usLastRecsAvail - usRecsAvail
	ja	$L001			; jump if anything freed

	pop	di			; restore di
$L00A:	ret				; return, nothing was freed

;
; Yes something has definitely been freed. How much?
$L001:	or	ax,ax			; usRecsAvail equals 0?? 
	jnz	$L010			; jump if usRecsAvail nonzero

;
; usRecsAvail == 0. That means everything is free, reset everything.
; This code can be optimized further, but now it clears the whole array
; which may cleanup things after (fatal!!) bugs.....
	mov	inact_timer,ax		; clear inactivity timer
	mov	usLastRecsAvail,ax	; usLastRecsAvail = 0
	mov	usOldestOffs,ax		; usOldestOffs = 0
	mov	usFreeOffs,ax		; usFreeOffs = 0
	mov	usCurrentSeg,1		; usCurrentSeg = 1
	mov	usOldestSeg,1		; usOldestSeg = 1
	mov	cx,nsel			; set cx to number of event segments
	mov	bx,aSel			; get begin segment descriptor array
$L002:	add	bx,SIZE SEGD		; bump pointer
	mov	[bx.SEGD.fAttr],SEG_MORE ; Reset segment flags to free
	mov	[bx.SEGD.usBegin],ax	; Reset usBegin to 0
	loop	$L002			; next segment descriptor
	jmp	$L020			; goto flush test...

;
; Something has been freed, but not everything. Follow the pointers
; until usLastRecsAvail becomes usRecsAvail
; On entry:
;    ax = usRecsAvail
;    cx = number of records that were freed (must be reclaimed)
; During loop:
;    bx = pointer in aSel (segment descriptor array)
;    cx = number of records that REMAIN to be freed (must be reclaimed)
;    dx = usCurrentSeg
;    es:di = pointer to event record
$L010:	mov	usLastRecsAvail,ax	; usLastRecsAvail = usRecsAvail
	mov	dx,usOldestSeg
	imul	bx,dx,SIZE SEGD		; bx = offset in aSel
	add	bx,aSel			; let bx point to the descriptor
	mov	di,usOldestOffs		; load offset oldest event record
	mov	ax,[bx.SEGD.usSeg0]	; get selector of oldest event
	mov	es,ax

$L011:	mov	ax,es:[di]		; get OFFSET(next)
	or	ax,ax			; test if next is in new segment
	jz	$L030			; jump if in new segment

	mov	[bx.SEGD.usBegin],ax	; set usBegin to new offset
	mov	di,ax			; and load new offset of new event

$L012:	loop	$L011			; repeat for all other freed records
	mov	usOldestOffs,di		; save the last new offset 

;
; Reclaim of memory has finished, now check if we were blocked and
; try to flush if that is true
$L020:	test	ufGlb,GF_BLOCKED
	jz	$L021

	push	si
	call	FlushBlockedQue
	pop	si

$L021:	pop	di
	ret

;
; es:di points to last event in this segment
$L030:	mov	[bx.SEGD.fAttr],SEG_MORE ; set usAttr to SEG_MORE
	mov	[bx.SEGD.usBegin],ax	; reset usBegin to 0
	mov	di,ax			; new event offset (0) to di
	call	IncSelD
	mov	usOldestSeg,dx		; store new usOldestSeg
	mov	ax,[bx.SEGD.usSeg0]	; load new ring 0 selector into ES
	mov	es,ax			; es:di is again pointer to event
	jmp	short $L012

ChkFreed ENDP

;************************************************************************
;* NAME:        MallocEvent
;* SYNOPSIS:    PEVENT MallocEvent(unsigned short usSize)
;*              unsigned short usSize     Size of abData array
;* DESCRIPTION: Will try to allocate the requested event record + usSize
;*              bytes for the abData array in the shared memory segments.
;*              Effectively that means 4(for ptr) + STD_EVENT + usSize bytes,
;*              rounded upwards to an integer multiple of 4.
;*              usFreeOffs and usCurrentSeg will be updated only if
;*              allocation was succesful.
;* CALL REGISTERS:
;*              ax        usSize
;* UNMODIFIED REGISTERS:
;*              ds bp
;* RETURNS:     Zero flag set      Not enough memory for usSize
;*              else               es:di holds pointer to allocated memory
;************************************************************************
MallocEvent PROC	NEAR

; During function execution:
;    bx = pointer in aSel (segment descriptor array)
;    cx = allocation length     ; later, new free OFFSET
;    dx = usCurrentSeg
;    es:di = pointer to event record
;    ds:si = previous segment descriptor pointer if split or skip
;
	add	ax,STD_EVENT_SIZE+3+4	; add default event record size
	and	ax,NOT 3		; truncate to multiple of 4 bytes
	mov	cx,ax			; move allocation size to cx
	mov	dx,usCurrentSeg		; fetch usCurrentSeg
	imul	bx,dx,SIZE SEGD		; bx = offset in aSel
	add	bx,aSel			; let bx point to the descriptor

	mov	ax,usLastRecsAvail	; get usRecsAvail
	or	ax,ax			; test if usRecsAvail not zero
	jnz	$L001			; jump if not 0

;
; Trivial case. Allocate first event record in segment
pat0:	xor	di,di			; set OFFSET to 0

; general pointer patch and return.....
pat1:	mov	ax,[bx.SEGD.usSeg0]	; get ring 0 selector
	mov	es,ax
pat3:	mov	es:[di],cx		; store offset in record
	mov	ax,[bx.SEGD.usSeg3]	; get new selector
	mov	es:[di+2],ax		; store selector in record
	or	ax,ax			; clear zero flag
	mov	usFreeOffs,cx		; set new Free offset
	ret

;
; Not first and only event in first segment
$L001:	add	cx,STD_EVENT_SIZE+4	; Add extra event record size
	mov	di,usFreeOffs		; prepare attempt for new offset
	test	[bx.SEGD.fAttr],SEG_MORE ; allocate at the end of the segment?
	jz	$L010			; jump if must be at the begin

;
; Allocate at the end of the segment. (Or in next...)
	mov	ax,di			; get begin of event record
	add	ax,cx			; add requested + extra size
	jc	$L004			; jump if not both in this segment

allok:	mov	cx,ax			; move calculated offset to cx
	sub	cx,STD_EVENT_SIZE+4	; restore to new free offset
	jmp	pat1

;
; Not both in this (current) segment. First try only
; requested record in current segment, free in next segment.
; If that fails try both in new segment.
$L004:	mov	si,bx			; save current segment descriptor
	call	IncSelD			; move to next segment
	sub	ax,STD_EVENT_SIZE+4	; subtract extra record size
	ja	$L006			; jump if that doesn't fit too

;
; Requested fits in current segment; try if extra record fits in
; new segment.
	test	[bx.SEGD.fAttr],SEG_MORE ; allocate first of the segment?
	jnz	$L005			; jump if extra is first in new segment
	mov	cx,[bx.SEGD.usBegin]	; get freed length in ne segment
	cmp	cx,STD_EVENT_SIZE+4	; extra fits here?
	jc	nomem

;
; Allocation by split. (Free in next segment)
$L005:	mov	usCurrentSeg,dx		; set new current segment index
	and	[si.SEGD.fAttr],NOT SEG_MORE ; remove SEG_MORE flag
	mov	ax,[si.SEGD.usSeg0]	; get ring 0 selector
	mov	es,ax
	xor	cx,cx			; new free offset is 0
	jmp	pat3

;
; Try to go with both requested and extra record to new segment.
$L006:	test	[bx.SEGD.fAttr],SEG_MORE ; allocate first of the segment?
	jnz	$L007			; jump if fits as first in new segment

;
; Try both at the begin, there is something at the end....
	cmp	cx,[bx.SEGD.usBegin]	; size both < usBegin
	jnc	nomem			; jump if there is not enough space

;
; Yes allocate in new segment (at offset 0) and mark last (full)
; segment descriptor as NOT SEG_MORE.
; Since we skipped the last part of the current segment we must
; generate a dummy record.
$L007:	mov	usCurrentSeg,dx		; set new current segment index
	and	[si.SEGD.fAttr],NOT SEG_MORE ; remove SEG_MORE flag
	sub	cx,STD_EVENT_SIZE+4	; restore to new free offset

	mov	ax,[si.SEGD.usSeg0]	; get ring 0 selector
	mov	es,ax
	xor	ax,ax			; prepare next offset
	mov	es:[di],ax		; store next offset (0000)
	mov	ax,[bx.SEGD.usSeg3]	; get next selector
	mov	es:[di+2],ax		; store next selector
	mov	es:[di._EVENT.usEventType],EVENT_DUMMY
					; make it an EVENT_DUMMY

	push	bx
	push	cx
	mov	ax,1			; just one record
	call	IncAvail		; signal DUMMY
	pop	cx
	pop	bx

	jmp	pat0			; handle as if first record

;
; This is an allocation of both requested and extra at the begin of the
; segment
$L010:	mov	ax,di			; get begin of event record
	add	ax,cx			; add requested size
	jc	nomem			; jump if doesn't fit in segment
	cmp	ax,[bx.SEGD.usBegin]	; check if fits before usBegin
	jbe	allok			; jump if it fits

;
; Not enough memory.......
nomem:	or	ufGlb,GF_BLOCKED	; set blocked flag
	xor	ax,ax			; set zero flag
	ret

MallocEvent ENDP

;************************************************************************
;* NAME:        C_MallocEvent
;* SYNOPSIS:    PEVENT C_MallocEvent(unsigned short usExtra)
;*              unsigned short usExtra      # number of extra bytes
;* DESCRIPTION: C language equivalent of MallocEvent, but conforms to C
;*              calling convention
;* CALL REGISTERS:
;* UNMODIFIED REGISTERS:
;*              ds bp si di
;* RETURNS:     dx:ax = 0 -->   not enough memory
;*              else            ptr
;************************************************************************
C_MallocEvent PROC NEAR C USES di si, usExtra :WORD

;	usExtra	= 4

;	push	bp
;	mov	bp,sp
;	push	si
;	push	di

	mov	ax,[usExtra]
	call	MallocEvent		; try to allocate
	jnz	ok
	xor	ax,ax
	xor	dx,dx			; prepare NULL pointer
	jmp	short done
ok:	mov	ax,di			; get offset
	mov	dx,es			; get selector

done:	ret
;	pop	di
;	pop	si
;	pop	bp
;	ret

C_MallocEvent ENDP

;************************************************************************
;* NAME:        FlushRcvQue
;* SYNOPSIS:    void FlushRcvQue(void)
;* DESCRIPTION: Flushes the RcvQue. Tries to allocate shared memory and
;*              copy the frame followed by a RcvRelease to the MAC.
;*              If there is not enough memory then the frame will be
;*              discarded, and the GF_BLOCKED flag in ufGlb will be set.
;* CALL REGISTERS:
;*              Interrupts must be disabled on entry
;* UNMODIFIED REGISTERS:
;*              ds bp
;* RETURNS:     Nothing
;************************************************************************
FlushRcvQue PROC NEAR

$L000:	mov	di,dataOFFSET RxQue
	call	QueueGet		; get oldest rcv entry
	jnz	$L001

	ret

$L001:	test	ufGlb,GF_BLOCKED	; are we already discarding?
	jnz	$L005			; jump to discard code if so

	push	si			; save the event pointer
	mov	ax,[si]._DELEV.usLen
	call	MallocEvent		; Try to allocate shared memory
	pop	si			; restore event ptr
	jz	$L005			; jump if allocation failed

$L002:	sti				; everything allocated; enable INT
	mov	es:[di]._EVENT.usEventType,EVENT_FRAME
	mov	ax,WORD PTR[si].DELEV.ulTime
	mov	WORD PTR es:[di]._EVENT.ulTime,ax
	mov	ax,WORD PTR[si].DELEV.ulTime+2
	mov	WORD PTR es:[di]._EVENT.ulTime+2,ax
	mov	ax,WORD PTR[si].DELEV.usFrameSize
	mov	es:[di]._EVENT.usRet1,ax
	mov	cx,[si].DELEV.usLen	; get copy size

	push	si			; save queue element pointer
	push	di			; save event pointer
	add	si,SIZE DELEV
	add	di,_EVENT.abData	; set es:di to frame destination
	lodsw				; get block count
	mov	bx,ax
	mov	ax,cx			; set remaining copy size
	xor	dx,dx			; reset total size

; now the frame copy action....
; register usage:
;   ax	remaining copy length
;   bx	blockcount
;   cx  copy size
;   dx  Total bytes copied
	or	bx,bx
	jz	$L009			; jump if all blocks copied

$L003:	push	ds			; save default data segment
	push	si			; save block descriptor pointer
	mov	cx,[si].BlockD.Len	; load length of this block
	cmp	ax,cx			; compare remaining and blocksize
	ja	$L003A			; jump if remaining > blocksize
	mov	cx,ax			; copy only remaining length
$L003A:	sub	ax,cx			; adjust remaining length
	add	dx,cx			; adjust copied length
	lds	si,[si].BlockD.DataPtr	; load new source pointer
	MYCOPY				; copy frame block data
	pop	si			; restore block descriptor pointer
	pop	ds			; restore default data segment
	or	ax,ax			; test if more remaining length
	jz	$L009			; jmp if done

	add	si,SIZE BlockD		; bump block descriptor pointer
	dec	bx			; did we do all blocks?
	jnz	$L003			; jump while more blocks to be copied

$L009:	pop	di			; restore event pointer
	pop	si			; restore queue element pointer
	mov	es:[di]._EVENT.usRet2,dx ; set BytesCopied

	mov	ax,1
	call	IncAvail		; notify ring 3

	jmp	$L006			; jump to release MAC memory

; Discard rcv event from si
$L005:	inc	usLostFrames
	
; Release rcv event from si
$L006:	push	[si]._DELEV.usHandleMAC
	push	LoMod.usDS		; MAC DS
	call	LoMod.pRcvRel		; ReceiveRelease

	or	ax,ax			; Test MAC return code
	jz	$L007
	int	3			; Beng!! Error from MAC!!
$L007:	cli
	mov	di,dataOFFSET RxFree
	call	QueueAdd
	jmp	$L000

FlushRcvQue ENDP

;************************************************************************
;* NAME:        FlushBlockedQue
;* SYNOPSIS:    FlushBlockedQue()
;*              
;* DESCRIPTION: Flushes the BlockedQue if anything is there and while
;*              there is enough space....
;*              The function will check to see if there is space in shared
;*              segments. If there is enough space then it will write an
;*              event record. When the function can flush the whole queue it
;*              will remove the GF_BLOCKED flag from ufGlb.
;*              During the copy of an event record the interrupts will be
;*              enabled. But the function will return with interrupts
;*              disabled!
;* CALL REGISTERS: none
;*              Interrupts must be disabled
;* UNMODIFIED REGISTERS:
;*              ds bp
;* RETURNS:     ax = 0 -->   no error
;*              else    error code
;************************************************************************
FlushBlockedQue PROC NEAR

$L000:	mov	di,dataOFFSET BlockedQue ; Load pointer to never lost queue
	call	QueuePeek		; find oldest element
	jnz	$L002			; jump if we have an element

;
; No more events in Blocked Queue; maybe some extra lost records
	mov	ax,usLostFrames
	mov	bx,usLostStatus
	or	ax,bx			; are both lost counters 0?
	jnz	$L001			; jump if both 0

	and	ufGlb,NOT GF_BLOCKED	; clear Blocked flag...
$L001A:	ret				; return after flushing whole queue

$L001:	xor	ax,ax			; standard record size
	call	MallocEvent
	jz	$L001A			; jump to exit still blocked
	mov	ax,usLostFrames
	mov	es:[di]._EVENT.usRet1,ax
	mov	ax,usLostStatus		; fetch counters to registers
	mov	es:[di]._EVENT.usRet2,ax
	xor	ax,ax
	mov	usLostStatus,ax
	mov	usLostFrames,ax		; reset counters
	and	ufGlb,NOT GF_BLOCKED	; clear Blocked flag...
	sti
	inc	ax			; ax=1 for IncAvail
	mov	es:[di]._EVENT.usEventType,EVENT_LOST_FRAMES
	call	IncAvail
	cli
	jmp	$L001A

;
; There is at least one event in the queue. First check its
; lost counters and generate events for that, then try the real event.
$L002:	mov	ax,[si]._NLEV.usLostSt
	mov	dx,[si]._NLEV.usLostFr
	or	ax,dx			; any lost events?
	jz	$L010			; jump to real event if not

	push	si
	call	GenLost
	pop	si			; restore event descriptor ptr
	jz	$L011			; jump if out of memory

;
; Try to generate event record for real event.
; si points to never lost event.
$L010:	call	GenNL
	jnz	$L000			; repeat whole function if OK

$L011:	ret				; return because blocked

FlushBlockedQue ENDP

;************************************************************************
;* NAME:        DirectEvent
;* SYNOPSIS:    void DirectEvent( pEvent )
;* DESCRIPTION: Adds pEvent to the blocked queue and flushes the Rcv
;*              queue and the BlockedQueue
;* CALL REGISTERS:
;*              ds:si (pEvenT) pointer to SEND or GEN_REQ descriptor
;*              Interrupts must be disabled (but will be enabled temporarily)
;* UNMODIFIED REGISTERS:
;*              ds bp
;* RETURNS:     Nothing
;*              Returns with interrupts disabled
;************************************************************************
DirectEvent PROC NEAR

	mov	ax,usLostFrames
	mov	[si._NLEV.usLostFr],ax	; copy lost frames
	mov	ax,usLostStatus
	mov	[si._NLEV.usLostSt],ax	; copy lost counters
	xor	ax,ax
	mov	usLostFrames,ax		; reset lost frames
	mov	usLostStatus,ax		; reset lost counters

$L000:	mov	di,dataOFFSET BlockedQue
	call	QueueAdd		; add this general request to queue

	call	FlushRcvQue

	call	FlushBlockedQue

	ret

DirectEvent ENDP

;************************************************************************
;* NAME:        GenLost
;* SYNOPSIS:    GenLost(struct _NLEV *pEvent)
;*              
;* DESCRIPTION: Generates EVENT_LOST_FRAMES record if space available
;*              in shared memory. The counters in the event descriptor
;*              are cleared if the event is generated.
;*              Interrupts will be temporarily enabled during the copy.
;* CALL REGISTERS:
;*              si         pEvent
;*              Interrupts must be disabled
;* UNMODIFIED REGISTERS:
;*              ds bp si
;* RETURNS:     Zero flag set:    Not enough memory; Not written
;*              Zero flag clear:  Copied to shared memory and freed.
;************************************************************************
GenLost	PROC NEAR

	push	si
	xor	ax,ax			; standard size event
	call	MallocEvent		; Try to allocate shared memory
	pop	si			; restore event descriptor ptr
	jz	$L001			; Jump (return) if allocation failed
	xor	cx,cx
	mov	ax,[si]._NLEV.usLostFr
	mov	bx,[si]._NLEV.usLostSt	; fetch counters to registers
	mov	[si]._NLEV.usLostSt,cx
	mov	[si]._NLEV.usLostFr,cx	; reset counters

	sti
	mov	es:[di]._EVENT.usEventType,EVENT_LOST_FRAMES
	mov	es:[di]._EVENT.usRet1,ax
	mov	es:[di]._EVENT.usRet2,bx
	mov	ax,1
	call	IncAvail
	cli

	xor	ax,ax			; set zero flag
	inc	ax			; clear zero flag
$L001:	ret

GenLost ENDP

;************************************************************************
;* NAME:        GenNL
;* SYNOPSIS:    void	GenNL(void)
;*              
;* DESCRIPTION: Tries to allocate memory for a "Never Lost" event
;*              and write it to shared memory. If written the queue
;*              element will be removed from the queue and added to the
;*              free list for that event (GenReq or Tx).
;*              Interrupts will be temporarily enabled during the copy.
;* CALL REGISTERS:
;*              Interrupts must be disabled
;* UNMODIFIED REGISTERS:
;*              ds bp
;* RETURNS:     Zero flag set:    Not enough memory; Not written
;*              Zero flag clear:  Copied to shared memory and freed.
;************************************************************************
GenNL	PROC NEAR

	xor	ax,ax
	call	MallocEvent		; Try to allocate shared memory
	jz	$L002			; jump if allocation failed
	push	di
	mov	di,dataOFFSET BlockedQue
	call	QueueGet		; remove event from blocked queue

	sti

	pop	di			; restore shared memory pointer
	push	si			; save event descriptor ptr
	add	si,_NLEV.usEventType
	add	di,_EVENT.usEventType	; bump pointers to usEventType
	mov	cx,STD_EVENT_SIZE/2
	rep	movsw			; copy event to shared memory
	pop	si			; restore event descriptor ptr

	mov	ax,1
	call	IncAvail

	mov	di,dataOFFSET GenReqFree
	cmp	[si]._NLEV.usEventType,EVENT_GENREQ
	jz	$L001
	mov	di,dataOFFSET TxFree	; no, SEND or MULTI_SEND
$L001:	cli
	call	QueueAdd		; release event descriptor
	or	di,di			; clear zero flag
$L002:	ret

GenNL	ENDP

;************************************************************************
;* NAME:        IncAvail
;* SYNOPSIS:    void IncAvail(cnt)
;*              cnt	number of records available
;* DESCRIPTION: Adds cnt to usRecsAvail (and usLastRecsAvail) to signal
;*              the ring 3 application that new records are available.
;*              If usRecsAvail is zero on entry it releases the semaphore.
;* CALL REGISTERS:
;*              ax    contains cnt
;*              Interrupts wil be enabled if ax not 0
;* UNMODIFIED REGISTERS:
;*              ds bp si di
;* RETURNS:     nothing
;************************************************************************
IncAvail PROC NEAR

	or	ax,ax
	jz	$L001
	les	bx,pTxSeg		; get pointer to Tx segment

	SAV_IF
	mov	dx,inact_user
	mov	inact_timer,dx		; Reload inactivity timer

	add	usLastRecsAvail,ax	; add extra # of records to LastAvail
	mov	cx,es:[bx].TXSEG.usRecsAvail ; get old value
	add	ax,cx			; add extra # of records to Avail
	mov	es:[bx].TXSEG.usRecsAvail,ax ; set new value
	REST_IF

	or	cx,cx			; test old value
	jnz	$L001			; jump if already records there

; No need to set pxFirst because that is done during NEUS_GIVE
;	add	bx,TXSEG.hsemRing0	; let es:bx point to hsem
	mov	ax,WORD PTR es:[bx+TXSEG.hsemRing0+2]
					; SemHandle Segment
	mov	bx,WORD PTR es:[bx+TXSEG.hsemRing0]
					; SemHandle offset
	mov	dl,DevHlp_SemClear
	call	[DevHelp]		; clear semaphore
	jnc	$L001
	int	3			; Beng!!!!! You're DEAD....

$L001:	ret

IncAvail ENDP

;
;
	PUBLIC	_inact_expired
_inact_expired PROC NEAR
	pusha
	call	ChkFreed
	popa
	ret
_inact_expired ENDP
;
; - - - - - - - - - - - - - - - - - - - - - - - - -
; Timer functions
	COUNT_P0 equ	040h
	COUNT_CMD equ	COUNT_P0+3
_time_upd proc near
	xor	ax,ax
	out	COUNT_CMD,al		; Latch Counter 0 register
	mov	dx,COUNT_P0		; Let dx point to counter 0 port
	in	al,dx			; Get Counter low byte
	mov	ah,al
	in	al,dx			; Get Counter high byte
	xchg	ah,al			; Change to correct order
	not	ax			; Flip all bits of DOWN counter

	cmp	[TimeLo],ax		; Check if new value lower
	mov	[TimeLo],ax		; Store new low value
	mov	dx,[TimeHi]		; get timer high count
	jb	$T002			; Jump if no wrap through 0
	inc	dx			; Increment timer high count
	mov	[TimeHi],dx		; Store timer high count
	jnc	$T002
	inc	[TimeVeryHi]
$T002:	ret
_time_upd endp

	PUBLIC	init_8253
init_8253 PROC NEAR
;	mov	al,036h			; Load counter 0 mode 3 (??for ISA??)
	mov	al,034h			; Load counter 0 mode 2 (for PS/2)
	out	COUNT_CMD,al
	xor	ax,ax
	out	COUNT_P0,al
	jmp	$+2
	out	COUNT_P0,al
	ret
init_8253 ENDP

;************************************************************************
;* NAME:        test_counter
;* SYNOPSIS:    USHORT test_counter(void)
;* DESCRIPTION: Returns a number that approximates the maximal counter
;*              value of the Couter 0 circuit (IRQ0 counter).
;*              This function contains some tricky things. First we
;*              read the counter twice to see if it changes. Probably
;*              it does, but return with 0 if it doesn't change.
;*              Then read the counter (which is a DOWN counter!) until
;*              the new value is larger. That means it has been reloaded.
;*              This reloaded value is returned.
;*              Although the value initially loaded into the counter is
;*              always the same (and large) we will "never" see that value
;*              if an interrupt handler for IRQ0 is installed, since the
;*              handler will use some time before we can read the counter.
;*              Therefore we initialise our first "read" counter to 0FFFFh.
;*              0FFFFh may be retuned if we disabled interrupts just
;*              before the counter fires and latch after it fired (chance
;*              smaller than 1/0FFFFh?).......
;*
;*   !!         Not a very usefull function........
;*              Mode initialization required anyway!
;* UNMODIFIED REGISTERS:
;*              ds bp si di
;*              will enable interrupts!!!
;* RETURNS:     ax = 0 -->   Counter is not running!
;*              else         Approximate Counter wrap value
;************************************************************************
test_counter PROC near C
	push	bp
	mov	bp,sp

	xor	ax,ax
	cli
	out	COUNT_CMD,al		; Latch Counter 0
	in	al,COUNT_P0		; Get low byte
	mov	bl,al
	in	al,COUNT_P0		; Get high byte
	mov	bh,al			; BX contains first Counter 0 value
	xor	ax,ax
	out	COUNT_CMD,al		; Latch Counter 0
	in	al,COUNT_P0		; Get low byte
	mov	ah,al
	in	al,COUNT_P0		; Get high byte
	sti
	xchg	al,ah			; Put back in correct order
	sub	ax,bx			; Subtract previous value
	jz	$L002			; Jump if equal (NOT counting)

	mov	bx,0FFFFh		; Pretend we just read the max value
$L001:	mov	cx,bx			; Save last read value
	xor	ax,ax
	cli
	out	COUNT_CMD,al		; Latch Counter 0
	in	al,COUNT_P0		; Get low byte
	mov	bl,al
	in	al,COUNT_P0		; Get high byte
	sti
	mov	bh,al			; BX contains first Counter 0 value
	cmp	cx,bx			; Compare previous and last counters
	jnc	$L001			; Jump if still decreasing

	mov	ax,bx			; Get reloaded Counter

$L002:	mov	sp,bp
	pop	bp
	ret
test_counter ENDP

;
; - - - - - - - - - - - - - - - - - - - - - - - - -
;
	PUBLIC	_trapit
_trapit	proc	near
	pushf
	test	WORD PTR [ufGlb],GF_DEB
	jz	noint
	int	3
noint:	popf
	ret
_trapit	endp

	PUBLIC	_GetStamp	; Yes I know, normally at the postoffice
_GetStamp proc NEAR

	call	_time_upd
	ret

;	BootMSEC = 4
;	les	bx,[pgisGlobTime]	; Get pointer to INFOSEG
;	mov	ax,es:[bx+BootMSEC]
;	mov	dx,es:[bx+BootMSEC+2]
;	ret

_GetStamp endp

	PUBLIC	_SaveSel
_SaveSel PROC	NEAR
	push	bp
	mov	bp,sp

	push	di
	push	si
	les	di,[bp+4]	; Get dest ptr
	mov	si,dataOFFSET ProtCS
	mov	cx,4
	rep movsw
	mov	ax,cs
	stosw
	mov	ax,ds
	stosw
	pop	si
	pop	di

	pop	bp
	ret
_SaveSel ENDP

_TEXT	ENDS
	.XLIST
	SUBTITLE
	.LIST

	END
