;************************************************************************
;* MODULE INFORMATION*
;**********************
;*     FILE     NAME:       neusque.asm
;*     SYSTEM   NAME:       
;*     ORIGINAL AUTHOR(S):  Paul Lemmers
;*     VERSION  NUMBER:     
;*     CREATION DATE:       1991/11/14
;*
;* DESCRIPTION: Queue operation primitives
;*              All functions are gathered here to allow queue management
;*              modifications without influencing to the rest of code.
;*              Some functions are so simple that a macro would have been
;*              perfect too......
;*
;************************************************************************
;* CHANGES INFORMATION **
;************************
;* REVISION:    $Revision:   1.1  $
;* WORKFILE:    $Workfile:   neusque.asm  $
;* LOGINFO:     $Log:   D:/CPROG/NDIS/NEUS/VCS/NEUSQUE.ASV  $
;*              
;*                 Rev 1.1   30 Mar 1993 09:40:48   PAUL
;*              MASM 6.00B prototypes
;*              
;*                 Rev 1.0   16 Feb 1992 18:25:12   PAUL
;*              Initial revision.
;************************************************************************/
;
	INCLUDE neussdef.inc
	NAME    NEUSQUE
	.XLIST
	.XCREF
	INCLUDE neusglb.inc
	.CREF
	SUBTITLE	'Queue handling functions'
	.LIST
	PAGE
_TEXT   SEGMENT
	PUBLIC	QueueGet, QueueAdd, QueuePeek, RxAlloc

;************************************************************************
;* NAME:        QueueGet
;* SYNOPSIS:    PQEL QueueGet(QDESC *pQD)
;*              QDESC *pQD       pointer to queue descriptor structure
;* DESCRIPTION: Removes an element from the queue described by pQD and
;*              returns the pointer to that element. Be careful, the
;*              element is no longer in any queue after this. Use
;*              QueueAdd() to add it to an other list!
;* CALL REGISTERS:
;*              interrupts disabled!!!
;*              ds           must be set to the drivers default data segment
;*              di           must hold pQD
;* UNMODIFIED REGISTERS:
;*              ds bp di
;*              es ax cx dx
;* RETURNS:     Zero flag set   Queue was empty when called; si=0
;*              else            si contains near pointer to "dangling"
;*                              queue element
;************************************************************************
QueueGet PROC NEAR

	ASSUME	DI:PTR QDESC
	ASSUME	SI:PTR _NLEV
	mov	si,[di.QDESC.Tail]	; fetch oldest to pEl
	or	si,si			; test if there is anything in queue
	jz	$L002			; jump if queue is empty

	mov	bx,[si._NLEV.next]	; get new tail pointer
	mov	[di.QDESC.Tail],bx	; save it in pQD
	or	bx,bx			; test if queue empty now
	jz	$L001			; jump if not empty
	mov	[si._NLEV.next],0	; pEl no longer in any queue
	jmp short $L002
$L001:	mov	[di.QDESC.Head],bx	; indicate pQD now empty
	or	si,si			; clear zero flag

$L002:	ret
	ASSUME	DI:NOTHING, SI:NOTHING

QueueGet ENDP



;************************************************************************
;* NAME:        QueueAdd
;* SYNOPSIS:    PQEL QueueAdd(QDESC *pQD, PQEL pEl)
;*              QDESC *pQD       pointer to queue descriptor structure
;*              PQEL   pEl       pointer to queue element
;* DESCRIPTION: Adds an element from the queue described by pQD and
;*              returns the pointer to that element.
;* CALL REGISTERS:
;*              interrupts disabled!!!
;*              ds           must be set to the drivers default data segment
;*              di           must hold pQD
;*              si           must hold pEl
;* UNMODIFIED REGISTERS:
;*              ds bp si di
;*              es ax cx dx
;* RETURNS:     pEl in si (which is where it already was)
;************************************************************************
QueueAdd PROC NEAR

	ASSUME	DI:PTR QDESC,SI:PTR _NLEV
	mov	bx,[di.QDESC.Tail]	; fetch oldest
	or	bx,bx			; test if there is anything in queue
	jnz	$L001			; jump if queue is not empty

	mov	[di.QDESC.Tail],si	; set first element in pQD
	jmp short $L002

$L001:	mov	bx,[di.QDESC.Head]	; fetch old head
	mov	[bx._NLEV.next],si	; pEl into previous head

$L002:	mov	[di.QDESC.Head],si	; Set new head
	ret
	ASSUME	DI:NOTHING, SI:NOTHING

QueueAdd ENDP



;************************************************************************
;* NAME:        QueuePeek
;* SYNOPSIS:    PQEL QueuePeek(QDESC *pQD)
;*              QDESC *pQD       pointer to queue descriptor structure
;* DESCRIPTION: Fetches the tail from the queue described by pQD.
;* CALL REGISTERS:
;*              interrupts disabled!!!
;*              ds           must be set to the drivers default data segment
;*              di           must hold pQD
;* UNMODIFIED REGISTERS:
;*              ds bp di
;*              es ax bx cx dx
;* RETURNS:     Zero flag set   Queue was empty when called; si=0
;*              else            si contains near pointer to Tail queue
;*                              element
;************************************************************************
QueuePeek PROC NEAR

	ASSUME	DI:PTR QDESC
	mov	si,[di.QDESC.Tail]	; fetch oldest to pEl
	or	si,si			; test if there is anything in queue
	ret
	ASSUME	DI:NOTHING

QueuePeek ENDP


;************************************************************************
;* NAME:        RxAlloc
;* SYNOPSIS:    PQEL RxAlloc(void)
;* DESCRIPTION: Allocates an element from the RxFree queue if they are
;*              still available and adds this element to the RxQueue and
;*              then returns the pointer to that element.
;* CALL REGISTERS:
;*              interrupts disabled!!!
;*              ds           must be set to the drivers default data segment
;* UNMODIFIED REGISTERS:
;*              ds bp
;*              es ax cx dx
;* RETURNS:     Zero flag set   Queue was empty when called; si=0
;*              else            si contains near pointer to "moved"
;*                              queue element
;************************************************************************
RxAlloc PROC NEAR

	mov	di,dataOFFSET RxFree	; load RxFree descriptor
	call	QueueGet		; Try to get element
	jz	$L001			; return if no free elements
	mov	di,dataOFFSET RxQue	; load RxQue descriptor
	call	QueueAdd		; add new element
	or	si,si			; make sure Zero flag is clear

$L001:	ret

RxAlloc ENDP

;************************************************************************
;* NAME:        Q_Get
;* SYNOPSIS:    struct _NLEV *Q_Get(struct _QDESC *pQ)
;*              struct _QDESC *pQ;      Poiter to Queue descriptor
;* DESCRIPTION: C language version of QueueGet function (saves "C" registers).
;* CALL REGISTERS:
;*              none
;* UNMODIFIED REGISTERS:
;*              es ds bp si di
;* RETURNS:     ax = 0 -->   error
;*              else         pointer to queue element (to be used as handle)
;************************************************************************
Q_Get	PROC	NEAR C USES di si pQ:PTR _QDESC

;	pQ	=	4

;	push	bp
;	mov	bp,sp
;	push	si
;	push	di

	mov	di,[pQ]			; get queue descriptor
	cli
	call	QueueGet
	sti
	mov	ax,si

;	pop	di
;	pop	si
;	pop	bp
	ret

Q_Get	ENDP

_TEXT	ENDS

	END
