/************************************************************************
** MODULE INFORMATION*
**********************
**     FILE     NAME:       neustrat.c
**     SYSTEM   NAME:       
**     ORIGINAL AUTHOR(S):  Paul Lemmers
**     VERSION  NUMBER:     
**     CREATION DATE:       1991/11/12
**
** DESCRIPTION: The strategy handler (including Open, Close and IOCtl)
**              of NEUS.
**
*************************************************************************
** CHANGES INFORMATION **
*************************
** REVISION:    $Revision:   1.6  $
** WORKFILE:    $Workfile:   neustrat.c  $
** LOGINFO:     $Log:   D:/CPROG/NDIS/NEUS/VCS/NEUSTRAT.C_V  $
**              
**                 Rev 1.6   30 Mar 1993 09:41:10   PAUL
**              Multisend removed. Link ring 0 selectors to default data
**              segment during Close processing.
**              
**                 Rev 1.5   20 Sep 1992 21:28:36   PAUL
**              DevHelp_SemHandle addition
**              
**                 Rev 1.4   15 Jul 1992 13:42:38   PAUL
**              Modifications for inactivity timer
**              
**                 Rev 1.3   12 Mar 1992 22:32:04   PAUL
**              High Resolution timer
**              
**                 Rev 1.2   05 Mar 1992 13:16:22   PAUL
**              MULTI_SEND implementation C part
**              
**                 Rev 1.1   21 Feb 1992 22:26:32   PAUL
**              Ahum, nobody knows why... See vdiff...
**              
**                 Rev 1.0   16 Feb 1992 18:26:22   PAUL
**              Initial revision.
*************************************************************************/

#include <os2def.h>
#include <ndisdef.h>

#include <cdev.h>
#include <devioctl.h>
#include <devlib.h>

#include <neus.h>
#include "neusglb.h"

/* ahum........ */
void _cdecl _disable(void);
void _cdecl _enable(void);

#define MYERR     (STAT_ERR | STAT_DEV | STAT_DONE)

UI	neusgive(PTXSEG pTx);
UI	Send(struct _iocparm _far *pParm, unsigned char _far *pData);
UI	GenReq(struct _iocparm _far *pParm);
UI	CpyTabs(char _far *pData);
unsigned short CopyT(LPBUF pTab, unsigned char _far *pData);
void	QueueInit(QDESC *pQD, unsigned short usSize);
int	checkseg(int n, SEL selX, unsigned short usSize);
void	cleanup(void);


void	c_strat(REQHDR  _far *req_p)
{

	switch ( req_p->command )
		{
		case C_INIT:
			Init((INIT_R _far *)req_p);
			break;
		case C_DEV_OPEN:
			Device_Open(req_p);
			break;
		case C_DEV_CLOSE:
			Device_Close(req_p);
			break;
		case C_GEN_IOCtl:
			Generic_IOCtl((IOCtl_R _far *)req_p);
			break;
		default:
			req_p->status = E_UNKN_COMM | STAT_ERR | STAT_DONE;
			break;
		}
}




void	Device_Open(REQHDR  _far *req_p)
{
	if ( open_cnt )
		{
		/* Whoops already open.......... */
		req_p->status = STAT_DONE | STAT_ERR | E_GEN_FAIL;
		}
	else if ( (test_counter() < 0xF000) || (InstallTick() != 0) )
		{
		req_p->status = MYERR | NER_NO_TIMER;
		}
	else
		{
		open_cnt++;
		req_p->status = STAT_DONE;

		/* Save xxxxx_num */
		}

	return;
}


void	Device_Close(REQHDR  _far *req_p)
{
	if ( open_cnt )
		/* O.K., but don't decrement below 0 */
		open_cnt--;

	DeInstallTick();

	cleanup();

	req_p->status = STAT_DONE;

	return;
}

/***
**
*/
void	Generic_IOCtl(IOCtl_R _far *reqp)
{
	USHORT	rstat; /* return status in request header */
	struct _iocparm _far *pParm;

	/* preset error code */
	if ( reqp->rdata.func_code & IGNORE_MASK )
		rstat = STAT_DONE;
	else
		rstat = E_UNKN_COMM | STAT_ERR | STAT_DONE;


	/* Select the task to perform and always mark header as */
	/* done. On error, return with preselected error code.  */
	if ( reqp->rdata.func_cat == NEUS_CATEGORY )
		{
		unsigned char code;
		pParm = (struct _iocparm _far *)reqp->rdata.parmp;

		code = reqp->rdata.func_code;
		code &= (SUBF_MASK | QUERY_MASK | PASS_TO_DEV);
		if ( (ufGlb & GF_MAC_OK) == 0 )
			{
			rstat = MYERR | NER_NOT_BOUND;
			}
		else if ( ((ufGlb & GF_GIVE) == 0) && (code != NEUS_GIVE) &&
				(code != NEUS_HIGHRES_TIME) )
			{
			rstat = MYERR | NER_NO_GIVE;
			}
		else
		switch ( code )
			{
			case NEUS_MOTOR:
				_disable();
				ChkFreed();
				_enable();
				rstat = STAT_DONE;
				break;

			case NEUS_GIVE:
				rstat = neusgive(reqp->rdata.datap);
				break;

			case NEUS_GET_CONFIG:
				rstat = CpyTabs(reqp->rdata.datap);
				break;

			case NEUS_MAX_LEN:
				_disable();
				ChkFreed();
				usMaxLen = pParm->usDatLen;
/*				gen_event(pParm); aaaah, this may block?? */
				_enable();
				rstat = STAT_DONE;
				break;

			case NEUS_GENREQ:
				_disable();
				ChkFreed();
				_enable();
				rstat = GenReq(pParm);
				break;

			case NEUS_SEND:
				_disable();
				ChkFreed();
				_enable();
				rstat = Send(pParm, reqp->rdata.datap);
				break;

			case NEUS_HIGHRES_TIME:
				_disable();
				pParm->ulParam2 = GetStamp();
				pParm->usParam1 = ticks;
				_enable();
				rstat = STAT_DONE;
				break;

			case NEUS_TYPE_LEN:
			default:
				rstat = (MYERR | 0x80);
				break;
			}
		}

	reqp->hdr.status = rstat;
}


/***
**
*/
UI	neusgive(PTXSEG pTx)
{
	UI	rstat = STAT_DONE;
	unsigned short cnt;
	HSEM	hsemTmp;

	/* Check if it is a duplicate GIVE */
	if ( ufGlb & GF_GIVE )
		return(MYERR | NER_DUP_GIVE);

	/* is pTx anywhere near??? */
	if ( dev_VerifyAccess(pTx, sizeof(TXSEG), 1) )
		return(MYERR | NER_INVALID_TXSEG);

	nsel = pTx->usRcvSegCnt;
	if ( (nsel<2) || (nsel>maxsel) )
		return(MYERR | NER_TOO_MANY_RXSEG);

	/* save semaphore handle and reset it. Otherwise cleanup releases it */
	hsemTmp = pTx->hsemRing0;
	pTx->hsemRing0 = 0L;

	/* Do Tx segment */
	if ( checkseg(0, SELECTOROF(pTx), pTx->usTxSize) )
		{
		cleanup();
		return(MYERR | NER_INVALID_TXSEG);
		}

	pTxSeg = MAKEP(aSel[0].usSeg0, OFFSETOF(pTx));
	if ( pTx->usTxSize == 0 )
		aSel[0].usBegin = 0xffff;
	else
		aSel[0].usBegin = pTx->usTxSize;

	/* do all Rx segments */
	for ( cnt=0 ; cnt < nsel ; cnt++ )
		{
		if ( checkseg(cnt+1, pTx->awRcvSels[cnt], 0) )
			{
			cleanup();
			return(MYERR | NER_INVALID_RXSEG);
			}

		aSel[cnt].fAttr = SEG_MORE;
		}

	pTx->pxFirst = MAKEP(pTx->awRcvSels[0], 0);
	pTx->usRecsAvail =
		usLastRecsAvail =
		usLostFrames =
		usLostStatus =
		usFreeOffs =
		usOldestOffs = 0;
	usCurrentSeg = usOldestSeg = 1;

	/* prepare queues */
	QueueInit(&GenReqFree, sizeof(struct _NLEV));
	QueueInit(&TxFree, sizeof(struct _NLEV));
	QueueInit(&RxFree, SIZE_RX);

	/* prepare semaphore */
	if ( hsemTmp == 0L )
		{
		/* use RAM semaphore */
		hsemTmp = &(pTxSeg->semRecsAvail);
		}

	if ( dev_SemHandle(&hsemTmp, 1) != 0 )
		{
		cleanup();
		return(MYERR | NER_INVALID_SEM);
		}
	else
		pTxSeg->hsemRing0 = hsemTmp;

	/* set GIVE flag in ufGlb...... */
	ufGlb |= GF_GIVE;

	return(rstat);
}


UI	Send(struct _iocparm _far *pParm, unsigned char _far *pData)
{
	UI	rstat = STAT_DONE;
	struct _NLEV *pEvent;
	struct _TxBufD MyTxD;

	if ( SELECTOROF(pData) != aSel[0].usSeg3 )
		return(MYERR | NER_INVALID_TXSEG);

	if ( (OFFSETOF(pData)+pParm->usDatLen) >= aSel[0].usBegin)
		return(MYERR | NER_INVALID_TXSEG);

	if ( (pEvent=Q_Get(&TxFree)) != NULL )
		{
		pEvent->usFlags = REQ_ACTIVE;
		pEvent->usEventType = EVENT_SEND;
		pEvent->usHandle3 = pParm->usHandle;
		pEvent->usRet2 = 0;
		MyTxD.TxImmedLen = 0;
		MyTxD.TxImmedPtr = 0L;
		MyTxD.TxDataCount = 1;
		MyTxD.TxArray[0].PtrType = '\2';
		MyTxD.TxArray[0].ResByte = '\0';
		MyTxD.TxArray[0].Data.Len = pParm->usDatLen;
		MyTxD.TxArray[0].Data.DataPtr =
			MAKEP(aSel[0].usSeg0, OFFSETOF(pData));
		do_Send(&MyTxD, pEvent);
		}
	else
		rstat = MYERR | NER_TOO_MANY;

	return(rstat);
}

UI	GenReq(struct _iocparm _far *pParm)
{
	UI	rstat = STAT_DONE;
	struct _NLEV *pEvent;

	switch ( pParm->usSub )
		{
		case OpenAdapter:
			/* Check if NULL ptr */
			if ( SELECTOROF(pParm->ulParam2) == 0 )
				/* don't convert, because NULL */
				break;

		case ReadErrorLog:
		case SetStationAddress:
		case AddMulticastAddress:
		case DeleteMulticastAddress:
		case SetFunctionalAddress:
			/* Check if in TX segment */
			if ( SELECTOROF(pParm->ulParam2) != aSel[0].usSeg3 )
				return(MYERR | NER_INVALID_TXSEG);

			/* convert to ring 0 selector */
			SELECTOROF(pParm->ulParam2) = aSel[0].usSeg0;
			break;
		}

	if ( (pEvent=Q_Get(&GenReqFree)) != NULL )
		{
		pEvent->usFlags = REQ_ACTIVE;
		pEvent->usEventType = EVENT_GENREQ;
		pEvent->usHandle3 = pParm->usHandle;
		pEvent->usRet2 = 0;
		do_GenReq(pParm, pEvent);
		}
	else
		rstat = MYERR | NER_TOO_MANY;

	return(rstat);
}

#define NOT_NULL(p) ((long)(p) != 0L)
UI	CpyTabs(char _far *pData)
{
	unsigned short usLen;
	UI	rstat = STAT_DONE;

	usLen = 0;
	if ( NOT_NULL(LoMod.pCCT) )
		{
		usLen += *(unsigned short _far *)LoMod.pCCT;

		if ( NOT_NULL(LoMod.pSSCT) )
			{
			usLen += *(unsigned short _far *)LoMod.pSSCT;
			}

		if ( NOT_NULL(LoMod.pSSST) )
			{
			usLen += *(unsigned short _far *)LoMod.pSSST;
			}
		}

	if ( dev_VerifyAccess(pData, usLen, 1) )
		{
		/* beng not enough */
		rstat = MYERR | NER_INVALID_SEG;
		}
	else
		{
		pData += CopyT(LoMod.pCCT, pData);

		pData += CopyT(LoMod.pSSCT, pData);

		pData += CopyT(LoMod.pSSST, pData);
		}

	return(rstat);
}

unsigned short CopyT(LPBUF pTab, unsigned char _far *pData)
{
	unsigned short usLen;

	if ( NOT_NULL(pTab) )
		{
		usLen = *(unsigned short _far *)pTab;
		}
	else
		usLen = 2;

	if ( usLen > 2 )
		{
		_disable();           /* No changes during copy! */
		dev_memcpy(pData, pTab, usLen);
		_enable();
		}
	else
		{
		*(unsigned short _far *)pData = usLen = 2;
		}

	return(usLen);
}

void	QueueInit(QDESC *pQD, unsigned short usSize)
{
	unsigned short n;
	PQEL	pEl;

	pEl = pQD->Tail = pQD->Mem;
	pEl->nlevent.next = (PQEL)((char *)pEl + usSize);

	for (n=1 ; n<pQD->q_size ; n++)
		{
		pEl = pEl->nlevent.next;
		pEl->nlevent.next = (PQEL)((char *)pEl + usSize);
		}

	pEl->nlevent.next = NULL;
	pQD->Head = pEl;
}

int	checkseg(int n, SEL selX, unsigned short usSize)
{
	int	retc = 0;
	unsigned long ulLock;
	unsigned long ulPhys;

	if ( dev_Lock(selX, 0x100, &ulLock) == 0 )
		{
		if ( dev_VerifyAccess(MAKEP(selX,0), usSize, 1) )
			{
			dev_Unlock(ulLock);
			retc = -1;
			}
		else
			{
			aSel[n].usSeg3 = selX;
			aSel[n].ulLock = ulLock;
			if ( dev_VirtToPhys(MAKEP(selX,0), &ulPhys) )
				retc = 2;
			else if ( dev_PhysToGDTSelector(ulPhys, usSize,
							awRing0Sels[n]) )
				retc = 3;
			else
				aSel[n].usSeg0 = awRing0Sels[n];
			}
		}
	else
		retc = 1;

	return(retc);
}

void	cleanup()
{
	unsigned short cnt;       /* loop counter */
	unsigned long ulPhysDS   /* physical address of DS segemnt */
			= 0L;

	_disable();
	inact_timer = 0;
	if ( ufGlb & GF_GIVE )
		{
		/* stop new activity */
		ufGlb &= ~GF_GIVE;

		_enable();

		if ( pTxSeg->hsemRing0 != 0L )
			{
			HSEM	hsemTmp;

			hsemTmp = pTxSeg->hsemRing0;
			pTxSeg->hsemRing0 = 0L;
			dev_SemHandle(&hsemTmp, 0);
			}

		/* check if any pending requests */
		/* and abort them........ */
		ufGlb &= ~GF_BLOCKED;

		/* Get physical address of default data segment */
		cnt = (USHORT)((ULONG)((void _far *)awRing0Sels)>>16);
		if ( dev_VirtToPhys( MAKEP(cnt, 0), &ulPhysDS) != 0 )
			ulPhysDS = 0L;    /* BUG; ERROR; but can't report */

		for ( cnt=0 ; cnt <= nsel ; cnt++ )
			{
			if ( aSel[cnt].ulLock != 0L )
				dev_Unlock(aSel[cnt].ulLock);

			if ( ulPhysDS != 0L )
				{
				if ( dev_PhysToGDTSelector(ulPhysDS, 8,
							awRing0Sels[cnt]) )
					ulPhysDS = 0L;    /* BUG; ERROR; */
				}
			}

		dev_memset(aSel, 0, (nsel+1)*sizeof(*aSel));
		}
	else
		_enable();
}
