/************************************************************************
** MODULE INFORMATION*
**********************
**     FILE     NAME:       protparm.c
**     SYSTEM   NAME:       NDIS\PROTPARM
**     ORIGINAL AUTHOR(S):  Paul Lemmers
**     VERSION  NUMBER:     
**     CREATION DATE:       1990/8/31
**
** DESCRIPTION: 
**              File with functions to parse protocol.ini info
**              as obtained from the GetProtocolManagerInfo call.
**              The INIT code of the driver should call:
**              
**              and then:
**              
**              This will write all information to the specified variables.
*************************************************************************
** CHANGES INFORMATION **
*************************
** REVISION:    $Revision:   1.3  $
** WORKFILE:    $Workfile:   protparm.c  $
** LOGINFO:     $Log:   D:/cprog/ndis/protparm/vcs/protparm.c_v  $
**              
**                 Rev 1.3   22 Sep 1991 18:47:08   PAUL
**              Backup tijdens/na MASM 6.0 conversie
**              
**                 Rev 1.2   11 Dec 1990 21:40:08   PAUL
**              - Centralised the error reporting into one function
**                and messages at the beginning of the file.
**              
**                 Rev 1.1   09 Dec 1990 19:55:32   PAUL
**              Adaptions for MSC 6.00
**              
**                 Rev 1.0   03 Sep 1990 13:42:34   PAUL
**              Initial revision.
*************************************************************************/

#include <os2def.h>

#include <ndisdef.h>
#include <devlib.h>

#include "protparm.h"

extern char ModSign[];
/* some local functions */
void	ParamCpy(PARMD *parmp, PARM_E _far *pp, unsigned short i);
void	OutProtErr(unsigned short erno, KEYW_E _far *keywp, ...);
#define keycmp(p1, p2)	dev_stricmp(p1,p2)

#if ! defined(NULL)
#define NULL ((void *)0)
#endif


char	*proterr[] = {
"%s1%02dE %Fs keyword ignored\r\n",
#define ERR_KEYW_IGN   0
"%s1%02dE %Fs to many parameters\r\n",
#define ERR_TOMANY     1
"%s1%02dE %Fs (index %d) invalid parameter type\r\n",
#define ERR_INV_PARM   2
"%s1%02dE %Fs (index %d) parameter out of range\r\n"
#define ERR_RANGE_PARM 3
};



/**************************************************************
** NAME:        DoSection
** SYNOPSIS:    int DoSection(MOD_E _far *ModP, PARMD *myparms,
**                                             unsigned short cnt)
**              MOD_E _far *ModP      Pointer to section
**              PARMD     *myparms   pointer to parameter description
**              unsigned short cnt   # of parameters in myparms
** DESCRIPTION: All keywords in the section pointed to by ModP will
**              be checked against myparms description.
**              This checking is done with IsValidKeyword() and
**              if valid then DoKeyword() is used to do the other
**              checks and storing into variables.
** RETURNS:     0 -->   no error
**              else    error code
**************************************************************/
int	DoSection(MOD_E _far *ModP, PARMD *myparms, unsigned short cnt)
{
	int	retc = 0;
	KEYW_E	_far *keywp;
	PARMD *parmp;

	keywp = &ModP->KeyWord1;

	while ( (long) keywp )
		{
		if ( (parmp=IsValidKeyword(keywp, myparms, cnt)) != NULL )
			{
			/* Yes, this is a valid entry */
			if ( DoKeyword(keywp, parmp) )
				{
				/* something very wrong with the keyword */
				retc = -1;
				}
			}
		else
			{
			OutProtErr(ERR_KEYW_IGN, keywp);
			}

		keywp = keywp->Next;
		}

	return(retc);
}


/**************************************************************
** NAME:        IsMySection
** SYNOPSIS:    int IsMySection(MOD_E _far *ModP, char *devname)
**              MOD_E _far *ModP       Pointer to section
**              char      *devname    Pointer to device driver name
** DESCRIPTION: This function checks if the section pointed to by
**              ModP, contains the "DRIVERNAME=devname" parameter.
**              All keywords are scanned until the keyword "DRIVERNAME"
**              is found. The the compare on devname is performed,
**              and the function will return that result.
** RETURNS:     0 -->   not my section
**              else    yes, this is my section
**************************************************************/
int  IsMySection(MOD_E _far *ModP, char *devname)
{
	KEYW_E	_far *keywp;

	keywp = &ModP->KeyWord1;

	while ( (long) keywp )
		{
		if ( keycmp("DRIVERNAME", keywp->Keyword) == 0 )
			{
			/* Yes, this is the DriverName entry */
			if ( (keywp->Parm1.ParamType == 1) &&
			     (keycmp(devname, keywp->Parm1.Value.ValC) == 0)
					)
				return(1);
			else
				{
				/* lets not be silly: one name per section */
				return(0);
				}
			}

		keywp = keywp->Next;
		}

	return(0);
}



/**************************************************************
** NAME:        IsValidKeyword
** SYNOPSIS:    PARMD *IsValidKeyword(KEYW_E _far *keywp, PARMD *parmp,
**                                   unsigned short cnt)
**              KEYW_E _far *keywp       pointer to keyword under test
**              PARMD      *parmp       pointer to parameter description
**              unsigned short cnt      # of parameters in parmp
** DESCRIPTION: Checks if the keyword pointed to by keywp is in the
**              array of allowed parameters.
** RETURNS:     NULL -->   Not Valid
**              else       Pointer to parameter description
**************************************************************/
PARMD *IsValidKeyword(KEYW_E _far *keywp, PARMD *parmp, unsigned short cnt)
{
	while ( cnt-- )
		{
		if ( keycmp(parmp->keyw, keywp->Keyword) )
			{
			/* No not the same keyword */
			parmp++;
			}
		else
			return(parmp);
		}

	/* did them all but no match */
	return(0);
}


/**************************************************************
** NAME:        DoKeyword
** SYNOPSIS:    int DoKeyword(KEYW_E _far *keywp, PARMD *parmp)
**              KEYW_E _far *keywp    pointer to found keyword
**              PARMD      *parmp    pointer to allowed keyword
** DESCRIPTION: This function checks the parameter and copies it to
**              the variable.
**              The first version is very limited.
**              It returns an error if the parameter type is invalid
**              but not for limit checks (not performed!) etc.
**              The actual number of parameters or the maximum number,
**              whichever is lowest, will be copied. No errors are
**              generated for an invalid number of parameters.
** RETURNS:     0 -->   no error
**              else    error code
**************************************************************/
int	DoKeyword(KEYW_E _far *keywp, PARMD *parmp)
{
	union	{
		char _far *cparm;
		PARM_E _far *pparm;
		} pp;
	unsigned short	i;
	unsigned short	retc = 0;
	unsigned short nparm;

	if ( parmp->funcp == 0 )
		{
		if ( parmp->parmsize != 0 )
			{
			pp.pparm = &keywp->Parm1;
			if ( (unsigned short)parmp->nparms < keywp->NumParams)
				{
				nparm = parmp->nparms;
				OutProtErr(ERR_TOMANY, keywp);
				}
			else
				nparm = keywp->NumParams;

			/* nparm can be 0, then no illegal references */
			for ( i=0 ; i<nparm ; i++ )
				{
				if ( pp.pparm->ParamType != parmp->type )
					{
					OutProtErr(ERR_INV_PARM,
						keywp, i);
					retc = !0;
					}
				else if ( pp.pparm->ParamType == 0 )
					{
					if ( (pp.pparm->Value.ValL >= parmp->min) &&
						(pp.pparm->Value.ValL <= parmp->max)
						)
						ParamCpy(parmp, pp.pparm, i);
					else
						{
						OutProtErr(ERR_RANGE_PARM,
							keywp, i);
						retc = !0;
						break; /* stop copy */
						}
					}
				else
					ParamCpy(parmp, pp.pparm, i);

				pp.cparm += pp.pparm->ParamLen + 4;
				}
			}
		}
	else
		retc = (*parmp->funcp)(keywp, parmp);

	return(retc);
}


/**************************************************************
** NAME:        ParamCpy
** SYNOPSIS:    void ParamCpy(PARMD *parmp, PARM_E _far *pp, unsigned short i)
**              PARMD      *parmp    Allowed keyword description
**              PARM_E _far *pp       Found keyword description
**              unsigned short i     index in destination array
** DESCRIPTION: This function does the real copy of parameter data.
**              It can copy data with size 1, 2 and 4 (char, int and long)
**              and arrays of char. The arrays af char are only copied if
**              they fit into the destination.
**
**              Watch it!
**              It will copy to index i in the destination array of
**              parmp->parmsize byte elements for non string types!
**
** RETURNS:     Nothing
**************************************************************/
void	ParamCpy(PARMD *parmp, PARM_E _far *pp, unsigned short i)
{
	switch ( parmp->type )
		{
		case 0:
			switch ( parmp->parmsize )
				{
				case 1:
					((char *)parmp->binp)[i] =
						(char)pp->Value.ValL;
					break;
				case 2:
					((short *)parmp->binp)[i] =
						(short)pp->Value.ValL;
					break;
				case 4:
					((long *)parmp->binp)[i] =
						pp->Value.ValL;
					break;
				}
			break;
		case 1:
			if ( pp->ParamLen <= parmp->parmsize )
				dev_strcpy((char *)parmp->binp, pp->Value.ValC);
			break;
		}

}


#define va_arg(ap,t) ((t _far *)(ap += sizeof(t)))[-1]
/**************************************************************
** NAME:        OutProtErr
** SYNOPSIS:    void OutProtErr(unsigned short erno,
**				KEYW_E _far *keywp, ...)
**              unsigned short erno;   error number
**              KEYW_E _far *keywp;    ptr to keyword structure
** DESCRIPTION: Function to write an error message about parameters
**              in protocol.ini to the screen
** RETURNS:     nothing
**************************************************************/
void	OutProtErr(unsigned short erno, KEYW_E _far *keywp, ...)
{
	char _far *ap;

	ap = (char _far *)&keywp + sizeof(keywp);

	switch ( erno )
		{
		case ERR_KEYW_IGN:
		case ERR_TOMANY:
			dev_printf(proterr[erno], ModSign, erno,
				keywp->Keyword);
			break;
		case ERR_INV_PARM:
		case ERR_RANGE_PARM:
			dev_printf(proterr[erno], ModSign, erno,
				keywp->Keyword, va_arg(ap, short));
			break;
		}
}
