/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>
#include <dtime.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ring.h>
#include <timer.h>
#include <mac.h>
#include <protocol.h>
#include <message.h>
#include <config.h>
#include <sysmib.h>
#include <maxmem.h>
#include <memory.h>
#include <channel.h>

#include "capturee.h"
#include "capturec.h"


IMPORT BOOLEAN BeholderStrictRMON;


IMPORT WORD	memoryPerc;


BOOLEAN CaptureConfigInit(VOID)
{
SHORT p;

    if (ConfigGetShort("beholder.capture.memoryperc", &p))
        memoryPerc = p;
    if (memoryPerc < 1 || memoryPerc > 75)
    {
        DnpapMessage(DMC_WARNING, CAPTURE_MEMPERC, "capture: bad value for beholder.capture.memoryperc, using default");
        memoryPerc = 75;
    }
    return TRUE;
}


BOOLEAN CaptureInit (BUFFER *Buffer)
{
    Buffer->ChannelIndex = 1;          /* first channel */
    Buffer->ChannelCallback = NULL;
    Buffer->FullStatus = 1;            /* spaceAvailable */
    Buffer->FullAction = 1;            /* lockWhenFull */
    Buffer->CaptureSliceSize = 100;
    Buffer->DownloadSliceSize = 100;
    Buffer->MaxOctetsRequested = -1;
    Buffer->MaxOctetsGranted = -1;
    Buffer->MaxOctetsAvailable = 0;
        
	Buffer->CapturedPkts = 0;
	
    Buffer->TurnOnTime = 0;
    Buffer->TurnOnTimeMsec = 0;
	
	BooleanSetAllTrue(Buffer->ObjectSet);
	BooleanSetFalse(Buffer->ObjectSet, CAPTURE_BOOLEAN_INDEX);
	BooleanSetFalse(Buffer->ObjectSet, CAPTURE_BOOLEAN_OWNER);

    return TRUE;
}


BOOLEAN CaptureStart (BUFFER *Buffer)
{
	if (BeholderStrictRMON && !BooleanCheckAllTrue(Buffer->ObjectSet))
		return FALSE;

    if (Buffer->RingBuf == NULL)
	{
	    Buffer->MaxOctetsAvailable = (MemoryGetAvail()/100)*memoryPerc;
   		if (Buffer->MaxOctetsAvailable > MemoryGetMaxChunk())
       		Buffer->MaxOctetsAvailable = MemoryGetMaxChunk();
		if (Buffer->MaxOctetsRequested == -1)
		{
			if ((Buffer->RingBuf = RingAlloc(Buffer->MaxOctetsAvailable, 
   				CaptureElems(Buffer->MaxOctetsAvailable, Buffer->CaptureSliceSize))) == NULL)
			{
				Buffer->MaxOctetsGranted = 0;
				return FALSE;
			}
			Buffer->MaxOctetsGranted = -1;
		}
		else
		{
			Buffer->MaxOctetsGranted = Buffer->MaxOctetsAvailable;
		 	if (Buffer->MaxOctetsGranted > Buffer->MaxOctetsRequested)
		 		Buffer->MaxOctetsGranted = Buffer->MaxOctetsRequested;
			if ((Buffer->RingBuf = RingAlloc(Buffer->MaxOctetsGranted, 
   				CaptureElems(Buffer->MaxOctetsGranted, Buffer->CaptureSliceSize))) == NULL)
			{
				Buffer->MaxOctetsGranted = 0;
				return FALSE;
			}
		}
	}
		
    Buffer->ChannelCallback = ChannelRegisterSink(Buffer->ChannelIndex, CapturePkt, CaptureDeRegister, Buffer);
        
	Buffer->CapturedPkts = 0;
	
    Buffer->TurnOnTime = SysTime();
    Buffer->TurnOnTimeMsec = TimeMsec();
	
    return TRUE;
}


BOOLEAN CaptureStop (BUFFER *Buffer)
{
    ChannelDeRegisterSink(Buffer->ChannelIndex, Buffer->ChannelCallback);
	Buffer->ChannelCallback = NULL;
	RingFree(Buffer->RingBuf);
	Buffer->RingBuf = NULL;
	
    Buffer->MaxOctetsAvailable = (MemoryGetAvail()/100)*memoryPerc;
   	if (Buffer->MaxOctetsAvailable > MemoryGetMaxChunk())
       	Buffer->MaxOctetsAvailable = MemoryGetMaxChunk();
	if (Buffer->MaxOctetsRequested == -1)
		Buffer->MaxOctetsGranted = -1;
	else
	{
		Buffer->MaxOctetsGranted = Buffer->MaxOctetsAvailable;
	 	if (Buffer->MaxOctetsGranted > Buffer->MaxOctetsRequested)
		 	Buffer->MaxOctetsGranted = Buffer->MaxOctetsRequested;
	}
			
	Buffer->CapturedPkts = 0;
	
    Buffer->TurnOnTime = 0;
    Buffer->TurnOnTimeMsec = 0;
	
    Buffer->FullStatus = 1;            /* spaceAvailable */
	
    return TRUE;
}

               
LONG CaptureElems(LONG available, LONG size)
{
    if (size == 0)
        size = 100;
    return (available/size)*2;          /* make sure number is even */
}


VOID CapturePkt(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *sink)
{
    BUFFER      *buffer = (BUFFER *)sink;
    PROT_OBJ    id =     {0, {1,3}};
	PROT_OBJ 	len =	 {0, {1,4}};
	PROT_OBJ 	caplen = {0, {1,5}};
    PROT_OBJ    time =   {0, {1,6}};
    PROT_OBJ    packet = {0, {1,7}};
    PROT_OBJ    status = {0, {1,8}};
    LONG        octetsCopied;
	ULONG		packetTime;
    CAPTURE		info;

    if (buffer->Status == SNMP_VALID && 
        buffer->ChannelIndex == channelIndex)
    {
        if (ProtGetField(pkt, &id) == FALSE ||
        	ProtGetField(pkt, &len) == FALSE ||
        	ProtGetField(pkt, &caplen) == FALSE ||
            ProtGetField(pkt, &time) == FALSE ||
            ProtGetField(pkt, &status) == FALSE ||
            ProtGetField(pkt, &packet) == FALSE) 
        {
            DnpapMessage(DMC_ERROR, CAPTURE_BADPACKET, "capture: bad packet");
            return;
        }

        if (buffer->FullStatus == 2 && buffer->FullAction == 1)
            return;                                         /* full & lock */
        if (buffer->CaptureSliceSize == 0 || 
            caplen.Syntax.LngInt < buffer->CaptureSliceSize)   /* snap packet */
            octetsCopied = caplen.Syntax.LngInt;
        else
            octetsCopied = buffer->CaptureSliceSize;                    
            
        if (octetsCopied + sizeof(CAPTURE) > RingAvailMem(buffer->RingBuf) ||
        	(RingGetElems(buffer->RingBuf) == RingGetMaxElems(buffer->RingBuf) && 
        	 RingSetMaxElems(buffer->RingBuf, RingGetMaxElems(buffer->RingBuf) + 10) == FALSE))
            buffer->FullStatus = 2;
        	
        if (buffer->FullStatus == 2 && buffer->FullAction == 1)
            return;

        if (octetsCopied + sizeof(CAPTURE) > buffer->MaxOctetsAvailable)
        	octetsCopied = buffer->MaxOctetsAvailable - sizeof(CAPTURE);
                                                    
        while (RingAvailMem(buffer->RingBuf) < octetsCopied + sizeof(CAPTURE) ||
        	   RingGetElems(buffer->RingBuf) == RingGetMaxElems(buffer->RingBuf))
        {
            if (RingGetMem(buffer->RingBuf, NULL, NULL) == FALSE)
            {   /* header */
                DnpapMessage(DMC_FATAL, CAPTURE_SYNC, "capture: ring buffer out of sync");
                DnpapExit(1);
            }
            if (RingGetMem(buffer->RingBuf, NULL, NULL) == FALSE)
            {   /* packet */
                DnpapMessage(DMC_FATAL, CAPTURE_SYNC, "capture: ring buffer out of sync");
                DnpapExit(1);
            }
            buffer->CapturedPkts--;
        }

		packetTime = time.Syntax.LngUns/1000;		   
		if (buffer->CapturedPkts == 0 || packetTime < buffer->TurnOnTimeMsec)
			buffer->TurnOnTimeMsec = packetTime - (TimeMsec() - buffer->TurnOnTimeMsec);
						   
        info.Time = packetTime - buffer->TurnOnTimeMsec;
        info.ID = id.Syntax.LngUns;                
        info.Len = (WORD) len.Syntax.LngInt+4;
		info.Status = (WORD) status.Syntax.LngInt;
        
        if (RingPutMem(buffer->RingBuf, (BYTE *) &info, sizeof(info)) == FALSE)
        {
            DnpapMessage(DMC_FATAL, CAPTURE_RING, "capture: ring buffer out of space");
            DnpapExit(1);
        }
        if (RingPutMem(buffer->RingBuf, packet.Syntax.Ptr, octetsCopied) == FALSE)
        {
            DnpapMessage(DMC_FATAL, CAPTURE_RING, "capture: ring buffer out of space");
            DnpapExit(1);
        }

        buffer->CapturedPkts++;            
    }

    return;
}


VOID CaptureDeRegister(ULONG channelIndex, VOID *sink)
{
    BUFFER      *buffer = (BUFFER *)sink;

    if (buffer->ChannelIndex == channelIndex)
	{
		ChannelDeRegisterSink(channelIndex, buffer->ChannelCallback);
		buffer->ChannelCallback = NULL;
	}
}
