/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>
#include <dtime.h>
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <snmp.h>
#include <mibsup.h>
#include <protocol.h>
#include <message.h>
#include <maxmem.h>
#include <ring.h>

#include "capturee.h"
#include "capturec.h"
#include "capture.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define CONTROLINDEXSIZE	1
#define CAPTUREINDEXSIZE  	2

#define CAPTURE_FIRST	1
#define CAPTURE_GET		2
#define CAPTURE_NXT		3

#define PACKET_SIZE_MAX		2048

static MIB_LOCAL *CaptureInstance = NULL;

EXPORT WORD memoryPerc = 75;

static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, LONG * elem);

static LONG CaptureSearch(WORD Request, SNMP_OBJECT * Obj, MIB_LOCAL * Local, WORD IdLen);

BOOLEAN CaptureMInit(VOID)
{
	MessageConfig(CAPTURE_ERROR, "Capture");
	return CaptureConfigInit();
}

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, LONG * elem)
{
	BOOLEAN first = FALSE;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case CAPTUREINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			if ((*elem = CaptureSearch(CAPTURE_NXT, obj, *local, idlen)) != -1)
			{
				obj->Id[idlen] = (*local)->Index;
				obj->IdLen = idlen + CAPTUREINDEXSIZE;
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL &&
			   (*elem = CaptureSearch(CAPTURE_FIRST, obj, *local, idlen)) == -1)
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen + 1] = 1;
		obj->IdLen = idlen + CAPTUREINDEXSIZE;
		break;
	case CONTROLINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + CONTROLINDEXSIZE;
		break;
	}
	return TRUE;
}

LONG CaptureSearch(WORD Request, SNMP_OBJECT * Obj, MIB_LOCAL * local, WORD IdLen)
{
	BYTE *p;
	LONG index;
	BUFFER *buffer = (BUFFER *) local->Data;

	if (buffer->CapturedPkts == 0)
		return -1;

	if (Obj->IdLen < IdLen + CAPTUREINDEXSIZE)
		Request = CAPTURE_FIRST;

	index = 2 * (Obj->Id[IdLen + 1] - 1);

	switch (Request)
	{
	case CAPTURE_GET:
		if (RingPeekMem(buffer->RingBuf, index, 0, &p, NULL, NULL) == TRUE)
			return index;
		break;
	case CAPTURE_FIRST:
		if (buffer->CapturedPkts > 0)
		{
			Obj->Id[IdLen + 1] = 1;
			return 0;
		}
		break;
	case CAPTURE_NXT:
		if (Obj->Id[IdLen + 1] < buffer->CapturedPkts)
		{
			Obj->Id[IdLen + 1]++;
			return index + 2;
		}
		break;
	}

	return -1;
}

WORD bufferControlIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD bufferControlChannelIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BUFFER *data;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BUFFER *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, CAPTURE_BOOLEAN_INDEX))
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = data->ChannelIndex;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (BUFFER *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		data->ChannelIndex = Obj->Syntax.LngInt;
		BooleanSetTrue(data->ObjectSet, CAPTURE_BOOLEAN_INDEX);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD bufferControlFullStatus(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BUFFER *data;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BUFFER *) local->Data;
		Obj->Syntax.LngInt = data->FullStatus;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD bufferControlFullAction(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BUFFER *data;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BUFFER *) local->Data;
		Obj->Syntax.LngInt = data->FullAction;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (BUFFER *) local->Data;
		if (Obj->Syntax.LngInt != 1 && Obj->Syntax.LngInt != 2)
			return SNMP_BADVALUE;
		data->FullAction = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD bufferControlCaptureSliceSize(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BUFFER *data;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BUFFER *) local->Data;
		Obj->Syntax.LngInt = data->CaptureSliceSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (BUFFER *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->Syntax.LngInt < 0 ||
			Obj->Syntax.LngInt > (data->MaxOctetsAvailable - sizeof(CAPTURE)) ||
			Obj->Syntax.LngInt > PACKET_SIZE_MAX)
			return SNMP_BADVALUE;
		data->CaptureSliceSize = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD bufferControlDownloadSliceSize(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BUFFER *data;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BUFFER *) local->Data;
		Obj->Syntax.LngInt = data->DownloadSliceSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (BUFFER *) local->Data;
		if (Obj->Syntax.LngInt < 1 || Obj->Syntax.LngInt > SNMP_SIZE_BUFCHR ||
			(data->CaptureSliceSize != 0 && Obj->Syntax.LngInt > data->CaptureSliceSize))
			return SNMP_BADVALUE;
		data->DownloadSliceSize = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD bufferControlDownloadOffset(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BUFFER *data;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BUFFER *) local->Data;
		Obj->Syntax.LngInt = data->DownloadOffset;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (BUFFER *) local->Data;
		if (Obj->Syntax.LngInt < 0 || Obj->Syntax.LngInt > data->CaptureSliceSize)
			return SNMP_BADVALUE;
		data->DownloadOffset = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD bufferControlMaxOctetsRequested(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BUFFER *data;
	Ring *newring;
	LONG octetsCopied;
	LONG maxmemchunk;
	BYTE *p;
	LONG i, n, len;
	BYTE packet[PACKET_SIZE_MAX];

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BUFFER *) local->Data;
		Obj->Syntax.LngInt = data->MaxOctetsRequested;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (BUFFER *) local->Data;
		maxmemchunk = (MemoryGetAvail() / 100) * memoryPerc;
		if (maxmemchunk > MemoryGetMaxChunk())
			maxmemchunk = MemoryGetMaxChunk();
		if (Obj->Syntax.LngInt == -1)
		{
			octetsCopied = maxmemchunk;
			if (octetsCopied <= data->MaxOctetsAvailable)
			{
				data->MaxOctetsRequested = -1;
				data->MaxOctetsGranted = -1;
				data->FullStatus = 1;
				return SNMP_NOERROR;
			}
			if (octetsCopied < data->CaptureSliceSize + sizeof(CAPTURE))
				return SNMP_BADVALUE;
			if ((newring = RingAlloc(octetsCopied, CaptureElems(octetsCopied, data->CaptureSliceSize))) == NULL)
				return SNMP_GENERROR;
			data->MaxOctetsRequested = -1;
			data->MaxOctetsGranted = -1;
		}
		else
		{
			if (Obj->Syntax.LngInt < 1)
				return SNMP_BADVALUE;
			if ((octetsCopied = Obj->Syntax.LngInt) > maxmemchunk)
				octetsCopied = maxmemchunk;
			if (octetsCopied < data->CaptureSliceSize + sizeof(CAPTURE))
				return SNMP_BADVALUE;
			if ((newring = RingAlloc(octetsCopied, CaptureElems(octetsCopied, data->CaptureSliceSize))) == NULL)
				return SNMP_GENERROR;
			data->MaxOctetsRequested = Obj->Syntax.LngInt;
			data->MaxOctetsGranted = octetsCopied;
		}

		if (octetsCopied > data->MaxOctetsAvailable)
			data->FullStatus = 1;
		data->MaxOctetsAvailable = octetsCopied;

		if (data->CapturedPkts > 0)
		{
			for (i = 2 * data->CapturedPkts - 1, n = 0; i > 0;)
			{
				if (RingPeekMem(data->RingBuf, i, 0, &p, NULL, &len) == FALSE)
				{
					DnpapMessage(DMC_FATAL, CAPTURE_SYNC, "capture: ring buffer out of sync");
					DnpapExit(1);
				}
				i--;
				n += len;
				if (RingPeekMem(data->RingBuf, i, 0, &p, NULL, &len) == FALSE)
				{
					DnpapMessage(DMC_FATAL, CAPTURE_SYNC, "capture: ring buffer out of sync");
					DnpapExit(1);
				}
				i--;
				n += len;
				if (n > octetsCopied)
					break;
			}

			i++;						/* backup one step to compensate i-- */
			if (n > octetsCopied)
				i += 2;					/* backup 2 steps forward */
			for (; i < 2 * data->CapturedPkts; i++)
			{
				len = 0;
				RingPeekMem(data->RingBuf, i, 0, &p, packet, &len);
				RingPutMem(newring, packet, len);
			}
		}

		RingFree(data->RingBuf);
		data->RingBuf = newring;
		data->CapturedPkts = RingGetElems(data->RingBuf) / 2;

		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}

	return SNMP_GENERROR;
}

WORD bufferControlMaxOctetsGranted(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BUFFER *data;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BUFFER *) local->Data;
		Obj->Syntax.LngInt = data->MaxOctetsGranted;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD bufferControlCapturedPackets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BUFFER *data;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BUFFER *) local->Data;
		Obj->Syntax.LngInt = data->CapturedPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD bufferControlTurnOnTime(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BUFFER *data;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BUFFER *) local->Data;
		Obj->Syntax.LngUns = data->TurnOnTime;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD bufferControlOwner(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BUFFER *data;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (BUFFER *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, CAPTURE_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		Obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (BUFFER *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, Obj->Syntax.BufChr, Obj->SyntaxLen);
		data->OwnerLen = Obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, CAPTURE_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD bufferControlStatus(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	BUFFER *data;

	local = MibRmon(Obj, CaptureInstance, IdLen, CONTROLINDEXSIZE);

	switch (Obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (BUFFER *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (BUFFER *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(Obj, &CaptureInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(BUFFER))) == NULL)
				{
					MibRemove(Obj, &CaptureInstance, IdLen, CONTROLINDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (BUFFER *) local->Data;
				memset(data, 0, sizeof(BUFFER));
				if (CaptureInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, CAPTURE_CREATE, "capture: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(Obj, &CaptureInstance, IdLen, CONTROLINDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (BUFFER *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (CaptureStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, CAPTURE_ACTIVE, "capture: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				else
					return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (CaptureStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, CAPTURE_DESTROY, "capture: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &CaptureInstance, IdLen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_INVALID:
				if (CaptureStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, CAPTURE_DESTROY, "capture: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &CaptureInstance, IdLen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (CaptureStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			case SNMP_VALID:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}

	return SNMP_GENERROR;
}

WORD captureBufferControlIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	LONG index = -1;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CAPTUREINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CAPTUREINDEXSIZE, &index) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (index == -1 && (index = CaptureSearch(CAPTURE_GET, Obj, local, IdLen)) == -1)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD captureBufferIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	LONG index = -1;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CAPTUREINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CAPTUREINDEXSIZE, &index) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (index == -1 && (index = CaptureSearch(CAPTURE_GET, Obj, local, IdLen)) == -1)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = Obj->Id[IdLen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD captureBufferPacketID(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	LONG index = -1;
	BUFFER *data;
	CAPTURE info;
	BYTE *p;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CAPTUREINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CAPTUREINDEXSIZE, &index) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (index == -1 && (index = CaptureSearch(CAPTURE_GET, Obj, local, IdLen)) == -1)
			return SNMP_NOSUCHNAME;
		data = (BUFFER *) local->Data;
		if (RingPeekMem(data->RingBuf, index, 0, &p, (BYTE *) & info, NULL) == FALSE)
			DnpapMessage(DMC_FATAL, CAPTURE_RING, "capture: missing expected ring buffer element");
		Obj->Syntax.LngInt = info.ID;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD captureBufferPacketData(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	LONG index = -1;
	BUFFER *data;
	BYTE *p;
	LONG len;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CAPTUREINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CAPTUREINDEXSIZE, &index) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (index == -1 && (index = CaptureSearch(CAPTURE_GET, Obj, local, IdLen)) == -1)
			return SNMP_NOSUCHNAME;
		data = (BUFFER *) local->Data;
		if (RingPeekMem(data->RingBuf, index + 1, 0, &p, NULL, &len) == FALSE)
			DnpapMessage(DMC_FATAL, CAPTURE_RING, "capture: missing expected ring buffer element");

		if (data->DownloadOffset + data->DownloadSliceSize <= len)
			Obj->SyntaxLen = len = data->DownloadSliceSize;
		else if (data->DownloadOffset <= len)
			Obj->SyntaxLen = len = len - data->DownloadOffset;
		else
			Obj->SyntaxLen = 0;
		RingPeekMem(data->RingBuf, index + 1, data->DownloadOffset, &p, Obj->Syntax.BufChr, &len);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD captureBufferPacketLength(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	LONG index = -1;
	BUFFER *data;
	BYTE *p;
	CAPTURE info;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CAPTUREINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CAPTUREINDEXSIZE, &index) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (index == -1 && (index = CaptureSearch(CAPTURE_GET, Obj, local, IdLen)) == -1)
			return SNMP_NOSUCHNAME;
		data = (BUFFER *) local->Data;
		if (RingPeekMem(data->RingBuf, index, 0, &p, (BYTE *) & info, NULL) == FALSE)
			DnpapMessage(DMC_FATAL, CAPTURE_RING, "capture: missing expected ring buffer element");
		Obj->Syntax.LngInt = info.Len;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD captureBufferPacketTime(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	LONG index = -1;
	BUFFER *data;
	BYTE *p;
	CAPTURE info;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CAPTUREINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CAPTUREINDEXSIZE, &index) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (index == -1 && (index = CaptureSearch(CAPTURE_GET, Obj, local, IdLen)) == -1)
			return SNMP_NOSUCHNAME;
		data = (BUFFER *) local->Data;
		if (RingPeekMem(data->RingBuf, index, 0, &p, (BYTE *) & info, NULL) == FALSE)
			DnpapMessage(DMC_FATAL, CAPTURE_RING, "capture: missing expected ring buffer element");
		Obj->Syntax.LngInt = info.Time;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD captureBufferPacketStatus(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	LONG index = -1;
	BUFFER *data;
	BYTE *p;
	CAPTURE info;

	if ((local = MibRmon(Obj, CaptureInstance, IdLen, CAPTUREINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CAPTUREINDEXSIZE, &index) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (index == -1 && (index = CaptureSearch(CAPTURE_GET, Obj, local, IdLen)) == -1)
			return SNMP_NOSUCHNAME;
		data = (BUFFER *) local->Data;
		if (RingPeekMem(data->RingBuf, index, 0, &p, (BYTE *) & info, NULL) == FALSE)
			DnpapMessage(DMC_FATAL, CAPTURE_RING, "capture: missing expected ring buffer element");
		Obj->Syntax.LngInt = info.Status;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

BOOLEAN CaptureRegisterAll(ULONG channelIndex)
{
	MIB_LOCAL *local;
	BUFFER *data;

	for (local = CaptureInstance; local != NULL; local = local->Next)
	{
		data = (BUFFER *) local->Data;
		if (data->ChannelIndex == channelIndex && data->ChannelCallback == NULL)
		{
			data->ChannelCallback = ChannelRegisterSink(data->ChannelIndex, CapturePkt, CaptureDeRegister, data);
		}
	}

	return TRUE;
}
