/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <stdlib.h>
#include <memory.h>
#include <dnpap.h>
#include <message.h>
#include <mac.h>
#include <event.h>
#include <trap.h>
#include <filter.h>
#include <capture.h>
#include <host.h>
#include <dhost.h>
#include <matrix.h>
#include <dmatrix.h>
#include <discnod.h>
#include <disccon.h>

#include "channele.h"
#include "channelc.h"

IMPORT BOOLEAN BeholderStrictRMON;


static VOID ChannelFillTrap(CHANNEL *channel);

static BOOLEAN ChannelMacCallback(MAC_COLL * collector, PROT_PKT * pkt);
static VOID DeRegisterAllFilters(CHANNEL *channel);
static VOID DeRegisterAllSinks(CHANNEL *channel);


static SNMP_OBJECT trapobj[] =
{
	{SNMP_PDU_TRAP,
	 {MIB_CHANNEL, MIB_INDEX_CHANNELINDEX}, MIB_LEN_CHANNEL + 2, SNMP_INTEGER, {0}},
	{SNMP_PDU_TRAP,
	 {MIB_CHANNEL, MIB_INDEX_CHANNELIFINDEX}, MIB_LEN_CHANNEL + 2, SNMP_INTEGER, {0}},
	{SNMP_PDU_TRAP,
	 {MIB_CHANNEL, MIB_INDEX_CHANNELACCEPTTYPE}, MIB_LEN_CHANNEL + 2, SNMP_INTEGER, {0}},
	{SNMP_PDU_TRAP,
	 {MIB_CHANNEL, MIB_INDEX_CHANNELMATCHES}, MIB_LEN_CHANNEL + 2, SNMP_COUNTER, {0}},
	{SNMP_PDU_TRAP,
	 {MIB_CHANNEL, MIB_INDEX_CHANNELEVENTSTATUS}, MIB_LEN_CHANNEL + 2, SNMP_INTEGER, {0}}
};
static INT traplen = sizeof(trapobj) / sizeof(trapobj[0]);


BOOLEAN ChannelInit(CHANNEL * channel)
{
	channel->IfIndex = 1;				/* first interface */
	channel->AcceptType = 1;			/* accept if filters TRUE */
	channel->DataControl = 2;			/* no flow through channel */
	channel->EventIndex = 0;
	channel->EventStatus = 1;			/* only one event is fired */
	channel->EventHook = NULL;
	channel->Matches = 0;
	                                                                            
    channel->TurnOnEventIndex = 0;
    channel->TurnOnCallback = NULL;
    channel->TurnOffEventIndex = 0;
    channel->TurnOffCallback = NULL;
    
    channel->Filters = NULL;
    channel->Sinks = NULL;
    
	if ((channel->Iface =
		 MacIfaceGet((WORD) channel->IfIndex)) == NULL)
	{
		DnpapMessage(DMC_ERROR, CHANNEL_NETINIT, "channel: network init");
		return FALSE;
	}

	BooleanSetAllTrue(channel->ObjectSet);
	BooleanSetFalse(channel->ObjectSet, CHANNEL_BOOLEAN_INDEX);
	BooleanSetFalse(channel->ObjectSet, CHANNEL_BOOLEAN_OWNER);

	return TRUE;
}


BOOLEAN ChannelStart(CHANNEL * channel)
{
	if (BeholderStrictRMON && !BooleanCheckAllTrue(channel->ObjectSet))
		return FALSE;
      
	if (channel->EventIndex > 0)
		channel->EventHook = EventRegisterGenerator(channel->EventIndex, ChannelDeRegisterGenerator, channel);
    if (channel->TurnOnEventIndex > 0)
        channel->TurnOnCallback = EventRegisterParty(channel->TurnOnEventIndex, ChannelEventGenerated, ChannelDeRegisterParty, channel);
    if (channel->TurnOffEventIndex > 0)
        channel->TurnOffCallback = EventRegisterParty(channel->TurnOffEventIndex, ChannelEventGenerated, ChannelDeRegisterParty, channel);
    
	FilterRegisterAll(channel->Index);
	CaptureRegisterAll(channel->Index);
	HostRegisterAll(channel->Index);
	DnpapHostRegisterAll(channel->Index);
	MatrixRegisterAll(channel->Index);
	DnpapMatrixRegisterAll(channel->Index);
	DiscNodeRegisterAll(channel->Index);
	DiscConnRegisterAll(channel->Index);

	channel->Coll.Rcve = ChannelMacCallback;
	channel->Coll.specific = channel;
	if (!MacCollRegister(&(channel->Coll)))
	{
		DnpapMessage(DMC_ERROR, CHANNEL_NETINIT, "channel: network init");
        EventDeRegisterGenerator(channel->EventIndex, channel->EventHook);
        channel->EventHook = NULL;
        EventDeRegisterParty(channel->TurnOnEventIndex, channel->TurnOnCallback);
        channel->TurnOnCallback = NULL;
        EventDeRegisterParty(channel->TurnOffEventIndex, channel->TurnOffCallback);
        channel->TurnOffCallback = NULL;
		return FALSE;
	}
    channel->IfType = channel->Iface->type;
    
	return TRUE;
}


BOOLEAN ChannelStop(CHANNEL * channel)
{
	MacCollRemove(&(channel->Coll));
	DeRegisterAllFilters(channel);
	if (channel->Filters != NULL)
		DnpapMessage(DMC_FATAL, CHANNEL_FILTERS, "channel: internal error, not all filters were removed");
	channel->Filters = NULL;
	DeRegisterAllSinks(channel);
	if (channel->Sinks != NULL)
		DnpapMessage(DMC_FATAL, CHANNEL_SINKS, "channel: internal error, not all sinks were removed");
	channel->Sinks = NULL;
    EventDeRegisterGenerator(channel->EventIndex, channel->EventHook);
	channel->EventHook = NULL;
    EventDeRegisterParty(channel->TurnOnEventIndex, channel->TurnOnCallback);
	channel->TurnOnCallback = NULL;
    EventDeRegisterParty(channel->TurnOffEventIndex, channel->TurnOffCallback);
	channel->TurnOffCallback = NULL;
	
	channel->Matches = 0;
	
	return TRUE;
}


BOOLEAN ChannelMacCallback(MAC_COLL * collector, PROT_PKT * pkt)
{
	BOOLEAN match, packetAccept;
    FilterCallback *filter;
    SinkCallback *sink;
    PROT_OBJ Interface = {0, {1,2}};
	CHANNEL *channel = collector->specific;
                                 
    if (channel->Status == SNMP_INVALID ||
        ProtGetField(pkt, &Interface) == FALSE ||
        Interface.Syntax.LngInt != channel->IfIndex)
        return TRUE;
        
    match = FALSE;
    for (filter = channel->Filters; filter != NULL; filter = filter->Next)
        if ((match = filter->Match(channel->Index, channel->IfType, pkt, filter->Filter)) == TRUE)
            break;
    
	if ((channel->AcceptType == 1 && match == TRUE) ||
		(channel->AcceptType == 2 && match == FALSE))
	{
		channel->Matches++;
		packetAccept = TRUE;
	}
	else
		packetAccept = FALSE;

	if (channel->DataControl == 1)		/* channel on */
	{
		if (packetAccept == TRUE)
		{
			switch (channel->EventStatus)
			{
			case 1:					/* event ready */
                ChannelFillTrap(channel);
				if (EventGenerate(channel->EventHook, "channel", SNMP_TRAP_ENTSPECIFIC, TRAP_SPECIFIC_PACKETMATCH, trapobj, traplen) == TRUE)
					channel->EventStatus = 2;
				break;
			case 3:					/* event always ready */
                ChannelFillTrap(channel);
				EventGenerate(channel->EventHook, "channel", SNMP_TRAP_ENTSPECIFIC, TRAP_SPECIFIC_PACKETMATCH, trapobj, traplen);
				break;
			case 2:					/* event fired once */
				break;
			}

            for (sink = channel->Sinks; sink != NULL; sink = sink->Next)
                sink->Dump(channel->Index, channel->IfType, pkt, sink->Sink);
		}
	}
    
	return TRUE;
}

BOOLEAN ChannelEventGenerated(ULONG eventIndex, VOID *party)
{
	CHANNEL *channel = (CHANNEL *) party;

	if (channel->Status != SNMP_VALID)
		return TRUE;
	if (eventIndex == channel->TurnOnEventIndex)
		channel->DataControl = 1;
	if (eventIndex == channel->TurnOffEventIndex)
		channel->DataControl = 2;
        
    return TRUE;
}

VOID ChannelFillTrap(CHANNEL *channel)
{
	INT i;

	/* set channel instance in trap varbinds */
	for (i = 0; i < traplen; i++)
		trapobj[i].Id[MIB_LEN_CHANNEL + 1] = channel->Index;

	/* set values in varbinds */
	trapobj[0].Syntax.LngInt = channel->Index;
	trapobj[1].Syntax.LngInt = channel->IfIndex;
	trapobj[2].Syntax.LngInt = channel->AcceptType;
	trapobj[3].Syntax.LngUns = channel->Matches;
	trapobj[4].Syntax.LngInt = channel->EventStatus;
}

VOID ChannelDeRegisterGenerator(ULONG eventIndex, VOID *generator)
{
	CHANNEL *channel = (CHANNEL *) generator;

	if (eventIndex == channel->EventIndex)
	{
		EventDeRegisterGenerator(eventIndex, channel->EventHook);
		channel->EventHook = NULL;
	}
}

VOID ChannelDeRegisterParty(ULONG eventIndex, VOID *party)
{
	CHANNEL *channel = (CHANNEL *) party;

	if (eventIndex == channel->TurnOnEventIndex)
	{
		EventDeRegisterParty(eventIndex, channel->TurnOnCallback);
		channel->TurnOnCallback = NULL;
	}
	if (eventIndex == channel->TurnOffEventIndex)
	{
		EventDeRegisterParty(eventIndex, channel->TurnOffCallback);
		channel->TurnOffCallback = NULL;
	}
}

FilterCallback *ChannelAddFilter(CHANNEL *channel, ChannelFilterCallback match, ChannelFilterDeRegisterCallback deregister, VOID *filter)
{
FilterCallback *callback;
     
    if ((callback = DnpapMalloc(sizeof(FilterCallback))) == NULL)
        return NULL;
    callback->Match = match;
    callback->Filter = filter;
    callback->DeRegister = deregister; 
    callback->Prev = NULL;
    callback->Next = channel->Filters;
    if (channel->Filters != NULL)
        channel->Filters->Prev = callback;
    channel->Filters = callback;
    
    return callback;
}

SinkCallback *ChannelAddSink(CHANNEL *channel, ChannelSinkCallback dump, ChannelSinkDeRegisterCallback deregister, VOID *sink)
{
SinkCallback *callback;

    if ((callback = DnpapMalloc(sizeof(SinkCallback))) == NULL)
        return NULL;
    callback->Dump = dump;
    callback->DeRegister = deregister; 
    callback->Sink = sink;
    callback->Prev = NULL;
    callback->Next = channel->Sinks;
    if (channel->Sinks != NULL)
        channel->Sinks->Prev = callback;
    channel->Sinks = callback;
    
    return callback;
}

BOOLEAN ChannelRemoveFilter(CHANNEL *channel, FilterCallback *callback)
{                          
    if (callback == NULL)
        return TRUE;
    if (callback->Next != NULL)
        callback->Next->Prev = callback->Prev;
    if (callback->Prev == NULL)
        channel->Filters = callback->Next;
    else
        callback->Prev->Next = callback->Next;
        
    DnpapFree(callback);
    
    return TRUE;
}

BOOLEAN ChannelRemoveSink(CHANNEL *channel, SinkCallback *callback)
{
    if (callback == NULL)
        return TRUE;
    if (callback->Next != NULL)
        callback->Next->Prev = callback->Prev;
    if (callback->Prev == NULL)
        channel->Sinks = callback->Next;
    else
        callback->Prev->Next = callback->Next;
        
    DnpapFree(callback);
    
    return TRUE;
}


VOID DeRegisterAllFilters(CHANNEL *channel)
{
FilterCallback *filter;

    for (filter = channel->Filters; filter != NULL; filter = channel->Filters)
        filter->DeRegister(channel->Index, filter->Filter);
}


VOID DeRegisterAllSinks(CHANNEL *channel)
{
SinkCallback *sink;

	for (sink = channel->Sinks; sink != NULL; sink = channel->Sinks)
		sink->DeRegister(channel->Index, sink->Sink);
}
