/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#ifndef _CHANNELC_H
#define _CHANNELC_H


#include <dnpap.h>
#include <boolset.h>
#include <mac.h>
#include <snmp.h>
#include <event.h>

#include "channeld.h"
#include "channelx.h"


#define CHANNEL_BOOLEAN_INDEX		1
#define CHANNEL_BOOLEAN_OWNER		2


typedef struct _CHANNEL
{
    ULONG       Index;
    ULONG       IfIndex;
    LONG        IfType;
    LONG        AcceptType;
    LONG        DataControl;
    ULONG       TurnOnEventIndex;
    PartyCallback *TurnOnCallback;
    ULONG       TurnOffEventIndex;
    PartyCallback *TurnOffCallback;
    ULONG       EventIndex;
    LONG        EventStatus;
    EventHook  *EventHook;
    LWORD       Matches;
    BYTE        Description[SNMP_SIZE_BUFCHR];
    WORD        DescriptionLen;
    BYTE        Owner[SNMP_SIZE_BUFCHR];
    WORD        OwnerLen;
    LONG        Status;

    MAC_COLL    Coll;
    MAC_IFACE  *Iface;  
    
    FilterCallback *Filters;
    SinkCallback *Sinks;

    BooleanSet  ObjectSet;
	
} CHANNEL;


struct FilterCallback_
{   
    ChannelFilterCallback Match;
    ChannelFilterDeRegisterCallback DeRegister;
    VOID *Filter;
    FilterCallback *Prev;
    FilterCallback *Next;
};


struct SinkCallback_
{   
    ChannelSinkCallback Dump;
    ChannelSinkDeRegisterCallback DeRegister;
    VOID *Sink;                 
    SinkCallback *Prev;
    SinkCallback *Next;
};


BOOLEAN ChannelInit (CHANNEL *channel);
BOOLEAN ChannelStart (CHANNEL *channel);
BOOLEAN ChannelStop (CHANNEL *channel);

BOOLEAN ChannelEventGenerated(ULONG eventIndex, VOID *party);
VOID ChannelDeRegisterGenerator(ULONG eventIndex, VOID *generator);
VOID ChannelDeRegisterParty(ULONG eventIndex, VOID *party);

FilterCallback *ChannelAddFilter(CHANNEL *channel, ChannelFilterCallback match, ChannelFilterDeRegisterCallback deregister, VOID *filter);
SinkCallback *ChannelAddSink(CHANNEL *channel, ChannelSinkCallback dump, ChannelSinkDeRegisterCallback deregister, VOID *sink);
BOOLEAN ChannelRemoveFilter(CHANNEL *channel, FilterCallback *callback);
BOOLEAN ChannelRemoveSink(CHANNEL *channel, SinkCallback *callback);


#endif
