/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mibsup.h>
#include <message.h>
#include <mac.h>

#include "channele.h"
#include "channelc.h"
#include "channel.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define INDEXSIZE   1

static MIB_LOCAL *ChannelInstance = NULL;

static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize);

BOOLEAN ChannelMInit(VOID)
{
	MessageConfig(CHANNEL_ERROR, "Channel");
	return TRUE;
}

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize)
{
	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->IdLen == idlen || (*local)->Index > obj->Id[idlen] ||
		(*local = (*local)->Next) != NULL)
	{
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + indexsize;
		return TRUE;
	}

	return FALSE;
}

WORD ChannelIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(Obj, ChannelInstance, IdLen, INDEXSIZE)) == NULL)
		return (SNMP_NOSUCHNAME);

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ChannelIfIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	MAC_IFACE *iface;
	CHANNEL *data;

	if ((local = MibRmon(Obj, ChannelInstance, IdLen, INDEXSIZE)) == NULL)
		return (SNMP_NOSUCHNAME);

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (CHANNEL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, CHANNEL_BOOLEAN_INDEX))
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = data->IfIndex;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (CHANNEL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if ((iface = MacIfaceGet((WORD) Obj->Syntax.LngInt)) == NULL)
			return SNMP_BADVALUE;
		if (iface->type != MAC_TYPE_ETHERNET_CSMACD && iface->type != MAC_TYPE_88023_CSMACD)
			return SNMP_BADVALUE;
		data->IfIndex = Obj->Syntax.LngInt;
		data->Iface = iface;
		BooleanSetTrue(data->ObjectSet, CHANNEL_BOOLEAN_INDEX);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ChannelAcceptType(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	CHANNEL *data;

	if ((local = MibRmon(Obj, ChannelInstance, IdLen, INDEXSIZE)) == NULL)
		return (SNMP_NOSUCHNAME);

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (CHANNEL *) local->Data;
		Obj->Syntax.LngInt = data->AcceptType;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (CHANNEL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->Syntax.LngInt != 1 && Obj->Syntax.LngInt != 2)
			return SNMP_BADVALUE;
		data->AcceptType = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ChannelDataControl(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	CHANNEL *data;

	if ((local = MibRmon(Obj, ChannelInstance, IdLen, INDEXSIZE)) == NULL)
		return (SNMP_NOSUCHNAME);

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (CHANNEL *) local->Data;
		Obj->Syntax.LngInt = data->DataControl;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (CHANNEL *) local->Data;
		if (Obj->Syntax.LngInt != 1 && Obj->Syntax.LngInt != 2)
			return SNMP_BADVALUE;
		data->DataControl = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ChannelTurnOnEventIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	CHANNEL *data;

	if ((local = MibRmon(Obj, ChannelInstance, IdLen, INDEXSIZE)) == NULL)
		return (SNMP_NOSUCHNAME);

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (CHANNEL *) local->Data;
		Obj->Syntax.LngInt = data->TurnOnEventIndex;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (CHANNEL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		data->TurnOnEventIndex = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ChannelTurnOffEventIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	CHANNEL *data;

	if ((local = MibRmon(Obj, ChannelInstance, IdLen, INDEXSIZE)) == NULL)
		return (SNMP_NOSUCHNAME);

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (CHANNEL *) local->Data;
		Obj->Syntax.LngInt = data->TurnOffEventIndex;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (CHANNEL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		data->TurnOffEventIndex = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ChannelEventIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	CHANNEL *data;

	if ((local = MibRmon(Obj, ChannelInstance, IdLen, INDEXSIZE)) == NULL)
		return (SNMP_NOSUCHNAME);

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (CHANNEL *) local->Data;
		Obj->Syntax.LngInt = data->EventIndex;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (CHANNEL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		data->EventIndex = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ChannelEventStatus(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	CHANNEL *data;

	if ((local = MibRmon(Obj, ChannelInstance, IdLen, INDEXSIZE)) == NULL)
		return (SNMP_NOSUCHNAME);

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (CHANNEL *) local->Data;
		Obj->Syntax.LngInt = data->EventStatus;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (CHANNEL *) local->Data;
		if (Obj->Syntax.LngInt != 1 && Obj->Syntax.LngInt != 3)
			return SNMP_BADVALUE;
		data->EventStatus = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ChannelMatches(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	CHANNEL *data;

	if ((local = MibRmon(Obj, ChannelInstance, IdLen, INDEXSIZE)) == NULL)
		return (SNMP_NOSUCHNAME);

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (CHANNEL *) local->Data;
		Obj->Syntax.LngInt = data->Matches;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ChannelDescription(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	CHANNEL *data;

	if ((local = MibRmon(Obj, ChannelInstance, IdLen, INDEXSIZE)) == NULL)
		return (SNMP_NOSUCHNAME);

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (CHANNEL *) local->Data;
		memcpy(Obj->Syntax.BufChr, data->Description, data->DescriptionLen);
		Obj->SyntaxLen = data->DescriptionLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (CHANNEL *) local->Data;
		memcpy(data->Description, Obj->Syntax.BufChr, Obj->SyntaxLen);
		data->DescriptionLen = Obj->SyntaxLen;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ChannelOwner(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	CHANNEL *data;

	if ((local = MibRmon(Obj, ChannelInstance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (CHANNEL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, CHANNEL_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		Obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (CHANNEL *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, Obj->Syntax.BufChr, Obj->SyntaxLen);
		data->OwnerLen = Obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, CHANNEL_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD ChannelStatus(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	CHANNEL *data;

	local = MibRmon(Obj, ChannelInstance, IdLen, INDEXSIZE);

	switch (Obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (CHANNEL *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (CHANNEL *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(Obj, &ChannelInstance, IdLen, INDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(CHANNEL))) == NULL)
				{
					MibRemove(Obj, &ChannelInstance, IdLen, INDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (CHANNEL *) local->Data;
				memset(data, 0, sizeof(CHANNEL));
				if (ChannelInit(data) == TRUE)
				{
					data->Index = local->Index;
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, CHANNEL_CREATE, "channel: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(Obj, &ChannelInstance, IdLen, INDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (CHANNEL *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (ChannelStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, CHANNEL_ACTIVE, "channel: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				else
					return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (ChannelStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, CHANNEL_DESTROY, "channel: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &ChannelInstance, IdLen, INDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_INVALID:
				if (ChannelStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, CHANNEL_DESTROY, "channel: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &ChannelInstance, IdLen, INDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (ChannelStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			case SNMP_VALID:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

FilterCallback *ChannelRegisterFilter(ULONG channelIndex, ChannelFilterCallback match, ChannelFilterDeRegisterCallback deregister, VOID * filter)
{
	MIB_LOCAL *local;
	CHANNEL *data;
	FilterCallback *callback;

	for (local = ChannelInstance; local != NULL; local = local->Next)
	{
		data = (CHANNEL *) local->Data;
		if (data->Index < channelIndex)
			continue;
		if (data->Index != channelIndex)
			return NULL;
		if ((callback = ChannelAddFilter(data, match, deregister, filter)) != NULL)
			return callback;
		break;
	}

	return NULL;
}

BOOLEAN ChannelDeRegisterFilter(ULONG channelIndex, FilterCallback * filter)
{
	MIB_LOCAL *local;
	CHANNEL *data;

	if (filter == NULL)
		return TRUE;

	for (local = ChannelInstance; local != NULL; local = local->Next)
	{
		data = (CHANNEL *) local->Data;
		if (data->Index < channelIndex)
			continue;
		if (data->Index != channelIndex)
			return FALSE;
		if (ChannelRemoveFilter(data, filter) == TRUE)
			return TRUE;
		break;
	}

	return FALSE;
}

SinkCallback *ChannelRegisterSink(ULONG channelIndex, ChannelSinkCallback dump, ChannelSinkDeRegisterCallback deregister, VOID * sink)
{
	MIB_LOCAL *local;
	CHANNEL *data;
	SinkCallback *callback;

	for (local = ChannelInstance; local != NULL; local = local->Next)
	{
		data = (CHANNEL *) local->Data;
		if (data->Index < channelIndex)
			continue;
		if (data->Index != channelIndex)
			return NULL;
		if ((callback = ChannelAddSink(data, dump, deregister, sink)) != NULL)
			return callback;
		break;
	}

	return NULL;
}

BOOLEAN ChannelDeRegisterSink(ULONG channelIndex, SinkCallback * sink)
{
	MIB_LOCAL *local;
	CHANNEL *data;

	if (sink == NULL)
		return TRUE;

	for (local = ChannelInstance; local != NULL; local = local->Next)
	{
		data = (CHANNEL *) local->Data;
		if (data->Index < channelIndex)
			continue;
		if (data->Index != channelIndex)
			return FALSE;
		if (ChannelRemoveSink(data, sink) == TRUE)
			return TRUE;
		break;
	}

	return FALSE;
}

BOOLEAN ChannelRegisterAllGenerators(ULONG eventIndex)
{
	MIB_LOCAL *local;
	CHANNEL *channel;

	for (local = ChannelInstance; local != NULL; local = local->Next)
	{
		channel = (CHANNEL *) local->Data;
		if (channel->EventIndex == eventIndex && channel->EventHook == NULL)
			channel->EventHook = EventRegisterGenerator(eventIndex, ChannelDeRegisterGenerator, channel);
	}

	return TRUE;
}

BOOLEAN ChannelRegisterAllParties(ULONG eventIndex)
{
	MIB_LOCAL *local;
	CHANNEL *channel;

	for (local = ChannelInstance; local != NULL; local = local->Next)
	{
		channel = (CHANNEL *) local->Data;
		if (channel->TurnOnEventIndex == eventIndex && channel->TurnOnCallback == NULL)
			channel->TurnOnCallback = EventRegisterParty(channel->TurnOnEventIndex, ChannelEventGenerated, ChannelDeRegisterParty, channel);
		if (channel->TurnOffEventIndex == eventIndex && channel->TurnOffCallback == NULL)
			channel->TurnOffCallback = EventRegisterParty(channel->TurnOffEventIndex, ChannelEventGenerated, ChannelDeRegisterParty, channel);
	}

	return TRUE;
}
