/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>
#include <log.h>
#include <alarm.h>
#include <channel.h>
#include <stdlib.h>

#include "event_d.h"
#include "event_e.h"
#include "event_x.h"
#include "event_c.h"


IMPORT BOOLEAN BeholderStrictRMON;


static CHAR MODULE[] = "event";


static VOID DeRegisterAllGenerators(Event* event);
static VOID DeRegisterAllParties(Event* event);


BOOLEAN EventCInit(Event* event)
{
    event->Index = 0;
	event->Description[0] = '\0';
    event->DescrLen = 0;
	event->Type = EVENT_NONE;
	event->Community[0] = '\0';
    event->CommLen = 0;
	event->LastTimeSent = 0;
	event->Owner[0] = '\0';
	event->OwnerLen = 0;
	event->Status = SNMP_INVALID;
	
	event->NrLogs = 0;
	event->LogPos = 0;
	event->LogIndex = 1;
	event->LogBounced = FALSE;
	event->Logs = NULL;
	             
    event->Generators = NULL;
    event->Parties = NULL;
    
	BooleanSetAllTrue(event->ObjectSet);
	BooleanSetFalse(event->ObjectSet, EVENT_BOOLEAN_COMMUNITY);
	BooleanSetFalse(event->ObjectSet, EVENT_BOOLEAN_OWNER);

    return TRUE;
}


BOOLEAN EventCStart(Event* event)
{
	if (BeholderStrictRMON && !BooleanCheckAllTrue(event->ObjectSet))
		return FALSE;

	AlarmRegisterAllGenerators(event->Index);
	ChannelRegisterAllGenerators(event->Index);
	ChannelRegisterAllParties(event->Index);

    return TRUE;
}


BOOLEAN EventCStop(Event* event)
{        
    DeRegisterAllGenerators(event);
	if (event->Generators != NULL)
		DnpapMessage(DMC_FATAL, EVENT_GENERATORS, "event: internal error, not all generators were removed");
	event->Generators = NULL;
    DeRegisterAllParties(event);
	if (event->Parties != NULL)
		DnpapMessage(DMC_FATAL, EVENT_PARTIES, "event: internal error, not all parties were removed");
	event->Parties = NULL;
	DelLog(event->Logs, event->NrLogs);
	event->NrLogs = 0;
	event->LogPos = 0;
	event->LogIndex = 1;
	event->LogBounced = FALSE;
	event->Logs = NULL;
	             
	event->LastTimeSent = 0;
	
    return TRUE;
}


EventHook *EventAddGenerator(Event* event, EventGeneratorDeRegisterCallback deregister, VOID *generator)
{
EventHook *hook;
     
    if ((hook = DnpapMalloc(sizeof(EventHook))) == NULL)
        return NULL;
    hook->Event = event;
    hook->DeRegister = deregister;
    hook->Generator = generator;
    hook->Prev = NULL;
    hook->Next = event->Generators;
    if (event->Generators != NULL)
        event->Generators->Prev = hook;
    event->Generators = hook;
    
    return hook;
}


BOOLEAN EventRemoveGenerator(Event* event, EventHook *hook)
{
    if (hook == NULL)
        return TRUE;
    if (hook->Next != NULL)
        hook->Next->Prev = hook->Prev;
    if (hook->Prev == NULL)
        event->Generators = hook->Next;
    else
        hook->Prev->Next = hook->Next;
        
    DnpapFree(hook);
    
    return TRUE;
}


PartyCallback *EventAddParty(Event* event, EventPartyCallback notify, EventPartyDeRegisterCallback deregister, VOID *party)
{
PartyCallback *callback;
     
    if ((callback = DnpapMalloc(sizeof(PartyCallback))) == NULL)
        return NULL;
    callback->Notify = notify;
    callback->DeRegister = deregister; 
    callback->Party = party; 
    callback->Prev = NULL;
    callback->Next = event->Parties;
    if (event->Parties != NULL)
        event->Parties->Prev = callback;
    event->Parties = callback;
    
    return callback;
}


BOOLEAN EventRemoveParty(Event* event, PartyCallback *callback)
{
    if (callback == NULL)
        return TRUE;
    if (callback->Next != NULL)
        callback->Next->Prev = callback->Prev;
    if (callback->Prev == NULL)
        event->Parties = callback->Next;
    else
        callback->Prev->Next = callback->Next;
        
    DnpapFree(callback);
    
    return TRUE;
}


BOOLEAN EventNotify(Event* event)
{             
BOOLEAN result;
PartyCallback *party;
        
    result = TRUE;
    for (party = event->Parties; party != NULL; party = party->Next)
        if (party->Notify(event->Index, party->Party) == FALSE)
            result = FALSE;
    return result;
}


VOID DeRegisterAllGenerators(Event* event)
{
EventHook *hook;

    for (hook = event->Generators; hook != NULL; hook = event->Generators)
        hook->DeRegister(event->Index, hook->Generator);
}


VOID DeRegisterAllParties(Event* event)
{
PartyCallback *party;

    for (party = event->Parties; party != NULL; party = event->Parties)
        party->DeRegister(event->Index, party->Party);
}


Event* NewEvent(VOID)
{
    return DnpapMalloc(sizeof(Event));
}


VOID DelEvent(Event* ptr)
{
    DnpapFree(ptr);
}
