/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <stdio.h>
#include <string.h>
#include <dnpap.h>
#include <mibsup.h>
#include <message.h>
#include <channel.h>

#include "filtere.h"
#include "filterc.h"

IMPORT BOOLEAN BeholderStrictRMON;



BOOLEAN FilterCInit(FILTER *filter)
{
    filter->ChannelIndex = 1;
    filter->ChannelCallback = NULL;

    filter->PktData = NULL;
    filter->PktDataMask = NULL;
    filter->PktDataNotMask = NULL;

    filter->PktDataLen = 0;
    filter->PktDataMaskLen = 0;
    filter->PktDataNotMaskLen = 0;
                        
	filter->PktStatus = -1;
	filter->PktStatusMask = -1;
	filter->PktStatusNotMask = -1;

	BooleanSetAllTrue(filter->ObjectSet);
	BooleanSetFalse(filter->ObjectSet, FILTER_BOOLEAN_CHANNEL);
	BooleanSetFalse(filter->ObjectSet, FILTER_BOOLEAN_OWNER);

    return TRUE;
}


BOOLEAN FilterCStart(FILTER *filter)
{
WORD i;
    
	if (BeholderStrictRMON && !BooleanCheckAllTrue(filter->ObjectSet))
		return FALSE;

    filter->PktDataNotMaskAllZero = TRUE;
    for (i = 0; i < filter->PktDataNotMaskLen; i++)
        if (filter->PktDataNotMask[i] != 0x00)
        {
            filter->PktDataNotMaskAllZero = FALSE;
            break;
        }
        
    filter->ChannelCallback = ChannelRegisterFilter(filter->ChannelIndex, FilterPkt, FilterDeRegister, filter);
        
    return TRUE;
}


BOOLEAN FilterCStop(FILTER *filter)
{       
    ChannelDeRegisterFilter(filter->ChannelIndex, filter->ChannelCallback);
	filter->ChannelCallback = NULL;
	
    DnpapFree(filter->PktData);
	filter->PktData = NULL;
    DnpapFree(filter->PktDataMask);
	filter->PktDataMask = NULL;
    DnpapFree(filter->PktDataNotMask);
	filter->PktDataNotMask = NULL;
    filter->PktDataLen = 0;
    filter->PktDataMaskLen = 0;
    filter->PktDataNotMaskLen = 0;
	
    return TRUE;
}


BOOLEAN FilterPkt(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *fptr)
{
FILTER *filter = (FILTER *) fptr;
PROT_OBJ len =    {0, {1,5}};
PROT_OBJ packet = {0, {1,7}};
PROT_OBJ status = {0, {1,8}};
BYTE scratch;
WORD bi, fb, lb;
BOOLEAN rule2fail, rule3fail;

    if (filter->Status == SNMP_VALID && filter->ChannelIndex == channelIndex)
    {
        if (ProtGetField(pkt, &len) == FALSE ||
        	ProtGetField(pkt, &status) == FALSE ||
            ProtGetField(pkt, &packet) == FALSE) 
        {
            DnpapMessage(DMC_ERROR, FILTER_BADPACKET, "filter: bad packet");
            return FALSE;
        }
    
		if (filter->PktStatus != -1)
		{
		LONG mask, notmask;

			mask = filter->PktStatusMask == -1 ?
						FILTER_STATUS_MAX : filter->PktStatusMask;
			notmask = filter->PktStatusNotMask == -1 ?
						0 : filter->PktStatusNotMask;

			rule2fail = FALSE;
			if (notmask == 0)
				rule3fail = FALSE;
			else
				rule3fail = TRUE;
			scratch = (status.Syntax.LngInt ^ filter->PktStatus) & mask;
			if ((scratch & ~notmask) != 0)
				rule2fail = TRUE;
			if (rule3fail == TRUE && (scratch & notmask) != 0)
				rule3fail = FALSE;
			if (!(rule2fail == FALSE && rule3fail == FALSE))
				return FALSE;
		}

        if (filter->PktDataLen + filter->PktDataOffset > len.Syntax.LngInt)
            return FALSE;       /*  rule 1 of packet matching  */
        if (filter->PktDataLen == 0 || filter->PktData == NULL)
            return TRUE;	    /*  no data, match everything */
        fb = (WORD)filter->PktDataOffset;
        lb = (WORD)filter->PktDataOffset+filter->PktDataLen;

        rule2fail = FALSE;
        if (filter->PktDataNotMaskAllZero == TRUE)
            rule3fail = FALSE;
        else
            rule3fail = TRUE;
        for (bi = fb; bi < lb; bi++)
        {
            scratch = (((BYTE*)packet.Syntax.Ptr)[bi] ^ filter->PktData[bi-fb]) & filter->PktDataMask[bi-fb];
            if ((scratch & ~filter->PktDataNotMask[bi-fb]) != 0)
            {
                rule2fail = TRUE;
                break;
            }
            if (rule3fail == TRUE && (scratch & filter->PktDataNotMask[bi-fb]) != 0)
                rule3fail = FALSE;
        }
        if (rule2fail == FALSE && rule3fail == FALSE)
            return TRUE;
    }

    return FALSE;
}


VOID FilterDeRegister(ULONG channelIndex, VOID *fptr)
{
FILTER *filter = (FILTER *) fptr;

    if (filter->ChannelIndex == channelIndex)
    {
		ChannelDeRegisterFilter(channelIndex, filter->ChannelCallback);
		filter->ChannelCallback = NULL;
	}
}
