/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <memory.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mibsup.h>
#include <message.h>
#include <protocol.h>
#include <channel.h>

#include "filtere.h"
#include "filterc.h"
#include "filter.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define INDEXSIZE   1

static MIB_LOCAL *FilterInstance = NULL;

static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize);

BOOLEAN FilterMInit(VOID)
{
	MessageConfig(FILTER_ERROR, "Filter");
	return TRUE;
}

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize)
{
	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->IdLen == idlen || (*local)->Index > obj->Id[idlen] ||
		(*local = (*local)->Next) != NULL)
	{
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + indexsize;
		return TRUE;
	}

	return FALSE;
}

WORD filterIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(Obj, FilterInstance, IdLen, INDEXSIZE)) == NULL)
		return (SNMP_NOSUCHNAME);

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD filterChannelIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER *data;

	if ((local = MibRmon(Obj, FilterInstance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (FILTER *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, FILTER_BOOLEAN_CHANNEL))
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = data->ChannelIndex;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (FILTER *) local->Data;
		data->ChannelIndex = Obj->Syntax.LngInt;
		BooleanSetTrue(data->ObjectSet, FILTER_BOOLEAN_CHANNEL);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD filterPktDataOffset(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER *data;

	if ((local = MibRmon(Obj, FilterInstance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (FILTER *) local->Data;
		Obj->Syntax.LngInt = data->PktDataOffset;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (FILTER *) local->Data;
		if (Obj->Syntax.LngInt < 0)
			return SNMP_BADVALUE;
		data->PktDataOffset = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD filterPktData(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER *data;
	BYTE *datap, *maskp, *notp;
	WORD i;

	if ((local = MibRmon(Obj, FilterInstance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (FILTER *) local->Data;
		memcpy(Obj->Syntax.BufChr, data->PktData, Obj->SyntaxLen = data->PktDataLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (FILTER *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->SyntaxLen > SNMP_SIZE_BUFCHR)
			return SNMP_BADVALUE;
		if ((datap = DnpapRealloc(data->PktData, Obj->SyntaxLen)) == NULL)
			return SNMP_GENERROR;
		if ((maskp = DnpapRealloc(data->PktDataMask, Obj->SyntaxLen)) == NULL)
		{
			DnpapFree(datap);
			return SNMP_GENERROR;
		}
		if ((notp = DnpapRealloc(data->PktDataNotMask, Obj->SyntaxLen)) == NULL)
		{
			DnpapFree(datap);
			DnpapFree(maskp);
			return SNMP_GENERROR;
		}
		data->PktData = datap;
		data->PktDataMask = maskp;
		data->PktDataNotMask = notp;
		for (i = data->PktDataMaskLen; i < Obj->SyntaxLen; i++)
			data->PktDataMask[i] = 0xff;
		if (Obj->SyntaxLen < data->PktDataMaskLen)
			data->PktDataMaskLen = Obj->SyntaxLen;
		for (i = data->PktDataNotMaskLen; i < Obj->SyntaxLen; i++)
			data->PktDataNotMask[i] = 0x00;
		if (Obj->SyntaxLen < data->PktDataNotMaskLen)
			data->PktDataNotMaskLen = Obj->SyntaxLen;
		memcpy(data->PktData, Obj->Syntax.BufChr, data->PktDataLen = Obj->SyntaxLen);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD filterPktDataMask(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER *data;
	WORD i;

	if ((local = MibRmon(Obj, FilterInstance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (FILTER *) local->Data;
		memcpy(Obj->Syntax.BufChr, data->PktDataMask, Obj->SyntaxLen = data->PktDataMaskLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (FILTER *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->SyntaxLen > SNMP_SIZE_BUFCHR || Obj->SyntaxLen > data->PktDataLen)
			return SNMP_BADVALUE;
		memcpy(data->PktDataMask, Obj->Syntax.BufChr, Obj->SyntaxLen);
		for (i = Obj->SyntaxLen; i < data->PktDataLen; i++)
			data->PktDataMask[i] = 0xff;
		data->PktDataMaskLen = Obj->SyntaxLen;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD filterPktDataNotMask(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER *data;
	WORD i;

	if ((local = MibRmon(Obj, FilterInstance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (FILTER *) local->Data;
		memcpy(Obj->Syntax.BufChr, data->PktDataNotMask, Obj->SyntaxLen = data->PktDataNotMaskLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (FILTER *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->SyntaxLen > SNMP_SIZE_BUFCHR || Obj->SyntaxLen > data->PktDataLen)
			return SNMP_BADVALUE;
		memcpy(data->PktDataNotMask, Obj->Syntax.BufChr, Obj->SyntaxLen);
		for (i = Obj->SyntaxLen; i < data->PktDataLen; i++)
			data->PktDataNotMask[i] = 0x00;
		data->PktDataNotMaskLen = Obj->SyntaxLen;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD filterPktStatus(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER *data;

	if ((local = MibRmon(Obj, FilterInstance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (FILTER *) local->Data;
		if (data->PktStatus == -1)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = data->PktStatus;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (FILTER *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->Syntax.LngInt < 0 || Obj->Syntax.LngInt > FILTER_STATUS_MAX)
			return SNMP_BADVALUE;
		data->PktStatus = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD filterPktStatusMask(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER *data;

	if ((local = MibRmon(Obj, FilterInstance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (FILTER *) local->Data;
		if (data->PktStatusMask == -1)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = data->PktStatusMask;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (FILTER *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->Syntax.LngInt < 0 || Obj->Syntax.LngInt > FILTER_STATUS_MAX)
			return SNMP_BADVALUE;
		data->PktStatusMask = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD filterPktStatusNotMask(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER *data;

	if ((local = MibRmon(Obj, FilterInstance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (FILTER *) local->Data;
		if (data->PktStatusNotMask == -1)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = data->PktStatusNotMask;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (FILTER *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->Syntax.LngInt < 0 || Obj->Syntax.LngInt > FILTER_STATUS_MAX)
			return SNMP_BADVALUE;
		data->PktStatusNotMask = Obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD filterOwner(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER *data;

	if ((local = MibRmon(Obj, FilterInstance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (FILTER *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, FILTER_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		Obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (FILTER *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, Obj->Syntax.BufChr, Obj->SyntaxLen);
		data->OwnerLen = Obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, FILTER_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD filterStatus(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER *data;

	local = MibRmon(Obj, FilterInstance, IdLen, INDEXSIZE);

	switch (Obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (FILTER *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (FILTER *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(Obj, &FilterInstance, IdLen, INDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(FILTER))) == NULL)
				{
					MibRemove(Obj, &FilterInstance, IdLen, INDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (FILTER *) local->Data;
				memset(data, 0, sizeof(FILTER));
				if (FilterCInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, FILTER_CREATE, "filter: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(Obj, &FilterInstance, IdLen, INDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (FILTER *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (FilterCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, FILTER_ACTIVE, "filter: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (FilterCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, FILTER_DESTROY, "filter: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &FilterInstance, IdLen, INDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_INVALID:
				if (FilterCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, FILTER_DESTROY, "filter: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &FilterInstance, IdLen, INDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (FilterCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			case SNMP_VALID:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

BOOLEAN FilterRegisterAll(ULONG channelIndex)
{
	MIB_LOCAL *local;
	FILTER *filter;

	for (local = FilterInstance; local != NULL; local = local->Next)
	{
		filter = (FILTER *) local->Data;
		if (filter->ChannelIndex == channelIndex && filter->ChannelCallback == NULL)
		{
			filter->ChannelCallback = ChannelRegisterFilter(filter->ChannelIndex, FilterPkt, FilterDeRegister, filter);
		}
	}

	return TRUE;
}
