/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <dnpap.h>
#include <config.h>
#include <maxmem.h>
#include <message.h>
#include <mac.h>
#include <snmp.h>
#include <mibrqs.h>
#include <mibsup.h>
#include <dmath.h>

#include "hist_e.h"
#include "hist_c.h"
#include "hist.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define CONTROLINDEXSIZE        1
#define HISTORYINDEXSIZE        2

extern LONG HistoryMaxNrBuckets;

static MIB_LOCAL *HistoryControlInstance = NULL;

static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, ETHER_HISTORY ** elem);

BOOLEAN HistoryMInit(VOID)
{
	LONG i, nriface, j, nrcoll, n;
	CHAR s[81];
	BOOLEAN v;
	LONG buckets, interval;

	MAC_IFACE *iface;

	LONG histsource[] = {1, 3, 6, 1, 2, 1, 2, 2, 1, 1, 1};
	CHAR histowner[] = "monitorHistory";

#define STATUSINDEX     11
	SNMP_OBJECT histobj[] =
	{
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 2, 1, 1, 7, 1}, 12, SNMP_INTEGER},
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 2, 1, 1, 6, 1}, 12, SNMP_DISPLAYSTR},
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 2, 1, 1, 2, 1}, 12, SNMP_OBJECTID},
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 2, 1, 1, 3, 1}, 12, SNMP_INTEGER},
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 2, 1, 1, 5, 1}, 12, SNMP_INTEGER}
	};
	WORD dummy;

	MessageConfig(HISTORY_ERROR, "History");
	if (HistoryConfigInit() == FALSE)
		return FALSE;

	n = 1;

	nriface = (LONG) MacIfaceCount();
	for (i = 1; i <= nriface; i++)
	{
		sprintf(s, "beholder.history.iface.%ld", i);
		if (ConfigGetBoolean(s, &v) == TRUE)
		{
			iface = MacIfaceGet((WORD) i);
			if (iface != NULL && v == TRUE &&
				(iface->type == MAC_TYPE_ETHERNET_CSMACD ||
				 iface->type == MAC_TYPE_88023_CSMACD))
			{
				nrcoll = 2;
				sprintf(s, "beholder.history.collectors.%ld", i);
				ConfigGetLong(s, &nrcoll);
				for (j = 1; j <= nrcoll; j++)
				{
					buckets = 50;
					if (j % 2 == 1)
						interval = 30;
					else
						interval = 1800;
					sprintf(s, "beholder.history.buckets.%ld.%ld", i, j);
					ConfigGetLong(s, &buckets);
					sprintf(s, "beholder.history.interval.%ld.%ld", i, j);
					ConfigGetLong(s, &interval);

					histobj[0].Id[STATUSINDEX] = n;
					histobj[1].Id[STATUSINDEX] = n;
					histobj[2].Id[STATUSINDEX] = n;
					histobj[3].Id[STATUSINDEX] = n;
					histobj[4].Id[STATUSINDEX] = n;

					histobj[0].Request = SNMP_PDU_SET;
					histobj[0].Syntax.LngInt = SNMP_INVALID;
					MibRequest(1, &histobj[0], &dummy);
					histobj[0].Request = SNMP_PDU_SET;
					histobj[0].Syntax.LngInt = SNMP_CREATEREQUEST;
					if (MibRequest(1, &histobj[0], &dummy) != SNMP_NOERROR)
					{
						DnpapMessage(DMC_WARNING, HISTORY_HISTINIT, "history: couldn't create history collector %ld (%ld,%ld)", n, i, j);
						continue;
					}
					histobj[1].Request = SNMP_PDU_SET;
					strncpy(histobj[1].Syntax.BufChr, histowner, histobj[1].SyntaxLen = strlen(histowner));
					histobj[2].Request = SNMP_PDU_SET;
					memcpy(histobj[2].Syntax.BufInt, histsource, sizeof(histsource));
					histobj[2].Syntax.BufInt[sizeof(histsource) / sizeof(histsource[0]) - 1] = i;
					histobj[2].SyntaxLen = sizeof(histsource) / sizeof(histsource[0]);
					histobj[3].Request = SNMP_PDU_SET;
					histobj[3].Syntax.LngInt = buckets;
					histobj[4].Request = SNMP_PDU_SET;
					histobj[4].Syntax.LngInt = interval;
					if (MibRequest(4, &histobj[1], &dummy) != SNMP_NOERROR)
					{
						DnpapMessage(DMC_WARNING, HISTORY_HISTINIT, "history: couldn't configure history collector %ld (%ld,%ld)", n, i, j);
						histobj[0].Request = SNMP_PDU_SET;
						histobj[0].Syntax.LngInt = SNMP_INVALID;
						MibRequest(1, &histobj[0], &dummy);
						continue;
					}
					histobj[0].Request = SNMP_PDU_SET;
					histobj[0].Syntax.LngInt = SNMP_VALID;
					if (MibRequest(1, &histobj[0], &dummy) != SNMP_NOERROR)
					{
						DnpapMessage(DMC_WARNING, HISTORY_HISTINIT, "history: couldn't set history collector %ld (%ld,%ld) to valid", n, i, j);
						histobj[0].Request = SNMP_PDU_SET;
						histobj[0].Syntax.LngInt = SNMP_INVALID;
						MibRequest(1, &histobj[0], &dummy);
						continue;
					}

					n++;
				}
			}
		}
	}

	return TRUE;
}

#define CONTROL_DATA    ((HISTORY_CONTROL*)(*local)->Data)

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, ETHER_HISTORY ** elem)
{
	HISTORY_CONTROL *data;
	BOOLEAN first = FALSE;
	LONG index;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case HISTORYINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			data = CONTROL_DATA;
			if ((index = obj->Id[idlen + 1] - data->BucketsDeleted) >= 1 &&
				((data->BucketsWrapped == TRUE && index < data->BucketsGranted) ||
				 (data->BucketsWrapped == FALSE && index < data->BucketsPos)))
			{							/* BucketsWrapped == FALSE and BucketsPos == 0 if Status != VALID  */
				obj->Id[idlen] = (*local)->Index;
				obj->Id[idlen + 1]++;
				obj->IdLen = idlen + HISTORYINDEXSIZE;
				*elem = EtherHistorySearch(CONTROL_DATA, obj, idlen);
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL &&
			   (CONTROL_DATA->BucketsGranted == 0 ||
				(CONTROL_DATA->BucketsWrapped == FALSE &&
				 CONTROL_DATA->BucketsPos == 0)))
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen + 1] = CONTROL_DATA->BucketsDeleted + 1;
		obj->IdLen = idlen + HISTORYINDEXSIZE;
		*elem = EtherHistorySearch(CONTROL_DATA, obj, idlen);
		break;
	case CONTROLINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + CONTROLINDEXSIZE;
		break;
	}
	return TRUE;
}

WORD HistoryControlIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HistoryControlDataSource(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HISTORY_CONTROL *data;
	MAC_IFACE *iface;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (HISTORY_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, HISTORY_BOOLEAN_DATASOURCE))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufInt, data->Source, data->SourceLen * sizeof(data->Source[0]));
		obj->SyntaxLen = data->SourceLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (HISTORY_CONTROL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (data->SourceLen != obj->SyntaxLen ||
			memcmp(data->Source, obj->Syntax.BufInt, (data->SourceLen - 1) * sizeof(data->Source[0])) != 0)
			return SNMP_BADVALUE;
		if ((iface = MacIfaceGet((WORD) obj->Syntax.BufInt[obj->SyntaxLen - 1])) == NULL)
			return SNMP_BADVALUE;
		if (iface->type != MAC_TYPE_ETHERNET_CSMACD && iface->type != MAC_TYPE_88023_CSMACD)
			return SNMP_BADVALUE;
		memcpy(data->Source, obj->Syntax.BufInt, obj->SyntaxLen * sizeof(data->Source[0]));
		data->SourceLen = obj->SyntaxLen;
		data->Iface = iface;
		BooleanSetTrue(data->ObjectSet, HISTORY_BOOLEAN_DATASOURCE);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HistoryControlBucketsRequested(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HISTORY_CONTROL *data;
	ETHER_HISTORY *etherentries;
	LONG i, ni, j;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (HISTORY_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->BucketsRequested;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (HISTORY_CONTROL *) local->Data;
		if (obj->Syntax.LngInt < 1)
			return SNMP_BADVALUE;
		data->BucketsGranted = LongMin(MemoryGetAvail(), MemoryGetMaxChunk()) / sizeof(ETHER_HISTORY);
		if (data->BucketsGranted > LongMin(obj->Syntax.LngInt, HistoryMaxNrBuckets))
			data->BucketsGranted = LongMin(obj->Syntax.LngInt, HistoryMaxNrBuckets);
		if (data->BucketsGranted == data->BucketsAllocated)
		{
			data->BucketsRequested = obj->Syntax.LngInt;
			return SNMP_NOERROR;
		}
		if ((etherentries = NewEtherHistory(data->BucketsGranted)) == NULL)
		{
			DelEtherHistory(data->EtherBuckets);
			data->EtherBuckets = NULL;
			data->BucketsPos = 0;
			data->BucketsGranted = 0;
			data->BucketsAllocated = 0;
			data->BucketsWrapped = FALSE;
		}
		else
		{
			if (data->BucketsWrapped == TRUE)
				ni = LongMin(data->BucketsGranted, data->BucketsAllocated);
			else
				ni = LongMin(data->BucketsGranted, data->BucketsPos);
			for (i = 0; i < ni; i++)
			{
				j = data->BucketsPos - ni + i;
				if (j < 0)
					j = j + data->BucketsAllocated;
				memcpy(etherentries + i, data->EtherBuckets + j, sizeof(ETHER_HISTORY));
			}
			DelEtherHistory(data->EtherBuckets);
			data->EtherBuckets = etherentries;
			data->BucketsPos = ni;
			data->BucketsAllocated = data->BucketsGranted;
			data->BucketsWrapped = FALSE;
		}
		data->BucketsRequested = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HistoryControlBucketsGranted(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HISTORY_CONTROL *data;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (HISTORY_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->BucketsGranted;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HistoryControlInterval(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HISTORY_CONTROL *data;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (HISTORY_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Interval;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (HISTORY_CONTROL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt < 1)
			return SNMP_BADVALUE;
		data->Interval = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HistoryControlOwner(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HISTORY_CONTROL *data;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (HISTORY_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, HISTORY_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (HISTORY_CONTROL *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, obj->Syntax.BufChr, obj->SyntaxLen);
		data->OwnerLen = obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, HISTORY_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HistoryControlStatus(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HISTORY_CONTROL *data;

	local = MibRmon(obj, HistoryControlInstance, idlen, CONTROLINDEXSIZE);

	switch (obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (HISTORY_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (HISTORY_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(obj, &HistoryControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(HISTORY_CONTROL))) == NULL)
				{
					MibRemove(obj, &HistoryControlInstance, idlen, CONTROLINDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (HISTORY_CONTROL *) local->Data;
				memset(data, 0, sizeof(HISTORY_CONTROL));
				if (HistoryCInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, HISTORY_CREATE, "history: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(obj, &HistoryControlInstance, idlen, CONTROLINDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (HISTORY_CONTROL *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (HistoryCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, HISTORY_ACTIVE, "history: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (HistoryCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, HISTORY_DESTROY, "history: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &HistoryControlInstance, idlen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				return SNMP_NOERROR;
			case SNMP_INVALID:
				if (HistoryCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, HISTORY_DESTROY, "history: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &HistoryControlInstance, idlen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (HistoryCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistorySampleIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = obj->Id[idlen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryIntervalStart(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->IntervalStart;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryDropEvents(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->DropEvents;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryOctets(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryPkts(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryBroadcastPkts(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->BroadcastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryMulticastPkts(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->MulticastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryCRCAlignErrors(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->CRCAlignErrors;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryUndersizePkts(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->UndersizePkts;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryOversizePkts(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->OversizePkts;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryFragments(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Fragments;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryJabbers(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Jabbers;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryCollisions(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Collisions;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherHistoryUtilization(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_HISTORY *data = NULL;

	if ((local = MibRmon(obj, HistoryControlInstance, idlen, HISTORYINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, HISTORYINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = EtherHistorySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Utilization;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
