/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */

#include <stdlib.h>
#include <memory.h>
#include <dnpap.h>
#include <config.h>
#include <message.h>
#include <mac.h>
#include <hash.h>
#include <sysmib.h>
#include <protocol.h>
#include <channel.h>
#include <topn.h>

#include "hoste.h"
#include "hostc.h"

IMPORT BOOLEAN BeholderStrictRMON;

static LONG HostMaxNrHosts = 2000;

static BOOLEAN HostCallback(MAC_COLL * collector, PROT_PKT * pkt);
static VOID HostHandlePkt(HOST_CONTROL * hostcontrol, int type, PROT_PKT * pkt);

static VOID HostNotify(HOST_CONTROL * hostcontrol, int event, HOST *host, WORD addrlen, int flags, LWORD size);

static BOOLEAN HostAddLexList(HOST_CONTROL * hostcontrol, HOST * host);
static BOOLEAN HostRemoveLexList(HOST_CONTROL * hostcontrol, HOST * host);

static BOOLEAN HostAddTimeList(HOST_CONTROL * hostcontrol, HOST * host);
static BOOLEAN HostRemoveTimeList(HOST_CONTROL * hostcontrol, HOST * host);

static HOST *HostAddLRUList(HOST_CONTROL * hostcontrol, HOST * host);
static BOOLEAN HostUpdateLRUList(HOST_CONTROL * hostcontrol, HOST * host);
static BOOLEAN HostRemoveLRUList(HOST_CONTROL * hostcontrol, HOST * host);

static BOOLEAN HostUpdateTimeTable(HOST_CONTROL * hostcontrol);

static VOID DelHosts(HOST_CONTROL * hostcontrol);

static int addrcmp(HOST * host1, HOST * host2, WORD addrlen, WORD len);

#ifdef DEBUG
static VOID HostCheckListsConsistency(HOST_CONTROL * host);
static VOID HostCheckListsRemove(HOST_CONTROL * hostcontrol, HOST * host);
static VOID HostCheckListsRemove2(HOST_CONTROL * hostcontrol, HOST * host);
static VOID HostCheckListsAdd(HOST_CONTROL * hostcontrol, HOST * host);
static VOID HostCheckListsAdd2(HOST_CONTROL * hostcontrol, HOST * host);

#endif

BOOLEAN HostConfigInit(VOID)
{
	ConfigGetLong("beholder.host.maxnrhosts", &HostMaxNrHosts);

	if (HostMaxNrHosts < 2)
	{
		DnpapMessage(DMC_WARNING, HOST_MAX, "hostcontrol: beholder.host.maxnrhosts < 2, setting it to 2");
		HostMaxNrHosts = 2;
	}

	return TRUE;
}

BOOLEAN HostCInit(HOST_CONTROL * hostcontrol)
{
	LONG source[] =
	{1, 3, 6, 1, 2, 1, 2, 2, 1, 1, 1};

	hostcontrol->Index = 0;
	memcpy(hostcontrol->Source, source, sizeof(source));
	hostcontrol->SourceLen = sizeof(source) / sizeof(source[0]);
	hostcontrol->Channel = FALSE;
    hostcontrol->ChannelCallback = NULL;
	hostcontrol->AddressLen = 0;
	hostcontrol->TableSize = 0;
	hostcontrol->LastDeleteTime = 0;
	hostcontrol->Owner[0] = '\0';
	hostcontrol->OwnerLen = 0;
	hostcontrol->Status = SNMP_INVALID;

	if ((hostcontrol->Iface =
		 MacIfaceGet((WORD) hostcontrol->Source[hostcontrol->SourceLen - 1])) == NULL)
	{
		DnpapMessage(DMC_ERROR, HOST_NETINIT, "hostcontrol: network initialisation failed");
		return (FALSE);
	}

	hostcontrol->Table = NULL;

	hostcontrol->LexList = NULL;

	hostcontrol->TimeList = NULL;
	hostcontrol->TimeLast = NULL;
	hostcontrol->TimeTable = NULL;
	hostcontrol->TimeListUpdated = FALSE;

	hostcontrol->LRUList = NULL;
	hostcontrol->LRULast = NULL;

    hostcontrol->TopNs = NULL;
    
	BooleanSetAllTrue(hostcontrol->ObjectSet);
	BooleanSetFalse(hostcontrol->ObjectSet, HOST_BOOLEAN_DATASOURCE);
	BooleanSetFalse(hostcontrol->ObjectSet, HOST_BOOLEAN_OWNER);

	return TRUE;
}

BOOLEAN HostCStart(HOST_CONTROL * hostcontrol)
{
	if (BeholderStrictRMON && !BooleanCheckAllTrue(hostcontrol->ObjectSet))
		return FALSE;

	if ((hostcontrol->Table = NewHash(HostMaxNrHosts / 2 + 11, NULL)) == NULL)
	{
		DnpapMessage(DMC_ERROR, HOST_HASHERR, "hostcontrol: can not create hashtable");
		return FALSE;
	}

	if ((hostcontrol->TimeTable = (HOST **) DnpapMalloc(HostMaxNrHosts * sizeof(HOST *))) == NULL)
	{
		DnpapMessage(DMC_ERROR, HOST_TIMER, "hostcontrol: can not create time ordered table");
		return FALSE;
	}

	if (hostcontrol->Channel == TRUE)
    {   
        hostcontrol->ChannelCallback = ChannelRegisterSink(hostcontrol->Source[hostcontrol->SourceLen - 1],
                                                           HostChannelPkt, HostDeRegister, hostcontrol);
    }
    else
	{
		hostcontrol->Coll.Rcve = HostCallback;
		hostcontrol->Coll.specific = hostcontrol;
		if (!MacCollRegister(&(hostcontrol->Coll)))
		{
			DnpapMessage(DMC_ERROR, HOST_NETERR, "hostcontrol: network initialisation failed");
			return FALSE;
		}
	}

	HostTopNRegisterAll(hostcontrol->Index);

	hostcontrol->AddressLen = 0;
	
	return TRUE;
}

BOOLEAN HostCStop(HOST_CONTROL * hostcontrol)
{
	HostNotify(hostcontrol, HOST_EVENT_STOP, NULL, 0, 0, 0);
	if (hostcontrol->TopNs != NULL)
		DnpapMessage(DMC_FATAL, HOST_TOPNS, "host: internal error, not all topNs were removed");
	hostcontrol->TopNs = NULL;

    if (hostcontrol->Channel == TRUE)
	{
        ChannelDeRegisterSink(hostcontrol->Source[hostcontrol->SourceLen - 1],
                              hostcontrol->ChannelCallback);
		hostcontrol->ChannelCallback = NULL;
	}
    else
    	MacCollRemove(&(hostcontrol->Coll));

	DelHash(hostcontrol->Table);
	hostcontrol->Table = NULL;

	DelHosts(hostcontrol);

	DnpapFree(hostcontrol->TimeTable);
	hostcontrol->TimeTable = NULL;

	hostcontrol->LexList = NULL;

	hostcontrol->TimeList = NULL;
	hostcontrol->TimeLast = NULL;
	hostcontrol->TimeListUpdated = FALSE;

	hostcontrol->LRUList = NULL;
	hostcontrol->LRULast = NULL;

	hostcontrol->TableSize = 0;

	hostcontrol->LastDeleteTime = 0;

	return TRUE;
}

BOOLEAN HostCallback(MAC_COLL * collector, PROT_PKT * pkt)
{
	HOST_CONTROL *hostcontrol = collector->specific;
	PROT_OBJ Interface =
	{0,
	 {1, 2}};

	if (hostcontrol->Status == SNMP_VALID &&
        ProtGetField(pkt, &Interface) == TRUE &&
		Interface.Syntax.LngInt == hostcontrol->Source[hostcontrol->SourceLen - 1])
		HostHandlePkt(hostcontrol, hostcontrol->Iface->type, pkt);

	return TRUE;
}

VOID HostChannelPkt(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *sink)
{
	HOST_CONTROL *hostcontrol = (HOST_CONTROL *) sink;

	if (hostcontrol->Status == SNMP_VALID && hostcontrol->Channel == TRUE &&
		hostcontrol->Source[hostcontrol->SourceLen - 1] == channelIndex)
		HostHandlePkt(hostcontrol, type, pkt);
}

VOID HostHandlePkt(HOST_CONTROL * hostcontrol, int type, PROT_PKT * pkt)
{
	HOST *host = NULL, *oldhost = NULL;
	BYTE EthBroadcast[ETH_SIZE_ADDR] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	PROT_OBJ EthSize =	{0, {1, 4}};
	PROT_OBJ EthDst =	{1, {2, 1}};
	PROT_OBJ EthSrc =	{1, {2, 2}};
	BYTE *src, *dst;
	WORD addrlen;
	LWORD size;
	int flags;

	src = NULL;
	dst = NULL;

	switch (type)
	{
	case MAC_TYPE_ETHERNET_CSMACD:
	case MAC_TYPE_88023_CSMACD:
		if (ProtGetField(pkt, &EthSize) == FALSE)
			return;
		size = EthSize.Syntax.LngUns + 4L;
		if (ProtGetField(pkt, &EthSrc) == FALSE)
			return;
		src = EthSrc.Syntax.BufChr;
		if (ProtGetField(pkt, &EthDst) == FALSE)
			return;
		dst = EthDst.Syntax.BufChr;
		if (EthSrc.SyntaxLen != ETH_SIZE_ADDR)
			src = NULL;
		if (EthDst.SyntaxLen != ETH_SIZE_ADDR)
			dst = NULL;
		addrlen = ETH_SIZE_ADDR;
		if (hostcontrol->AddressLen == 0)
			hostcontrol->AddressLen = addrlen;
		else if (hostcontrol->AddressLen != addrlen)
		{
			DnpapMessage(DMC_FATAL, HOST_INCONSISTENT, "host: inconsistent address length encountered");
			DnpapExit(1);
		}
		break;
	default:
		return;
	}
	
	flags = 0;
	
	if ((host = HashSearch(hostcontrol->Table, src, addrlen)) == NULL)
	{
		/* first try to add the new host  */
		if ((host = DnpapMalloc(sizeof(HOST))) != NULL)
		{
			memset(host, 0, sizeof(HOST));
			memcpy(host->Address, src, addrlen);
#ifdef DEBUG
			HostCheckListsAdd(hostcontrol, host);
#endif
			if (HashAdd(hostcontrol->Table, host->Address, addrlen, host) == NULL)
			{
				DnpapMessage(DMC_WARNING, HOST_NADD, "hostcontrol: host could not be added to the hash table");
				DnpapFree(host);
				host = NULL;
			}
			else
			{
#ifdef DEBUG
				if (HashSearch(hostcontrol->Table, src, addrlen) == NULL)
					DnpapMessage(DMC_FATAL, HOST_ERROR, "host that was just added can not be retrieved");
#endif
				HostAddLexList(hostcontrol, host);
				HostAddTimeList(hostcontrol, host);
				oldhost = HostAddLRUList(hostcontrol, host);
#ifdef DEBUG
				HostCheckListsAdd2(hostcontrol, host);
#endif
				hostcontrol->TableSize++;

				if (hostcontrol->TableSize > (LONG) HostMaxNrHosts)
				{
#ifdef DEBUG
					HostCheckListsRemove2(hostcontrol, oldhost);
#endif
					HostRemoveLRUList(hostcontrol, oldhost);
					HostRemoveTimeList(hostcontrol, oldhost);
					HostRemoveLexList(hostcontrol, oldhost);
					HashRemove(hostcontrol->Table, oldhost->Address, addrlen);
#ifdef DEBUG
					HostCheckListsRemove(hostcontrol, oldhost);
					memset(oldhost->Address, 0, HOST_SIZE_ADDR);
#endif
					HostNotify(hostcontrol, HOST_EVENT_RESET, oldhost, hostcontrol->AddressLen, 0, 0);
					DnpapFree(oldhost);
					hostcontrol->LastDeleteTime = SysTime();
					hostcontrol->TableSize--;
#ifdef DEBUG
					if (HashSearch(hostcontrol->Table, src, addrlen) == NULL)
						DnpapMessage(DMC_FATAL, HOST_ERROR, "host that was just added can not be retrieved");
#endif
				}
			}
		}
	}

	if (host != NULL)
	{
		flags = HOST_FLAG_OUT;
		
		host->OutPkts++;
		host->OutOctets += size;
		switch (type)
		{
		case MAC_TYPE_ETHERNET_CSMACD:
		case MAC_TYPE_88023_CSMACD:
			if (!memcmp(dst, &EthBroadcast, addrlen))
			{
				flags |= HOST_FLAG_BROADCAST;
				host->OutBroadcastPkts++;
			}
			else if (dst[0] & 0x01)
			{
				flags |= HOST_FLAG_MULTICAST;
				host->OutMulticastPkts++;
			}
			break;
		default:
			break;
		}

		HostUpdateLRUList(hostcontrol, host);
		HostNotify(hostcontrol, HOST_EVENT_UPDATE, host, hostcontrol->AddressLen, flags, size);
	}

#ifdef DEBUG
	HostCheckListsConsistency(hostcontrol);
#endif

	flags = 0;
	
	if ((host = HashSearch(hostcontrol->Table, dst, addrlen)) == NULL)
	{
		if ((host = DnpapMalloc(sizeof(HOST))) != NULL)
		{
			memset(host, 0, sizeof(HOST));
			memcpy(host->Address, dst, addrlen);
#ifdef DEBUG
			HostCheckListsAdd(hostcontrol, host);
#endif
			if (HashAdd(hostcontrol->Table, host->Address, addrlen, host) == NULL)
			{
				DnpapMessage(DMC_WARNING, HOST_NADD, "hostcontrol: host could not be added to the hash table");
				DnpapFree(host);
				host = NULL;
			}
			else
			{
#ifdef DEBUG
				if (HashSearch(hostcontrol->Table, dst, addrlen) == NULL)
					DnpapMessage(DMC_FATAL, HOST_ERROR, "host that was just added can not be retrieved");
#endif
				HostAddLexList(hostcontrol, host);
				HostAddTimeList(hostcontrol, host);
				oldhost = HostAddLRUList(hostcontrol, host);
#ifdef DEBUG
				HostCheckListsAdd2(hostcontrol, host);
#endif
				hostcontrol->TableSize++;

				if (hostcontrol->TableSize > (LONG) HostMaxNrHosts)
				{
#ifdef DEBUG
					HostCheckListsRemove2(hostcontrol, oldhost);
#endif
					HostRemoveLRUList(hostcontrol, oldhost);
					HostRemoveTimeList(hostcontrol, oldhost);
					HostRemoveLexList(hostcontrol, oldhost);
					HashRemove(hostcontrol->Table, oldhost->Address, addrlen);
#ifdef DEBUG
					HostCheckListsRemove(hostcontrol, oldhost);
					memset(oldhost->Address, 0, HOST_SIZE_ADDR);
#endif
					HostNotify(hostcontrol, HOST_EVENT_RESET, oldhost, hostcontrol->AddressLen, 0, 0);
					DnpapFree(oldhost);
					hostcontrol->LastDeleteTime = SysTime();
					hostcontrol->TableSize--;
#ifdef DEBUG
					if (HashSearch(hostcontrol->Table, dst, addrlen) == NULL)
						DnpapMessage(DMC_FATAL, HOST_ERROR, "host that was just added can not be retrieved");
#endif
				}
			}
		}
	}

	if (host != NULL)
	{
		flags = HOST_FLAG_IN;
		
		host->InPkts++;
		host->InOctets += size;

		HostUpdateLRUList(hostcontrol, host);
		HostNotify(hostcontrol, HOST_EVENT_UPDATE, host, hostcontrol->AddressLen, flags, size);
	}

#ifdef DEBUG
	HostCheckListsConsistency(hostcontrol);
#endif

	return;
}

VOID HostDeRegister(ULONG channelIndex, VOID *sink)
{
HOST_CONTROL *data = (HOST_CONTROL *) sink;

	if (data->Channel == TRUE && data->Source[data->SourceLen - 1] == channelIndex)
	{
        ChannelDeRegisterSink(data->Source[data->SourceLen - 1],
                              data->ChannelCallback);
		data->ChannelCallback = NULL;
	}
}

VOID HostNotify(HOST_CONTROL * hostcontrol, int event, HOST *host, WORD addrlen, int flags, LWORD size)
{
TopNCallback *topn;

	switch (event)
	{
	case HOST_EVENT_RESET:
	case HOST_EVENT_UPDATE:
		for (topn = hostcontrol->TopNs; topn != NULL; topn = topn->Next)
    		topn->Notify(hostcontrol->Index, event, host, addrlen, flags, size, topn->TopN);
    	break;
	case HOST_EVENT_STOP:		/* TopN removes its callback when called */
		while ((topn = hostcontrol->TopNs) != NULL)
    		topn->Notify(hostcontrol->Index, event, host, addrlen, 0, 0, topn->TopN);
    	break;
   	}
}

TopNCallback *HostAddTopN(HOST_CONTROL *hostcontrol, HostTopNCallback notify, VOID *topn)
{
TopNCallback *callback;

    if ((callback = DnpapMalloc(sizeof(TopNCallback))) == NULL)
        return NULL;
    callback->Notify = notify;
    callback->TopN = topn;
    callback->Prev = NULL;
    callback->Next = hostcontrol->TopNs;
    if (hostcontrol->TopNs != NULL)
        hostcontrol->TopNs->Prev = callback;
    hostcontrol->TopNs = callback;
    
    return callback;
}

BOOLEAN HostRemoveTopN(HOST_CONTROL *hostcontrol, TopNCallback *callback)
{
    if (callback == NULL)
        return TRUE;
    if (callback->Next != NULL)
        callback->Next->Prev = callback->Prev;
    if (callback->Prev == NULL)
        hostcontrol->TopNs = callback->Next;
    else
        callback->Prev->Next = callback->Next;
        
    DnpapFree(callback);
    
    return TRUE;
}

BOOLEAN HostAddLexList(HOST_CONTROL * hostcontrol, HOST * host)
{
	HOST *p, *q;
	WORD addrlen = hostcontrol->AddressLen;

	if (hostcontrol->LexList == NULL)
	{
		hostcontrol->LexList = host;
		host->LexPrev = NULL;
		host->LexNext = NULL;
		return TRUE;
	}

	for (p = hostcontrol->LexList, q = NULL; p != NULL; q = p, p = p->LexNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) >= 0)
			break;
	}

	if (q != NULL)
		q->LexNext = host;
	else
		hostcontrol->LexList = host;

	if (p != NULL)
		p->LexPrev = host;

	host->LexPrev = q;
	host->LexNext = p;

	return TRUE;
}

BOOLEAN HostAddTimeList(HOST_CONTROL * hostcontrol, HOST * host)
{
	if (hostcontrol->TimeList == NULL)
	{
		hostcontrol->TimeList = host;
		hostcontrol->TimeLast = host;
		host->TimePrev = NULL;
		host->TimeNext = NULL;

		hostcontrol->TimeListUpdated = TRUE;

		return TRUE;
	}

	host->TimePrev = hostcontrol->TimeLast;
	host->TimeNext = NULL;
	hostcontrol->TimeLast->TimeNext = host;
	hostcontrol->TimeLast = host;

	hostcontrol->TimeListUpdated = TRUE;

	return TRUE;
}

HOST *HostAddLRUList(HOST_CONTROL * hostcontrol, HOST * host)
{
	if (hostcontrol->LRUList == NULL)
	{
		hostcontrol->LRUList = host;
		hostcontrol->LRULast = host;
		host->LRUPrev = NULL;
		host->LRUNext = NULL;

		return NULL;
	}

	host->LRUNext = hostcontrol->LRUList;
	host->LRUPrev = NULL;
	hostcontrol->LRUList->LRUPrev = host;
	hostcontrol->LRUList = host;

	return hostcontrol->LRULast;
}

BOOLEAN HostUpdateLRUList(HOST_CONTROL * hostcontrol, HOST * host)
{
	if (hostcontrol->TableSize > 1 && hostcontrol->LRUList != host)
	{
		HostRemoveLRUList(hostcontrol, host);
		HostAddLRUList(hostcontrol, host);
	}
	return TRUE;
}

BOOLEAN HostRemoveLexList(HOST_CONTROL * hostcontrol, HOST * host)
{
	if (host->LexPrev != NULL)
		host->LexPrev->LexNext = host->LexNext;
	else
		hostcontrol->LexList = host->LexNext;
	if (host->LexNext != NULL)
		host->LexNext->LexPrev = host->LexPrev;

	host->LexPrev = NULL;
	host->LexNext = NULL;

	return TRUE;
}

BOOLEAN HostRemoveTimeList(HOST_CONTROL * hostcontrol, HOST * host)
{
	if (host->TimePrev != NULL)
		host->TimePrev->TimeNext = host->TimeNext;
	else
		hostcontrol->TimeList = host->TimeNext;
	if (host->TimeNext != NULL)
		host->TimeNext->TimePrev = host->TimePrev;
	else
		hostcontrol->TimeLast = host->TimePrev;

	hostcontrol->TimeListUpdated = TRUE;

	host->TimePrev = NULL;
	host->TimeNext = NULL;

	return TRUE;
}

BOOLEAN HostRemoveLRUList(HOST_CONTROL * hostcontrol, HOST * host)
{
	if (host->LRUPrev != NULL)
		host->LRUPrev->LRUNext = host->LRUNext;
	else
		hostcontrol->LRUList = host->LRUNext;
	if (host->LRUNext != NULL)
		host->LRUNext->LRUPrev = host->LRUPrev;
	else
		hostcontrol->LRULast = host->LRUPrev;

	host->LRUPrev = NULL;
	host->LRUNext = NULL;

	return TRUE;
}

VOID DelHosts(HOST_CONTROL * hostcontrol)
{
	HOST *host1, *host2;

	for (host2 = hostcontrol->LexList; host2 != NULL; host2 = host1)
	{
		host1 = host2->LexNext;
		DnpapFree(host2);
		hostcontrol->TableSize--;
	}
	hostcontrol->LexList = NULL;
}

HOST *HostSearchNext(HOST_CONTROL * hostcontrol, HOST * host, WORD len)
{
	HOST *p;
	WORD addrlen = hostcontrol->AddressLen;

	if (len == addrlen &&
		(p = HashSearch(hostcontrol->Table, host->Address, len)) != NULL)
		return p->LexNext;
	else
	{
		for (p = hostcontrol->LexList; p != NULL; p = p->LexNext)
		{
			if (addrcmp(p, host, addrlen, len) >= 0)
				break;
		}
		return p;
	}
	return NULL;
}

HOST *HostSearchHash(HOST_CONTROL * hostcontrol, SNMP_OBJECT * obj, WORD idlen)
{
	WORD i, addrlen = hostcontrol->AddressLen;
	BYTE address[HOST_SIZE_ADDR];

	if (obj->Id[idlen + 1] != addrlen)
		return NULL;
	for (i = 0; i < addrlen; i++)
		address[i] = (BYTE) obj->Id[idlen + 2 + i];
	return HashSearch(hostcontrol->Table, address, addrlen);
}

HOST *HostTimeSearch(HOST_CONTROL * hostcontrol, SNMP_OBJECT * obj, WORD idlen)
{
	if (obj->Id[idlen + 1] < 1 || obj->Id[idlen + 1] > hostcontrol->TableSize)
		return NULL;
	if (hostcontrol->TimeListUpdated == TRUE)
		HostUpdateTimeTable(hostcontrol);
	return hostcontrol->TimeTable[obj->Id[idlen + 1] - 1];
}

LONG HostSearchCreationOrder(HOST_CONTROL * hostcontrol, HOST * host)
{
	if (hostcontrol->TimeListUpdated == TRUE)
		HostUpdateTimeTable(hostcontrol);
	return host->TimeListPtr - hostcontrol->TimeTable + 1;
}

BOOLEAN HostUpdateTimeTable(HOST_CONTROL * hostcontrol)
{
	LONG i;
	HOST *p;

	for (p = hostcontrol->TimeList, i = 0; p != NULL; p = p->TimeNext, i++)
	{
		hostcontrol->TimeTable[i] = p;
		p->TimeListPtr = hostcontrol->TimeTable + i;
	}

	if (i != hostcontrol->TableSize)
	{
		DnpapMessage(DMC_FATAL, HOST_UPDTIME, "hostcontrol: internal error in time list and table");
		DnpapExit(HOST_UPDTIME);
	}

	hostcontrol->TimeListUpdated = FALSE;

	return TRUE;
}

int addrcmp(HOST * host1, HOST * host2, WORD addrlen, WORD len)
{
	INT i1, i2;

	BYTE *a1 = host1->Address, *a2 = host2->Address;

	if (len > addrlen)
		len = addrlen;

	for (i1 = 0, i2 = 0; len > 0; i1++, i2++, len--)
	{
		if (a1[i1] < a2[i2])
			return -1;
		else if (a1[i1] > a2[i2])
			return 1;
	}

	return 0;
}

#ifdef DEBUG

VOID HostCheckListsConsistency(HOST_CONTROL * hostcontrol)
{
	HOST *p;
	LONG count;
	WORD addrlen = hostcontrol->AddressLen;

	for (p = hostcontrol->LexList, count = 0; p != NULL; p = p->LexNext, count++)
		if (HashSearch(hostcontrol->Table, p->Address, addrlen) != p)
			DnpapMessage(DMC_FATAL, HOST_ERROR, "hostcontrol: internal error in lexical list");
	if (count != hostcontrol->TableSize)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "hostcontrol: wrong number of elements in lexical list");

	for (p = hostcontrol->TimeList, count = 0; p != NULL; p = p->TimeNext, count++)
		if (HashSearch(hostcontrol->Table, p->Address, addrlen) != p)
			DnpapMessage(DMC_FATAL, HOST_ERROR, "hostcontrol: internal error in time list");
	if (count != hostcontrol->TableSize)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "hostcontrol: wrong number of elements in time list");

	for (p = hostcontrol->LRUList, count = 0; p != NULL; p = p->LRUNext, count++)
		if (HashSearch(hostcontrol->Table, p->Address, addrlen) != p)
			DnpapMessage(DMC_FATAL, HOST_ERROR, "hostcontrol: internal error in LRU list");
	if (count != hostcontrol->TableSize)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "hostcontrol: wrong number of elements in time list");

	if (hostcontrol->TimeListUpdated == TRUE)
		HostUpdateTimeTable(hostcontrol);

	for (p = hostcontrol->TimeList, count = 0; p != NULL; p = p->TimeNext, count++)
		if (hostcontrol->TimeTable[count] != p || p->TimeListPtr != hostcontrol->TimeTable + count)
			DnpapMessage(DMC_FATAL, HOST_ERROR, "hostcontrol: internal error in time table");
}

VOID HostCheckListsAdd(HOST_CONTROL * hostcontrol, HOST * host)
{
	HOST *p;
	WORD addrlen = hostcontrol->AddressLen;

	if (HashSearch(hostcontrol->Table, host->Address, addrlen) != NULL)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "host already in hash table");

	for (p = hostcontrol->LexList; p != NULL; p = p->LexNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) == 0)
			DnpapMessage(DMC_FATAL, HOST_ERROR, "host already in lexical list");
	}
	for (p = hostcontrol->TimeList; p != NULL; p = p->TimeNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) == 0)
			DnpapMessage(DMC_FATAL, HOST_ERROR, "host already in time list");
	}
	for (p = hostcontrol->LRUList; p != NULL; p = p->LRUNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) == 0)
			DnpapMessage(DMC_FATAL, HOST_ERROR, "host already in LRU list");
	}
}

VOID HostCheckListsAdd2(HOST_CONTROL * hostcontrol, HOST * host)
{
	HOST *p;
	WORD addrlen = hostcontrol->AddressLen;

	if (HashSearch(hostcontrol->Table, host->Address, addrlen) == NULL)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "host not added to hash table");

	for (p = hostcontrol->LexList; p != NULL; p = p->LexNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) == 0)
			break;
	}
	if (p == NULL)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "host not added to lexical list");
	for (p = hostcontrol->TimeList; p != NULL; p = p->TimeNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) == 0)
			break;
	}
	if (p == NULL)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "host not added to time list");
	for (p = hostcontrol->LRUList; p != NULL; p = p->LRUNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) == 0)
			break;
	}
	if (p == NULL)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "host not added to LRU list");
}

VOID HostCheckListsRemove(HOST_CONTROL * hostcontrol, HOST * host)
{
	HOST *p;
	WORD addrlen = hostcontrol->AddressLen;

	if (HashSearch(hostcontrol->Table, host->Address, addrlen) != NULL)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "host that was just removed can still be retrieved");

	for (p = hostcontrol->LexList; p != NULL; p = p->LexNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) == 0)
			DnpapMessage(DMC_FATAL, HOST_ERROR, "host that was just removed can still be retrieved from lexical list");
	}
	for (p = hostcontrol->TimeList; p != NULL; p = p->TimeNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) == 0)
			DnpapMessage(DMC_FATAL, HOST_ERROR, "host that was just removed can still be retrieved from time list");
	}
	for (p = hostcontrol->LRUList; p != NULL; p = p->LRUNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) == 0)
			DnpapMessage(DMC_FATAL, HOST_ERROR, "host that was just removed can still be retrieved from LRU list");
	}
}

VOID HostCheckListsRemove2(HOST_CONTROL * hostcontrol, HOST * host)
{
	HOST *p;
	WORD addrlen = hostcontrol->AddressLen;

	if (HashSearch(hostcontrol->Table, host->Address, addrlen) == NULL)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "host already removed");

	for (p = hostcontrol->LexList; p != NULL; p = p->LexNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) == 0)
			break;
	}
	if (p == NULL)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "host already removed from lexical list");
	for (p = hostcontrol->TimeList; p != NULL; p = p->TimeNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) == 0)
			break;
	}
	if (p == NULL)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "host already removed from time list");
	for (p = hostcontrol->LRUList; p != NULL; p = p->LRUNext)
	{
		if (addrcmp(p, host, addrlen, addrlen) == 0)
			break;
	}
	if (p == NULL)
		DnpapMessage(DMC_FATAL, HOST_ERROR, "host already removed from LRU list");
}

#endif
