/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */

#include <dnpap.h>
#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <message.h>
#include <mibsup.h>
#include <mibrqs.h>
#include <mac.h>

#include "hoste.h"
#include "hostc.h"
#include "host.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define CONTROLINDEXSIZE   1
#define HOSTINDEXSIZE      7			/* not the real size, only used as switch value */
#define TIMEINDEXSIZE      2

static MIB_LOCAL *HostControlInstance = NULL;

static MIB_LOCAL *HostMibRmon(SNMP_OBJECT * obj, MIB_LOCAL * local, WORD idlen);
static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, HOST ** elem);

static HOST *HostGetFirstNext(HOST_CONTROL * hostcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first);

BOOLEAN HostMInit(VOID)
{
	LONG i, nriface;
	CHAR s[81];
	BOOLEAN v;

	MAC_IFACE *iface;

	LONG hostsource[] =
	{1, 3, 6, 1, 2, 1, 2, 2, 1, 1, 1};
	CHAR hostowner[] = "monitorHost";

#define STATUSINDEX     11
	SNMP_OBJECT hostobj[] =
	{
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 4, 1, 1, 6, 1}, 12, SNMP_INTEGER},
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 4, 1, 1, 5, 1}, 12, SNMP_DISPLAYSTR},
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 4, 1, 1, 2, 1}, 12, SNMP_OBJECTID}
	};
	WORD dummy;

	MessageConfig(HOST_ERROR, "Host");
	if (HostConfigInit() == FALSE)
		return FALSE;

	nriface = (LONG) MacIfaceCount();
	for (i = 1; i <= nriface; i++)
	{
		sprintf(s, "beholder.host.iface.%ld", i);
		if (ConfigGetBoolean(s, &v) == TRUE)
		{
			iface = MacIfaceGet((WORD) i);
			if (iface != NULL && v == TRUE &&
				(iface->type == MAC_TYPE_ETHERNET_CSMACD ||
				 iface->type == MAC_TYPE_88023_CSMACD))
			{
				hostobj[0].Id[STATUSINDEX] = i;
				hostobj[1].Id[STATUSINDEX] = i;
				hostobj[2].Id[STATUSINDEX] = i;

				hostobj[0].Request = SNMP_PDU_SET;
				hostobj[0].Syntax.LngInt = SNMP_INVALID;
				MibRequest(1, &hostobj[0], &dummy);
				hostobj[0].Request = SNMP_PDU_SET;
				hostobj[0].Syntax.LngInt = SNMP_CREATEREQUEST;
				if (MibRequest(1, &hostobj[0], &dummy) != SNMP_NOERROR)
				{
					DnpapMessage(DMC_WARNING, HOST_HOSTINIT, "host: couldn't create host collector %ld", i);
					continue;
				}
				hostobj[1].Request = SNMP_PDU_SET;
				strncpy(hostobj[1].Syntax.BufChr, hostowner, hostobj[1].SyntaxLen = strlen(hostowner));
				hostobj[2].Request = SNMP_PDU_SET;
				memcpy(hostobj[2].Syntax.BufInt, hostsource, sizeof(hostsource));
				hostobj[2].Syntax.BufInt[sizeof(hostsource) / sizeof(hostsource[0]) - 1] = i;
				hostobj[2].SyntaxLen = sizeof(hostsource) / sizeof(hostsource[0]);
				if (MibRequest(2, &hostobj[1], &dummy) != SNMP_NOERROR)
				{
					DnpapMessage(DMC_WARNING, HOST_HOSTINIT, "host: couldn't set owner of host collector %ld", i);
					hostobj[0].Request = SNMP_PDU_SET;
					hostobj[0].Syntax.LngInt = SNMP_INVALID;
					MibRequest(1, &hostobj[0], &dummy);
					continue;
				}
				hostobj[0].Request = SNMP_PDU_SET;
				hostobj[0].Syntax.LngInt = SNMP_VALID;
				if (MibRequest(1, &hostobj[0], &dummy) != SNMP_NOERROR)
				{
					DnpapMessage(DMC_WARNING, HOST_HOSTINIT, "host: couldn't set host collector %ld to valid", i);
					hostobj[0].Request = SNMP_PDU_SET;
					hostobj[0].Syntax.LngInt = SNMP_INVALID;
					MibRequest(1, &hostobj[0], &dummy);
					continue;
				}
			}
		}
	}

	return TRUE;
}

MIB_LOCAL *HostMibRmon(SNMP_OBJECT * obj, MIB_LOCAL * local, WORD idlen)
{
	WORD addrlen;

	if (local == NULL)
		return NULL;
	if (obj->Request != SNMP_PDU_NEXT)
	{
		if (obj->IdLen == idlen)
			return NULL;
		/* first search the right table */
		while (local != NULL && local->Index < obj->Id[idlen])
			local = local->Next;
		if (local == NULL)
			return NULL;
		addrlen = ((HOST_CONTROL *) local->Data)->AddressLen;
		if (local->Index != obj->Id[idlen] || addrlen == 0)
			return NULL;
		if (obj->IdLen != idlen + 1 + 1 + addrlen ||
			obj->Id[idlen + 1] != addrlen)
			return NULL;
		return local;
	}
	if (obj->IdLen == idlen)			/* first instance */
		return local;
	while (local != NULL && local->Index < obj->Id[idlen])
		local = local->Next;
	if (local == NULL || local->Index < obj->Id[idlen])
		return NULL;
	return local;
}

#define CONTROL_DATA    ((HOST_CONTROL *) (*local)->Data)

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, HOST ** elem)
{
	BOOLEAN first = FALSE;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case HOSTINDEXSIZE:
		while (*local != NULL && (*elem = HostGetFirstNext(CONTROL_DATA, obj, idlen, first)) == NULL)
		{
			first = TRUE;
			*local = (*local)->Next;
		}
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		break;
	case TIMEINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			if (obj->Id[idlen + 1] < CONTROL_DATA->TableSize)
			{
				obj->Id[idlen] = (*local)->Index;
				obj->Id[idlen + 1]++;
				obj->IdLen = idlen + TIMEINDEXSIZE;
				*elem = HostTimeSearch(CONTROL_DATA, obj, idlen);
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL && CONTROL_DATA->TableSize == 0)
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen + 1] = 1;
		obj->IdLen = idlen + TIMEINDEXSIZE;
		*elem = HostTimeSearch(CONTROL_DATA, obj, idlen);
		break;
	case CONTROLINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + CONTROLINDEXSIZE;
		break;
	}
	return TRUE;
}

HOST *HostGetFirstNext(HOST_CONTROL * hostcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first)
{
	HOST host, *p;
	INT i, len;
	WORD addrlen = hostcontrol->AddressLen;

	if (hostcontrol->TableSize == 0)
		return NULL;

	if (first == FALSE)
	{
		if (obj->IdLen > idlen + 1)
		{
			if (obj->Id[idlen + 1] > addrlen)
				return NULL;
			if (obj->Id[idlen + 1] < addrlen)
				first = TRUE;
		}
		else
			first = TRUE;
	}

	if (first == TRUE)
	{
		p = hostcontrol->LexList;
		for (i = 0; i < addrlen; i++)
			obj->Id[idlen + 2 + i] = (LONG) p->Address[i];
		obj->Id[idlen + 1] = addrlen;
		obj->IdLen = idlen + 1 + 1 + addrlen;
		return p;
	}

	len = obj->IdLen - idlen - 2;
	if (len < 0)
		len = 0;
	else if (len > addrlen)
		len = addrlen;

	for (i = 0; i < len; i++)
		host.Address[i] = (BYTE) obj->Id[idlen + 2 + i];

	p = HostSearchNext(hostcontrol, &host, len);

	if (p != NULL)
	{
		for (i = 0; i < addrlen; i++)
			obj->Id[idlen + 2 + i] = (LONG) p->Address[i];
		obj->Id[idlen + 1] = addrlen;
		obj->IdLen = idlen + 1 + 1 + addrlen;
	}
	return p;
}

WORD hostControlIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostControlDataSource(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST_CONTROL *data;
	MAC_IFACE *iface;
	LONG IfaceOid[] =
	{1, 3, 6, 1, 2, 1, 2, 2, 1, 1, 1};
	WORD IfaceOidLen = sizeof(IfaceOid) / sizeof(LONG);
	LONG ChannelOid[] =
	{1, 3, 6, 1, 2, 1, 16, 7, 2, 1, 1, 1};
	WORD ChannelOidLen = sizeof(ChannelOid) / sizeof(LONG);

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (HOST_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, HOST_BOOLEAN_DATASOURCE))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufInt, data->Source, data->SourceLen * sizeof(data->Source[0]));
		Obj->SyntaxLen = data->SourceLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (HOST_CONTROL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (IfaceOidLen == Obj->SyntaxLen &&
			memcmp(IfaceOid, Obj->Syntax.BufInt, (IfaceOidLen - 1) * sizeof(IfaceOid[0])) == 0)
		{
			data->Channel = FALSE;
			if ((iface = MacIfaceGet((WORD) Obj->Syntax.BufInt[Obj->SyntaxLen - 1])) == NULL)
				return SNMP_BADVALUE;
			if (iface->type != MAC_TYPE_ETHERNET_CSMACD && iface->type != MAC_TYPE_88023_CSMACD)
				return SNMP_BADVALUE;
			memcpy(data->Source, Obj->Syntax.BufInt, Obj->SyntaxLen * sizeof(data->Source[0]));
			data->SourceLen = Obj->SyntaxLen;
			data->Iface = iface;
			BooleanSetTrue(data->ObjectSet, HOST_BOOLEAN_DATASOURCE);
			return SNMP_NOERROR;
		}
		else if (ChannelOidLen == Obj->SyntaxLen &&
				 memcmp(ChannelOid, Obj->Syntax.BufInt, (ChannelOidLen - 1) * sizeof(ChannelOid[0])) == 0)
		{
			data->Channel = TRUE;
			memcpy(data->Source, Obj->Syntax.BufInt, Obj->SyntaxLen * sizeof(data->Source[0]));
			data->SourceLen = Obj->SyntaxLen;
			BooleanSetTrue(data->ObjectSet, HOST_BOOLEAN_DATASOURCE);
			return SNMP_NOERROR;
		}
		return SNMP_BADVALUE;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostControlTableSize(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST_CONTROL *data;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (HOST_CONTROL *) local->Data;
		Obj->Syntax.LngInt = data->TableSize;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostControlLastDeleteTime(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST_CONTROL *data;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (HOST_CONTROL *) local->Data;
		Obj->Syntax.LngUns = data->LastDeleteTime;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostControlOwner(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST_CONTROL *data;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (HOST_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, HOST_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		Obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (HOST_CONTROL *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, Obj->Syntax.BufChr, Obj->SyntaxLen);
		data->OwnerLen = Obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, HOST_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostControlStatus(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST_CONTROL *data;

	local = MibRmon(Obj, HostControlInstance, IdLen, CONTROLINDEXSIZE);

	switch (Obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (HOST_CONTROL *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, CONTROLINDEXSIZE, NULL) == FALSE)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (HOST_CONTROL *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(Obj, &HostControlInstance, IdLen, CONTROLINDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(HOST_CONTROL))) == NULL)
				{
					MibRemove(Obj, &HostControlInstance, IdLen, CONTROLINDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (HOST_CONTROL *) local->Data;
				if (HostCInit(data) == TRUE)
				{
					data->Index = local->Index;
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, HOST_CREATE, "host: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(Obj, &HostControlInstance, IdLen, CONTROLINDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (HOST_CONTROL *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (HostCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, HOST_START, "host: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (HostCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, HOST_DESTROY, "host: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &HostControlInstance, IdLen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				return SNMP_NOERROR;
			case SNMP_INVALID:
				if (HostCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, HOST_DESTROY, "host: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &HostControlInstance, IdLen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (HostCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostAddress(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = HostMibRmon(Obj, HostControlInstance, IdLen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostSearchHash(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufChr, data->Address, Obj->SyntaxLen = ((HOST_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostCreationOrder(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = HostMibRmon(Obj, HostControlInstance, IdLen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostSearchHash(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = HostSearchCreationOrder(local->Data, data);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = HostMibRmon(Obj, HostControlInstance, IdLen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostSearchHash(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostInPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = HostMibRmon(Obj, HostControlInstance, IdLen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostSearchHash(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->InPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostOutPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = HostMibRmon(Obj, HostControlInstance, IdLen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostSearchHash(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostInOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = HostMibRmon(Obj, HostControlInstance, IdLen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostSearchHash(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->InOctets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostOutOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = HostMibRmon(Obj, HostControlInstance, IdLen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostSearchHash(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutOctets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostOutErrors(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = HostMibRmon(Obj, HostControlInstance, IdLen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostSearchHash(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutErrors;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostOutBroadcastPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = HostMibRmon(Obj, HostControlInstance, IdLen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostSearchHash(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutBroadcastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostOutMulticastPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = HostMibRmon(Obj, HostControlInstance, IdLen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, HOSTINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostSearchHash(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutMulticastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostTimeAddress(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostTimeSearch(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufChr, data->Address, Obj->SyntaxLen = ((HOST_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostTimeCreationOrder(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostTimeSearch(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = Obj->Id[IdLen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostTimeIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostTimeSearch(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostTimeInPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostTimeSearch(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->InPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostTimeOutPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostTimeSearch(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostTimeInOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostTimeSearch(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->InOctets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostTimeOutOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostTimeSearch(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutOctets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostTimeOutErrors(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostTimeSearch(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutErrors;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostTimeOutBroadcastPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostTimeSearch(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutBroadcastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD hostTimeOutMulticastPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	HOST *data = NULL;

	if ((local = MibRmon(Obj, HostControlInstance, IdLen, TIMEINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, TIMEINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = HostTimeSearch(local->Data, Obj, IdLen)) == NULL)
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngUns = data->OutMulticastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

TopNCallback *HostRegisterTopN(ULONG hostindex, HostTopNCallback notify, VOID * topn)
{
	MIB_LOCAL *local;
	HOST_CONTROL *data;
	TopNCallback *callback;

	for (local = HostControlInstance; local != NULL; local = local->Next)
	{
		data = (HOST_CONTROL *) local->Data;
		if (data->Index < hostindex)
			continue;
		if (data->Index != hostindex)
			return NULL;
		if ((callback = HostAddTopN(data, notify, topn)) != NULL)
			return callback;
		break;
	}

	return NULL;
}

BOOLEAN HostDeRegisterTopN(ULONG hostindex, TopNCallback * topn)
{
	MIB_LOCAL *local;
	HOST_CONTROL *data;

	for (local = HostControlInstance; local != NULL; local = local->Next)
	{
		data = (HOST_CONTROL *) local->Data;
		if (data->Index < hostindex)
			continue;
		if (data->Index != hostindex)
			return FALSE;
		if (HostRemoveTopN(data, topn) == TRUE)
			return TRUE;
		break;
	}

	return FALSE;
}

BOOLEAN HostRegisterAll(ULONG hostindex)
{
	MIB_LOCAL *local;
	HOST_CONTROL *data;

	for (local = HostControlInstance; local != NULL; local = local->Next)
	{
		data = (HOST_CONTROL *) local->Data;
		if (data->Channel == TRUE && data->Source[data->SourceLen - 1] == hostindex &&
			data->ChannelCallback == NULL)
		{
			data->ChannelCallback = ChannelRegisterSink(data->Source[data->SourceLen - 1],
														HostChannelPkt, HostDeRegister, data);
		}
	}

	return TRUE;
}
