/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>
#include <memory.h>
#include <dmath.h>
#include <maxmem.h>
#include <snmp.h>
#include <stdlib.h>

#include "log.h"

static CHAR MODULE[] = "log";

WORD LogEventIndex(SNMP_OBJECT * obj, WORD idlen, Log * log)
{
	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = obj->Id[idlen];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_NOSUCHNAME;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD LogIndex(SNMP_OBJECT * obj, WORD idlen, Log * log)
{
	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = obj->Id[idlen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_NOSUCHNAME;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD LogTime(SNMP_OBJECT * obj, WORD idlen, Log * log)
{
	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
	case SNMP_PDU_GET:
		obj->Syntax.LngUns = log->Time;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_NOSUCHNAME;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD LogDescription(SNMP_OBJECT * obj, WORD idlen, Log * log)
{
	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
	case SNMP_PDU_GET:
		memcpy(obj->Syntax.BufChr, log->Description, obj->SyntaxLen = log->DescrLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_NOSUCHNAME;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

Log **NewLog(LONG entries, LONG maxentries)
{
	return ReNewLog(NULL, entries, 0, maxentries);
}

Log **ReNewLog(Log ** ptr, LONG entries, LONG oldentries, LONG maxentries)
{
	Log **p;
	LONG i, n, nold;

	if (entries > maxentries)
		return NULL;
	n = ((entries - 1) / MINALLOCLOGS) * MINALLOCLOGS + MINALLOCLOGS;
	nold = (oldentries <= 0 ? 0 :
			((oldentries - 1) / MINALLOCLOGS) * MINALLOCLOGS + MINALLOCLOGS);
	if (ptr == NULL || abs(n - nold) >= MINALLOCLOGS)
	{
		if ((p = DnpapMalloc(n * sizeof(Log *))) == NULL)
			return NULL;
		if (n > nold)
		{
			for (i = 0; i < nold; i++)
				p[i] = ptr[i];
			for (i = nold; i < n; i++)
				p[i] = NULL;
			for (i = nold; i < n; i++)
			{
				p[i] = DnpapMalloc(sizeof(Log));
				if (p[i] == NULL)
				{
					DelLog(p, n);
					return NULL;
				}
			}
		}
		else
		{
			for (i = 0; i < n; i++)
				p[i] = ptr[i];
			for (i = n; i < nold; i++)
				DnpapFree(ptr[i]);
		}
		DnpapFree(ptr);
		ptr = p;
	}
	return ptr;
}

VOID DelLog(Log ** ptr, LONG entries)
{
	LONG i, n;

	if (ptr == NULL)
		return;
	n = ((entries - 1) / MINALLOCLOGS) * MINALLOCLOGS + MINALLOCLOGS;
	for (i = 0; i < n; i++)
		DnpapFree(ptr[i]);
	DnpapFree(ptr);
}
