/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#ifndef _MATRIX_C_H
#define _MATRIX_C_H


#include <dnpap.h>
#include <boolset.h>
#include <snmp.h>
#include <hash.h>
#include <channel.h>


#define MATRIX_BOOLEAN_DATASOURCE		1
#define MATRIX_BOOLEAN_OWNER			2


typedef struct _SRCDST
{
    BYTE        SrcDst[2*HOST_SIZE_ADDR];
    LWORD       Pkts;
    LWORD       Octets;
    LWORD       Errors;

    struct _SRCDST  *SrcDstPrev;
    struct _SRCDST  *SrcDstNext;
    
    struct _SRCDST  *DstSrcPrev;
    struct _SRCDST  *DstSrcNext;

    struct _SRCDST  *LRUPrev;
    struct _SRCDST  *LRUNext;
    
} SRCDST;


typedef struct _MATRIX_CONTROL
{
    ULONG           Source[SNMP_SIZE_BUFINT];
    WORD            SourceLen;
    BOOLEAN         Channel;
    SinkCallback   *ChannelCallback;
	WORD			AddressLen;
    LONG            TableSize;
    ULONG           LastDeleteTime;
    BYTE            Owner[SNMP_SIZE_BUFCHR];
    WORD            OwnerLen;
    LONG            Status;
    
    MAC_COLL        Coll;
    MAC_IFACE       *Iface;
    
	HashTable		*Table;
	
	SRCDST			*SrcDstList;
	SRCDST			*DstSrcList;

    SRCDST          *LRUList;
    SRCDST          *LRULast;
    
	BooleanSet       ObjectSet;
    
} MATRIX_CONTROL;


BOOLEAN MatrixConfigInit(VOID);

BOOLEAN MatrixCInit(MATRIX_CONTROL *matrixcontrol);
BOOLEAN MatrixCStart(MATRIX_CONTROL *matrixcontrol);
BOOLEAN MatrixCStop(MATRIX_CONTROL *matrixcontrol);

VOID MatrixChannelPkt(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *sink);
VOID MatrixDeRegister(ULONG channelIndex, VOID *sink);

SRCDST* MatrixSearchNext(MATRIX_CONTROL *matrixcontrol, BYTE *srcdst, WORD len, BOOLEAN SDorder);
SRCDST* MatrixSearchHash(MATRIX_CONTROL *matrixcontrol, SNMP_OBJECT *obj, WORD idlen, BOOLEAN SDorder);


#endif
