/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>
#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <message.h>
#include <mibsup.h>
#include <mibrqs.h>
#include <mac.h>

#include "matrix_d.h"
#include "matrix_e.h"
#include "matrix_c.h"
#include "matrix.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define CONTROLINDEXSIZE   1
#define SRCDSTINDEXSIZE   13			/* not the real size, only used as switch value */

static MIB_LOCAL *MatrixControlInstance = NULL;

static MIB_LOCAL *MatrixMibRmon(SNMP_OBJECT * obj, MIB_LOCAL * local, WORD idlen);
static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, SRCDST ** elem, BOOLEAN SDorder);
static SRCDST *SrcDstGetFirstNext(MATRIX_CONTROL * matrixcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first, BOOLEAN SDorder);

BOOLEAN MatrixMInit(VOID)
{
	LONG i, nriface;
	CHAR s[81];
	BOOLEAN v;

	MAC_IFACE *iface;

	LONG matrixsource[] =
	{1, 3, 6, 1, 2, 1, 2, 2, 1, 1, 1};
	CHAR matrixowner[] = "monitorMatrix";

#define STATUSINDEX     11
	SNMP_OBJECT matrixobj[] =
	{
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 6, 1, 1, 6, 1}, 12, SNMP_INTEGER},
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 6, 1, 1, 5, 1}, 12, SNMP_DISPLAYSTR},
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 6, 1, 1, 2, 1}, 12, SNMP_OBJECTID}
	};
	WORD dummy;

	MessageConfig(MATRIX_ERROR, "Matrix");
	if (MatrixConfigInit() == FALSE)
		return FALSE;

	nriface = (LONG) MacIfaceCount();
	for (i = 1; i <= nriface; i++)
	{
		sprintf(s, "beholder.matrix.iface.%ld", i);
		if (ConfigGetBoolean(s, &v) == TRUE)
		{
			iface = MacIfaceGet((WORD) i);
			if (iface != NULL && v == TRUE &&
				(iface->type == MAC_TYPE_ETHERNET_CSMACD ||
				 iface->type == MAC_TYPE_88023_CSMACD))
			{
				matrixobj[0].Id[STATUSINDEX] = i;
				matrixobj[1].Id[STATUSINDEX] = i;
				matrixobj[2].Id[STATUSINDEX] = i;

				matrixobj[0].Request = SNMP_PDU_SET;
				matrixobj[0].Syntax.LngInt = SNMP_INVALID;
				MibRequest(1, &matrixobj[0], &dummy);
				matrixobj[0].Request = SNMP_PDU_SET;
				matrixobj[0].Syntax.LngInt = SNMP_CREATEREQUEST;
				if (MibRequest(1, &matrixobj[0], &dummy) != SNMP_NOERROR)
				{
					DnpapMessage(DMC_WARNING, MATRIX_MATRIXINIT, "matrix: couldn't create matrix collector %ld", i);
					continue;
				}
				matrixobj[1].Request = SNMP_PDU_SET;
				strncpy(matrixobj[1].Syntax.BufChr, matrixowner, matrixobj[1].SyntaxLen = strlen(matrixowner));
				matrixobj[2].Request = SNMP_PDU_SET;
				memcpy(matrixobj[2].Syntax.BufInt, matrixsource, sizeof(matrixsource));
				matrixobj[2].Syntax.BufInt[sizeof(matrixsource) / sizeof(matrixsource[0]) - 1] = i;
				matrixobj[2].SyntaxLen = sizeof(matrixsource) / sizeof(matrixsource[0]);
				if (MibRequest(2, &matrixobj[1], &dummy) != SNMP_NOERROR)
				{
					DnpapMessage(DMC_WARNING, MATRIX_MATRIXINIT, "matrix: couldn't set owner of matrix collector %ld", i);
					matrixobj[0].Request = SNMP_PDU_SET;
					matrixobj[0].Syntax.LngInt = SNMP_INVALID;
					MibRequest(1, &matrixobj[0], &dummy);
					continue;
				}
				matrixobj[0].Request = SNMP_PDU_SET;
				matrixobj[0].Syntax.LngInt = SNMP_VALID;
				if (MibRequest(1, &matrixobj[0], &dummy) != SNMP_NOERROR)
				{
					DnpapMessage(DMC_WARNING, MATRIX_MATRIXINIT, "matrix: couldn't set matrix collector %ld to valid", i);
					matrixobj[0].Request = SNMP_PDU_SET;
					matrixobj[0].Syntax.LngInt = SNMP_INVALID;
					MibRequest(1, &matrixobj[0], &dummy);
					continue;
				}
			}
		}
	}

	return TRUE;
}

MIB_LOCAL *MatrixMibRmon(SNMP_OBJECT * obj, MIB_LOCAL * local, WORD idlen)
{
	WORD addrlen;

	if (local == NULL)
		return NULL;
	if (obj->Request != SNMP_PDU_NEXT)
	{
		if (obj->IdLen == idlen)
			return NULL;
		/* first search the right table */
		while (local != NULL && local->Index < obj->Id[idlen])
			local = local->Next;
		if (local == NULL)
			return NULL;
		addrlen = ((MATRIX_CONTROL *) local->Data)->AddressLen;
		if (local->Index != obj->Id[idlen] || addrlen == 0)
			return NULL;
		if (obj->IdLen != idlen + 1 + 2 + 2 * addrlen ||
			obj->Id[idlen + 1] != addrlen || obj->Id[idlen + 1 + addrlen + 1] != addrlen)
			return NULL;
		return local;
	}
	if (obj->IdLen == idlen)			/* first instance */
		return local;
	while (local != NULL && local->Index < obj->Id[idlen])
		local = local->Next;
	if (local == NULL || local->Index < obj->Id[idlen])
		return NULL;
	return local;
}

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, SRCDST ** elem, BOOLEAN SDorder)
{
	BOOLEAN first = FALSE;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case SRCDSTINDEXSIZE:
		while (*local != NULL && (*elem = SrcDstGetFirstNext((MATRIX_CONTROL *) (*local)->Data, obj, idlen, first, SDorder)) == NULL)
		{
			first = TRUE;
			*local = (*local)->Next;
		}
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		break;
	case CONTROLINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + CONTROLINDEXSIZE;
		break;
	}
	return TRUE;
}

SRCDST *SrcDstGetFirstNext(MATRIX_CONTROL * matrixcontrol, SNMP_OBJECT * obj, WORD idlen, BOOLEAN first, BOOLEAN SDorder)
{
	BYTE srcdst[2 * HOST_SIZE_ADDR];
	SRCDST *p;
	INT i, len;
	BOOLEAN nextconn, firstconn;
	WORD addrlen = matrixcontrol->AddressLen;

	if (matrixcontrol->TableSize == 0)
		return NULL;

	nextconn = FALSE;
	firstconn = FALSE;

	if (first == FALSE)
	{
		if (obj->IdLen > idlen + 1)
		{
			if (obj->Id[idlen + 1] > addrlen)
				return NULL;
			if (obj->Id[idlen + 1] < addrlen)
				first = TRUE;
		}
		else
			first = TRUE;

		if (obj->IdLen > idlen + 1 + addrlen + 1)
		{
			if (obj->Id[idlen + 1 + addrlen + 1] > addrlen)
				nextconn = TRUE;
			if (obj->Id[idlen + 1 + addrlen + 1] < addrlen)
				firstconn = TRUE;
		}
	}

	if (first == TRUE)
	{
		if (SDorder == TRUE)
		{
			p = matrixcontrol->SrcDstList;
			for (i = 0; i < addrlen; i++)
			{
				obj->Id[idlen + 2 + i] = (LONG) p->SrcDst[i];
				obj->Id[idlen + 3 + addrlen + i] = (LONG) p->SrcDst[addrlen + i];
				obj->Id[idlen + 1] = addrlen;
				obj->Id[idlen + 1 + addrlen + 1] = addrlen;
				obj->IdLen = idlen + 1 + 2 + 2 * addrlen;
			}
			return p;
		}
		else
		{
			p = matrixcontrol->DstSrcList;
			for (i = 0; i < addrlen; i++)
			{
				obj->Id[idlen + 2 + i] = (LONG) p->SrcDst[addrlen + i];
				obj->Id[idlen + 3 + addrlen + i] = (LONG) p->SrcDst[i];
				obj->Id[idlen + 1] = addrlen;
				obj->Id[idlen + 1 + addrlen + 1] = addrlen;
				obj->IdLen = idlen + 1 + 2 + 2 * addrlen;
			}
			return p;
		}
	}

	len = obj->IdLen - idlen - 2;
	if (len < 0)
		len = 0;
	else if (len > addrlen)
		len--;
	if (len > 2 * addrlen)
		len = 2 * addrlen;

	/* fill in srcdst in both orders: SDorder or reverse  */
	for (i = 0; i < len; i++)
		srcdst[i] = (BYTE) obj->Id[idlen + 2 + (i < addrlen ? 0 : 1) + i];

	if (firstconn == TRUE)
		len = addrlen;					/* force finding of first connection */
	if (nextconn == TRUE)
	{
		len = addrlen;
		/* force finding of next connection */
		for (i = addrlen - 1; i >= 0; i--)
			if (++srcdst[i] != 0)
				break;
		if (i < 0 && srcdst[0] == 0)
			return NULL;
	}

	p = MatrixSearchNext(matrixcontrol, srcdst, len, SDorder);

	if (p != NULL)
	{
		if (SDorder == TRUE)
		{
			for (i = 0; i < addrlen; i++)
			{
				obj->Id[idlen + 2 + i] = (LONG) p->SrcDst[i];
				obj->Id[idlen + 3 + addrlen + i] = (LONG) p->SrcDst[addrlen + i];
			}
		}
		else
		{
			for (i = 0; i < addrlen; i++)
			{
				obj->Id[idlen + 2 + i] = (LONG) p->SrcDst[addrlen + i];
				obj->Id[idlen + 3 + addrlen + i] = (LONG) p->SrcDst[i];
			}
		}

		obj->Id[idlen + 1] = addrlen;
		obj->Id[idlen + 1 + addrlen + 1] = addrlen;
		obj->IdLen = idlen + 1 + 2 + 2 * addrlen;
	}

	return p;
}

WORD MatrixControlIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, MatrixControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL, FALSE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixControlDataSource(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	MATRIX_CONTROL *data;
	MAC_IFACE *iface;
	LONG IfaceOid[] =
	{1, 3, 6, 1, 2, 1, 2, 2, 1, 1, 1};
	WORD IfaceOidLen = sizeof(IfaceOid) / sizeof(LONG);
	LONG ChannelOid[] =
	{1, 3, 6, 1, 2, 1, 16, 7, 2, 1, 1, 1};
	WORD ChannelOidLen = sizeof(ChannelOid) / sizeof(LONG);

	if ((local = MibRmon(obj, MatrixControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL, FALSE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (MATRIX_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, MATRIX_BOOLEAN_DATASOURCE))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufInt, data->Source, data->SourceLen * sizeof(data->Source[0]));
		obj->SyntaxLen = data->SourceLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (MATRIX_CONTROL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (IfaceOidLen == obj->SyntaxLen &&
			memcmp(IfaceOid, obj->Syntax.BufInt, (IfaceOidLen - 1) * sizeof(IfaceOid[0])) == 0)
		{
			data->Channel = FALSE;
			if ((iface = MacIfaceGet((WORD) obj->Syntax.BufInt[obj->SyntaxLen - 1])) == NULL)
				return SNMP_BADVALUE;
			if (iface->type != MAC_TYPE_ETHERNET_CSMACD && iface->type != MAC_TYPE_88023_CSMACD)
				return SNMP_BADVALUE;
			memcpy(data->Source, obj->Syntax.BufInt, obj->SyntaxLen * sizeof(data->Source[0]));
			data->SourceLen = obj->SyntaxLen;
			data->Iface = iface;
			BooleanSetTrue(data->ObjectSet, MATRIX_BOOLEAN_DATASOURCE);
			return SNMP_NOERROR;
		}
		else if (ChannelOidLen == obj->SyntaxLen &&
				 memcmp(ChannelOid, obj->Syntax.BufInt, (ChannelOidLen - 1) * sizeof(ChannelOid[0])) == 0)
		{
			data->Channel = TRUE;
			memcpy(data->Source, obj->Syntax.BufInt, obj->SyntaxLen * sizeof(data->Source[0]));
			data->SourceLen = obj->SyntaxLen;
			BooleanSetTrue(data->ObjectSet, MATRIX_BOOLEAN_DATASOURCE);
			return SNMP_NOERROR;
		}
		return SNMP_BADVALUE;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixControlTableSize(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	MATRIX_CONTROL *data;

	if ((local = MibRmon(obj, MatrixControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL, FALSE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (MATRIX_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->TableSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixControlLastDeleteTime(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	MATRIX_CONTROL *data;

	if ((local = MibRmon(obj, MatrixControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL, FALSE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (MATRIX_CONTROL *) local->Data;
		obj->Syntax.LngUns = data->LastDeleteTime;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixControlOwner(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	MATRIX_CONTROL *data;

	if ((local = MibRmon(obj, MatrixControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL, FALSE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (MATRIX_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, MATRIX_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (MATRIX_CONTROL *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, obj->Syntax.BufChr, obj->SyntaxLen);
		data->OwnerLen = obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, MATRIX_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixControlStatus(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	MATRIX_CONTROL *data;

	local = MibRmon(obj, MatrixControlInstance, idlen, CONTROLINDEXSIZE);

	switch (obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (MATRIX_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL, FALSE) == FALSE)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (MATRIX_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(obj, &MatrixControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(MATRIX_CONTROL))) == NULL)
				{
					MibRemove(obj, &MatrixControlInstance, idlen, CONTROLINDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (MATRIX_CONTROL *) local->Data;
				memset(data, 0, sizeof(MATRIX_CONTROL));
				if (MatrixCInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, MATRIX_CREATE, "matrix: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(obj, &MatrixControlInstance, idlen, CONTROLINDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (MATRIX_CONTROL *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (MatrixCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, MATRIX_START, "matrix: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (MatrixCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, MATRIX_DESTROY, "matrix: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &MatrixControlInstance, idlen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				return SNMP_NOERROR;
			case SNMP_INVALID:
				if (MatrixCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, MATRIX_DESTROY, "matrix: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &MatrixControlInstance, idlen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (MatrixCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixSDSourceAddress(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	SRCDST *data = NULL;

	if ((local = MatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = MatrixSearchHash(local->Data, obj, idlen, TRUE)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->SrcDst, obj->SyntaxLen = ((MATRIX_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixSDDestAddress(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	SRCDST *data = NULL;

	if ((local = MatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = MatrixSearchHash(local->Data, obj, idlen, TRUE)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->SrcDst + ((MATRIX_CONTROL *) local->Data)->AddressLen, obj->SyntaxLen = ((MATRIX_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixSDIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	SRCDST *data = NULL;

	if ((local = MatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = MatrixSearchHash(local->Data, obj, idlen, TRUE)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixSDPkts(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	SRCDST *data = NULL;

	if ((local = MatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = MatrixSearchHash(local->Data, obj, idlen, TRUE)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixSDOctets(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	SRCDST *data = NULL;

	if ((local = MatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = MatrixSearchHash(local->Data, obj, idlen, TRUE)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixSDErrors(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	SRCDST *data = NULL;

	if ((local = MatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, TRUE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = MatrixSearchHash(local->Data, obj, idlen, TRUE)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Errors;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixDSSourceAddress(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	SRCDST *data = NULL;

	if ((local = MatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = MatrixSearchHash(local->Data, obj, idlen, FALSE)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->SrcDst, obj->SyntaxLen = ((MATRIX_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixDSDestAddress(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	SRCDST *data = NULL;

	if ((local = MatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = MatrixSearchHash(local->Data, obj, idlen, FALSE)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->SrcDst + ((MATRIX_CONTROL *) local->Data)->AddressLen, obj->SyntaxLen = ((MATRIX_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixDSIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	SRCDST *data = NULL;

	if ((local = MatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = MatrixSearchHash(local->Data, obj, idlen, TRUE)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixDSPkts(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	SRCDST *data = NULL;

	if ((local = MatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = MatrixSearchHash(local->Data, obj, idlen, FALSE)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixDSOctets(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	SRCDST *data = NULL;

	if ((local = MatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = MatrixSearchHash(local->Data, obj, idlen, FALSE)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD MatrixDSErrors(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	SRCDST *data = NULL;

	if ((local = MatrixMibRmon(obj, MatrixControlInstance, idlen)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, SRCDSTINDEXSIZE, &data, FALSE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = MatrixSearchHash(local->Data, obj, idlen, FALSE)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngUns = data->Errors;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

BOOLEAN MatrixRegisterAll(ULONG channelIndex)
{
	MIB_LOCAL *local;
	MATRIX_CONTROL *data;

	for (local = MatrixControlInstance; local != NULL; local = local->Next)
	{
		data = (MATRIX_CONTROL *) local->Data;
		if (data->Channel == TRUE && data->Source[data->SourceLen - 1] == channelIndex &&
			data->ChannelCallback == NULL)
		{
			data->ChannelCallback = ChannelRegisterSink(data->Source[data->SourceLen - 1],
														MatrixChannelPkt, MatrixDeRegister, data);
		}
	}

	return TRUE;
}
