/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <config.h>
#include <snmp.h>
#include <mibrqs.h>
#include <mibsup.h>
#include <message.h>
#include <mac.h>

#include "state.h"
#include "statc.h"
#include "stat.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define INDEXSIZE   1

static MIB_LOCAL *etherStats = NULL;

static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize);

BOOLEAN EtherStatsMInit(VOID)
{
	LONG i, nriface;
	CHAR s[81];
	BOOLEAN v;

	MAC_IFACE *iface;

	LONG statsource[] =
	{1, 3, 6, 1, 2, 1, 2, 2, 1, 1, 1};
	CHAR statowner[] = "monitorEtherStats";

#define STATUSINDEX     11
	SNMP_OBJECT statobj[] =
	{
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 1, 1, 1, 21, 1}, 12, SNMP_INTEGER},
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 1, 1, 1, 20, 1}, 12, SNMP_DISPLAYSTR},
		{SNMP_PDU_SET,
		 {1, 3, 6, 1, 2, 1, 16, 1, 1, 1, 2, 1}, 12, SNMP_OBJECTID}
	};
	WORD dummy;

	MessageConfig(STATISTICS_ERROR, "EtherStat");

	nriface = (LONG) MacIfaceCount();
	for (i = 1; i <= nriface; i++)
	{
		sprintf(s, "beholder.stat.iface.%ld", i);
		if (ConfigGetBoolean(s, &v) == TRUE)
		{
			iface = MacIfaceGet((WORD) i);
			if (iface != NULL && v == TRUE &&
				(iface->type == MAC_TYPE_ETHERNET_CSMACD ||
				 iface->type == MAC_TYPE_88023_CSMACD))
			{
				statobj[0].Id[STATUSINDEX] = i;
				statobj[1].Id[STATUSINDEX] = i;
				statobj[2].Id[STATUSINDEX] = i;

				statobj[0].Request = SNMP_PDU_SET;
				statobj[0].Syntax.LngInt = SNMP_INVALID;
				MibRequest(1, &statobj[0], &dummy);
				statobj[0].Request = SNMP_PDU_SET;
				statobj[0].Syntax.LngInt = SNMP_CREATEREQUEST;
				if (MibRequest(1, &statobj[0], &dummy) != SNMP_NOERROR)
				{
					DnpapMessage(DMC_WARNING, STAT_STATINIT, "etherStats: couldn't create etherStats collector %ld", i);
					continue;
				}
				statobj[1].Request = SNMP_PDU_SET;
				strncpy(statobj[1].Syntax.BufChr, statowner, statobj[1].SyntaxLen = strlen(statowner));
				statobj[2].Request = SNMP_PDU_SET;
				memcpy(statobj[2].Syntax.BufInt, statsource, sizeof(statsource));
				statobj[2].Syntax.BufInt[sizeof(statsource) / sizeof(statsource[0]) - 1] = i;
				statobj[2].SyntaxLen = sizeof(statsource) / sizeof(statsource[0]);
				if (MibRequest(2, &statobj[1], &dummy) != SNMP_NOERROR)
				{
					DnpapMessage(DMC_WARNING, STAT_STATINIT, "etherStats: couldn't set owner of etherStats collector %ld", i);
					statobj[0].Request = SNMP_PDU_SET;
					statobj[0].Syntax.LngInt = SNMP_INVALID;
					MibRequest(1, &statobj[0], &dummy);
					continue;
				}
				statobj[0].Request = SNMP_PDU_SET;
				statobj[0].Syntax.LngInt = SNMP_VALID;
				if (MibRequest(1, &statobj[0], &dummy) != SNMP_NOERROR)
				{
					DnpapMessage(DMC_WARNING, STAT_STATINIT, "etherStats: couldn't set etherStats collector %ld to valid", i);
					statobj[0].Request = SNMP_PDU_SET;
					statobj[0].Syntax.LngInt = SNMP_INVALID;
					MibRequest(1, &statobj[0], &dummy);
					continue;
				}
			}
		}
	}

	return TRUE;
}

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize)
{
	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->IdLen == idlen || (*local)->Index > obj->Id[idlen] ||
		(*local = (*local)->Next) != NULL)
	{
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + indexsize;
		return TRUE;
	}

	return FALSE;
}

WORD EtherStatsIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;
	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsDataSource(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;
	MAC_IFACE *iface;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, STAT_BOOLEAN_DATASOURCE))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufInt, data->Source, data->SourceLen * sizeof(LONG));
		Obj->SyntaxLen = data->SourceLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (ETHER_STATS *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (data->SourceLen != Obj->SyntaxLen ||
			memcmp(data->Source, Obj->Syntax.BufInt, (data->SourceLen - 1) * sizeof(LONG)))
			return SNMP_BADVALUE;
		if ((iface = MacIfaceGet((WORD) Obj->Syntax.BufInt[Obj->SyntaxLen - 1])) == NULL)
			return SNMP_BADVALUE;
		if (iface->type != MAC_TYPE_ETHERNET_CSMACD && iface->type != MAC_TYPE_88023_CSMACD)
			return SNMP_BADVALUE;
		memcpy(data->Source, Obj->Syntax.BufInt, Obj->SyntaxLen * sizeof(LONG));
		data->SourceLen = Obj->SyntaxLen;
		data->Iface = iface;
		BooleanSetTrue(data->ObjectSet, STAT_BOOLEAN_DATASOURCE);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsDropEvents(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.LostPkts + stat.DiscardedPkts - (data->BaseStat.LostPkts + data->BaseStat.DiscardedPkts);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsOctets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
WORD EtherStatsPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsBroadcastPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.BroadcastPkts - data->BaseStat.BroadcastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsMulticastPkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.MulticastPkts - data->BaseStat.MulticastPkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsCRCAlignErrors(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.CRCAlignErrors - data->BaseStat.CRCAlignErrors;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsUndersizePkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.UndersizePkts - data->BaseStat.UndersizePkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsOversizePkts(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.OversizePkts - data->BaseStat.OversizePkts;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsFragments(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.Fragments - data->BaseStat.Fragments;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsJabbers(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.Jabbers - data->BaseStat.Jabbers;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsCollisions(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;
	MAC_STAT stat;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		if (data->Status != SNMP_VALID)
		{
			Obj->Syntax.LngUns = 0UL;
			return SNMP_NOERROR;
		}
		MacStatistics(data->Iface, &stat);
		Obj->Syntax.LngUns = stat.Collisions - data->BaseStat.Collisions;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsPkts64Octets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts64Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsPkts65to127Octets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts65to127Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsPkts128to255Octets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts128to255Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsPkts256to511Octets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts256to511Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsPkts512to1023Octets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts512to1023Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsPkts1024to1518Octets(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		Obj->Syntax.LngUns = data->Pkts1024to1518Octets;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsOwner(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;

	if ((local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (ETHER_STATS *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, STAT_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		Obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (ETHER_STATS *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, Obj->Syntax.BufChr, Obj->SyntaxLen);
		data->OwnerLen = Obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, STAT_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD EtherStatsStatus(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	ETHER_STATS *data;

	local = MibRmon(Obj, etherStats, IdLen, INDEXSIZE);

	switch (Obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (ETHER_STATS *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (ETHER_STATS *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(Obj, &etherStats, IdLen, INDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(ETHER_STATS))) == NULL)
				{
					MibRemove(Obj, &etherStats, IdLen, INDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (ETHER_STATS *) local->Data;
				memset(data, 0, sizeof(ETHER_STATS));
				if (EtherStatsCInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, STAT_CREATE, "etherStats: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(Obj, &etherStats, IdLen, INDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (ETHER_STATS *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (EtherStatsCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, STAT_ACTIVE, "etherStats: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				else
					return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (EtherStatsCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, STAT_DESTROY, "etherStats: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &etherStats, IdLen, INDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_INVALID:
				if (EtherStatsCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, STAT_DESTROY, "etherStats: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &etherStats, IdLen, INDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (EtherStatsCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			case SNMP_VALID:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
