/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <limits.h>
#include <dnpap.h>
#include <dmath.h>
#include <memory.h>
#include <maxmem.h>
#include <mibsup.h>
#include <stdlib.h>
#include <sysmib.h>
#include <timer.h>
#include <host.h>

#include "topn_d.h"
#include "topn_e.h"
#include "topn_c.h"

IMPORT BOOLEAN BeholderStrictRMON;

static LONG TopNMaxNrEntries = 100;

static BOOLEAN TopNEntryReset(TOPN_CONTROL *topncontrol, ULONG hostcontrolindex, HOST * host, WORD addrlen);
static BOOLEAN TopNEntryUpdate(TOPN_CONTROL *topncontrol, ULONG hostcontrolindex, HOST * host, WORD addrlen, int flags, LWORD size);
static BOOLEAN TopNHostStop(TOPN_CONTROL *topncontrol, ULONG hostcontrolindex);

static VOID TopNTimerCallback(TIMER_DESCR * timer, ULONG now, VOID * param);

static VOID DelEntries(TOPN_CONTROL * topncontrol);


BOOLEAN TopNCInit(TOPN_CONTROL * topncontrol)
{
	topncontrol->HostIndex = 1;
	topncontrol->HostCallback = NULL;
	topncontrol->AddressLen = 0;
	topncontrol->RateBase = 1;
	topncontrol->TimeRemaining = 0;
	topncontrol->Duration = 0;
	topncontrol->RequestedSize = 10;
	topncontrol->GrantedSize = TopNGranted(topncontrol->RequestedSize);
	topncontrol->StartTime = 0;
	topncontrol->Owner[0] = '\0';
	topncontrol->OwnerLen = 0;
	topncontrol->Status = SNMP_INVALID;

	topncontrol->TableSize = 0;
	topncontrol->Table = NULL;

	topncontrol->TopNEntries = NULL;
	topncontrol->TopNLast = NULL;
	topncontrol->TopNArray = NULL;

	topncontrol->Timer = NULL;

	BooleanSetAllTrue(topncontrol->ObjectSet);
	BooleanSetFalse(topncontrol->ObjectSet, TOPN_BOOLEAN_DATASOURCE);
	BooleanSetFalse(topncontrol->ObjectSet, TOPN_BOOLEAN_OWNER);

	return TRUE;
}

BOOLEAN TopNCStart(TOPN_CONTROL * topncontrol)
{
	if (BeholderStrictRMON && !BooleanCheckAllTrue(topncontrol->ObjectSet))
		return FALSE;

	if (topncontrol->TimeRemaining > 0)
	{
		topncontrol->HostCallback = HostRegisterTopN(topncontrol->HostIndex, HostTopNNotify, topncontrol);
		
		if ((topncontrol->Table = NewHash(TopNMaxNrEntries / 2 + 11, NULL)) == NULL)
		{
			DnpapMessage(DMC_ERROR, TOPN_TIMER, "topN: can not create hashtable");
        	HostDeRegisterTopN(topncontrol->HostIndex, topncontrol->HostCallback);
        	topncontrol->HostCallback = NULL;
			return FALSE;
		}

		if ((topncontrol->Timer = TimerRegister(TopNTimerCallback, topncontrol, 1000, TIMER_FOREVER, TIMER_TYPE_RECOVER)) == NULL)
		{
			DnpapMessage(DMC_ERROR, TOPN_TIMER, "topN: can not register a timer");
        	HostDeRegisterTopN(topncontrol->HostIndex, topncontrol->HostCallback);
        	topncontrol->HostCallback = NULL;
			return FALSE;
		}

		topncontrol->GrantedSize = TopNGranted(topncontrol->RequestedSize);
		
		topncontrol->AddressLen = 0;
        
		topncontrol->StartTime = SysTime();
	}

	return TRUE;
}

BOOLEAN TopNCStop(TOPN_CONTROL * topncontrol)
{
	HostDeRegisterTopN(topncontrol->HostIndex, topncontrol->HostCallback);
	topncontrol->HostCallback = NULL;
	
	TimerRemove(topncontrol->Timer);
	topncontrol->Timer = NULL;
	
	DelEntries(topncontrol);

	DelHash(topncontrol->Table);
	topncontrol->TableSize = 0;
	topncontrol->Table = NULL;

	DnpapFree(topncontrol->TopNArray);
	topncontrol->TopNEntries = NULL;
	topncontrol->TopNLast = NULL;
	topncontrol->TopNArray = NULL;

	topncontrol->TimeRemaining = 0;
	topncontrol->Duration = 0;
	topncontrol->StartTime = 0;

	topncontrol->GrantedSize = TopNGranted(topncontrol->RequestedSize);
	
	return TRUE;
}
 

LONG TopNGranted(LONG requested)
{
FLOAT factor;

	factor = (FLOAT)LongMin(MemoryGetMaxChunk(), MemoryGetAvail()) / 
				(100*requested*sizeof(TOPN_ENTRY));
				
	if (factor >= 1.0)
		return requested;
	return factor*requested;
}


VOID TopNTimerCallback(TIMER_DESCR * timer, ULONG now, VOID * param)
{
	TOPN_CONTROL *topncontrol = param;

	topncontrol->TimeRemaining--;
	if (topncontrol->TimeRemaining <= 0)
	{
		if (FillEntryTable(topncontrol, topncontrol->GrantedSize) != TRUE)
			DnpapMessage(DMC_WARNING, TOPN_FILLFAIL, "topN: creation of entry array failed");
		HostDeRegisterTopN(topncontrol->HostIndex, topncontrol->HostCallback);
		topncontrol->HostCallback = NULL;
		TimerRemove(topncontrol->Timer);
		topncontrol->Timer = NULL;
	}
}

VOID HostTopNNotify(ULONG hostindex, int event, HOST *host, WORD addrlen, int flags, LWORD size, VOID *topn)
{
	switch (event)
	{
	case HOST_EVENT_RESET:
		TopNEntryReset(topn, hostindex, host, addrlen);
		break;
	case HOST_EVENT_UPDATE:
		TopNEntryUpdate(topn, hostindex, host, addrlen, flags, size);
		break;
	case HOST_EVENT_STOP:
		TopNHostStop(topn, hostindex);
		break;
	}
}

BOOLEAN TopNEntryReset(TOPN_CONTROL *topncontrol, ULONG hostcontrolindex, HOST * host, WORD addrlen)
{
	TOPN_ENTRY *entry;

	if (topncontrol->Status == SNMP_VALID && topncontrol->HostIndex == hostcontrolindex &&
		topncontrol->TimeRemaining > 0)
	{
		if (topncontrol->AddressLen == 0)
			topncontrol->AddressLen = addrlen;
		else if (topncontrol->AddressLen != addrlen)
		{
			DnpapMessage(DMC_FATAL, TOPN_INCONSISTENT, "topncontrol: inconsistent address length encountered");
			DnpapExit(1);
		}

		if ((entry = HashSearch(topncontrol->Table, host->Address, addrlen)) != NULL)
			entry->PrevValue = 0;
	}
	return TRUE;
}

BOOLEAN TopNEntryUpdate(TOPN_CONTROL *topncontrol, ULONG hostcontrolindex, HOST * host, WORD addrlen, int flags, LWORD size)
{
	TOPN_ENTRY *entry;

	if (topncontrol->Status == SNMP_VALID && topncontrol->HostIndex == hostcontrolindex &&
		topncontrol->TimeRemaining > 0)
	{
		if (topncontrol->AddressLen == 0)
			topncontrol->AddressLen = addrlen;
		else if (topncontrol->AddressLen != addrlen)
		{
			DnpapMessage(DMC_FATAL, TOPN_INCONSISTENT, "topncontrol: inconsistent address length encountered");
			DnpapExit(1);
		}

		switch (topncontrol->RateBase)
		{
		case TOPN_RATE_INPKTS:
		case TOPN_RATE_INOCTETS:
			if ((flags & HOST_FLAG_IN) == 0)
				return TRUE;
			break;
		case TOPN_RATE_OUTPKTS:
		case TOPN_RATE_OUTOCTETS:
		case TOPN_RATE_OUTERRORS:
		case TOPN_RATE_OUTBROADCASTPKTS:
		case TOPN_RATE_OUTMULTICASTPKTS:
			if ((flags & HOST_FLAG_OUT) == 0)
				return TRUE;
			break;
		}
		
		switch (topncontrol->RateBase)
		{
		case TOPN_RATE_OUTERRORS:
			if ((flags & HOST_FLAG_ERROR) == 0)
				return TRUE;			
			break;
		case TOPN_RATE_OUTBROADCASTPKTS:
			if ((flags & HOST_FLAG_BROADCAST) == 0)
				return TRUE;			
			break;
		case TOPN_RATE_OUTMULTICASTPKTS:
			if ((flags & HOST_FLAG_MULTICAST) == 0)
				return TRUE;
			break;
		}
		
		if ((entry = HashSearch(topncontrol->Table, host->Address, addrlen)) == NULL)
		{
			if ((entry = DnpapMalloc(sizeof(TOPN_ENTRY))) == NULL)
				return FALSE;
			else
			{
				memset(entry, 0, sizeof(TOPN_ENTRY));
				memcpy(entry->Address, host->Address, addrlen);

				entry->PrevValue = 0;
				switch (topncontrol->RateBase)
				{
				case TOPN_RATE_INPKTS:
				case TOPN_RATE_OUTPKTS:
				case TOPN_RATE_OUTERRORS:
				case TOPN_RATE_OUTBROADCASTPKTS:
				case TOPN_RATE_OUTMULTICASTPKTS:
					entry->Rate = 1;
					break;
				case TOPN_RATE_INOCTETS:
					entry->Rate = size;
					entry->PrevValue = host->InOctets;
					break;
				case TOPN_RATE_OUTOCTETS:
					entry->Rate = size;
					entry->PrevValue = host->OutOctets;
					break;
				}

				if (HashAdd(topncontrol->Table, entry->Address, addrlen, entry) == NULL)
				{
					DnpapFree(entry);
					return FALSE;
				}
				else
				{
					EntryAddList(topncontrol, entry);
					topncontrol->TableSize++;
				}
			}
		}
		else
		{
			switch (topncontrol->RateBase)
			{
			case TOPN_RATE_INPKTS:
			case TOPN_RATE_OUTPKTS:
			case TOPN_RATE_OUTERRORS:
			case TOPN_RATE_OUTBROADCASTPKTS:
			case TOPN_RATE_OUTMULTICASTPKTS:
				entry->Rate++;
				break;
			case TOPN_RATE_INOCTETS:
				if (entry->PrevValue > host->InOctets)
					entry->Rate += host->InOctets + (ULONG_MAX - entry->PrevValue);
				else
					entry->Rate += host->InOctets - entry->PrevValue;
				entry->PrevValue = host->InOctets;
				break;
			case TOPN_RATE_OUTOCTETS:
				if (entry->PrevValue > host->OutOctets)
					entry->Rate += host->OutOctets + (ULONG_MAX - entry->PrevValue);
				else
					entry->Rate += host->OutOctets - entry->PrevValue;
				entry->PrevValue = host->OutOctets;
				break;
			}
			EntryUpdateList(topncontrol, entry);
		}
	}
	
	return TRUE;
}

BOOLEAN TopNHostStop(TOPN_CONTROL *topncontrol, ULONG hostcontrolindex)
{
	if (topncontrol->Status == SNMP_VALID && topncontrol->HostIndex == hostcontrolindex)
	{
		if (TopNCStop(topncontrol) == FALSE)
			return FALSE;
	}
	return TRUE;
}

TOPN_ENTRY *TopNEntrySearch(TOPN_CONTROL * topncontrol, SNMP_OBJECT * obj, WORD idlen)
{
	if (topncontrol->Status == SNMP_VALID && topncontrol->TimeRemaining == 0 &&
		topncontrol->TopNArray != NULL)
		return topncontrol->TopNArray[obj->Id[idlen + 1] - 1];
	return NULL;
}

BOOLEAN FillEntryTable(TOPN_CONTROL * topncontrol, LONG reqsize)
{
	TOPN_ENTRY *entry;
	TOPN_ENTRY **array;
	LONG count;
	LONG arraysize = min(reqsize, topncontrol->TableSize);

	if (arraysize == 0 && topncontrol->Status == SNMP_VALID)
	{
		/*
		DnpapMessage(DMC_MESSAGE, TOPN_ARRAY, "topN: table is empty");
		 */
		return TRUE;
	}

	if ((array = (TOPN_ENTRY **) DnpapMalloc(arraysize * sizeof(TOPN_ENTRY *))) == NULL)
	{
		DnpapMessage(DMC_ERROR, TOPN_ARRAY, "topN: can not create entry array of size %ld", arraysize);
		return FALSE;
	}
	DnpapFree(topncontrol->TopNArray);
	memset(topncontrol->TopNArray = array, 0, (size_t) arraysize * sizeof(TOPN_ENTRY *));

	for (entry = topncontrol->TopNEntries, count = 0;
		 entry != NULL && count < arraysize;
		 entry = entry->Next, count++)
		topncontrol->TopNArray[count] = entry;

	return TRUE;
}

VOID DelEntries(TOPN_CONTROL * topncontrol)
{
	TOPN_ENTRY *entry1, *entry2;

	for (entry2 = topncontrol->TopNEntries; entry2 != NULL; entry2 = entry1)
	{
		entry1 = entry2->Next;
		DnpapFree(entry2);
		topncontrol->TableSize--;
	}
	topncontrol->TopNEntries = NULL;
	topncontrol->TopNLast = NULL;
}

BOOLEAN EntryAddList(TOPN_CONTROL * topncontrol, TOPN_ENTRY * entry)
{
	if (topncontrol->TopNEntries == NULL)
	{
		topncontrol->TopNEntries = entry;
		topncontrol->TopNLast = entry;
		entry->Prev = NULL;
		entry->Next = NULL;
		return TRUE;
	}

	entry->Prev = topncontrol->TopNLast;
	entry->Next = NULL;
	topncontrol->TopNLast->Next = entry;
	topncontrol->TopNLast = entry;

	return EntryUpdateList(topncontrol, entry);
}

BOOLEAN EntryUpdateList(TOPN_CONTROL * topncontrol, TOPN_ENTRY * entry)
{
	TOPN_ENTRY *p;
	BOOLEAN update = FALSE;

	if (topncontrol->TopNEntries == entry)
		return TRUE;

	for (p = entry->Prev; p != NULL; p = p->Prev)
	{
		if (p->Rate >= entry->Rate)
			break;
		update = TRUE;
	}

	if (update == FALSE)
		return TRUE;

	entry->Prev->Next = entry->Next;
	if (entry->Next != NULL)
		entry->Next->Prev = entry->Prev;
	else
		topncontrol->TopNLast = entry->Prev;

	if (p != NULL)
	{
		entry->Prev = p;
		entry->Next = p->Next;
		p->Next->Prev = entry;
		p->Next = entry;
	}
	else
	{
		entry->Prev = p;
		entry->Next = topncontrol->TopNEntries;
		topncontrol->TopNEntries->Prev = entry;
		topncontrol->TopNEntries = entry;
	}

	return TRUE;
}
