/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <limits.h>
#include <dnpap.h>
#include <dmath.h>
#include <stdlib.h>
#include <memory.h>
#include <message.h>
#include <mibsup.h>
#include <host.h>

#include "topn_d.h"
#include "topn_e.h"
#include "topn_c.h"
#include "topn.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define CONTROLINDEXSIZE   1
#define TOPNINDEXSIZE      2

static MIB_LOCAL *TopNControlInstance = NULL;

static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, TOPN_ENTRY ** elem);

BOOLEAN TopNMInit(VOID)
{
	MessageConfig(TOPN_ERROR, "TopN");
	return TRUE;
}

#define CONTROL_DATA    ((TOPN_CONTROL*)(*local)->Data)

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize, TOPN_ENTRY ** elem)
{
	BOOLEAN first = FALSE;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case TOPNINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			if (CONTROL_DATA->Status == SNMP_VALID &&
				CONTROL_DATA->TimeRemaining == 0 &&
				CONTROL_DATA->TableSize > 0 &&
				CONTROL_DATA->TopNArray != NULL &&
				obj->Id[idlen + 1] < min(CONTROL_DATA->RequestedSize, CONTROL_DATA->TableSize))
			{
				obj->Id[idlen] = (*local)->Index;
				obj->Id[idlen + 1]++;
				obj->IdLen = idlen + TOPNINDEXSIZE;
				*elem = CONTROL_DATA->TopNArray[obj->Id[idlen + 1] - 1];
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL &&
			   (CONTROL_DATA->Status != SNMP_VALID ||
				CONTROL_DATA->TimeRemaining > 0 ||
				CONTROL_DATA->TableSize == 0 ||
				CONTROL_DATA->TopNArray == NULL))
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen + 1] = 1;
		obj->IdLen = idlen + TOPNINDEXSIZE;
		*elem = CONTROL_DATA->TopNArray[0];
		break;
	case CONTROLINDEXSIZE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + CONTROLINDEXSIZE;
		break;
	}
	return TRUE;
}

WORD HostTopNControlIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNHostIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_CONTROL *data;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (TOPN_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->HostIndex;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (TOPN_CONTROL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		data->HostIndex = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNRateBase(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_CONTROL *data;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (TOPN_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->RateBase;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (TOPN_CONTROL *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt < TOPN_RATE_BASEMIN || obj->Syntax.LngInt > TOPN_RATE_BASEMAX)
			return SNMP_BADVALUE;
		data->RateBase = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNTimeRemaining(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_CONTROL *data;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (TOPN_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->TimeRemaining;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (TOPN_CONTROL *) local->Data;
		if (obj->Syntax.LngInt < 1)
			return SNMP_BADVALUE;
		if (data->Status == SNMP_VALID)
		{
			if (TopNCStop(data) == FALSE)
			{
				DnpapMessage(DMC_ERROR, TOPN_STOP, "topN: collector %ld could not be stopped", local->Index);
				return SNMP_GENERROR;
			}
			/* this must come before TopNCStart() */
			data->TimeRemaining = obj->Syntax.LngInt;
			data->Duration = data->TimeRemaining;
			if (TopNCStart(data) == FALSE)
			{
				DnpapMessage(DMC_ERROR, TOPN_START, "topN: collector %ld could not be started", local->Index);
				return SNMP_BADVALUE;
			}
		}
		else
		{
			data->TimeRemaining = obj->Syntax.LngInt;
			data->Duration = data->TimeRemaining;
		}
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNDuration(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_CONTROL *data;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (TOPN_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Duration;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNRequestedSize(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_CONTROL *data;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (TOPN_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->RequestedSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (TOPN_CONTROL *) local->Data;
		if (obj->Syntax.LngInt < 0)
			return SNMP_BADVALUE;
		if (FillEntryTable(data, obj->Syntax.LngInt) != TRUE)
			return SNMP_GENERROR;
		data->RequestedSize = obj->Syntax.LngInt;
		data->GrantedSize = TopNGranted(data->RequestedSize);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNGrantedSize(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_CONTROL *data;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (TOPN_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->GrantedSize;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNStartTime(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_CONTROL *data;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (TOPN_CONTROL *) local->Data;
		obj->Syntax.LngUns = data->StartTime;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNOwner(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_CONTROL *data;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (TOPN_CONTROL *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, TOPN_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (TOPN_CONTROL *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, obj->Syntax.BufChr, obj->SyntaxLen);
		data->OwnerLen = obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, TOPN_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNStatus(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_CONTROL *data;

	local = MibRmon(obj, TopNControlInstance, idlen, CONTROLINDEXSIZE);

	switch (obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (TOPN_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, CONTROLINDEXSIZE, NULL) == FALSE)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (TOPN_CONTROL *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(obj, &TopNControlInstance, idlen, CONTROLINDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(TOPN_CONTROL))) == NULL)
				{
					MibRemove(obj, &TopNControlInstance, idlen, CONTROLINDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (TOPN_CONTROL *) local->Data;
				if (TopNCInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, TOPN_CREATE, "topN: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(obj, &TopNControlInstance, idlen, CONTROLINDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (TOPN_CONTROL *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (TopNCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, TOPN_START, "topN: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (TopNCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, TOPN_DESTROY, "topN: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &TopNControlInstance, idlen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				return SNMP_NOERROR;
			case SNMP_INVALID:
				if (TopNCStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, TOPN_DESTROY, "topN: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &TopNControlInstance, idlen, CONTROLINDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (TopNCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNReport(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_ENTRY *data = NULL;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, TOPNINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, TOPNINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = TopNEntrySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNIndex(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_ENTRY *data = NULL;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, TOPNINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, TOPNINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = TopNEntrySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = obj->Id[idlen + 1];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNAddress(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_ENTRY *data = NULL;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, TOPNINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, TOPNINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = TopNEntrySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Address, obj->SyntaxLen = ((TOPN_CONTROL *) local->Data)->AddressLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

WORD HostTopNRate(SNMP_OBJECT * obj, WORD idlen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	TOPN_ENTRY *data = NULL;

	if ((local = MibRmon(obj, TopNControlInstance, idlen, TOPNINDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(obj, &local, idlen, TOPNINDEXSIZE, &data) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		if (data == NULL && (data = TopNEntrySearch(local->Data, obj, idlen)) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = data->Rate;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}

BOOLEAN HostTopNRegisterAll(ULONG hostindex)
{
	MIB_LOCAL *local;
	TOPN_CONTROL *data;

	for (local = TopNControlInstance; local != NULL; local = local->Next)
	{
		data = (TOPN_CONTROL *) local->Data;
		if (data->HostIndex == hostindex && data->HostCallback == NULL)
		{
			data->HostCallback = HostRegisterTopN(data->HostIndex, HostTopNNotify, data);
		}
	}

	return TRUE;
}
