/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <stdio.h>
#include <string.h>
#include <dnpap.h>
#include <mibsup.h>
#include <message.h>
#include <channel.h>

#include "filter2e.h"
#include "filter2c.h"

IMPORT BOOLEAN BeholderStrictRMON;



BOOLEAN Filter2CInit(FILTER2 *filter)
{
    filter->ChannelIndex = 1;
    filter->ChannelCallback = NULL;

    filter->Program.bf_insns = 0;
    filter->Program.bf_len = 0;
                        
	BooleanSetAllTrue(filter->ObjectSet);
	BooleanSetFalse(filter->ObjectSet, FILTER2_BOOLEAN_CHANNEL);
	BooleanSetFalse(filter->ObjectSet, FILTER2_BOOLEAN_OWNER);

    return TRUE;
}


BOOLEAN Filter2CStart(FILTER2 *filter)
{
	if (BeholderStrictRMON && !BooleanCheckAllTrue(filter->ObjectSet))
		return FALSE;

    filter->ChannelCallback = ChannelRegisterFilter(filter->ChannelIndex, Filter2Pkt, Filter2DeRegister, filter);
        
    return TRUE;
}


BOOLEAN Filter2CStop(FILTER2 *filter)
{       
    ChannelDeRegisterFilter(filter->ChannelIndex, filter->ChannelCallback);
	filter->ChannelCallback = NULL;
	
    DnpapFree(filter->Program.bf_insns);
	filter->Program.bf_len = NULL;
	
    return TRUE;
}


BOOLEAN Filter2Pkt(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *fptr)
{
FILTER2 *filter = (FILTER2 *) fptr;
PROT_OBJ len =		{0, {1,4}};
PROT_OBJ caplen =	{0, {1,5}};
PROT_OBJ packet =	{0, {1,7}};

    if (filter->Status == SNMP_VALID && filter->ChannelIndex == channelIndex)
    {
        if (ProtGetField(pkt, &len) == FALSE ||
        	ProtGetField(pkt, &caplen) == FALSE ||
            ProtGetField(pkt, &packet) == FALSE) 
        {
            DnpapMessage(DMC_ERROR, FILTER2_BADPACKET,"filter: bad packet");
            return FALSE;
        }
    
        if (filter->Program.bf_len == 0 || filter->Program.bf_insns == NULL)
            return TRUE;	    /*  no data, match everything */

		if (bpf_filter(filter->Program.bf_insns, (BYTE *)packet.Syntax.Ptr,
						len.Syntax.LngInt, caplen.Syntax.LngInt))
			return TRUE;
    }

    return FALSE;
}


VOID Filter2DeRegister(ULONG channelIndex, VOID *fptr)
{
FILTER2 *filter = (FILTER2 *) fptr;

    if (filter->ChannelIndex == channelIndex)
    {
		ChannelDeRegisterFilter(channelIndex, filter->ChannelCallback);
		filter->ChannelCallback = NULL;
	}
}
