/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <memory.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mibsup.h>
#include <message.h>
#include <protocol.h>
#include <channel.h>

#include "filter2e.h"
#include "filter2c.h"
#include "filter2.h"

IMPORT BOOLEAN BeholderStrictRMON;

#define INDEXSIZE   1

static MIB_LOCAL *Filter2Instance = NULL;

static BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize);

BOOLEAN Filter2MInit(VOID)
{
	MessageConfig(FILTER2_ERROR, "Filter2");
	return TRUE;
}

BOOLEAN RmonNext(SNMP_OBJECT * obj, MIB_LOCAL ** local, WORD idlen, WORD indexsize)
{
	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->IdLen == idlen || (*local)->Index > obj->Id[idlen] ||
		(*local = (*local)->Next) != NULL)
	{
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + indexsize;
		return TRUE;
	}

	return FALSE;
}

WORD filter2Index(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(Obj, Filter2Instance, IdLen, INDEXSIZE)) == NULL)
		return (SNMP_NOSUCHNAME);

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		Obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	}
	return SNMP_GENERROR;
}

WORD filter2ChannelIndex(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER2 *data;

	if ((local = MibRmon(Obj, Filter2Instance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (FILTER2 *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, FILTER2_BOOLEAN_CHANNEL))
			return SNMP_NOSUCHNAME;
		Obj->Syntax.LngInt = data->ChannelIndex;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (FILTER2 *) local->Data;
		data->ChannelIndex = Obj->Syntax.LngInt;
		BooleanSetTrue(data->ObjectSet, FILTER2_BOOLEAN_CHANNEL);
		return SNMP_NOERROR;
	}
	return SNMP_GENERROR;
}

WORD filter2Program(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER2 *data;
	INSN *datap;

	if ((local = MibRmon(Obj, Filter2Instance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (FILTER2 *) local->Data;
		memcpy(Obj->Syntax.BufChr, data->Program.bf_insns, Obj->SyntaxLen = data->Program.bf_len * sizeof(INSN));
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (FILTER2 *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (Obj->SyntaxLen > SNMP_SIZE_BUFCHR || Obj->SyntaxLen == 0 ||
			(Obj->SyntaxLen % sizeof(INSN)) != 0)
			return SNMP_BADVALUE;
		if ((datap = DnpapRealloc(data->Program.bf_insns, Obj->SyntaxLen)) == NULL)
			return SNMP_GENERROR;
		data->Program.bf_insns = datap;
		data->Program.bf_len = Obj->SyntaxLen / sizeof(INSN);
		memcpy(data->Program.bf_insns, Obj->Syntax.BufChr, Obj->SyntaxLen);
		return SNMP_NOERROR;
	}
	return SNMP_GENERROR;
}

WORD filter2Owner(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER2 *data;

	if ((local = MibRmon(Obj, Filter2Instance, IdLen, INDEXSIZE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (Obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (FILTER2 *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, FILTER2_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(Obj->Syntax.BufChr, data->Owner, data->OwnerLen);
		Obj->SyntaxLen = data->OwnerLen;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (FILTER2 *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, Obj->Syntax.BufChr, Obj->SyntaxLen);
		data->OwnerLen = Obj->SyntaxLen;
		BooleanSetTrue(data->ObjectSet, FILTER2_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	}
	return SNMP_GENERROR;
}

WORD filter2Status(SNMP_OBJECT * Obj, WORD IdLen, VOID * param)
{
	MIB_LOCAL *local = NULL;
	FILTER2 *data;

	local = MibRmon(Obj, Filter2Instance, IdLen, INDEXSIZE);

	switch (Obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (FILTER2 *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (RmonNext(Obj, &local, IdLen, INDEXSIZE) == FALSE)
		{
			Obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (FILTER2 *) local->Data;
		Obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(Obj, &Filter2Instance, IdLen, INDEXSIZE)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = DnpapMalloc(sizeof(FILTER2))) == NULL)
				{
					MibRemove(Obj, &Filter2Instance, IdLen, INDEXSIZE);
					return SNMP_GENERROR;
				}
				data = (FILTER2 *) local->Data;
				memset(data, 0, sizeof(FILTER2));
				if (Filter2CInit(data) == TRUE)
				{
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, FILTER2_CREATE, "filter: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				DnpapFree(local->Data);
				MibRemove(Obj, &Filter2Instance, IdLen, INDEXSIZE);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (FILTER2 *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (Filter2CStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, FILTER2_ACTIVE, "filter: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (Filter2CStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, FILTER2_DESTROY, "filter: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &Filter2Instance, IdLen, INDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (Obj->Syntax.LngInt)
			{
			case SNMP_INVALID:
				if (Filter2CStop(data) == TRUE)
				{
					DnpapFree(local->Data);
					DnpapMessage(DMC_MESSAGE, FILTER2_DESTROY, "filter: collector %ld destroyed", local->Index);
					if (MibRemove(Obj, &Filter2Instance, IdLen, INDEXSIZE) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (Filter2CStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			case SNMP_VALID:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	}
	return SNMP_GENERROR;
}

BOOLEAN Filter2RegisterAll(ULONG channelIndex)
{
	MIB_LOCAL *local;
	FILTER2 *filter;

	for (local = Filter2Instance; local != NULL; local = local->Next)
	{
		filter = (FILTER2 *) local->Data;
		if (filter->ChannelIndex == channelIndex && filter->ChannelCallback == NULL)
		{
			filter->ChannelCallback = ChannelRegisterFilter(filter->ChannelIndex, Filter2Pkt, Filter2DeRegister, filter);
		}
	}

	return TRUE;
}
