/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <stdio.h>
#include <dnpap.h>
#include <memory.h>
#include <string.h>
#include <message.h>
#include <config.h>   
#include <dkvm.h>
#include <mac.h>
#include <chain.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <net/if.h>
#include <bpf.h>

#include "bbpf.h"
 

extern Kvm *kvm;


typedef struct ifnet 	IFNET;

typedef struct bpf_hdr BPFHDR;


BPF_DESCR *bpfDescrList = 0;


static void    Rcve(NW_DG *nw, char *addr, BYTE *frame, int length, void *parm);
static BOOLEAN Send(MAC_IFACE *iface, CHAIN *chain);
static BOOLEAN Stat(MAC_IFACE *iface, MAC_STAT *stats);

static BOOLEAN GetIfnetOffset(CHAR *name, ULONG *offset);
static BOOLEAN GetIfnet(ULONG offset, IFNET *ifnet);

static char *Iface(char *descr);


void BpfExit(void)
{
    BPF_DESCR      *bpf;
    
    while (bpfDescrList!=0)
    {
        bpf = bpfDescrList;
        bpfDescrList = bpfDescrList->next;
                    
        MacIfaceRemove(bpf->mac);
        NwDgClose(bpf->nw);
        DnpapFree(bpf);
    }
}


BOOLEAN BpfInit(void)
{
    BYTE            name[255];
    BYTE            addr[255];
    CHAR            *descr;
    MAC_IFACE       *mac;
    BPF_DESCR       *bpf; 
    WORD            i;
    ULONG			offset = 0;
	IFNET			ifnet;
	NW_IFCONFIG		ifconfig;
    static BOOLEAN  init = FALSE;
    
    if (!init)
    {
		MessageConfig(BPF_ERROR, "Bpf");

        i = 1;
        sprintf(name, "bpf.iface.name.%d", i);
        while (ConfigGetString(name, &descr))
        {
            mac = DnpapMalloc(sizeof(MAC_IFACE));
            if (mac!=0)
            {                        
                bpf=DnpapMalloc(sizeof(BPF_DESCR)); 
                memset(bpf, 0, sizeof(BPF_DESCR));
                if (bpf!=0)
                {
                    bpf->mac=mac;
 
#ifdef WORKING_BIOCSBLEN
                    sprintf(addr,"bpf:%s:8192", descr);
#else
                    sprintf(addr,"bpf:%s:1016", descr);
#endif
                    bpf->nw=NwDgAccept(addr, Rcve, bpf);  
                    if (bpf->nw != NULL && NwDgIfConfig(bpf->nw, &ifconfig) == TRUE)
                    {
						if (GetIfnetOffset(descr, &offset) == FALSE || GetIfnet(offset, &ifnet) == FALSE)
							DnpapMessage(DMC_FATAL, BPF_ERROR, "bpf: interface not found in kernel memory");

						bpf->offset        = offset;

                        mac->specific      = bpf;
                        sprintf(mac->descr, "bpf.%s", descr);
                        mac->Send          = Send;
                        mac->Stat          = Stat;
                        mac->type          = ifconfig.type;
                        mac->speed         = ifconfig.speed;
                        mac->mtu           = ifconfig.mtu;
                        mac->statusOper    = TRUE;
                        mac->statusAdmin   = TRUE;
                        mac->arp           = MAC_ARP_NONE;  
                        memcpy(mac->addr, ifconfig.address, ifconfig.addrlen);
                        memcpy(mac->addrBroadcast, ifconfig.broadcast, ifconfig.addrlen);
						mac->addrLength    = ifconfig.addrlen;
						mac->frameId       = 0;
                        MacIfaceRegister(mac); 
                    }
					else
					{
						DnpapMessage(DMC_ERROR, BPF_ERROR, "bpf: can not open %s", descr);
					}
                    
                    bpf->next=bpfDescrList;
                    bpfDescrList=bpf;
                }
            }    
            
            i++;
            sprintf(name, "bpf.iface.name.%d", i);
        }

        DnpapAtExit(BpfExit);

        init = TRUE;
    }

    return init;
}


static void Rcve(NW_DG *nw, char *addr, BYTE *frame, int length, void *parm)
{
    BPF_DESCR           *bpf;
    CHAIN               chain;
    MAC_INFO            info; 
    BYTE                *p;
	BPFHDR				*bpfhdr;
        
    bpf=(BPF_DESCR *)parm;
    
    bpfhdr = (BPFHDR *) frame;
    p = frame + bpfhdr->bh_hdrlen;
    
    info.copied      = length; 
    info.length      = bpfhdr->bh_datalen;
    info.time        = bpfhdr->bh_tstamp.tv_sec*1000000UL + bpfhdr->bh_tstamp.tv_usec;
    info.promiscuous = ((p[0] & (BYTE)0x01)!=(BYTE)0x01 &&
                            memcmp(p, bpf->mac->addr, bpf->mac->addrLength)!=0);
	info.status      = 0;

    bpf->Pkts++;
    bpf->Octets += info.length;
        
    if (info.length > info.copied)
        bpf->TruncatedPkts++;
    if (!info.promiscuous)
        bpf->StackedPkts++;
    if (memcmp(p,"\xff\xff\xff\xff\xff\xff",bpf->mac->addrLength)==0)
        bpf->BroadcastPkts++;
    else if(p[0] & (BYTE)0x01)
        bpf->MulticastPkts++;
	if (MacUndersize(bpf->mac->type, info.length))
	{
		info.status |= MAC_STATUS_UNDERSIZE;
        bpf->UndersizePkts++;           
	}
	else
	if (MacOversize(bpf->mac->type, info.length))
	{
		info.status |= MAC_STATUS_OVERSIZE;
        bpf->OversizePkts++;           
	}

    if (ChainAlloc(&chain, p, length, length, 0, 0)==0)
        return;
  
    MacRcve(bpf->mac, &chain, &info);
}        


static BOOLEAN Send(MAC_IFACE *iface, CHAIN *chain)
{
    BPF_DESCR  *bpf;
    BYTE       frame[BPF_FRAME];
    WORD       length;
    BYTE       addr[32];
    
    bpf=(BPF_DESCR *)iface->specific;

    length = ChainLength(chain);
    if (!ChainCopy(chain, frame, length))
        return FALSE;
                   
    sprintf(addr,"bpf:%s", Iface(iface->descr));
              
    bpf->OutPkts++;
    if (memcmp(frame, "\xff\xff\xff\xff\xff\xff", iface->addrLength) == 0)
        bpf->OutBroadcastPkts++;
    else if(frame[0] & (BYTE)0x01)
        bpf->OutMulticastPkts++;

    if (!NwDgSendTo(0, addr, frame, length))
	{
		bpf->OutDiscards++;
        return FALSE;
	}
	bpf->OutOctets += length;

    return TRUE;
}


static BOOLEAN Stat(MAC_IFACE *iface, MAC_STAT *stats)
{
IFNET ifnet;
BPF_DESCR *bpf;
NwStat *dgStats, *ifaceStats; 
NwStat *coStats;
BYTE addr[255];

    bpf=(BPF_DESCR *)iface->specific;

	/* get Nw overall statistics pointers */
	dgStats = NwDgStatistics(NULL);
	coStats = NwCoStatistics(NULL);
	
	if (bpf->offset != 0)
	{
		if (GetIfnet(bpf->offset, &ifnet) == FALSE)
			DnpapMessage(DMC_FATAL, BPF_ERROR, "bpf: failed to get interface data");
		else
		{
			bpf->OutQLen    = ifnet.if_snd.ifq_len;
			bpf->InErrors   = ifnet.if_ierrors;
			bpf->OutErrors  = ifnet.if_oerrors;
			bpf->Collisions = ifnet.if_collisions;
		}
	}
	
    sprintf(addr,"bpf:%s", Iface(iface->descr));
	ifaceStats = NwDgStatistics(addr);

    stats->LostPkts         = bpf->LostPkts      ;
    stats->Octets           = bpf->Octets        ;
    stats->Pkts             = bpf->Pkts          ;
    stats->BroadcastPkts    = bpf->BroadcastPkts ;
    stats->MulticastPkts    = bpf->MulticastPkts ;
    stats->CRCAlignErrors   = bpf->CRCAlignErrors;         
    stats->UndersizePkts    = bpf->UndersizePkts ;            
    stats->OversizePkts     = bpf->OversizePkts  ;             
    stats->Fragments        = bpf->Fragments     ;
    stats->Jabbers          = bpf->Jabbers       ;         
    stats->BufferedPkts     = bpf->BufferedPkts  ;
    stats->StackedPkts      = bpf->StackedPkts   ;
    stats->DiscardedPkts    = ifaceStats->InDiscards ;
    stats->TruncatedPkts    = bpf->TruncatedPkts ;

	/* order of execution is important */
    stats->inPkts           = bpf->Pkts          ;
	stats->inOctets         = bpf->Octets + bpf->Pkts*4;
    stats->inDiscards       = ifaceStats->InDiscards ;
    stats->inNUcastPkts     = bpf->BroadcastPkts + bpf->MulticastPkts;
    stats->inUcastPkts      = bpf->Pkts - stats->inNUcastPkts;

	/* order of execution is important */
    stats->outPkts          = dgStats->OutPkts + coStats->OutPkts;
    stats->outDiscards      = dgStats->OutDiscards + coStats->OutDiscards;
	stats->outOctets        = dgStats->OutOctets + coStats->OutOctets + stats->outPkts*4;
    stats->outNUcastPkts    = bpf->OutBroadcastPkts + bpf->OutMulticastPkts;
    stats->outUcastPkts     = stats->outPkts - stats->outNUcastPkts;

    stats->outQLen          = bpf->OutQLen       ;
    stats->inErrors         = bpf->InErrors      ;
    stats->outErrors        = bpf->OutErrors     ;
    stats->Collisions       = bpf->Collisions    ;

    return TRUE;
}


BOOLEAN GetIfnetOffset(CHAR *name, ULONG *offset)
{
IFNET ifnet, *ifnetaddr;
CHAR ifnetname[16];
CHAR tmpname[32];

	if (KvmReadSym(kvm, "_ifnet", &ifnetaddr, sizeof(IFNET *)) < 0)
		return FALSE;
	
	while (ifnetaddr != NULL)
	{
		if (KvmRead(kvm, (ULONG) ifnetaddr, &ifnet, sizeof(IFNET)) < 0)
			return FALSE;
		if (KvmRead(kvm, (ULONG) ifnet.if_name, (BYTE *) ifnetname, 16) < 0)
			return FALSE;

		sprintf(tmpname, "%.16s%d", ifnetname, ifnet.if_unit);
		if (strcmp(tmpname, name) == 0)
		{
			*offset = (ULONG) ifnetaddr;
			return TRUE;
		}

		ifnetaddr = ifnet.if_next;
	}

	*offset = 0;

	return FALSE;
}


BOOLEAN GetIfnet(ULONG offset, IFNET *ifnet)
{
	if (KvmRead(kvm, offset, ifnet, sizeof(IFNET)) < 0)
		return FALSE;
	return TRUE;
}


char *Iface(char *descr)
{
char *p;

    if ((p = strchr(descr, '.')) == NULL)
        return descr;
    return p+1;
}
