\documentstyle[lreport]{report}


\begin{document}

\newcommand{\Subscript}{{\it Subscript\/}}

\title{\Subscript}
\author{R.~Kooijman}
\maketitle


\pagenumbering{roman}
\tableofcontents
\pagebreak
\pagenumbering{arabic}


%
% Subscript MIB
%

\chapter{\Subscript\ MIB}

The \Subscript source table consists of two sub tables. One to
control source and one for the sources itself.

@d Subscript source table
@{
@<Subscript source control doc@>
@<Subscript source control table@>
@<Subscript source sources doc@>
@<Subscript source sources table@>
@}

@d Subscript run table
@{
@<Subscript run control doc@>
@<Subscript run control table@>
@<Subscript run log output doc@>
@<Subscript run log output table@>
@}

@d Subscript source control doc
@{
    -- The Subscript source control table consists of control
    -- entries that can be created and deleted dynamically or
    -- that are statically present in an agent as a result of
    -- its start-up configuration.
    -- Entries that are still in use by one or more run instances
    -- can not be deleted.
@}

@d Subscript source control table
@{
    subscriptSourceControlTable OBJECT-TYPE
        SYNTAX SEQUENCE OF SubscriptSourceControlEntry
        ACCESS not-accessible
        STATUS mandatory
        DESCRIPTION
            "A list of Subscript control entries."
        ::= { subscript 1 }
@}

@d Subscript source control table
@{
    subscriptSourceControlEntry OBJECT-TYPE
        SYNTAX SubscriptSourceControlEntry
        ACCESS not-accessible
        STATUS mandatory
        DESCRIPTION
            "Subscript control entry."
        INDEX { subscriptSourceControlIndex }
        ::= { subscriptSourceControlTable 1 }
@}

@d Subscript source control doc
@{
    -- A Subscript source control entry defines variables for a script's
    -- instance index, filename, filestatus, language, number of
    -- source lines, its owner and status.
@}

@d Subscript source control table
@{
    SubscriptSourceControlEntry ::= SEQUENCE {
        subscriptSourceControlIndex           INTEGER (1..65535),
        subscriptSourceControlFileName        FileNameString,
        subscriptSourceControlFileStatus      INTEGER,
        subscriptSourceControlLanguage        INTEGER,
        subscriptSourceControlLines           INTEGER,
        subscriptSourceControlOwner           OwnerString,
        subscriptSourceControlStatus          EntryStatus
    }
@}

@d Subscript source control table
@{
    -- The instance index of an control entry. It is defined
    -- on creation time when a create set request is issued
    -- to the corresponding status variable.

        subscriptSourceControlIndex OBJECT-TYPE
            SYNTAX INTEGER (1..65535)
            ACCESS read-only
            STATUS mandatory
            DESCRIPTION
				"Control instance index."
            ::= { subscriptSourceControlEntry 1 }
@}

@d SubscriptSourceControlIndex prototype
@{
WORD SubscriptSourceControlIndex(SNMP_OBJECT * obj, WORD idlen, VOID *param);
@}

@d SubscriptSourceControlIndex code
@{
WORD SubscriptSourceControlIndex(SNMP_OBJECT * obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, SubscriptSourceControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptSourceMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript source control table
@{
    -- A Subscript source can be saved on an agent for later retrieval.
    -- A script is identified in the agent's secondary non-volatile
    -- storage by its owner and filename.
    -- If there are no source lines defined yet and if this entry combined
    -- with the owner value point to an already existing script, then
    -- the source lines of this script will be loaded automatically.

        subscriptSourceControlFileName OBJECT-TYPE
            SYNTAX FileNameString
            ACCESS read-write
            STATUS mandatory
            DESCRIPTION
				"Filename of script on secondary storage."
            ::= { subscriptSourceControlEntry 2 }
@}

@d SubscriptSourceControlFileName prototype
@{
WORD SubscriptSourceControlFileName(SNMP_OBJECT * obj, WORD idlen, VOID *param);
@}

@d SubscriptSourceControlFileName code
@{
WORD SubscriptSourceControlFileName(SNMP_OBJECT * obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Script *data;
	SourceList *source;

	if ((local = MibRmon(obj, SubscriptSourceControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptSourceMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Script *) local->Data;
		memcpy(obj->Syntax.BufChr, data->FileName, obj->SyntaxLen = data->FileNameLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Script *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		memcpy(data->FileName, obj->Syntax.BufChr, data->FileNameLen = obj->SyntaxLen);
        /* run FileExists() to update FileStatus */
        if (FileExists(data) == TRUE && data->Source->Lines == 0)
        {
            if ((source = FileRead(data)) == NULL)
            	return SNMP_GENERROR;
            SourceListDel(data->Source);
            data->Source = source;
        }
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript source control table
@{
    -- Storing and retrieving Subscript sources from secondary
    -- non-volatile memory is controlled by subscriptSourceControlFileStatus.
    -- When this value is retrieved it may indicate that the
    -- identifying tuple (owner, filename) has an invalid value, or that
    -- a script already exists under this name or that is a new one.
    -- A script can be saved, overwritten or reloaded. Saving a script
    -- will not overwrite an existing one, use overwrite(2) to
    -- accomplish that.

        subscriptSourceControlFileStatus OBJECT-TYPE
            SYNTAX INTEGER {
                        save(1),
                        overwrite(2),
                        reload(3),
                        exists(4),
                        new(5),
                        invalid(6)
                    }
            ACCESS read-write
            STATUS mandatory
            DESCRIPTION
				"File status."
            ::= { subscriptSourceControlEntry 3 }
@}

@d SubscriptSourceControlFileStatus prototype
@{
WORD SubscriptSourceControlFileStatus(SNMP_OBJECT * obj, WORD idlen, VOID *param);
@}

@d SubscriptSourceControlFileStatus code
@{
WORD SubscriptSourceControlFileStatus(SNMP_OBJECT * obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Script *data;
	SourceList *source;

	if ((local = MibRmon(obj, SubscriptSourceControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptSourceMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Script *) local->Data;
        /* FileStatus is changed in File.*() functions */
        if (data->FileStatus == FILE_STATUS_NULL)
            FileExists(data);
		obj->Syntax.LngInt = data->FileStatus;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Script *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
        if (obj->Syntax.LngInt < FILE_STATUS_SAVE ||
            obj->Syntax.LngInt > FILE_STATUS_RELOAD)
            return SNMP_BADVALUE;
        switch (obj->Syntax.LngInt)
        {
        case FILE_STATUS_SAVE:
            if (FileExists(data) == TRUE)
                return SNMP_BADVALUE;
            if (FileWrite(data) == FALSE)
                return SNMP_BADVALUE;
            break;
        case FILE_STATUS_OVERWRITE:
            if (FileWrite(data) == FALSE)
                return SNMP_BADVALUE;
            break;
        case FILE_STATUS_RELOAD:
            if (FileExists(data) == FALSE)
                return SNMP_BADVALUE;
            if ((source = FileRead(data)) == NULL)
                return SNMP_GENERROR;
            SourceListDel(data->Source);
            data->Source = source;
            break;
        default:
            return SNMP_GENERROR;
        }
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript source control table
@{
     -- subscriptSourceControlLanguage must be set to identify
	 -- the language the script is written in. This can be used by
	 -- run instances to start up the relevant interpreter or compiler.

        subscriptSourceControlLanguage OBJECT-TYPE
            SYNTAX INTEGER {
                        scheme(1)
                        perl(2)
                    }
            ACCESS read-write
            STATUS mandatory
            DESCRIPTION
				"Script language."
            ::= { subscriptSourceControlEntry 4 }
@}

@d SubscriptSourceControlLanguage prototype
@{
WORD SubscriptSourceControlLanguage(SNMP_OBJECT * obj, WORD idlen, VOID *param);
@}

@d SubscriptSourceControlLanguage code
@{
WORD SubscriptSourceControlLanguage(SNMP_OBJECT * obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Script *data;

	if ((local = MibRmon(obj, SubscriptSourceControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptSourceMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Script *) local->Data;
        if (data->Language == SCRIPT_LANGUAGE_NULL)
            return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = data->Language;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Script *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		switch (obj->Syntax.LngInt)
		{
		case SCRIPT_LANGUAGE_SCHEME:
		case SCRIPT_LANGUAGE_PERL:
			break;
		default:
            return SNMP_BADVALUE;
		}
		data->Language = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript source control table
@{
     -- Number of source lines in the script.

        subscriptSourceControlLines OBJECT-TYPE
            SYNTAX INTEGER
            ACCESS read-only
            STATUS mandatory
            DESCRIPTION
				"Number of lines in script."
            ::= { subscriptSourceControlEntry 5 }
@}

@d SubscriptSourceControlLines prototype
@{
WORD SubscriptSourceControlLines(SNMP_OBJECT * obj, WORD idlen, VOID *param);
@}

@d SubscriptSourceControlLines code
@{
WORD SubscriptSourceControlLines(SNMP_OBJECT * obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Script *data;

	if ((local = MibRmon(obj, SubscriptSourceControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptSourceMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Script *) local->Data;
		obj->Syntax.LngInt = data->Source->Lines;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript source control table
@{
    -- Owner of this Subscript source control entry. It is
    -- also used to identify a script on secondary storage.
    -- If the 4-field owner specification of rfcXXXX is used, it is
    -- suggested to use only the management application and user ID
    -- in the owner part of the script's identification.

        subscriptSourceControlOwner OBJECT-TYPE
            SYNTAX OwnerString
            ACCESS read-write
            STATUS mandatory
            DESCRIPTION
				"Owner of this script entry."
            ::= { subscriptSourceControlEntry 6 }
@}

@d SubscriptSourceControlOwner prototype
@{
WORD SubscriptSourceControlOwner(SNMP_OBJECT * obj, WORD idlen, VOID *param);
@}

@d SubscriptSourceControlOwner code
@{
WORD SubscriptSourceControlOwner(SNMP_OBJECT * obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Script *data;
	SourceList *source;

	if ((local = MibRmon(obj, SubscriptSourceControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptSourceMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Script *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, SUBSCRIPT_SOURCE_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Owner, obj->SyntaxLen = data->OwnerLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Script *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, obj->Syntax.BufChr, data->OwnerLen = obj->SyntaxLen);
        /* run FileExists() to update FileStatus */
        if (FileExists(data) == TRUE && data->Source->Lines == 0)
        {
            if ((source = FileRead(data)) == NULL)
            	return SNMP_GENERROR;
            SourceListDel(data->Source);
            data->Source = source;
        }
		BooleanSetTrue(data->ObjectSet, SUBSCRIPT_SOURCE_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript source control table
@{
    -- The status of this Subscript source control entry. By setting
    -- this entry, new instances can be created, validated and destroyed.
    -- This mechanism is identical to the one used by the RMON MIB.

        subscriptSourceControlStatus OBJECT-TYPE
            SYNTAX EntryStatus
            ACCESS read-write
            STATUS mandatory
            DESCRIPTION
				"Status of this script entry."
            ::= { subscriptSourceControlEntry 7 }
@}

@d SubscriptSourceControlStatus prototype
@{
WORD SubscriptSourceControlStatus(SNMP_OBJECT *obj, WORD idlen, VOID *param);
@}

@d SubscriptSourceControlStatus code
@{
WORD SubscriptSourceControlStatus(SNMP_OBJECT *obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Script *data;

	local = MibRmon(obj, SubscriptSourceControlInstance, idlen, INDEX_SIZE_CONTROL);

	switch (obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (Script *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (SubscriptSourceMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (Script *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(obj, &SubscriptSourceControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = ScriptNew()) == NULL)
				{
					MibRemove(obj, &SubscriptSourceControlInstance, idlen, INDEX_SIZE_CONTROL);
					return SNMP_GENERROR;
				}
				data = (Script *) local->Data;
				if (SubscriptSourceCInit(data) == TRUE)
				{
					data->Index = local->Index;
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, SUBSCRIPT_CREATE, "subscriptSource: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				ScriptDel(local->Data);
				MibRemove(obj, &SubscriptSourceControlInstance, idlen, INDEX_SIZE_CONTROL);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (Script *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (SubscriptSourceCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, SUBSCRIPT_ACTIVE, "subscriptSource: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (SubscriptSourceCStop(data) == TRUE)
				{
					ScriptDel(local->Data);
					DnpapMessage(DMC_MESSAGE, SUBSCRIPT_DESTROY, "subscriptSource: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &SubscriptSourceControlInstance, idlen, INDEX_SIZE_CONTROL) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_INVALID:
				if (SubscriptSourceCStop(data) == TRUE)
				{
					ScriptDel(local->Data);
					DnpapMessage(DMC_MESSAGE, SUBSCRIPT_DESTROY, "subscriptSource: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &SubscriptSourceControlInstance, idlen, INDEX_SIZE_CONTROL) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (SubscriptSourceCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->FileStatus = FILE_STATUS_NULL;
            	SourceListDel(data->Source);
				data->Source = SourceListNew();
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			case SNMP_VALID:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript source sources doc
@{
    -- The Subscript source table consists of entries that form
    -- a subtable to the corresponding control table. These
    -- entries contain information about the source lines of a script
    -- including the source lines themselves.
@}

@d Subscript source sources table
@{
    subscriptSourceTable OBJECT-TYPE
        SYNTAX SEQUENCE OF SubscriptSourceEntry
        ACCESS not-accessible
        STATUS mandatory
        DESCRIPTION
            "A list of Subscript sources entries."
        ::= { subscript 2 }
@}

@d Subscript source sources doc
@{
    -- Source line entries are indexed by control index and source
    -- line number.
@}

@d Subscript source sources table
@{
    subscriptSourceEntry OBJECT-TYPE
        SYNTAX SubscriptSourceEntry
        ACCESS not-accessible
        STATUS mandatory
        DESCRIPTION
            "Subscript source entry."
        INDEX { subscriptSourceIndex , subscriptSourceLineNumber }
        ::= { subscriptSourceTable 1 }
@}

@d Subscript source sources doc
@{
    -- A Subscript source entry contains information about the
    -- control instance these lines belong to, the line number and
    -- the source line itself.
@}

@d Subscript source sources table
@{
    SubscriptSourceEntry ::= SEQUENCE {
        subscriptSourceIndex                   INTEGER,
        subscriptSourceLineNumber              INTEGER,
        subscriptSourceLine                    DisplayString
    }
@}

@d Subscript source sources table
@{
    -- The index of the control instance this source line belongs to.

        subscriptSourceIndex OBJECT-TYPE
            SYNTAX INTEGER
            ACCESS read-only
            STATUS mandatory
            DESCRIPTION
				"Index of the control entry to which
				this source line belongs."
            ::= { subscriptSourceEntry 1 }
@}

@d SubscriptSourceIndex prototype
@{
WORD SubscriptSourceIndex(SNMP_OBJECT *obj, WORD idlen, VOID *param);
@}

@d SubscriptSourceIndex code
@{
WORD SubscriptSourceIndex(SNMP_OBJECT *obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Script *data;
	SourceLine *line = NULL;

	if ((local = MibRmon(obj, SubscriptSourceControlInstance, idlen, INDEX_SIZE_SOURCE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptSourceMibNext(obj, &local, idlen, INDEX_SIZE_SOURCE, &line) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Script *) local->Data;
		if (line == NULL && (line = SourceListGetLine(data->Source, obj->Id[idlen+1])) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript source sources table
@{
    -- The line number of this source line. Source lines start at 1
    -- and have consecutive values.

        subscriptSourceLineNumber OBJECT-TYPE
            SYNTAX INTEGER
            ACCESS read-only
            STATUS mandatory
            DESCRIPTION
				"Linenumber of this source line."
            ::= { subscriptSourceEntry 2 }
@}

@d SubscriptSourceLineNumber prototype
@{
WORD SubscriptSourceLineNumber(SNMP_OBJECT *obj, WORD idlen, VOID *param);
@}

@d SubscriptSourceLineNumber code
@{
WORD SubscriptSourceLineNumber(SNMP_OBJECT *obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Script *data;
	SourceLine *line = NULL;

	if ((local = MibRmon(obj, SubscriptSourceControlInstance, idlen, INDEX_SIZE_SOURCE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptSourceMibNext(obj, &local, idlen, INDEX_SIZE_SOURCE, &line) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Script *) local->Data;
		if (line == NULL && (line = SourceListGetLine(data->Source, obj->Id[idlen+1])) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = obj->Id[idlen+1];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript source sources table
@{
    -- The contents of this source line. By setting a source line,
    -- it can be given a new contents. By setting it to "DELETE"
    -- a line can be deleted. This can only be done on the last
    -- line since the linenumbers always start at 1 and have consecutive
    -- values.
    -- Obviously, the syntax must conform to the rules as defined
    -- by the language that has been set.

        subscriptSourceLine OBJECT-TYPE
            SYNTAX DisplayString
            ACCESS read-write
            STATUS mandatory
            DESCRIPTION
				"Source line."
            ::= { subscriptSourceEntry 3 }
@}

@d SubscriptSourceLine prototype
@{
WORD SubscriptSourceLine(SNMP_OBJECT *obj, WORD idlen, VOID *param);
@}

@d SubscriptSourceLine code
@{
WORD SubscriptSourceLine(SNMP_OBJECT *obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Script *data;
	SourceLine *line = NULL;

	if ((local = MibRmon(obj, SubscriptSourceControlInstance, idlen, INDEX_SIZE_SOURCE)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptSourceMibNext(obj, &local, idlen, INDEX_SIZE_SOURCE, &line) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Script *) local->Data;
		if (line == NULL && (line = SourceListGetLine(data->Source, obj->Id[idlen+1])) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, line->Line, obj->SyntaxLen = strlen(line->Line));
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Script *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
        if (obj->SyntaxLen == strlen(SOURCE_STRING_DELETE) && strncmp(obj->Syntax.BufChr, SOURCE_STRING_DELETE, strlen(SOURCE_STRING_DELETE)) == 0)
        {
            if (SourceListDelLine(data->Source, obj->Id[idlen+1]) == FALSE)
    			return SNMP_BADVALUE;
        }
        else
        {
	    	if (SourceListAddLine(data->Source, obj->Id[idlen+1], obj->Syntax.BufChr, (int)obj->SyntaxLen) == FALSE)
    			return SNMP_BADVALUE;
        }
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript run control doc
@{
    -- The Subscript run control table consists of control entries
    -- that each contain their own run of a given source script.
    -- Entries can be created and deleted dynamically.
    -- The compiler or interpreter that is being used depends on the
    -- language of the source script.
@}

@d Subscript run control table
@{
    subscriptRunControlTable OBJECT-TYPE
        SYNTAX SEQUENCE OF SubscriptRunControlEntry
        ACCESS not-accessible
        STATUS mandatory
        DESCRIPTION
            "A list of Subscript sources entries."
        ::= { subscript 3 }
@}

@d Subscript run control table
@{
    subscriptRunControlEntry OBJECT-TYPE
        SYNTAX SubscriptRunControlEntry
        ACCESS not-accessible
        STATUS mandatory
        DESCRIPTION
            "Run control entry."
        INDEX { subscriptRunControlIndex }
        ::= { subscriptRunControlTable 1 }
@}

@d Subscript run control doc
@{
    -- A Subscript run control entry defines variables for the
    -- instance index, source script number, run command, number of
    -- run-time output lines, the owner and status of a run instance.
@}

@d Subscript run control table
@{
    SubscriptRunControlEntry ::= SEQUENCE {
        subscriptRunControlIndex               INTEGER,
        subscriptRunControlScript              INTEGER,
        subscriptRunControlCommand             INTEGER,
        subscriptRunControlLines               INTEGER,
        subscriptRunControlOwner               OwnerString,
        subscriptRunControlStatus              EntryStatus
    }
@}

@d Subscript run control table
@{
    -- The instance index of an control entry. It is defined
    -- on creation time when a create set request is issued
    -- to the corresponding status variable.

        subscriptRunControlIndex OBJECT-TYPE
            SYNTAX INTEGER (1..65535)
            ACCESS read-only
            STATUS mandatory
            DESCRIPTION
				"Run control index."
            ::= { subscriptRunControlEntry 1 }
@}

@d SubscriptRunControlIndex prototype
@{
WORD SubscriptRunControlIndex(SNMP_OBJECT * obj, WORD idlen, VOID *param);
@}

@d SubscriptRunControlIndex code
@{
WORD SubscriptRunControlIndex(SNMP_OBJECT * obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;

	if ((local = MibRmon(obj, SubscriptRunControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptRunMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript run control table
@{
    -- subscriptRunControlScript points to the instance index of
    -- the source that is being used by this run entry. That source
    -- can not be deleted while a run entry is still using it.

        subscriptRunControlScript OBJECT-TYPE
            SYNTAX INTEGER
            ACCESS read-write
            STATUS mandatory
            DESCRIPTION
				"Script that is being compiled/parsed by
				this run control entry."
            ::= { subscriptRunControlEntry 2 }
@}

@d SubscriptRunControlScript prototype
@{
WORD SubscriptRunControlScript(SNMP_OBJECT * obj, WORD idlen, VOID *param);
@}

@d SubscriptRunControlScript code
@{
WORD SubscriptRunControlScript(SNMP_OBJECT * obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Run *data;

	if ((local = MibRmon(obj, SubscriptRunControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptRunMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Run *) local->Data;
        if (data->Script == 0)
    		return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = data->Script;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Run *) local->Data;
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
		if (obj->Syntax.LngInt < 1)
			return SNMP_BADVALUE;
		data->Script = obj->Syntax.LngInt;
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript run control table
@{
    -- A script can register 5 commands that can be called externally
    -- to initialize or change the state of a script.
    -- init(1) is called implicitly when a run instance is set to
    -- valid and exit(5) is called when it is destroyed.
    -- During its initialization, a script can register certain
    -- MIB variables whose values can serve as arguments. A manager
    -- can set this values and then issue a start(2) to run
    -- the script. stop(3) and resume(4) can be used to temporarily
    -- suspend the execution of a script.
    -- init(1) and exit(5) can not be issued externally. Retrieving
    -- this variable, reveals which command was last executed.

        subscriptRunControlCommand OBJECT-TYPE
            SYNTAX INTEGER {
                        init(1),
                        start(2),
                        stop(3),
                        resume(4),
                        exit(5)
                    }
            ACCESS read-write
            STATUS mandatory
            DESCRIPTION
				"Run control entry command."
            ::= { subscriptRunControlEntry 3 }
@}

@d SubscriptRunControlCommand prototype
@{
WORD SubscriptRunControlCommand(SNMP_OBJECT * obj, WORD idlen, VOID *param);
@}

@d SubscriptRunControlCommand code
@{
WORD SubscriptRunControlCommand(SNMP_OBJECT * obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Run *data;

	if ((local = MibRmon(obj, SubscriptRunControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptRunMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Run *) local->Data;
		if (data->Status != SNMP_VALID || data->Command == RUN_COMMAND_NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = data->Command;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Run *) local->Data;
		if (data->Status != SNMP_VALID)
			return SNMP_NOSUCHNAME;
		switch (obj->Syntax.LngInt)
		{
     /* case RUN_COMMAND_INIT:
      *  	if (data->Command != RUN_COMMAND_NULL)
      *  	    return SNMP_BADVALUE;
      *  	break;
	  */
		case RUN_COMMAND_START:
			if (data->Command != RUN_COMMAND_INIT)
				return SNMP_BADVALUE;
			break;
		case RUN_COMMAND_STOP:
			if (data->Command != RUN_COMMAND_START && data->Command != RUN_COMMAND_RESUME)
				return SNMP_BADVALUE;
			break;
		case RUN_COMMAND_RESUME:
			if (data->Command != RUN_COMMAND_STOP)
				return SNMP_BADVALUE;
			break;
     /* case RUN_COMMAND_EXIT:
	  * 	if (data->Command != RUN_COMMAND_STOP)
	  * 		return SNMP_BADVALUE;
	  *     break;
	  */
		default:
			return SNMP_BADVALUE;
		}
  	 	if (RunDoCommand(data, obj->Syntax.LngInt) == FALSE)
   	 	    return SNMP_GENERROR;
   	 	/* data->Command is set by RunDoCommand */
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript run control table
@{
    -- The number of run-time log entries written so far during the
    -- execution of a script. The contents of these entries are in the
    -- Subscript log subtable.

        subscriptRunControlLines OBJECT-TYPE
            SYNTAX INTEGER
            ACCESS read-write
            STATUS mandatory
            DESCRIPTION
				"Number of log entries."
            ::= { subscriptRunControlEntry 4 }
@}

@d SubscriptRunControlLines prototype
@{
WORD SubscriptRunControlLines(SNMP_OBJECT * obj, WORD idlen, VOID *param);
@}

@d SubscriptRunControlLines code
@{
WORD SubscriptRunControlLines(SNMP_OBJECT * obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Run *data;

	if ((local = MibRmon(obj, SubscriptRunControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptRunMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Run *) local->Data;
		obj->Syntax.LngInt = data->Logs->Lines;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript run control table
@{
        subscriptRunControlOwner OBJECT-TYPE
            SYNTAX OwnerString
            ACCESS read-write
            STATUS mandatory
            DESCRIPTION
				"Owner of this control entry."
            ::= { subscriptRunControlEntry 5 }
@}

@d SubscriptRunControlOwner prototype
@{
WORD SubscriptRunControlOwner(SNMP_OBJECT * obj, WORD idlen, VOID *param);
@}

@d SubscriptRunControlOwner code
@{
WORD SubscriptRunControlOwner(SNMP_OBJECT * obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Run *data;

	if ((local = MibRmon(obj, SubscriptRunControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptRunMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Run *) local->Data;
		if (BeholderStrictRMON && !BooleanCheck(data->ObjectSet, SUBSCRIPT_RUN_BOOLEAN_OWNER))
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, data->Owner, obj->SyntaxLen = data->OwnerLen);
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		data = (Run *) local->Data;
#ifdef RMON_OWNER_STRICT
		if (data->Status != SNMP_UNDERCREATION)
			return SNMP_READONLY;
#endif
		memcpy(data->Owner, obj->Syntax.BufChr, data->OwnerLen = obj->SyntaxLen);
		BooleanSetTrue(data->ObjectSet, SUBSCRIPT_RUN_BOOLEAN_OWNER);
		return SNMP_NOERROR;
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript run control table
@{
    -- The status of this Subscript run control entry. By setting
    -- this entry, new instances can be created, validated and destroyed.
    -- This mechanism is identical to the one used by the RMON MIB.
    -- Validating an entry has the side effect that the init(1) function
    -- of a script is called. Likewise, invalidating this entry will
    -- issue an exit(5) call to a script.

        subscriptRunControlStatus OBJECT-TYPE
            SYNTAX EntryStatus
            ACCESS read-write
            STATUS mandatory
            DESCRIPTION
				"Status of this control entry."
            ::= { subscriptRunControlEntry 6 }
@}

@d SubscriptRunControlStatus prototype
@{
WORD SubscriptRunControlStatus(SNMP_OBJECT *obj, WORD idlen, VOID *param);
@}

@d SubscriptRunControlStatus code
@{
WORD SubscriptRunControlStatus(SNMP_OBJECT *obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Run *data;

	local = MibRmon(obj, SubscriptRunControlInstance, idlen, INDEX_SIZE_CONTROL);

	switch (obj->Request)
	{
	case SNMP_PDU_GET:
		if (local == NULL)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (Run *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_NEXT:
		if (SubscriptRunMibNext(obj, &local, idlen, INDEX_SIZE_CONTROL, NULL) == FALSE)
		{
			obj->Syntax.LngInt = SNMP_INVALID;
			return SNMP_NOSUCHNAME;
		}
		data = (Run *) local->Data;
		obj->Syntax.LngInt = data->Status;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		if (local == NULL)
		{
			switch (obj->Syntax.LngInt)
			{
			case SNMP_CREATEREQUEST:
				if ((local = MibInsert(obj, &SubscriptRunControlInstance, idlen, INDEX_SIZE_CONTROL)) == NULL)
					return SNMP_GENERROR;
				if ((local->Data = RunNew()) == NULL)
				{
					MibRemove(obj, &SubscriptRunControlInstance, idlen, INDEX_SIZE_CONTROL);
					return SNMP_GENERROR;
				}
				data = (Run *) local->Data;
				if (SubscriptRunCInit(data) == TRUE)
				{
					data->Index = local->Index;
					data->Status = SNMP_UNDERCREATION;
					DnpapMessage(DMC_MESSAGE, SUBSCRIPT_CREATE, "subscriptRun: collector %ld created", local->Index);
					return SNMP_NOERROR;
				}
				RunDel(local->Data);
				MibRemove(obj, &SubscriptRunControlInstance, idlen, INDEX_SIZE_CONTROL);
				return SNMP_GENERROR;
			default:
				return SNMP_NOSUCHNAME;
			}
		}
		data = (Run *) local->Data;
		switch (data->Status)
		{
		case SNMP_UNDERCREATION:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_VALID:
				if (SubscriptRunCStart(data) == TRUE)
				{
					data->Status = SNMP_VALID;
					DnpapMessage(DMC_MESSAGE, SUBSCRIPT_ACTIVE, "subscriptRun: collector %ld active", local->Index);
					return SNMP_NOERROR;
				}
				return SNMP_BADVALUE;
			case SNMP_INVALID:
				if (SubscriptRunCStop(data) == TRUE)
				{
					RunDel(local->Data);
					DnpapMessage(DMC_MESSAGE, SUBSCRIPT_DESTROY, "subscriptRun: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &SubscriptRunControlInstance, idlen, INDEX_SIZE_CONTROL) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		case SNMP_VALID:
			switch (obj->Syntax.LngInt)
			{
			case SNMP_INVALID:
				if (SubscriptRunCStop(data) == TRUE)
				{
					RunDel(local->Data);
					DnpapMessage(DMC_MESSAGE, SUBSCRIPT_DESTROY, "subscriptRun: collector %ld destroyed", local->Index);
					if (MibRemove(obj, &SubscriptRunControlInstance, idlen, INDEX_SIZE_CONTROL) == TRUE)
						return SNMP_NOERROR;
				}
				return SNMP_GENERROR;
			case SNMP_UNDERCREATION:
				if (SubscriptRunCStop(data) == FALSE)
					return SNMP_GENERROR;
				data->Command = RUN_COMMAND_NULL;
            	LogListDel(data->Logs);
				data->Logs = LogListNew();
				data->LogLine = 1;
				data->Status = SNMP_UNDERCREATION;
				return SNMP_NOERROR;
			case SNMP_VALID:
				return SNMP_NOERROR;
			default:
				return SNMP_BADVALUE;
			}
		}
	case SNMP_PDU_COMMIT:
		return SNMP_NOERROR;
	case SNMP_PDU_UNDO:
		return SNMP_GENERROR;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript run log output doc
@{
    -- The Subscript run table consists of entries that form
    -- a subtable to the corresponding control table. These
    -- entries contain information about run-time log output that
    -- was made while executing the script.
@}

@d Subscript run log output table
@{
    subscriptRunLogTable OBJECT-TYPE
        SYNTAX SEQUENCE OF SubscriptRunLogEntry
        ACCESS not-accessible
        STATUS mandatory
        DESCRIPTION
            "A list of Subscript run log entries."
        ::= { subscript 4 }
@}

@d Subscript run log output doc
@{
    -- Run-time log entries are indexed by control index and log line
    -- number.
@}

@d Subscript run log output table
@{
    subscriptRunLogEntry OBJECT-TYPE
        SYNTAX SubscriptRunLogEntry
        ACCESS not-accessible
        STATUS mandatory
        DESCRIPTION
            "Run log entry."
        INDEX { subscriptRunLogIndex , subscriptRunLogLineNumber }
        ::= { subscriptRunLogTable 1 }
@}

@d Subscript run log output doc
@{
    -- A Subscript log entry contains information about the
    -- control instance these lines belong to, the line number and
    -- the log line itself.
@}

@d Subscript run log output table
@{
    SubscriptRunLogEntry ::= SEQUENCE {
        subscriptRunLogIndex                 INTEGER,
        subscriptRunLogLineNumber            INTEGER,
        subscriptRunLogLine                  DisplayString
    }
@}

@d Subscript run log output table
@{
    -- The index of the control instance this log entry belongs to.

        subscriptRunLogIndex OBJECT-TYPE
            SYNTAX INTEGER
            ACCESS read-only
            STATUS mandatory
            DESCRIPTION
				"Index of the run control entry to which this run log
				line belongs."
            ::= { subscriptRunLogEntry 1 }
@}

@d SubscriptRunLogIndex prototype
@{
WORD SubscriptRunLogIndex(SNMP_OBJECT *obj, WORD idlen, VOID *param);
@}

@d SubscriptRunLogIndex code
@{
WORD SubscriptRunLogIndex(SNMP_OBJECT *obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Run *data;
	LogLine *line = NULL;

	if ((local = MibRmon(obj, SubscriptRunControlInstance, idlen, INDEX_SIZE_ERROR)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptRunMibNext(obj, &local, idlen, INDEX_SIZE_ERROR, &line) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Run *) local->Data;
		if (line == NULL && (line = LogListGetLine(data->Logs, obj->Id[idlen+1])) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = local->Index;
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript run log output table
@{
    -- The line number of this log line. Line numbers have a
    -- value of 1 or greater but do not have to be consecutive.

        subscriptRunLogLineNumber OBJECT-TYPE
            SYNTAX INTEGER
            ACCESS read-only
            STATUS mandatory
            DESCRIPTION
				"Linenumber of this run log."
            ::= { subscriptRunLogEntry 2 }
@}

@d SubscriptRunLogLineNumber prototype
@{
WORD SubscriptRunLogLineNumber(SNMP_OBJECT *obj, WORD idlen, VOID *param);
@}

@d SubscriptRunLogLineNumber code
@{
WORD SubscriptRunLogLineNumber(SNMP_OBJECT *obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Run *data;
	LogLine *line = NULL;

	if ((local = MibRmon(obj, SubscriptRunControlInstance, idlen, INDEX_SIZE_ERROR)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptRunMibNext(obj, &local, idlen, INDEX_SIZE_ERROR, &line) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Run *) local->Data;
		if (line == NULL && (line = LogListGetLine(data->Logs, obj->Id[idlen+1])) == NULL)
			return SNMP_NOSUCHNAME;
		obj->Syntax.LngInt = obj->Id[idlen+1];
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	}
	return SNMP_GENERROR;
}
@}

@d Subscript run log output table
@{
    -- The contents of this log line. The format is not fixed
    -- and is mainly meant as visual output for developers.

        subscriptRunLogLine OBJECT-TYPE
            SYNTAX DisplayString
            ACCESS read-only
            STATUS mandatory
            DESCRIPTION
				"Run log line."
            ::= { subscriptRunLogEntry 3 }
@}

@d SubscriptRunLogLine prototype
@{
WORD SubscriptRunLogLine(SNMP_OBJECT *obj, WORD idlen, VOID *param);
@}

@d SubscriptRunLogLine code
@{
WORD SubscriptRunLogLine(SNMP_OBJECT *obj, WORD idlen, VOID *param)
{
	MIB_LOCAL *local = NULL;
	Run *data;
	LogLine *line = NULL;

	if ((local = MibRmon(obj, SubscriptRunControlInstance, idlen, INDEX_SIZE_ERROR)) == NULL)
		return SNMP_NOSUCHNAME;

	switch (obj->Request)
	{
	case SNMP_PDU_NEXT:
		if (SubscriptRunMibNext(obj, &local, idlen, INDEX_SIZE_ERROR, &line) == FALSE)
			return SNMP_NOSUCHNAME;
	case SNMP_PDU_GET:
		data = (Run *) local->Data;
		if (line == NULL && (line = LogListGetLine(data->Logs, obj->Id[idlen+1])) == NULL)
			return SNMP_NOSUCHNAME;
		memcpy(obj->Syntax.BufChr, line->Line, obj->SyntaxLen = strlen(line->Line));
		return SNMP_NOERROR;
	case SNMP_PDU_SET:
		return SNMP_READONLY;
	}
	return SNMP_GENERROR;
}
@}


@d Miscellaneous MIB definitions
@{
@<Define Subscript SUBSCRIPT-IDENTITY macro@>
@<Hook Subscript subscript table in the MIB tree@>
@<Hook Subscript arguments and results tables in the MIB tree@>

@<FileNameString definition@>
@}


@d Hook Subscript subscript table in the MIB tree
@{
    subscript   OBJECT IDENTIFIER ::= { dnpap 4 }
@}

@d Hook Subscript arguments and results tables in the MIB tree
@{
    subscriptArguments     OBJECT IDENTIFIER ::= { subscript 5 }
    subscriptResults       OBJECT IDENTIFIER ::= { subscript 6 }
    subscriptDescriptions  OBJECT IDENTIFIER ::= { subscript 7 }
@}


@d FileNameString definition
@{
    FileNameString ::= DisplayString
@}


@d Define Subscript SUBSCRIPT-IDENTITY macro
@{
	SUBSCRIPT-IDENTITY MACRO ::=
	BEGIN
		TYPE NOTATION ::=
					"LAST-UPDATED" value(Update UTCTime)
					"ORGANIZATION" Text
					"CONTACT-INFO" Text
					"SUBSCRIPT" Text
					"DESCRIPTION" Text
					RevisionPart

		VALUE NOTATION ::=
					value(VALUE OBJECT IDENTIFIER)

		RevisionPart ::=
					Revisions
				  | empty

		Revisions ::=
					Revision
				  | Revisions Revision

		Revision ::=
					"REVISION" value(Update UTCTime)
					"DESCRIPTION" Text

		Text ::= """" string """"
	END
@}


Finally save the pieces in the right order.

@o subscript.mib -t
@{
@<Miscellaneous MIB definitions@>
@<Subscript source table@>
@<Subscript run table@>
@}


%
% Subscript code
%

\chapter{\Subscript\ code}

\section{\Subscript\ MIB code}

@o subscript.h -dt
@{
#ifndef _SUBSCRIPT_H
#define _SUBSCRIPT_H


#include <dnpap.h>
#include <snmp.h>

#include "subscriptx.h"


#define SOURCE_STRING_DELETE	"DELETE"


IMPORT BOOLEAN BeholderStrictRMON;


@<SubscriptMInit prototype@>

@<SubscriptSourceControlIndex prototype@>
@<SubscriptSourceControlFileName prototype@>
@<SubscriptSourceControlFileStatus prototype@>
@<SubscriptSourceControlLanguage prototype@>
@<SubscriptSourceControlLines prototype@>
@<SubscriptSourceControlOwner prototype@>
@<SubscriptSourceControlStatus prototype@>

@<SubscriptSourceIndex prototype@>
@<SubscriptSourceLineNumber prototype@>
@<SubscriptSourceLine prototype@>

@<SubscriptRunControlIndex prototype@>
@<SubscriptRunControlScript prototype@>
@<SubscriptRunControlCommand prototype@>
@<SubscriptRunControlLines prototype@>
@<SubscriptRunControlOwner prototype@>
@<SubscriptRunControlStatus prototype@>

@<SubscriptRunLogIndex prototype@>
@<SubscriptRunLogLineNumber prototype@>
@<SubscriptRunLogLine prototype@>


#endif
@}

@o subscriptm.c -dt
@{
@<Subscript MIB module include files@>

@<Subscript MIB module definitions@>

@<Subscript MIB module static variables@>

@<Subscript MIB module static functions@>

@<SubscriptMInit code@>

@<SubscriptSourceMibNext code@>
@<SubscriptRunMibNext code@>

@<SubscriptSourceControlIndex code@>
@<SubscriptSourceControlFileName code@>
@<SubscriptSourceControlFileStatus code@>
@<SubscriptSourceControlLanguage code@>
@<SubscriptSourceControlLines code@>
@<SubscriptSourceControlOwner code@>
@<SubscriptSourceControlStatus code@>

@<SubscriptSourceIndex code@>
@<SubscriptSourceLineNumber code@>
@<SubscriptSourceLine code@>

@<SubscriptRunControlIndex code@>
@<SubscriptRunControlScript code@>
@<SubscriptRunControlCommand code@>
@<SubscriptRunControlLines code@>
@<SubscriptRunControlOwner code@>
@<SubscriptRunControlStatus code@>

@<SubscriptRunLogIndex code@>
@<SubscriptRunLogLineNumber code@>
@<SubscriptRunLogLine code@>

@<SubscriptRegisterRun code@>
@<SubscriptDeRegisterRun code@>
@}


@d Subscript MIB module include files
@{
#include <dnpap.h>
#include <boolset.h>
#include <message.h>
#include <mibsup.h>
#include <mibrqs.h>

#include "subscriptl.h"
#include "subscriptc.h"
#include "subscript.h"
@}

@d Subscript MIB module definitions
@{
#define INDEX_SIZE_CONTROL    1
#define INDEX_SIZE_SOURCE     2
#define INDEX_SIZE_ERROR      2
@}

@d Subscript MIB module static variables
@{
static MIB_LOCAL *SubscriptSourceControlInstance = NULL;
static MIB_LOCAL *SubscriptRunControlInstance = NULL;
@}

@d Subscript MIB module static functions
@{
static BOOLEAN SubscriptSourceMibNext(SNMP_OBJECT *obj, MIB_LOCAL **local, WORD idlen, WORD indexsize, SourceLine **line);
static BOOLEAN SubscriptRunMibNext(SNMP_OBJECT *obj, MIB_LOCAL **local, WORD idlen, WORD indexsize, LogLine **line);
@}


@d SubscriptMInit prototype
@{
BOOLEAN SubscriptMInit(VOID);
@}

@d SubscriptMInit code
@{
BOOLEAN SubscriptMInit(VOID)
{
	MessageConfig(SUBSCRIPT_ERROR, "Subscript");
	if (SubscriptConfigInit() == FALSE)
		return FALSE;
	return TRUE;
}
@}


@d SubscriptSourceMibNext code
@{
#define SOURCECONTROL_DATA	((Script *)(*local)->Data)

BOOLEAN SubscriptSourceMibNext(SNMP_OBJECT *obj, MIB_LOCAL **local, WORD idlen, WORD indexsize, SourceLine **line)
{
BOOLEAN first = FALSE;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case INDEX_SIZE_SOURCE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			if ((*line = SourceListGetNextLine(SOURCECONTROL_DATA->Source, obj->Id[idlen+1])) != NULL)
			{
				obj->Id[idlen] = (*local)->Index;
				obj->Id[idlen+1] = (*line)->LineNumber;
				obj->IdLen = idlen + INDEX_SIZE_SOURCE;
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL && SOURCECONTROL_DATA->Source->Lines == 0)
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		*line = SourceListGetNextLine(SOURCECONTROL_DATA->Source, 0);
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen+1] = (*line)->LineNumber;
		obj->IdLen = idlen + INDEX_SIZE_SOURCE;
		break;
	case INDEX_SIZE_CONTROL:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + INDEX_SIZE_CONTROL;
		break;
	}
	return TRUE;
}
@}

@d SubscriptRunMibNext code
@{
#define RUNCONTROL_DATA	((Run *)(*local)->Data)

BOOLEAN SubscriptRunMibNext(SNMP_OBJECT *obj, MIB_LOCAL **local, WORD idlen, WORD indexsize, LogLine **line)
{
BOOLEAN first = FALSE;

	if (*local == NULL || indexsize < 1)
		return FALSE;

	if (obj->Id[idlen] < (*local)->Index)
		first = TRUE;

	switch (indexsize)
	{
	case INDEX_SIZE_SOURCE:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			if ((*line = LogListGetNextLine(RUNCONTROL_DATA->Logs, obj->Id[idlen+1])) != NULL)
			{
				obj->Id[idlen] = (*local)->Index;
				obj->Id[idlen+1] = (*line)->LineNumber;
				obj->IdLen = idlen + INDEX_SIZE_SOURCE;
				break;
			}
			*local = (*local)->Next;
		}
		while (*local != NULL && RUNCONTROL_DATA->Logs->Lines == 0)
			*local = (*local)->Next;
		if (*local == NULL)
			return FALSE;
		*line = LogListGetNextLine(RUNCONTROL_DATA->Logs, 0);
		obj->Id[idlen] = (*local)->Index;
		obj->Id[idlen+1] = (*line)->LineNumber;
		obj->IdLen = idlen + INDEX_SIZE_SOURCE;
		break;
	case INDEX_SIZE_CONTROL:
		if (first == FALSE && obj->IdLen >= idlen + indexsize)
		{
			/* only go to next collector if requested index was higher or equal */
			if (obj->Id[idlen] >= (*local)->Index)
				*local = (*local)->Next;
			if (*local == NULL)
				return FALSE;
		}
		obj->Id[idlen] = (*local)->Index;
		obj->IdLen = idlen + INDEX_SIZE_CONTROL;
		break;
	}
	return TRUE;
}
@}


@d SubscriptRegisterRun prototype
@{
ScriptHook *SubscriptRegisterRun(ULONG scriptIndex);
@}

@d SubscriptRegisterRun code
@{
ScriptHook *SubscriptRegisterRun(ULONG scriptIndex)
{
ScriptHook *hook;
MIB_LOCAL *local;
Script *data;

	if (scriptIndex == 0)
		return NULL;

	for (local = SubscriptSourceControlInstance; local != NULL; local = local->Next)
	{
		data = (Script *) local->Data;
        if (local->Index < scriptIndex)
            continue;
        if (local->Index != scriptIndex)
            return NULL;
        if (data->Status != SNMP_VALID)
            return NULL;
        if ((hook = SubscriptAddRun(data)) != NULL)
            return hook;
        break;
	}

    return NULL;
}
@}

@d SubscriptDeRegisterRun prototype
@{
BOOLEAN SubscriptDeRegisterRun(ULONG scriptIndex, ScriptHook *hook);
@}

@d SubscriptDeRegisterRun code
@{
BOOLEAN SubscriptDeRegisterRun(ULONG scriptIndex, ScriptHook *hook)
{
	MIB_LOCAL *local;
	Script *data;

	if (scriptIndex == 0 || hook == NULL)
		return TRUE;

	for (local = SubscriptSourceControlInstance; local != NULL; local = local->Next)
	{
		data = (Script *) local->Data;
        if (local->Index < scriptIndex)
            continue;
        if (local->Index != scriptIndex)
            return FALSE;
        if (SubscriptRemoveRun(data, hook) == TRUE)
            return TRUE;
        break;
	}

    return FALSE;
}
@}


\section{\Subscript\ additional export information}

@o subscriptx.h -dt
@{
#ifndef _SUBSCRIPTX_H
#define _SUBSCRIPTX_H

typedef struct ScriptHook_ ScriptHook;

@<SubscriptRegisterRun prototype@>
@<SubscriptDeRegisterRun prototype@>

#endif
@}


\section{\Subscript\ collector code}

@o subscriptc.h -dt
@{
#ifndef _SUBSCRIPTC_H
#define _SUBSCRIPTC_H


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dnpap.h>
#include <config.h>
#include <boolset.h>
#include <snmp.h>

#include "subscriptx.h"

@<Subscript error definitions@>
@<Subscript language definitions@>
@<Subscript file definitions@>
@<Subscript type definitions@>
@<Subscript variable definitions@>

@<SubscriptConfigInit prototype@>

@<SubscriptSourceCInit prototype@>
@<SubscriptSourceCStart prototype@>
@<SubscriptSourceCStop prototype@>

@<SubscriptRunCInit prototype@>
@<SubscriptRunCStart prototype@>
@<SubscriptRunCStop prototype@>

@<RunDoCommand prototype@>

@<RunGets prototype@>
@<RunPuts prototype@>

@<Subscript file function prototypes@>

@<Subscript source function prototypes@>
@<Subscript run function prototypes@>


#endif
@}

@o subscriptc.c -dt
@{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <dnpap.h>
#include <config.h>

#include "subscriptl.h"
#include "subscriptc.h"

@<Subscript collector static function prototypes@>

@<Subscript collector static variables@>

@<SubscriptConfigInit code@>

@<SubscriptSourceCInit code@>
@<SubscriptSourceCStart code@>
@<SubscriptSourceCStop code@>

@<SubscriptRunCInit code@>
@<SubscriptRunCStart code@>
@<SubscriptRunCStop code@>

@<RunDoCommand code@>

@<RunGets code@>
@<RunPuts code@>

@<SubscriptAddRun code@>
@<SubscriptRemoveRun code@>

@<SubscriptSourceLanguage code@>
@<SubscriptSourceList code@>

@<SubscriptRunCommand code@>
@<SubscriptRunIndex code@>
@<SubscriptRunScript code@>
@}

@d Subscript collector static variables
@{
static CHAR *StaticLog = "error";

static CHAR *ScriptDirectory = NULL;
@}


@d SubscriptConfigInit prototype
@{
BOOLEAN SubscriptConfigInit(VOID);
@}

@d SubscriptConfigInit code
@{
BOOLEAN SubscriptConfigInit(VOID)
{
	ScriptDirectory = SCRIPT_DIRECTORY;
	ConfigGetString("subscript.directory", &ScriptDirectory);
	return TRUE;
}
@}

@d SubscriptSourceCInit prototype
@{
BOOLEAN SubscriptSourceCInit(Script *script);
@}

@d SubscriptSourceCInit code
@{
BOOLEAN SubscriptSourceCInit(Script *script)
{
	return TRUE;
}
@}

@d SubscriptSourceCStart prototype
@{
BOOLEAN SubscriptSourceCStart(Script *script);
@}

@d SubscriptSourceCStart code
@{
BOOLEAN SubscriptSourceCStart(Script *script)
{
    if (script->Source->Lines == 0)
		return FALSE;

	switch (script->Language)
	{
	case SCRIPT_LANGUAGE_SCHEME:
	case SCRIPT_LANGUAGE_PERL:
		break;
	default:
		return FALSE;
	}

	return TRUE;
}
@}

@d SubscriptSourceCStop prototype
@{
BOOLEAN SubscriptSourceCStop(Script *script);
@}

@d SubscriptSourceCStop code
@{
BOOLEAN SubscriptSourceCStop(Script *script)
{
	if (script->Runs != NULL)
		return FALSE;
	return TRUE;
}
@}


@d SubscriptRunCInit prototype
@{
BOOLEAN SubscriptRunCInit(Run *run);
@}

@d SubscriptRunCInit code
@{
BOOLEAN SubscriptRunCInit(Run *run)
{
	return TRUE;
}
@}

@d SubscriptRunCStart prototype
@{
BOOLEAN SubscriptRunCStart(Run *run);
@}

@d SubscriptRunCStart code
@{
BOOLEAN SubscriptRunCStart(Run *run)
{
    if (run->Script == 0)
		return FALSE;
	if ((run->ScriptHook = SubscriptRegisterRun(run->Script)) == NULL)
		return FALSE;
	switch (SubscriptSourceLanguage(run->ScriptHook))
	{
	case SCRIPT_LANGUAGE_SCHEME:
		run->Languages.Scheme = SchemeInit(RunGets, RunPuts, run);
    	if (run->Languages.Scheme == NULL)
	    {
		    SubscriptDeRegisterRun(run->Script, run->ScriptHook);
			return FALSE;
	    }
	    if (SchemeCall(run->Languages.Scheme, RUN_COMMAND_INIT) == FALSE)
	    {
			SchemeExit(run->Languages.Scheme);
		    SubscriptDeRegisterRun(run->Script, run->ScriptHook);
			return FALSE;
	    }
	    run->Command = RUN_COMMAND_INIT;
		break;
	case SCRIPT_LANGUAGE_PERL:
		run->Languages.Perl = PerlInit(RunGets, RunPuts, run);
    	if (run->Languages.Perl == NULL)
	    {
		    SubscriptDeRegisterRun(run->Script, run->ScriptHook);
			return FALSE;
	    }
	    if (PerlCall(run->Languages.Perl, RUN_COMMAND_INIT) == FALSE)
	    {
			PerlExit(run->Languages.Perl);
		    SubscriptDeRegisterRun(run->Script, run->ScriptHook);
			return FALSE;
	    }
	    run->Command = RUN_COMMAND_INIT;
		break;
	default:
		return FALSE;
	}
	return TRUE;
}
@}

@d SubscriptRunCStop prototype
@{
BOOLEAN SubscriptRunCStop(Run *run);
@}

@d SubscriptRunCStop code
@{
BOOLEAN SubscriptRunCStop(Run *run)
{
	switch (SubscriptSourceLanguage(run->ScriptHook))
	{
	case SCRIPT_LANGUAGE_SCHEME:
	    SchemeCall(run->Languages.Scheme, RUN_COMMAND_EXIT);
	    run->Command = RUN_COMMAND_EXIT;
		SchemeExit(run->Languages.Scheme);
		run->Languages.Scheme = NULL;
		break;
	case SCRIPT_LANGUAGE_PERL:
	    PerlCall(run->Languages.Perl, RUN_COMMAND_EXIT);
	    run->Command = RUN_COMMAND_EXIT;
		PerlExit(run->Languages.Perl);
		run->Languages.Perl = NULL;
		break;
	default:
		return FALSE;
	}
	SubscriptDeRegisterRun(run->Script, run->ScriptHook);
	return TRUE;
}
@}


@d RunDoCommand prototype
@{
BOOLEAN RunDoCommand(Run *run, int command);
@}

@d RunDoCommand code
@{
BOOLEAN RunDoCommand(Run *run, int command)
{
BOOLEAN result;

	switch (SubscriptSourceLanguage(run->ScriptHook))
	{
	case SCRIPT_LANGUAGE_SCHEME:
		result = SchemeCall(run->Languages.Scheme, command);
        if (result == FALSE)
			return FALSE;
	    run->Command = command;
		break;
	case SCRIPT_LANGUAGE_PERL:
		result = PerlCall(run->Languages.Perl, command);
        if (result == FALSE)
			return FALSE;
	    run->Command = command;
		break;
	default:
		return FALSE;
	}

	return TRUE;
}
@}


@d RunGets prototype
@{
CHAR *RunGets(VOID *run, int line);
@}

@d RunGets code
@{
CHAR *RunGets(VOID *r, int line)
{
Run *run = (Run *) r;
SourceLine *l;

	if ((l = SourceListGetLine(SubscriptSourceList(run->ScriptHook), line)) == NULL)
		return NULL;
	return l->Line;
}
@}

@d RunPuts prototype
@{
int RunPuts(VOID *run, int line, CHAR *log);
@}

@d RunPuts code
@{
int RunPuts(VOID *r, int line, CHAR *log)
{
Run *run = (Run *) r;
int len;

	if (line == 0)
		line = run->LogLine++;
	if (LogListAddLine(run->Logs, line, log, len = strlen(log)) == FALSE)
		return -1;
	return len;
}
@}

@d Subscript source function prototypes
@{
ScriptHook *SubscriptAddRun(Script *script);
@}

@d SubscriptAddRun code
@{
ScriptHook *SubscriptAddRun(Script *script)
{
ScriptHook *hook;

    if ((hook = DnpapMalloc(sizeof(ScriptHook))) == NULL)
        return NULL;
    hook->Script = script;
    hook->Prev = NULL;
    hook->Next = script->Runs;
    if (script->Runs != NULL)
        script->Runs->Prev = hook;
        script->Runs = hook;

    return hook;
}
@}


@d Subscript source function prototypes
@{
BOOLEAN SubscriptRemoveRun(Script *script, ScriptHook *hook);
@}

@d SubscriptRemoveRun code
@{
BOOLEAN SubscriptRemoveRun(Script *script, ScriptHook *hook)
{
    if (hook == NULL)
        return TRUE;
    if (hook->Next != NULL)
        hook->Next->Prev = hook->Prev;
    if (hook->Prev == NULL)
        script->Runs = hook->Next;
    else
        hook->Prev->Next = hook->Next;

    DnpapFree(hook);

    return TRUE;
}
@}

@d Subscript source function prototypes
@{
int SubscriptSourceLanguage(ScriptHook *hook);
@}

@d SubscriptSourceLanguage code
@{
int SubscriptSourceLanguage(ScriptHook *hook)
{
	if (hook == NULL)
		return SCRIPT_LANGUAGE_NULL;
	return hook->Script->Language;
}
@}


@d Subscript source function prototypes
@{
SourceList *SubscriptSourceList(ScriptHook *hook);
@}

@d SubscriptSourceList code
@{
SourceList *SubscriptSourceList(ScriptHook *hook)
{
	if (hook == NULL)
		return NULL;
	return hook->Script->Source;
}
@}


@d Subscript run function prototypes
@{
int SubscriptRunCommand(Run *run);
@}

@d SubscriptRunCommand code
@{
int SubscriptRunCommand(Run *run)
{
	if (run == NULL)
		return RUN_COMMAND_NULL;
	return run->Command;
}
@}

@d Subscript run function prototypes
@{
int SubscriptRunIndex(Run *run);
@}

@d SubscriptRunIndex code
@{
int SubscriptRunIndex(Run *run)
{
	if (run == NULL)
		return 0;
	return run->Index;
}
@}

@d Subscript run function prototypes
@{
int SubscriptRunScript(Run *run);
@}

@d SubscriptRunScript code
@{
int SubscriptRunScript(Run *run)
{
	if (run == NULL)
		return 0;
	return run->Script;
}
@}

@d Subscript error definitions
@{
enum SubscriptErrors
{
    SUBSCRIPT_CREATE = SUBSCRIPT_ERROR,
    SUBSCRIPT_ACTIVE, SUBSCRIPT_DESTROY,
	SUBSCRIPT_CONTEXT
};
@}

@d Subscript language definitions
@{
#define SCRIPT_LANGUAGE_NULL		0
#define SCRIPT_LANGUAGE_SCHEME		1
#define SCRIPT_LANGUAGE_PERL		2
@}

@d Subscript command definitions
@{
#define RUN_COMMAND_NULL			0
#define RUN_COMMAND_INIT			1
#define RUN_COMMAND_START			2
#define RUN_COMMAND_STOP			3
#define RUN_COMMAND_RESUME			4
#define RUN_COMMAND_EXIT			5
@}

@d Subscript file definitions
@{
#define SCRIPT_DIRECTORY		"/usr/local/project/scripts"

#define FILE_STATUS_NULL			0
#define FILE_STATUS_SAVE			1
#define FILE_STATUS_OVERWRITE		2
#define FILE_STATUS_RELOAD			3
#define FILE_STATUS_EXISTS			4
#define FILE_STATUS_NEW				5
#define FILE_STATUS_INVALID			6
@}

@d Subscript type definitions
@{
@<SourceLine type definition@>
@<SourceList type definition@>

@<Script type definition@>

@<LogLine type definition@>
@<LogList type definition@>

@<Run type definition@>

@<ScriptHook structure definition@>
@}

@d SourceLine type definition
@{
typedef struct SourceLine_ SourceLine;

struct SourceLine_
{
    int LineNumber;
	CHAR *Line;

	SourceLine *Prev;
	SourceLine *Next;
};
@}

@d SourceList type definition
@{
typedef struct SourceList_
{
	int Lines;
	SourceLine *Head;
	SourceLine *Tail;
	SourceLine *Current;

} SourceList;
@}

@d Script type definition
@{
typedef struct Script_
{
    ULONG Index;
	CHAR FileName[SNMP_SIZE_BUFCHR];
    WORD FileNameLen;
	LONG FileStatus;
    LONG Language;
	CHAR Owner[SNMP_SIZE_BUFCHR];
    WORD OwnerLen;
	LONG Status;

	BooleanSet ObjectSet;

	SourceList *Source;

	ScriptHook *Runs;

} Script;
@}

@d LogLine type definition
@{
typedef struct LogLine_ LogLine;

struct LogLine_
{
    int LineNumber;
	CHAR *Line;

	LogLine *Next;
};
@}

@d LogList type definition
@{
typedef struct LogList_
{
	int Lines;
	LogLine *Head;
	LogLine *Current;

} LogList;
@}

@d Run type definition
@{
typedef struct Run_
{
    ULONG Index;
    ULONG Script;
    ScriptHook *ScriptHook;
    LONG Command;
	CHAR Owner[SNMP_SIZE_BUFCHR];
    WORD OwnerLen;
	LONG Status;

	BooleanSet ObjectSet;

	LogList *Logs;
	int LogLine;

	union
	{
		VOID *Scheme;
		VOID *Perl;
	} Languages;

} Run;
@}

@d ScriptHook structure definition
@{
struct ScriptHook_
{
    Script *Script;
    ScriptHook *Prev;
    ScriptHook *Next;
};
@}

@d Subscript variable definitions
@{
#define SUBSCRIPT_SOURCE_BOOLEAN_LANGUAGE	0
#define SUBSCRIPT_SOURCE_BOOLEAN_OWNER		1

#define SUBSCRIPT_RUN_BOOLEAN_SCRIPT		0
#define SUBSCRIPT_RUN_BOOLEAN_OWNER			1
@}


@d Subscript source function prototypes
@{
BOOLEAN SourceListAddLine(SourceList *list, int line, CHAR *data, int len);
@}

@o subscriptc.c -dt
@{
BOOLEAN SourceListAddLine(SourceList *list, int line, CHAR *data, int len)
{
SourceLine *l;
CHAR *new;

    if (data == NULL)
        return FALSE;

	if (line < 1 || line > list->Lines + 1)
		return FALSE;

	if (line == list->Lines + 1)
	{
		if ((l = SourceLineNew()) == NULL)
			return FALSE;
        l->LineNumber = line;
		if (list->Tail != NULL)
			list->Tail->Next = l;
		else
			list->Head = l;
		l->Prev = list->Tail;
		list->Tail = l;
		list->Lines++;
	}
	else
	if ((l = SourceListGetLine(list, line)) == NULL)
		return FALSE;

	if ((new = DnpapRealloc(l->Line, (ULONG)len+1)) == NULL)
		return FALSE;

	strncpy(l->Line = new, data, len);
    l->Line[len]='\0';

	return TRUE;
}
@}


@d Subscript source function prototypes
@{
BOOLEAN SourceListDelLine(SourceList *list, int line);
@}

@o subscriptc.c -dt
@{
BOOLEAN SourceListDelLine(SourceList *list, int line)
{
SourceLine *last;

	if (line < 1 || line != list->Lines)
		return FALSE;

    last = list->Tail;
	list->Current = list->Tail = list->Tail->Prev;

	if (list->Tail == NULL)
		list->Head = NULL;
	else
		list->Tail->Next = NULL;
	list->Lines--;

	SourceLineDel(last);

	return TRUE;
}
@}

@d Subscript source function prototypes
@{
SourceList *SourceListNew(VOID);
@}

@o subscriptc.c -dt
@{
SourceList *SourceListNew(VOID)
{
SourceList *list;

	if ((list = DnpapMalloc(sizeof(SourceList))) == NULL)
		return NULL;

	list->Lines = 0;
	list->Head = NULL;
	list->Tail = NULL;
	list->Current = NULL;

	return list;
}
@}

@d Subscript source function prototypes
@{
VOID SourceListDel(SourceList *list);
@}

@o subscriptc.c -dt
@{
VOID SourceListDel(SourceList *list)
{
	while (list->Lines > 0)
		SourceListDelLine(list, list->Lines);
	DnpapFree(list);
}
@}


@d Subscript collector static function prototypes
@{
SourceLine *SourceLineNew(VOID);
@}

@o subscriptc.c -dt
@{
SourceLine *SourceLineNew(VOID)
{
SourceLine *line;

	if ((line = DnpapMalloc(sizeof(SourceLine))) == NULL)
		return NULL;

	line->LineNumber = 0;
	line->Line = NULL;
	line->Prev = NULL;
	line->Next = NULL;

	return line;
}
@}


@d Subscript collector static function prototypes
@{
VOID SourceLineDel(SourceLine *line);
@}

@o subscriptc.c -dt
@{
VOID SourceLineDel(SourceLine *line)
{
	DnpapFree(line);
}
@}


@d Subscript run function prototypes
@{
LogList *LogListNew(VOID);
@}

@o subscriptc.c -dt
@{
LogList *LogListNew(VOID)
{
LogList *list;

	if ((list = DnpapMalloc(sizeof(LogList))) == NULL)
		return NULL;

	list->Lines = 0;
	list->Head = NULL;
	list->Current = NULL;

	return list;
}
@}

@d Subscript collector static function prototypes
@{
LogLine *LogLineNew(VOID);
@}

@o subscriptc.c -dt
@{
LogLine *LogLineNew(VOID)
{
LogLine *log;

	if ((log = DnpapMalloc(sizeof(LogLine))) == NULL)
		return NULL;

	log->LineNumber = 0;
	log->Line = NULL;
	log->Next = NULL;

	return log;
}
@}


@d Subscript collector static function prototypes
@{
VOID LogLineDel(LogLine *log);
@}

@o subscriptc.c -dt
@{
VOID LogLineDel(LogLine *log)
{
    if (log->Line != StaticLog)
        DnpapFree(log->Line);
	DnpapFree(log);
}
@}


@d Subscript run function prototypes
@{
VOID LogListDel(LogList *logs);
@}

@o subscriptc.c -dt
@{
VOID LogListDel(LogList *logs)
{
	while (logs->Lines > 0)
		LogListDelFirstLine(logs);
	DnpapFree(logs);
}
@}


@d Subscript run function prototypes
@{
VOID LogListDelLines(LogList *logs);
@}

@o subscriptc.c -dt
@{
VOID LogListDelLines(LogList *logs)
{
	while (logs->Lines > 0)
		LogListDelFirstLine(logs);
}
@}


@d Subscript source function prototypes
@{
SourceLine *SourceListGetLine(SourceList *list, int line);
@}

@o subscriptc.c -dt
@{
SourceLine *SourceListGetLine(SourceList *list, int line)
{
	if (line < 1 || line > list->Lines)
		return NULL;
    if (list->Current != NULL && list->Current->LineNumber == line)
        return list->Current;
	for (list->Current = list->Head; list->Current != NULL; list->Current = list->Current->Next)
        if (list->Current->LineNumber >= line)
            break;
    if (list->Current != NULL && list->Current->LineNumber == line)
    	return list->Current;
    return NULL;
}
@}


@d Subscript source function prototypes
@{
SourceLine *SourceListGetNextLine(SourceList *list, int line);
@}

@o subscriptc.c -dt
@{
SourceLine *SourceListGetNextLine(SourceList *list, int line)
{
    if (line < 1)
        return list->Current = list->Head;
	if ((list->Current = SourceListGetLine(list, line)) == NULL)
		return NULL;
	return list->Current = list->Current->Next;
}
@}


@d Subscript run function prototypes
@{
LogLine *LogListGetLine(LogList *list, int line);
@}

@o subscriptc.c -dt
@{
LogLine *LogListGetLine(LogList *list, int line)
{
	if (line < 1)
		return NULL;
    if (list->Current != NULL && list->Current->LineNumber == line)
        return list->Current;
	for (list->Current = list->Head; list->Current != NULL; list->Current = list->Current->Next)
        if (list->Current->LineNumber >= line)
            break;
    if (list->Current != NULL && list->Current->LineNumber == line)
    	return list->Current;
	return NULL;
}
@}


@d Subscript run function prototypes
@{
LogLine *LogListGetNextLine(LogList *list, int line);
@}

@o subscriptc.c -dt
@{
LogLine *LogListGetNextLine(LogList *list, int line)
{
    if (line < 1)
        return list->Current = list->Head;
	if ((list->Current = LogListGetLine(list, line)) == NULL)
		return NULL;
	return list->Current = list->Current->Next;
}
@}


@d Subscript run function prototypes
@{
BOOLEAN LogListAddLine(LogList *list, int line, CHAR *log, int len);
@}

@o subscriptc.c -dt
@{
BOOLEAN LogListAddLine(LogList *list, int line, CHAR *log, int len)
{
LogLine **l, *entry;
CHAR *new;

    if (log == NULL)
        return FALSE;

	for (l = &list->Head; *l != NULL; l = &(*l)->Next)
		if (line <= (*l)->LineNumber)
			break;
    entry = *l;
	if (entry == NULL || line != entry->LineNumber)
	{
		if ((entry = LogLineNew()) == NULL)
			return FALSE;
        entry->LineNumber = line;
		entry->Next = *l;
		*l = entry;
        list->Lines++;
	}

	if (entry->Line == NULL)
	{
		if ((entry->Line = DnpapMalloc((ULONG)len+1)) == NULL)
		{
			/* let log string point to well-known static string */
			entry->Line = StaticLog;
			return FALSE;
		}
		strncpy(entry->Line, log, len);
        entry->Line[len] = '\0';
	}
	else
	{
		if ((new = DnpapRealloc(entry->Line, strlen(entry->Line) + strlen("\n") + len + 1)) == NULL)
			return FALSE;
		entry->Line = new;
		sprintf(entry->Line + strlen(entry->Line), "\n%*s", len, log);
	}

	return TRUE;
}
@}


@d Subscript run function prototypes
@{
BOOLEAN LogListDelFirstLine(LogList *list);
@}

@o subscriptc.c -dt
@{
BOOLEAN LogListDelFirstLine(LogList *list)
{
LogLine *line;

    if (list->Head == NULL)
        return FALSE;
    line = list->Head;
    list->Current = list->Head = list->Head->Next;

    list->Lines--;

    LogLineDel(line);

    return TRUE;
}
@}


@d Subscript source function prototypes
@{
Script *ScriptNew(VOID);
@}

@o subscriptc.c -dt
@{
Script *ScriptNew(VOID)
{
Script *script;

    if ((script = DnpapMalloc(sizeof(Script))) == NULL)
        return NULL;

    script->Index = 0;
	script->FileName[0] = '\0';
    script->FileNameLen = 0;
	script->FileStatus = FILE_STATUS_NULL;
    script->Language = SCRIPT_LANGUAGE_NULL;
    script->Owner[0] = '\0';
    script->OwnerLen = 0;
	script->Status = SNMP_INVALID;

    if ((script->Source = SourceListNew()) == NULL)
    {
        ScriptDel(script);
        return NULL;
    }

	script->Runs = NULL;

    return script;
}
@}


@d Subscript source function prototypes
@{
VOID ScriptDel(Script *script);
@}

@o subscriptc.c -dt
@{
VOID ScriptDel(Script *script)
{
    SourceListDel(script->Source);
    script->Source = NULL;
    DnpapFree(script);
}
@}


@d Subscript run function prototypes
@{
Run *RunNew(VOID);
@}

@o subscriptc.c -dt
@{
Run *RunNew(VOID)
{
Run *run;

    if ((run = DnpapMalloc(sizeof(Run))) == NULL)
        return NULL;

    run->Index = 0;
    run->Script = 0;
    run->ScriptHook = NULL;
    run->Command = RUN_COMMAND_NULL;
    run->Owner[0] = '\0';
    run->OwnerLen = 0;
	run->Status = SNMP_INVALID;

    if ((run->Logs = LogListNew()) == NULL)
    {
        RunDel(run);
        return NULL;
    }
	run->LogLine = 1;

    return run;
}
@}


@d Subscript run function prototypes
@{
VOID RunDel(Run *run);
@}

@o subscriptc.c -dt
@{
VOID RunDel(Run *run)
{
    LogListDel(run->Logs);
    run->Logs = NULL;
    DnpapFree(run);
}
@}



\section{File Functions}


@d Subscript file function prototypes
@{
SourceList *FileRead(Script *script);
@}

@o subscriptc.c -dt
@{
SourceList *FileRead(Script *script)
{
SourceList *source;
CHAR name[SNMP_SIZE_BUFCHR];
CHAR line[SNMP_SIZE_BUFCHR];
FILE *f;
int i;

    script->FileStatus = FILE_STATUS_INVALID;

    if (script->OwnerLen == 0 || script->FileNameLen == 0)
        return NULL;

    @<Copy script data to C string filename@>

    if ((source = SourceListNew()) == NULL)
        return FALSE;

    script->FileStatus = FILE_STATUS_NEW;

    if ((f = fopen(name, "r")) == NULL)
    {
        SourceListDel(source);
        return NULL;
    }

    script->FileStatus = FILE_STATUS_EXISTS;

    for (i = 1; fgets(line, sizeof(line), f) != (char *)EOF; i++)
    {
        if (SourceListAddLine(source, i, line, (int)strlen(line)) == FALSE)
        {
            fclose(f);
            SourceListDel(source);
            return NULL;
        }
    }

    fclose(f);

    return source;
}
@}


@d Subscript file function prototypes
@{
BOOLEAN FileWrite(Script *script);
@}

@o subscriptc.c -dt
@{
BOOLEAN FileWrite(Script *script)
{
CHAR name[SNMP_SIZE_BUFCHR];
FILE *f;
int i;
SourceLine *line;

    script->FileStatus = FILE_STATUS_INVALID;

    if (script->OwnerLen == 0 || script->FileNameLen == 0)
        return FALSE;

    @<Copy script data to C string filename@>

    script->FileStatus = FILE_STATUS_NEW;

    if ((f = fopen(name, "w")) == NULL)
        return FALSE;

    script->FileStatus = FILE_STATUS_EXISTS;

    for (i = 1; i <= script->Source->Lines; i++)
    {
        if ((line = SourceListGetLine(script->Source, i)) == NULL ||
            fputs(line->Line, f) == EOF)
        {
            fclose(f);
            return FALSE;
        }
    }

    fclose(f);

    return TRUE;
}
@}


@d Subscript file function prototypes
@{
BOOLEAN FileExists(Script *script);
@}

@o subscriptc.c -dt
@{
BOOLEAN FileExists(Script *script)
{
CHAR name[SNMP_SIZE_BUFCHR];
FILE *f;

    script->FileStatus = FILE_STATUS_INVALID;

    if (script->OwnerLen == 0 || script->FileNameLen == 0)
        return FALSE;

    @<Copy script data to C string filename@>

    script->FileStatus = FILE_STATUS_NEW;

    if ((f = fopen(name, "r")) == NULL)
        return FALSE;
    fclose(f);

    script->FileStatus = FILE_STATUS_EXISTS;

    return TRUE;
}
@}

@d Copy script data to C string filename
@{
{
CHAR  owner[SNMP_SIZE_BUFCHR],
	 *o1 = NULL, *o2 = NULL, *o3 = NULL, *o4 = NULL,
	 *p;

	strncpy(owner, script->Owner, script->OwnerLen);
	owner[script->OwnerLen] = '\0';
	o1 = strtok(owner, " \t");
	if (o1 != NULL)
		o2 = strtok(NULL, " \t");
	if (o2 != NULL)
		o3 = strtok(NULL, " \t");
	if (o3 != NULL)
		o4 = strtok(NULL, " \t");

	if (o2 != NULL && o3 != NULL)
	{
		if ((p = strchr(o2, '(')) != NULL)
			*p = '\0';
		if ((p = strchr(o3, '(')) != NULL)
			*p = '\0';
    	sprintf(name, "%s/%s/%s/%*s", ScriptDirectory,
                                	  o2, o3,
                                	  script->FileNameLen, script->FileName);
	}
	else
	if (o1 != NULL)
	{
    	sprintf(name, "%s/%s/%*s", ScriptDirectory,
                              	   o1,
                                   script->FileNameLen, script->FileName);
	}
	else
	{
    	sprintf(name, "%s/%*s", ScriptDirectory,
                                script->FileNameLen, script->FileName);
	}
}
@}



%
% Language functions
%

\chapter{Languages}

Currently SubScript supports the Scheme and Perl language.
For Scheme we used Brent Benson's LibScheme library and for Perl
we just took the Perl 5 library.

@o subscriptl.h -dt
@{
#ifndef _SUBSCRIPTL_H_
#define _SUBSCRIPTL_H_


#ifdef SUBSCRIPT_PERL
#include <perl/INTERN.h>
#include <perl/perl.h>
#undef WORD
#endif
#include <dnpap.h>
#include <hash.h>
#include <mibsup.h>
#include <protocol.h>
#include <mac.h>
#include <timer.h>
#include <channel.h>
#include <event.h>
#include <nw.h>
#include <nwsnmpdef.h>
#include <scheme.h>

@<Language function prototype definitions@>
@<Language type definitions@>

@<Scheme size definitions@>

@<Scheme type definitions@>
@<Perl type definitions@>

@<Scheme macro definitions@>

@<Subscript command definitions@>

@<SchemeInit prototype@>
@<SchemeCall prototype@>
@<SchemeExit prototype@>

@<PerlInit prototype@>
@<PerlCall prototype@>
@<PerlExit prototype@>

@<SchemeInputGetc prototype@>
@<SchemeInputUnGetc prototype@>
@<SchemeInputPoll prototype@>
@<SchemeInputClose prototype@>
@<SchemeOutputPuts prototype@>
@<SchemeOutputClose prototype@>

@<AddCallback prototype@>
@<SearchCallback prototype@>
@<RemoveCallback prototype@>
@<CutCallback prototype@>

@<SchemeContextInit prototype@>
@<SchemeContextSwitch prototype@>
@<SchemeContextCurrent prototype@>

@<PerlContextInit prototype@>
@<PerlContextSwitch prototype@>
@<PerlContextCurrent prototype@>

#endif
@}

@o subscriptl.c -dt
@{
#ifdef SUBSCRIPT_PERL
#include <perl/INTERN.h>
#include <perl/perl.h>
#undef WORD
#endif
#include <string.h>
#include <dnpap.h>
#include <sysmib.h>
#include <agent.h>
#include <mibsup.h>
#include <mibrqs.h>
#include <agent.h>
#include <scheme.h>

#include "subscriptc.h"
#include "subscriptl.h"

@<Scheme static variables@>

#ifdef SUBSCRIPT_PERL
@<Perl static variables@>
#endif

@<SchemeNew prototype@>
@<SchemeDel prototype@>

#ifdef SUBSCRIPT_PERL
@<Perl external extensions callback prototype@>
@<PerlNew prototype@>
@<PerlDel prototype@>
#endif

@<SchemeMakeAssoc prototype@>
@<SchemeAssocSize prototype@>
@<SchemeAssocRef prototype@>
@<SchemeAssocSet prototype@>
@<SchemeDelAssoc prototype@>
@<HashCleanEntry prototype@>

@<SchemeRunCommandRegister prototype@>

@<SchemeMacRegister prototype@>
@<SchemeMacDeRegister prototype@>
@<SchemeMac prototype@>
@<SchemeDeMac prototype@>
@<SchemeMacCallback prototype@>

@<SchemeTimerRegister prototype@>
@<SchemeTimerDeRegister prototype@>
@<SchemeTimer prototype@>
@<SchemeDeTimer prototype@>
@<SchemeTimerCallback prototype@>

@<SchemeChannelRegister prototype@>
@<SchemeChannelDeRegister prototype@>
@<SchemeChannel prototype@>
@<SchemeDeChannel prototype@>
@<SchemeChannelCallback prototype@>
@<SchemeChannelDeRegisterCallback prototype@>

@<SchemeFilterRegister prototype@>
@<SchemeFilterDeRegister prototype@>
@<SchemeFilter prototype@>
@<SchemeDeFilter prototype@>
@<SchemeFilterCallback prototype@>
@<SchemeFilterDeRegisterCallback prototype@>

@<SchemeGeneratorRegister prototype@>
@<SchemeGeneratorSend prototype@>
@<SchemeEventGenerate prototype@>
@<SchemeGeneratorDeRegister prototype@>
@<SchemeGenerator prototype@>
@<SchemeDeGenerator prototype@>
@<SchemeGeneratorDeRegisterCallback prototype@>

@<SchemeEventRegister prototype@>
@<SchemeEventDeRegister prototype@>
@<SchemeEvent prototype@>
@<SchemeDeEvent prototype@>
@<SchemeEventCallback prototype@>
@<SchemeEventDeRegisterCallback prototype@>

@<SchemeMibRequest prototype@>
@<SchemeMibGet prototype@>
@<SchemeMibSet prototype@>
@<SchemeMibNext prototype@>
@<SchemeMibCallback prototype@>
@<SchemeMib prototype@>
@<SchemeDeMib prototype@>
@<SchemeMibRegister prototype@>
@<SchemeMibDeRegister prototype@>
@<SchemeMibRegisterArgs prototype@>
@<SchemeMibRegisterResults prototype@>

@<Varbinds2Objects prototype@>
@<Objects2Varbinds prototype@>

@<SnmpObjectPutId prototype@>
@<SnmpObjectPutType prototype@>
@<SnmpObjectPutSyntax prototype@>
@<SnmpObjectGetId prototype@>
@<SnmpObjectGetType prototype@>
@<SnmpObjectGetSyntax prototype@>
@<SnmpObjectGetRequest prototype@>

@<SchemeSnmpConnect prototype@>
@<SchemeSnmp prototype@>
@<SchemeDeSnmp prototype@>
@<SchemeSnmpSend prototype@>
@<SchemeSnmpDisConnect prototype@>
@<SchemeSnmpCallback prototype@>

@<SchemeTrapRegister prototype@>
@<SchemeTrap prototype@>
@<SchemeDeTrap prototype@>
@<SchemeTrapDeRegister prototype@>
@<SchemeTrapCallback prototype@>

@<SchemeTrapSend prototype@>

@<SuspendCallbacks prototype@>
@<RemoveAllCallbacks prototype@>

@<scheme_make_packet prototype@>

@<SchemeInit code@>
@<SchemeCall code@>
@<SchemeExit code@>

@<PerlInit code@>
@<PerlCall code@>
@<PerlExit code@>
#ifdef SUBSCRIPT_PERL
@<Perl external extensions callback code@>
#endif

@<SchemeMakeAssoc code@>
@<SchemeAssocSize code@>
@<SchemeAssocRef code@>
@<SchemeAssocSet code@>
@<SchemeDelAssoc code@>
@<HashCleanEntry code@>

@<SchemeRunCommandRegister code@>

@<SchemeMacRegister code@>
@<SchemeMacDeRegister code@>
@<SchemeMac code@>
@<SchemeDeMac code@>
@<SchemeMacCallback code@>

@<SchemeTimerRegister code@>
@<SchemeTimerDeRegister code@>
@<SchemeTimer code@>
@<SchemeDeTimer code@>
@<SchemeTimerCallback code@>

@<SchemeChannelRegister code@>
@<SchemeChannelDeRegister code@>
@<SchemeChannel code@>
@<SchemeDeChannel code@>
@<SchemeChannelCallback code@>
@<SchemeChannelDeRegisterCallback code@>

@<SchemeFilterRegister code@>
@<SchemeFilterDeRegister code@>
@<SchemeFilter code@>
@<SchemeDeFilter code@>
@<SchemeFilterCallback code@>
@<SchemeFilterDeRegisterCallback code@>

@<SchemeGeneratorRegister code@>
@<SchemeGeneratorSend code@>
@<SchemeEventGenerate code@>
@<SchemeGeneratorDeRegister code@>
@<SchemeGenerator code@>
@<SchemeDeGenerator code@>
@<SchemeGeneratorDeRegisterCallback code@>

@<SchemeEventRegister code@>
@<SchemeEventDeRegister code@>
@<SchemeEvent code@>
@<SchemeDeEvent code@>
@<SchemeEventCallback code@>
@<SchemeEventDeRegisterCallback code@>

@<SchemeMibRequest code@>
@<SchemeMibGet code@>
@<SchemeMibSet code@>
@<SchemeMibNext code@>
@<SchemeMibCallback code@>
@<SchemeMib code@>
@<SchemeDeMib code@>
@<SchemeMibRegister code@>
@<SchemeMibDeRegister code@>
@<SchemeMibRegisterArgs code@>
@<SchemeMibRegisterResults code@>

@<Varbinds2Objects code@>
@<Objects2Varbinds code@>

@<SnmpObjectPutId code@>
@<SnmpObjectPutType code@>
@<SnmpObjectPutSyntax code@>
@<SnmpObjectGetId code@>
@<SnmpObjectGetType code@>
@<SnmpObjectGetSyntax code@>
@<SnmpObjectGetRequest code@>

@<SchemeSnmpConnect code@>
@<SchemeSnmp code@>
@<SchemeDeSnmp code@>
@<SchemeSnmpSend code@>
@<SchemeSnmpDisConnect code@>
@<SchemeSnmpCallback code@>

@<SchemeTrapRegister code@>
@<SchemeTrap code@>
@<SchemeDeTrap code@>
@<SchemeTrapDeRegister code@>
@<SchemeTrapCallback code@>

@<SchemeTrapSend code@>

@<SuspendCallbacks code@>
@<RemoveAllCallbacks code@>
@<AddCallback code@>
@<SearchCallback code@>
@<RemoveCallback code@>
@<CutCallback code@>

@<SchemeNew code@>
@<SchemeDel code@>

#ifdef SUBSCRIPT_PERL
@<PerlNew code@>
@<PerlDel code@>
#endif

@<SchemeContextInit code@>
@<SchemeContextSwitch code@>
@<SchemeContextCurrent code@>

#ifdef SUBSCRIPT_PERL
@<PerlContextInit code@>
@<PerlContextSwitch code@>
@<PerlContextCurrent code@>
#endif

@<SchemeInputGetc code@>
@<SchemeInputUnGetc code@>
@<SchemeInputPoll code@>
@<SchemeInputClose code@>
@<SchemeOutputPuts code@>
@<SchemeOutputClose code@>

@<scheme_make_packet code@>
@}


@d Language function prototype definitions
@{
typedef CHAR *(*RunGetsFunction)(VOID *run, int line);
typedef int (*RunPutsFunction)(VOID *run, int line, CHAR *log);
@}


@d Language type definitions
@{
typedef struct LanguageCallback_ LanguageCallback;

typedef struct _LANGUAGE_MIB_OBJECT
{
    MIB_OBJECT          *object;
    int                 offset;
} LANGUAGE_MIB_OBJECT;

struct LanguageCallback_
{
	int id;
	BOOLEAN suspend;

	VOID *language;
	LanguageCallback *prev;
	LanguageCallback *next;

	VOID *callback;
	VOID *param;
	union
	{
		TIMER_DESCR  *timer;
		struct
		{
			int interface;
			MAC_COLL coll;
		} mac;
		struct
		{
			int index;
			SinkCallback  *callback;
		} channel;
		struct
		{
			int index;
			FilterCallback  *callback;
		} filter;
		struct
		{
			int index;
			EventHook  *hook;
		} generator;
		struct
		{
			int index;
			PartyCallback  *callback;
		} event;
        LANGUAGE_MIB_OBJECT mib;
		struct
		{
			int request;
			NW_DG *conn;
		} snmp;
		NW_DG *accept;
	} callbacks;
};
@}


\section{Scheme}

@d Scheme type definitions
@{
typedef struct Scheme_
{
    VOID				*run;
    RunGetsFunction		rungets;
    RunPutsFunction		runputs;

    Scheme_Env			*env;

    Scheme_Object		*init;
    Scheme_Object		*start;
    Scheme_Object		*stop;
    Scheme_Object		*resume;
    Scheme_Object		*exit;

	LanguageCallback	*macs;
	LanguageCallback	*timers;
	LanguageCallback	*filters;
	LanguageCallback	*channels;
	LanguageCallback	*generators;
	LanguageCallback	*events;
	LanguageCallback	*mibs;
	LanguageCallback	*snmps;
	LanguageCallback	*traps;

	Scheme_Object		*stdinport;
	Scheme_Object		*stdoutport;
    CHAR				*stdindata;
    int					stdinline;
    int					stdinpos;
    CHAR				*stdoutdata;
    int					stdoutline;
    int					stdoutpos;

	Scheme_Object		*hash_type;
	Scheme_Object		*packet_type;
	Scheme_Object		*varbind_type;

} Scheme;
@}


@d Scheme type definitions
@{
typedef struct Scheme_Assoc_Entry_
{
	BYTE key[SCHEME_HASH_SIZE_KEY];
	Scheme_Object *data;

} Scheme_Assoc_Entry;
@}


@d Scheme size definitions
@{
#define SCHEME_HASH_SIZE_KEY		32
@}


@d Scheme macro definitions
@{
#define SCHEME_HASHP(obj)		(SCHEME_TYPE(obj) == scheme->hash_type)
#define SCHEME_VARBINDP(obj)	(SCHEME_TYPE(obj) == scheme->varbind_type)
@}


@d Scheme static variables
@{
Scheme *CurrentScheme = NULL;

char scheme_string_buf[1024];
@}


@d SchemeInit prototype
@{
VOID *SchemeInit(RunGetsFunction rungets,
                 RunPutsFunction runputs,
                 VOID *run);
@}

@d SchemeInit code
@{
VOID *SchemeInit(RunGetsFunction rungets,
                 RunPutsFunction runputs,
				 VOID *run)
{
    Scheme *scheme;
    Scheme_Object *obj;
	int expression;

	SchemeContextInit();

    /* allocate a Scheme structure */
    if ((scheme = SchemeNew()) == NULL)
        return NULL;
    scheme->rungets = rungets;
    scheme->runputs = runputs;
    scheme->run = run;

	scheme->hash_type = scheme_make_type("<hash>");
	scheme->packet_type = scheme_make_type("<packet>");
    scheme_add_global("make-hash", scheme_make_prim(SchemeMakeAssoc), scheme->env);
    scheme_add_global("hash-size", scheme_make_prim(SchemeAssocSize), scheme->env);
    scheme_add_global("hash-ref", scheme_make_prim(SchemeAssocRef), scheme->env);
    scheme_add_global("hash-set!", scheme_make_prim(SchemeAssocSet), scheme->env);
    scheme_add_global("del-hash", scheme_make_prim(SchemeDelAssoc), scheme->env);

    scheme_add_global("command-register", scheme_make_prim(SchemeRunCommandRegister), scheme->env);

    scheme_add_global("mac-register", scheme_make_prim(SchemeMacRegister), scheme->env);
    scheme_add_global("mac-deregister", scheme_make_prim(SchemeMacDeRegister), scheme->env);
    scheme_add_global("timer-register", scheme_make_prim(SchemeTimerRegister), scheme->env);
    scheme_add_global("timer-deregister", scheme_make_prim(SchemeTimerDeRegister), scheme->env);
    scheme_add_global("channel-register", scheme_make_prim(SchemeChannelRegister), scheme->env);
    scheme_add_global("channel-deregister", scheme_make_prim(SchemeChannelDeRegister), scheme->env);
    scheme_add_global("filter-register", scheme_make_prim(SchemeFilterRegister), scheme->env);
    scheme_add_global("filter-deregister", scheme_make_prim(SchemeFilterDeRegister), scheme->env);
    scheme_add_global("generator-register", scheme_make_prim(SchemeGeneratorRegister), scheme->env);
    scheme_add_global("generator-send", scheme_make_prim(SchemeGeneratorSend), scheme->env);
    scheme_add_global("generator-deregister", scheme_make_prim(SchemeGeneratorDeRegister), scheme->env);
    scheme_add_global("event-register", scheme_make_prim(SchemeEventRegister), scheme->env);
    scheme_add_global("event-deregister", scheme_make_prim(SchemeEventDeRegister), scheme->env);

	scheme->varbind_type = scheme_define_struct(scheme->env, "varbind", "id", "type", "syntax", NULL);

	scheme_add_global("mib-get", scheme_make_prim(SchemeMibGet), scheme->env);
	scheme_add_global("mib-set", scheme_make_prim(SchemeMibSet), scheme->env);
	scheme_add_global("mib-next", scheme_make_prim(SchemeMibNext), scheme->env);
	scheme_add_global("snmp-connect", scheme_make_prim(SchemeSnmpConnect), scheme->env);
	scheme_add_global("snmp-send", scheme_make_prim(SchemeSnmpSend), scheme->env);
	scheme_add_global("snmp-deconnect", scheme_make_prim(SchemeSnmpDisConnect), scheme->env);
	scheme_add_global("trap-register", scheme_make_prim(SchemeTrapRegister), scheme->env);
	scheme_add_global("trap-deregister", scheme_make_prim(SchemeTrapDeRegister), scheme->env);
	scheme_add_global("trap-send", scheme_make_prim(SchemeTrapSend), scheme->env);
	scheme_add_global("mib-register-args",
					  scheme_make_prim(SchemeMibRegisterArgs), scheme->env);
	scheme_add_global("mib-register-results",
					  scheme_make_prim(SchemeMibRegisterResults), scheme->env);
	scheme_add_global("mib-deregister",
					  scheme_make_prim(SchemeMibDeRegister), scheme->env);

    scheme->stdindata = NULL;
    scheme->stdinline = 0;
    scheme->stdinpos = 0;
    scheme->stdoutdata = NULL;
    scheme->stdoutline = 0;
    scheme->stdoutpos = 0;

    if ((scheme->stdinport = scheme_alloc_object()) == NULL ||
		(SCHEME_PTR_VAL(scheme->stdinport) =
			scheme_make_input_port(NULL, scheme,
				SchemeInputGetc, SchemeInputUnGetc,
				SchemeInputPoll, SchemeInputClose)) == NULL)
	{
		SchemeDel(scheme);
		return NULL;
	}
	SCHEME_TYPE(scheme->stdinport) = scheme_input_port_type;

    if ((scheme->stdoutport = scheme_alloc_object()) == NULL ||
		(SCHEME_PTR_VAL(scheme->stdoutport) =
			scheme_make_output_port(NULL, scheme,
				SchemeOutputPuts, SchemeOutputClose)) == NULL)
	{
		SchemeDel(scheme);
		return NULL;
	}
	SCHEME_TYPE(scheme->stdoutport) = scheme_output_port_type;

    if ((scheme = SchemeContextSwitch(scheme)) == NULL)
	{
		SchemeDel(scheme);
		return NULL;
	}

    for (expression = 1; ; expression++)
    {
        obj = scheme_read(scheme->stdinport);
        if (obj == scheme_eof)
            break;
        obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env) ,0);
        if (obj == NULL)
		{
			sprintf(scheme_string_buf, "error evaluating expression %d", expression);
			scheme_write_string(scheme_string_buf, scheme->stdoutport);
			SchemeDel(scheme);
			return NULL;
		}
		else
        {
            scheme_write(obj, scheme->stdoutport);
        }
    }

    return scheme;
}
@}


@d SchemeContextInit prototype
@{
VOID SchemeContextInit(VOID);
@}

@d SchemeContextInit code
@{
VOID SchemeContextInit(VOID)
{
	CurrentScheme = NULL;
}
@}


@d SchemeContextSwitch prototype
@{
Scheme *SchemeContextSwitch(Scheme *scheme);
@}

@d SchemeContextSwitch code
@{
Scheme *SchemeContextSwitch(Scheme *scheme)
{
	CurrentScheme = scheme;
	if (CurrentScheme == NULL)
	{
		DnpapMessage(DMC_FATAL, SUBSCRIPT_CONTEXT, "subscript: no context available");
		return NULL;
	}
	scheme_stdin_port = CurrentScheme->stdinport;
	scheme_stdout_port = CurrentScheme->stdoutport;
	scheme_stderr_port = CurrentScheme->stdoutport;
	scheme_set_current_ports(scheme_stdin_port, scheme_stdout_port);
	return CurrentScheme;
}
@}


@d SchemeContextCurrent prototype
@{
Scheme *SchemeContextCurrent(VOID);
@}

@d SchemeContextCurrent code
@{
Scheme *SchemeContextCurrent(VOID)
{
	return CurrentScheme;
}
@}


@d SchemeCall prototype
@{
BOOLEAN SchemeCall(Scheme *scheme, int command);
@}

@d SchemeCall code
@{
BOOLEAN SchemeCall(Scheme *scheme, int command)
{
    Scheme_Object *obj = NULL;

	if ((scheme = SchemeContextSwitch(scheme)) == NULL)
		return FALSE;

    switch (command)
    {
    case RUN_COMMAND_INIT:
        if (scheme->init == NULL)
            return TRUE;
		if (!SCHEME_PROCP(scheme->init))
            return FALSE;
        obj = scheme_make_pair(scheme->init, scheme_null);
        obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env), 0);
        break;
    case RUN_COMMAND_START:
        if (scheme->start == NULL)
            return TRUE;
		if (!SCHEME_PROCP(scheme->start))
            return FALSE;
        obj = scheme_make_pair(scheme->start, scheme_null);
        obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env), 0);
        break;
    case RUN_COMMAND_STOP:
    	SuspendCallbacks(scheme, SCRIPT_LANGUAGE_SCHEME, TRUE);
        if (scheme->stop == NULL)
            return TRUE;
		if (!SCHEME_PROCP(scheme->stop))
            return FALSE;
        obj = scheme_make_pair(scheme->stop, scheme_null);
        obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env), 0);
        break;
    case RUN_COMMAND_RESUME:
    	SuspendCallbacks(scheme, SCRIPT_LANGUAGE_SCHEME, FALSE);
        if (scheme->resume == NULL)
            return TRUE;
		if (!SCHEME_PROCP(scheme->resume))
            return FALSE;
        obj = scheme_make_pair(scheme->resume, scheme_null);
        obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env), 0);
        break;
    case RUN_COMMAND_EXIT:
        if (scheme->exit == NULL)
            return TRUE;
		if (!SCHEME_PROCP(scheme->exit))
            return FALSE;
        obj = scheme_make_pair(scheme->exit, scheme_null);
        obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env), 0);
        break;
    default:
        return FALSE;
    }

	if (obj == NULL)
	{
		sprintf(scheme_string_buf, "error executing command %d", command);
		scheme_write_string(scheme_string_buf, scheme->stdoutport);
        return FALSE;
	}

	return TRUE;
}
@}


@d SchemeExit prototype
@{
VOID SchemeExit(Scheme *scheme);
@}

@d SchemeExit code
@{
VOID SchemeExit(Scheme *scheme)
{
	if ((scheme = SchemeContextSwitch(scheme)) == NULL)
		return;

   	RemoveAllCallbacks(scheme, SCRIPT_LANGUAGE_SCHEME);

	SchemeDel(scheme);
}
@}


@d SchemeNew prototype
@{
Scheme *SchemeNew(VOID);
@}

@d SchemeNew code
@{
Scheme *SchemeNew(VOID)
{
    Scheme *scheme;

    if ((scheme = DnpapMalloc(sizeof(Scheme))) == NULL)
        return NULL;

    scheme->env = NULL;
    scheme->init = NULL;
    scheme->start = NULL;
    scheme->stop = NULL;
    scheme->resume = NULL;
    scheme->exit = NULL;

    scheme->macs = NULL;
    scheme->timers = NULL;
    scheme->filters = NULL;
    scheme->channels = NULL;
    scheme->generators = NULL;
    scheme->events = NULL;
    scheme->mibs = NULL;
    scheme->snmps = NULL;
    scheme->traps = NULL;

    scheme->stdinport = NULL;
    scheme->stdoutport = NULL;
    scheme->stdindata = NULL;
    scheme->stdinline = 0;
    scheme->stdinpos = 0;
    scheme->stdoutdata = NULL;
    scheme->stdoutline = 0;
    scheme->stdoutpos = 0;

	scheme->hash_type = NULL;
	scheme->packet_type = NULL;
	scheme->varbind_type = NULL;

    /* create a new scheme environment */
    if ((scheme->env = scheme_basic_env()) == NULL)
    {
        SchemeDel(scheme);
        return NULL;
    }

    return scheme;
}
@}


@d SchemeDel prototype
@{
VOID SchemeDel(Scheme *scheme);
@}

@d SchemeDel code
@{
VOID SchemeDel(Scheme *scheme)
{
    DnpapFree(scheme);
}
@}


@d SchemeInputGetc prototype
@{
int SchemeInputGetc(Scheme_Input_Port *port);
@}

@d SchemeInputGetc code
@{
int SchemeInputGetc(Scheme_Input_Port *port)
{
Scheme *scheme = (Scheme *)port->port_data;

	scheme->stdinpos++;

	while (scheme->stdinline == 0 || scheme->stdindata == NULL ||
			scheme->stdindata[scheme->stdinpos] == '\0')
	{
		if ((scheme->stdindata = scheme->rungets(scheme->run, scheme->stdinline+1)) == NULL)
			return EOF;
		scheme->stdinline++;
		scheme->stdinpos = 0;
	}

	return scheme->stdindata[scheme->stdinpos];
}
@}


@d SchemeInputUnGetc prototype
@{
void SchemeInputUnGetc(int c, Scheme_Input_Port *port);
@}

@d SchemeInputUnGetc code
@{
void SchemeInputUnGetc(int c, Scheme_Input_Port *port)
{
Scheme *scheme = (Scheme *)port->port_data;

	if (scheme->stdinline == 0 || scheme->stdindata == NULL ||
		c == '\0' || c == EOF)
		return;

	scheme->stdindata[scheme->stdinpos] = c;

	scheme->stdinpos--;

	while (scheme->stdinpos < 0)
	{
		if (scheme->stdinline > 1)
		{
			scheme->stdinline--;
			scheme->stdindata = scheme->rungets(scheme->run, scheme->stdinline);
			scheme->stdinpos = strlen(scheme->stdindata)-1;
		}
		else
		{
			scheme->stdinline = 0;
			scheme->stdindata = NULL;
			scheme->stdinpos = 0;
		}
	}
}
@}


@d SchemeInputPoll prototype
@{
int SchemeInputPoll(Scheme_Input_Port *port);
@}

@d SchemeInputPoll code
@{
int SchemeInputPoll(Scheme_Input_Port *port)
{
int c;

	if ((c = SchemeInputGetc(port)) == EOF)
		return 0;
	SchemeInputUnGetc(c, port);
	return 1;
}
@}


@d SchemeInputClose prototype
@{
void SchemeInputClose(Scheme_Input_Port *port);
@}

@d SchemeInputClose code
@{
void SchemeInputClose(Scheme_Input_Port *port)
{
	/* do nothing */
}
@}


@d SchemeOutputPuts prototype
@{
void SchemeOutputPuts(char *s, Scheme_Output_Port *port);
@}

@d SchemeOutputPuts code
@{
void SchemeOutputPuts(char *s, Scheme_Output_Port *port)
{
Scheme *scheme = (Scheme *)port->port_data;

	scheme->runputs(scheme->run, ++scheme->stdoutline, s);
}
@}


@d SchemeOutputClose prototype
@{
void SchemeOutputClose(Scheme_Output_Port *port);
@}

@d SchemeOutputClose code
@{
void SchemeOutputClose(Scheme_Output_Port *port)
{
	/* do nothing */
}
@}


@d SchemeMakeAssoc prototype
@{
Scheme_Object *SchemeMakeAssoc(int argc, Scheme_Object *argv[]);
@}

@d SchemeMakeAssoc code
@{
Scheme_Object *SchemeMakeAssoc(int argc, Scheme_Object *argv[])
{
	Scheme_Object *hash;
	HashTable *h;
    Scheme *scheme;

    if ((scheme = SchemeContextCurrent()) == NULL)
    	return scheme_null;

    if (argc != 1 || !SCHEME_INTP(argv[0]))
    {
    	scheme_write_string("make-hash: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_null;
    }

	if ((h = NewHash(SCHEME_INT_VAL(argv[0]), NULL)) == NULL)
		return scheme_null;
	if ((hash = scheme_alloc_object()) == NULL)
	{
		DelHash(h);
		return scheme_null;
	}
	SCHEME_TYPE(hash) = scheme->hash_type;
	SCHEME_PTR_VAL(hash) = h;

	return hash;
}
@}


@d SchemeAssocSize prototype
@{
Scheme_Object *SchemeAssocSize(int argc, Scheme_Object *argv[]);
@}

@d SchemeAssocSize code
@{
Scheme_Object *SchemeAssocSize(int argc, Scheme_Object *argv[])
{
    Scheme *scheme;

    if ((scheme = SchemeContextCurrent()) == NULL)
    	return scheme_null;

    if (argc != 1 || !SCHEME_HASHP(argv[0]))
    {
    	scheme_write_string("hash-size: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_null;
    }

	return scheme_make_integer(HashSize((HashTable *)SCHEME_PTR_VAL(argv[0])));
}
@}


@d SchemeAssocRef prototype
@{
Scheme_Object *SchemeAssocRef(int argc, Scheme_Object *argv[]);
@}

@d SchemeAssocRef code
@{
Scheme_Object *SchemeAssocRef(int argc, Scheme_Object *argv[])
{
	Scheme_Assoc_Entry *elem;
    Scheme *scheme;
	int i, keylen = 0;
	BYTE vector[SCHEME_HASH_SIZE_KEY];
	BYTE *key = NULL;
	Scheme_Object **elems;

    if ((scheme = SchemeContextCurrent()) == NULL)
    	return scheme_null;

    if (argc != 2 || !SCHEME_HASHP(argv[0]) ||
		(!SCHEME_STRINGP(argv[1]) && !SCHEME_VECTORP(argv[1])))
    {
    	scheme_write_string("hash-ref: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_null;
    }

	if (SCHEME_STRINGP(argv[1]))
	{
		key =  SCHEME_STR_VAL(argv[1]);
		keylen = strlen(key);
	}
	else
	if (SCHEME_VECTORP(argv[1]))
	{
		key = vector;
		keylen = SCHEME_VEC_SIZE(argv[1]);
		elems = SCHEME_VEC_ELS(argv[1]);
		for (i = 0; i < keylen; i++)
		{
			if (SCHEME_INTP(elems[i]))
				vector[i] = SCHEME_INT_VAL(elems[i]);
			else
			if (SCHEME_CHARP(elems[i]))
				vector[i] = SCHEME_CHAR_VAL(elems[i]);
			else
				return scheme_null;
		}
	}

	if ((elem = HashSearch((HashTable *)SCHEME_PTR_VAL(argv[0]), key, keylen)) == NULL)
		return scheme_null;

	return elem->data;
}
@}


@d SchemeAssocSet prototype
@{
Scheme_Object *SchemeAssocSet(int argc, Scheme_Object *argv[]);
@}

@d SchemeAssocSet code
@{
Scheme_Object *SchemeAssocSet(int argc, Scheme_Object *argv[])
{
    Scheme *scheme;
	HashTable *hash;
	int i, keylen = 0;
	BYTE vector[SCHEME_HASH_SIZE_KEY];
	BYTE *key = NULL;
	Scheme_Object **elems;
	Scheme_Assoc_Entry *elem;

    if ((scheme = SchemeContextCurrent()) == NULL)
    	return scheme_null;

    if (argc != 3 || !SCHEME_HASHP(argv[0]) ||
		(!SCHEME_STRINGP(argv[1]) && !SCHEME_VECTORP(argv[1])))
    {
    	scheme_write_string("hash-set!: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_null;
    }
	if (SCHEME_STRINGP(argv[1]))
	{
		key =  SCHEME_STR_VAL(argv[1]);
		keylen = strlen(key);
	}
	else
	if (SCHEME_VECTORP(argv[1]))
	{
		key = vector;
		keylen = SCHEME_VEC_SIZE(argv[1]);
		elems = SCHEME_VEC_ELS(argv[1]);
		for (i = 0; i < keylen; i++)
		{
			if (SCHEME_INTP(elems[i]))
				vector[i] = SCHEME_INT_VAL(elems[i]);
			else
			if (SCHEME_CHARP(elems[i]))
				vector[i] = SCHEME_CHAR_VAL(elems[i]);
			else
				return scheme_null;
		}
	}
    if (keylen > SCHEME_HASH_SIZE_KEY)
    {
    	scheme_write_string("hash-set!: key too long", scheme->stdoutport);
    	return scheme_null;
	}

	hash = (HashTable *)SCHEME_PTR_VAL(argv[0]);
	if ((elem = HashSearch(hash, key, keylen)) != NULL)
	{
		elem->data = argv[2];
		return argv[2];
	}
	if ((elem = DnpapMalloc(sizeof(Scheme_Assoc_Entry))) == NULL)
		return scheme_null;
	memcpy(elem->key, key, keylen);
	elem->data = argv[2];
	if (HashAdd(hash, elem->key, keylen, elem) == NULL)
	{
		DnpapFree(elem);
		return scheme_null;
	}

	return argv[2];
}
@}


@d SchemeDelAssoc prototype
@{
Scheme_Object *SchemeDelAssoc(int argc, Scheme_Object *argv[]);
@}

@d SchemeDelAssoc code
@{
Scheme_Object *SchemeDelAssoc(int argc, Scheme_Object *argv[])
{
    Scheme *scheme;
	HashTable *hash;

    if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

    if (argc != 1 || !SCHEME_HASHP(argv[0]))
    {
    	scheme_write_string("del-hash: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	hash = (HashTable *)SCHEME_PTR_VAL(argv[0]);
	CleanHash(hash, HashCleanEntry);
	DelHash(hash);

	return scheme_true;
}
@}

@d HashCleanEntry prototype
@{
static VOID HashCleanEntry(HashTable *ht, VOID *entry);
@}

@d HashCleanEntry code
@{
VOID HashCleanEntry(HashTable *ht, VOID *entry)
{
Scheme_Assoc_Entry *elem = (Scheme_Assoc_Entry *) entry;

	HashRemove(ht, elem->key, strlen(elem->key));
}
@}


@d scheme_make_packet prototype
@{
Scheme_Object *scheme_make_packet(PROT_PKT *pkt);
@}

@d scheme_make_packet code
@{
Scheme_Object *scheme_make_packet(PROT_PKT *pkt)
{
Scheme_Object *packet;
Scheme *scheme;

    if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_null;

	if ((packet = scheme_alloc_object()) == NULL)
		return scheme_null;
	SCHEME_TYPE(packet) = scheme->packet_type;
	SCHEME_PTR_VAL(packet) = pkt;

	return packet;
}
@}


@d SchemeRunCommandRegister prototype
@{
Scheme_Object *SchemeRunCommandRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeRunCommandRegister code
@{
Scheme_Object *SchemeRunCommandRegister(int argc, Scheme_Object *argv[])
{
    Scheme *scheme;

    if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

	if (SubscriptRunCommand(scheme->run) != RUN_COMMAND_NULL)
    {
    	scheme_write_string("command-register: can only be called during initialization", scheme->stdoutport);
    	return scheme_false;
    }

    if (argc != 2 || !SCHEME_STRINGP(argv[0]) || !SCHEME_PROCP(argv[1]))
    {
    	scheme_write_string("command-register: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

    if (strcasecmp("init", SCHEME_STR_VAL(argv[0])) == 0)
        scheme->init = argv[1];
    else
    if (strcasecmp("start", SCHEME_STR_VAL(argv[0])) == 0)
        scheme->start = argv[1];
    else
    if (strcasecmp("stop", SCHEME_STR_VAL(argv[0])) == 0)
        scheme->stop = argv[1];
    else
    if (strcasecmp("resume", SCHEME_STR_VAL(argv[0])) == 0)
        scheme->resume = argv[1];
    else
    if (strcasecmp("exit", SCHEME_STR_VAL(argv[0])) == 0)
        scheme->exit = argv[1];
    else
        return scheme_false;

    return scheme_true;
}
@}


@d SchemeMacRegister prototype
@{
Scheme_Object *SchemeMacRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeMacRegister code
@{
Scheme_Object *SchemeMacRegister(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
LanguageCallback *mac;
int command;

    if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

	if ((command = SubscriptRunCommand(scheme->run)) != RUN_COMMAND_INIT &&
		command != RUN_COMMAND_STOP)
    {
    	scheme_write_string("mac-register: can only be called during 'start' or 'resume'", scheme->stdoutport);
    	return scheme_false;
    }

    if (argc != 3 || !SCHEME_PROCP(argv[0]) || !SCHEME_INTP(argv[1]))
    {
    	scheme_write_string("mac-register: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if ((mac = SchemeMac(scheme, argv[0], SCHEME_INT_VAL(argv[1]), argv[2])) == NULL)
    {
    	scheme_write_string("mac-register: can not register mac", scheme->stdoutport);
    	return scheme_false;
    }

	if (AddCallback(&scheme->macs, mac) == FALSE)
    {
    	scheme_write_string("mac-register: can not add mac", scheme->stdoutport);
    	return scheme_false;
    }

    return scheme_make_integer(mac->id);
}
@}

@d SchemeMac prototype
@{
LanguageCallback *SchemeMac(Scheme *scheme, Scheme_Object *obj, int interface, Scheme_Object *param);
@}

@d SchemeMac code
@{
LanguageCallback *SchemeMac(Scheme *scheme, Scheme_Object *obj, int interface, Scheme_Object *param)
{
LanguageCallback *mac;

	if (MacIfaceGet(interface) == NULL)
		return NULL;
	if ((mac = DnpapMalloc(sizeof(LanguageCallback))) == NULL)
		return NULL;
	mac->id = (int) mac;
	mac->language = scheme;
	mac->callback = obj;
	mac->suspend = FALSE;
	mac->param = param;
	mac->callbacks.mac.coll.Rcve = SchemeMacCallback;
	mac->callbacks.mac.coll.specific = mac;
	mac->callbacks.mac.interface = interface;
	mac->prev = NULL;
	mac->next = NULL;
	if (MacCollRegister(&mac->callbacks.mac.coll) == FALSE)
	{
    	DnpapFree(mac);
        return NULL;
    }
	return mac;
}
@}

@d SchemeDeMac prototype
@{
BOOLEAN SchemeDeMac(LanguageCallback *mac);
@}

@d SchemeDeMac code
@{
BOOLEAN SchemeDeMac(LanguageCallback *mac)
{
    MacCollRemove(&mac->callbacks.mac.coll);
    DnpapFree(mac);
	return TRUE;
}
@}

@d SchemeMacDeRegister prototype
@{
Scheme_Object *SchemeMacDeRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeMacDeRegister code
@{
Scheme_Object *SchemeMacDeRegister(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
LanguageCallback *mac;

    if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

    if (argc != 1 || !SCHEME_INTP(argv[0]))
    {
    	scheme_write_string("mac-deregister: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if ((mac = RemoveCallback(&scheme->macs, SCHEME_INT_VAL(argv[0]))) == NULL)
	   	return scheme_false;
	SchemeDeMac(mac);

   	return scheme_true;
}
@}

@d SchemeMacCallback prototype
@{
BOOLEAN SchemeMacCallback(MAC_COLL *coll, PROT_PKT *pkt);
@}

@d SchemeMacCallback code
@{
BOOLEAN SchemeMacCallback(MAC_COLL *coll, PROT_PKT *pkt)
{
LanguageCallback *mac = (LanguageCallback *) coll->specific;
Scheme *scheme;
Scheme_Object *obj;
PROT_OBJ interface = {0, {1,2}};

    if ((scheme = SchemeContextSwitch(mac->language)) == NULL)
		return FALSE;

    if (mac->callback == NULL || mac->suspend == TRUE)
        return TRUE;
	if (ProtGetField(pkt, &interface) == FALSE)
		return TRUE;
    if (mac->callbacks.mac.interface != interface.Syntax.LngInt)
		return TRUE;

    obj = scheme_make_pair(mac->param, scheme_null);
    obj = scheme_make_pair(scheme_make_packet(pkt), obj);
    obj = scheme_make_pair(scheme_make_integer(mac->callbacks.mac.interface), obj);
    obj = scheme_make_pair(scheme_make_integer(mac->id), obj);
    obj = scheme_make_pair(mac->callback, obj);
    obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env), 0);

    if (obj == NULL)
	{
		scheme_write_string("error executing mac callback", scheme->stdoutport);
		return FALSE;
	}

    return TRUE;
}
@}


@d SchemeTimerRegister prototype
@{
Scheme_Object *SchemeTimerRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeTimerRegister code
@{
Scheme_Object *SchemeTimerRegister(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
LanguageCallback *timer;
int command;

    if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

	if ((command = SubscriptRunCommand(scheme->run)) != RUN_COMMAND_INIT &&
		command != RUN_COMMAND_STOP)
    {
    	scheme_write_string("timer-register: can only be called during 'start' or 'resume'", scheme->stdoutport);
    	return scheme_false;
    }

    if (argc != 4 || !SCHEME_PROCP(argv[0]) || !SCHEME_INTP(argv[1]) || !SCHEME_INTP(argv[2]))
    {
    	scheme_write_string("timer-register: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if ((timer = SchemeTimer(scheme, argv[0], SCHEME_INT_VAL(argv[1]), TIMER_FOREVER, SCHEME_INT_VAL(argv[2]), argv[3])) == NULL)
    {
    	scheme_write_string("timer-register: can not register timer", scheme->stdoutport);
    	return scheme_false;
    }

	if (AddCallback(&scheme->timers, timer) == FALSE)
    {
    	scheme_write_string("timer-register: can not add timer", scheme->stdoutport);
    	return scheme_false;
    }

	return scheme_make_integer(timer->id);
}
@}

@d SchemeTimer prototype
@{
LanguageCallback *SchemeTimer(Scheme *scheme, Scheme_Object *obj, int period, ULONG count, int type, Scheme_Object *param);
@}

@d SchemeTimer code
@{
LanguageCallback *SchemeTimer(Scheme *scheme, Scheme_Object *obj, int period, ULONG count, int type, Scheme_Object *param)
{
LanguageCallback *timer;

	if ((timer = DnpapMalloc(sizeof(LanguageCallback))) == NULL)
		return NULL;
	timer->id = (int) timer;
	timer->language = scheme;
	timer->callback = obj;
	timer->param = param;
	timer->suspend = FALSE;
	timer->prev = NULL;
	timer->next = NULL;
	timer->callbacks.timer = TimerRegister(SchemeTimerCallback, timer,
										   (ULONG)period, count, type);
	if (timer->callbacks.timer == NULL)
	{
		DnpapFree(timer);
		return NULL;
	}
	return timer;
}
@}

@d SchemeDeTimer prototype
@{
BOOLEAN SchemeDeTimer(LanguageCallback *timer);
@}

@d SchemeDeTimer code
@{
BOOLEAN SchemeDeTimer(LanguageCallback *timer)
{
	TimerRemove(timer->callbacks.timer);
    DnpapFree(timer);
	return TRUE;
}
@}

@d SchemeTimerDeRegister prototype
@{
Scheme_Object *SchemeTimerDeRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeTimerDeRegister code
@{
Scheme_Object *SchemeTimerDeRegister(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
LanguageCallback *timer;

    if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

    if (argc != 1 || !SCHEME_INTP(argv[0]))
    {
    	scheme_write_string("timer-deregister: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if ((timer = RemoveCallback(&scheme->timers, SCHEME_INT_VAL(argv[0]))) == NULL)
	   	return scheme_false;
	SchemeDeTimer(timer);

   	return scheme_true;
}
@}

@d SchemeTimerCallback prototype
@{
VOID SchemeTimerCallback(TIMER_DESCR *descr, ULONG now, VOID *param);
@}

@d SchemeTimerCallback code
@{
VOID SchemeTimerCallback(TIMER_DESCR *descr, ULONG now, VOID *param)
{
LanguageCallback *timer = (LanguageCallback *) param;
Scheme *scheme;
Scheme_Object *obj;

    if ((scheme = SchemeContextSwitch(timer->language)) == NULL)
		return;

    if (timer->callback == NULL || timer->suspend == TRUE)
        return;

    obj = scheme_make_pair(timer->param, scheme_null);
    obj = scheme_make_pair(scheme_make_integer(now), obj);
    obj = scheme_make_pair(scheme_make_integer(timer->id), obj);
    obj = scheme_make_pair(timer->callback, obj);
    obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env), 0);

	if (obj == NULL)
	{
		scheme_write_string("error executing timer callback", scheme->stdoutport);
	}
}
@}


@d SchemeChannelRegister prototype
@{
Scheme_Object *SchemeChannelRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeChannelRegister code
@{
Scheme_Object *SchemeChannelRegister(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
LanguageCallback *channel;
int command;

    if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

	if ((command = SubscriptRunCommand(scheme->run)) != RUN_COMMAND_INIT &&
		command != RUN_COMMAND_STOP)
    {
    	scheme_write_string("channel-register: can only be called during 'start' or 'resume'", scheme->stdoutport);
    	return scheme_false;
    }

    if (argc != 3 || !SCHEME_PROCP(argv[0]) || !SCHEME_INTP(argv[1]))
    {
    	scheme_write_string("channel-register: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if ((channel = SchemeChannel(scheme, argv[0], SCHEME_INT_VAL(argv[1]), argv[2])) == NULL)
    {
    	scheme_write_string("channel-register: can not register channel", scheme->stdoutport);
    	return scheme_false;
    }

	if (AddCallback(&scheme->channels, channel) == FALSE)
    {
    	scheme_write_string("channel-register: can not add channel", scheme->stdoutport);
    	return scheme_false;
    }

	return scheme_make_integer(channel->id);
}
@}

@d SchemeChannel prototype
@{
LanguageCallback *SchemeChannel(Scheme *scheme, Scheme_Object *obj, int channelIndex, Scheme_Object *param);
@}

@d SchemeChannel code
@{
LanguageCallback *SchemeChannel(Scheme *scheme, Scheme_Object *obj, int channelIndex, Scheme_Object *param)
{
LanguageCallback *channel;

	if ((channel = DnpapMalloc(sizeof(LanguageCallback))) == NULL)
		return NULL;
	channel->id = (int) channel;
	channel->language = scheme;
	channel->callback = obj;
	channel->param = param;
	channel->suspend = FALSE;
	channel->prev = NULL;
	channel->next = NULL;
	channel->callbacks.channel.index = channelIndex;
	channel->callbacks.channel.callback = ChannelRegisterSink(channelIndex, SchemeChannelCallback, SchemeChannelDeRegisterCallback, channel);
	if (channel->callbacks.channel.callback == NULL)
	{
		DnpapFree(channel);
		return NULL;
	}
	return channel;
}
@}

@d SchemeDeChannel prototype
@{
BOOLEAN SchemeDeChannel(LanguageCallback *channel);
@}

@d SchemeDeChannel code
@{
BOOLEAN SchemeDeChannel(LanguageCallback *channel)
{
	ChannelDeRegisterSink(channel->callbacks.channel.index, channel->callbacks.channel.callback);
    DnpapFree(channel);
	return TRUE;
}
@}

@d SchemeChannelDeRegister prototype
@{
Scheme_Object *SchemeChannelDeRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeChannelDeRegister code
@{
Scheme_Object *SchemeChannelDeRegister(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
LanguageCallback *channel;

	if ((scheme = SchemeContextCurrent()) == NULL)
    	return scheme_false;

    if (argc != 1 || !SCHEME_INTP(argv[0]))
    {
    	scheme_write_string("channel-deregister: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if ((channel = RemoveCallback(&scheme->channels, SCHEME_INT_VAL(argv[0]))) == NULL)
	   	return scheme_false;
	SchemeDeChannel(channel);

   	return scheme_true;
}
@}

@d SchemeChannelCallback prototype
@{
VOID SchemeChannelCallback(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *param);
@}

@d SchemeChannelCallback code
@{
VOID SchemeChannelCallback(ULONG channelIndex, int type, PROT_PKT *pkt, VOID *param)
{
LanguageCallback *channel = (LanguageCallback *) param;
Scheme *scheme;
Scheme_Object *obj;

    if ((scheme = SchemeContextSwitch(channel->language)) == NULL)
    	return;

    if (channel->callback == NULL || channel->suspend == TRUE)
        return;

    obj = scheme_make_pair(channel->param, scheme_null);
    obj = scheme_make_pair(scheme_make_packet(pkt), obj);
    obj = scheme_make_pair(scheme_make_integer(type), obj);
    obj = scheme_make_pair(scheme_make_integer(channelIndex), obj);
    obj = scheme_make_pair(scheme_make_integer(channel->id), obj);
    obj = scheme_make_pair(channel->callback, obj);
    obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env), 0);

	if (obj == NULL)
	{
		scheme_write_string("error executing channel callback", scheme->stdoutport);
	}
}
@}

@d SchemeChannelDeRegisterCallback prototype
@{
VOID SchemeChannelDeRegisterCallback(ULONG channelIndex, VOID *param);
@}

@d SchemeChannelDeRegisterCallback code
@{
VOID SchemeChannelDeRegisterCallback(ULONG channelIndex, VOID *param)
{
LanguageCallback *channel = (LanguageCallback *) param;
Scheme *scheme;

    if ((scheme = SchemeContextSwitch(channel->language)) == NULL)
		return;

	if (RemoveCallback(&scheme->channels, channel->id) == channel)
		SchemeDeChannel(channel);
}
@}

@d SchemeFilterRegister prototype
@{
Scheme_Object *SchemeFilterRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeFilterRegister code
@{
Scheme_Object *SchemeFilterRegister(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
LanguageCallback *filter;
int command;

    if ((scheme = SchemeContextCurrent()) == NULL)
        return scheme_false;

	if ((command = SubscriptRunCommand(scheme->run)) != RUN_COMMAND_INIT &&
		command != RUN_COMMAND_STOP)
    {
    	scheme_write_string("filter-register: can only be called during 'start' or 'resume'", scheme->stdoutport);
    	return scheme_false;
    }

    if (argc != 3 || !SCHEME_PROCP(argv[0]) || !SCHEME_INTP(argv[1]))
    {
    	scheme_write_string("filter-register: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if ((filter = SchemeFilter(scheme, argv[0], SCHEME_INT_VAL(argv[1]), argv[2])) == NULL)
    {
    	scheme_write_string("filter-register: can not register filter", scheme->stdoutport);
    	return scheme_false;
    }

	if (AddCallback(&scheme->filters, filter) == FALSE)
    {
    	scheme_write_string("filter-register: can not add filter", scheme->stdoutport);
    	return scheme_false;
    }

	return scheme_make_integer(filter->id);
}
@}

@d SchemeFilter prototype
@{
LanguageCallback *SchemeFilter(Scheme *scheme, Scheme_Object *obj, int filterIndex, Scheme_Object *param);
@}

@d SchemeFilter code
@{
LanguageCallback *SchemeFilter(Scheme *scheme, Scheme_Object *obj, int filterIndex, Scheme_Object *param)
{
LanguageCallback *filter;

	if ((filter = DnpapMalloc(sizeof(LanguageCallback))) == NULL)
		return NULL;
	filter->id = (int) filter;
	filter->language = scheme;
	filter->callback = obj;
	filter->param = param;
	filter->suspend = FALSE;
	filter->prev = NULL;
	filter->next = NULL;
	filter->callbacks.filter.index = filterIndex;
	filter->callbacks.filter.callback = ChannelRegisterFilter(filterIndex, SchemeFilterCallback, SchemeFilterDeRegisterCallback, filter);
	if (filter->callbacks.filter.callback == NULL)
	{
		DnpapFree(filter);
		return NULL;
	}
	return filter;
}
@}

@d SchemeDeFilter prototype
@{
BOOLEAN SchemeDeFilter(LanguageCallback *filter);
@}

@d SchemeDeFilter code
@{
BOOLEAN SchemeDeFilter(LanguageCallback *filter)
{
	ChannelDeRegisterFilter(filter->callbacks.filter.index, filter->callbacks.filter.callback);
    DnpapFree(filter);
	return TRUE;
}
@}

@d SchemeFilterDeRegister prototype
@{
Scheme_Object *SchemeFilterDeRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeFilterDeRegister code
@{
Scheme_Object *SchemeFilterDeRegister(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
LanguageCallback *filter;

    if ((scheme = SchemeContextCurrent()) == NULL)
    	return scheme_false;

    if (argc != 1 || !SCHEME_INTP(argv[0]))
    {
    	scheme_write_string("filter-deregister: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if ((filter = RemoveCallback(&scheme->filters, SCHEME_INT_VAL(argv[0]))) == NULL)
	   	return scheme_false;
	SchemeDeFilter(filter);

   	return scheme_true;
}
@}

@d SchemeFilterCallback prototype
@{
BOOLEAN SchemeFilterCallback(ULONG filterIndex, int type, PROT_PKT *pkt, VOID *param);
@}

@d SchemeFilterCallback code
@{
BOOLEAN SchemeFilterCallback(ULONG filterIndex, int type, PROT_PKT *pkt, VOID *param)
{
LanguageCallback *filter = (LanguageCallback *) param;
Scheme *scheme;
Scheme_Object *obj;

    if ((scheme = SchemeContextSwitch(filter->language)) == NULL)
    	return FALSE;

    if (filter->callback == NULL || filter->suspend == TRUE)
        return FALSE;

    obj = scheme_make_pair(filter->param, scheme_null);
    obj = scheme_make_pair(scheme_make_packet(pkt), obj);
    obj = scheme_make_pair(scheme_make_integer(type), obj);
    obj = scheme_make_pair(scheme_make_integer(filterIndex), obj);
    obj = scheme_make_pair(scheme_make_integer(filter->id), obj);
    obj = scheme_make_pair(filter->callback, obj);
    obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env), 0);

	if (obj == NULL)
	{
		scheme_write_string("error executing filter callback", scheme->stdoutport);
		return FALSE;
	}

	if (!SCHEME_BOOLP(obj) || obj == scheme_false)
		return FALSE;

	return TRUE;
}
@}


@d SchemeFilterDeRegisterCallback prototype
@{
VOID SchemeFilterDeRegisterCallback(ULONG filterIndex, VOID *param);
@}

@d SchemeFilterDeRegisterCallback code
@{
VOID SchemeFilterDeRegisterCallback(ULONG filterIndex, VOID *param)
{
LanguageCallback *filter = (LanguageCallback *) param;
Scheme *scheme;

    if ((scheme = SchemeContextSwitch(filter->language)) == NULL)
		return;

	if (RemoveCallback(&scheme->filters, filter->id) == filter)
		SchemeDeFilter(filter);
}
@}


@d SchemeGeneratorRegister prototype
@{
Scheme_Object *SchemeGeneratorRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeGeneratorRegister code
@{
Scheme_Object *SchemeGeneratorRegister(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
LanguageCallback *generator;
int command;

    if ((scheme = SchemeContextCurrent()) == NULL)
        return scheme_false;

	if ((command = SubscriptRunCommand(scheme->run)) != RUN_COMMAND_INIT &&
		command != RUN_COMMAND_STOP)
    {
    	scheme_write_string("generator-register: can only be called during 'start' or 'resume'", scheme->stdoutport);
    	return scheme_false;
    }

    if (argc != 1 || !SCHEME_INTP(argv[0]))
    {
    	scheme_write_string("generator-register: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if ((generator = SchemeGenerator(scheme, SCHEME_INT_VAL(argv[0]))) == NULL)
    {
    	scheme_write_string("generator-register: can not register generator", scheme->stdoutport);
    	return scheme_false;
    }

	if (AddCallback(&scheme->generators, generator) == FALSE)
    {
    	scheme_write_string("generator-register: can not add generator", scheme->stdoutport);
    	return scheme_false;
    }

	return scheme_make_integer(generator->id);
}
@}

@d SchemeGenerator prototype
@{
LanguageCallback *SchemeGenerator(Scheme *scheme, int eventIndex);
@}

@d SchemeGenerator code
@{
LanguageCallback *SchemeGenerator(Scheme *scheme, int eventIndex)
{
LanguageCallback *generator;

	if ((generator = DnpapMalloc(sizeof(LanguageCallback))) == NULL)
		return NULL;
	generator->id = (int) generator;
	generator->language = scheme;
	generator->callback = NULL;
	generator->param = NULL;
	generator->suspend = FALSE;
	generator->prev = NULL;
	generator->next = NULL;
	generator->callbacks.generator.index = eventIndex;
	generator->callbacks.generator.hook = EventRegisterGenerator(eventIndex, SchemeGeneratorDeRegisterCallback, generator);
	if (generator->callbacks.generator.hook == NULL)
	{
		DnpapFree(generator);
		return NULL;
	}
	return generator;
}
@}

@d SchemeDeGenerator prototype
@{
BOOLEAN SchemeDeGenerator(LanguageCallback *generator);
@}

@d SchemeDeGenerator code
@{
BOOLEAN SchemeDeGenerator(LanguageCallback *generator)
{
	EventDeRegisterGenerator(generator->callbacks.generator.index, generator->callbacks.generator.hook);
    DnpapFree(generator);
	return TRUE;
}
@}

@d SchemeGeneratorSend prototype
@{
Scheme_Object *SchemeGeneratorSend(int argc, Scheme_Object *argv[]);
@}

@d SchemeGeneratorSend code
@{
Scheme_Object *SchemeGeneratorSend(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
int command;

    if ((scheme = SchemeContextCurrent()) == NULL)
        return scheme_false;

	if ((command = SubscriptRunCommand(scheme->run)) != RUN_COMMAND_INIT &&
		command != RUN_COMMAND_STOP)
    {
    	scheme_write_string("generator-send: can only be called during 'start' or 'resume'", scheme->stdoutport);
    	return scheme_false;
    }

    if (argc != 5 || !SCHEME_INTP(argv[0]) || !SCHEME_STRINGP(argv[1]) ||
		!SCHEME_INTP(argv[2]) || !SCHEME_INTP(argv[3]) ||
		!SCHEME_VECTORP(argv[4]))
    {
    	scheme_write_string("generator-send: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if (SchemeEventGenerate(scheme, SCHEME_INT_VAL(argv[0]),
		SCHEME_STR_VAL(argv[1]), SCHEME_INT_VAL(argv[2]),
		SCHEME_INT_VAL(argv[3]), argv[4]) == FALSE)
    {
    	scheme_write_string("generator-send: can not send event", scheme->stdoutport);
    	return scheme_false;
    }

	return scheme_true;
}
@}


@d SchemeEventGenerate prototype
@{
BOOLEAN SchemeEventGenerate(Scheme *scheme, int id, char *generator, int code, int specific, Scheme_Object *varbinds);
@}

@d SchemeEventGenerate code
@{
BOOLEAN SchemeEventGenerate(Scheme *scheme, int id, char *generator, int code, int specific, Scheme_Object *varbinds)
{
EventHook *hook;
SNMP_OBJECT *objects;
WORD nrobjects;
LanguageCallback *entry;

	if ((entry = SearchCallback(scheme->generators, id)) == NULL)
		return FALSE;
	if ((hook = entry->callbacks.generator.hook) == NULL)
		return FALSE;
	if (!SCHEME_VECTORP(varbinds))
		return FALSE;
	nrobjects = SCHEME_VEC_SIZE(varbinds);
	if ((objects = DnpapMalloc(nrobjects*sizeof(SNMP_OBJECT))) == NULL)
		return FALSE;
	if (Varbinds2Objects(SNMP_PDU_SET, varbinds, objects, nrobjects) == FALSE)
	{
		DnpapFree(objects);
		return FALSE;
	}
	if (EventGenerate(hook, generator, code, specific, objects, nrobjects) == FALSE)
	{
		DnpapFree(objects);
		return FALSE;
	}
	DnpapFree(objects);
	return TRUE;
}
@}


@d SchemeGeneratorDeRegister prototype
@{
Scheme_Object *SchemeGeneratorDeRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeGeneratorDeRegister code
@{
Scheme_Object *SchemeGeneratorDeRegister(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
LanguageCallback *generator;

    if ((scheme = SchemeContextCurrent()) == NULL)
    	return scheme_false;

    if (argc != 1 || !SCHEME_INTP(argv[0]))
    {
    	scheme_write_string("generator-deregister: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if ((generator = RemoveCallback(&scheme->generators, SCHEME_INT_VAL(argv[0]))) == NULL)
	   	return scheme_false;
	SchemeDeGenerator(generator);

   	return scheme_true;
}
@}


@d SchemeGeneratorDeRegisterCallback prototype
@{
VOID SchemeGeneratorDeRegisterCallback(ULONG generatorIndex, VOID *param);
@}

@d SchemeGeneratorDeRegisterCallback code
@{
VOID SchemeGeneratorDeRegisterCallback(ULONG generatorIndex, VOID *param)
{
LanguageCallback *generator = (LanguageCallback *) param;
Scheme *scheme;

    if ((scheme = SchemeContextSwitch(generator->language)) == NULL)
    	return;

	if (RemoveCallback(&scheme->generators, generator->id) == generator)
		SchemeDeGenerator(generator);
}
@}


@d SchemeEventRegister prototype
@{
Scheme_Object *SchemeEventRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeEventRegister code
@{
Scheme_Object *SchemeEventRegister(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
LanguageCallback *event;
int command;

    if ((scheme = SchemeContextCurrent()) == NULL)
        return scheme_false;

	if ((command = SubscriptRunCommand(scheme->run)) != RUN_COMMAND_INIT &&
		command != RUN_COMMAND_STOP)
    {
    	scheme_write_string("event-register: can only be called during 'start' or 'resume'", scheme->stdoutport);
    	return scheme_false;
    }

    if (argc != 3 || !SCHEME_PROCP(argv[0]) || !SCHEME_INTP(argv[1]))
    {
    	scheme_write_string("event-register: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if ((event = SchemeEvent(scheme, argv[0], SCHEME_INT_VAL(argv[1]), argv[2])) == NULL)
    {
    	scheme_write_string("event-register: can not register event", scheme->stdoutport);
    	return scheme_false;
    }

	if (AddCallback(&scheme->events, event) == FALSE)
    {
    	scheme_write_string("event-register: can not add event", scheme->stdoutport);
    	return scheme_false;
    }

	return scheme_make_integer(event->id);
}
@}

@d SchemeEvent prototype
@{
LanguageCallback *SchemeEvent(Scheme *scheme, Scheme_Object *obj, int eventIndex, Scheme_Object *param);
@}

@d SchemeEvent code
@{
LanguageCallback *SchemeEvent(Scheme *scheme, Scheme_Object *obj, int eventIndex, Scheme_Object *param)
{
LanguageCallback *event;

	if ((event = DnpapMalloc(sizeof(LanguageCallback))) == NULL)
		return NULL;
	event->id = (int) event;
	event->language = scheme;
	event->callback = obj;
	event->param = param;
	event->suspend = FALSE;
	event->prev = NULL;
	event->next = NULL;
	event->callbacks.event.index = eventIndex;
	event->callbacks.event.callback = EventRegisterParty(eventIndex, SchemeEventCallback, SchemeEventDeRegisterCallback, event);
	if (event->callbacks.event.callback == NULL)
	{
		DnpapFree(event);
		return NULL;
	}
	return event;
}
@}

@d SchemeDeEvent prototype
@{
BOOLEAN SchemeDeEvent(LanguageCallback *event);
@}

@d SchemeDeEvent code
@{
BOOLEAN SchemeDeEvent(LanguageCallback *event)
{
	EventDeRegisterParty(event->callbacks.event.index, event->callbacks.event.callback);
    DnpapFree(event);
	return TRUE;
}
@}

@d SchemeEventDeRegister prototype
@{
Scheme_Object *SchemeEventDeRegister(int argc, Scheme_Object *argv[]);
@}

@d SchemeEventDeRegister code
@{
Scheme_Object *SchemeEventDeRegister(int argc, Scheme_Object *argv[])
{
Scheme *scheme;
LanguageCallback *event;

    if ((scheme = SchemeContextCurrent()) == NULL)
    	return scheme_false;

    if (argc != 1 || !SCHEME_INTP(argv[0]))
    {
    	scheme_write_string("event-deregister: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_false;
    }

	if ((event = RemoveCallback(&scheme->events, SCHEME_INT_VAL(argv[0]))) == NULL)
	   	return scheme_false;
	SchemeDeEvent(event);

   	return scheme_true;
}
@}

@d SchemeEventCallback prototype
@{
BOOLEAN SchemeEventCallback(ULONG eventIndex, VOID *param);
@}

@d SchemeEventCallback code
@{
BOOLEAN SchemeEventCallback(ULONG eventIndex, VOID *param)
{
LanguageCallback *event = (LanguageCallback *) param;
Scheme *scheme;
Scheme_Object *obj;

    if ((scheme = SchemeContextSwitch(event->language)) == NULL)
    	return FALSE;

    if (event->callback == NULL || event->suspend == TRUE)
        return FALSE;

    obj = scheme_make_pair(event->param, scheme_null);
    obj = scheme_make_pair(scheme_make_integer(eventIndex), obj);
    obj = scheme_make_pair(scheme_make_integer(event->id), obj);
    obj = scheme_make_pair(event->callback, obj);
    obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env), 0);

	if (obj == NULL)
	{
		scheme_write_string("error executing event callback", scheme->stdoutport);
		return FALSE;
	}

	if (!SCHEME_BOOLP(obj) || obj == scheme_false)
		return FALSE;

	return TRUE;
}
@}

@d SchemeEventDeRegisterCallback prototype
@{
VOID SchemeEventDeRegisterCallback(ULONG eventIndex, VOID *param);
@}

@d SchemeEventDeRegisterCallback code
@{
VOID SchemeEventDeRegisterCallback(ULONG eventIndex, VOID *param)
{
LanguageCallback *event = (LanguageCallback *) param;
Scheme *scheme;

    if ((scheme = SchemeContextSwitch(event->language)) == NULL)
    	return;

	if (RemoveCallback(&scheme->events, event->id) == event)
		SchemeDeEvent(event);
}
@}


@d Varbinds2Objects prototype
@{
static BOOLEAN Varbinds2Objects(int request, Scheme_Object *varbinds, SNMP_OBJECT *objects, WORD nrobjects);
@}

@d Varbinds2Objects code
@{
BOOLEAN Varbinds2Objects(int request, Scheme_Object *varbinds, SNMP_OBJECT *objects, WORD nrobjects)
{
Scheme *scheme;
Scheme_Object *varbind, *id, *type, *syntax;
int i;

	if ((scheme = SchemeContextCurrent()) == NULL)
		return FALSE;

	for (i = 0; i < nrobjects; i++)
	{
		varbind = SCHEME_VEC_ELS(varbinds)[i];
		if (!SCHEME_VARBINDP(varbind))
			return FALSE;
		objects[i].Request = request;
		id = SCHEME_VEC_ELS(varbind)[0];
		if (SnmpObjectPutId(&objects[i], id, 0) == FALSE)
			return FALSE;
		if (request == SNMP_PDU_SET)
		{
			type = SCHEME_VEC_ELS(varbind)[1];
			syntax = SCHEME_VEC_ELS(varbind)[2];
			if (SnmpObjectPutType(&objects[i], type) == FALSE)
				return FALSE;
			if (SnmpObjectPutSyntax(&objects[i], syntax) == FALSE)
				return FALSE;
		}
	}
	return TRUE;
}
@}


@d Objects2Varbinds prototype
@{
Scheme_Object *Objects2Varbinds(SNMP_OBJECT *objects, WORD nrobjects);
@}

@d Objects2Varbinds code
@{
Scheme_Object *Objects2Varbinds(SNMP_OBJECT *objects, WORD nrobjects)
{
Scheme_Object *varbinds, *varbind;
Scheme *scheme;
int i;

	if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_null;
	if (objects == NULL || nrobjects < 1)
		return scheme_null;
	if ((varbinds = scheme_make_vector(nrobjects, scheme_null)) == NULL)
		return scheme_null;
	for (i = 0; i < nrobjects; i++)
	{
		if ((varbind = scheme_make_vector(3, scheme_null)) == NULL)
			return scheme_null;
		SCHEME_TYPE(varbind) = scheme->varbind_type;
		SCHEME_VEC_ELS(varbind)[0] = SnmpObjectGetId(&objects[i], 0);
		SCHEME_VEC_ELS(varbind)[1] = SnmpObjectGetType(&objects[i]);
		SCHEME_VEC_ELS(varbind)[2] = SnmpObjectGetSyntax(&objects[i]);
		SCHEME_VEC_ELS(varbinds)[i] = varbind;
	}
	return varbinds;
}
@}


@d SnmpObjectPutId prototype
@{
static BOOLEAN SnmpObjectPutId(SNMP_OBJECT * snmp, Scheme_Object * id, int offset);
@}

@d SnmpObjectPutId code
@{
BOOLEAN SnmpObjectPutId(SNMP_OBJECT * snmp, Scheme_Object * id, int offset)
{
	int i;

	if (!SCHEME_VECTORP(id))
		return FALSE;

	snmp->IdLen = SCHEME_VEC_SIZE(id) + offset;

	for (i = offset; i < snmp->IdLen; i++)
	{
		Scheme_Object *elm = SCHEME_VEC_ELS(id)[i - offset];

		if (!SCHEME_INTP(elm))
			return FALSE;
		snmp->Id[i] = SCHEME_INT_VAL(elm);
	}
	return TRUE;
}
@}


@d SnmpObjectPutType prototype
@{
static BOOLEAN SnmpObjectPutType(SNMP_OBJECT * snmp, Scheme_Object * type);
@}

@d SnmpObjectPutType code
@{
BOOLEAN SnmpObjectPutType(SNMP_OBJECT * snmp, Scheme_Object * type)
{
	if (!SCHEME_INTP(type))
		return FALSE;

	snmp->Type = SCHEME_INT_VAL(type);
	return TRUE;
}
@}


@d SnmpObjectPutSyntax prototype
@{
static BOOLEAN SnmpObjectPutSyntax(SNMP_OBJECT * snmp, Scheme_Object * syntax);
@}

@d SnmpObjectPutSyntax code
@{
BOOLEAN SnmpObjectPutSyntax(SNMP_OBJECT * snmp, Scheme_Object * syntax)
{
	int i;

	switch (snmp->Type)
	{
	case SNMP_INTEGER:					/* LngInt */
	case SNMP_IPADDR:					/* LngUns */
	case SNMP_COUNTER:					/* LngUns */
	case SNMP_GAUGE:					/* LngUns */
	case SNMP_TIMETICKS:				/* LngUns */
		if (!SCHEME_INTP(syntax))
			return FALSE;
		snmp->Syntax.LngInt = SCHEME_INT_VAL(syntax);
		break;
	case SNMP_OPAQUE:					/* BufChr */
	case SNMP_OCTETSTR:					/* BufChr */
		if (!SCHEME_VECTORP(syntax))
			return FALSE;
		snmp->SyntaxLen = SCHEME_VEC_SIZE(syntax);
		for (i = 0; i < snmp->SyntaxLen; i++)
		{
			Scheme_Object *elm = SCHEME_VEC_ELS(syntax)[i];

			if (!SCHEME_CHARP(elm))
				return FALSE;
			snmp->Syntax.BufChr[i] = SCHEME_CHAR_VAL(elm);
		}
		break;
	case SNMP_OBJECTID:					/* BufInt */
		if (!SCHEME_VECTORP(syntax))
			return FALSE;
		snmp->SyntaxLen = SCHEME_VEC_SIZE(syntax);
		for (i = 0; i < snmp->SyntaxLen; i++)
		{
			Scheme_Object *elm = SCHEME_VEC_ELS(syntax)[i];

			if (!SCHEME_INTP(elm))
				return FALSE;
			snmp->Syntax.BufInt[i] = SCHEME_INT_VAL(elm);
		}
		break;
	case SNMP_NULL:						/* Null   */
	default:
		break;
	}
	return TRUE;
}
@}


@d SnmpObjectGetId prototype
@{
static Scheme_Object *SnmpObjectGetId(SNMP_OBJECT * snmp, int offset);
@}

@d SnmpObjectGetId code
@{
Scheme_Object *SnmpObjectGetId(SNMP_OBJECT * snmp, int offset)
{
	int i;
	Scheme_Object *id;

	id = scheme_make_vector(snmp->IdLen - offset, scheme_null);
	for (i = offset; i < snmp->IdLen; i++)
	{
		SCHEME_VEC_ELS(id)[i - offset] = scheme_make_integer(snmp->Id[i]);
	}
	return id;
}
@}


@d SnmpObjectGetType prototype
@{
static Scheme_Object *SnmpObjectGetType(SNMP_OBJECT * snmp);
@}

@d SnmpObjectGetType code
@{
Scheme_Object *SnmpObjectGetType(SNMP_OBJECT * snmp)
{
	Scheme_Object *type;

	type = scheme_make_integer(snmp->Type);
	return type;
}
@}


@d SnmpObjectGetSyntax prototype
@{
static Scheme_Object *SnmpObjectGetSyntax(SNMP_OBJECT * snmp);
@}

@d SnmpObjectGetSyntax code
@{
Scheme_Object *SnmpObjectGetSyntax(SNMP_OBJECT * snmp)
{
	Scheme_Object *syntax;
	int i;

	switch (snmp->Type)
	{
	case SNMP_INTEGER:					/* LngInt */
	case SNMP_IPADDR:					/* LngUns */
	case SNMP_COUNTER:					/* LngUns */
	case SNMP_GAUGE:					/* LngUns */
	case SNMP_TIMETICKS:				/* LngUns */
		syntax = scheme_make_integer(snmp->Syntax.LngInt);
		break;
	case SNMP_OPAQUE:					/* BufChr */
	case SNMP_OCTETSTR:					/* BufChr */
		syntax = scheme_make_vector(snmp->SyntaxLen, scheme_null);
		for (i = 0; i < snmp->SyntaxLen; i++)
		{
			SCHEME_VEC_ELS(syntax)[i] = scheme_make_char(snmp->Syntax.BufChr[i]);
		}
		break;
	case SNMP_OBJECTID:					/* BufInt */
		syntax = scheme_make_vector(snmp->SyntaxLen, scheme_null);
		for (i = 0; i < snmp->SyntaxLen; i++)
		{
			SCHEME_VEC_ELS(syntax)[i] = scheme_make_integer(snmp->Syntax.BufInt[i]);
		}
		break;
	case SNMP_NULL:						/* Null   */
	default:
		syntax = scheme_null;
		break;
	}
	return syntax;
}
@}


@d SnmpObjectGetRequest prototype
@{
static Scheme_Object *SnmpObjectGetRequest(SNMP_OBJECT * snmp);
@}

@d SnmpObjectGetRequest code
@{
Scheme_Object *SnmpObjectGetRequest(SNMP_OBJECT * snmp)
{
	Scheme_Object *request;

	request = scheme_make_integer(snmp->Request);
	return request;
}
@}


@d SchemeMibRequest prototype
@{
static Scheme_Object *SchemeMibRequest(int argc, Scheme_Object * argv[], int request);
@}

@d SchemeMibRequest code
@{
Scheme_Object *SchemeMibRequest(int argc, Scheme_Object * argv[], int request)
{
    Scheme *scheme;
	SNMP_OBJECT objects[AGENT_LIST_SIZE];
	int nrobjects;
	WORD status, index;

    if ((scheme = SchemeContextCurrent()) == NULL)
    	return scheme_null;

    if (argc != 1 || !SCHEME_VECTORP(argv[0]))
    {
    	scheme_write_string("mib-request: arguments do not match either in type or number", scheme->stdoutport);
    	return scheme_null;
    }

	if (Varbinds2Objects(request, argv[0], objects, nrobjects = SCHEME_VEC_SIZE(argv[0])) == FALSE)
	{
		scheme_write_string("mib-request: error in varbinds", scheme->stdoutport);
		return scheme_null;
	}

	if ((status = MibRequest(nrobjects, objects, &index)) != SNMP_NOERROR)
		return scheme_make_integer(status);

	return Objects2Varbinds(objects, nrobjects);
}
@}


@d SchemeMibGet prototype
@{
Scheme_Object *SchemeMibGet(int argc, Scheme_Object * argv[]);
@}

@d SchemeMibGet code
@{
Scheme_Object *SchemeMibGet(int argc, Scheme_Object * argv[])
{
	return SchemeMibRequest(argc, argv, SNMP_PDU_GET);
}
@}


@d SchemeMibSet prototype
@{
Scheme_Object *SchemeMibSet(int argc, Scheme_Object * argv[]);
@}

@d SchemeMibSet code
@{
Scheme_Object *SchemeMibSet(int argc, Scheme_Object * argv[])
{
	return SchemeMibRequest(argc, argv, SNMP_PDU_SET);
}
@}


@d SchemeMibNext prototype
@{
Scheme_Object *SchemeMibNext(int argc, Scheme_Object * argv[]);
@}

@d SchemeMibNext code
@{
Scheme_Object *SchemeMibNext(int argc, Scheme_Object * argv[])
{
	return SchemeMibRequest(argc, argv, SNMP_PDU_NEXT);
}
@}


@d SchemeMibCallback prototype
@{
WORD SchemeMibCallback(SNMP_OBJECT * snmp, WORD idlen, VOID * param);
@}

@d SchemeMibCallback code
@{
WORD SchemeMibCallback(SNMP_OBJECT * snmp, WORD idlen, VOID * param)
{
	LANGUAGE_MIB_OBJECT *object;
	LanguageCallback *mib = (LanguageCallback *) param;
	Scheme *scheme;
	Scheme_Object *call, *args, *result, *varbind, *request, *id, *type, *syntax;

	if ((scheme = SchemeContextSwitch(mib->language)) == NULL)
		return SNMP_GENERROR;

	object = &mib->callbacks.mib;

	if ((request = SnmpObjectGetRequest(snmp)) == NULL)
		return SNMP_GENERROR;

	if ((id = SnmpObjectGetId(snmp, object->offset)) == NULL)
		return SNMP_GENERROR;
	if ((type = SnmpObjectGetType(snmp)) == NULL)
		return SNMP_GENERROR;
	if ((syntax = SnmpObjectGetSyntax(snmp)) == NULL)
		return SNMP_GENERROR;

	varbind = scheme_make_vector(3, scheme_null);
	SCHEME_TYPE(varbind) = scheme->varbind_type;
	SCHEME_VEC_ELS(varbind)[0] = id;
	SCHEME_VEC_ELS(varbind)[1] = type;
	SCHEME_VEC_ELS(varbind)[2] = syntax;

	args = scheme_make_pair(mib->param, scheme_null);
	args = scheme_make_pair(varbind, args);
	args = scheme_make_pair(request, args);
	args = scheme_make_pair(scheme_make_integer(mib->id), args);
	call = scheme_make_pair(mib->callback, args);

	result = SCHEME_CATCH_ERROR(scheme_eval(call, scheme->env), 0);

	if (result == NULL)
	{
		scheme_write_string("error executing mib callback", scheme->stdoutport);
		return SNMP_GENERROR;
	}

	if (SCHEME_INTP(result))
		return SCHEME_INT_VAL(result);

	if (!SCHEME_VARBINDP(result))
		return SNMP_GENERROR;

	id = SCHEME_VEC_ELS(result)[0];
	type = SCHEME_VEC_ELS(result)[1];
	syntax = SCHEME_VEC_ELS(result)[2];

	if (!SnmpObjectPutId(snmp, id, object->offset))
		return SNMP_GENERROR;
	if (!SnmpObjectPutType(snmp, type))
		return SNMP_GENERROR;
	if (!SnmpObjectPutSyntax(snmp, syntax))
		return SNMP_GENERROR;

	return SNMP_NOERROR;
}
@}


@d SchemeMib prototype
@{
LanguageCallback *SchemeMib(Scheme * scheme, int argc, Scheme_Object * argv[], LONG * tree, INT treeLen);
@}

@d SchemeMib code
@{
LanguageCallback *SchemeMib(Scheme * scheme, int argc, Scheme_Object * argv[], LONG * tree, INT treeLen)
{
	LanguageCallback *mib;
	Scheme_Object *elm;
	INT i, varLen, idLen;
	LONG *id;

	varLen = SCHEME_VEC_SIZE(argv[0]);
	idLen = treeLen + 1 + varLen + 1;

	if ((id = DnpapMalloc(idLen * sizeof(id[0]))) == NULL)
		return NULL;

	for (i = 0; i < treeLen; i++)
	{
		id[i] = tree[i];
	}
	id[treeLen] = SubscriptRunScript(scheme->run);
	for (i = 0; i < varLen; i++)
	{
		if (!SCHEME_INTP(elm = SCHEME_VEC_ELS(argv[0])[i]))
		{
			DnpapFree(id);
			return NULL;
		}
		id[treeLen + 1 + i] = SCHEME_INT_VAL(elm);
	}
	id[treeLen + 1 + varLen] = SubscriptRunIndex(scheme->run);

	if ((mib = DnpapMalloc(sizeof(LanguageCallback))) == NULL)
		return NULL;
	mib->id = (int) mib;
	mib->language = scheme;
	mib->callback = argv[1];
	mib->param = argv[4];
	mib->suspend = FALSE;
	mib->prev = NULL;
	mib->next = NULL;
	mib->callbacks.mib.object = MibRegister(id, idLen, SchemeMibCallback,
										SCHEME_INT_VAL(argv[2]),
										SCHEME_INT_VAL(argv[3]), mib);
	DnpapFree(id);
	if (mib->callbacks.mib.object == NULL)
	{
		DnpapFree(mib);
		return NULL;
	}
	mib->callbacks.mib.offset = treeLen + 1;

	return mib;
}
@}


@d SchemeDeMib prototype
@{
BOOLEAN SchemeDeMib(LanguageCallback * mib);
@}

@d SchemeDeMib code
@{
BOOLEAN SchemeDeMib(LanguageCallback * mib)
{
	MibDeRegister(mib->callbacks.mib.object);
	DnpapFree(mib);
	return TRUE;
}
@}


@d SchemeMibRegister prototype
@{
static Scheme_Object *SchemeMibRegister(int argc, Scheme_Object * argv[], LONG * tree, INT treeLen);
@}

@d SchemeMibRegister code
@{
Scheme_Object *SchemeMibRegister(int argc, Scheme_Object * argv[], LONG * tree, INT treeLen)
{
	Scheme *scheme;
	LanguageCallback *mib;
	int command;

	if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

	if (argc != 5 ||
		!SCHEME_VECTORP(argv[0]) ||
		!SCHEME_PROCP(argv[1]) ||
		!SCHEME_INTP(argv[2]) ||
		!SCHEME_INTP(argv[3]))
	{
		scheme_write_string("mib-register: arguments do not match either in type or number", scheme->stdoutport);
		return scheme_false;
	}

	if ((mib = SchemeMib(scheme, argc, argv, tree, treeLen)) == NULL)
	{
		scheme_write_string("mib-register: can not register mib", scheme->stdoutport);
		return scheme_false;
	}

	if (AddCallback(&scheme->mibs, mib) == FALSE)
	{
		scheme_write_string("mib-register: can not add mib", scheme->stdoutport);
		return scheme_false;
	}

	return scheme_make_integer(mib->id);
}
@}


@d SchemeMibDeRegister prototype
@{
Scheme_Object *SchemeMibDeRegister(int argc, Scheme_Object * argv[]);
@}

@d SchemeMibDeRegister code
@{
Scheme_Object *SchemeMibDeRegister(int argc, Scheme_Object * argv[])
{
	Scheme *scheme;
	LanguageCallback *mib;

	if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

	if (argc != 1 || !SCHEME_INTP(argv[0]))
	{
		scheme_write_string("mib-deregister: arguments do not match either in type or number", scheme->stdoutport);
		return scheme_false;
	}

	if ((mib = RemoveCallback(&scheme->mibs, SCHEME_INT_VAL(argv[0]))) == NULL)
		return scheme_false;
	SchemeDeMib(mib);

	return scheme_true;
}
@}


@d SchemeMibRegisterArgs prototype
@{
Scheme_Object *SchemeMibRegisterArgs(int argc, Scheme_Object * argv[]);
@}

@d SchemeMibRegisterArgs code
@{
Scheme_Object *SchemeMibRegisterArgs(int argc, Scheme_Object * argv[])
{
	static LONG tree[] =
	{DNPAP_MIB, 4, 5};

	return SchemeMibRegister(argc, argv, tree, sizeof(tree) / sizeof(tree[0]));
}
@}


@d SchemeMibRegisterResults prototype
@{
Scheme_Object *SchemeMibRegisterResults(int argc, Scheme_Object * argv[]);
@}

@d SchemeMibRegisterResults code
@{
Scheme_Object *SchemeMibRegisterResults(int argc, Scheme_Object * argv[])
{
	static LONG tree[] =
	{DNPAP_MIB, 4, 6};

	return SchemeMibRegister(argc, argv, tree, sizeof(tree) / sizeof(tree[0]));
}
@}


@d Scheme static variables
@{
INT SchemeSnmpId = 0;
@}


@d SchemeSnmpConnect prototype
@{
Scheme_Object *SchemeSnmpConnect(int argc, Scheme_Object * argv[]);
@}

@d SchemeSnmpConnect code
@{
Scheme_Object *SchemeSnmpConnect(int argc, Scheme_Object * argv[])
{
	Scheme *scheme;
	LanguageCallback *snmp;
	int command;

	if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

	if ((command = SubscriptRunCommand(scheme->run)) != RUN_COMMAND_INIT &&
		command != RUN_COMMAND_STOP)
	{
		scheme_write_string("snmp-connect: can only be called during 'start' or 'resume'", scheme->stdoutport);
		return scheme_false;
	}

	if (argc != 6 ||
		!SCHEME_PROCP(argv[0]) ||
		!SCHEME_STRINGP(argv[1]) ||
		!SCHEME_INTP(argv[2]) ||
		!SCHEME_STRINGP(argv[3]) ||
		!SCHEME_STRINGP(argv[4]))
	{
		scheme_write_string("snmp-connect: arguments do not match either in type or number", scheme->stdoutport);
		return scheme_false;
	}

	if ((snmp = SchemeSnmp(scheme, argv[0], SCHEME_STR_VAL(argv[1]), SCHEME_INT_VAL(argv[2]), SCHEME_STR_VAL(argv[3]), SCHEME_STR_VAL(argv[4]), argv[5])) == NULL)
	{
		scheme_write_string("snmp-connect: can not connect", scheme->stdoutport);
		return scheme_false;
	}

	if (AddCallback(&scheme->snmps, snmp) == FALSE)
	{
		scheme_write_string("snmp-connect: can not add snmp", scheme->stdoutport);
		return scheme_false;
	}

	return scheme_make_integer(snmp->id);
}
@}


@d SchemeSnmp prototype
@{
LanguageCallback *SchemeSnmp(Scheme *scheme, Scheme_Object *obj, char *host, int port, char *request, char *community, Scheme_Object *param);
@}

@d SchemeSnmp code
@{
LanguageCallback *SchemeSnmp(Scheme *scheme, Scheme_Object *obj, char *host, int port, char *request, char *community, Scheme_Object *param)
{
LanguageCallback *snmp;
char addr[NWSNMP_SIZE_ADDRESS];

	if ((snmp = DnpapMalloc(sizeof(LanguageCallback))) == NULL)
		return NULL;
	snmp->id = (int) snmp;
	snmp->language = scheme;
	snmp->callback = obj;
	snmp->param = param;
	snmp->suspend = FALSE;
	snmp->prev = NULL;
	snmp->next = NULL;
	sprintf(addr, "snmp:%s:%d:%s:%s", host, port, request, community);
	if (stricmp(request, "get") == 0)
		snmp->callbacks.snmp.request = SNMP_PDU_GET;
	else
	if (stricmp(request, "set") == 0)
		snmp->callbacks.snmp.request = SNMP_PDU_SET;
	else
	if (stricmp(request, "next") == 0)
		snmp->callbacks.snmp.request = SNMP_PDU_NEXT;
	else
	{
		DnpapFree(snmp);
		return NULL;
	}
	snmp->callbacks.snmp.conn = NwDgConnect(addr, SchemeSnmpCallback, snmp);
	if (snmp->callbacks.snmp.conn == NULL)
	{
		DnpapFree(snmp);
		return NULL;
	}
	return snmp;
}
@}


@d SchemeSnmpCallback prototype
@{
static VOID SchemeSnmpCallback(NW_DG *dg, CHAR *addr, BYTE *frame, INT length, VOID *param);
@}

@d SchemeSnmpCallback code
@{
VOID SchemeSnmpCallback(NW_DG *dg, CHAR *addr, BYTE *frame, INT length, VOID *param)
{
LanguageCallback *snmp = (LanguageCallback *) param;
Scheme *scheme;
Scheme_Object *obj;
NwSnmpFrame *pdu;
SNMP_OBJECT *objects;
WORD nrobjects;

    if ((scheme = SchemeContextSwitch(snmp->language)) == NULL)
    	return;

	if (snmp->callback == NULL || snmp->suspend == TRUE)
		return;

	pdu = (NwSnmpFrame *)frame;
	nrobjects = pdu->NrObjects;
	objects = pdu->Objects;

	obj = scheme_make_pair(snmp->param, scheme_null);
	obj = scheme_make_pair(Objects2Varbinds(objects, nrobjects), obj);
	obj = scheme_make_pair(scheme_make_integer(pdu->Pdu.Request.Index), obj);
	obj = scheme_make_pair(scheme_make_integer(pdu->Pdu.Request.Status), obj);
	obj = scheme_make_pair(scheme_make_integer(snmp->id), obj);
	obj = scheme_make_pair(snmp->callback, obj);
	obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env), 0);

	if (obj == NULL)
	{
		scheme_write_string("error executing snmp callback", scheme->stdoutport);
	}
}
@}


@d SchemeDeSnmp prototype
@{
BOOLEAN SchemeDeSnmp(LanguageCallback *snmp);
@}

@d SchemeDeSnmp code
@{
BOOLEAN SchemeDeSnmp(LanguageCallback *snmp)
{
	NwDgClose(snmp->callbacks.snmp.conn);
    DnpapFree(snmp);
	return TRUE;
}
@}


@d SchemeSnmpSend prototype
@{
Scheme_Object *SchemeSnmpSend(int argc, Scheme_Object * argv[]);
@}

@d SchemeSnmpSend code
@{
Scheme_Object *SchemeSnmpSend(int argc, Scheme_Object * argv[])
{
	Scheme *scheme;
	LanguageCallback *snmp;
	NwSnmpFrame frame;

	if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

	if (argc != 2 || !SCHEME_INTP(argv[0]) || !SCHEME_VECTORP(argv[1]))
	{
		scheme_write_string("snmp-send: arguments do not match either in type or number", scheme->stdoutport);
		return scheme_false;
	}

	if ((snmp = SearchCallback(scheme->snmps, SCHEME_INT_VAL(argv[0]))) == NULL)
	{
		scheme_write_string("snmp-send: unknown snmp connection id", scheme->stdoutport);
		return scheme_false;
	}

	frame.Pdu.Request.Id = SchemeSnmpId++;
	frame.Pdu.Request.Status = 0;
	frame.Pdu.Request.Index = 0;

	frame.NrObjects = SCHEME_VEC_SIZE(argv[1]);
	if ((frame.Objects = DnpapMalloc(frame.NrObjects*sizeof(SNMP_OBJECT))) == NULL)
		return scheme_false;

	if (Varbinds2Objects(snmp->callbacks.snmp.request, argv[1], frame.Objects, frame.NrObjects) == FALSE)
	{
		DnpapFree(frame.Objects);
		return scheme_false;
	}

	if (NwDgSend(snmp->callbacks.snmp.conn, (BYTE *)&frame,
			sizeof(frame)+frame.NrObjects*sizeof(SNMP_OBJECT)) == FALSE)
	{
		DnpapFree(frame.Objects);
		return scheme_false;
	}

	DnpapFree(frame.Objects);

	return scheme_true;
}
@}


@d SchemeSnmpDisConnect prototype
@{
Scheme_Object *SchemeSnmpDisConnect(int argc, Scheme_Object * argv[]);
@}

@d SchemeSnmpDisConnect code
@{
Scheme_Object *SchemeSnmpDisConnect(int argc, Scheme_Object * argv[])
{
	Scheme *scheme;
	LanguageCallback *snmp;

	if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

	if (argc != 1 || !SCHEME_INTP(argv[0]))
	{
		scheme_write_string("snmp-disconnect: arguments do not match either in type or number", scheme->stdoutport);
		return scheme_false;
	}

	if ((snmp = RemoveCallback(&scheme->snmps, SCHEME_INT_VAL(argv[0]))) == 0)
		return scheme_false;
	SchemeDeSnmp(snmp);

	return scheme_true;
}
@}


@d SchemeTrapRegister prototype
@{
Scheme_Object *SchemeTrapRegister(int argc, Scheme_Object * argv[]);
@}

@d SchemeTrapRegister code
@{
Scheme_Object *SchemeTrapRegister(int argc, Scheme_Object * argv[])
{
	Scheme *scheme;
	LanguageCallback *trap;
	int command;

	if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

	if ((command = SubscriptRunCommand(scheme->run)) != RUN_COMMAND_INIT &&
		command != RUN_COMMAND_STOP)
	{
		scheme_write_string("trap-register: can only be called during 'start' or 'resume'", scheme->stdoutport);
		return scheme_false;
	}

	if (argc != 3 ||
		!SCHEME_PROCP(argv[0]) ||
		!SCHEME_INTP(argv[1]))
	{
		scheme_write_string("trap-register: arguments do not match either in type or number", scheme->stdoutport);
		return scheme_false;
	}

	if ((trap = SchemeTrap(scheme, argv[0], SCHEME_INT_VAL(argv[1]), argv[2])) == NULL)
	{
		scheme_write_string("trap-register: can not register accept", scheme->stdoutport);
		return scheme_false;
	}

	if (AddCallback(&scheme->traps, trap) == FALSE)
	{
		scheme_write_string("trap-register: can not add trap", scheme->stdoutport);
		return scheme_false;
	}

	return scheme_make_integer(trap->id);
}
@}


@d SchemeTrap prototype
@{
LanguageCallback *SchemeTrap(Scheme *scheme, Scheme_Object *obj, int port, Scheme_Object *param);
@}

@d SchemeTrap code
@{
LanguageCallback *SchemeTrap(Scheme *scheme, Scheme_Object *obj, int port, Scheme_Object *param)
{
LanguageCallback *trap;
char addr[NWSNMP_SIZE_ADDRESS];

	if ((trap = DnpapMalloc(sizeof(LanguageCallback))) == NULL)
		return NULL;
	trap->id = (int) trap;
	trap->language = scheme;
	trap->callback = obj;
	trap->param = param;
	trap->suspend = FALSE;
	trap->prev = NULL;
	trap->next = NULL;
	sprintf(addr, "snmp:%d", port);
	trap->callbacks.accept = NwDgAccept(addr, SchemeTrapCallback, trap);
	if (trap->callbacks.accept == NULL)
	{
		DnpapFree(trap);
		return NULL;
	}
	return trap;
}
@}


@d SchemeTrapCallback prototype
@{
static VOID SchemeTrapCallback(NW_DG *dg, CHAR *addr, BYTE *frame, INT length, VOID *param);
@}

@d SchemeTrapCallback code
@{
VOID SchemeTrapCallback(NW_DG *dg, CHAR *addr, BYTE *frame, INT length, VOID *param)
{
LanguageCallback *trap = (LanguageCallback *) param;
Scheme *scheme;
Scheme_Object *obj;
NwSnmpFrame *pdu;
SNMP_OBJECT *objects;
WORD nrobjects;

    if ((scheme = SchemeContextSwitch(trap->language)) == NULL)
    	return;

	if (trap->callback == NULL || trap->suspend == TRUE)
		return;
	pdu = (NwSnmpFrame *)frame;
	nrobjects = pdu->NrObjects;
	objects = pdu->Objects;
    obj = scheme_make_pair(trap->param, scheme_null);
    obj = scheme_make_pair(Objects2Varbinds(objects, nrobjects), obj);
    obj = scheme_make_pair(scheme_make_integer(trap->id), obj);
    obj = scheme_make_pair(trap->callback, obj);
	obj = SCHEME_CATCH_ERROR(scheme_eval(obj, scheme->env), 0);

	if (obj == NULL)
	{
		scheme_write_string("error executing trap callback", scheme->stdoutport);
	}
}
@}


@d SchemeDeTrap prototype
@{
BOOLEAN SchemeDeTrap(LanguageCallback *trap);
@}

@d SchemeDeTrap code
@{
BOOLEAN SchemeDeTrap(LanguageCallback *trap)
{
	NwDgClose(trap->callbacks.accept);
    DnpapFree(trap);
	return TRUE;
}
@}


@d SchemeTrapDeRegister prototype
@{
Scheme_Object *SchemeTrapDeRegister(int argc, Scheme_Object * argv[]);
@}

@d SchemeTrapDeRegister code
@{
Scheme_Object *SchemeTrapDeRegister(int argc, Scheme_Object * argv[])
{
	Scheme *scheme;
	LanguageCallback *trap;

	if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

	if (argc != 1 || !SCHEME_INTP(argv[0]))
	{
		scheme_write_string("trap-deregister: arguments do not match either in type or number", scheme->stdoutport);
		return scheme_false;
	}

	if ((trap = RemoveCallback(&scheme->traps, SCHEME_INT_VAL(argv[0]))) == 0)
		return scheme_false;
	SchemeDeTrap(trap);

	return scheme_true;
}
@}


@d SchemeTrapSend prototype
@{
Scheme_Object *SchemeTrapSend(int argc, Scheme_Object * argv[]);
@}

@d SchemeTrapSend code
@{
Scheme_Object *SchemeTrapSend(int argc, Scheme_Object * argv[])
{
	Scheme *scheme;
	INT general, specific, commLen, listLen, i;
	CHAR *comm;
	SNMP_OBJECT *list = NULL;
	Scheme_Object *id, *syntax, *type;

	if ((scheme = SchemeContextCurrent()) == NULL)
		return scheme_false;

	if (argc != 4 ||
		!SCHEME_STRINGP(argv[0]) ||
		!SCHEME_INTP(argv[1]) ||
		!SCHEME_INTP(argv[2]) ||
		!SCHEME_VECTORP(argv[3]))
	{
		scheme_write_string("trap-send: arguments do not match either in type or number", scheme->stdoutport);
		return scheme_false;
	}

	comm = SCHEME_STR_VAL(argv[0]);
	commLen = strlen(comm);
	general = SCHEME_INT_VAL(argv[1]);
	specific = SCHEME_INT_VAL(argv[2]);

	listLen = SCHEME_VEC_SIZE(argv[3]);
	if (listLen != 0 &&
		(list = DnpapMalloc(listLen * sizeof(SNMP_OBJECT))) == NULL)
	{
		return scheme_false;
	}
	for (i = 0; i < listLen; i++)
	{
		Scheme_Object *elm = SCHEME_VEC_ELS(argv[3])[i];

		if (!SCHEME_VARBINDP(elm))
		{
			DnpapFree(list);
			return scheme_false;
		}

		id = SCHEME_VEC_ELS(elm)[0];
		type = SCHEME_VEC_ELS(elm)[1];
		syntax = SCHEME_VEC_ELS(elm)[2];

		if (!SnmpObjectPutId(list + i, id, 0))
		{
			DnpapFree(list);
			return scheme_false;
		}
		if (!SnmpObjectPutType(list + i, type))
		{
			DnpapFree(list);
			return scheme_false;
		}
		if (!SnmpObjectPutSyntax(list + i, syntax))
		{
			DnpapFree(list);
			return scheme_false;
		}
	}

	if (!AgentSendTrap(comm, commLen, SysTime(), general, specific, list, listLen))
	{
		DnpapFree(list);
		return scheme_false;
	}

	DnpapFree(list);
	return scheme_true;
}
@}


@d AddCallback prototype
@{
BOOLEAN AddCallback(LanguageCallback **list, LanguageCallback *entry);
@}

@d AddCallback code
@{
BOOLEAN AddCallback(LanguageCallback **list, LanguageCallback *entry)
{
	entry->prev = NULL;
	entry->next = *list;
	if (*list != NULL)
		(*list)->prev = entry;
	*list = entry;
	return TRUE;
}
@}


@d SearchCallback prototype
@{
LanguageCallback *SearchCallback(LanguageCallback *list, int id);
@}

@d SearchCallback code
@{
LanguageCallback *SearchCallback(LanguageCallback *list, int id)
{
LanguageCallback *entry;

	for (entry = list; entry != NULL; entry = entry->next)
		if (entry->id == id)
			break;

	return entry;
}
@}


@d RemoveCallback prototype
@{
LanguageCallback *RemoveCallback(LanguageCallback **list, int id);
@}

@d RemoveCallback code
@{
LanguageCallback *RemoveCallback(LanguageCallback **list, int id)
{
LanguageCallback *entry;

	entry = SearchCallback(*list, id);
	CutCallback(list, entry);

	return entry;
}
@}


@d CutCallback prototype
@{
VOID CutCallback(LanguageCallback **list, LanguageCallback *entry);
@}

@d CutCallback code
@{
VOID CutCallback(LanguageCallback **list, LanguageCallback *entry)
{
	if (entry == NULL)
		return;
	if (entry->prev != NULL)
		entry->prev->next = entry->next;
	else
		*list = entry->next;
	if (entry->next != NULL)
		entry->next->prev = entry->prev;
}
@}


@d SuspendCallbacks prototype
@{
VOID SuspendCallbacks(VOID *language, int lang, BOOLEAN suspend);
@}

@d SuspendCallbacks code
@{
VOID SuspendCallbacks(VOID *language, int lang, BOOLEAN suspend)
{
LanguageCallback *callback;
LanguageCallback *macs = NULL, *timers = NULL, *channels = NULL,
				*filters = NULL, *generators = NULL, *events = NULL,
				*mibs = NULL, *snmps = NULL, *traps = NULL;

	switch (lang)
	{
	case SCRIPT_LANGUAGE_SCHEME:
		macs = ((Scheme *)language)->macs;
		timers = ((Scheme *)language)->timers;
		channels = ((Scheme *)language)->channels;
		filters = ((Scheme *)language)->filters;
		generators = ((Scheme *)language)->generators;
		events = ((Scheme *)language)->events;
		mibs = ((Scheme *)language)->mibs;
		snmps = ((Scheme *)language)->snmps;
		traps = ((Scheme *)language)->traps;
		break;
	case SCRIPT_LANGUAGE_PERL:
#	ifdef SUBSCRIPT_PERL
		macs = ((Perl *)language)->macs;
		timers = ((Perl *)language)->timers;
		channels = ((Perl *)language)->channels;
		filters = ((Perl *)language)->filters;
		generators = ((Perl *)language)->generators;
		events = ((Perl *)language)->events;
		mibs = ((Perl *)language)->mibs;
		snmps = ((Perl *)language)->snmps;
		traps = ((Perl *)language)->traps;
#	endif
		break;
	default:
		return;
	}

	for (callback = macs; callback != NULL; callback = callback->next)
		callback->suspend = suspend;
	for (callback = timers; callback != NULL; callback = callback->next)
		callback->suspend = suspend;
	for (callback = channels; callback != NULL; callback = callback->next)
		callback->suspend = suspend;
	for (callback = filters; callback != NULL; callback = callback->next)
		callback->suspend = suspend;
	for (callback = generators; callback != NULL; callback = callback->next)
		callback->suspend = suspend;
	for (callback = events; callback != NULL; callback = callback->next)
		callback->suspend = suspend;
	for (callback = mibs; callback != NULL; callback = callback->next)
		callback->suspend = suspend;
	for (callback = snmps; callback != NULL; callback = callback->next)
		callback->suspend = suspend;
	for (callback = traps; callback != NULL; callback = callback->next)
		callback->suspend = suspend;
}
@}


@d RemoveAllCallbacks prototype
@{
VOID RemoveAllCallbacks(VOID *language, int lang);
@}

@d RemoveAllCallbacks code
@{
VOID RemoveAllCallbacks(VOID *language, int lang)
{
LanguageCallback *callback;
Scheme *scheme = NULL;
Perl *perl = NULL;

	switch (lang)
	{
	case SCRIPT_LANGUAGE_SCHEME:
		scheme = (Scheme *) language;
		for (callback = scheme->macs; callback != NULL; callback = scheme->macs)
		{
			CutCallback(&scheme->macs, callback);
			SchemeDeMac(callback);
		}
		for (callback = scheme->timers; callback != NULL; callback = scheme->timers)
		{
			CutCallback(&scheme->timers, callback);
			SchemeDeTimer(callback);
		}
		for (callback = scheme->channels; callback != NULL; callback = scheme->channels)
		{
			CutCallback(&scheme->channels, callback);
			SchemeDeChannel(callback);
		}
		for (callback = scheme->filters; callback != NULL; callback = scheme->filters)
		{
			CutCallback(&scheme->filters, callback);
			SchemeDeFilter(callback);
		}
		for (callback = scheme->generators; callback != NULL; callback = scheme->generators)
		{
			CutCallback(&scheme->generators, callback);
			SchemeDeGenerator(callback);
		}
		for (callback = scheme->events; callback != NULL; callback = scheme->events)
		{
			CutCallback(&scheme->events, callback);
			SchemeDeEvent(callback);
		}
		for (callback = scheme->mibs; callback != 0; callback = scheme->mibs)
		{
			CutCallback(&scheme->mibs, callback);
			SchemeDeMib(callback);
		}
		for (callback = scheme->traps; callback != 0; callback = scheme->traps)
		{
			CutCallback(&scheme->traps, callback);
			SchemeDeSnmp(callback);
		}
		break;
	case SCRIPT_LANGUAGE_PERL:
#	ifdef SUBSCRIPT_PERL
		perl = (Perl *) language;
		for (callback = perl->macs; callback != NULL; callback = perl->macs)
		{
			CutCallback(&perl->macs, callback);
			SchemeDeMac(callback);
		}
		for (callback = perl->timers; callback != NULL; callback = perl->timers)
		{
			CutCallback(&perl->timers, callback);
			SchemeDeTimer(callback);
		}
		for (callback = perl->channels; callback != NULL; callback = perl->channels)
		{
			CutCallback(&perl->channels, callback);
			SchemeDeChannel(callback);
		}
		for (callback = perl->filters; callback != NULL; callback = perl->filters)
		{
			CutCallback(&perl->filters, callback);
			SchemeDeFilter(callback);
		}
		for (callback = perl->generators; callback != NULL; callback = perl->generators)
		{
			CutCallback(&perl->generators, callback);
			SchemeDeGenerator(callback);
		}
		for (callback = perl->events; callback != NULL; callback = perl->events)
		{
			CutCallback(&perl->events, callback);
			SchemeDeEvent(callback);
		}
		for (callback = perl->mibs; callback != 0; callback = perl->mibs)
		{
			CutCallback(&perl->mibs, callback);
			SchemeDeMib(callback);
		}
		for (callback = perl->traps; callback != 0; callback = perl->traps)
		{
			CutCallback(&perl->traps, callback);
			SchemeDeSnmp(callback);
		}
#	endif
		break;
	default:
		return;
	}
}
@}


\section{Perl}

@d Perl type definitions
@{
typedef struct Perl_
{
#ifdef SUBSCRIPT_PERL
    VOID				*run;
    RunGetsFunction		rungets;
    RunPutsFunction		runputs;

    PerlInterpreter		*env;

    OP					*init;
    OP					*start;
    OP					*stop;
    OP					*resume;
    OP					*exit;

	LanguageCallback	*macs;
	LanguageCallback	*timers;
	LanguageCallback	*filters;
	LanguageCallback	*channels;
	LanguageCallback	*generators;
	LanguageCallback	*events;
	LanguageCallback	*mibs;
	LanguageCallback	*snmps;
	LanguageCallback	*traps;
#endif

} Perl;
@}


@d Perl static variables
@{
Perl *CurrentPerl = NULL;
@}


@d PerlInit prototype
@{
VOID *PerlInit(RunGetsFunction rungets,
               RunPutsFunction runputs,
			   VOID *run);
@}

@d PerlInit code
@{
VOID *PerlInit(RunGetsFunction rungets,
               RunPutsFunction runputs,
			   VOID *run)
{
#ifdef SUBSCRIPT_PERL
    Perl *perl;
	int result;

	PerlContextInit();

    /* allocate a Perl structure */
    if ((perl = PerlNew()) == NULL)
        return NULL;
    perl->rungets = rungets;
    perl->runputs = runputs;
    perl->run = run;

	result = perl_parse(perl->env, perl_ext_init, 0, NULL, NULL);

	if (result)
	{
	}

    return perl;
#else
	return NULL;
#endif
}
@}


@d Perl external extensions callback prototype
@{
static VOID perl_ext_init(VOID);
@}

@d Perl external extensions callback code
@{
VOID perl_ext_init(VOID)
{
	char *file = __FILE__;
	{
		extern void boot_DynaLoader(CV *cv);
		newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
	}
}
@}


@d PerlCall prototype
@{
BOOLEAN PerlCall(Perl *perl, int command);
@}

@d PerlCall code
@{
BOOLEAN PerlCall(Perl *perl, int command)
{
#ifdef SUBSCRIPT_PERL
	if ((perl = PerlContextSwitch(perl)) == NULL)
		return FALSE;

    switch (command)
    {
    case RUN_COMMAND_INIT:
        if (perl->init == NULL)
            return TRUE;
        break;
    case RUN_COMMAND_START:
        if (perl->start == NULL)
            return TRUE;
        break;
    case RUN_COMMAND_STOP:
    	SuspendCallbacks(perl, SCRIPT_LANGUAGE_PERL, TRUE);
        if (perl->stop == NULL)
            return TRUE;
        break;
    case RUN_COMMAND_RESUME:
    	SuspendCallbacks(perl, SCRIPT_LANGUAGE_PERL, FALSE);
        if (perl->resume == NULL)
            return TRUE;
        break;
    case RUN_COMMAND_EXIT:
        if (perl->exit == NULL)
            return TRUE;
        break;
    default:
        return FALSE;
    }

	return TRUE;
#else
	return FALSE;
#endif
}
@}


@d PerlExit prototype
@{
VOID PerlExit(Perl *perl);
@}

@d PerlExit code
@{
VOID PerlExit(Perl *perl)
{
#ifdef SUBSCRIPT_PERL
	if ((perl = PerlContextSwitch(perl)) == NULL)
		return;

   	RemoveAllCallbacks(perl, SCRIPT_LANGUAGE_PERL);

	perl_destruct(perl->env, 0);
	perl_free(perl->env);

	PerlDel(perl);
#else
#endif
}
@}


@d PerlNew prototype
@{
Perl *PerlNew(VOID);
@}

@d PerlNew code
@{
Perl *PerlNew(VOID)
{
    Perl *perl;

    if ((perl = DnpapMalloc(sizeof(Perl))) == NULL)
        return NULL;

    perl->env = NULL;
    perl->init = NULL;
    perl->start = NULL;
    perl->stop = NULL;
    perl->resume = NULL;
    perl->exit = NULL;

    /* create a new perl environment */
    if ((perl->env = perl_alloc()) == NULL)
    {
        PerlDel(perl);
        return NULL;
    }

	perl_construct(perl->env);

    return perl;
}
@}


@d PerlDel prototype
@{
VOID PerlDel(Perl *perl);
@}

@d PerlDel code
@{
VOID PerlDel(Perl *perl)
{
    DnpapFree(perl);
}
@}


@d PerlContextInit prototype
@{
VOID PerlContextInit(VOID);
@}

@d PerlContextInit code
@{
VOID PerlContextInit(VOID)
{
	CurrentPerl = NULL;
}
@}


@d PerlContextSwitch prototype
@{
Perl *PerlContextSwitch(Perl *perl);
@}

@d PerlContextSwitch code
@{
Perl *PerlContextSwitch(Perl *perl)
{
	CurrentPerl = perl;
	if (CurrentPerl == NULL)
	{
		DnpapMessage(DMC_FATAL, SUBSCRIPT_CONTEXT, "subscript: no context available");
		return NULL;
	}
	return CurrentPerl;
}
@}


@d PerlContextCurrent prototype
@{
Perl *PerlContextCurrent(VOID);
@}

@d PerlContextCurrent code
@{
Perl *PerlContextCurrent(VOID)
{
	return CurrentPerl;
}
@}


\section{Makefile Dependencies}

Makefile dependencies for Scheme and Perl language support.

@d Makefile dependencies
@{
doc: $(SYS).dvi

$(SYS).dvi: $(SYS).tex
	$(TEX) $(SYS).tex
	$(WEAVE) $(SYS).w
	$(TEX) $(SYS).tex

$(SYS).tex: $(SYS).w
	$(WEAVE) $(SYS).w

obj: $(SYS)m.o $(SYS)c.o $(SYS)l.o

$(SYS)m.o: $(SYS)m.c $(SYS)c.h $(SYS)x.h $(SYS)l.c $(SYS).h
$(SYS)c.o: $(SYS)c.c $(SYS)c.h $(SYS)x.h $(SYS)l.c
$(SYS)l.o: $(SYS)l.c $(SYS)c.h
@}




%
% Makefiles
%

\chapter{Makefiles}

First include the standard Makefile rules file.

@d Makefile include
@{
include ../../include/gcc_unx.std
@}

@d Makefile settings
@{
MCFLAGS = -I/data/d4/develop

SYS	= subscript
SYSLIB	= subscript

WEAVE = nuweb -o
TANGLE = nuweb -t

TEX = latex

TOUCH = touch

.SUFFIXES: .w .tex .h .c .dvi .o .ps
@}

@o gcc_unx.mak -t
@{
@<Makefile include@>
@<Makefile settings@>

all: tangle.ok obj #doc

install: all $(INCDIR)/$(SYS).h $(INCDIR)/$(SYS)x.h $(INCDIR)/$(SYS)l.h $(LIBDIR)/lib$(SYSLIB).a

$(INCDIR)/$(SYS).h: $(SYS).h
	$(CP) $(SYS).h $(INCDIR)/$(SYS).h

$(INCDIR)/$(SYS)x.h: $(SYS)x.h
	$(CP) $(SYS)x.h $(INCDIR)/$(SYS)x.h

$(INCDIR)/$(SYS)l.h: $(SYS)l.h
	$(CP) $(SYS)l.h $(INCDIR)/$(SYS)l.h

$(LIBDIR)/lib$(SYSLIB).a: $(SYS)m.o $(SYS)c.o $(SYS)l.o
	$(AR) $(LIBDIR)/lib$(SYSLIB).a $(SYS)m.o $(SYS)c.o $(SYS)l.o
	ranlib $(LIBDIR)/lib$(SYSLIB).a

@<Makefile clean rule@>

tangle.ok: $(SYS).w
	$(TANGLE) $(SYS).w
	$(TOUCH) tangle.ok

@<Makefile dependencies@>
@}


@d Makefile clean rule
@{
clean:
	- $(RM) *.exe
	- $(RM) *.obj
	- $(RM) *.o
	- $(RM) *.i
	- $(RM) *.lib
	- $(RM) *.man
	- $(RM) *.err
	- $(RM) *.bak
@}



\end{document}
