
#line 1997 "subscript.w"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <dnpap.h>
#include <config.h>

#include "subscriptl.h"
#include "subscriptc.h"


#line 2698 "subscript.w"

SourceLine *SourceLineNew(VOID);

#line 2722 "subscript.w"

VOID SourceLineDel(SourceLine *line);

#line 2758 "subscript.w"

LogLine *LogLineNew(VOID);

#line 2781 "subscript.w"

VOID LogLineDel(LogLine *log);

#line 2008 "subscript.w"



#line 2039 "subscript.w"

static CHAR *StaticLog = "error";

static CHAR *ScriptDirectory = NULL;

#line 2010 "subscript.w"



#line 2052 "subscript.w"

BOOLEAN SubscriptConfigInit(VOID)
{
	ScriptDirectory = SCRIPT_DIRECTORY;
	ConfigGetString("subscript.directory", &ScriptDirectory);
	return TRUE;
}

#line 2012 "subscript.w"



#line 2067 "subscript.w"

BOOLEAN SubscriptSourceCInit(Script *script)
{
	return TRUE;
}

#line 2014 "subscript.w"


#line 2080 "subscript.w"

BOOLEAN SubscriptSourceCStart(Script *script)
{
    if (script->Source->Lines == 0)
		return FALSE;

	switch (script->Language)
	{
	case SCRIPT_LANGUAGE_SCHEME:
	case SCRIPT_LANGUAGE_PERL:
		break;
	default:
		return FALSE;
	}

	return TRUE;
}

#line 2015 "subscript.w"


#line 2105 "subscript.w"

BOOLEAN SubscriptSourceCStop(Script *script)
{
	if (script->Runs != NULL)
		return FALSE;
	return TRUE;
}

#line 2016 "subscript.w"



#line 2121 "subscript.w"

BOOLEAN SubscriptRunCInit(Run *run)
{
	return TRUE;
}

#line 2018 "subscript.w"


#line 2134 "subscript.w"

BOOLEAN SubscriptRunCStart(Run *run)
{
    if (run->Script == 0)
		return FALSE;
	if ((run->ScriptHook = SubscriptRegisterRun(run->Script)) == NULL)
		return FALSE;
	switch (SubscriptSourceLanguage(run->ScriptHook))
	{
	case SCRIPT_LANGUAGE_SCHEME:
		run->Languages.Scheme = SchemeInit(RunGets, RunPuts, run);
    	if (run->Languages.Scheme == NULL)
	    {
		    SubscriptDeRegisterRun(run->Script, run->ScriptHook);
			return FALSE;
	    }
	    if (SchemeCall(run->Languages.Scheme, RUN_COMMAND_INIT) == FALSE)
	    {
			SchemeExit(run->Languages.Scheme);
		    SubscriptDeRegisterRun(run->Script, run->ScriptHook);
			return FALSE;
	    }
	    run->Command = RUN_COMMAND_INIT;
		break;
	case SCRIPT_LANGUAGE_PERL:
		run->Languages.Perl = PerlInit(RunGets, RunPuts, run);
    	if (run->Languages.Perl == NULL)
	    {
		    SubscriptDeRegisterRun(run->Script, run->ScriptHook);
			return FALSE;
	    }
	    if (PerlCall(run->Languages.Perl, RUN_COMMAND_INIT) == FALSE)
	    {
			PerlExit(run->Languages.Perl);
		    SubscriptDeRegisterRun(run->Script, run->ScriptHook);
			return FALSE;
	    }
	    run->Command = RUN_COMMAND_INIT;
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

#line 2019 "subscript.w"


#line 2186 "subscript.w"

BOOLEAN SubscriptRunCStop(Run *run)
{
	switch (SubscriptSourceLanguage(run->ScriptHook))
	{
	case SCRIPT_LANGUAGE_SCHEME:
	    SchemeCall(run->Languages.Scheme, RUN_COMMAND_EXIT);
	    run->Command = RUN_COMMAND_EXIT;
		SchemeExit(run->Languages.Scheme);
		run->Languages.Scheme = NULL;
		break;
	case SCRIPT_LANGUAGE_PERL:
	    PerlCall(run->Languages.Perl, RUN_COMMAND_EXIT);
	    run->Command = RUN_COMMAND_EXIT;
		PerlExit(run->Languages.Perl);
		run->Languages.Perl = NULL;
		break;
	default:
		return FALSE;
	}
	SubscriptDeRegisterRun(run->Script, run->ScriptHook);
	return TRUE;
}

#line 2020 "subscript.w"



#line 2218 "subscript.w"

BOOLEAN RunDoCommand(Run *run, int command)
{
BOOLEAN result;

	switch (SubscriptSourceLanguage(run->ScriptHook))
	{
	case SCRIPT_LANGUAGE_SCHEME:
		result = SchemeCall(run->Languages.Scheme, command);
        if (result == FALSE)
			return FALSE;
	    run->Command = command;
		break;
	case SCRIPT_LANGUAGE_PERL:
		result = PerlCall(run->Languages.Perl, command);
        if (result == FALSE)
			return FALSE;
	    run->Command = command;
		break;
	default:
		return FALSE;
	}

	return TRUE;
}

#line 2022 "subscript.w"



#line 2252 "subscript.w"

CHAR *RunGets(VOID *r, int line)
{
Run *run = (Run *) r;
SourceLine *l;

	if ((l = SourceListGetLine(SubscriptSourceList(run->ScriptHook), line)) == NULL)
		return NULL;
	return l->Line;
}

#line 2024 "subscript.w"


#line 2270 "subscript.w"

int RunPuts(VOID *r, int line, CHAR *log)
{
Run *run = (Run *) r;
int len;

	if (line == 0)
		line = run->LogLine++;
	if (LogListAddLine(run->Logs, line, log, len = strlen(log)) == FALSE)
		return -1;
	return len;
}

#line 2025 "subscript.w"



#line 2290 "subscript.w"

ScriptHook *SubscriptAddRun(Script *script)
{
ScriptHook *hook;

    if ((hook = DnpapMalloc(sizeof(ScriptHook))) == NULL)
        return NULL;
    hook->Script = script;
    hook->Prev = NULL;
    hook->Next = script->Runs;
    if (script->Runs != NULL)
        script->Runs->Prev = hook;
        script->Runs = hook;

    return hook;
}

#line 2027 "subscript.w"


#line 2315 "subscript.w"

BOOLEAN SubscriptRemoveRun(Script *script, ScriptHook *hook)
{
    if (hook == NULL)
        return TRUE;
    if (hook->Next != NULL)
        hook->Next->Prev = hook->Prev;
    if (hook->Prev == NULL)
        script->Runs = hook->Next;
    else
        hook->Prev->Next = hook->Next;

    DnpapFree(hook);

    return TRUE;
}

#line 2028 "subscript.w"



#line 2339 "subscript.w"

int SubscriptSourceLanguage(ScriptHook *hook)
{
	if (hook == NULL)
		return SCRIPT_LANGUAGE_NULL;
	return hook->Script->Language;
}

#line 2030 "subscript.w"


#line 2355 "subscript.w"

SourceList *SubscriptSourceList(ScriptHook *hook)
{
	if (hook == NULL)
		return NULL;
	return hook->Script->Source;
}

#line 2031 "subscript.w"



#line 2371 "subscript.w"

int SubscriptRunCommand(Run *run)
{
	if (run == NULL)
		return RUN_COMMAND_NULL;
	return run->Command;
}

#line 2033 "subscript.w"


#line 2386 "subscript.w"

int SubscriptRunIndex(Run *run)
{
	if (run == NULL)
		return 0;
	return run->Index;
}

#line 2034 "subscript.w"


#line 2401 "subscript.w"

int SubscriptRunScript(Run *run)
{
	if (run == NULL)
		return 0;
	return run->Script;
}

#line 2035 "subscript.w"


#line 2589 "subscript.w"

BOOLEAN SourceListAddLine(SourceList *list, int line, CHAR *data, int len)
{
SourceLine *l;
CHAR *new;

    if (data == NULL)
        return FALSE;

	if (line < 1 || line > list->Lines + 1)
		return FALSE;

	if (line == list->Lines + 1)
	{
		if ((l = SourceLineNew()) == NULL)
			return FALSE;
        l->LineNumber = line;
		if (list->Tail != NULL)
			list->Tail->Next = l;
		else
			list->Head = l;
		l->Prev = list->Tail;
		list->Tail = l;
		list->Lines++;
	}
	else
	if ((l = SourceListGetLine(list, line)) == NULL)
		return FALSE;

	if ((new = DnpapRealloc(l->Line, (ULONG)len+1)) == NULL)
		return FALSE;

	strncpy(l->Line = new, data, len);
    l->Line[len]='\0';

	return TRUE;
}

#line 2635 "subscript.w"

BOOLEAN SourceListDelLine(SourceList *list, int line)
{
SourceLine *last;

	if (line < 1 || line != list->Lines)
		return FALSE;

    last = list->Tail;
	list->Current = list->Tail = list->Tail->Prev;

	if (list->Tail == NULL)
		list->Head = NULL;
	else
		list->Tail->Next = NULL;
	list->Lines--;

	SourceLineDel(last);

	return TRUE;
}

#line 2664 "subscript.w"

SourceList *SourceListNew(VOID)
{
SourceList *list;

	if ((list = DnpapMalloc(sizeof(SourceList))) == NULL)
		return NULL;

	list->Lines = 0;
	list->Head = NULL;
	list->Tail = NULL;
	list->Current = NULL;

	return list;
}

#line 2687 "subscript.w"

VOID SourceListDel(SourceList *list)
{
	while (list->Lines > 0)
		SourceListDelLine(list, list->Lines);
	DnpapFree(list);
}

#line 2703 "subscript.w"

SourceLine *SourceLineNew(VOID)
{
SourceLine *line;

	if ((line = DnpapMalloc(sizeof(SourceLine))) == NULL)
		return NULL;

	line->LineNumber = 0;
	line->Line = NULL;
	line->Prev = NULL;
	line->Next = NULL;

	return line;
}

#line 2727 "subscript.w"

VOID SourceLineDel(SourceLine *line)
{
	DnpapFree(line);
}

#line 2741 "subscript.w"

LogList *LogListNew(VOID)
{
LogList *list;

	if ((list = DnpapMalloc(sizeof(LogList))) == NULL)
		return NULL;

	list->Lines = 0;
	list->Head = NULL;
	list->Current = NULL;

	return list;
}

#line 2763 "subscript.w"

LogLine *LogLineNew(VOID)
{
LogLine *log;

	if ((log = DnpapMalloc(sizeof(LogLine))) == NULL)
		return NULL;

	log->LineNumber = 0;
	log->Line = NULL;
	log->Next = NULL;

	return log;
}

#line 2786 "subscript.w"

VOID LogLineDel(LogLine *log)
{
    if (log->Line != StaticLog)
        DnpapFree(log->Line);
	DnpapFree(log);
}

#line 2802 "subscript.w"

VOID LogListDel(LogList *logs)
{
	while (logs->Lines > 0)
		LogListDelFirstLine(logs);
	DnpapFree(logs);
}

#line 2818 "subscript.w"

VOID LogListDelLines(LogList *logs)
{
	while (logs->Lines > 0)
		LogListDelFirstLine(logs);
}

#line 2833 "subscript.w"

SourceLine *SourceListGetLine(SourceList *list, int line)
{
	if (line < 1 || line > list->Lines)
		return NULL;
    if (list->Current != NULL && list->Current->LineNumber == line)
        return list->Current;
	for (list->Current = list->Head; list->Current != NULL; list->Current = list->Current->Next)
        if (list->Current->LineNumber >= line)
            break;
    if (list->Current != NULL && list->Current->LineNumber == line)
    	return list->Current;
    return NULL;
}

#line 2856 "subscript.w"

SourceLine *SourceListGetNextLine(SourceList *list, int line)
{
    if (line < 1)
        return list->Current = list->Head;
	if ((list->Current = SourceListGetLine(list, line)) == NULL)
		return NULL;
	return list->Current = list->Current->Next;
}

#line 2874 "subscript.w"

LogLine *LogListGetLine(LogList *list, int line)
{
	if (line < 1)
		return NULL;
    if (list->Current != NULL && list->Current->LineNumber == line)
        return list->Current;
	for (list->Current = list->Head; list->Current != NULL; list->Current = list->Current->Next)
        if (list->Current->LineNumber >= line)
            break;
    if (list->Current != NULL && list->Current->LineNumber == line)
    	return list->Current;
	return NULL;
}

#line 2897 "subscript.w"

LogLine *LogListGetNextLine(LogList *list, int line)
{
    if (line < 1)
        return list->Current = list->Head;
	if ((list->Current = LogListGetLine(list, line)) == NULL)
		return NULL;
	return list->Current = list->Current->Next;
}

#line 2915 "subscript.w"

BOOLEAN LogListAddLine(LogList *list, int line, CHAR *log, int len)
{
LogLine **l, *entry;
CHAR *new;

    if (log == NULL)
        return FALSE;

	for (l = &list->Head; *l != NULL; l = &(*l)->Next)
		if (line <= (*l)->LineNumber)
			break;
    entry = *l;
	if (entry == NULL || line != entry->LineNumber)
	{
		if ((entry = LogLineNew()) == NULL)
			return FALSE;
        entry->LineNumber = line;
		entry->Next = *l;
		*l = entry;
        list->Lines++;
	}

	if (entry->Line == NULL)
	{
		if ((entry->Line = DnpapMalloc((ULONG)len+1)) == NULL)
		{
			/* let log string point to well-known static string */
			entry->Line = StaticLog;
			return FALSE;
		}
		strncpy(entry->Line, log, len);
        entry->Line[len] = '\0';
	}
	else
	{
		if ((new = DnpapRealloc(entry->Line, strlen(entry->Line) + strlen("\n") + len + 1)) == NULL)
			return FALSE;
		entry->Line = new;
		sprintf(entry->Line + strlen(entry->Line), "\n%*s", len, log);
	}

	return TRUE;
}

#line 2968 "subscript.w"

BOOLEAN LogListDelFirstLine(LogList *list)
{
LogLine *line;

    if (list->Head == NULL)
        return FALSE;
    line = list->Head;
    list->Current = list->Head = list->Head->Next;

    list->Lines--;

    LogLineDel(line);

    return TRUE;
}

#line 2993 "subscript.w"

Script *ScriptNew(VOID)
{
Script *script;

    if ((script = DnpapMalloc(sizeof(Script))) == NULL)
        return NULL;

    script->Index = 0;
	script->FileName[0] = '\0';
    script->FileNameLen = 0;
	script->FileStatus = FILE_STATUS_NULL;
    script->Language = SCRIPT_LANGUAGE_NULL;
    script->Owner[0] = '\0';
    script->OwnerLen = 0;
	script->Status = SNMP_INVALID;

    if ((script->Source = SourceListNew()) == NULL)
    {
        ScriptDel(script);
        return NULL;
    }

	script->Runs = NULL;

    return script;
}

#line 3029 "subscript.w"

VOID ScriptDel(Script *script)
{
    SourceListDel(script->Source);
    script->Source = NULL;
    DnpapFree(script);
}

#line 3045 "subscript.w"

Run *RunNew(VOID)
{
Run *run;

    if ((run = DnpapMalloc(sizeof(Run))) == NULL)
        return NULL;

    run->Index = 0;
    run->Script = 0;
    run->ScriptHook = NULL;
    run->Command = RUN_COMMAND_NULL;
    run->Owner[0] = '\0';
    run->OwnerLen = 0;
	run->Status = SNMP_INVALID;

    if ((run->Logs = LogListNew()) == NULL)
    {
        RunDel(run);
        return NULL;
    }
	run->LogLine = 1;

    return run;
}

#line 3079 "subscript.w"

VOID RunDel(Run *run)
{
    LogListDel(run->Logs);
    run->Logs = NULL;
    DnpapFree(run);
}

#line 3099 "subscript.w"

SourceList *FileRead(Script *script)
{
SourceList *source;
CHAR name[SNMP_SIZE_BUFCHR];
CHAR line[SNMP_SIZE_BUFCHR];
FILE *f;
int i;

    script->FileStatus = FILE_STATUS_INVALID;

    if (script->OwnerLen == 0 || script->FileNameLen == 0)
        return NULL;

    
#line 3222 "subscript.w"
    
    {
    CHAR  owner[SNMP_SIZE_BUFCHR],
    	 *o1 = NULL, *o2 = NULL, *o3 = NULL, *o4 = NULL,
    	 *p;
    
    	strncpy(owner, script->Owner, script->OwnerLen);
    	owner[script->OwnerLen] = '\0';
    	o1 = strtok(owner, " \t");
    	if (o1 != NULL)
    		o2 = strtok(NULL, " \t");
    	if (o2 != NULL)
    		o3 = strtok(NULL, " \t");
    	if (o3 != NULL)
    		o4 = strtok(NULL, " \t");
    
    	if (o2 != NULL && o3 != NULL)
    	{
    		if ((p = strchr(o2, '(')) != NULL)
    			*p = '\0';
    		if ((p = strchr(o3, '(')) != NULL)
    			*p = '\0';
        	sprintf(name, "%s/%s/%s/%*s", ScriptDirectory,
                                    	  o2, o3,
                                    	  script->FileNameLen, script->FileName);
    	}
    	else
    	if (o1 != NULL)
    	{
        	sprintf(name, "%s/%s/%*s", ScriptDirectory,
                                  	   o1,
                                       script->FileNameLen, script->FileName);
    	}
    	else
    	{
        	sprintf(name, "%s/%*s", ScriptDirectory,
                                    script->FileNameLen, script->FileName);
    	}
    }
    
#line 3113 "subscript.w"


    if ((source = SourceListNew()) == NULL)
        return FALSE;

    script->FileStatus = FILE_STATUS_NEW;

    if ((f = fopen(name, "r")) == NULL)
    {
        SourceListDel(source);
        return NULL;
    }

    script->FileStatus = FILE_STATUS_EXISTS;

    for (i = 1; fgets(line, sizeof(line), f) != (char *)EOF; i++)
    {
        if (SourceListAddLine(source, i, line, (int)strlen(line)) == FALSE)
        {
            fclose(f);
            SourceListDel(source);
            return NULL;
        }
    }

    fclose(f);

    return source;
}

#line 3151 "subscript.w"

BOOLEAN FileWrite(Script *script)
{
CHAR name[SNMP_SIZE_BUFCHR];
FILE *f;
int i;
SourceLine *line;

    script->FileStatus = FILE_STATUS_INVALID;

    if (script->OwnerLen == 0 || script->FileNameLen == 0)
        return FALSE;

    
#line 3222 "subscript.w"
    
    {
    CHAR  owner[SNMP_SIZE_BUFCHR],
    	 *o1 = NULL, *o2 = NULL, *o3 = NULL, *o4 = NULL,
    	 *p;
    
    	strncpy(owner, script->Owner, script->OwnerLen);
    	owner[script->OwnerLen] = '\0';
    	o1 = strtok(owner, " \t");
    	if (o1 != NULL)
    		o2 = strtok(NULL, " \t");
    	if (o2 != NULL)
    		o3 = strtok(NULL, " \t");
    	if (o3 != NULL)
    		o4 = strtok(NULL, " \t");
    
    	if (o2 != NULL && o3 != NULL)
    	{
    		if ((p = strchr(o2, '(')) != NULL)
    			*p = '\0';
    		if ((p = strchr(o3, '(')) != NULL)
    			*p = '\0';
        	sprintf(name, "%s/%s/%s/%*s", ScriptDirectory,
                                    	  o2, o3,
                                    	  script->FileNameLen, script->FileName);
    	}
    	else
    	if (o1 != NULL)
    	{
        	sprintf(name, "%s/%s/%*s", ScriptDirectory,
                                  	   o1,
                                       script->FileNameLen, script->FileName);
    	}
    	else
    	{
        	sprintf(name, "%s/%*s", ScriptDirectory,
                                    script->FileNameLen, script->FileName);
    	}
    }
    
#line 3164 "subscript.w"


    script->FileStatus = FILE_STATUS_NEW;

    if ((f = fopen(name, "w")) == NULL)
        return FALSE;

    script->FileStatus = FILE_STATUS_EXISTS;

    for (i = 1; i <= script->Source->Lines; i++)
    {
        if ((line = SourceListGetLine(script->Source, i)) == NULL ||
            fputs(line->Line, f) == EOF)
        {
            fclose(f);
            return FALSE;
        }
    }

    fclose(f);

    return TRUE;
}

#line 3196 "subscript.w"

BOOLEAN FileExists(Script *script)
{
CHAR name[SNMP_SIZE_BUFCHR];
FILE *f;

    script->FileStatus = FILE_STATUS_INVALID;

    if (script->OwnerLen == 0 || script->FileNameLen == 0)
        return FALSE;

    
#line 3222 "subscript.w"
    
    {
    CHAR  owner[SNMP_SIZE_BUFCHR],
    	 *o1 = NULL, *o2 = NULL, *o3 = NULL, *o4 = NULL,
    	 *p;
    
    	strncpy(owner, script->Owner, script->OwnerLen);
    	owner[script->OwnerLen] = '\0';
    	o1 = strtok(owner, " \t");
    	if (o1 != NULL)
    		o2 = strtok(NULL, " \t");
    	if (o2 != NULL)
    		o3 = strtok(NULL, " \t");
    	if (o3 != NULL)
    		o4 = strtok(NULL, " \t");
    
    	if (o2 != NULL && o3 != NULL)
    	{
    		if ((p = strchr(o2, '(')) != NULL)
    			*p = '\0';
    		if ((p = strchr(o3, '(')) != NULL)
    			*p = '\0';
        	sprintf(name, "%s/%s/%s/%*s", ScriptDirectory,
                                    	  o2, o3,
                                    	  script->FileNameLen, script->FileName);
    	}
    	else
    	if (o1 != NULL)
    	{
        	sprintf(name, "%s/%s/%*s", ScriptDirectory,
                                  	   o1,
                                       script->FileNameLen, script->FileName);
    	}
    	else
    	{
        	sprintf(name, "%s/%*s", ScriptDirectory,
                                    script->FileNameLen, script->FileName);
    	}
    }
    
#line 3207 "subscript.w"


    script->FileStatus = FILE_STATUS_NEW;

    if ((f = fopen(name, "r")) == NULL)
        return FALSE;
    fclose(f);

    script->FileStatus = FILE_STATUS_EXISTS;

    return TRUE;
}
