
#line 1951 "subscript.w"

#ifndef _SUBSCRIPTC_H
#define _SUBSCRIPTC_H


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dnpap.h>
#include <config.h>
#include <boolset.h>
#include <snmp.h>

#include "subscriptx.h"


#line 2411 "subscript.w"

enum SubscriptErrors
{
    SUBSCRIPT_CREATE = SUBSCRIPT_ERROR,
    SUBSCRIPT_ACTIVE, SUBSCRIPT_DESTROY,
	SUBSCRIPT_CONTEXT
};

#line 1966 "subscript.w"


#line 2421 "subscript.w"

#define SCRIPT_LANGUAGE_NULL		0
#define SCRIPT_LANGUAGE_SCHEME		1
#define SCRIPT_LANGUAGE_PERL		2

#line 1967 "subscript.w"


#line 2438 "subscript.w"

#define SCRIPT_DIRECTORY		"/usr/local/project/scripts"

#define FILE_STATUS_NULL			0
#define FILE_STATUS_SAVE			1
#define FILE_STATUS_OVERWRITE		2
#define FILE_STATUS_RELOAD			3
#define FILE_STATUS_EXISTS			4
#define FILE_STATUS_NEW				5
#define FILE_STATUS_INVALID			6

#line 1968 "subscript.w"


#line 2451 "subscript.w"


#line 2466 "subscript.w"

typedef struct SourceLine_ SourceLine;

struct SourceLine_
{
    int LineNumber;
	CHAR *Line;

	SourceLine *Prev;
	SourceLine *Next;
};

#line 2452 "subscript.w"


#line 2480 "subscript.w"

typedef struct SourceList_
{
	int Lines;
	SourceLine *Head;
	SourceLine *Tail;
	SourceLine *Current;

} SourceList;

#line 2453 "subscript.w"



#line 2492 "subscript.w"

typedef struct Script_
{
    ULONG Index;
	CHAR FileName[SNMP_SIZE_BUFCHR];
    WORD FileNameLen;
	LONG FileStatus;
    LONG Language;
	CHAR Owner[SNMP_SIZE_BUFCHR];
    WORD OwnerLen;
	LONG Status;

	BooleanSet ObjectSet;

	SourceList *Source;

	ScriptHook *Runs;

} Script;

#line 2455 "subscript.w"



#line 2514 "subscript.w"

typedef struct LogLine_ LogLine;

struct LogLine_
{
    int LineNumber;
	CHAR *Line;

	LogLine *Next;
};

#line 2457 "subscript.w"


#line 2527 "subscript.w"

typedef struct LogList_
{
	int Lines;
	LogLine *Head;
	LogLine *Current;

} LogList;

#line 2458 "subscript.w"



#line 2538 "subscript.w"

typedef struct Run_
{
    ULONG Index;
    ULONG Script;
    ScriptHook *ScriptHook;
    LONG Command;
	CHAR Owner[SNMP_SIZE_BUFCHR];
    WORD OwnerLen;
	LONG Status;

	BooleanSet ObjectSet;

	LogList *Logs;
	int LogLine;

	union
	{
		VOID *Scheme;
		VOID *Perl;
	} Languages;

} Run;

#line 2460 "subscript.w"



#line 2564 "subscript.w"

struct ScriptHook_
{
    Script *Script;
    ScriptHook *Prev;
    ScriptHook *Next;
};

#line 2462 "subscript.w"


#line 1969 "subscript.w"


#line 2574 "subscript.w"

#define SUBSCRIPT_SOURCE_BOOLEAN_LANGUAGE	0
#define SUBSCRIPT_SOURCE_BOOLEAN_OWNER		1

#define SUBSCRIPT_RUN_BOOLEAN_SCRIPT		0
#define SUBSCRIPT_RUN_BOOLEAN_OWNER			1

#line 1970 "subscript.w"



#line 2047 "subscript.w"

BOOLEAN SubscriptConfigInit(VOID);

#line 1972 "subscript.w"



#line 2062 "subscript.w"

BOOLEAN SubscriptSourceCInit(Script *script);

#line 1974 "subscript.w"


#line 2075 "subscript.w"

BOOLEAN SubscriptSourceCStart(Script *script);

#line 1975 "subscript.w"


#line 2100 "subscript.w"

BOOLEAN SubscriptSourceCStop(Script *script);

#line 1976 "subscript.w"



#line 2116 "subscript.w"

BOOLEAN SubscriptRunCInit(Run *run);

#line 1978 "subscript.w"


#line 2129 "subscript.w"

BOOLEAN SubscriptRunCStart(Run *run);

#line 1979 "subscript.w"


#line 2181 "subscript.w"

BOOLEAN SubscriptRunCStop(Run *run);

#line 1980 "subscript.w"



#line 2213 "subscript.w"

BOOLEAN RunDoCommand(Run *run, int command);

#line 1982 "subscript.w"



#line 2247 "subscript.w"

CHAR *RunGets(VOID *run, int line);

#line 1984 "subscript.w"


#line 2265 "subscript.w"

int RunPuts(VOID *run, int line, CHAR *log);

#line 1985 "subscript.w"



#line 3094 "subscript.w"

SourceList *FileRead(Script *script);

#line 3146 "subscript.w"

BOOLEAN FileWrite(Script *script);

#line 3191 "subscript.w"

BOOLEAN FileExists(Script *script);

#line 1987 "subscript.w"



#line 2285 "subscript.w"

ScriptHook *SubscriptAddRun(Script *script);

#line 2310 "subscript.w"

BOOLEAN SubscriptRemoveRun(Script *script, ScriptHook *hook);

#line 2334 "subscript.w"

int SubscriptSourceLanguage(ScriptHook *hook);

#line 2350 "subscript.w"

SourceList *SubscriptSourceList(ScriptHook *hook);

#line 2584 "subscript.w"

BOOLEAN SourceListAddLine(SourceList *list, int line, CHAR *data, int len);

#line 2630 "subscript.w"

BOOLEAN SourceListDelLine(SourceList *list, int line);

#line 2659 "subscript.w"

SourceList *SourceListNew(VOID);

#line 2682 "subscript.w"

VOID SourceListDel(SourceList *list);

#line 2828 "subscript.w"

SourceLine *SourceListGetLine(SourceList *list, int line);

#line 2851 "subscript.w"

SourceLine *SourceListGetNextLine(SourceList *list, int line);

#line 2988 "subscript.w"

Script *ScriptNew(VOID);

#line 3024 "subscript.w"

VOID ScriptDel(Script *script);

#line 1989 "subscript.w"


#line 2366 "subscript.w"

int SubscriptRunCommand(Run *run);

#line 2381 "subscript.w"

int SubscriptRunIndex(Run *run);

#line 2396 "subscript.w"

int SubscriptRunScript(Run *run);

#line 2736 "subscript.w"

LogList *LogListNew(VOID);

#line 2797 "subscript.w"

VOID LogListDel(LogList *logs);

#line 2813 "subscript.w"

VOID LogListDelLines(LogList *logs);

#line 2869 "subscript.w"

LogLine *LogListGetLine(LogList *list, int line);

#line 2892 "subscript.w"

LogLine *LogListGetNextLine(LogList *list, int line);

#line 2910 "subscript.w"

BOOLEAN LogListAddLine(LogList *list, int line, CHAR *log, int len);

#line 2963 "subscript.w"

BOOLEAN LogListDelFirstLine(LogList *list);

#line 3040 "subscript.w"

Run *RunNew(VOID);

#line 3074 "subscript.w"

VOID RunDel(Run *run);

#line 1990 "subscript.w"



#endif
