#include <dnpap.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "dkvm.h"


Kvm *KvmOpen(INT flags)
{
Kvm *kvm;

#ifdef SUNOS
	if ((kvm = kvm_open(NULL, NULL, NULL, flags, NULL)) == NULL)
		return NULL;
#else
INT fd;

	if ((fd = open (KMEMDEVICE, flags, 0)) < 0)
		return NULL;
	if ((kvm = DnpapMalloc(sizeof(Kvm))) == NULL)
		return NULL;
	*kvm = fd;
#endif
	return kvm;
}


VOID KvmClose(Kvm *kvm)
{
#ifdef SUNOS
	if (kvm != NULL)
		kvm_close(kvm);
#else
	if (kvm != NULL && *kvm != 0)
		close(*kvm);
	DnpapFree(kvm);
#endif
}


INT KvmNList(Kvm *kvm, NList *nl)
{
#ifdef SUNOS
	return kvm_nlist(kvm, nl);
#else
	return nlist(KMEMNLIST, nl);
#endif
}


INT KvmRead(Kvm *kvm, ULONG address, VOID *buf, UINT len)
{
#ifdef SUNOS
	return kvm_read(kvm, address, buf, len);
#else
	if (lseek(*kvm, address, SEEK_SET) == -1)
		return -1;
	return read(*kvm, buf, len);
#endif
}


INT KvmReadSym(Kvm *kvm, CHAR *name, VOID *buf, UINT len)
{
NList nl[] = { { name },
			   { NULL } };

	if (KvmNList(kvm, nl) < 0 || nl[0].n_type == 0)
		return -1;
	return KvmRead(kvm, nl[0].n_value, buf, len);
}


INT KvmWrite(Kvm *kvm, ULONG address, VOID *buf, UINT len)
{
#ifdef SUNOS
	return kvm_write(kvm, address, buf, len);
#else
	if (lseek(*kvm, address, SEEK_SET) == -1)
		return -1;
	return write(*kvm, buf, len);
#endif
}


INT KvmWriteSym(Kvm *kvm, CHAR *name, VOID *buf, UINT len)
{
NList nl[] = { { name },
			   { NULL } };

	if (KvmNList(kvm, nl) < 0 || nl[0].n_type == 0)
		return -1;
	return KvmWrite(kvm, nl[0].n_value, buf, len);
}
