/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <dnpap.h>
#include <memory.h>
#include <maxmem.h>

#include "ring.h"


static LONG RingNormPos(Ring *ring, LONG pos);



Ring *RingAlloc(LONG size, LONG elems)
{
Ring *ring;

	if (elems < 0 || elems > size)
		return FALSE;                               
        
	if ((ring = DnpapMalloc(sizeof(Ring))) == NULL)
		return NULL;
		
	ring->Size = size;
	ring->Start = 0;
	ring->Stop = 0;
	
	if ((ring->PosList = DnpapMalloc(elems*sizeof(RingPos))) == NULL)
	{
		RingFree(ring);
		return NULL;
	}
	
	ring->PosCount = 0;
	ring->PosMax = elems;
	
	if ((ring->Buffer = DnpapMalloc(size)) == NULL)
	{
		RingFree(ring);
		return NULL;
	}
		
	return ring;
}


BOOLEAN RingSetMaxElems(Ring *ring, LONG elems)
{
RingPos *poslist;

	if (ring == NULL || elems < ring->PosCount || elems > ring->Size)
		return FALSE;
		
	if ((poslist = DnpapRealloc(ring->PosList, elems*sizeof(RingPos))) == NULL)
		return FALSE;
	ring->PosList = poslist;
	ring->PosMax = elems;
	
	return TRUE;
}


LONG RingGetMaxElems(Ring *ring)
{						
	if (ring == NULL)
		return 0;
	return ring->PosMax;
}


LONG RingGetElems(Ring *ring)
{
	if (ring == NULL)
		return 0;
	return ring->PosCount;
}


VOID RingFree(Ring *ring)
{
	if (ring == NULL)
		return;
	DnpapFree(ring->PosList);
	DnpapFree(ring->Buffer);
	DnpapFree(ring);
}


BOOLEAN RingPutMem(Ring *ring, BYTE *mem, LONG len)
{
LONG sublen;

	if (ring == NULL || len > RingAvailMem(ring) || ring->PosCount >= ring->PosMax)
		return FALSE;
		
	if (len > (sublen = ring->Size - ring->Stop))
	{
		memcpy(ring->Buffer + ring->Stop, mem, sublen); 
		memcpy(ring->Buffer, mem + sublen, len - sublen); 
	}
	else
	{
		memcpy(ring->Buffer + ring->Stop, mem, len); 
	}
	
	ring->PosList[ring->PosCount].Start = ring->Stop;
	ring->PosList[ring->PosCount].Len = len;
	ring->PosCount++;

	ring->Stop = RingNormPos(ring, ring->Stop + len);
		
	return TRUE;
}


BOOLEAN RingGetMem(Ring *ring, BYTE *mem, LONG *len)
{
BYTE *p;

	if (ring == NULL || ring->PosCount == 0)
		return FALSE;

	if (len != NULL)                                                
	    *len = 0;
		
	if (RingPeekMem(ring, 0, 0, &p, mem, len) == FALSE)
		return FALSE;
			
	ring->Start = RingNormPos(ring, ring->Start + ring->PosList[0].Len);
	
	memcpy(ring->PosList, ring->PosList+1, (ring->PosCount-1)*sizeof(RingPos));
	ring->PosCount--;
	
	return TRUE;
}


BOOLEAN RingPeekMem(Ring *ring, LONG elem, LONG offset, BYTE **p, BYTE *mem, LONG *len)
{
LONG start, sublen, n;

	if (ring == NULL || elem >= ring->PosCount || offset > ring->PosList[elem].Len)
		return FALSE;
                                                             
    start = RingNormPos(ring, ring->PosList[elem].Start + offset);
    
    n = ring->PosList[elem].Len - offset;
    if (mem != NULL && len != NULL &&
        *len > 0 && offset + *len < ring->PosList[elem].Len)
        n = *len;
        
	if (n > (sublen = ring->Size - start))
	{
		if (mem != NULL)
		{
			memcpy(mem, ring->Buffer + start, sublen);
			memcpy(mem + sublen, ring->Buffer, n - sublen);
		}
		*p = NULL;
	}
	else
	{
		if (mem != NULL)
			memcpy(mem, ring->Buffer + start, n);
		*p = ring->Buffer + start;
	}
	    
    if (len != NULL)
        *len = n;
        
	return TRUE;
}


LONG RingAvailMem(Ring *ring)
{
LONG used;
						
	if (ring == NULL)
		return 0;
	used = ring->Stop - ring->Start;
	if (used == 0 && ring->PosCount > 0)
		return 0;
	else
	if (used >= 0)
		return ring->Size - used;
	else
		return -used;
}


LONG RingNormPos(Ring *ring, LONG pos)
{
	if (ring == NULL)
		return -1;
	return pos % ring->Size;
}
