/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <memory.h>
#include <dnpap.h>
#include <message.h>
#ifdef NWNIT
#include "nwnit/nwnit.h"
#endif
#ifdef NWPF
#include "nwpf/nwpf.h"
#endif
#ifdef NWDLPI
#include "nwdlpi/nwdlpi.h"
#endif
#ifdef NWETH
#include "nweth/nweth.h"
#endif
#ifdef NWICMP
#include "nwicmp/nwicmp.h"
#endif
#ifdef NWUDP
#include "nwudp/nwudp.h"
#endif
#ifdef NWTCP
#include "nwtcp/nwtcp.h"
#endif
#ifdef NWSNMP
#include "nwsnmp/nwsnmp.h"
#endif
#ifdef NWFORE
#include "nwfore/nwfore.h"
#endif
#ifdef NWBPF
#include "nwbpf/nwbpf.h"
#endif
#ifdef NWENI
#include "nweni/nweni.h"
#endif
#include "nw.h"


static NwStat nwDgStat, tmpDgStat;
static NwStat nwCoStat, tmpCoStat;


static NW_DG_IFACE *nwDgIface[] =
{
#ifdef NWNIT
	&NitDgIface,
#endif
#ifdef NWPF
	&PfDgIface,
#endif
#ifdef NWDLPI
	&DlpiDgIface,
#endif
#ifdef NWETH
	&EthDgIface,
#endif
#ifdef NWICMP
	&IcmpDgIface,
#endif
#ifdef NWUDP
	&nwUdpIface,
#endif
#ifdef NWSNMP
	&SnmpDgIface,
#endif
#ifdef NWFORE
	&nwForeIface,
#endif
#ifdef NWBPF
	&BpfDgIface,
#endif
#ifdef NWENI
	&EniDgIface,
#endif
	NULL
};


static NW_CO_IFACE *nwCoIface[] =
{
#ifdef	NWTCP
	&nwTcpIface,
#endif
	NULL
};


#define IfaceCount(ifaces)  (sizeof(ifaces) / sizeof(ifaces[0]) - 1)


static BOOLEAN Parse(char *addr, char *head);
static NW_DG_IFACE *DgFindIface(char *addr);
static NW_CO_IFACE *CoFindIface(char *addr);



BOOLEAN NwInit(void)
{
	BOOLEAN rc = TRUE;

	int i, count;

	MessageConfig(NW_ERROR, "Nw");
	
	count = IfaceCount(nwDgIface);

	for (i = 0; i < count; i++)
		if (nwDgIface[i]->init() == FALSE)
			rc = FALSE;

	count = IfaceCount(nwCoIface);

	for (i = 0; i < count; i++)
		if (nwCoIface[i]->init() == FALSE)
			rc = FALSE;

	return rc;
}

BOOLEAN NwExit(void)
{
	BOOLEAN rc = TRUE;

	int i, count;

	count = IfaceCount(nwDgIface);

	for (i = 0; i < count; i++)
		if (nwDgIface[i]->exit() == FALSE)
			rc = FALSE;

	count = IfaceCount(nwCoIface);

	for (i = 0; i < count; i++)
		if (nwCoIface[i]->exit() == FALSE)
			rc = FALSE;

	return rc;
}

NwStat *NwDgStatistics(char *addr)
{
NW_DG_IFACE *iface;
int i, count;
NwStat *stats;

	if (addr == NULL || strlen(addr) == 0)
		return &nwDgStat;
                        
    if (strcmp(addr, "*") == 0)
    {            
        memset(&tmpDgStat, 0, sizeof(tmpDgStat));

    	count = IfaceCount(nwDgIface);
	    for (i = 0; i < count; i++)         
        {
		    if (nwDgIface[i]->statistics == 0)
                continue;
            stats = nwDgIface[i]->statistics(addr);
            tmpDgStat.InPkts += stats->InPkts;
            tmpDgStat.InOctets += stats->InOctets;         
            tmpDgStat.InDiscards += stats->InDiscards;       
            tmpDgStat.InErrors += stats->InErrors; 
            tmpDgStat.InUnknownProtos += stats->InUnknownProtos;  
            tmpDgStat.OutPkts += stats->OutPkts;  
            tmpDgStat.OutOctets += stats->OutOctets;
            tmpDgStat.OutDiscards += stats->OutDiscards;      
            tmpDgStat.OutErrors += stats->OutErrors;
            tmpDgStat.OutQLen += stats->OutQLen;  
        }
        return &tmpDgStat;
    }
    
	iface = DgFindIface(addr);
	if (iface == 0)
		return 0;

	if (iface->statistics == 0)
		return 0;

	return iface->statistics(addr);
}

NwStat *NwCoStatistics(char *addr)
{
NW_CO_IFACE *iface;
int i, count;
NwStat *stats;

	if (addr == NULL || strlen(addr) == 0)
		return &nwCoStat;

    if (strcmp(addr, "*") == 0)
    {
        memset(&tmpCoStat, 0, sizeof(tmpCoStat));

    	count = IfaceCount(nwCoIface);
	    for (i = 0; i < count; i++)         
        {
		    if (nwCoIface[i]->statistics == 0)
                continue;
            stats = nwCoIface[i]->statistics(addr);
            tmpCoStat.InPkts += stats->InPkts;
            tmpCoStat.InOctets += stats->InOctets;         
            tmpCoStat.InDiscards += stats->InDiscards;       
            tmpCoStat.InErrors += stats->InErrors; 
            tmpCoStat.InUnknownProtos += stats->InUnknownProtos;  
            tmpCoStat.OutPkts += stats->OutPkts;  
            tmpCoStat.OutOctets += stats->OutOctets;
            tmpCoStat.OutDiscards += stats->OutDiscards;      
            tmpCoStat.OutErrors += stats->OutErrors;
            tmpCoStat.OutQLen += stats->OutQLen;  
        }
        return &tmpCoStat;
    }
    
	iface = CoFindIface(addr);
	if (iface == 0)
		return 0;

	if (iface->statistics == 0)
		return 0;

	return iface->statistics(addr);
}

NW_DG *NwDgAccept(char *addr, NW_DG_RECV_CALL recv, void *parm)
{
	NW_DG_IFACE *iface;

	iface = DgFindIface(addr);
	if (iface == 0)
		return 0;

	if (iface->accept == 0)
		return 0;

	return iface->accept(addr, recv, parm);
}

NW_DG *NwDgConnect(char *addr, NW_DG_RECV_CALL recv, void *parm)
{
	NW_DG_IFACE *iface;

	iface = DgFindIface(addr);
	if (iface == 0)
		return 0;

	if (iface->connect == 0)
		return 0;

	return iface->connect(addr, recv, parm);
}

BOOLEAN NwDgSend(NW_DG * dg, BYTE * frame, int length)
{
BOOLEAN result;

	if (dg == NULL)
		return FALSE;

	if (dg->iface->send == 0)
		return FALSE;

	nwDgStat.OutPkts++;
	result = dg->iface->send(dg, frame, length);
	if (result == FALSE)
	{
		nwDgStat.OutDiscards++;
		return FALSE;
	}
	nwDgStat.OutOctets += length;
	return result;
}

BOOLEAN NwDgSendTo(NW_DG * dg, char *addr, BYTE * frame, int length)
{
BOOLEAN result;
NW_DG_IFACE *iface;

	if (dg == 0)
		iface = DgFindIface(addr);
	else
		iface = dg->iface;

	if (iface == 0)
		return FALSE;

	if (iface->sendto == 0)
		return FALSE;

	nwDgStat.OutPkts++;
	result = iface->sendto(dg, addr, frame, length);
	if (result == FALSE)
	{
		nwDgStat.OutDiscards++;
		return FALSE;
	}
	nwDgStat.OutOctets += length;
	return result;
}

BOOLEAN NwDgClose(NW_DG * dg)
{
	if (dg == NULL)
		return TRUE;

	if (dg->iface->close == 0)
		return FALSE;

	return dg->iface->close(dg);
}

BOOLEAN  NwDgIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig)
{
	if (dg == NULL)
		return FALSE;

	if (dg->iface->ifconfig == 0)
		return FALSE;

	return dg->iface->ifconfig(dg, ifconfig);
}

NW_AC *NwCoAccept(char *addr, NW_CO_RECV_CALL recv, NW_CO_CLOSE_CALL close, NW_CO_OPEN_CALL open, void *parm)
{
	NW_CO_IFACE *iface;

	iface = CoFindIface(addr);
	if (iface == 0)
		return 0;

	if (iface->accept == 0)
		return 0;

	return iface->accept(addr, recv, close, open, parm);
}

NW_CO *NwCoConnect(char *addr, NW_CO_RECV_CALL recv, NW_CO_CLOSE_CALL close, void *parm)
{
	NW_CO_IFACE *iface;

	iface = CoFindIface(addr);
	if (iface == 0)
		return 0;

	if (iface->connect == 0)
		return 0;

	return iface->connect(addr, recv, close, parm);
}

BOOLEAN NwCoSend(NW_CO * co, BYTE * frame, int length)
{
BOOLEAN result;

	if (co == NULL)
		return FALSE;
		
	if (co->iface->send == 0)
		return FALSE;

	nwCoStat.OutPkts++;
	result = co->iface->send(co, frame, length);
	if (result == FALSE)
	{
		nwCoStat.OutDiscards++;
		return FALSE;
	}
	nwCoStat.OutOctets += length;
	return result;
}

BOOLEAN NwCoClose(NW_CO * co)
{
	if (co == NULL)
		return FALSE;
		
	if (co->iface->close == 0)
		return FALSE;

	return co->iface->close(co);
}

BOOLEAN NwCoRelease(NW_AC * ac)
{
	if (ac == NULL)
		return FALSE;
		
	if (ac->iface->release == 0)
		return FALSE;

	return ac->iface->release(ac);
}

BOOLEAN  NwCoIfConfig(NW_CO *co, NW_IFCONFIG *ifconfig)
{
	if (co == NULL)
		return FALSE;

	if (co->iface->ifconfig == 0)
		return FALSE;

	return co->iface->ifconfig(co, ifconfig);
}

static BOOLEAN Parse(char *addr, char *head)
{
	while (*addr != ':' && *addr != 0)
		*head++ = *addr++;
	*head = 0;

	return TRUE;
}

static NW_DG_IFACE *DgFindIface(char *addr)
{
	char descr[16];
	int i, count;

	if (!Parse(addr, descr))
		return 0;

	count = sizeof(nwDgIface) / sizeof(NW_DG_IFACE *) - 1;

	for (i = 0; i < count; i++)
	{
		if (strcmp(descr, nwDgIface[i]->descr) == 0)
			return nwDgIface[i];
	}

	return 0;
}

static NW_CO_IFACE *CoFindIface(char *addr)
{
	char descr[16];
	int i, count;

	if (!Parse(addr, descr))
		return 0;

	count = sizeof(nwCoIface) / sizeof(NW_CO_IFACE *) - 1;

	for (i = 0; i < count; i++)
	{
		if (strcmp(descr, nwCoIface[i]->descr) == 0)
			return nwCoIface[i];
	}

	return 0;
}
