/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#ifndef _NW_H_
#define _NW_H_


#include <dnpap.h>


#define MAC_TYPE_OTHER              	1
#define MAC_TYPE_REGULAR_1822       	2
#define MAC_TYPE_HDH_1822           	3
#define MAC_TYPE_DDN_X25            	4
#define MAC_TYPE_RFC877_X25         	5
#define MAC_TYPE_ETHERNET_CSMACD    	6
#define MAC_TYPE_88023_CSMACD       	7
#define MAC_TYPE_88024_TOKENBUS     	8
#define MAC_TYPE_88025_TOKENRING    	9
#define MAC_TYPE_88026_MAN          	10
#define MAC_TYPE_SOFT_LOOPBACK      	24

#define MAC_MTU_ETHERNET_CSMACD			1500
#define MAC_MTU_88023_CSMACD			1500

#define MAC_SPEED_ETHERNET_CSMACD		10000000
#define MAC_SPEED_88023_CSMACD			10000000

#define MAC_ADDRLEN_ETHERNET_CSMACD		6
#define MAC_ADDRLEN_88023_CSMACD		6



typedef struct timeval      TIMEVAL;
typedef struct sockaddr     SOCKADDR;
typedef struct sockaddr_in  SOCKADDR_IN;
typedef struct in_addr      IN_ADDR;
typedef struct strbuf       STRBUF;
typedef struct strioctl     STRIOCTL;
typedef struct ifreq        IFREQ;
typedef struct protoent     PROTOENT;
typedef struct hostent      HOSTENT;
#ifndef LINUX
typedef struct ip           IP;
typedef struct icmp         ICMP;
#else
#define ip_hl    ihl
#define icmp_id  un.echo.id
typedef struct iphdr        IP;
typedef struct icmphdr      ICMP;
#endif


typedef struct NwStat_
{
	ULONG InPkts;
	ULONG InOctets;
    ULONG InDiscards;
    ULONG InErrors;
    ULONG InUnknownProtos;
    ULONG OutPkts;
    ULONG OutOctets;
    ULONG OutDiscards;
    ULONG OutErrors;
    ULONG OutQLen;

} NwStat;


typedef struct _NW_DG NW_DG;
typedef struct _NW_CO NW_CO; 
typedef struct _NW_AC NW_AC; 
typedef struct _NW_DG_IFACE NW_DG_IFACE;
typedef struct _NW_CO_IFACE NW_CO_IFACE;
typedef struct _NW_IFCONFIG NW_IFCONFIG;


typedef void  (*NW_DG_RECV_CALL)    (NW_DG *dg, char *addr, BYTE *frame, int length, void *parm);

typedef void  (*NW_CO_RECV_CALL)    (NW_CO *co, BYTE *frame, int length, void *parm);
typedef void* (*NW_CO_OPEN_CALL)    (NW_CO *co, char *addr, void *parm);
typedef void  (*NW_CO_CLOSE_CALL)   (NW_CO *co, void *parm);

BOOLEAN  NwInit      (void); 
BOOLEAN  NwExit      (void);

NwStat  *NwDgStatistics(char *addr);
NwStat  *NwCoStatistics(char *addr);

NW_DG   *NwDgAccept  (char *addr, NW_DG_RECV_CALL rcv, void *parm);
NW_DG   *NwDgConnect (char *addr, NW_DG_RECV_CALL rcv, void *parm);
BOOLEAN  NwDgSend    (NW_DG *dg, BYTE *frame, int length);
BOOLEAN  NwDgSendTo  (NW_DG *dg, char *addr, BYTE *frame, int length);
BOOLEAN  NwDgClose   (NW_DG *dg);
BOOLEAN  NwDgIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig);
 

NW_AC   *NwCoAccept  (char *addr, NW_CO_RECV_CALL rcv, NW_CO_CLOSE_CALL clse, NW_CO_OPEN_CALL opn, void *parm);
NW_CO   *NwCoConnect (char *addr, NW_CO_RECV_CALL rcv, NW_CO_CLOSE_CALL clse, void *parm);
BOOLEAN  NwCoSend    (NW_CO *co, BYTE *frame, int length);
BOOLEAN  NwCoClose   (NW_CO *co);
BOOLEAN  NwCoRelease (NW_AC *ac);
BOOLEAN  NwCoIfConfig(NW_CO *co, NW_IFCONFIG *ifconfig);



struct _NW_DG_IFACE
{
    char   *descr;
    BOOLEAN (*init)      (void); 
    BOOLEAN (*exit)      (void); 
    NW_DG  *(*accept)    (char *addr, NW_DG_RECV_CALL rcv, void *parm); 
    NW_DG  *(*connect)   (char *addr, NW_DG_RECV_CALL rcv, void *parm);
    BOOLEAN (*send)      (NW_DG *dg, BYTE *frame, int length);  
    BOOLEAN (*sendto)    (NW_DG *dg, char *addr, BYTE *frame, int length);  
    BOOLEAN (*close)     (NW_DG *dg);
	BOOLEAN (*ifconfig)  (NW_DG *dg, NW_IFCONFIG *ifconfig);
	NwStat *(*statistics)(char *addr);
};


struct _NW_CO_IFACE
{
    char    *descr;
    BOOLEAN  (*init)    (void); 
    BOOLEAN  (*exit)    (void); 
    NW_AC   *(*accept)  (char *addr, NW_CO_RECV_CALL rcv, NW_CO_CLOSE_CALL clse, NW_CO_OPEN_CALL opn, void *parm);
    NW_CO   *(*connect) (char *addr, NW_CO_RECV_CALL rcv, NW_CO_CLOSE_CALL clse, void *parm);
    BOOLEAN  (*send)    (NW_CO *co, BYTE *frame, int length);
    BOOLEAN  (*close)   (NW_CO *co);
    BOOLEAN  (*release) (NW_AC *ac);
	BOOLEAN  (*ifconfig)(NW_CO *co, NW_IFCONFIG *ifconfig);
	NwStat  *(*statistics)(char *addr);
};


struct _NW_DG
{
    NW_DG_IFACE         *iface;   
    char                addr[256];
    NW_DG_RECV_CALL     recv;
    void                *parm;
};


struct _NW_CO
{            
    NW_CO_IFACE         *iface;
    char                addr[256];
    NW_CO_RECV_CALL     recv; 
    NW_CO_CLOSE_CALL    close;
    void                *parm;
};


struct _NW_AC
{            
    NW_CO_IFACE         *iface;
    char                addr[256];
    NW_CO_RECV_CALL     recv; 
    NW_CO_CLOSE_CALL    close;
    NW_CO_OPEN_CALL     open;
    void                *parm;  
};


struct _NW_IFCONFIG
{
	INT	type;
	INT mtu;
	INT speed;
	BYTE address[16];
	BYTE broadcast[16];
	WORD addrlen;
};


INT IpCheckSum(USHORT *ptr, INT nbytes);


#endif
