/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <limits.h>
#include <memory.h>
#include <string.h>
#include <stropts.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <acp.h>

#include <dnpap.h>
#include <block.h>
#include <config.h>
#include <message.h>

#include "nw.h"
#include "nweni.h"


/* nweni.c - NwEni module functions */


typedef struct _NwEni
{
	INT		Socket;
	ATM_CONNECTION Params;

} NwEni;


typedef struct _NwEniDg
{
    NW_DG	NwInfo;
    NwEni	EniInfo;

    BLOCK	Block;

    struct _NwEniDg *Prev;
    struct _NwEniDg *Next;

} NwEniDg;



static BOOLEAN NwEniInit(VOID);
static BOOLEAN NwEniExit(VOID);

static NW_DG  *NwEniAccept(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static NW_DG  *NwEniConnect(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static BOOLEAN NwEniSend(NW_DG *dg, BYTE *frame, INT len);
static BOOLEAN NwEniSendTo(NW_DG *dg, char *addr, BYTE *frame, INT len);
static BOOLEAN NwEniClose(NW_DG *dg);
static BOOLEAN NwEniIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig);
static NwStat *NwEniStatistics(char *addr);


NW_DG_IFACE EniDgIface = { "eni",
                           NwEniInit,   NwEniExit,
                           NwEniAccept, NwEniConnect,
                           NwEniSend,   NwEniSendTo,
                           NwEniClose,  NwEniIfConfig,
                           NwEniStatistics
                        };


static NwEniDg *NwEniDgList = NULL;

static NwStat nwStat;


static BOOLEAN NwEniCloseDg(NwEniDg *dg);

static BOOLEAN NwEniDgAddList(NwEniDg *dg);
static BOOLEAN NwEniDgRemoveList(NwEniDg *dg);

static NwEniDg* NewNwEniDg(VOID);
static VOID DelNwEniDg(NwEniDg *dg);

static BOOLEAN NwEniRead(NwEniDg *dg);

static int NwEniBlockCallback(BLOCK *blk, int reason, VOID *param);

static BOOLEAN Addr2Conn(CHAR *addr, ATM_CONNECTION *conn);


/* Initializes the NwEni module
 * Initializes the NwEni module
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwEniInit(VOID)
{
    return TRUE;
}


/* Exits the NwEni module
 * Closes all opened sockets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwEniExit(VOID)
{
BOOLEAN rc = TRUE;
NwEniDg *dg;

    for (dg = NwEniDgList; dg != NULL; dg = NwEniDgList)
        rc &= NwEniClose((NW_DG *)dg);

    return rc;
}


NwStat *NwEniStatistics(char *addr)
{
    return &nwStat;
}


/* Opens a socket for receiving packets
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG *NwEniAccept(char *addr				/* accept specification: 'eni:...' */,
                   NW_DG_RECV_CALL recv		/* receive callback */,
                   VOID *param				/* extra parameter you want to be forwarded to recv() */)
{
	return NwEniConnect(addr, recv, param);
}


/* Opens a socket for receiving and sending packets from and to a specific host
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG  *NwEniConnect(char *addr				/* connect specification: 'eni:...' */,
                     NW_DG_RECV_CALL recv	/* receive callback */,
                     VOID *param			/* extra parameter you want to forward to recv() */)
{
NwEniDg *dg;
int status;
ATM_CONNECTION conn;

    if (Addr2Conn(addr, &conn) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWENI_PARAM, "nweni: invalid parameter list");
        return NULL;
    }

	if ((dg = NewNwEniDg()) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWENI_NEWENI, "nweni: memory allocation failed");
		return NULL;
    }

	dg->EniInfo.Params = conn;

    if ((status = acp_pvc_open(&dg->EniInfo.Params, &dg->EniInfo.Socket, O_RDWR)) != ACP_RC_SUCCESSFUL)
    {
        DnpapMessage(DMC_ERROR, NWENI_OPEN, "nweni: open failed, %s", acp_error(status));
    	NwEniCloseDg(dg);
        return NULL;
    }

	if (ioctl(dg->EniInfo.Socket, I_SRDOPT, RMSGN) < 0)
	{
        DnpapMessage(DMC_ERROR, NWENI_OPEN, "nweni: can not set discrete message option");
    	NwEniCloseDg(dg);
        return NULL;
	}

    dg->Block.sckt = dg->EniInfo.Socket;

    if (BlockRegister(&dg->Block, dg, NwEniBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWENI_BLOCK, "nweni: block registration failed");
    	NwEniCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &EniDgIface;
    sprintf(dg->NwInfo.addr, 
		"eni:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d",
		dg->EniInfo.Params.id.port_index,
		dg->EniInfo.Params.id.vpi,
		dg->EniInfo.Params.id.vci,
		dg->EniInfo.Params.virtual_connection_type,
		dg->EniInfo.Params.pvc_svc_type,
		dg->EniInfo.Params.aal_type,
		dg->EniInfo.Params.aal_mode,
		dg->EniInfo.Params.tx_max_cpcs_sdu_size,
		dg->EniInfo.Params.rx_max_cpcs_sdu_size,
		dg->EniInfo.Params.deliver_corrupted_data,
		dg->EniInfo.Params.tx_requested_rate.td_type,
		dg->EniInfo.Params.tx_requested_rate.parms.noclp_noscr.clp01_pcr,
		dg->EniInfo.Params.tx_requested_rate.be_indicator,
		dg->EniInfo.Params.tx_acceptable_rate.td_type,
		dg->EniInfo.Params.tx_acceptable_rate.parms.noclp_noscr.clp01_pcr,
		dg->EniInfo.Params.tx_acceptable_rate.be_indicator,
		dg->EniInfo.Params.rx_requested_rate.td_type,
		dg->EniInfo.Params.rx_requested_rate.parms.noclp_noscr.clp01_pcr,
		dg->EniInfo.Params.rx_requested_rate.be_indicator,
		dg->EniInfo.Params.rx_acceptable_rate.td_type,
		dg->EniInfo.Params.rx_acceptable_rate.parms.noclp_noscr.clp01_pcr,
		dg->EniInfo.Params.rx_acceptable_rate.be_indicator
	);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;

    if (NwEniDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWENI_ADDLIST, "nweni: could not add to list");
    	NwEniCloseDg(dg);
		return NULL;
	}

    return (NW_DG *)dg;
}


/* Closes a NwEni datagram socket
 * Closes the specified NwEni datagram socket
 * Returns TRUE for success and FALSE for failure
 */
BOOLEAN NwEniClose(NW_DG *dg   /* an existing NwEni datagram socket */)
{
NwEniDg *enidg = (NwEniDg *)dg;

    NwEniDgRemoveList(enidg);
    NwEniCloseDg(enidg);

    return TRUE;
}


BOOLEAN NwEniCloseDg(NwEniDg *dg)
{
    if (dg->EniInfo.Socket != 0)
	{
    	BlockDeRegister(&dg->Block);
        close(dg->EniInfo.Socket);
	}

    DelNwEniDg(dg);

    return TRUE;
}


/* Sends a packet over an existing NW datagram
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwEniSend(NW_DG *dg    /* valid NW datagram */,
                  BYTE *frame  /* packet */,
                  INT len      /* packet length */)
{
NwEniDg *enidg = (NwEniDg *) dg;

    if (len > NWENI_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWENI_SEND, "nweni: packet too long");
        return FALSE;
    }

    if (write(enidg->EniInfo.Socket, frame, len) < 0)
    {
        DnpapMessage(DMC_ERROR, NWENI_SEND, "nweni: write failed");
        return FALSE;
    }

    return TRUE;
}


/* Sends a packet to a specified address
 * This function can be used to sent one-shot only packets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwEniSendTo(NW_DG *dg    /* not supported, set to NULL */,
                    char *addr   /* address specification: 'eni:...' */,
                    BYTE *frame  /* packet */,
                    INT len      /* packet length */)
{
    if (dg != NULL)
        DnpapMessage(DMC_WARNING, NWENI_SEND, "nweni: sendto over existing connection has no effect");

    if (len > NWENI_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWENI_SEND, "nweni: packet too long");
        return FALSE;
    }

    return FALSE;				/* sendto() is not implemented yet */
}


BOOLEAN NwEniIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig)
{
NwEniDg *enidg = (NwEniDg *) dg;

	return TRUE;
}


int NwEniBlockCallback(BLOCK *blk, int reason, VOID *param)
{
NwEniDg *dg = (NwEniDg *) param;

	switch (reason)
	{
	case BLOCKEVENT_READ:
		if (NwEniRead(dg) != TRUE)
			DnpapMessage(DMC_ERROR, NWENI_READ, "nweni: read failed");
		break;
	case BLOCKEVENT_WRITE:
		DnpapMessage(DMC_ERROR, NWENI_WEVENT, "nweni: write event");
		break;
	case BLOCKEVENT_EXCEPT:
		DnpapMessage(DMC_ERROR, NWENI_EEVENT, "nweni: except event");
		break;
	default:
		DnpapMessage(DMC_ERROR, NWENI_NEVENT, "nweni: unknown block event");
		return 1;
	}

	return 0;
}


BOOLEAN NwEniRead(NwEniDg *dg)
{
BYTE buf[NWENI_SIZE_PACKET];
int len;

    if ((len = read(dg->EniInfo.Socket, buf, sizeof(buf))) < 0)
    	return FALSE;

	if (dg->NwInfo.recv != NULL)
		dg->NwInfo.recv((NW_DG *)dg, dg->NwInfo.addr, buf, len, dg->NwInfo.parm);

	return TRUE;
}


BOOLEAN NwEniDgAddList(NwEniDg *dg)
{
    if (NwEniDgList != NULL)
		NwEniDgList->Prev = dg;

    dg->Next = NwEniDgList;
    dg->Prev = NULL;

    NwEniDgList = dg;

    return TRUE;
}


BOOLEAN NwEniDgRemoveList(NwEniDg *dg)
{
    if (dg->Prev != NULL)
        dg->Prev->Next = dg->Next;
    else
        NwEniDgList = dg->Next;

    if (dg->Next != NULL)
        dg->Next->Prev = dg->Prev;

    return TRUE;
}


NwEniDg* NewNwEniDg(VOID)
{
NwEniDg *dg;

	if ((dg = DnpapMalloc(sizeof(NwEniDg))) == NULL)
		return NULL;

    dg->NwInfo.iface = NULL;
	dg->NwInfo.recv = NULL;
    dg->NwInfo.parm = NULL;

	dg->EniInfo.Socket = 0;

    dg->Next = NULL;
    dg->Prev = NULL;

	return dg;
}


VOID DelNwEniDg(NwEniDg *dg)
{
    DnpapFree(dg);
}


BOOLEAN Addr2Conn(CHAR *addr, ATM_CONNECTION *conn)
{
CHAR *p, buf[NWENI_SIZE_ENTRY];
int i;

	strcpy(buf, addr);

	memset(conn, 0, sizeof(*conn));

	for (i = 0, p = strtok(buf, ":"); p != NULL; i++, p = strtok(NULL, ":"))
	{
		switch (i)
		{
		case 0:					/* assume it starts with "eni:" */
			break;
		case 1:
            conn->id.port_index = atoi(p);
			break;
		case 2:
            conn->id.vpi = atoi(p);
			break;
		case 3:
            conn->id.vci = atoi(p);
			break;
		case 4:
            conn->virtual_connection_type = atoi(p);
			break;
		case 5:
            conn->pvc_svc_type = atoi(p);
			break;
		case 6:
            conn->aal_type = atoi(p);
			break;
		case 7:
            conn->aal_mode = atoi(p);
			break;
		case 8:
            conn->tx_max_cpcs_sdu_size = atoi(p);
			break;
		case 9:
            conn->rx_max_cpcs_sdu_size = atoi(p);
			break;
		case 10:
            conn->deliver_corrupted_data = atoi(p);
			break;
		case 11:
            conn->tx_requested_rate.td_type = atoi(p);
			break;
		case 12:
            conn->tx_requested_rate.parms.noclp_noscr.clp01_pcr = atoi(p);
			break;
		case 13:
            conn->tx_requested_rate.be_indicator = atoi(p);
			break;
		case 14:
            conn->tx_acceptable_rate.td_type = atoi(p);
			break;
		case 15:
            conn->tx_acceptable_rate.parms.noclp_noscr.clp01_pcr = atoi(p);
			break;
		case 16:
            conn->tx_acceptable_rate.be_indicator = atoi(p);
			break;
		case 17:
            conn->rx_requested_rate.td_type = atoi(p);
			break;
		case 18:
            conn->rx_requested_rate.parms.noclp_noscr.clp01_pcr = atoi(p);
			break;
		case 19:
            conn->rx_requested_rate.be_indicator = atoi(p);
			break;
		case 20:
            conn->rx_acceptable_rate.td_type = atoi(p);
			break;
		case 21:
            conn->rx_acceptable_rate.parms.noclp_noscr.clp01_pcr = atoi(p);
			break;
		case 22:
            conn->rx_acceptable_rate.be_indicator = atoi(p);
			break;
		case 23:				/* too many parameters */
			return FALSE;
		}
	}

	if (i < 23)					/* too few parameters */
		return FALSE;

	return TRUE;
}
