/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <block.h>

#ifndef BSD_SELECT
#define BSD_SELECT
#endif

#ifdef OS2
#include <types.h>
#include <netlib.h>
#include <sys/select.h>					/* contains FD_*** macros */
#else
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#endif
#include <netdb.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <dinet.h>
#if defined(SUNOS) || defined(SOLARIS)
#include <stropts.h>
#endif
#include <errno.h>
#include <fore_atm/fore_atm_user.h>

#include "nw.h"
#include "nwfore.h"

typedef struct _NwFore
{
	CHAR dev[NW_FORE_DEVICE_SIZE];
	int socket;
	Atm_info info;
	Vpvc vpvc;
	Aal_type aal;
	Atm_conn_resource qos;
} NwFore;

typedef struct _NW_FORE_DG NW_FORE_DG;

struct _NW_FORE_DG
{
	NW_DG  NwInfo;
	NwFore ForeInfo;

	BLOCK block;

	NW_FORE_DG *next;
};

static NwStat nwStat;

static BOOLEAN Init(void);
static BOOLEAN Exit(void);
static NW_DG *Accept(char *addr, NW_DG_RECV_CALL rcv, void *parm);
static NW_DG *Connect(char *addr, NW_DG_RECV_CALL rcv, void *parm);
static BOOLEAN Close(NW_DG * dg);
static BOOLEAN Send(NW_DG * dg, BYTE * frame, int length);
static BOOLEAN SendTo(NW_DG * dg, char *addr, BYTE * frame, int length);
static BOOLEAN IfConfig(NW_DG * dg, NW_IFCONFIG * ifconfig);
static NwStat *Statistics(char *addr);
static int RecvCall(BLOCK * block, int reason, void *param);
static BOOLEAN Addr2Sock(char *addr, char *dev, Vpvc *vpvc, Aal_type *aal, Atm_conn_resource *qos);
static BOOLEAN Sock2Addr(char *addr, NW_FORE_DG *dg, Vpvc vpvc);

NW_DG_IFACE nwForeIface =
{
	"fore",
	Init,
	Exit,
	Accept,
	Connect,
	Send,
	SendTo,
	Close,
	IfConfig,
	Statistics
};

static NW_FORE_DG *nwForeDgList = 0;
static int nwForeSocket = 0;

static BOOLEAN init = FALSE;

BOOLEAN Init(void)
{
	if (init)
		return TRUE;

	init = TRUE;

	return init;
}

BOOLEAN Exit(void)
{
	NW_FORE_DG *atm, *p;

	p = nwForeDgList;

	while (p != 0)
	{
		atm = p;
		p = p->next;

		BlockDeRegister(&atm->block);
		atm_close(atm->ForeInfo.socket);

		DnpapFree(atm);
	}

	return TRUE;
}

NwStat *Statistics(char *addr)
{
	return &nwStat;
}

NW_DG *Accept(char *addr, NW_DG_RECV_CALL rcv, void *parm)
{
	NW_FORE_DG *atm;

	if ((atm = (NW_FORE_DG *) DnpapMalloc(sizeof(NW_FORE_DG))) == 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: accept: malloc failed");
		return 0;
	}

	if (Addr2Sock(addr, atm->ForeInfo.dev, &atm->ForeInfo.vpvc, &atm->ForeInfo.aal, &atm->ForeInfo.qos) == FALSE)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: accept: address translation failed");
		return NULL;
	}

	if ((atm->ForeInfo.socket = atm_open(atm->ForeInfo.dev, O_RDWR, &atm->ForeInfo.info)) < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: accept: socket failed");
		return 0;
	}

	if (atm_bind_pvc(atm->ForeInfo.socket, atm->ForeInfo.vpvc, atm->ForeInfo.aal, &atm->ForeInfo.qos) < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: accept: can not bind server socket");
		return 0;
	}

	atm->block.sckt = atm->ForeInfo.socket;
	if (BlockRegister(&(atm->block), atm, RecvCall) != 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: accept: failed to register server block");
		return 0;
	}

	atm->NwInfo.iface = &nwForeIface;
	sprintf(atm->NwInfo.addr, "fore:%s:%d:%d:%d:%d", atm->ForeInfo.dev, atm->ForeInfo.vpvc, atm->ForeInfo.aal, atm->ForeInfo.qos.peak_bandwidth, atm->ForeInfo.qos.mean_burst);
	atm->NwInfo.recv = rcv;
	atm->NwInfo.parm = parm;

	atm->next = nwForeDgList;
	nwForeDgList = atm;

	return (NW_DG *) atm;
}

NW_DG *Connect(char *addr, NW_DG_RECV_CALL rcv, void *parm)
{
	NW_FORE_DG *atm;

	if ((atm = (NW_FORE_DG *) DnpapMalloc(sizeof(NW_FORE_DG))) == 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: connect: malloc failed");
		return 0;
	}

	if (Addr2Sock(addr, atm->ForeInfo.dev, &atm->ForeInfo.vpvc, &atm->ForeInfo.aal, &atm->ForeInfo.qos) == FALSE)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: connect: address translation failed");
		return NULL;
	}

	if ((atm->ForeInfo.socket = atm_open(atm->ForeInfo.dev, O_RDWR, &atm->ForeInfo.info)) < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: connect: socket failed");
		return 0;
	}

	if (atm_connect_pvc(atm->ForeInfo.socket, atm->ForeInfo.vpvc, atm->ForeInfo.aal, &atm->ForeInfo.qos) < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: connect: can not bind server socket");
		return 0;
	}

	atm->block.sckt = atm->ForeInfo.socket;
	if (BlockRegister(&(atm->block), atm, RecvCall) != 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: connect: failed to register server block");
		return 0;
	}

	atm->NwInfo.iface = &nwForeIface;
	sprintf(atm->NwInfo.addr, "fore:%s:%d:%d:%d:%d", atm->ForeInfo.dev, atm->ForeInfo.vpvc, atm->ForeInfo.aal, atm->ForeInfo.qos.peak_bandwidth, atm->ForeInfo.qos.mean_burst);
	atm->NwInfo.recv = rcv;
	atm->NwInfo.parm = parm;

	atm->next = nwForeDgList;
	nwForeDgList = atm;

	return (NW_DG *) atm;
}

BOOLEAN Close(NW_DG * dg)
{
	NW_FORE_DG **p, *atm;

	atm = (NW_FORE_DG *) dg;

	for (p = &nwForeDgList; *p != 0 && *p != atm; p = &(*p)->next) ;

	if (*p == 0)
		return FALSE;

	*p = atm->next;

	BlockDeRegister(&atm->block);
	atm_close(atm->ForeInfo.socket);

	DnpapFree(atm);

	return TRUE;
}

BOOLEAN Send(NW_DG * dg, BYTE * frame, int length)
{
	int n;
	NW_FORE_DG *atm = (NW_FORE_DG *) dg;

	if (atm == NULL)
		return FALSE;

	n = atm_send(atm->ForeInfo.socket, frame, length);

	if (n < 0 || n != length)
		return FALSE;

	return TRUE;
}

BOOLEAN SendTo(NW_DG * dg, char *addr, BYTE * frame, int length)
{
	int n;
	NW_FORE_DG *atm = (NW_FORE_DG *) dg, tmp;

	if (atm != NULL)
	{
		n = atm_sendto(atm->ForeInfo.socket, frame, length, atm->ForeInfo.vpvc);
	}
	else
	{
		if (Addr2Sock(addr, tmp.ForeInfo.dev, &tmp.ForeInfo.vpvc, &tmp.ForeInfo.aal, &tmp.ForeInfo.qos) == FALSE)
		{
			DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: sendto: address translation failed");
			return FALSE;
		}
		if ((tmp.ForeInfo.socket = atm_open(tmp.ForeInfo.dev, O_RDWR, &tmp.ForeInfo.info)) < 0)
		{
			DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: sendto: socket failed");
			return FALSE;
		}
		if (atm_connect_pvc(tmp.ForeInfo.socket, tmp.ForeInfo.vpvc, tmp.ForeInfo.aal, &tmp.ForeInfo.qos) < 0)
		{
			DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: sendto: can not bind server socket");
			return 0;
		}
		n = atm_sendto(tmp.ForeInfo.socket, frame, length, tmp.ForeInfo.vpvc);
		atm_close(tmp.ForeInfo.socket);
	}

	if (n < 0 || n != length)
		return FALSE;

	return TRUE;
}

BOOLEAN IfConfig(NW_DG * dg, NW_IFCONFIG * ifconfig)
{
	return FALSE;
}

int RecvCall(BLOCK * block, int reason, void *parm)
{
	NW_FORE_DG *atm;
	BYTE buffer[NW_FORE_BUFFER_SIZE];
	int n;
	Vpvc vpvc;

	atm = (NW_FORE_DG *) parm;
	switch (reason)
	{
	case BLOCKEVENT_READ:
		n = atm_recvfrom(atm->ForeInfo.socket, buffer, sizeof(buffer), &vpvc);
		if (n < 0)
		{
			DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: recvcall: recvfrom error");
			return -1;
		}
		atm->NwInfo.recv((NW_DG *) atm, atm->NwInfo.addr, buffer, n, atm->NwInfo.parm);
		break;
	case BLOCKEVENT_WRITE:
		/* ignore these events */
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: recvcall: unexpected BLOCKEVENT_WRITE");
		break;
	case BLOCKEVENT_EXCEPT:
		/* ignore these events */
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: recvcall: unexpected BLOCKEVENT_EXCEPT");
		break;
	default:
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwfore: recvcall: unknown block event");
		break;
	}
	return 0;
}


BOOLEAN Addr2Sock(char *addr, char *dev, Vpvc *vpvc, Aal_type *aal, Atm_conn_resource *qos)
{
char *p, *q;
int bandwidth, burst;

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NW_ERROR, "nwfore: invalid parameter list");
        return FALSE;
    }   

	strcpy(dev, "/dev/");
    strncat(dev, addr, p-addr);
    dev[5+p-addr] = '\0';

	if ((q = strchr(++p, ':')) == NULL)
	{
        DnpapMessage(DMC_ERROR, NW_ERROR, "nwfore: invalid parameter list");
        return FALSE;
	}
	*vpvc = atoi(p);
	if ((p = strchr(++q, ':')) == NULL)
	{
        DnpapMessage(DMC_ERROR, NW_ERROR, "nwfore: invalid parameter list");
        return FALSE;
	}
	*aal = atoi(q);
	if ((q = strchr(++p, ':')) == NULL)
	{
        DnpapMessage(DMC_ERROR, NW_ERROR, "nwfore: invalid parameter list");
        return FALSE;
	}
	bandwidth = atoi(p);
	burst = atoi(++q);

	qos->peak_bandwidth = bandwidth;
	qos->mean_bandwidth = bandwidth/2;
	qos->mean_burst = burst;

	return TRUE;
}


BOOLEAN Sock2Addr(char *addr, NW_FORE_DG *dg, Vpvc vpvc)
{
	NW_FORE_DG *atm = (NW_FORE_DG *) dg;

	sprintf(addr, "fore:%s:%d:%d:%d:%d", atm->ForeInfo.dev, vpvc, atm->ForeInfo.aal, atm->ForeInfo.qos.peak_bandwidth, atm->ForeInfo.qos.mean_burst);

	return TRUE;
}
