/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <limits.h>
#include <memory.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>
#include <netinet/ip_var.h>
#include <netdb.h>
               
#include <dnpap.h>
#include <block.h>
#include <config.h>
#include <message.h>

#include "nw.h"
#include "nwpf.h"
 
            
/* nwpf.c - NwPf module functions */

            
typedef struct _NwPf
{
	INT Socket;
	CHAR *Interface;
    BYTE  Address[NWPF_SIZE_ADDRESS];
	INT BufSize;
	INT SnapLen;
	TIMEVAL Timeout;
	INT Flags;
	BYTE *Buf;
	LONG  Pos;
	LONG  End;
	         
} NwPf;
    

typedef struct _NwPfDg
{          
    NW_DG  NwInfo;
    NwPf   PfInfo;
    
    BLOCK          *Block;
    
    struct _NwPfDg *Prev;
    struct _NwPfDg *Next;
    
} NwPfDg;

      
static BOOLEAN NwPfInit(VOID);
static BOOLEAN NwPfExit(VOID);

static NW_DG  *NwPfAccept(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static NW_DG  *NwPfConnect(char *addr, NW_DG_RECV_CALL recv, VOID *param);
static BOOLEAN NwPfSend(NW_DG *dg, BYTE *frame, INT len);
static BOOLEAN NwPfSendTo(NW_DG *dg, char *addr, BYTE *frame, INT len);
static BOOLEAN NwPfClose(NW_DG *dg);
static BOOLEAN NwPfIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig);
static NwStat *NwPfStatistics(char *addr);


NW_DG_IFACE PfDgIface = { "pf", 
                          NwPfInit,   NwPfExit, 
                          NwPfAccept, NwPfConnect, 
                          NwPfSend,   NwPfSendTo,
                          NwPfClose,  NwPfIfConfig,
                          NwPfStatistics
                        };


static NwPfDg *NwPfDgList = NULL;

static NwStat nwStat;


static INT pfsocket;

static BOOLEAN NwPfCloseDg(NwPfDg *dg);

static BOOLEAN NwPfDgAddList(NwPfDg *dg);
static BOOLEAN NwPfDgRemoveList(NwPfDg *dg);

static NwPfDg* NewNwPfDg(CHAR *interface, INT bufsize);
static VOID DelNwPfDg(NwPfDg *dg);

static BOOLEAN NwPfRead(NwPfDg *dg);

static BOOLEAN NwPfSetup(NwPfDg *dg, CHAR *interface, INT flags, INT bufsize, INT snaplen, TIMEVAL *timeout);

static int NwPfBlockCallback(BLOCK *blk, int reason, VOID *param);



/* Initializes the NwPf module
 * Initializes the NwPf messages and opens a socket for general use
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwPfInit(VOID)
{       
    if ((pfsocket = pfopen(NWPFDEVICE, O_WRONLY)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWPF_OPEN, "nwpf: open failed");
        return FALSE;
    }

    return TRUE;
}


/* Exits the NwPf module
 * Closes all opened sockets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwPfExit(VOID)
{             
BOOLEAN rc = TRUE;
NwPfDg *dg;

    for (dg = NwPfDgList; dg != NULL; dg = NwPfDgList)
        rc &= NwPfClose((NW_DG *)dg);

    if (pfsocket != 0)
        close(pfsocket);
                
    return rc;
}


NwStat *NwPfStatistics(char *addr)
{            
    return &nwStat;
}

        
/* Opens a socket for receiving promiscuous packets
 * Opens the specified interface with capacity 'bufsize' and snapshot
 * length 'snaplen'. 'bufsize' space is automatically allocated.
 * Set 'snaplen' to 0 for no snapshot at all
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG *NwPfAccept(char *addr            /* accept specification: 'pf:(char*)interface:(ulong)bufsize:(ulong)snaplen' */,
                  NW_DG_RECV_CALL recv  /* receive callback */,                                                         
                  VOID *param           /* extra parameter you want to be forwarded to recv() */)                       
{   
NwPfDg *dg;
CHAR interface[NWPF_SIZE_INTERFACE];
INT bufsize, snaplen;
INT flags;
TIMEVAL timeout;
CHAR *p;

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWPF_PARAM, "nwpf: invalid parameter list");
        return NULL;
    }   
    
    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
	if (sscanf(p, ":%d:%d", &bufsize, &snaplen) != 2)
	{
        DnpapMessage(DMC_ERROR, NWPF_PARAM, "nwpf: invalid parameters");
	    return NULL;
	}
	                           
	flags = ENPROMISC | NWPF_DEFAULT_FLAGS;
	
    /*  set buffer timeout at 1/10 of a second  */
	timeout.tv_sec = 0;
	timeout.tv_usec = 100000UL;

	if ((dg = NewNwPfDg(interface, bufsize)) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWPF_NEWPF, "nwpf: memory allocation failed");
		return NULL;
    }

    if ((dg->PfInfo.Socket = pfopen(interface, O_RDONLY)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWPF_OPEN, "nwpf: open failed");
    	NwPfCloseDg(dg);
        return NULL;
    }

    if (NwPfSetup(dg, interface, flags, bufsize, snaplen, &timeout) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWPF_SETUP, "nwpf: setup failed");
    	NwPfCloseDg(dg);
        return NULL;
    }
    
    dg->Block->sckt = dg->PfInfo.Socket;
    
    if (BlockRegister(dg->Block, dg, NwPfBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWPF_BLOCK, "nwpf: block registration failed");
    	NwPfCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &PfDgIface;                                
    sprintf(dg->NwInfo.addr, "pf:%s:%d:%d", interface, bufsize, snaplen);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;    
    
	strcpy(dg->PfInfo.Interface, interface);
	dg->PfInfo.BufSize = bufsize;
	dg->PfInfo.SnapLen = snaplen;
	dg->PfInfo.Timeout.tv_sec = timeout.tv_sec;
	dg->PfInfo.Timeout.tv_usec = timeout.tv_usec;
	dg->PfInfo.Flags = flags;
	dg->PfInfo.Pos = 0;
	dg->PfInfo.End = 0;
             
    if (NwPfDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWPF_ADDLIST, "nwpf: could not add to list");
    	NwPfCloseDg(dg);
		return NULL;
	}
    
    if (ioctl(dg->PfInfo.Socket, EIOCFLUSH, 0) < 0)
	{
        DnpapMessage(DMC_ERROR, NWPF_IOCTL, "nwpf: read flush failed");
    	NwPfCloseDg(dg);
		return NULL;
	}
    
    return (NW_DG *)dg;
}


/* Opens a socket for receiving and sending packets from and to a specific host
 * Opens the specified interface with capacity 'bufsize' and snapshot
 * length 'snaplen'. 'bufsize' space is automatically allocated.
 * Set 'snaplen' to 0 for no snapshot at all. The Ethernet address is recorded 
 * and the receive callback is only called for packets to or from this address
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG  *NwPfConnect(char *addr            /* connect specification: 'pf:(char*)interface:(ulong)bufsize:(ulong)snaplen:(int[6])ethernet' */, 
                    NW_DG_RECV_CALL recv  /* receive callback */, 
                    VOID *param           /* extra parameter you want to forward to recv() */)
{   
NwPfDg *dg;
CHAR interface[NWPF_SIZE_INTERFACE];
INT bufsize, snaplen;
INT flags;
TIMEVAL timeout;
CHAR *p;
UINT a1, a2, a3, a4, a5, a6;

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWPF_PARAM, "nwpf: invalid parameter list");
        return NULL;
    }   
    
    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
	if (sscanf(p, ":%d:%d:%u.%u.%u.%u.%u.%u", &bufsize, &snaplen, &a1, &a2, &a3, &a4, &a5, &a6) != 8)
	{
        DnpapMessage(DMC_ERROR, NWPF_PARAM, "nwpf: invalid parameters");
	    return NULL;
	}
	
	flags = NWPF_DEFAULT_FLAGS;
	
    /*  set buffer timeout at 1/10 of a second  */
	timeout.tv_sec = 0;
	timeout.tv_usec = 100000UL;

	if ((dg = NewNwPfDg(interface, bufsize)) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWPF_NEWPF, "nwpf: memory allocation failed");
		return NULL;
    }

    if ((dg->PfInfo.Socket = open(NWPFDEVICE, O_RDWR)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWPF_OPEN, "nwpf: open failed");
    	NwPfCloseDg(dg);
        return NULL;
    }

    if (NwPfSetup(dg, interface, flags, bufsize, snaplen, &timeout) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWPF_SETUP, "nwpf: setup failed");
    	NwPfCloseDg(dg);
        return NULL;
    }
    
    dg->Block->sckt = dg->PfInfo.Socket;
    
    if (BlockRegister(dg->Block, dg, NwPfBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWPF_BLOCK, "nwpf: block registration failed");
    	NwPfCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &PfDgIface;                                
    sprintf(dg->NwInfo.addr, "pf:%s:%d:%d:%u.%u.%u.%u.%u.%u", 
                             interface, bufsize, snaplen, a1, a2, a3, a4, a5, a6);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;    
	
	strcpy(dg->PfInfo.Interface, interface);
    dg->PfInfo.Address[0] = a1;
    dg->PfInfo.Address[1] = a2;
    dg->PfInfo.Address[2] = a3;
    dg->PfInfo.Address[3] = a4;
    dg->PfInfo.Address[4] = a5;
    dg->PfInfo.Address[5] = a6;
	dg->PfInfo.BufSize = bufsize;
	dg->PfInfo.SnapLen = snaplen;
	dg->PfInfo.Timeout.tv_sec = timeout.tv_sec;
	dg->PfInfo.Timeout.tv_usec = timeout.tv_usec;
	dg->PfInfo.Flags = flags;
	dg->PfInfo.Pos = 0;
	dg->PfInfo.End = 0;
             
    if (NwPfDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWPF_ADDLIST, "nwpf: could not add to list");
    	NwPfCloseDg(dg);
		return NULL;
	}
    
    if (ioctl(dg->PfInfo.Socket, EIOCFLUSH, 0) < 0)
	{
        DnpapMessage(DMC_ERROR, NWPF_IOCTL, "nwpf: read flush failed");
    	NwPfCloseDg(dg);
		return NULL;
	}
    
    return (NW_DG *)dg;
}


BOOLEAN NwPfClose(NW_DG *dg)
{                   
NwPfDg *nitdg = (NwPfDg *)dg;

    NwPfDgRemoveList(nitdg);
    NwPfCloseDg(nitdg);

    return TRUE;
}


BOOLEAN NwPfCloseDg(NwPfDg *dg)
{
    if (dg->Block != NULL)    
        BlockDeRegister(dg->Block);
    
    if (dg->PfInfo.Socket != 0)
        close(dg->PfInfo.Socket);
        
    DelNwPfDg(dg);  
    
    return TRUE;
}    
    

/* Sends a packet over an existing NW datagram
 * The destination address is checked against the recorded Ethernet address
 * to allow only packets that are sent to that address
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwPfSend(NW_DG *dg    /* valid NW datagram */, 
                 BYTE *frame  /* packet */, 
                 INT len      /* packet length */)
{
NwPfDg *nitdg = (NwPfDg *) dg;	
BYTE noaddr[NWPF_SIZE_ADDRESS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    if (len > NWPF_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWPF_SEND, "nwpf: packet too long");
        return FALSE;
    }

    if (memcmp(nitdg->PfInfo.Address, noaddr, NWPF_SIZE_ADDRESS) == 0)
    {
        DnpapMessage(DMC_ERROR, NWPF_SEND, "nwpf: can not send over accept socket");
        return FALSE;
    }
            
    memcpy(frame, nitdg->PfInfo.Address, NWPF_SIZE_ADDRESS);

    if (write(nitdg->PfInfo.Socket, frame, len) < 0)
    {
        DnpapMessage(DMC_ERROR, NWPF_SEND, "nwpf: write failed");
        return FALSE;
    }

    return TRUE;
}


/* Sends a packet to a specified address
 * This function can be used to sent one-shot only packets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwPfSendTo(NW_DG *dg    /* not supported, set to NULL */, 
                   char *addr   /* address specification: 'pf:(char*)interface:(int[6])ethernet' */, 
                   BYTE *frame  /* packet */, 
                   INT len      /* packet length */)
{
IFREQ ifr;
CHAR interface[NWPF_SIZE_INTERFACE];
CHAR *p;
UINT a1, a2, a3, a4, a5, a6;
BYTE address[NWPF_SIZE_ADDRESS];
         
    if (dg != NULL)
        DnpapMessage(DMC_WARNING, NWPF_SEND, "nwpf: sendto over existing connection has no effect");
    
    if (pfsocket == 0)
    {
        DnpapMessage(DMC_ERROR, NWPF_INIT, "nwpf: not initialized");
        return FALSE;
    }
    
    if (len > NWPF_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWPF_SEND, "nwpf: packet too long");
        return FALSE;
    }

    if ((addr = strchr(addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWPF_PARAM, "nwpf: invalid parameter list");
        return FALSE;
    }   
    
	if ((p = strchr(++addr, ':')) == NULL)
	{
		p = addr+strlen(addr);
	}
	else
	{
    	if (sscanf(p, ":%u.%u.%u.%u.%u.%u", &a1, &a2, &a3, &a4, &a5, &a6) != 6)
    	{
            DnpapMessage(DMC_ERROR, NWPF_PARAM, "nwpf: invalid parameters");
    	    return FALSE;
    	}
    
        address[0] = a1;	
        address[1] = a2;	
        address[2] = a3;	
        address[3] = a4;	
        address[4] = a5;	
        address[5] = a6;	
        
        memcpy(frame, address, NWPF_SIZE_ADDRESS);
    }
    
    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
    memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));
    strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name));
    ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';
    if (ioctl(pfsocket, EIOCSETIF, &ifr) < 0)
	{
        DnpapMessage(DMC_ERROR, NWPF_IOCTL, "nwpf: interface bind failed");
		return FALSE;
	}
    
    if (write(pfsocket, frame, len) < 0)
    {
        DnpapMessage(DMC_ERROR, NWPF_SEND, "nwpf: write failed");
        return FALSE;
    }

    return TRUE;
}


BOOLEAN NwPfIfConfig(NW_DG *dg, NW_IFCONFIG *ifconfig)
{
NwPfDg *pfdg = (NwPfDg *) dg;	
ENDEVP devp;

    if (ioctl(pfdg->PfInfo.Socket, EIOCDEVP, (VOID *)&devp) < 0)
		return FALSE;

	switch (devp.end_dev_type)	
	{	
	case ENDT_10MB:
		ifconfig->type = MAC_TYPE_ETHERNET_CSMACD;
		ifconfig->mtu = devp.end_MTU;
		ifconfig->speed = MAC_SPEED_ETHERNET_CSMACD;
		ifconfig->addrlen = devp.end_addr_len;
		memcpy(ifconfig->address, devp.end_addr, ifconfig->addrlen);
		memcpy(ifconfig->broadcast, devp.end_broadaddr, ifconfig->addrlen);
		break;
	default:
		DnpapMessage(DMC_ERROR, NWPF_CONFIG, "nwpf: unknown interface type");
		return FALSE;
	}
		
	return TRUE;
}


int NwPfBlockCallback(BLOCK *blk, int reason, VOID *param)
{
NwPfDg *dg = (NwPfDg *) param;	
	
	switch (reason)
	{
	case BLOCKEVENT_READ:
	case BLOCKEVENT_EXCEPT:
		if (NwPfRead(dg) != TRUE)
			DnpapMessage(DMC_ERROR, NWPF_READ, "nwpf: read failed");
		break;
	case BLOCKEVENT_WRITE:
		DnpapMessage(DMC_ERROR, NWPF_WEVENT, "nwpf: write event");
		break;
/*	case BLOCKEVENT_EXCEPT:
		DnpapMessage(DMC_ERROR, NWPF_EEVENT, "nwpf: except event");
		break;*/
	default:
		DnpapMessage(DMC_ERROR, NWPF_NEVENT, "nwpf: unknown block event");
		return 1;
	}
	
	return 0;
}


BOOLEAN NwPfRead(NwPfDg *dg)
{
BYTE *cp;
ENSTAMP *hp;
USHORT msglen;
BYTE noaddr[NWPF_SIZE_ADDRESS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    if (!(dg->PfInfo.Pos < dg->PfInfo.End))
    {
    	if ((dg->PfInfo.End = read(dg->PfInfo.Socket, dg->PfInfo.Buf, dg->PfInfo.BufSize)) < 0)
    		return FALSE;
            
    	dg->PfInfo.Pos = 0;
    }
		
	while (dg->PfInfo.Pos < dg->PfInfo.End)
	{
		cp = dg->PfInfo.Buf+dg->PfInfo.Pos;
		
		hp = (ENSTAMP *)cp;      
	    msglen = hp->ens_count;
        if (dg->PfInfo.SnapLen > 0 && msglen > dg->PfInfo.SnapLen)
            msglen = dg->PfInfo.SnapLen;                   

		dg->PfInfo.Pos += hp->ens_stamplen + ENALIGN(msglen);

        if (memcmp(dg->PfInfo.Address, noaddr, NWPF_SIZE_ADDRESS) != 0 &&
            memcmp(dg->PfInfo.Address, cp, NWPF_SIZE_ADDRESS) != 0 &&
			memcmp(dg->PfInfo.Address, cp+NWPF_SIZE_ADDRESS, NWPF_SIZE_ADDRESS) != 0)
            continue;
        
		if (dg->NwInfo.recv != NULL)
			dg->NwInfo.recv((NW_DG *)dg, dg->NwInfo.addr, cp, msglen, dg->NwInfo.parm);
	}
	
	return TRUE;
}


BOOLEAN NwPfDgAddList(NwPfDg *dg)
{
    if (NwPfDgList != NULL)
		NwPfDgList->Prev = dg;
        
    dg->Next = NwPfDgList;
    dg->Prev = NULL;
    
    NwPfDgList = dg;
        
    return TRUE;
}


BOOLEAN NwPfDgRemoveList(NwPfDg *dg)
{   
    if (dg->Prev != NULL)
        dg->Prev->Next = dg->Next;
    else
        NwPfDgList = dg->Next;
        
    if (dg->Next != NULL)
        dg->Next->Prev = dg->Prev;
        
    return TRUE;
}


NwPfDg* NewNwPfDg(CHAR *interface, INT bufsize)
{
NwPfDg *dg;

	if ((dg = DnpapMalloc(sizeof(NwPfDg))) == NULL)
		return NULL;
	
	dg->Block = NULL;
    
    dg->NwInfo.iface = NULL;
	dg->NwInfo.recv = NULL;
    dg->NwInfo.parm = NULL;    
    
    dg->PfInfo.Interface = NULL;
    memset(dg->PfInfo.Address, 0, NWPF_SIZE_ADDRESS);
    dg->PfInfo.Buf = NULL;
	dg->PfInfo.Socket = 0;		
	dg->PfInfo.BufSize = 0;
	dg->PfInfo.SnapLen = 0;
	dg->PfInfo.Timeout.tv_sec = 0;
	dg->PfInfo.Timeout.tv_usec = 0;
	dg->PfInfo.Flags = 0;
	dg->PfInfo.Pos = 0;
	dg->PfInfo.End = 0;
    
    dg->Next = NULL;
    dg->Prev = NULL;
	
    if ((dg->Block = DnpapMalloc(sizeof(BLOCK))) == NULL)
		return NULL;
    
	if ((dg->PfInfo.Interface = DnpapMalloc(strlen(interface)+1)) == NULL)
    {
    	DnpapFree(dg->Block);
		return NULL;
	}
	
	if ((dg->PfInfo.Buf = DnpapMalloc(bufsize)) == NULL)
    {
    	DnpapFree(dg->Block);
    	DnpapFree(dg->PfInfo.Interface);
		return NULL;
	}
	
	return dg;
}


VOID DelNwPfDg(NwPfDg *dg)
{
	DnpapFree(dg->Block);
	
    DnpapFree(dg->PfInfo.Interface);
    DnpapFree(dg->PfInfo.Buf);
    
    DnpapFree(dg);
}


BOOLEAN NwPfSetup(NwPfDg *dg, CHAR *interface, INT flags, INT bufsize, INT snaplen, TIMEVAL *timeout)
{
IFREQ ifr;
ENFILTER filter;

    flags |= NWPF_DEFAULT_FLAGS;
                   
    if (ioctl(dg->PfInfo.Socket, EIOCSETW, &bufsize) < 0)
	{
        DnpapMessage(DMC_ERROR, NWPF_IOCTL, "nwpf: bufsize option set failed");
		return FALSE;
	}

    if (ioctl(dg->PfInfo.Socket, EIOCSRTIMEOUT, timeout) < 0)
	{
        DnpapMessage(DMC_ERROR, NWPF_IOCTL, "nwpf: timeout option set failed");
		return FALSE;
	}
    
    memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));
    strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name));
    ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';
    if (ioctl(dg->PfInfo.Socket, EIOCSETIF, &ifr) < 0)
	{
        DnpapMessage(DMC_ERROR, NWPF_IOCTL, "nwpf: interface bind failed");
		return FALSE;
	}
    
    if (snaplen > 0)
    {
        if (ioctl(dg->PfInfo.Socket, EIOCTRUNCATE, &snaplen) < 0)
    	{
            DnpapMessage(DMC_ERROR, NWPF_IOCTL, "nwpf: snapshot length option set failed");
    		return FALSE;
    	}
    }
                                   
    if (flags != 0)
    {
        if (ioctl(dg->PfInfo.Socket, EIOCMBIS, &flags) < 0)
    	{
            DnpapMessage(DMC_ERROR, NWPF_IOCTL, "nwpf: flags set failed");
    		return FALSE;
    	}
    }
    
	filter.enf_Priority = 37;	/* anything > 2 */
	filter.enf_FilterLen = 0;	/* means "always true" */
	if (ioctl(dg->PfInfo.Socket, EIOCSETF, &filter) < 0)
	{
        DnpapMessage(DMC_ERROR, NWPF_IOCTL, "nwpf: filter option set failed");
		return FALSE;
	}
    
    return TRUE;
}
