/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <block.h>

#ifndef BSD_SELECT
#define BSD_SELECT
#endif

#ifdef OS2
#include <types.h>
#include <netlib.h>
#include <sys/select.h>					/* contains FD_*** macros */
#else
#include <sys/types.h>
#define sock_init()	0
#define soclose(x)	close(x)
#include <unistd.h>
#endif
#include <netdb.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <dinet.h>
#if defined(SUNOS) || defined(SOLARIS)
#include <stropts.h>
#endif
#include <errno.h>

#include "nw.h"
#include "nwtcp.h"

typedef struct _NW_TCP_CO NW_TCP_CO;
typedef struct _NW_TCP_AC NW_TCP_AC;

struct _NW_TCP_CO
{
	NW_CO nw;
	int socket;
	BLOCK block;
	NW_TCP_CO *next;
};

struct _NW_TCP_AC
{
	NW_AC nw;
	int socket;
	BLOCK block;
	NW_TCP_AC *next;
};

static NwStat nwStat;

static BOOLEAN Init(void);
static BOOLEAN Exit(void);
static NW_AC *Accept(char *addr, NW_CO_RECV_CALL rcv, NW_CO_CLOSE_CALL clse, NW_CO_OPEN_CALL opn, void *parm);
static NW_CO *Connect(char *addr, NW_CO_RECV_CALL rcv, NW_CO_CLOSE_CALL clse, void *parm);
static BOOLEAN Send(NW_CO * co, BYTE * frame, int length);
static BOOLEAN Close(NW_CO * co);
static BOOLEAN Release(NW_AC * ac);
static BOOLEAN IfConfig(NW_CO * co, NW_IFCONFIG * ifconfig);
static NwStat *Statistics(char *addr);
static int RecvCall(BLOCK * block, int reason, void *parm);
static int OpenCall(BLOCK * block, int reason, void *parm);
static struct sockaddr_in *Addr2Sock(char *addr);
static char *Sock2Addr(struct sockaddr_in *sa);

NW_CO_IFACE nwTcpIface =
{
	"tcp",
	Init,
	Exit,
	Accept,
	Connect,
	Send,
	Close,
	Release,
	IfConfig,
	Statistics
};

static NW_TCP_CO *nwTcpCoList = 0;
static NW_TCP_AC *nwTcpAcList = 0;

BOOLEAN Init(void)
{
	static BOOLEAN init = FALSE;

	if (init)
		return TRUE;

	if (sock_init() != 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: init: sock_init failed");
		return FALSE;
	}

	init = TRUE;

	return init;
}

BOOLEAN Exit(void)
{
	NW_TCP_CO *co, *p;
	NW_TCP_AC *ac, *q;

	p = nwTcpCoList;

	while (p != 0)
	{
		co = p;
		p = p->next;

		BlockDeRegister(&co->block);
		soclose(co->socket);
		DnpapFree(co);
	}
	q = nwTcpAcList;

	while (q != 0)
	{
		ac = q;
		q = q->next;

		BlockDeRegister(&ac->block);
		soclose(ac->socket);
		DnpapFree(ac);
	}

	return TRUE;
}

NwStat *Statistics(char *addr)
{
	return &nwStat;
}

NW_AC *Accept(char *addr, NW_CO_RECV_CALL rcv, NW_CO_CLOSE_CALL clse, NW_CO_OPEN_CALL opn, void *parm)
{
	NW_TCP_AC *tcp;
	struct sockaddr_in *sa;

	tcp = (NW_TCP_AC *) DnpapMalloc(sizeof(NW_TCP_AC));
	if (tcp == 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: accept: malloc failed");
		return 0;
	}

	tcp->socket = socket(AF_INET, SOCK_STREAM, 0);
	if (tcp->socket < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: accept: socket failed");
		return 0;
	}

	sa = Addr2Sock(addr);
	if (sa == 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: accept: couldn't make socket address");
		return 0;
	}

	if (bind(tcp->socket, (struct sockaddr *) sa, sizeof(struct sockaddr_in)) < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: accept: can't bind server socket");
		return 0;
	}

	listen(tcp->socket, NW_TCP_LISTEN);

	tcp->block.sckt = tcp->socket;
	if (BlockRegister(&(tcp->block), tcp, OpenCall) != 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: accept: failed to register server block");
		return 0;
	}

	tcp->nw.iface = &nwTcpIface;
	strcpy(tcp->nw.addr, addr);
	tcp->nw.recv = rcv;
	tcp->nw.close = clse;
	tcp->nw.open = opn;
	tcp->nw.parm = parm;

	tcp->next = nwTcpAcList;
	nwTcpAcList = tcp;

	return (NW_AC *) tcp;
}

NW_CO *Connect(char *addr, NW_CO_RECV_CALL rcv, NW_CO_CLOSE_CALL clse, void *parm)
{
	NW_TCP_CO *tcp;
	struct sockaddr_in *sa;

	tcp = (NW_TCP_CO *) DnpapMalloc(sizeof(NW_TCP_CO));
	if (tcp == 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: connect: malloc failed");
		return 0;
	}

	tcp->socket = socket(AF_INET, SOCK_STREAM, 0);
	if (tcp->socket < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: connect: socket failed");
		return 0;
	}

	sa = Addr2Sock(addr);
	if (sa == 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: connect: couldn't make socket address");
		return 0;
	}

	if (connect(tcp->socket, (struct sockaddr *) sa, sizeof(struct sockaddr_in)) < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: connect: can't bind server socket");
		return 0;
	}

	tcp->block.sckt = tcp->socket;
	if (BlockRegister(&(tcp->block), tcp, RecvCall) != 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: connect: failed to register server block");
		return 0;
	}

	tcp->nw.iface = &nwTcpIface;
	strcpy(tcp->nw.addr, addr);
	tcp->nw.recv = rcv;
	tcp->nw.close = clse;
	tcp->nw.parm = parm;

	tcp->next = nwTcpCoList;
	nwTcpCoList = tcp;

	return (NW_CO *) tcp;
}

BOOLEAN Send(NW_CO * co, BYTE * frame, int length)
{
	int n;
	NW_TCP_CO *tcp;

	tcp = (NW_TCP_CO *) co;

	n = send(tcp->socket, frame, length, 0);

	if (n < 0 || n != length)
		return FALSE;

	return TRUE;
}

BOOLEAN Close(NW_CO * co)
{
	NW_TCP_CO **p, *tcp;

	tcp = (NW_TCP_CO *) co;

	for (p = &nwTcpCoList; *p != 0 && *p != tcp; p = &(*p)->next) ;

	if (*p == 0)
		return FALSE;

	*p = tcp->next;

	BlockDeRegister(&tcp->block);
	soclose(tcp->socket);

	DnpapFree(tcp);

	return TRUE;
}

BOOLEAN Release(NW_AC * ac)
{
	NW_TCP_AC **p, *tcp;

	tcp = (NW_TCP_AC *) ac;

	for (p = &nwTcpAcList; *p != 0 && *p != tcp; p = &(*p)->next) ;

	if (*p == 0)
		return FALSE;

	*p = tcp->next;

	BlockDeRegister(&tcp->block);
	soclose(tcp->socket);

	DnpapFree(tcp);

	return TRUE;
}

BOOLEAN IfConfig(NW_CO * co, NW_IFCONFIG * ifconfig)
{
	return FALSE;
}

int RecvCall(BLOCK * block, int reason, void *parm)
{
	NW_TCP_CO *tcp;
	BYTE buffer[NW_TCP_BUFFER_SIZE];
	struct sockaddr_in sa;
	int n, m;

	tcp = (NW_TCP_CO *) parm;
	switch (reason)
	{
	case BLOCKEVENT_READ:
		sa.sin_family = AF_INET;
		m = sizeof(struct sockaddr_in);

		n = read(tcp->socket, buffer, sizeof(buffer));
		if (n < 0)
		{
			DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: recvcall: recvfrom error");
			return -1;
		}

		if (n == 0)
		{
			tcp->nw.close((NW_CO *) tcp, tcp->nw.parm);
			Close((NW_CO *) tcp);
		}
		else
		{
			tcp->nw.recv((NW_CO *) tcp, buffer, n, tcp->nw.parm);
		}
		break;
	case BLOCKEVENT_WRITE:
		/* ignore these events */
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: recvcall: unexpected BLOCKEVENT_WRITE");
		break;
	case BLOCKEVENT_EXCEPT:
		/* ignore these events */
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: recvcall: unexpected BLOCKEVENT_EXCEPT");
		break;
	default:
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: recvcall: unknown block event");
		break;
	}
	return 0;
}

int OpenCall(BLOCK * block, int reason, void *parm)
{
	NW_TCP_AC *ac;
	NW_TCP_CO *tcp;
	struct sockaddr_in sa;
	int m, s;
	char *addr;
	void *p;

	ac = (NW_TCP_AC *) parm;
	switch (reason)
	{
	case BLOCKEVENT_READ:
		sa.sin_family = AF_INET;
		m = sizeof(struct sockaddr_in);

		s = accept(ac->socket, (struct sockaddr *) &sa, &m);
		if (s < 0)
		{
			DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: opencall: accept failed");
			return -1;
		}

		tcp = (NW_TCP_CO *) DnpapMalloc(sizeof(NW_TCP_CO));
		if (tcp == 0)
		{
			soclose(s);
			DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: opencall: malloc failed");
			return -1;
		}

		tcp->socket = s;

		tcp->block.sckt = s;
		if (BlockRegister(&(tcp->block), tcp, RecvCall) != 0)
		{
			DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: opencall: failed to register block");
			return -1;
		}

		addr = Sock2Addr(&sa);
		if (addr == 0)
		{
			DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: opencall: couldn't translate address");
			return -1;
		}

		tcp->nw.iface = &nwTcpIface;
		strcpy(tcp->nw.addr, addr);
		tcp->nw.recv = ac->nw.recv;
		tcp->nw.close = ac->nw.close;
		tcp->nw.parm = ac->nw.parm;

		tcp->next = nwTcpCoList;
		nwTcpCoList = tcp;

		if ((p = ac->nw.open((NW_CO *) tcp, addr, ac->nw.parm)) != NULL)
			tcp->nw.parm = p;

		break;
	case BLOCKEVENT_WRITE:
		/* ignore these events */
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: opencall: unexpected BLOCKEVENT_WRITE");
		break;
	case BLOCKEVENT_EXCEPT:
		/* ignore these events */
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: opencall: unexpected BLOCKEVENT_EXCEPT");
		break;
	default:
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: opencall: unknown block event");
		break;
	}
	return 0;
}

struct sockaddr_in *Addr2Sock(char *addr)
{
	static struct sockaddr_in sa;
	char *p, *q;
	char string[32];
	struct hostent *ent;

	p = addr;

	q = string;
	while (*p != ':' && *p != 0)
		*q++ = tolower(*p++);
	*q++ = 0;

	if (strcmp(string, "tcp") != 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: addr2sock: wrong address family");
		return 0;
	}

	if (*p++ != ':')
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: addr2sock: colon expected");
		return 0;
	}

	q = string;
	while (*p != ':' && *p != 0)
		*q++ = *p++;
	*q++ = 0;

	if (strcmp(string, "*") == 0)
	{
		sa.sin_addr.s_addr = INADDR_ANY;
	}
	else
	{
		sa.sin_addr.s_addr = inet_addr(string);
		if (sa.sin_addr.s_addr == -1)
		{
			ent = gethostbyname(string);
			if (ent != 0)
			{
				sa.sin_addr.s_addr = ((struct in_addr *) (VOID *) ent->h_addr_list[0])->s_addr;
			}
			else
			{
				DnpapMessage(DMC_ERROR, NW_ERROR, "nwtcp: addr2sock: can't resolve host");
				return 0;
			}
		}
	}

	if (*p++ != ':')
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: addr2sock: colon expected");
		return 0;
	}

	q = string;
	while (*p != ':' && *p != 0)
		*q++ = *p++;
	*q++ = 0;

	sa.sin_port = htons(atoi(string));

	sa.sin_family = AF_INET;

	return &sa;
}

char *Sock2Addr(struct sockaddr_in *sa)
{
	static char addr[32];

	if (sa->sin_family != AF_INET)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwtcp: sock2addr: wrong address family");
		return 0;
	}

	sprintf(addr, "tcp:%s:%d", Inet_NtoA(sa->sin_addr.s_addr), ntohs(sa->sin_port));

	return addr;
}
