/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <block.h>

#ifndef BSD_SELECT
#define BSD_SELECT
#endif

#ifdef OS2
#include <types.h>
#include <netlib.h>
#include <sys/select.h>					/* contains FD_*** macros */
#else
#include <sys/types.h>
#define sock_init()	0
#define soclose(x)	close(x)
#include <unistd.h>
#endif
#include <netdb.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <dinet.h>
#if defined(SUNOS) || defined(SOLARIS)
#include <stropts.h>
#endif
#include <errno.h>

#include "nw.h"
#include "nwudp.h"

typedef struct _NW_UDP_DG NW_UDP_DG;

struct _NW_UDP_DG
{
	NW_DG nw;
	int socket;
	BLOCK block;
	NW_UDP_DG *next;
};

static NwStat nwStat;

static BOOLEAN Init(void);
static BOOLEAN Exit(void);
static NW_DG *Accept(char *addr, NW_DG_RECV_CALL rcv, void *parm);
static NW_DG *Connect(char *addr, NW_DG_RECV_CALL rcv, void *parm);
static BOOLEAN Close(NW_DG * dg);
static BOOLEAN Send(NW_DG * dg, BYTE * frame, int length);
static BOOLEAN SendTo(NW_DG * dg, char *addr, BYTE * frame, int length);
static BOOLEAN IfConfig(NW_DG * dg, NW_IFCONFIG * ifconfig);
static NwStat *Statistics(char *addr);
static int RecvCall(BLOCK * block, int reason, void *param);
static struct sockaddr_in *Addr2Sock(char *addr);
static char *Sock2Addr(struct sockaddr_in *sa);

NW_DG_IFACE nwUdpIface =
{
	"udp",
	Init,
	Exit,
	Accept,
	Connect,
	Send,
	SendTo,
	Close,
	IfConfig,
	Statistics
};

static NW_UDP_DG *nwUdpDgList = 0;
static int nwUdpSocket = 0;

BOOLEAN Init(void)
{
	static BOOLEAN init = FALSE;
	struct sockaddr_in sa;

	if (init)
		return TRUE;

	if (sock_init() != 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: init: sock_init failed");
		return FALSE;
	}

	nwUdpSocket = socket(AF_INET, SOCK_DGRAM, 0);
	if (nwUdpSocket < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: init: socket failed");
		return FALSE;
	}
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = INADDR_ANY;
	sa.sin_port = htons(0);
	if (bind(nwUdpSocket, (struct sockaddr *) &sa, sizeof(struct sockaddr_in)) < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: init: bind failed");
		return FALSE;
	}

	init = TRUE;

	return init;
}

BOOLEAN Exit(void)
{
	NW_UDP_DG *udp, *p;

	p = nwUdpDgList;

	while (p != 0)
	{
		udp = p;
		p = p->next;

		BlockDeRegister(&udp->block);
		soclose(udp->socket);
		DnpapFree(udp);
	}

	soclose(nwUdpSocket);

	return TRUE;
}

NwStat *Statistics(char *addr)
{
	return &nwStat;
}

NW_DG *Accept(char *addr, NW_DG_RECV_CALL rcv, void *parm)
{
	NW_UDP_DG *udp;
	struct sockaddr_in *sa;

	udp = (NW_UDP_DG *) DnpapMalloc(sizeof(NW_UDP_DG));
	if (udp == 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: accept: malloc failed");
		return 0;
	}

	udp->socket = socket(AF_INET, SOCK_DGRAM, 0);
	if (udp->socket < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: accept: socket failed");
		return 0;
	}

	sa = Addr2Sock(addr);
	if (sa == 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: accept: couldn't make socket address");
		return 0;
	}

	if (bind(udp->socket, (struct sockaddr *) sa, sizeof(struct sockaddr_in)) < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: accept: can't bind server socket");
		return 0;
	}

	udp->block.sckt = udp->socket;
	if (BlockRegister(&(udp->block), udp, RecvCall) != 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: accept: failed to register server block");
		return 0;
	}

	udp->nw.iface = &nwUdpIface;
	strcpy(udp->nw.addr, addr);
	udp->nw.recv = rcv;
	udp->nw.parm = parm;

	udp->next = nwUdpDgList;
	nwUdpDgList = udp;

	return (NW_DG *) udp;
}

NW_DG *Connect(char *addr, NW_DG_RECV_CALL rcv, void *parm)
{
	NW_UDP_DG *udp;
	struct sockaddr_in *sa;

	udp = (NW_UDP_DG *) DnpapMalloc(sizeof(NW_UDP_DG));
	if (udp == 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: connect: malloc failed");
		return 0;
	}

	udp->socket = socket(AF_INET, SOCK_DGRAM, 0);
	if (udp->socket < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: connect: socket failed");
		return 0;
	}

	sa = Addr2Sock(addr);
	if (sa == 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: connect: couldn't make socket address");
		return 0;
	}

	if (connect(udp->socket, (struct sockaddr *) sa, sizeof(struct sockaddr_in)) < 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: connect: can't bind server socket");
		return 0;
	}

	udp->block.sckt = udp->socket;
	if (BlockRegister(&(udp->block), udp, RecvCall) != 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: connect: failed to register server block");
		return 0;
	}

	udp->nw.iface = &nwUdpIface;
	strcpy(udp->nw.addr, addr);
	udp->nw.recv = rcv;
	udp->nw.parm = parm;

	udp->next = nwUdpDgList;
	nwUdpDgList = udp;

	return (NW_DG *) udp;
}

BOOLEAN Close(NW_DG * dg)
{
	NW_UDP_DG **p, *udp;

	udp = (NW_UDP_DG *) dg;

	for (p = &nwUdpDgList; *p != 0 && *p != udp; p = &(*p)->next) ;

	if (*p == 0)
		return FALSE;

	*p = udp->next;

	BlockDeRegister(&udp->block);
	soclose(udp->socket);

	DnpapFree(udp);

	return TRUE;
}

BOOLEAN Send(NW_DG * dg, BYTE * frame, int length)
{
	int n, s;

	if (dg != 0)
		s = ((NW_UDP_DG *) dg)->socket;
	else
		return FALSE;

	n = send(s, frame, length, 0);

	if (n < 0 || n != length)
		return FALSE;

	return TRUE;
}

BOOLEAN SendTo(NW_DG * dg, char *addr, BYTE * frame, int length)
{
	int n, s;
	struct sockaddr_in *sa;

	if (dg != 0)
		s = ((NW_UDP_DG *) dg)->socket;
	else
		s = nwUdpSocket;

	sa = Addr2Sock(addr);
	if (sa == 0)
		return FALSE;

	n = sendto(s, frame, length, 0, (struct sockaddr *) sa, sizeof(struct sockaddr_in));

	if (n < 0 || n != length)
		return FALSE;

	return TRUE;
}

BOOLEAN IfConfig(NW_DG * dg, NW_IFCONFIG * ifconfig)
{
	return FALSE;
}

int RecvCall(BLOCK * block, int reason, void *parm)
{
	NW_UDP_DG *udp;
	BYTE buffer[NW_UDP_BUFFER_SIZE];
	struct sockaddr_in sa;
	int n, m;
	char *addr;

	udp = (NW_UDP_DG *) parm;
	switch (reason)
	{
	case BLOCKEVENT_READ:
		sa.sin_family = AF_INET;
		m = sizeof(struct sockaddr_in);

		n = recvfrom(udp->socket, buffer, sizeof(buffer), 0,
					 (struct sockaddr *) &sa, &m);
		if (n < 0)
		{
			DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: recvcall: recvfrom error");
			return -1;
		}
		addr = Sock2Addr(&sa);
		if (addr == 0)
		{
			DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: recvcall: couldn't translate address");
			return -1;
		}
		udp->nw.recv((NW_DG *) udp, addr, buffer, n, udp->nw.parm);
		break;
	case BLOCKEVENT_WRITE:
		/* ignore these events */
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: recvcall: unexpected BLOCKEVENT_WRITE");
		break;
	case BLOCKEVENT_EXCEPT:
		/* ignore these events */
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: recvcall: unexpected BLOCKEVENT_EXCEPT");
		break;
	default:
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: recvcall: unknown block event");
		break;
	}
	return 0;
}

struct sockaddr_in *Addr2Sock(char *addr)
{
	static struct sockaddr_in sa;
	char *p, *q;
	char string[32];
	struct hostent *ent;

	p = addr;

	q = string;
	while (*p != ':' && *p != 0)
		*q++ = tolower(*p++);
	*q++ = 0;

	if (strcmp(string, "udp") != 0)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: addr2sock: wrong address family");
		return 0;
	}

	if (*p++ != ':')
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: addr2sock: colon expected");
		return 0;
	}

	q = string;
	while (*p != ':' && *p != 0)
		*q++ = *p++;
	*q++ = 0;

	if (strcmp(string, "*") == 0)
	{
		sa.sin_addr.s_addr = INADDR_ANY;
	}
	else
	{
		sa.sin_addr.s_addr = inet_addr(string);
		if (sa.sin_addr.s_addr == -1)
		{
			ent = gethostbyname(string);
			if (ent != 0)
			{
				sa.sin_addr.s_addr = ((struct in_addr *) (VOID *) ent->h_addr_list[0])->s_addr;
			}
			else
			{
				DnpapMessage(DMC_ERROR, NW_ERROR, "nwudp: addr2sock: can't resolve host");
				return 0;
			}
		}
	}

	if (*p++ != ':')
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: addr2sock: colon expected");
		return 0;
	}

	q = string;
	while (*p != ':' && *p != 0)
		*q++ = *p++;
	*q++ = 0;

	sa.sin_port = htons(atoi(string));

	sa.sin_family = AF_INET;

	return &sa;
}

char *Sock2Addr(struct sockaddr_in *sa)
{
	static char addr[32];

	if (sa->sin_family != AF_INET)
	{
		DnpapMessage(DMC_WARNING, NW_ERROR, "nwudp: sock2addr: wrong address family");
		return 0;
	}

	sprintf(addr, "udp:%s:%d", Inet_NtoA(sa->sin_addr.s_addr), ntohs(sa->sin_port));

	return addr;
}
