package buffer;
#########################################################################
## MODULE INFORMATION#
######################
##     FILE     NAME:       buffer.pl
##     SYSTEM   NAME:       buffer
##     ORIGINAL AUTHOR(S):  Jan van Oorschot
##     VERSION  NUMBER:     1.00
##     CREATION DATE:       Thu Jul 14 13:02:49 1994
##
## DESCRIPTION: create and destroy buffers
##              
#########################################################################
## CHANGE INFORMATION ##
#########################
##      JvO - Jul 14, 1994:  Created
#########################################################################
require "spook.pl";	   
require "rmon/rmon.pl";

sub bufferCreatorCB
{
	local($cnr,$colnam,$colnr) = @_;
	local($varName , $res , $iface , $int);

	# Data to be set for an buffer table
	#
	# bufferControlChannelIndex       --> id of channel to use
	# bufferControlFullStatus         --> spaceAvailable(1)
	#                                     full(2)
	#                                     (ReadOnly)
	# bufferControlFullAction         --> lockWhenFull(1)
	#                                     wrapWhenFull(2)
	# bufferControlCaptureSliceSize   --> max # octets saved per frame
	# bufferControlDownloadSliceSize  --> max # octets returned per frame
	# bufferControlDownloadOffset     --> offset returned
	# bufferControlMaxOctetsRequested --> max # octets in this buffer, requested
	#                                     (-1 is unlimited)
	# bufferControlMaxOctetsGranted   --> max # octets in this buffer, granted
	#                                     (ReadOnly)
	# bufferControlCapturedPackets    --> # packets currently in buffer
	#                                     (ReadOnly)
	# bufferControlTurnOnTime         --> sysUptime when buffer was turned on
	#                                     (ReadOnly)
        #
	$snmpCommand="bufferControlChannelIndex[$colnr]=$bufferControlChannelIndex";
	$res=&main'snmp_set($cnr,$snmpCommand);
        if($res != $bufferControlChannelIndex)
        {
             return 0;
        }
	$snmpCommand = "bufferControlFullAction[$colnr]=$bufferControlFullAction";
	$res=&main'snmp_set($cnr,$snmpCommand);
        if($res != $bufferControlFullAction)
        {
             return 0;
        }
	$snmpCommand = "bufferControlCaptureSliceSize[$colnr]=$bufferControlCaptureSliceSize";
	$res=&main'snmp_set($cnr,$snmpCommand);
        if($res != $bufferControlCaptureSliceSize)
        {
             return 0;
        }
	$snmpCommand = "bufferControlDownloadSliceSize[$colnr]=$bufferControlDownloadSliceSize";
	$res=&main'snmp_set($cnr,$snmpCommand);
        if($res != $bufferControlDownloadSliceSize)
        {
             return 0;
        }
	$snmpCommand = "bufferControlDownloadOffset[$colnr]=$bufferControlDownloadOffset";
	$res=&main'snmp_set($cnr,$snmpCommand);
        if($res != $bufferControlDownloadOffset)
        {
             return 0;
        }
	$snmpCommand = "bufferControlMaxOctetsRequested[$colnr]=$bufferControlMaxOctetsRequested";
	$res=&main'snmp_set($cnr,$snmpCommand);
        if($res != $bufferControlMaxOctetsRequested)
        {
             return 0;
        }
	return 1;
}	   

# create the etherStats collector with the correct inFun
sub main'bufferCreator
{
    local(
	  $cnr,$owner,
	  $bufferControlChannelIndex,
	  $bufferControlFullAction,
	  $bufferControlCaptureSliceSize,
	  $bufferControlDownloadSliceSize,
	  $bufferControlDownloadOffset,
	  $bufferControlMaxOctetsRequested
	  ) =@_;
    $colnr = &main'create_new_collector(
				    $cnr,
				    "bufferControl",
				    $owner,
				    buffer'bufferCreatorCB) || return 0;
    return( $colnr);
}

# destroy the host collector
sub main'bufferTerminator
{
    local($conn,$coll, $owner) = @_;

    return &main'remove_collector($conn,"bufferControl",$coll,$owner);
}

#
# receive a row of the table, for the use with snmp_tbl
#
sub bufferGardnerCB
{
	local($i,$j,@repl) = @_;
	if($j == 1)
	{
	    print $localFile "\n$localId $localTable ";
	}
	print $localFile "$repl[2] ";
}

#
# request the host table
#
# input variables:
#      
#      1: $conn        spook connection number
#      4: $coll        collector number of host collector
#      5: $id          identification string
#      6: $file        file handle to write to
#
sub main'bufferGardner
{
    local( $conn, $coll, $id, $file) = @_;
    local( @res );

    $localId = $id;
    $localFile = $file;

    $localTable = "bufferControl";
    @res = &main'snmp_get( $conn,
         (
              "bufferControlIndex[$coll]",                  # 0
              "bufferControlChannelIndex[$coll]",           # 1
              "bufferControlFullStatus[$coll]",             # 2
              "bufferControlFullAction[$coll]",             # 3
              "bufferControlCaptureSliceSize[$coll]",       # 4
              "bufferControlDownloadSliceSize[$coll]",      # 5
              "bufferControlDownloadOffset[$coll]",         # 6
              "bufferControlMaxOctetsRequested[$coll]",     # 7 
              "bufferControlMaxOctetsGranted[$coll]",       # 8
              "bufferControlCapturedPackets[$coll]",        # 9
              "bufferControlTurnOnTime[$coll]",             # 10
              "bufferControlOwner[$coll]",                  # 11
              "bufferControlStatus[$coll]"                 # 12
         ));
    $time = time();
    print $file "\n$id hostControlStart $time $res[1] $res[2] $res[3] $res[4] $res[5] $res[6] $res[7] $res[8] $res[9] $res[10] $res[11] $res[12]";

    $localTable = "captureBuffer";
    &main'snmp_tbl($conn,buffer'bufferGardnerCB,
         (
              "captureBufferControlIndex[$coll]",
              "captureBufferIndex[$coll]",
              "captureBufferPacketID[$coll]",
              "captureBufferPacketData[$coll]",
              "captureBufferPacketLength[$coll]",
              "captureBufferPacketTime[$coll]",
              "captureBufferPacketStatus[$coll]"
         ));

    $localTable = "bufferControl";
    @res = &main'snmp_get( $conn,
         (
              "bufferControlIndex[$coll]",                  # 0
              "bufferControlChannelIndex[$coll]",           # 1
              "bufferControlFullStatus[$coll]",             # 2
              "bufferControlFullAction[$coll]",             # 3
              "bufferControlCaptureSliceSize[$coll]",       # 4
              "bufferControlDownloadSliceSize[$coll]",      # 5
              "bufferControlDownloadOffset[$coll]",         # 6
              "bufferControlMaxOctetsRequested[$coll]",     # 7 
              "bufferControlMaxOctetsGranted[$coll]",       # 8
              "bufferControlCapturedPackets[$coll]",        # 9
              "bufferControlTurnOnTime[$coll]",             # 10
              "bufferControlOwner[$coll]",                  # 11
              "bufferControlStatus[$coll]"                 # 12
         ));
    $time = time();
    print $file "\n$id hostControlStart $time $res[1] $res[2] $res[3] $res[4] $res[5] $res[6] $res[7] $res[8] $res[9] $res[10] $res[11] $res[12]";
  
    return 1;
}
1;





