package matrix;
#
#	This module creates and interrogates a
#       Matrix Collector.
#       Such a collector examines all frames and builds a
#       node array containing the Ethernet
#       information of all ethernet matrixs it has seen
#       
#       This modules implements the following functions:
#
#       matrixCreator($connection,$ownerstring,$interface-number)
#
#               . creates a matrix collector
#                   <$connection> should be a Spook connection, and
#                 should point to the correct RMON monitor with
#                 the appropriate community string set.
#                   <$ownerstring> is the identifier for
#                 this collector
#                   <interface-number> is an integer identifying
#                 the network interface in the RMON monitor to be 
#                 used for building the node array.
#               . returns a integer representing the collector number
#                 or 0 if and error occured.
#
#       matrixTerminator($connection,$collector,$ownerstring)
# 
#               . terminates a matrix Collector
#                   <$connection> should be a correct
#                 spook for this operation.
#               .   <$collector> should be the collector number
#                 for the collector to be destroyed.
#               .   <ownerstring> should be the string
#                 identifying the collector
#
#        
#       matrixGardner($connection,$collector,$id-string,$file-hdnl)
#
#               . collects the date from the matrix
#                 Collector. The data is stored in a file in an
#                 ASCI format, each line containing the data
#                 about one connection
#                   <$connection> shoould be a correct connection
#                 for this operation
#                   <$collector> is the collector id 
#                   <file-hndl> file handle of file where
#                   results are written to.
#                   The format of this file can be derived by looking
#                   at the source code of this perl script :-).
#                   But in short, every line contains a row
#                   from one of the tables in the Discovery Collector.
#                   The first and the last line contain starting
#                   and stop time and status info of the collector. 
#####################################################################
require "spook.pl";	   
require "rmon/rmon.pl";
#
# define the functions
#
#
# function iniFun will be called during initialisation
# of the collector. This is the place to set all kind
# of collector specific values. In the case of etherStats, 
# this is only the ethernet-interface number etherStatsDataSource
#
$iniIface = "ifIndex[1]";
sub matrixCreatorCB
{
	local($connNr , $collName , $collNr) = @_;
	local($SNMPRequest , $res);
	
	$SNMPRequest = $collName . "DataSource\[$collNr\]=$iniIface";
	$res = &main'snmp_set($connNr,$SNMPRequest);
	if($res ne $iniIface)
	{
		return 0;
	}
	return 1;
}	   

# create the etherStats collector with the correct inFun
sub main'matrixCreator
{
    local($conn,$owner,$iface) = @_;

    $iniIface = $iface;
    
    $collNr = &main'create_new_collector(
				    $conn,
				    "matrixControl",
				    $owner,
				    matrix'matrixCreatorCB) || return 0;
    return( $collNr) ;
}

# destroy the matrix collector
sub main'matrixTerminator
{
    local($conn,$coll, $owner) = @_;

    return &main'remove_collector($conn,"matrixControl",$coll,$owner);
}

#
# receive a row of the table, for the use with snmp_tbl
#
sub matrixGardnerCB
{
	local($i,$j,@repl) = @_;
	if($j == 1)
	{
	    print $localFile "\n$localId $localTable ";
	}
	print $localFile "$repl[2] ";
}

#
# request the matrix table
#
# input variables:
#      
#      1: $conn        spook connection number
#      4: $coll        collector number of host collector
#      5: $id          identification string
#      6: $file        file handle to write to
#
sub main'matrixGardner
{
    local( $conn, $coll, $id, $file) = @_;
    local( @res );
    &main'snmp_connect($conn) || return 0;

    $localId = $id;
    $localFile = $file;

    $localTable = "matrixControl";
    @res = &main'snmp_get($conn,
         (
	    "sysUpTime[0]",
            "matrixControlIndex[$coll]",
            "matrixControlDataSource[$coll]",
            "matrixControlOwner[$coll]",
            "matrixControlStatus[$coll]",
            "matrixControlTableSize[$coll]",
            "matrixControlLastDeleteTime[$coll]"
         ));
    $time = time();
    print $file "\n$id matrixControlStart $time $res[0] $res[1] $res[2] $res[3] $res[4] $res[5] $res[6]";  

    $localTable = "matrixSD";
    &main'snmp_tbl($conn,matrix'matrixGardnerCB,
         (
            "matrixSDSourceAddress[$coll]",
            "matrixSDDestAddress[$coll]",
            "matrixSDIndex[$coll]",
            "matrixSDPkts[$coll]",
            "matrixSDOctets[$coll]",
            "matrixSDErrors[$coll]"
         ));

    $localTable = "matrixDS";
    &main'snmp_tbl($conn,matrix'matrixGardnerCB,
         (
            "matrixDSSourceAddress[$coll]",
            "matrixDSDestAddress[$coll]",
            "matrixDSIndex[$coll]",
            "matrixDSPkts[$coll]",
            "matrixDSOctets[$coll]",
            "matrixDSErrors[$coll]"
         ));

    $localTable = "matrixControl";
    @res = &main'snmp_get($conn,
         (
	    "sysUpTime[0]",
            "matrixControlIndex[$coll]",
            "matrixControlDataSource[$coll]",
            "matrixControlOwner[$coll]",
            "matrixControlStatus[$coll]",
            "matrixControlTableSize[$coll]",
            "matrixControlLastDeleteTime[$coll]"
         ));
    $time = time();
    print $file "\n$id matrixControlEnd $time $res[0] $res[1] $res[2] $res[3] $res[4] $res[5] $res[6]";  
    return 1;
}
1;





