package rmon;
#
#
require "tricklet.pl";
require "spook.pl";

# Return a correct ownerstring
# a quote from rfc1271:
#
#              OwnerString ::= DisplayString
#              -- This data type is used to model an administratively
#              -- assigned name of the owner of a resource. This
#              -- information is taken from the NVT ASCII character set.
#              -- It is suggested that this name contain one or more
#              -- of the following:
#              -- IP address, management station name, network manager's
#              -- name, location, or phone number.
#              -- In some cases the agent itself will be the owner of
#              -- an entry.  In these cases, this string shall be set
#              -- to a string starting with 'monitor'.
sub main'ownerstring
{
    local($login,$name,$aliases,$addrtype,$length,@addrs,$os);
    local($name,$pwd,$uid,$gid,$qta,$commend,$gcos,$dir,$shell);
    $login = getlogin();
    ($name,$pwd,$uid,$gid,$qta,$commend,$gcos,$dir,$shell)=getpwnam($login);
    $gcos =~ s/ /_/g;
    $host = `hostname`;chop($host);
    ($name,$aliases,$addrtype,$length,@addrs) = gethostbyname($host);
    ($a,$b,$c,$d) = unpack('C4',$addrs[0]);
    $addr = "$a.$b.$c.$d";
    @parts = split('\/', $0);
    $script = $parts[$#parts];
    $os = 
	$addr . 
	"(" . $name . ") " . 
	$script . " " .
	$login . 
	"(" . $gcos . ") " .
	&main'time2str(time());
    return $os;
}
$main'iniOwnerString = &main'ownerstring;	# initialise it for this script
#**********************************************************************
# list of all possible collector types
#**********************************************************************
@main'RMONTables = (
	"etherStats",
	"historyControl",
	"hostControl",
	"hostTopN",
	"matrixControl",
	"filter",
	"channel",
	"bufferControl",
	"event",
	"alarm",
	"beholderInfo",
	"neuro",
	"discNodeControl",
	"discConnControl",
	"subscriptRunControl",
	"subscriptSourceControl"
);

@collector_list = ();
 
sub rmon'collector_list_tbl
{
	local($i,$j,@repl) = @_;
	push(@collector_list,($repl[1]));	# remember index
}
#**********************************************************************
#* Name:	list_collectors
#* Synopsis:	list list_collectors(int connNr, char *collName);
#
#* Description:	This fuction return a list of the active
#*		collectors of the type <collName> on <connNr>.
#* Returns:	list of integers
#**********************************************************************
sub main'list_collectors
{
	local($nr,$varName) = @_;
	local($res,$varStat );
					     
	$varStat = $varName . "Status\[\]";
       	@collector_list = ();
	&main'snmp_tbl($nr,rmon'collector_list_tbl,$varStat);
	return @collector_list;
}
sub rmon'get_new_collector_number
{
	local($i,$varName) = @_;
	local($lst);
	@lst = &main'list_collectors($i,$varName);
	if( @lst < 1)
	{
	 	return 1;
	}
	else
	{
	  	return $collector_list[@lst - 1]+1;
	}
}
#**********************************************************************
#* Name:	create_new_collector
#* Synopsis:	int create_new_collector(
#*			int connNr, 
#*			char *collName
#*			char *ownerString
#*			fun  *iniFun);
#
#* Description:	This fuction creates a new collector of type <collName>
#*		on <connNr>. It also sets <ownerString>.
#*		The function <iniFun> will be called during
#*		creation time so that the caller can set
#*		some collector startup values. <iniFun> will
#*		be called as:
#*
#*			&$iniFun(in connNr,char*collName,int collNr);
#*
#* Returns:	number of new collector, 0 iff error
#**********************************************************************
sub main'create_new_collector
{
	local($i,$collName,$ownerString,$iniFun) = @_;
	local($nr,$res,$varStat , $varOwn);
	
					     
	$nr = &rmon'get_new_collector_number($i,$collName);
	$varStat = $collName . "Status\[$nr\]";
	$varOwn = $collName . "Owner\[$nr\]";
	$res = &main'snmp_set($i,"$varStat=2");
	if($res != 2)
	{
		return undef;
	}
	$res = &main'snmp_get($i,$varStat);
	if($res != 3)
	{
		return undef;
	}
	$res = &main'snmp_set($i,"$varOwn=\"$ownerString\"");
	if( &$iniFun( $i, $collName, $nr ) == 0 )
	{
	    $res = &main'snmp_set($i,"$varStat=4");
	    return undef;
	}
	$res = &main'snmp_set($i,"$varStat=1");
	if($res != 1)
	{
		return undef;
	}
	return $nr;
}
#**********************************************************************
#* Name:	remove_collector
#* Synopsis:	int remove_collector(
#*			int connNr, 
#*			char *collName,
#*			int collNr;
#*			char *ownerString);
#*
#* Description:	This fuction removes a collector of type <collName>,
#*		number <collNr>
#*		on <connNr>. It checks for <ownerString>.
#* Returns:	1 iff success, 0 iff error
#**********************************************************************
sub main'remove_collector
{
	local($nr,$collName,$collNr,$ownerString) = @_;
	local($res,$own,$varStat,$varOwn);
	local($cnt);
	$varStat = $collName . "Status\[$collNr\]";
	$varOwn = $collName . "Owner\[$collNr\]";
		
	$res = undef;
	if( ! defined($res = &main'snmp_get($nr,$varOwn)))
        {
             $main'TrickletErrStr = "RMON: remove_collector\nStdspook:$main'TrickletErrStr\n";
	     return undef;
	}
	
	$own = "\"" . $ownerString . "\"";
	if( $res ne $own)
	{
	    $main'TrickletErrStr = "RMON: wrong ownerstring <$res>!=<$own>\n";
	    return undef;
	}
	$res = &main'snmp_set($nr,"$varStat=4");
	if($res != 4)
	{
	    $main'TrickletErrStr = "RMON: could not remove row\n";
	    return undef;
	}
	return 1;
}
1;




