package shh;
#########################################################################
## MODULE INFORMATION#
######################
##     FILE     NAME:       shh.pl
##     SYSTEM   NAME:       shh
##     ORIGINAL AUTHOR(S):  Jan van Oorschot
##     VERSION  NUMBER:     1.00
##     CREATION DATE:       Thu Jul 21 12:38:29 1994
##
## DESCRIPTION: A Sherlock Appliction. 
##              Controls a permanent history row in one RMON monitor
##              Is implemented as a state machine with fixed state
##              variables.
##              
##              
#########################################################################
## CHANGE INFORMATION ##
#########################
##      JvO - Jul 21, 1994:  Created
#########################################################################
require "spook.pl";
require "fsm.pl";
require "rmon/rmon.pl";
require "rmon/history.pl";

#######
# State Variables
#######
# persistent state variables
@LocalVars =
    (
     "shh'PrevTimer",		# time() the applic should be woken
     "shh'RMON",		# name of RMON monitor
     "shh'RMONIP",		# IP address of RMON monitor
     "shh'Comm",		# community string to use
     "shh'HisSource",		# history group source
     "shh'HisInterval",		# history group interval
     "shh'HisBuckets",		# history group nr. buckets
     "shh'WorkDir",		# directory to store files in
     "shh'Owner",		# current owner string used
     "shh'HisRow",		# current history row nr
     "shh'LastIDX",		# index of last row retrieved
     "shh'Dirty"		# indicates that normal
				# flow was interupted
     );

#global variables
$PF = "shh'FSM";		# ID of this module for the 
				# finit state machine engine
$shhMaxSizeDataFile = 10000;	# data files should be < 10K

#
# Definition of the Finite State Machine, to be called once !!!!!
#
sub shhInitialise
{
    # create and initialise the Finit State Machine for this module
    &main'FSM_Create($PF);

    &main'FSM_State($PF,"Init",shh'Cleanup_Environment);
    &main'FSM_Transition($PF,"Init",0,"Init2");

    &main'FSM_State($PF,"Init2",shh'Initialise_Environment);
    &main'FSM_Transition($PF,"Init2",0,"Init3");
    &main'FSM_Transition($PF,"Init2",1,"Init4");

    &main'FSM_State($PF,"Init3",shh'Normal_Finish);
    &main'FSM_Transition($PF,"Init3",0,"exit");

    &main'FSM_State($PF,"Init4",shh'Error_Finish);
    &main'FSM_Transition($PF,"Init4",0,"exit");

    &main'FSM_State($PF,"Timer",shh'Check_Environment);
    &main'FSM_Transition($PF,"Timer",0,"Timer2");
    &main'FSM_Transition($PF,"Timer",1,"exit");
    &main'FSM_Transition($PF,"Timer",2,"Init2");

    &main'FSM_State($PF,"Timer2",shh'Collect_Data);
    &main'FSM_Transition($PF,"Timer2",0,"Timer3");
    &main'FSM_Transition($PF,"Timer2",1,"Timer3");

    &main'FSM_State($PF,"Timer3",shh'Normal_Finish);
    &main'FSM_Transition($PF,"Timer3",0,"exit");

    &main'FSM_State($PF,"Timer4",shh'Terminate_Environment);
    &main'FSM_Transition($PF,"Timer4",0,"Timer5");

    &main'FSM_State($PF,"Timer5",shh'Error_Finish);
    &main'FSM_Transition($PF,"Timer5",0,"exit");

    &main'FSM_State($PF,"Trap",shh'Handle_Cold_Trap);
    &main'FSM_Transition($PF,"Trap",0,"Trap2");
    &main'FSM_Transition($PF,"Trap",1,"Trap3");
    
    &main'FSM_State($PF,"Trap2",shh'Trap_NotMe_Finish);
    &main'FSM_Transition($PF,"Trap2",0,"exit");

    &main'FSM_State($PF,"Trap3",shh'Trap_ItsMe_Finish);
    &main'FSM_Transition($PF,"Trap3",0,"Init2");

    &main'FSM_State($PF,"Exit",shh'Terminate_Environment);
    &main'FSM_Transition($PF,"Exit",0,"exit");

    &main'FSM_Resolve($PF);

    # create the Task Data Space that is used by processes running this FSM
    &main'FSM_TaskInit($PF,@LocalVars);
}

############
# Local Functions (all maintain local state variables, and return signals
############
sub Initialise_Environment
{
    # create history row
    $Owner = &main'ownerstring();
    $LastIDX = 0;
    $HisRow = &main'historyCreator(
			   1,
			   $Owner,
			   $HisSource,
			   $HisInterval,
			   $HisBuckets);
    if($HisRow == 0)
    {
	&main'SherlockLog("$ID: cant create history row");
	return 1;
    }
    # create status file
    &Write_Status_File("$WorkDir/shh.$ID.status",
		       ($RMON,$Comm,$Owner,$HisRow));
    $PrevTime = time();  # start counting now
    return 0;
}

sub Cleanup_Environment
{
    local($fn,@vars);
    # cleanup history row using status file
    $fn = "$WorkDir/shh.$ID.status";
    if( -f $fn)
    {
	@vars = &Read_Status_File("$WorkDir/shh.$ID.status");
	&main'SherlockLog("Removing stale row ($vars[3],$vars[2])");
	&main'historyTerminator(1,$vars[3],$vars[2]);
    }
    # cleanup work files
    unlink("$WorkDir/shh.$ID.status");
    unlink("$WorkDir/shh.$ID.work");
    unlink("$WorkDir/shh.$ID");
    return 0;
}

sub Check_Environment
{
    # check current history row
    $own = &main'snmp_get(1,"historyControlOwner[$HisRow]");
    if(! defined($own) )
    {
        # error during the get, determine it's cause
        if($main'TrickletSpookErr == 11)
	{
	    # time out, retry it a bit later (?)
            &main'SherlockLog("$ID: snmp timeout, waiting");
            $NextTimer = time() + 10;
            return 1;
	}
	elsif($main'TrickletSpookErr == 12)
	{
	    # wrong answer, create new row directly
            &main'SherlockLog("$ID: row corrupted, restarting");
	    return 2;
	}
        else
	{
	    # very strange, try again later
            &main'SherlockLog("$ID: strange Spook error, restarting");
            return 2;
	}
    }
    if( ("\"" . $Owner . "\"") ne $own)
    {
	# somebody else stolen our row, create new directly
        &main'SherlockLog("$ID, wrong ownerstring, restarting");
	return 2;
    }
    return 0;
}

sub Terminate_Environment
{
    local($fn);
    # remove status and work files
    if( (-f "$WorkDir/shh.$ID") && (-s "$WorkDir/shh.$ID" > 0))
    {
        $fn = &Next_Filename($WorkDir,"shh.$ID");
	rename("$WorkDir/shh.$ID",$fn);
    }
    unlink("$WorkDir/shh.$ID");
    unlink("$WorkDir/shh.$ID.status");
    unlink("$WorkDir/shh.$ID.work");
    # cleanup current history row
    &main'historyTerminator(1,$HisRow,$Owner);
    return 0;
}

sub Collect_Data
{
    local($id,$row,$idx,$strt,$drp,$oct,$utl,$pkts,$brdc,$mcst);
    local($ltim,$rtim,@rest,$otim);
    local($rt);
    local($wf) = "$WorkDir/shh.$ID.work";
    local($df) = "$WorkDir/shh.$ID";
    # interogate history rows
    if( ! (open(FIL,">$wf")))
    {
	&main'SherlockLog("$ID: Collect_Data cant create $wf");
	return 1;
    }
    $rt = &main'historyGardner(1,$HisRow,$ID, shh'FIL);
    close(FIL);
    if($rt == 0)
    {
         return 1;
    }
    # refine data
    open(FIL,$wf);
    if(! (open(OFIL,">>$df")))
    {
	&main'SherlockLog("$ID: Collect_Data cant create $df");
	unlink($wf);
	return 1;
    }
    $line = <FIL>;
    $line = <FIL>;
    ($id,$row,$ltim,$rtim,@rest) = split(' ',$line);
    $dtim = $ltim - int($rtim/100);
    $ready = 0;
    while(! $ready)
    {
	$line = <FIL>;
	($id,$row,$idx,$strt,$drp,$oct,$utl,$pkts,$brdc,$mcst)
	    = split(' ',$line);
	if($row eq 'etherHistory')
	{
	    if($idx > $LastIDX)
	    {
		$obtim = $dtim + int($strt/100);
		$oetim = $obtim + $HisInterval;
		print OFIL "$idx $obtim $oetim $drp $oct $utl $pkts $brdc $mcst\n";
		$LastIDX = $idx;
	    }
	}
	else
	{
	    $ready = 1;
	}
    }
    close(OFIL);
    # check if there's enough in the file, if so, rename it
    if((-s $df) > $shhMaxSizeDataFile)
    {
	# find a free number in the working directory
        $fn = &Next_Filename($WorkDir,"shh.$ID");
	rename($df,$fn);
    }
    unlink($wf);
    return 0;
}

sub Handle_Cold_Trap
{
    # verify that it is a cold-start trap for this application
    # Trap[3] = dutepp2.et.tudelft.nl(130.161.144.67) --> RMON agent
    # Trap[6] = coldStart(0)                          --> exact match needed
    local($trnam,$trnum) = $Trap[6] =~ /^(\S*)\((\d+)\)$/;
    if( $trnum != 0 )
    {
	return 0;
    }
    ($name,$ip) = $Trap[3] =~ /^(\S*)\((\S*)\)$/;
    if($ip ne $RMONIP)
    {
	return 0;
    }
    &main'SherlockLog("$ID: cold start, reinitialise");
    return 1;
}


sub Normal_Finish
{
    # calculate next timer
    $NextTimer = $PrevTimer + ($HisInterval * $HisBuckets);
    $PrevTimer = $NextTimer;
    $Dirty = 0;
    return 0;
}

sub Error_Finish
{
    # in an error situatie, but please try another time
    $NextTimer = $PrevTimer + ($HisInterval * $HisBuckets);
    $PrevTimer = $NextTimer;
    $Dirty = 1;
    return 0;
}

sub Fatal_Finish
{
    # fatal, remove this process from Sherlock
    $NextTimer = -1;
    $Dirty = 1;
    return 0;
}

sub Trap_NotMe_Finish
{
    # just pretend this never happened
    $NextTimer = 0;
    return 0;
}

sub Trap_ItsMe_Finish
{
    # make sure it all starts over real soon now
    return 0;
}
###########
# Utility functions
###########
sub Read_Status_File
{
    local($fn) = @_;
    if(!open(FIL,$fn))
    {
	return 0;
    }
    $line = <FIL>;
    close(FIL);
    return &main'line2lst($line);
}

sub Write_Status_File
{
    local($fn,@vars) = @_;
    local($line);
    if(!open(FIL,">$fn"))
    {
	return 0;
    }
    $line = &main'lst2line(@vars);
    print FIL $line;
    close(FIL);
    return 1;
}

sub Next_Filename
{
    local($dir,$fn) = @_;
    local($nr,$nnr,@tfiles,$tfile);
    $nr = 1;
    opendir(TDIR,$dir);
    @tfiles = grep(/^$fn\.\d+$/,readdir(TDIR));
    closedir(TDIR);
    foreach $tfile (@tfiles)
    {
	($nnr) = $tfile =~ /^$fn\.(\d+)$/;
	if( $nnr >= $nr)
	{
	    $nr = $nnr+1;
	}
    }
    return "$dir/$fn.$nr";
}

###########
# Function Entry Points of this Sherlock Application
###########
sub main'shhInit
{
    ($ID,$Arg) = @_;
    # Initialise the state variables (argument list, derived, dynamic)
    (
     $RMON,
     $Comm,
     $HisSource,
     $HisInterval,
     $HisBuckets
     ) = &main'line2lst($Arg);      #'
    $RMONIP = &main'name2ip($RMON); #'
    $WorkDir = $main'sherDir;       #'
    $Owner = "";
    $HisRow = 0;
    $LastIDX = 0;
    $Dirty = 0;
    $PrevTimer = $NextTimer = time();

    &main'snmp_community(1,$Comm);
    &main'snmp_host(1,$RMON);
    if( ! &main'snmp_connect(1))
    {
	$main'TrickletErrStr = "cant connect<$Comm,$RMON>\n$main'TrickletErrStr";
        $main'TrickletErrNo = 420;
	return -1;
    }

    if( !&main'FSM_Run($PF,"Init"))
    {
        &main'SherlockLog("FSM Run failed: $main'TrickletErrStr");
    }
    &main'FSM_TaskSwitchOut($PF,$ID);

    if( ! &main'snmp_disconnect(1))
    {
	$main'TrickletErrStr = "cant disconnect\n$main'TrickletErrStr";
        $main'TrickletErrNo = 421;
    }

    return $NextTimer;
}

sub main'shhTimer
{
    ($ID,$Tim,$Arg) = @_;
    &main'FSM_TaskSwitchIn($PF,$ID);

    &main'snmp_community(1,$Comm);
    &main'snmp_host(1,$RMON);
    if( ! &main'snmp_connect(1))
    {
	$main'TrickletErrStr = "cant connect<$Comm,$RMON>\n$main'TrickletErrStr";
        $main'TrickletErrNo = 422;
	return -1;
    }

    if($Dirty)
    {
         if( !&main'FSM_Run($PF,"Init"))
         {
             &main'SherlockLog("FSM Run failed: $main'TrickletErrStr");
         }
    }
    else
    {
         if( !&main'FSM_Run($PF,"Timer"))
         {
             &main'SherlockLog("FSM Run failed: $main'TrickletErrStr");
         }
    }
    &main'FSM_TaskSwitchOut($PF,$ID);
    if( ! &main'snmp_disconnect(1))
    {
	$main'TrickletErrStr = "cant disconnect\n$main'TrickletErrStr";
        $main'TrickletErrNo = 423;
	$NextTimer =-1;
    }
    return $NextTimer;
}

sub main'shhTrap
{
    ($ID,$Arg,@Trap) = @_;
    &main'FSM_TaskSwitchIn($PF,$ID);

    &main'snmp_community(1,$Comm);
    &main'snmp_host(1,$RMON);
    if( ! &main'snmp_connect(1))
    {
	$main'TrickletErrStr = "cant connect<$Comm,$RMON>\n$main'TrickletErrStr";
        $main'TrickletErrNo = 424;
	return -1;
    }

    if( !&main'FSM_Run($PF,"Trap"))
    {
        &main'SherlockLog("FSM Run failed: $main'TrickletErrStr");
    }
    &main'FSM_TaskSwitchOut($PF,$ID);

    if( ! &main'snmp_disconnect(1))
    {
	$main'TrickletErrStr = "cant disconnect\n$main'TrickletErrStr";
        $main'TrickletErrNo = 425;
	$NextTimer =-1;
    }

    return $NextTimer;
}

sub main'shhExit
{
    ($ID,$Arg) = @_;
    &main'FSM_TaskSwitchIn($PF,$ID);

    &main'snmp_community(1,$Comm);
    &main'snmp_host(1,$RMON);
    if( ! &main'snmp_connect(1))
    {
	$main'TrickletErrStr = "cant connect<$Comm,$RMON>\n$main'TrickletErrStr";
        $main'TrickletErrNo = 426;
	return -1;
    }
    if( !&main'FSM_Run($PF,"Exit"))
    {
        &main'SherlockLog("FSM Run failed: $main'TrickletErrStr");
    }
    &main'FSM_TaskSwitchOut($PF,$ID);

    if( ! &main'snmp_disconnect(1))
    {
	$main'TrickletErrStr = "cant disconnect\n$main'TrickletErrStr";
        $main'TrickletErrNo = 427;
	$NextTimer =-1;
    }

    return $NextTimer;
}

&shhInitialise();
1;






