# xray.pl -- a Sherlock application
#
# call with:
# probe community interface topN interval localMAC hosts...
#
# for example:
# dutepp6 secret ifIndex[1] 10 30 08:00:20:19:51:94 dutepp0 dutepp2 dutepp6 dutepp9 dutepp13
#
# Requires a ping command that understands a count option and
# an arp command that can list all entries with the Ethernet address 
# in a separate field.

package xray;

require "spook.pl";
require "fsm.pl";
require "rmon/rmon.pl";
require "rmon/statistics.pl";
require "rmon/host.pl";
require "rmon/hosttopn.pl";
require "rmon/matrix.pl";

$PINGCMD = 'ping -c 1';		# ping only once
$ARPCMD = 'arp -a';			# show all arp entries
$ARPFIELD = 3;				# field 3 is Ethernet address

#
# State variables
#

@LocalVars =
(
    "xray'ID",              # ID
    "xray'probe",           # probe name
    "xray'ip",              # probe IP address
    "xray'comm",            # community
    "xray'interface",       # interface
    "xray'topn",            # top N
    "xray'interval",        # interval
    "xray'localmac",        # local MAC address
    "xray'verbose",         # verbose
    "xray'owner",           # owner
    "xray'timer",           # timer value
    "xray'retry",           # retry count
    "xray'datadir",
    "xray'statIndex",
    "xray'statOwner",
    "xray'hostIndex",
    "xray'hostOwner",
    "xray'hostTopNIndex1",
    "xray'hostTopNIndex2",
    "xray'hostTopNIndex3",
    "xray'hostTopNIndex4",
    "xray'hostTopNIndex5",
    "xray'hostTopNIndex6",
    "xray'hostTopNIndex7",
    "xray'matrixIndex",
    "xray'matrixOwner",
);

         

#
# State machine
#

$PF = "xray'FSM";

sub Init
{
    &main'FSM_Create($PF);

    &main'FSM_State($PF,"Init",xray'ProbeInit);
    &main'FSM_Transition($PF,"Init",0,"Return");
    &main'FSM_Transition($PF,"Init",1,"Error");
    &main'FSM_Transition($PF,"Init",2,"Exit");

    &main'FSM_State($PF,"Timer",xray'ProbeTimer);
    &main'FSM_Transition($PF,"Timer",0,"Return");
    &main'FSM_Transition($PF,"Timer",1,"Restart");
    &main'FSM_Transition($PF,"Timer",2,"Exit");

    &main'FSM_State($PF,"Restart",xray'ProbeRestart);
    &main'FSM_Transition($PF,"Restart",0,"Clean");
    &main'FSM_Transition($PF,"Restart",1,"Exit");   
    
    &main'FSM_State($PF,"Clean",xray'ProbeExit);
    &main'FSM_Transition($PF,"Clean",0,"Init");
    &main'FSM_Transition($PF,"Clean",1,"Exit");   
    
    &main'FSM_State($PF,"Trap",xray'ProbeTrap);
    &main'FSM_Transition($PF,"Trap",0,"Return0");
    &main'FSM_Transition($PF,"Trap",1,"Clean");
    &main'FSM_Transition($PF,"Trap",2,"Exit");
                                        
    &main'FSM_State($PF,"Exit",xray'ProbeExit);
    &main'FSM_Transition($PF,"Exit",0,"Quit");

    &main'FSM_State($PF,"Return",xray'Return);
    &main'FSM_Transition($PF,"Return",0,"exit");

    &main'FSM_State($PF,"Return0",xray'Return0);
    &main'FSM_Transition($PF,"Return0",0,"exit");

    &main'FSM_State($PF,"Error",xray'Error);
    &main'FSM_Transition($PF,"Error",0,"exit");

    &main'FSM_State($PF,"Quit",xray'Quit);
    &main'FSM_Transition($PF,"Quit",0,"exit");

    &main'FSM_Resolve($PF);

    &main'FSM_TaskInit($PF,@LocalVars);
}

sub search_datasource_collector
{
local($conn,$group,$datasource,$owner) = @_;
local($index,$source,$own);

    foreach $index (&main'list_collectors($conn,$group))
    {                                     
        ($source,$own) = &main'snmp_get($conn,"$group"."DataSource[$index];"."$group"."Owner[$index]");
        $own = &remove_quotes($own);
        if (($own =~ /$owner/) && ($source eq $datasource))
        {
            return ($index,$own);
        }
    }
    return 0;
}

sub remove_quotes
{
local($owner) = @_;

    $own =~ s/^\"(.*)\"$/\1/;
    return $own;
}

sub ArraySwitchIn
{
local(@arrays) = @_;

	foreach $array (@arrays)
	{		 
		eval "\@$array = \@$array" . "_$ID";
	}
}
        
sub AssocArraySwitchIn
{
local(@arrays) = @_;	   

	foreach $array (@arrays)
	{
		eval "\%$array = \%$array" . "_$ID";
	}
}
        
sub ArraySwitchOut
{
local(@arrays) = @_;

	foreach $array (@arrays)
	{	 
		eval "\@$array" . "_$ID = \@$array";
	}
}
        
sub AssocArraySwitchOut
{
local(@arrays) = @_;

	foreach $array (@arrays)
	{								 
		eval "\%$array" . "_$ID = \%$array";
	}
}
        
sub ProbeInit
{          
local($index,$source,$own);
local($rc);

    $owner = &main'ownerstring();

    ($statIndex,$statOwner) = &search_datasource_collector(1,"etherStats",$interface,'^monitor');
    if ($statIndex == 0)
    {
        $statIndex = &main'etherStatsCreator(1,$owner,$interface);
        $statOwner = $owner;
    }
    ($hostIndex,$hostOwner) = &search_datasource_collector(1,"hostControl",$interface,'^monitor');
    if ($hostIndex == 0)
    {
        $hostIndex = &main'hostCreator(1,$owner,$interface);
        $hostOwner = $owner;
    }
    $hostTopNIndex1 = &main'hostTopNCreator(1,$owner,$hostIndex,1,0,$topn);
    $hostTopNIndex2 = &main'hostTopNCreator(1,$owner,$hostIndex,2,0,$topn);
    $hostTopNIndex3 = &main'hostTopNCreator(1,$owner,$hostIndex,3,0,$topn);
    $hostTopNIndex4 = &main'hostTopNCreator(1,$owner,$hostIndex,4,0,$topn);
    $hostTopNIndex5 = &main'hostTopNCreator(1,$owner,$hostIndex,5,0,$topn);
    $hostTopNIndex6 = &main'hostTopNCreator(1,$owner,$hostIndex,6,0,$topn);
    $hostTopNIndex7 = &main'hostTopNCreator(1,$owner,$hostIndex,7,0,$topn);
    ($matrixIndex,$matrixOwner) = &search_datasource_collector(1,"matrixControl",$interface,'^monitor');
    if ($matrixIndex == 0)
    {
        $matrixIndex = &main'matrixCreator(1,$owner,$interface);
        $matrixOwner = $owner;
    }
    
    if ($statIndex == 0 || 
        $hostIndex == 0 ||
        $hostTopNIndex1 == 0 ||
        $hostTopNIndex2 == 0 ||
        $hostTopNIndex3 == 0 ||
        $hostTopNIndex4 == 0 ||
        $hostTopNIndex5 == 0 ||
        $hostTopNIndex6 == 0 ||
        $hostTopNIndex7 == 0 ||
        $matrixIndex == 0)
    {
        &main'SherlockLog("$ID: can not get access to necessary rows");
        return 1;
    }

    if (($rc = &ProbeTopNRestart()) == 0)
    {          
        &TimerSet(1);   # allow 1 extra second when we are
        return 0;       # this quick after ProbeTopNRestart()
    }
    return $rc;
}

sub ProbeTimer
{
local($own);
local($rc);

    $own = &main'snmp_get(1,"etherStatsOwner[$statIndex]");
    $own = &remove_quotes($own);
    if ($statOwner ne $own)
    {
        &main'SherlockLog("$ID: wrong etherStatsOwner <$own>, restarting");
        return 1;
    }
    $own = &main'snmp_get(1,"hostControlOwner[$hostIndex]");
    $own = &remove_quotes($own);
    if ($hostOwner ne $own)
    {
        &main'SherlockLog("$ID: wrong hostControlOwner <$own>, restarting");
        return 1;
    }
    $own = &main'snmp_get(1,"hostTopNOwner[$hostTopNIndex1]");
    $own = &remove_quotes($own);
    if ($owner ne $own)
    {
        &main'SherlockLog("$ID: wrong hostTopNOwner <1,$own>, restarting");
        return 1;
    }
    $own = &main'snmp_get(1,"hostTopNOwner[$hostTopNIndex2]");
    $own = &remove_quotes($own);
    if ($owner ne $own)
    {
        &main'SherlockLog("$ID: wrong hostTopNOwner <2,$own>, restarting");
        return 1;
    }
    $own = &main'snmp_get(1,"hostTopNOwner[$hostTopNIndex3]");
    $own = &remove_quotes($own);
    if ($owner ne $own)
    {
        &main'SherlockLog("$ID: wrong hostTopNOwner <3,$own>, restarting");
        return 1;
    }
    $own = &main'snmp_get(1,"hostTopNOwner[$hostTopNIndex4]");
    $own = &remove_quotes($own);
    if ($owner ne $own)
    {
        &main'SherlockLog("$ID: wrong hostTopNOwner <4,$own>, restarting");
        return 1;
    }
    $own = &main'snmp_get(1,"hostTopNOwner[$hostTopNIndex5]");
    $own = &remove_quotes($own);
    if ($owner ne $own)
    {
        &main'SherlockLog("$ID: wrong hostTopNOwner <5,$own>, restarting");
        return 1;
    }
    $own = &main'snmp_get(1,"hostTopNOwner[$hostTopNIndex6]");
    $own = &remove_quotes($own);
    if ($owner ne $own)
    {
        &main'SherlockLog("$ID: wrong hostTopNOwner <6,$own>, restarting");
        return 1;
    }
    $own = &main'snmp_get(1,"hostTopNOwner[$hostTopNIndex7]");
    $own = &remove_quotes($own);
    if ($owner ne $own)
    {
        &main'SherlockLog("$ID: wrong hostTopNOwner <7,$own>, restarting");
        return 1;
    }
    $own = &main'snmp_get(1,"matrixControlOwner[$matrixIndex]");
    $own = &remove_quotes($own);
    if ($matrixOwner ne $own)
    {
        &main'SherlockLog("$ID: wrong matrixControlOwner <$own>, restarting");
        return 1;
    }
    
    &DataReset();
    
    if (($rc = &DataCollect()) == 0)
    {
        if (($rc = &ProbeTopNRestart()) == 0)
        {
            &TimerSet(0);
            if (($rc = &DataSave()) == 0)
            {
                return 0;
            }
        }
    }

    &DataReset();
    
    return $rc;
}
                   
sub DataCollect
{    
    &main'SherlockLog("$ID: start collecting") if ($verbose);
    
    ($SysUpTime) = &main'snmp_get(1,"sysUpTime");
    
    (
        $DropEvents,$Octets,$Pkts,$BroadcastPkts,$MulticastPkts,
        $CRCAlignErrors,$UndersizePkts,$OversizePkts,$Fragments,$Jabbers,$Collisions,
        $Pkts64,$Pkts65,$Pkts128,$Pkts256,$Pkts512,$Pkts1024
    ) = 
        &main'snmp_get(1,"etherStatsDropEvents[$statIndex];".
                         "etherStatsOctets[$statIndex];".
                         "etherStatsPkts[$statIndex];".
                         "etherStatsBroadcastPkts[$statIndex];".
                         "etherStatsMulticastPkts[$statIndex];".
                         "etherStatsCRCAlignErrors[$statIndex];".
                         "etherStatsUndersizePkts[$statIndex];".
                         "etherStatsOversizePkts[$statIndex];".
                         "etherStatsFragments[$statIndex];".
                         "etherStatsJabbers[$statIndex];".
                         "etherStatsCollisions[$statIndex];".
                         "etherStatsPkts64Octets[$statIndex];".
                         "etherStatsPkts65to127Octets[$statIndex];".
                         "etherStatsPkts128to255Octets[$statIndex];".
                         "etherStatsPkts256to511Octets[$statIndex];".
                         "etherStatsPkts512to1023Octets[$statIndex];".
                         "etherStatsPkts1024to1518Octets[$statIndex]");

    while (($mac,$macobjid) = each %macobjids)
    {
        (
            $HostInPkts{$mac},$HostOutPkts{$mac},
            $HostInOctets{$mac},$HostOutOctets{$mac},
            $HostOutErrors{$mac},
            $HostOutBroadcastPkts{$mac},
            $HostOutMulticastPkts{$mac}
        ) = 
            &main'snmp_get(1,"hostInPkts[$hostIndex.$macobjid];".
                             "hostOutPkts[$hostIndex.$macobjid];".
                             "hostInOctets[$hostIndex.$macobjid];".
                             "hostOutOctets[$hostIndex.$macobjid];".
                             "hostOutErrors[$hostIndex.$macobjid];".
                             "hostOutBroadcastPkts[$hostIndex.$macobjid];".
                             "hostOutMulticastPkts[$hostIndex.$macobjid]");
        if (undef == &main'snmp_tbl_row(1,xray'ProbeMatrixSDTable, $mac,
                            "matrixSDDestAddress[$matrixIndex.$macobjid];".
                            "matrixSDPkts[$matrixIndex.$macobjid];".
                            "matrixSDOctets[$matrixIndex.$macobjid];".
                            "matrixSDErrors[$matrixIndex.$macobjid]"))
        {
            return 1;
        }
        if (undef == &main'snmp_tbl_row(1,xray'ProbeMatrixDSTable, $mac,
                            "matrixDSSourceAddress[$matrixIndex.$macobjid];".
                            "matrixDSPkts[$matrixIndex.$macobjid];".
                            "matrixDSOctets[$matrixIndex.$macobjid];".
                            "matrixDSErrors[$matrixIndex.$macobjid]"))
        {
            return 1;
        }
        
        # reset entries so they at least exist
        $TopNAddressIndex1{$mac} = 0;
        $TopNAddressRate1{$mac} = 0;
        $TopNAddressIndex2{$mac} = 0;
        $TopNAddressRate2{$mac} = 0;
        $TopNAddressIndex3{$mac} = 0;
        $TopNAddressRate3{$mac} = 0;
        $TopNAddressIndex4{$mac} = 0;
        $TopNAddressRate4{$mac} = 0;
        $TopNAddressIndex5{$mac} = 0;
        $TopNAddressRate5{$mac} = 0;
        $TopNAddressIndex6{$mac} = 0;
        $TopNAddressRate6{$mac} = 0;
        $TopNAddressIndex7{$mac} = 0;
        $TopNAddressRate7{$mac} = 0;
    }
    
    if (undef == &main'snmp_tbl_row(1,xray'ProbeTopNTable, 1,
                        "hostTopNIndex[$hostTopNIndex1];hostTopNAddress[$hostTopNIndex1];hostTopNRate[$hostTopNIndex1]"))
    {
        return 1;
    }
    if (undef == &main'snmp_tbl_row(1,xray'ProbeTopNTable, 2,
                        "hostTopNIndex[$hostTopNIndex2];hostTopNAddress[$hostTopNIndex2];hostTopNRate[$hostTopNIndex2]"))
    {
        return 1;
    }
    if (undef == &main'snmp_tbl_row(1,xray'ProbeTopNTable, 3,
                        "hostTopNIndex[$hostTopNIndex3];hostTopNAddress[$hostTopNIndex3];hostTopNRate[$hostTopNIndex3]"))
    {
        return 1;
    }
    if (undef == &main'snmp_tbl_row(1,xray'ProbeTopNTable, 4,
                        "hostTopNIndex[$hostTopNIndex4];hostTopNAddress[$hostTopNIndex4];hostTopNRate[$hostTopNIndex4]"))
    {
        return 1;
    }
    if (undef == &main'snmp_tbl_row(1,xray'ProbeTopNTable, 5,
                        "hostTopNIndex[$hostTopNIndex5];hostTopNAddress[$hostTopNIndex5];hostTopNRate[$hostTopNIndex5]"))
    {
        return 1;
    }
    if (undef == &main'snmp_tbl_row(1,xray'ProbeTopNTable, 6,
                        "hostTopNIndex[$hostTopNIndex6];hostTopNAddress[$hostTopNIndex6];hostTopNRate[$hostTopNIndex6]"))
    {
        return 1;
    }
    if (undef == &main'snmp_tbl_row(1,xray'ProbeTopNTable, 7,
                        "hostTopNIndex[$hostTopNIndex7];hostTopNAddress[$hostTopNIndex7];hostTopNRate[$hostTopNIndex7]"))
    {
        return 1;
    }
    return 0;
}

sub ProbeMatrixSDTable
{
local($ether,$row,$dest,$pkts,$octets,$errors) = @_;

    $MatrixSDConns{$ether} = $row;
    $MatrixSDDestination{$ether.$row} = $dest;
    $MatrixSDPkts{$ether.$row} = $pkts;
    $MatrixSDOctets{$ether.$row} = $octets;
    $MatrixSDErrors{$ether.$row} = $errors;
}

sub ProbeMatrixDSTable
{
local($ether,$row,$source,$pkts,$octets,$errors) = @_;
                                                     
    $MatrixDSConns{$ether} = $row;
    $MatrixDSSource{$ether.$row} = $source;
    $MatrixDSPkts{$ether.$row} = $pkts;
    $MatrixDSOctets{$ether.$row} = $octets;
    $MatrixDSErrors{$ether.$row} = $errors;
}

sub ProbeTopNTable
{
local($type,$row,$index,$address,$rate) = @_;

    eval "\$TopNAddressIndex$type{\$address} = \$index;";
    eval "\$TopNAddress$type[\$index] = \$address;";
    eval "\$TopNAddressRate$type{\$address} = \$rate;";
    eval "\$TopNRate$type[\$index] = \$rate;";
}

sub ProbeTopNRestart
{
    &main'SherlockLog("$ID: restarting topNs") if ($verbose);
    
    if (undef == &main'snmp_set(1,"hostTopNTimeRemaining[$hostTopNIndex1]=$interval;".
                       "hostTopNTimeRemaining[$hostTopNIndex2]=$interval;".
                       "hostTopNTimeRemaining[$hostTopNIndex3]=$interval;".
                       "hostTopNTimeRemaining[$hostTopNIndex4]=$interval;".
                       "hostTopNTimeRemaining[$hostTopNIndex5]=$interval;".
                       "hostTopNTimeRemaining[$hostTopNIndex6]=$interval;".
                       "hostTopNTimeRemaining[$hostTopNIndex7]=$interval"))
    {
        &main'SherlockLog("$ID: can not restart topNs");
        return 1;
    }            
    return 0;           
}

sub DataSave
{                  
local($sec,$min,$hour,$mday,$mon,$year) = localtime();
local($timestamp);
local($host,$mac,$index,$address);

    $year += 1900 if $year >= 70;                 
    $year += 2000 if $year < 70;                 
    $timestamp = sprintf("%4d-%02d-%02d.%02d:%02d:%02d", $year,$mon+1,$mday,$hour,$min,$sec);
    $fn = "$datadir/xray.sample.$ID.$timestamp";
    
    &main'SherlockLog("$ID: saving data in $fn") if ($verbose);
    
    if (!open(DATA,">$fn"))
    {
        &main'SherlockLog("$ID: can not open data file");
    }
    else
    {
        print DATA "$timestamp: xray $probe $comm $interface $topn $interval $localmac @hosts\n";
        print DATA "sysUpTime: $SysUpTime\n";
        print DATA "etherStats: $DropEvents $Octets $Pkts ".
                   "$BroadcastPkts $MulticastPkts ".
                   "$CRCAlignErrors ".
                   "$UndersizePkts $OversizePkts ".
                   "$Fragments $Jabbers $Collisions ".
                   "$Pkts64 $Pkts65 $Pkts128 $Pkts256 $Pkts512 $Pkts1024\n";

        print DATA "topN1:\n";
        foreach $index (1..$#TopNAddress1)
        {
            print DATA "$TopNAddress1[$index] $index $TopNRate1[$index]\n";
        }
        print DATA "topN2:\n";
        foreach $index (1..$#TopNAddress2)
        {
            print DATA "$TopNAddress2[$index] $index $TopNRate2[$index]\n";
        }
        print DATA "topN3:\n";
        foreach $index (1..$#TopNAddress3)
        {
            print DATA "$TopNAddress3[$index] $index $TopNRate3[$index]\n";
        }
        print DATA "topN4:\n";
        foreach $index (1..$#TopNAddress4)
        {
            print DATA "$TopNAddress4[$index] $index $TopNRate4[$index]\n";
        }
        print DATA "topN5:\n";
        foreach $index (1..$#TopNAddress5)
        {
            print DATA "$TopNAddress5[$index] $index $TopNRate5[$index]\n";
        }
        print DATA "topN6:\n";
        foreach $index (1..$#TopNAddress6)
        {
            print DATA "$TopNAddress6[$index] $index $TopNRate6[$index]\n";
        }
        print DATA "topN7:\n";
        foreach $index (1..$#TopNAddress7)
        {
            print DATA "$TopNAddress7[$index] $index $TopNRate7[$index]\n";
        }

        foreach $host (@hosts)
        {
            $mac = $macs{$host};

            print DATA "address: $host $mac\n";
            print DATA "host: ".
                       "$HostInPkts{$mac} ".
                       "$HostOutPkts{$mac} ".
                       "$HostInOctets{$mac} ".
                       "$HostOutOctets{$mac} ".
                       "$HostOutErrors{$mac} ".
                       "$HostOutBroadcastPkts{$mac} ".
                       "$HostOutMulticastPkts{$mac}\n";
                               
            print DATA "topN: $TopNAddressIndex1{$mac} $TopNAddressRate1{$mac}; ".
                       "$TopNAddressIndex2{$mac} $TopNAddressRate2{$mac}; ".
                       "$TopNAddressIndex3{$mac} $TopNAddressRate3{$mac}; ".
                       "$TopNAddressIndex4{$mac} $TopNAddressRate4{$mac}; ".
                       "$TopNAddressIndex5{$mac} $TopNAddressRate5{$mac}; ".
                       "$TopNAddressIndex6{$mac} $TopNAddressRate6{$mac}; ".
                       "$TopNAddressIndex7{$mac} $TopNAddressRate7{$mac}\n";

            print DATA "matrixSD:\n";
            foreach $conn (1..$MatrixSDConns{$mac})
            {
                print DATA "$MatrixSDDestination{$mac.$conn} ".
                           "$MatrixSDPkts{$mac.$conn} ".
                           "$MatrixSDOctets{$mac.$conn} ".
                           "$MatrixSDErrors{$mac.$conn}\n";
            }
            print DATA "matrixDS:\n";
            foreach $conn (1..$MatrixDSConns{$mac})
            {
                print DATA "$MatrixDSSource{$mac.$conn} ".
                           "$MatrixDSPkts{$mac.$conn} ".
                           "$MatrixDSOctets{$mac.$conn} ".
                           "$MatrixDSErrors{$mac.$conn}\n";
            }
        }
    }
    
    close(DATA);
    
    return 0;
}

sub DataReset
{
    undef $SysUpTime;

    undef $DropEvents;
    undef $Octets;
    undef $Pkts;
    undef $BroadcastPkts;
    undef $MulticastPkts;
    undef $CRCAlignErrors;
    undef $UndersizePkts;
    undef $OversizePkts;
    undef $Fragments;
    undef $Jabbers;
    undef $Collisions;
    undef $Pkts64;
    undef $Pkts65;
    undef $Pkts128;
    undef $Pkts256;
    undef $Pkts512;
    undef $Pkts1024;
    
    undef %HostInPkts;
    undef %HostOutPkts;
    undef %HostInOctets;
    undef %HostOutOctets;
    undef %HostOutErrors;
    undef %HostOutBroadcasts;
    undef %HostOutMulticasts;

    undef %MatrixSDConns;
    undef %MatrixSDDestination;
    undef %MatrixSDPkts;
    undef %MatrixSDOctets;
    undef %MatrixSDErrors;
    undef %MatrixDSConns;
    undef %MatrixDSSource;
    undef %MatrixDSPkts;
    undef %MatrixDSOctets;
    undef %MatrixDSErrors;

    undef %TopNAddressIndex1;
    undef @TopNAddress1;
    undef %TopNAddressRate1;
    undef @TopNRate1;

    undef %TopNAddressIndex2;
    undef @TopNAddress2;
    undef %TopNAddressRate2;
    undef @TopNRate2;

    undef %TopNAddressIndex3;
    undef @TopNAddress3;
    undef %TopNAddressRate3;
    undef @TopNRate3;

    undef %TopNAddressIndex4;
    undef @TopNAddress4;
    undef %TopNAddressRate4;
    undef @TopNRate4;

    undef %TopNAddressIndex5;
    undef @TopNAddress5;
    undef %TopNAddressRate5;
    undef @TopNRate5;

    undef %TopNAddressIndex6;
    undef @TopNAddress6;
    undef %TopNAddressRate6;
    undef @TopNRate6;

    undef %TopNAddressIndex7;
    undef @TopNAddress7;
    undef %TopNAddressRate7;
    undef @TopNRate7;
}
    
sub ProbeRestart
{
    if ($retry <= 0)
    {
        return 1;
    }
    $retry--;
    return 0;
}

sub ProbeTrap
{
    # verify that it is a coldStart trap for this application
    # Trap[3] = host.domain.country(a.b.c.d) --> RMON agent
    # Trap[6] = coldStart(0)                 --> exact match needed
    local($trnam,$trnum) = $Trap[6] =~ /^(\S*)\((\d+)\)$/;
    &main'SherlockLog("$ID: cold start received, checking") if ($verbose);
    if ($trnum != 0)
    {
        return 0;
    }
    ($name,$addr) = $Trap[3] =~ /^(\S*)\((\S*)\)$/;
    if ($addr ne $ip)
    {
        return 0;
    }
    &main'SherlockLog("$ID: cold start, initialize");
    return 1;
}

sub ProbeExit
{
    if ($statIndex != 0 && ($statOwner eq $owner))
    {
        &main'etherStatsTerminator(1,$statIndex,$statOwner);
    }
    if ($hostIndex != 0 && ($hostOwner eq $owner))
    {
        &main'hostTerminator(1,$hostIndex,$hostOwner);
    }
    if ($hostTopNIndex1 != 0)
    {
        &main'hostTopNTerminator(1,$hostTopNIndex1,$owner);
    }
    if ($hostTopNIndex2 != 0)
    {
        &main'hostTopNTerminator(1,$hostTopNIndex2,$owner);
    }
    if ($hostTopNIndex3 != 0)
    {
        &main'hostTopNTerminator(1,$hostTopNIndex3,$owner);
    }
    if ($hostTopNIndex4 != 0)
    {
        &main'hostTopNTerminator(1,$hostTopNIndex4,$owner);
    }
    if ($hostTopNIndex5 != 0)
    {
        &main'hostTopNTerminator(1,$hostTopNIndex5,$owner);
    }
    if ($hostTopNIndex6 != 0)
    {
        &main'hostTopNTerminator(1,$hostTopNIndex6,$owner);
    }
    if ($hostTopNIndex7 != 0)
    {
        &main'hostTopNTerminator(1,$hostTopNIndex7,$owner);
    }
    if ($matrixIndex != 0 && ($matrixOwner eq $owner))
    {
        &main'matrixTerminator(1,$matrixIndex,$matrixOwner);
    }
    return 0;
}

sub TimerSet
{           
local($offset) = @_;

    $timer = time()+$offset;
}

sub Return
{
    # calculate next timer
    &main'SherlockLog("$ID: return") if ($verbose);
    $timer = $timer + $interval;
    return 0;
}

sub Return0
{
    &main'SherlockLog("$ID: returning with timer unchanged") if ($verbose);
    $timer = 0;
    return 0;
}

sub Error
{
    # error, try another time
    &main'SherlockLog("$ID: error return, trying again later");
    $timer = $timer + $interval;
    return 0;
}

sub Quit
{
    # quit process
    &main'SherlockLog("$ID: quit");
    $timer = 0;
    return 0;
}


  
#
# Utility functions
#

sub macnormalize
{
local($mac) = @_;   
                                    
    return join(':',grep(((length($_) < 2) && ($_ = "0".$_)) || 1,split(/:/,$mac)));
}

sub mac2objid
{             
local($mac) = @_;                                 
local($objid,$len);
                                    
    $objid = join('.',grep((($_ = hex($_)) || 1) && ++$len,split(/:/,$mac)));
    return "$len.$objid";
}

sub hostnormalize
{
local(@hosts) = @_;
                                 
    grep((($_) = gethostbyname($_)) || 1, @hosts);
    return @hosts;
}

sub getmacs
{           
local(@cmds) = @_;
local(@hosts) = @_;
local(@cmd,$host,%macs,$mac);
local($line);

    grep(($_ = "$PINGCMD $_") || 1, @cmds);
    $cmd = join(";",@cmds);
    `$cmd`;
    
    @arp = split(/^/,`$ARPCMD`);
    foreach $host (@hosts)
    {                         
        @line = grep(/$host/, @arp);
        @line = split(/[ \t\n]+/,$line[0]);
        if ($#line < $ARPFIELD)
        {
            $mac = $localmac;
        }
		else
		{
        	$mac = $line[$ARPFIELD];
		}
        $mac = &macnormalize($mac);
        $macs{$host} = $mac;
    }
    return %macs;
}

sub macs2objid
{        
local(@macs) = @_;
local(%macobjids);
local($mac);

    foreach $mac (@macs)
    {
        $macobjids{$mac} = &mac2objid($mac);
    }
    return %macobjids;
}



#
# Sherlock entry points
#

sub main'xrayInit
{                       
local($host,$mac);

    ($ID,$Argv) = @_;
	
	$verbose = 0;
    $interface = 1;
    $topn = 10;
    $interval = 30;
    (
        $probe,
        $comm,
        $interface,
        $topn,
        $interval,
		$localmac,
        @hosts
    ) = &main'line2lst($Argv);
    $ip = &main'name2ip($probe);
    $owner = "";
    $statIndex = 0;
    $statOwner = "";
    $hostIndex = 0;
    $hostOwner = "";
    $hostTopNIndex1 = 0;
    $hostTopNIndex2 = 0;
    $hostTopNIndex3 = 0;
    $hostTopNIndex4 = 0;
    $hostTopNIndex5 = 0;
    $hostTopNIndex6 = 0;
    $hostTopNIndex7 = 0;
    $matrixIndex = 0;
    $matrixOwner = "";
    $timer = 0;
    $retry = 10;
              
    $datadir = $main'sherDir;
    
    @hosts = &hostnormalize(@hosts);
    %macs = &getmacs(@hosts);
    if (!defined(%macs))       
    {
        &main'SherlockLog("$ID: can not resolve all MAC addresses");
        return -1;
    }
    %macobjids = &macs2objid(values(%macs));
    
    &TimerSet(0);
    
    &main'snmp_host(1,$probe);
    &main'snmp_community(1,$comm);
    if (!&main'snmp_connect(1))
    {
        $main'TrickletErrStr = "can not connect <$probe,$comm>\n$main'TrickletErrStr";
        $main'TrickletErrNo = 420;
        return -1;
    }

    if (!&main'FSM_Run($PF,"Init"))
    {
        &main'SherlockLog("$ID: FSM Init run failed: $main'TrickletErrStr");
        return -1;
    }
    
    &main'FSM_TaskSwitchOut($PF,$ID);
	&ArraySwitchOut(('hosts'));
	&AssocArraySwitchOut(('macs', 'macobjids'));

    if (!&main'snmp_disconnect(1))
    {
        $main'TrickletErrStr = "can not disconnect <$probe,$comm>\n$main'TrickletErrStr";
        $main'TrickletErrNo = 421;
        return -1;
    }

    return $timer;
}

sub main'xrayTimer
{
    ($ID,$Timer,$Argv) = @_;

    &main'FSM_TaskSwitchIn($PF,$ID);
	&ArraySwitchIn(('hosts'));
	&AssocArraySwitchIn(('macs', 'macobjids'));

    &main'snmp_host(1,$probe);
    &main'snmp_community(1,$comm);
    if (!&main'snmp_connect(1))
    {
        $main'TrickletErrStr = "can not connect <$probe,$comm>\n$main'TrickletErrStr";
        $main'TrickletErrNo = 422;
        return -1;
    }

    if (!&main'FSM_Run($PF,"Timer"))
    {
        &main'SherlockLog("$ID: FSM Timer run failed: $main'TrickletErrStr");
        return -1;
    }
    
    &main'FSM_TaskSwitchOut($PF,$ID);
	&ArraySwitchOut(('hosts'));
	&AssocArraySwitchOut(('macs', 'macobjids'));
    
    if (!&main'snmp_disconnect(1))
    {
        $main'TrickletErrStr = "can not disconnect\n$main'TrickletErrStr";
        $main'TrickletErrNo = 423;
        return -1;
    }
    
    return $timer;
}

sub main'xrayTrap
{
    ($ID,$Argv,@Trap) = @_;

    &main'FSM_TaskSwitchIn($PF,$ID);
	&ArraySwitchIn(('hosts'));
	&AssocArraySwitchIn(('macs', 'macobjids'));

    &main'snmp_host(1,$probe);
    &main'snmp_community(1,$comm);
    if (! &main'snmp_connect(1))
    {
        $main'TrickletErrStr = "can not connect <$probe,$comm>\n$main'TrickletErrStr";
        $main'TrickletErrNo = 424;
        return -1;
    }

    if (!&main'FSM_Run($PF,"Trap"))
    {
        &main'SherlockLog("$ID: FSM Trap run failed: $main'TrickletErrStr");
        return -1;
    }

    &main'FSM_TaskSwitchOut($PF,$ID);
	&ArraySwitchOut(('hosts'));
	&AssocArraySwitchOut(('macs', 'macobjids'));

    if (! &main'snmp_disconnect(1))
    {
        $main'TrickletErrStr = "can not disconnect\n$main'TrickletErrStr";
        $main'TrickletErrNo = 425;
        return -1;
    }

    return $timer;
}

sub main'xrayExit
{
    ($ID,$Argv) = @_;

    &main'FSM_TaskSwitchIn($PF,$ID);
	&ArraySwitchIn(('hosts'));
	&AssocArraySwitchIn(('macs', 'macobjids'));

    &main'snmp_host(1,$probe);
    &main'snmp_community(1,$comm);
    if (!&main'snmp_connect(1))
    {
        $main'TrickletErrStr = "can not connect <$probe,$comm>\n$main'TrickletErrStr";
        $main'TrickletErrNo = 426;
        return -1;
    }

    if (!&main'FSM_Run($PF,"Exit"))
    {
        &main'SherlockLog("$ID: FSM Exit run failed: $main'TrickletErrStr");
        return -1;
    }

    &main'FSM_TaskSwitchOut($PF,$ID);
	&ArraySwitchOut(('hosts'));
	&AssocArraySwitchOut(('macs', 'macobjids'));

    if (!&main'snmp_disconnect(1))
    {
        $main'TrickletErrStr = "can not disconnect\n$main'TrickletErrStr";
        $main'TrickletErrNo = 427;
        return -1;
    }

    return $timer;
}

&Init();

# &main'xrayInit("dutepp6", "dutepp6 secret ifIndex[1] 10 30 08:00:20:19:51:94 dutepp0 dutepp2 dutepp6 dutepp9 dutepp13");

1;
