#!/usr/local/bin/perl		--*- perl -*--
package spook;
#**********************************************************************
#* NAME:	spook.pl
#* Description:	This module contains the subroutines
#*		to start and interrogate the spook
#*		snmp requestor.
#**********************************************************************

#**********************************************************************
#* Name:	snmp_open
#* Synopsis:	void snmp_open( int nr )
#* Description:	This fuction creates snmp connection <nr>.
#* Returns:	none
#**********************************************************************
sub main'snmp_open
{
	local($nr) = @_;
	local($resp,$w1,$r1,$w2,$r2);
	$w1 = "wh10" . $nr;
	$r1 = "rh10" . $nr;
	$w2 = "wh20" . $nr;
	$r2 = "rh20" . $nr;
	pipe($r1,$w1) || die "can't pipe";
	pipe($r2,$w2) || die "can't pipe";
	if ($pid = fork())
	{
		# parent code  
		close($r1);		# parent should
		close($w2);		# parent should 
		select($w1); $| = 1;	# make unbuffered
		select(STDOUT); $|=1;
		sleep(1);		# give daemon time to start
		return 1;
	} 
	elsif (defined $pid)
	{
	    # child code
	    close($r2);
	    close($w1);
	    select($w2); $| = 1;
	    open(STDOUT,">&$w2") || die "can't redirect STDOUT";
	    open(STDIN,"<&$r1")  || die "can't redirect STDIN";
	    select(STDOUT);	$| = 1;
	    if( !(exec 'stdspook'))
	    {
		$TrickletErrStr =  "Spook: can't exec stdspook\nPerl: $!\n";
		$TrickletErrNo  = 1;
		return undef;
	    }
	}
	elsif ($! =~ /o more process/)
	{
		# nor more processes, try again later
		sleep 5;
		redo FORK;
	}
	else
	{
	    $TrickletErrStr =  "Spook: can't fork stdspook\nPerl: $!\n";
	    $TrickletErrNo  = 2;
	    return undef;
	}
}

#**********************************************************************
#* Name:	snmp_close
#* Synopsis:	void snmp_close( int nr )
#* Description:	This fuction closes snmp connection <nr> that
#*		was opened using &snmp_open.
#* Returns:	none
#**********************************************************************
sub main'snmp_close
{
	local($nr) = @_;
	local($w1,$r2,$resp);
	
	$w1 = "wh10" . $nr;
	$r2 = "rh20" . $nr;
	print $w1 ".\n";
	$resp = <$r2>;
	close($w1);
	close($r2);
}
#**********************************************************************
#* Name:	snmp_connect
#* Synopsis:	void snmp_connect( int nr )
#* Description:	This fuction connects Spook handle <nr>
#*              to the host that has been defined for this handle.
#* Returns:	none
#**********************************************************************
sub main'snmp_connect
{
	local($nr) = @_;
	local($w1,$r2,$resp);
	
	$w1 = "wh10" . $nr;
	$r2 = "rh20" . $nr;
	print $w1 "[\n";
	$resp = <$r2>;
	if(substr($resp,0,2) ne "[+")  
	{ 
	    chop $resp;
	    $main'TrickletErrNo = 3;
            $main'TrickletErrStr = "Spook:snmp_connect failed\nStdspook: $resp\n";
	    return undef; 
	};
	return 1;
}
#**********************************************************************
#* Name:	snmp_disconnect
#* Synopsis:	boolean snmp_disconnect( int nr )
#* Description:	This fuction disconnects Spook handle <nr>
#*              from its current host.
#* Returns:	none
#**********************************************************************
sub main'snmp_disconnect
{
	local($nr) = @_;
	local($w1,$r2,$resp);
	
	$w1 = "wh10" . $nr;
	$r2 = "rh20" . $nr;
	print $w1 "]\n";
	$resp = <$r2>;
	if(substr($resp,0,2) ne "]+")  
	{
	    chop $resp;
	    $main'TrickletErrNo = 4;
            $main'TrickletErrStr = "Spook:snmp_disconnect failed\nStdspook: $resp\n";
	    return undef; 
	};
	return 1;
}
#**********************************************************************
#* Name:	snmp_community
#* Synopsis:	void snmp_community( int nr , char *comm)
#* Description:	This fuction sets the community string for 
#*		snmp connection <nr>.
#* Returns:	none
#**********************************************************************
sub main'snmp_community
{
	local($nr,$comm) = @_;
	local($w1,$r2,$resp);
	
	$w1 = "wh10" . $nr;
	$r2 = "rh20" . $nr;
	print $w1 "=community $comm\n";
	$resp = <$r2>;
	if(substr($resp,0,2) ne "=+")  
	{ 
	    $main'TrickletErrNo = 5;
            $main'TrickletErrStr = "Spook:snmp_community failed\nStdspook: $resp\n";
	    return undef;
	};
	return 1;
}
#**********************************************************************
#* Name:	snmp_host
#* Synopsis:	void snmp_host( int nr , char *host )
#* Description:	This fuction sets the host for 
#*		snmp connection <nr>.
#* Returns:	none
#**********************************************************************
sub main'snmp_host
{
	local($nr,$host) = @_;
	local($w1,$r2,$resp);
	
	$w1 = "wh10" . $nr;
	$r2 = "rh20" . $nr;
	print $w1 "=host $host\n";
	$resp = <$r2>;
	if(substr($resp,0,2) ne "=+")  
	{ 
	    $main'TrickletErrNo = 6;
            $main'TrickletErrStr = "Spook:snmp_host failed\nStdSpook: $resp\n";
	    return undef; 
	};
	return 1;
}
#**********************************************************************
#* Name:	snmp_medium
#* Synopsis:	void snmp_medium( int nr , char *medium )
#* Description:	This fuction sets the medium for 
#*		snmp connection <nr>.
#* Returns:	none
#**********************************************************************
sub main'snmp_medium
{
	local($nr,$medium) = @_;
	local($w1,$r2,$resp);
	
	$w1 = "wh10" . $nr;
	$r2 = "rh20" . $nr;
	print $w1 "=medium $medium\n";
	$resp = <$r2>;
	if(substr($resp,0,2) ne "=+")  
	{ 
	    $main'TrickletErrNo = 7;
            $main'TrickletErrStr = "Spook:snmp_medium failed\nStdspook: $resp\n";
	    return undef; 
	};
	return 1;
}
#**********************************************************************
#* Name:	snmp_commandline_open
#* Synopsis:	list = int snmp_commandline_open(int nr, int larg,ARGV)
#
#* Description:	This fuction return a list of the active
#*		collectors of the type <collName> on <connNr>.
#* Returns:	list of integers
#**********************************************************************
sub main'snmp_commandline_open
{
    local($nr,$larg,@varg) = @_;
    local($len,$iniHost,$iniComm,$iniMed,@rest);
    $len = @varg;
    if( $larg > $len)
    {
	# not enough arguments
	$main'TrickletErrNo = 8;
        $main'TrickletErrStr = "Spook:snmp_commandline_open, not enough args\n";
	return -1;
    }
    @stdargs = splice(@varg,$larg);
    ($iniHost,$iniComm,$iniMed,@rest) = @stdargs;
    
    if(!&main'snmp_open($nr))
    {
	$main'TrickletErrNo = 9;
        $main'TrickletErrStr = "Spook:snmp_commandline_open, snmp_open failed\n$main'TrickletErrStr";
	return undef;
    }
    if($iniHost ne "") 
    { 
	&main'snmp_host($nr,$iniHost);
    }
    if($iniMed ne "")  
    {
	&main'snmp_medium($nr,$iniMed);
    }
    if(!&main'snmp_connect($nr))
    {
	$main'TrickletErrNo = 10;
        $main'TrickletErrStr = "spook:snmp_commandline_open, snmp_connect failed\n$main'TrickletErrStr";
	return undef;
    }
    if($iniComm ne "") 
    {
	&main'snmp_community($nr,$iniComm);
    }
    $main'iniHost = $iniHost;
    $main'iniComm = $iniComm;
    $main'iniMed  = $iniMed;
    return 1;
}
#**********************************************************************
#* Name:	snmp_get
#* Synopsis:	void snmp_get( int nr, list *value)
#* Description:	This fuction does one SNMP get command over connection
#*		<nr>. This command will contain the MIB varaibles
#*		given in <list>.
#*		if $value is empty, the last SNMP request
#*		will be repeated. This doesn't have to
#*		be the last get request, but can also be the
#*		last set request.
#* Returns:	perl value
#**********************************************************************
sub main'snmp_get
{
	local($nr,@list) = @_;
	local($req,$ans,@rlst,$var,@ret);
	
	if($#list == -1)
	{
		$req = '';
	}
	else
	{
		if($#list != 0) 
		{
		    $req = join(";",@list);
		}
		else
		{
		    $req = $list[0];
		}
	}
	if( ! defined(($ans = &snmp_raw_get($nr,$req))))
	{
	    $main'TrickletErrNo = 11;
            ( 
                $main'TrickletSpookErr,
                $main'TrickletSNMPErr,
                $main'TrickletSNMPVar
            ) = $main'TrickletErrStr =~ /^.- <(\d+),(\d+),(\d+)>/;
            $main'TrickletErrStr = "Spook:snmp_get, get failed\nStdspook: <$main'TrickletErrStr>\n";
	     return undef;
	}
#	print("DEBUG: snmp_raw_get returned <$nr><$req><$ans><$main'TrickletErrStr>\n");
	@ret = ();
	foreach $var ( &snmp_rsp2lst($ans))
	{
		@rlst = &snmp_var2lst($var);
		push(@ret,$rlst[2]);
	}
	return @ret;
}
#**********************************************************************
#* Name:	snmp_getnext
#* Synopsis:	void snmp_getnext( int nr, list *value )
#* Description:	This fuction does one SNMP getnext command over connection
#*		<nr>. This command will contain the MIB varaibles
#*		given in <list>.
#*		if $value is empty, the last SNMP request
#*		will be repeated. This doesn't have to
#*		be the last get request, but can also be the
#*		last set request.
#* Returns:	list of triples (name,index,value)
#**********************************************************************
sub main'snmp_getnext
{
	local($nr,@list) = @_;
	local($req,$ans,@rlst,$var,@ret);
	
	if($#list == -1)
	{
		$req = '';
	}
	else
	{
		if($#list != 0) 
			{$req = join(";",@list);}
		else
			{$req = $list[0];}		   
	}
	if( ! defined(($ans = &snmp_raw_getnext($nr,$req))))
	{
	    $main'TrickletErrNo = 12;
            ( 
                $main'TrickletSpookErr,
                $main'TrickletSNMPErr,
                $main'TrickletSNMPVar
            ) = $main'TrickletErrStr =~ /^.- <(\d+),(\d+),(\d+)>/;
            $main'TrickletErrStr = "Spook:snmp_getnext, getnext failed\nStdspook: $main'TrickletErrStr\n";
	    return undef;
	}
	@ret = ();
	foreach $var ( &snmp_rsp2lst($ans))
	{
		@rlst = &snmp_var2lst($var);
		if(defined($rlst[0]))
		{
		    push(@ret,@rlst);
		}
	}
	return @ret;
}
#**********************************************************************
#* Name:	snmp_set
#* Synopsis:	void snmp_set( int nr, list *value )
#* Description:	This fuction does one SNMP set command over connection
#*		<nr>. This command will contain the MIB varaibles
#*		given in <list>.
#* Returns:	perl value
#**********************************************************************
sub main'snmp_set
{
	local($nr,@list) = @_;
	local($req,$ans,@rlst,$var,@ret);
	
	if($#list != 0) 
		{$req = join(";",@list);}
	else
		{$req = $list[0];}
	if( !defined(($ans = &snmp_raw_set($nr,$req))))
	{
	    $main'TrickletErrNo = 12;
            ( 
                $main'TrickletSpookErr,
                $main'TrickletSNMPErr,
                $main'TrickletSNMPVar
            ) = $main'TrickletErrStr =~ /^.- <(\d+),(\d+),(\d+)>/;
            $main'TrickletErrStr = "Spook:snmp_set, set failed\nStdspook: $main'TrickletErrStr\n";
	    return undef;
	}
	@rlst = split(";",$ans);	# dangerous, no ';' in answers
	foreach $var ( &snmp_rsp2lst($ans))
	{
		@rlst = &snmp_var2lst($var);
		push(@ret,$rlst[2]);
	}
	return @ret;
}
#**********************************************************************
#* Name:	snmp_tbl
#* Synopsis:	void snmp_tbl( int nr, funct fun, list *value )
#* Description:	This fuction does SNMP table request over connection
#*		<nr>. This command will contain the MIB varaibles
#*		given in <list>.
#*		The answer will be passed to function <fun>
#*		in the form
#*			(# row , #column , $name, $index, $val)
#* Returns:	none
#**********************************************************************
sub main'snmp_tbl
{
	local($nr,$fun,@list) = @_;
	local($w1,$r2,$resp,$request,$ind,$var,@tmp);
	$w1 = "wh10" . $nr;
	$r2 = "rh20" . $nr;
	if($#list != 0) 
		{$request = join(";",@list);}
	else
		{$request = $list[0];}
	print $w1 "*$request\n";
	$resp = <$r2>; 
	if(substr($resp,0,2) ne "*+") 
	{
	    chop $resp;
	    $main'TrickletErrNo = 12;
            ( 
                $main'TrickletSpookErr,
                $main'TrickletSNMPErr,
                $main'TrickletSNMPVar
            ) = $main'TrickletErrStr =~ /^.- <(\d+),(\d+),(\d+)>/;
            $main'TrickletErrStr = "Spook:snmp_tbl request failed\nStdspook: $resp\n";
	    return undef; 
	};
	
	# loop until spook indicates that is it all over
		
	$i = 1;
	do
	{
		print $w1 "?\n";
		$resp = <$r2>; # should give return code
		$ind = substr($resp,0,2);
		if($ind eq "?-") 
		{ 
		    chop $resp;
		    $main'TrickletErrNo = 13;
            ( 
                $main'TrickletSpookErr,
                $main'TrickletSNMPErr,
                $main'TrickletSNMPVar
            ) = $main'TrickletErrStr =~ /^.- <(\d+),(\d+),(\d+)>/;
                    $main'TrickletErrStr = "Spook:snmp_tbl reply failed\nStdspook: $resp\n";
		    return undef; 
		};
		if( $ind eq "?+")
		{
			$j = 1;
			$ans = substr($resp,2);
			foreach $var ( &snmp_rsp2lst($ans))
			{
				@rlst = &snmp_var2lst($var);
				@tmp = ($i,$j,@rlst);
				&$fun(@tmp);
				$j++;
			}
			$i++;
		}
	} until $ind eq "?.";
}
#**********************************************************************
#* Name:	snmp_tbl_p
#* Synopsis:	void snmp_tbl_p(int nr,funct fun,void $param,list *value)
#* Description:	This fuction does SNMP table request over connection
#*		<nr>. This command will contain the MIB varaibles
#*		given in <list>.
#*		The answer will be passed to function <fun>
#*		in the form
#*			($param,#row,#column,$name,$index,$val)
#*              <param> can be an arbitrary value, to be used by
#*              the callback <fun> to identify the table
#* Returns:	none
#**********************************************************************
sub main'snmp_tbl_p
{
    local($nr,$fun,$param,@list) = @_;
    local($w1,$r2,$resp,$request,$ind,$var,@tmp);
    $w1 = "wh10" . $nr;
    $r2 = "rh20" . $nr;
    if($#list != 0) 
        {$request = join(";",@list);}
    else
        {$request = $list[0];}
    print $w1 "*$request\n";
    $resp = <$r2>; 
    if(substr($resp,0,2) ne "*+") 
    {
        chop $resp;
        $main'TrickletErrNo = 12;
            ( 
                $main'TrickletSpookErr,
                $main'TrickletSNMPErr,
                $main'TrickletSNMPVar
            ) = $main'TrickletErrStr =~ /^.- <(\d+),(\d+),(\d+)>/;
            $main'TrickletErrStr = "Spook:snmp_tbl request failed\nStdspook: $resp\n";
        return undef; 
    };
    
    # loop until spook indicates that is it all over
        
    $i = 1;
    do
    {
        print $w1 "?\n";
        $resp = <$r2>; # should give return code
        $ind = substr($resp,0,2);
        if($ind eq "?-") 
        { 
            chop $resp;
            $main'TrickletErrNo = 13;
            ( 
                $main'TrickletSpookErr,
                $main'TrickletSNMPErr,
                $main'TrickletSNMPVar
            ) = $main'TrickletErrStr =~ /^.- <(\d+),(\d+),(\d+)>/;
                    $main'TrickletErrStr = "Spook:snmp_tbl reply failed\nStdspook: $resp\n";
            return undef; 
        };
        if( $ind eq "?+")
        {
            $j = 1;
            $ans = substr($resp,2);
            foreach $var ( &snmp_rsp2lst($ans))
            {
                @rlst = &snmp_var2lst($var);
                @tmp = ($param,$i,$j,@rlst);
                &$fun(@tmp);
                $j++;
            }
            $i++;
        }
    } until $ind eq "?.";
}

#**********************************************************************
#* Name:	snmp_tbl_row
#* Synopsis:	void snmp_tbl_row(int nr,funct fun,void $param,list *value )
#* Description:	This fuction does SNMP table request over connection
#*		<nr>. This command will contain the MIB varaibles
#*		given in <list>.
#*		The answer will be passed to function <fun>
#*		in the form
#*			($param,# row , @values)
#*              <param> can be an arbitrary value, to be used by
#*              the callback <fun> to identify the table
#* Returns:	none
#**********************************************************************
sub main'snmp_tbl_row
{
    local($nr,$fun,$param,@list) = @_;
    local($w1,$r2,$resp,$request,$ind,$var,@tmp);
    $w1 = "wh10" . $nr;
    $r2 = "rh20" . $nr;
    if($#list != 0) 
        {$request = join(";",@list);}
    else
        {$request = $list[0];}
    print $w1 "*$request\n";
    $resp = <$r2>; 
    if(substr($resp,0,2) ne "*+") 
    {
        chop $resp;
        $main'TrickletErrNo = 12;
            ( 
                $main'TrickletSpookErr,
                $main'TrickletSNMPErr,
                $main'TrickletSNMPVar
            ) = $main'TrickletErrStr =~ /^.- <(\d+),(\d+),(\d+)>/;
            $main'TrickletErrStr = "Spook:snmp_tbl request failed\nStdspook: $resp\n";
        return undef; 
    };
    
    # loop until spook indicates that is it all over
        
    $i = 1;
    do
    {
        print $w1 "?\n";
        $resp = <$r2>; # should give return code
        $ind = substr($resp,0,2);
        if($ind eq "?-") 
        { 
            chop $resp;
            $main'TrickletErrNo = 13;
            ( 
                $main'TrickletSpookErr,
                $main'TrickletSNMPErr,
                $main'TrickletSNMPVar
            ) = $main'TrickletErrStr =~ /^.- <(\d+),(\d+),(\d+)>/;
                    $main'TrickletErrStr = "Spook:snmp_tbl reply failed\nStdspook: $resp\n";
            return undef; 
        };
        if( $ind eq "?+")
        {
            @ret = ();
            foreach $var ( &snmp_rsp2lst(substr($resp,2)))
            {
                @rlst = &snmp_var2lst($var);
                push(@ret,$rlst[2]);
            }
            &$fun($param,$i,@ret);
            $i++;
        }
    } until $ind eq "?.";
}

#**********************************************************************
#* Name:	snmp_info
#* Synopsis:	void snmp_info( int nr , char *mibvar )
#* Description:	This fuction retrieves MIB information about
#*		MIB variables <mibvar>. 
#* Returns:	It returns a list of four elements:
#*			(char * mibvar,
#*			 char * type,
#*			 char * access,
#*			 char * status)
#*		Where <type> is one of:
#*			TimeTicks
#*			Counter
#*			Aggregate
#*			INTEGER
#*			Gauge
#*			OctetString
#*			DisplayString
#*			ObjectID
#*			IpAddress
#*		And <access> is one of:
#*			read-only
#*			write-only
#*			read-write
#*			not-accessible
#*		And <status> is one of:
#*			mandatory
#*			deprecated
#*			optional
#**********************************************************************
sub main'snmp_info
{
	local($nr,$mibvar) = @_;
	local($w1,$r2,$resp,@ret);
	$w1 = "wh10" . $nr;
	$r2 = "rh20" . $nr;
	print $w1 "\$$mibvar\n";
	$resp = <$r2>; 
	@ret = $resp =~ /^\$([+-])(\S+) (\S+) (\S+) (\S+).?$/;
}


#**********************************************************************
# Below this line are only local routines
#**********************************************************************

#**********************************************************************
#* Name:	snmp_rsp2lst
#* Synopsis:	void snmp_rsp2lst( char *resp )
#* Description:	This fuction translates a (Tricklet) response
#*		into a list of SNMP varaibles.
#* Returns:	list with values.
#**********************************************************************
sub snmp_rsp2lst
{				  
	local($rsp) = @_;
	split(";",$rsp); 
}
#**********************************************************************
#* Name:	snmp_var2lst
#* Synopsis:	void snmp_var2lst( char *var)
#* Description:	This fuction translates a (Tricklet) variable
#*		into a list of the form:
#*			($name , $index , $value)
#* Returns:	list 
#**********************************************************************
sub snmp_var2lst
{				  
	local($var) = @_;
	local($a,$b,$c) = $var =~ /^(.+)\[(.+)\]=(.+)$/;
	return($a,$b,$c);
}

#**********************************************************************
#* Name:	snmp_trans
#* Synopsis:	void snmp_trans( char *type, char *value )
#* Description:	This fuction translates a (Tricklet) SNMP type
#*		to a perl usable value. <type> constains the
#*		type of SNMP variable, and <value> its value.
#* Returns:	perl value
#**********************************************************************
sub snmp_trans
{
	local($type,$value) = @_;
	
	if($type eq "OctetString")
	{
		$value =~ /^\"(.*)\"/; # "
		return $1;
	}
	return $value;
}

#**********************************************************************
#* Name:	snmp_parse
#* Synopsis:	void snmp_parse( char *resp )
#* Description:	This fuction parse a spook answer into
#*              a perl list. The list consists of pairs of
#*		three strings
#*			(name,index,value)
#*		was opened using &snmp_open.
#* Returns:	none
#**********************************************************************
sub snmp_parse
{
	@ret = $resp =~ /^\?([+-])(.+)\[(.+)\]=(.+)$/;
	@info = &snmp_info($nr,$mibvar);
	&snmp_trans($info[2],$ret[3]);
}

sub snmp_raw_get
{
	local($nr,$request) = @_;
	local($w1,$r2,$resp,@ret,@info);
	$w1 = "wh10" . $nr;
	$r2 = "rh20" . $nr;
	if(! ($request eq '') )
	{
		print $w1 "<$request\n";
		$resp = <$r2>; 
		if(substr($resp,0,2) ne "<+")  
		{
		    chop $resp;
                    $main'TrickletErrStr = $resp;
 		    return undef; 
		};
	}
	print $w1 "?\n";
	$resp = <$r2>; # should give return code
	if(substr($resp,0,2) ne "?+")  
        { 
	        chop $resp;
                $main'TrickletErrStr = $resp;
                return undef; 
        };
#	print "DEBUG: snmp_raw_get: <$nr><$request><$resp>\n";
	return substr($resp,2);
}
sub snmp_raw_getnext
{
	local($nr,$request) = @_;
	local($w1,$r2,$resp,@ret,@info);
	$w1 = "wh10" . $nr;
	$r2 = "rh20" . $nr;
	if(! ($request eq '') )
	{
		print $w1 "}$request\n";
		$resp = <$r2>; 
		if(substr($resp,0,2) ne "}+")  
		{ 
		    chop $resp;
		    $main'TrickletErrStr = $resp;
		    return undef; 
		};
	}
	print $w1 "?\n";
	$resp = <$r2>; # should give return code
	if(substr($resp,0,2) ne "?+")  
	{ 
	    chop $resp;
            $main'TrickletErrStr = $resp;
	    return undef; 
	};
	return substr($resp,2);
}
sub snmp_raw_set
{
	local($nr,$request) = @_;
	local($w1,$r2,$resp,@ret,@info);
	$w1 = "wh10" . $nr;
	$r2 = "rh20" . $nr;
	print $w1 ">$request\n";
	$resp = <$r2>; 
	print $w1 "?\n";
	$resp = <$r2>; # should give return code
	if(substr($resp,0,2) ne "?+") 
	{
	    chop $resp;
	    $main'TrickletErrStr = $resp;
	    return undef; 
	};
	return substr($resp,2);
}
1;		



