package trap;
#########################################################################
## MODULE INFORMATION#
######################
##     FILE     NAME:       trap.pl
##     SYSTEM   NAME:       trap
##     ORIGINAL AUTHOR(S):  Jan van Oorschot
##     VERSION  NUMBER:     1.00
##     CREATION DATE:       Tue Jul 19 11:03:51 1994
##
## DESCRIPTION: A Tricklet library to work with snmp traps
##              using the tricklet utility snmp-trap. Notice
##              that this utility should be run by root or
##              ownded by root and have the 's bit on. Here
##              is the normal mode of this program:
##
##              -rwsr-xr-x 1 root 352256 Jul 15 14:32 snmp-trap*
##              
#########################################################################
## CHANGE INFORMATION ##
#########################
##      JvO - Jul 19, 1994:  Created
#########################################################################
#**********************************************************************
#* Name:	trap_open
#* Synopsis:	void trap_open( int nr  )
#* Description:	This fuction creates the trap deamon. The file handle
#*              that is returned can be used to read traps receive 
#*              by the host. Each line receive from that handle
#*              is in the format defined by the 'snmp-trap'
#*              util.
#* Returns:	the handle of the trap (read-only) file.
#**********************************************************************
require "tricklet.pl";
%TrapFileSize = ();
%TrapFileHandle = ();
$LastHandle = "trap0";
sub main'trap_open
{
    # check for existence, and remember the current file size
    local($file) = @_;

    if( ! -e $file )
    {
	$TrapFileSize{$file} = 0;
    }
    else
    {
	$TrapFileSize{$file} = -s $file;
    }
    $TrapFileHandle{$file} = $LastHandle++;
    return 1;
}

sub main'trap_check
{
    local($file) = @_;
    local($size,$osize,$hndl);

    if( ! -e $file)
    {
	$TrapFileSize{$file} = 0;
	return 0;
    }
    $osize = $TrapFileSize{$file};
    $size = -s $file;
    if($osize == $size)
    {
	return 0;
    }
    if($osize > $size)
    {
	$TrapFileSize{$file} = 0;
    }
    # return the opened file handle
    $hndl = $TrapFileHandle{$file};
    open($hndl,$file) || return 0;
    seek($hndl,$TrapFileSize{$file},0);
    $TrapFileSize{$file} = $size;
    return $hndl;
}


sub main'trap_receive
{
    local($file,$hndl) = @_;
    local($line,@arr);

    if(! ($line = <$hndl>) )
    {
	close($hndl);
	$TrapFileSize{$file} = -s $file;
	return (0);
    }
    if( ! ($line =~ /trap:/)  )
    {
	close($hndl);
	chop $line;
	$main'TrickletErrStr = "Trap: Illegal trap file $line\n";
	$main'TrickletErrNo = 303;
	return undef;
    }
    @arr = &main'line2lst($line);
    while( ! (($line = <$hndl>) =~/^end/) )
    {
        @arr = ( @arr , &main'line2lst($line));
    }
    return @arr;
}

1;
