package tricklet;
#########################################################################
## MODULE INFORMATION#
######################
##     FILE     NAME:       tricklet.pl
##     SYSTEM   NAME:       tricklet
##     ORIGINAL AUTHOR(S):  Jan van Oorschot
##     VERSION  NUMBER:     1.00
##     CREATION DATE:       Mon Jul 18 10:18:46 1994
##
## DESCRIPTION: Handy Perl subroutines to work with networks
##              and SNMP.
#########################################################################
## CHANGE INFORMATION ##
#########################
##      JvO - Jul 18, 1994:  Created
#########################################################################

# the two global error variables, read these after you 
# receive an error from Tricklet
$main'TrickletErrNo = 0;
$main'TrickletErrStr = "";


# get a variable from the Tricklet configuration system
sub main'ini2val
{
    local($var,$prg) = @_;
    local($line,$var2,$type,$val,$file);
    if( $prg eq "")
    {
	$prg = "global";
    }
    $line = `ini2cfg -s $var $prg`;
    if($line eq "")
    {
	$TrickletErrStr = "Tricklet: No IniVariable [$var,$prg]";
	$TrickletErrNo  = 1;
	return undef;
    }
    ($var2,$type,$val,$file)=$line=~/(\S+)\s*=\s*\((\S+)\)\s*(.+)\s+#\s*(\S*)/;
    $val =~ s/\s*$//;
    return $val;
}

#$ret = &ini2val("mgmt.medium","sherlock");
    
# convert a hex representation of an IP address to a decimal form
sub main'iph2ipa
{
    local($iph) = @_;
    local($d1,$d2,$d3,$d4);

    ($d1,$d2,$d3,$d4) = split(/:/,$iph);
    return( join('.',hex($d1),hex($d2),hex($d3),hex($d4)));
}

#convert a decimal representation of an IP address to a hex form
sub main'ipa2iph
{
    local($ipa) = @_;
    local($d1,$d2,$d3,$d4);
    
    ($d1,$d2,$d3,$d4) = split(/\./,$ipa);
    return( sprintf("%02x:%02x:%02x:%02x",$d1,$d2,$d3,$d4));
}			

# get the real name of an ip address
sub main'ip2name
{
    local($addr) = @_;
    local($n,$name,$aliases,$addrtype,$length,@addrs);

    if( $addr =~ /:/)
    {
	$addr = &main'iph2ipa($addr);
    }
    $n = pack("C4",split(/\./,$addr));
    ($name,$aliases,$addrtype,$length,@addrs) =gethostbyaddr($n,2);
    if( $name eq "")
    {
	return($addr);
    }
    return($name);
}

# get the ip address of a named host
sub main'name2ip
{
    local($name) = @_;
    local($lname,$aliases,$addrtype,$length,@addrs) = gethostbyname($name);
    local($a,$b,$c,$d) = unpack('C4',$addrs[0]);
    return sprintf("%d.%d.%d.%d",$a,$b,$c,$d);
}

# get the service name
sub main'serv2name
{
    local($port,$proto) = @_;
    local($name,$aliases,$p,$pr);
    ($name,$aliases,$p,$pr) = getservbyport($port,$proto);
    if($name eq "")
    {
	return($port);
    }
    return($name);
}

sub main'lst2line
{
    local(@lst) = @_;
    local($line);
    foreach $elem (@lst)
    {
	$line = $line . '"' . $elem . '" ';
    }
    chop $line;
    return $line;
}

sub main'line2lst
{
    local($str) = @_;
    local(@line,$ptr,$busy,$state,$len,$curstr,@lst);
    @line = split(//,$str);
    $ptr = 0;
    $busy=1;
    $state = 1;
    $len = $#line + 1;
    while($busy)
    {
	if($state == 1)
	{
	    # not reading anything
	    if($ptr >= $len)
	    {
		# end of line
		$busy = 0;
	    }
	    elsif($line[$ptr] =~ /^\s/)
	    {
		$ptr++;
	    }
	    elsif($line[$ptr] eq '"')
	    {
		# start of a string
		$ptr++;
		$curstr = "";
		$state=2;
	    }
	    else
	    {
		# start of a normal value
		$curval="";
		$state=3;
	    }
	}
	elsif($state == 2)
	{
	    # reading string
	    if($ptr >= $len)
	    {
		push(@lst,$curstr);
		$busy=0;
	    }
	    elsif($line[$ptr] eq '"')
	    {
		push(@lst,$curstr);
		$ptr++;
		$state=1;
	    }
	    elsif($line[$ptr] eq '\\')
	    {
		$ptr++;
		$curstr .= "\\$line[$ptr++]";
	    }
	    else
	    {
		$curstr .= $line[$ptr++];
	    }
	}
	elsif($state == 3)
	{
	    # reading normal scallar
	    if($ptr >= $len)
	    {
		push(@lst,$curval);
		$busy = 0;
	    }
	    elsif($line[$ptr] =~ /^\s/)
	    {
		push(@lst,$curval);
		$state=1;
	    }
	    else
	    {
		$curval .= $line[$ptr++];
	    }
	}
    }
    return @lst;
}
# calculate the TimeZone characteristics of this machine
# code butchered from <timelocal.pl>
@epoch = localtime(0);
$tzmin = $epoch[2] * 60 + $epoch[1];	# minutes east of GMT
if ($tzmin > 0) {
    $tzmin = 24 * 60 - $tzmin;		# minutes west of GMT
    $tzmin -= 24 * 60 if $epoch[5] == 70;	# account for the date line
}
if($tzmin < 0)
{
    $tzmin = $tzmin * -1;
    $TZSIGN = "+";
}
else
{
    $TZSIGN = "-";
}
$TZHOURS = int($tzmin / 60);
$TZMIN = $tzmin % 60;

sub main'time2str
{
    # convert a time() value to a string accoording to rfc1443
    # DISPLAY-HINT "2d-1d-1d,1d:1d:1d.1d,1a1d:1d"
    local($tim) = @_;
    @gmt = gmtime($tim);
    @lct = localtime($tim);
    $str = sprintf("%2d-%1d-%1d,%1d:%1d:%1d.%1d,%1s%1d:%1d",
            $lct[5] + 1900,     # WATCH OUT 2000 !!!!!
	    $lct[4] + 1,
	    $lct[3],
	    $lct[2],
	    $lct[1],
	    $lct[0],
	    0,
	    $TZSIGN,
	    $TZHOURS,
	    $TZMIN);
    return $str;
}
1;
