#!/usr/local/bin/perl

# This program is a support script for the SubScript facilities
# in the Beholder TNG.
# It uploads a SubScript source and starts it.
#
# DNPAP group of the Technical University Delft, The Netherlands.

require "spook.pl";
require "rmon.pl";

#
# handle the arguments
#
if (@ARGV != 5)
{
	die "usage: supload <script> <script-nr> <run-nr> [<host> [<community>]]\n" ,
        "    where:\n",
        "        <script>        name of script file\n",
        "        <script-nr>     number of script source\n",
        "        <run-nr>        number of run instance\n",
        "        <host>          name of RMON monitor\n" ,
        "        <community>     read-write community name";
}
($scriptfile, $snr, $rnr, $iniHost, $iniComm, @rest) = @ARGV;

#
# open the script file
#
if (! open(INPUT,$scriptfile))
{
	die "can not open script file";
}

#
# open a SNMP connection 
#
if (! &snmp_open(1))
{
	die "can not open a Spook connection\n";
}
if ($iniHost ne "")
{
	&snmp_host(1,$iniHost);
}
if ($iniMed ne "")
{
	&snmp_medium(1,$iniMed);
}
if (! &snmp_connect(1))
{
	die "could not connect to $iniHost";
}
if ($iniComm ne "")
{
	&snmp_community(1,$iniComm);
}

#
# set owner string
#
$owner = &main'ownerstring();

#
# open a SubScript collector
#
if (! &snmp_set(1,"subscriptSourceControlStatus[$snr]=2"))
{
	die "could not create source instance";
}
if (! &snmp_set(1,"subscriptSourceControlLanguage[$snr]=1"))
{
	&snmp_set(1,"subscriptSourceControlStatus[$snr]=4");
	die "could not set source language";
}
if (! &snmp_set(1,"subscriptSourceControlOwner[$snr]=\"$owner\""))
{
	&snmp_set(1,"subscriptSourceControlStatus[$snr]=4");
	die "could not set source owner";
}

#
# read the script, and store it as script in the beholder
#
$cnt = 1;
while($str = <INPUT>)
{
    chop $str;
	$str =~ s/"/\\"/g;
    if (! &snmp_set(1,"subscriptSourceLine[$snr.$cnt] = \"$str\""))
	{
		&snmp_set(1,"subscriptSourceControlStatus[$snr]=4");
		die "could write source line $snr.$cnt";
	}
    $cnt++;
}

#
# validate the SubScript source instance
#
if (! &snmp_set(1,"subscriptSourceControlStatus[$snr]=1"))
{
	&snmp_set(1,"subscriptSourceControlStatus[$snr]=4");
	die "could not validate source instance";
}

#
# create a SubScript run instance and configure it
#
if (! &snmp_set(1,"subscriptRunControlStatus[$rnr]=2"))
{
	die "could not create run instance";
}
if (! &snmp_set(1,"subscriptRunControlScript[$rnr]=$snr"))
{
	&snmp_set(1,"subscriptRunControlStatus[$rnr]=4");
	&snmp_set(1,"subscriptSourceControlStatus[$snr]=4");
	die "could not set run script";
}
if (! &snmp_set(1,"subscriptRunControlOwner[$rnr]=\"$owner\""))
{
	&snmp_set(1,"subscriptRunControlStatus[$rnr]=4");
	&snmp_set(1,"subscriptSourceControlStatus[$snr]=4");
	die "could not set run owner";
}

#
# validate the SubScript run instance and start the script
#
if (! &snmp_set(1,"subscriptRunControlStatus[$rnr]=1"))
{
	&snmp_set(1,"subscriptRunControlStatus[$rnr]=4");
	&snmp_set(1,"subscriptSourceControlStatus[$snr]=4");
	die "could not validate run instance";
}
if (! &snmp_set(1,"subscriptRunControlCommand[$rnr]=2"))
{
	&snmp_set(1,"subscriptRunControlStatus[$rnr]=4");
	&snmp_set(1,"subscriptSourceControlStatus[$snr]=4");
	die "could not start run instance";
}

sleep(1);

&snmp_tbl_row(1, main'callback, 0, "subscriptRunLogLine[]");

#
# close the SNMP connection
#
&snmp_disconnect(1);
&snmp_close(1);

sub callback
{
local($param, $row, @values) = @_;

	print "@values\n";
}
