set snmperrstr(0) "No error"
set snmperrstr(1) "Too big"
set snmperrstr(2) "No such name"
set snmperrstr(3) "Bad value"
set snmperrstr(4) "Read only"
set snmperrstr(5) "Generic error"


proc SnmpInit {} {
	global snmphost
	global snmpcommunity
	global snmpvarbind
	global snmpvarlist
	global snmperror

	if [catch {open "|stdspook" "r+"} fd] {
		error "can't open spook"
		return -1
	}

	set snmphost($fd)    	{}
	set snmpcommunity($fd)  {}
	set snmpvarbind($fd)  	{}
	set snmpvarlist($fd)  	{}
	set snmperror($fd)    	{}

	return $fd
}


proc SnmpExit {fd} {
	if [catch {close $fd}] {
		error "can't close spook"
		return -1
	}

	set snmphost($fd)    	{}
	set snmpcommunity($fd)  {}
	set snmpvarbind($fd)  	{}
	set snmpvarlist($fd)  	{}
	set snmperror($fd)    	{}

	return 0
}


proc Snmp {fd line} {
	global snmpvarbind
	global snmpvarlist
	global snmperror
	global snmperrstr

	# puts "Snmp argument: $line"
	puts $fd $line; flush $fd
	if {[gets $fd response] < 0} {
		return -3
	}
	# puts "Snmp response: $response"

	switch -glob $response {
	?+*     {
			set snmpvarbind($fd) [SnmpResponse2Varbind $response]
			set snmpvarlist($fd) [Varbind2List $snmpvarbind($fd)]
			set snmperror($fd)   "" 
			set ret 0
		}
	?-*     {
			set snmpvarbind($fd) ""
			set snmpvarlist($fd) ""
			set snmperror($fd)   [string range $response 3 end]
			set ret -1
		}
	\?.     {
			set snmpvarbind($fd) ""
			set snmpvarlist($fd) ""
			set snmperror($fd)   ""
			set ret 1
		}
	default {
			set snmpvarbind($fd) ""
			set snmpvarlist($fd) ""
			set snmperror($fd)   "unknown"
			set ret -2
		}
	}
	if {[scan $snmperror($fd) "SNMP error <%d> in variable <%d>" code var] == 2} {
		if {$var > 0} {
			set snmperror($fd) "Variable $var: $snmperrstr($code)"
		} else {
			set snmperror($fd) $snmperrstr($code)
		}
	}
	return $ret
}


proc SnmpResponse2Varbind {response} {
	set rawlist [split [string range $response 2 end] \;]
	for {set i 0; set varbind {}} {$i < [llength $rawlist]} {incr i} {
		lappend varbind [string trim [lindex $rawlist $i] " "]
	}
	return $varbind

}


proc SnmpHost {fd host} {
	if ![Snmp $fd "= host $host"] {
		set snmphost($fd) $host
	}
}


proc SnmpCommunity {fd community} {
	if ![Snmp $fd "= community $community"] {
		set snmpcommunity($fd) $community
	}
}


proc SnmpOpen {fd} {
	set snmpvarbind($fd)  	{}
	set snmpvarlist($fd)  	{}
	set snmperror($fd)    	{}
	return [Snmp $fd "\["]
}


proc SnmpClose {fd} {
	return [Snmp $fd "\]"]
}


proc SnmpKill {fd} {
	return [Snmp $fd "."]
}


proc SnmpGet {fd list} {
	return [Snmp $fd "< [List2Varbind $list]"]
}


proc SnmpSet {fd list} {
	return [Snmp $fd "> [List2Varbind $list]"]
}


proc SnmpNext {fd list} {
	return [Snmp $fd "\} [List2Varbind $list]"]
}


proc SnmpTable {fd list} {
	return [Snmp $fd "* [List2Varbind $list]"]
}


proc SnmpRawGet {fd varbind} {
	return [Snmp $fd "< $varbind"]
}


proc SnmpRawSet {fd varbind} {
	return [Snmp $fd "> $varbind"]
}


proc SnmpRawNext {fd varbind} {
	return [Snmp $fd "\} $varbind"]
}


proc SnmpRawTable {fd varbind} {
	return [Snmp $fd "* $varbind"]
}


proc SnmpExec {fd} {
	return [Snmp $fd "?"]
}
