/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dnpap.h>
#include <message.h>
#include <config.h>
#include <block.h>
#include <timer.h>
#include <dtime.h>
#include <nw.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "nwxmp.h"

#define SIZE_HOST		256
#define SIZE_VAR		256
#define SIZE_VALUE		256
#define SIZE_TIME		256

VOID CB_Timer_Burst(TIMER_DESCR * timer, ULONG now, VOID * parm);
VOID CB_Timer_Request(TIMER_DESCR * timer, ULONG now, VOID * parm);
VOID CB_ATM_Recv(NW_DG * dg, CHAR * addr, BYTE * frame, INT length, VOID * param);
VOID CB_ATM_Recv_Request(
							NW_DG * dg, CHAR * addr, BYTE * frame, INT length, VOID * param);

static NW_DG *ATM_Conn;
static NW_DG *ATM_Conn_Req;

static BYTE fr[MAX_LEN];
static char *cpHost, *cpHostReq;
static int iL1, iNrFrames, iTransFrames, iNrRuns, iID;
static ULONG ulInter;
static int iReady;
static ULONG TotalBurstTime;
static int iCurNr;

static TIMER_DESCR *handle1;

typedef struct _ThroughResult THROUGHRESULT;

struct _ThroughResult
{
	int id;
	int l1;
	int nrFrames;
	int transFrames;
	ULONG inter;
	int cnt;
	ULONG time;
};

static THROUGHRESULT *results;
static int resNr;

int main(int argc, char **argv)
{
	int i;

	/* read and check the arguments */
	if (argc != 8)
	{
		fprintf(stderr,
				"call as: nwthrough <echo-port> <nrFrames> <transFrames> <inter> <nrRuns> <len> <id>\n");
		exit(1);
	}

	cpHost = argv[1];
	iNrFrames = atoi(argv[2]);
	iTransFrames = atoi(argv[3]);
	ulInter = atol(argv[4]);
	iNrRuns = atoi(argv[5]);
	iL1 = atoi(argv[6]);
	iID = atoi(argv[7]);

	if (
		   (iL1 > MAX_LEN) || (iL1 < sizeof(FRAME)) ||
		   (iNrFrames > MAX_NR))
	{
		fprintf(stderr, " error in parameters\n");
		exit(2);
	}

	/* allocate the results array */
	if ((results = (THROUGHRESULT *) malloc(iNrRuns * sizeof(THROUGHRESULT))) == NULL)
	{
		fprintf(stderr, "could not allocate result array\n");
		exit(3);
	}
	resNr = 0;

	/* Initialise all subsystems */

	if (ConfigInit(argc, argv) == FALSE || ConfigLoad() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "ConfigInit failed");
		DnpapExit(1);
	}

	MessageInit("stdio:out", MSGFILE, MSGSTDIO, NULL);

	if (NwInit() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwInit failed");
		DnpapExit(1);
	}

	if (BlockInit() != 0)
		DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "BlockInit failed");

	if (!TimerInit())
		exit(1);

	/* open a ATM connection with the target host's echo port */
	if ((ATM_Conn = NwDgConnect(cpHost, CB_ATM_Recv, NULL)) == NULL)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwDgConnect failed<%s>",
					 cpHost);
		DnpapExit(1);
	}
	/* initiate the first Burst event */
	iCurNr = 0;
	if ((handle1 = TimerRegister(
									CB_Timer_Burst,
									NULL,	/* optional argument */
									100L,	/* duration timer in ms */
									1,	/* number of times to fire timer */
									TIMER_TYPE_RECOVER)) == 0)
		exit(1);
	/* Start the event handling */
	iReady = 0;
	while (!iReady)
		if (Block() != 0)
			DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "Block exit !");

	/* print the results */
	for (i = 0; i < resNr; i++)
	{
		printf("%d %d %d %d %d %lu %d %lu\n",
			   TYPE_THROUGH,
			   results[i].id,
			   results[i].l1,
			   results[i].nrFrames,
			   results[i].transFrames,
			   results[i].inter,
			   results[i].cnt,
			   results[i].time);
	}
	return 0;
}

VOID CB_Timer_Burst(TIMER_DESCR * timer, ULONG now, VOID * parm)
{
	FRAME *fr1;
	int i, totalFrames;
	ULONG StartBurstTime, StartNextFrame;

	if (!(iCurNr < iNrRuns))
	{
		iReady = 1;
		return;
	}
	/* send the next burst */
	iID++;
	fr1 = (FRAME *) fr;

	fr1->type = htonl(TYPE_THROUGH);
	fr1->id = htonl(iID);
	totalFrames = iNrFrames + iTransFrames;
	for (i = 0; i < totalFrames; i++)
	{
		StartNextFrame = TimeUsec() + ulInter;
		if (i == iTransFrames)
		{
			StartBurstTime = TimeUsec();
		}
		while (StartNextFrame > TimeUsec()) ;
		NwDgSend(ATM_Conn, fr, iL1);
	}
	TotalBurstTime = TimeUsec() - StartBurstTime;
        printf("%d\n",TotalBurstTime);
	iCurNr++;
	/* schedule a Request */
	if ((handle1 = TimerRegister(
									CB_Timer_Request,
									NULL,	/* optional argument */
									100L,	/* duration timer in ms */
									1,	/* number of times to fire timer */
									TIMER_TYPE_RECOVER)) == 0)
	{
		fprintf(stderr, "error scheduling a Request\n");
	}
}

VOID CB_Timer_Request(TIMER_DESCR * timer, ULONG now, VOID * parm)
{
	FRAME *fr1;

	/* request info from the echoer */
	fr1 = (FRAME *) fr;
	fr1->id = htonl(iID);
	NwDgSend(ATM_Conn_Req, fr, sizeof(FRAME));
	/* schedule the next burst */
	if ((handle1 = TimerRegister(
									CB_Timer_Burst,
									NULL,	/* optional argument */
									1000L,	/* duration timer in ms */
									1,	/* number of times to fire timer */
									TIMER_TYPE_RECOVER)) == 0)
	{
		fprintf(stderr, "error scheduling a Request\n");
	}
}

VOID CB_ATM_Recv(NW_DG * dg, CHAR * addr, BYTE * frame, INT length, VOID * param)
{
	fprintf(stderr, "Should not receive anything here\n");
}

VOID CB_ATM_Recv_Request(
							NW_DG * dg,
							CHAR * addr,
							BYTE * frame,
							INT length,
							VOID * param)
{
	FRAME *fr1;

	/* first handle the request reply frame */
	fr1 = (FRAME *) frame;
	if (ntohl(fr1->id) != iID)
	{
		fprintf(stderr, "receiving wrong ID from echoer\n");
	}
	else
	{
		results[resNr].id = ntohl(fr1->id);
		results[resNr].l1 = iL1;
		results[resNr].nrFrames = iNrFrames;
		results[resNr].transFrames = iTransFrames;
		results[resNr].inter = ulInter;
		results[resNr].cnt = ntohl(fr1->cnt);
		results[resNr].time = TotalBurstTime;
		resNr++;
	}

}
