/************************************************************************
** MODULE INFORMATION*
**********************
**     FILE     NAME:       nwxmp.c
**     SYSTEM   NAME:       nwxmp
**     ORIGINAL AUTHOR(S):  Jan van Oorschot
**     VERSION  NUMBER:     1.00
**     CREATION DATE:       Tue Dec  7 17:26:42 1993
**
** DESCRIPTION: Example application using the NW, Block and DNPAP
**              libraries.
*************************************************************************
** CHANGE INFORMATION **
*************************
**      JvO - Dec 7, 1993:  Created
*************************************************************************/
/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dnpap.h>
#include <message.h>
#include <config.h>
#include <mibdf.h>
#include <block.h>
#include <timer.h>
#include <dtime.h>
#include <nw.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "nwxmp.h"

#define SIZE_HOST		256
#define SIZE_VAR		256
#define SIZE_VALUE		256
#define SIZE_TIME		256


static NW_DG *UDP_Conn;

VOID CB_Timer(TIMER_DESCR *timer, ULONG now, VOID *parm);
VOID CB_UDP_Recv(NW_DG *dg,CHAR *addr,BYTE *frame,INT length, VOID *param);

int BlockCnt;

int main(int argc, char **argv)
{
    TIMER_DESCR *handle1;

    /* Initialise all subsystems */

	if (ConfigInit(argc, argv) == FALSE || ConfigLoad() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "ConfigInit failed");
		DnpapExit(1);
	}

	MessageInit("stdio:out", MSGFILE, MSGSTDIO, NULL);

	if (NwInit() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwInit failed");
		DnpapExit(1);
	}

	if (BlockInit() != 0)
		DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "BlockInit failed");

    if (!TimerInit())
	exit(1);

    /* open a UDP connection with the target host's echo port */
    if((UDP_Conn = NwDgConnect(argv[1],CB_UDP_Recv,NULL)) == NULL)
    {
	DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwDgConnect failed<%s>",
		     argv[1]);
	DnpapExit(1);
    }
    /* initiate a time which will fire every one-in-a-while */
    if ((handle1=TimerRegister(
			       CB_Timer,
			       0,
			       1000L,
			       TIMER_FOREVER,
			       TIMER_TYPE_RECOVER))==0)
        exit(1);
    /* Start the event handling*/
    BlockCnt = 10;
    while (BlockCnt>0)
	if (Block() != 0)
	    DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "Block exit !");
    return 0;
}

VOID CB_Timer(TIMER_DESCR *timer, ULONG now, VOID *parm)
{
    BYTE  fr[MAX_LEN];
    FRAME *fr1;
    /* send a new datagram to the destination */
    fr1 = (FRAME *) fr;
    fr1->t1 = TimeUsec();
    fr1->l1 = 100;
    fr1->l2 = 200;
    if(!NwDgSend(UDP_Conn,(BYTE *) fr1,fr1->l1))
    {	
	DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwDgCSend failed");
	DnpapExit(1);
    }       
    BlockCnt--;
}



VOID CB_UDP_Recv(NW_DG *dg,CHAR *addr,BYTE *frame,INT length, VOID *param)
{
    FRAME *fr;

    fr = (FRAME *) frame;

    fr->t4 = TimeUsec();
    printf("%s[%d]: %ld %ld %ld %ld %ld %ld\n",
	   addr,
	   length,
	   fr->t1,
	   fr->t2,
	   fr->t3,
	   fr->t4,
	   fr->l1,
	   fr->l2);
}
