/************************************************************************
** MODULE INFORMATION*
**********************
**     FILE     NAME:       nwecho.c
**     SYSTEM   NAME:       nwecho
**     ORIGINAL AUTHOR(S):  Jan van Oorschot
**     VERSION  NUMBER:     1.00
**     CREATION DATE:       Tue Dec  7 17:26:42 1993
**
** DESCRIPTION: Echo deamon for the generalised roundtrip measurements.

*************************************************************************
** CHANGE INFORMATION **
*************************
**      JvO - Dec 7, 1993:  Created
*************************************************************************/
/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dnpap.h>
#include <message.h>
#include <config.h>
#include <block.h>
#include <timer.h>
#include <dtime.h>
#include <nw.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "stack.h"

#define SIZE_HOST		256
#define SIZE_VAR		256
#define SIZE_VALUE		256
#define SIZE_TIME		256


static NW_DG *UDP_Conn_Echo;
static NW_DG *UDP_Conn_Through;
static NW_DG *UDP_Conn_Request;

VOID CB_UDP_Echo(NW_DG *dg,CHAR *addr,BYTE *frame,INT length, VOID *param);
VOID CB_UDP_Through(NW_DG *dg,CHAR *addr,BYTE *frame,INT length, VOID *param);
VOID CB_UDP_Request(NW_DG *dg,CHAR *addr,BYTE *frame,INT length, VOID *param);

static int Ready;
static int curID,curCnt;
static char *cpHostEcho,*cpHostThrough,*cpHostRequest;
int main(int argc, char **argv)
{
    /* check parameters */
    if(argc != 4)
    {
	fprintf(stderr,
		"call as: nwecho <echo-port> <through_port> <request_port>\n");
	exit(1);
    }
    cpHostEcho = argv[1];
    cpHostThrough = argv[2];
    cpHostRequest = argv[3];

    /* Initialise all subsystems */
    if (ConfigInit(argc, argv) == FALSE || ConfigLoad() == FALSE)
    {
	DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "ConfigInit failed");
	DnpapExit(1);
    }
    
    MessageInit("stdio:out", MSGFILE, MSGSTDIO, NULL);
    
    if (NwInit() == FALSE)
    {
	DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwInit failed");
	DnpapExit(1);
    }

    if (BlockInit() != 0)
	DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "BlockInit failed");

    if (!TimerInit())
	exit(1);

    /* initialse static info */
    curID = 999999;
    curCnt = 0;
    /* open a UDP connections */
    if((UDP_Conn_Echo = NwDgAccept(cpHostEcho,CB_UDP_Echo,NULL)) == NULL)
    {
	DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwDgConnect failed<%s>",
		     cpHostEcho);
	DnpapExit(1);
    }
    if((UDP_Conn_Through=NwDgAccept(cpHostThrough,CB_UDP_Through,NULL))==NULL)
    {
	DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwDgConnect failed<%s>",
		     cpHostThrough);
	DnpapExit(1);
    }
    if((UDP_Conn_Request=NwDgAccept(cpHostRequest,CB_UDP_Request,NULL))==NULL)
    {
	DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwDgConnect failed<%s>",
		     cpHostRequest);
	DnpapExit(1);
    }
    /* Start the event handling*/
    Ready = 0;
    while (!Ready)
	if (Block() != 0)
	    DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "Block exit !");
    return 0;
}

/* structure that will contain the info to be send */

static BYTE  frame2[MAX_LEN];

VOID CB_UDP_Echo(NW_DG *dg,CHAR *addr,BYTE *frame,INT length, VOID *param)
{
    FRAME *fr1,*fr2;
    ULONG tt1;

    tt1 = TimeUsec();		/* remember time of receive */
    
    memcpy(frame2,frame,length);
    fr1 = (FRAME *) frame;
    fr2 = (FRAME *) frame2;
    if(curID != ntohl(fr1->id))
    {
	curID = ntohl(fr1->id);
	curCnt = 1;
    }
    else
    {
	curCnt++;
    }
    fr2->t2 = htonl(tt1);
    fr2->t3 = htonl(TimeUsec());
    NwDgSendTo(dg,addr,frame2,htonl(fr1->l2));
}

VOID CB_UDP_Through(NW_DG *dg,CHAR *addr,BYTE *frame,INT length, VOID *param)
{
    FRAME *fr1;

    fr1 = (FRAME *) frame;

    if(curID != ntohl(fr1->id))
    {
	curID = ntohl(fr1->id);
	curCnt = 1;
    }
    else
    {
	curCnt++;
    }
}

VOID CB_UDP_Request(NW_DG *dg,CHAR *addr,BYTE *frame,INT length, VOID *param)
{
    FRAME *fr1,*fr2;
    ULONG tt1;

    tt1 = TimeUsec();		/* remember time of receive */
    
    memcpy(frame2,frame,length);
    fr1 = (FRAME *) frame;
    fr2 = (FRAME *) frame2;
    if(curID != ntohl(fr1->id))
    {
	fr2->cnt = 0;
    }
    else
    {
	fr2->cnt = htonl(curCnt);
    }
    fr2->t2 = htonl(tt1);
    fr2->t3 = htonl(TimeUsec());
    NwDgSendTo(dg,addr,frame2,length);
}

