/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dnpap.h>
#include <message.h>
#include <mibdf.h>
#include <config.h>


static INT depth, maxdepth;


INT main(INT argc, CHAR **argv);
static VOID Dump(CONST CHAR *name);
static INT MibID2String(LONG *id, INT len, CHAR *oid);


INT main(INT argc, CHAR **argv)
{
INT i, j;

	if (ConfigInit(argc, argv) == FALSE || ConfigLoad() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "ConfigInit failed");
		DnpapExit(1);
	}

	MessageInit("stdio:out", MSGFILE, MSGSTDIO, NULL);
	MessageConfig(TRICKLET_ERROR, "Snmp-Dump");

	if (argc < 2)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "usage: snmp-dump [-d depth] <MIB variable name> ...");
		DnpapExit(2);
	}

	if (MibDfInit() != 0 || MibDfFullMib() != 0)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "MibDfInit failed");
		DnpapExit(1);
	}

	depth = 0;
	maxdepth = 10000;

	for (i = 1; i < argc; i++)
	{
		if (argv[i][0] == '-')
		{
			for (j = 1; j < strlen(argv[i]); j++)
			{
				switch (argv[i][j])
				{
				case 'd':
					if (i+1 < argc)
					{
						if ((maxdepth = atoi(argv[++i])) < 1)
						{
							DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "invalid value for -d");
							DnpapExit(2);
						}
					}
					else
					{
						DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "missing value for -d");
						DnpapExit(2);
					}
					break;
				default:
					DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "invalid option %c", argv[i][j]);
					DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "usage: snmp-dump [-d depth] <MIB variable name>");
					DnpapExit(2);
					break;
				}
			}
		}
		else
		{
			Dump(argv[i]);
		}
	}

	return 0;
}


VOID Dump(CONST CHAR *name)
{
CONST CHAR *child;
LONG id[256];
INT len;
static CHAR info[256], oid[256] = "";

	depth++;
	if ((len = MibDfName2ID(id, name)) < 0)
		return;
	if (MibDfIDInfo(id, len, info) < 0)
		return;
	MibID2String(id, len, oid);
	printf("%s %s\n", info, oid);
	if (depth < maxdepth)
	{
		child = MibDfGetFirstChild(name);
		while (child != NULL)
		{
			Dump(child);
			child = MibDfGetNextChild(child);
		}
	}
	depth--;
}


INT MibID2String(LONG *id, INT len, CHAR *oid)
{
INT i;

    for (i = 0; i < len-1; i++)
    {
        sprintf(oid, "%ld.", id[i]);
        oid += strlen(oid);
    }
    if (i < len)
    {
        sprintf(oid, "%ld", id[i]);
        oid += strlen(oid);
    }

	if (i == 0)
		return -1;
	return i;
}
