/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <limits.h>

#include <dnpap.h>
#include <message.h>
#include <maxmem.h>
#include <dmemory.h>
#include <config.h>
#include <mibdf.h>
#include <block.h>
#include <snmp.h>
#include <mibsup.h>
#include <nw.h>
#include <nwsnmpdef.h>


#define NUMBER_VARBINDS			NWSNMP_SIZE_OBJECTS

#define EMUL_ID_VERBOSE			{ DNPAP_MIB, 5, 1, 0 }
#define EMUL_ID_STATUS			{ DNPAP_MIB, 5, 2, 0 }
#define EMUL_IDLEN				( DNPAP_MIBLEN+3 )

#define EMUL_VERBOSE	1
#define EMUL_QUIET		2

#define EMUL_VALID		1
#define EMUL_CONFIG		2

static INT verbose = EMUL_VERBOSE;
static INT status = EMUL_VALID;


typedef struct MibObject_ MibObject;
struct MibObject_
{
	SNMP_OBJECT	Object;
	int			IdLen;
	WORD		Support;
	CHAR		*Script;
};

static MibObject EmulVerbose = 
{
	{ SNMP_PDU_SET, EMUL_ID_VERBOSE, EMUL_IDLEN, SNMP_INTEGER, { EMUL_QUIET }, 0 }, EMUL_IDLEN, MIB_READ|MIB_WRITE, NULL };
static MibObject EmulStatus = 
{
	{ SNMP_PDU_SET, EMUL_ID_STATUS, EMUL_IDLEN, SNMP_INTEGER, { EMUL_VALID }, 0 }, EMUL_IDLEN, MIB_READ|MIB_WRITE, NULL };

#define MIB_ALLOC_CHUNK		10
static MibObject **MibList = NULL;
static int MibAllocated = 0;
static int MibSize = 0;


static NW_DG *acceptdg = NULL;

static CHAR *Status[] =
{
	"noError",
	"tooBig",
	"noSuchName",
	"badValue",
	"readOnly",
	"genError"
};

#define SIZE_HOST		256
#define SIZE_VAR		256
#define SIZE_VALUE		256
#define SIZE_TIME		256
#define SIZE_COMMAND	256
#define SIZE_LINE		512
#define SIZE_NAME		128

#ifndef abs
#define abs(x)		((x) < 0 ? -(x) : (x))
#endif

static VOID Callback(NW_DG * dg, CHAR *addr, BYTE * frame, INT length, VOID * param);

static BOOLEAN CheckRequest(INT mode, INT type, 
							 INT nrobjects, SNMP_OBJECT *objects, 
							 INT *status, INT *index);

static int MibListSearch(SNMP_OBJECT *object, int *compare);
static int MibListNext(int index);
static BOOLEAN MibListAdd(int index, MibObject *object);
static MibObject *MibListRemove(int index);

static WORD MibObjectAdd(SNMP_OBJECT *object);
static WORD MibObjectAddVerbatim(MibObject *object);
static WORD MibObjectRemove(SNMP_OBJECT *object);

static VOID MibObjectSetValue(MibObject *obj, SNMP_OBJECT *object);
static VOID MibObjectGetValue(MibObject *obj, SNMP_OBJECT *object);

static WORD MibObjectSet(SNMP_OBJECT *object);
static WORD MibObjectNext(SNMP_OBJECT *object);
static WORD MibObjectGet(SNMP_OBJECT *object);

static WORD MibObjectDo(INT request, CHAR *script, SNMP_OBJECT *object, int idlen, BOOLEAN first);

BOOLEAN ParseMibVar(WORD *type, SNMP_OBJECT *ret, char *line);
BOOLEAN Varbind2Object(CHAR *name, SNMP_OBJECT *object);
INT MibString2Id(CHAR *s, LONG *id, INT maxlen);

static BOOLEAN MibObjectInfo(SNMP_OBJECT *object, const char **name, int *type, int *mode, int *idlen);

static int VerboseCheck(INT type, SNMP_OBJECT *object);
static int StatusCheck(INT type, SNMP_OBJECT *object);

static INT PrintMibVar(SNMP_OBJECT *object, CHAR *buffer);
static INT PrintMibName(LONG *id, INT idlen, CHAR *var);
static INT PrintTimeTicks(ULONG timeticks, CHAR *str);
static INT PrintHost(ULONG address, CHAR *host);
static INT PrintHost2(CHAR *address, INT port, CHAR *host);

static INT ToPrint(BYTE *b, INT len);

static int ObjectCmp(SNMP_OBJECT *obj1, SNMP_OBJECT *obj2);
static int IdCmp(SNMP_OBJECT *obj1, MibObject *obj2);


int main(int argc, char **argv)
{
INT port;
CHAR address[NWSNMP_SIZE_ADDRESS];

	if (ConfigInit(argc, argv) == FALSE || ConfigLoad() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "ConfigInit failed");
		DnpapExit(1);
	}

    MemoryInit(8L*1024*1024, 48L*1024);

	MessageInit("stdio:out", MSGFILE, MSGSTDIO, NULL);
	MessageConfig(TRICKLET_ERROR, "Mib-Emul");

	port = 161;

	if (argc >= 2)
	{
		if ((port = atoi(argv[1])) == 0)
		{
			DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "invalid port number");
			DnpapExit(2);
		}
		if (argc > 2)
		{
			DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "too many arguments, only port number can be specified");
		}
	}

	if (MibDfInit() != 0 || MibDfFullMib() != 0)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "MibDfInit failed");
		DnpapExit(1);
	}

	if (NwInit() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwInit failed");
		DnpapExit(1);
	}

#ifdef OS2
	if (!BlockInit())
		DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "BlockInit failed");
#else
	if (BlockInit() != 0)
		DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "BlockInit failed");
#endif

	sprintf(address, "snmp:%d", port);
	if ((acceptdg = NwDgAccept(address, Callback, NULL)) == NULL)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwDgAccept failed");
		DnpapExit(1);
	}

	MibObjectAddVerbatim(&EmulVerbose);
	MibObjectAddVerbatim(&EmulStatus);

#ifdef OS2
	if (!Block())
		DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "Block exit !");
#else
	while (1)
		if (Block() != 0)
			DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "Block exit !");
#endif
}


VOID Callback(NW_DG * dg, CHAR *addr, BYTE * frame, INT length, VOID * param)
{
NwSnmpFrame *pdu = (NwSnmpFrame *)(VOID *)frame;
INT i;
CHAR address[NWSNMP_SIZE_ADDRESS];
CHAR host[NWSNMP_SIZE_HOST];
INT port;
CHAR request[NWSNMP_SIZE_REQUEST];
WORD type;
CHAR community[NWSNMP_SIZE_COMMUNITY];
INT commlen;
CHAR sender[SIZE_HOST];
CHAR line[SIZE_LINE];
int new;

	if (Addr2Snmp(addr, host, &port, request, &type, community, &commlen) == FALSE)
	{
        DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "invalid parameter in SNMP address");
        return;
	}

	verbose = EmulVerbose.Object.Syntax.LngInt;
	status = EmulStatus.Object.Syntax.LngInt;

	if (verbose == EMUL_VERBOSE)
	{
		PrintHost2(host, port, sender);
		ToPrint(community, commlen);
		printf("%s: %s \"%*s\" ", request, sender, commlen, community);
	}

	if (type == SNMP_PDU_GET || type == SNMP_PDU_SET || type == SNMP_PDU_NEXT)
	{
		if (verbose == EMUL_VERBOSE)
		{
			printf("%d\n", pdu->Pdu.Request.Id);
		}
	}
	else
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "invalid PDU request type (%d) received", type);
	}

	if (verbose == EMUL_VERBOSE)
	{
		for (i = 0; i < pdu->NrObjects; i++)
		{
			PrintMibVar(&(pdu->Objects[i]), line);
			puts(line);
		}

		puts("end");

		fflush(stdout);
	}

	pdu->Pdu.Request.Status = SNMP_NOERROR;
	pdu->Pdu.Request.Index = 0;

	if (CheckRequest(status, type, pdu->NrObjects, pdu->Objects, 
			&pdu->Pdu.Request.Status, &pdu->Pdu.Request.Index) == TRUE)
	{
		for (i = 0;  i < pdu->NrObjects && 
				pdu->Pdu.Request.Status == SNMP_NOERROR; i++)
		{
			if ((new = VerboseCheck(type, &pdu->Objects[i])) >= 0)
			{
				pdu->Pdu.Request.Status = new;
			}
			else
			if ((new = StatusCheck(type, &pdu->Objects[i])) >= 0)
			{
				pdu->Pdu.Request.Status = new;
			}
			else
			{
				switch (status)
				{
				case EMUL_CONFIG:
					switch (type)
					{
					case SNMP_PDU_SET:
						pdu->Pdu.Request.Status = MibObjectAdd(&pdu->Objects[i]);
						break;
					case SNMP_PDU_GET:
						pdu->Pdu.Request.Status = MibObjectRemove(&pdu->Objects[i]);
						break;
					}
					break;
				case EMUL_VALID:
					switch (type)
					{
					case SNMP_PDU_SET:
						pdu->Pdu.Request.Status = MibObjectSet(&pdu->Objects[i]);
						break;
					case SNMP_PDU_NEXT:
						pdu->Pdu.Request.Status = MibObjectNext(&pdu->Objects[i]);
						break;
					case SNMP_PDU_GET:
						pdu->Pdu.Request.Status = MibObjectGet(&pdu->Objects[i]);
						break;
					}
					break;
				}
			}
		}

		if (pdu->Pdu.Request.Status != SNMP_NOERROR)
		{
			pdu->Pdu.Request.Index = i;
		}
	}

	if (verbose == EMUL_VERBOSE)
	{
		printf("response: %s(%d) %d\n", Status[pdu->Pdu.Request.Status],
					pdu->Pdu.Request.Status, pdu->Pdu.Request.Index);

		for (i = 0; i < pdu->NrObjects; i++)
		{
			PrintMibVar(&(pdu->Objects[i]), line);
			puts(line);
		}

		puts("end\n");

		fflush(stdout);
	}

	sprintf(address, "snmp:%s:%d:%s:%*s", host, port, "response", commlen, community);
	if (NwDgSendTo(acceptdg, address, frame, length) == FALSE)
	{
		DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "returning response failed");
	}
}


BOOLEAN CheckRequest(INT mode, INT type, 
					 INT nrobjects, SNMP_OBJECT *objects, 
					 INT *status, INT *index)
{
int i, ind, compare;
MibObject *obj;

	switch (mode)
	{
	case EMUL_CONFIG:
		if (type == SNMP_PDU_NEXT)
		{
			*status = SNMP_GENERROR;
			*index = 1;
			return FALSE;
		}
		break;
	case EMUL_VALID:
		/* only SETs are checked */
		if (type == SNMP_PDU_SET)
		{
			for (i = 0; i < nrobjects; i++)
			{
				if ((ind = MibListSearch(&objects[i], &compare)) < 0)
					return SNMP_GENERROR;
				if (compare != 0)
				{
					*status = SNMP_NOSUCHNAME;
					*index = i+1;
					return FALSE;
				}
				obj = MibList[ind];
				if ((obj->Support & MIB_WRITE) == 0)
				{
					*status = SNMP_NOSUCHNAME;
					*index = i+1;
					return FALSE;
				}
			}
		}
		break;
	default:
		*status = SNMP_GENERROR;
		*index = 0;
		return FALSE;
		break;
	}

	return TRUE;
}


int MibListSearch(SNMP_OBJECT *object, int *compare)
{
int index, first, last;

	first = 0;
	last = MibSize-1;
	
	while (first < last)
	{
		index = first + (last - first)/2;
		
		switch(*compare = IdCmp(object, MibList[index]))
		{
		case -2:
		case -1:
			last = index;
			break;
		case 0:
		case 1:
			first = last = index;
			break;
		case 2:
			first = index+1;
			break;
		}
	}
	
	return first;
}


int MibListSize(void)
{
	return MibSize;
}


int MibListNext(int index)
{
	return (index < MibSize-1) ? index+1 : -1;
}


BOOLEAN MibListAdd(int index, MibObject *object)
{
MibObject **miblist;

	if (MibSize == MibAllocated)
	{
		if ((miblist = DnpapRealloc(MibList, (MibAllocated + MIB_ALLOC_CHUNK)*sizeof(MibList[0]))) == NULL)
			return FALSE;
		MibList = miblist;
		MibAllocated += MIB_ALLOC_CHUNK;
	}

	memmove(MibList + index + 1, MibList + index, 
			(MibSize - index)*sizeof(MibList[0]));

	MibList[index] = object;

	MibSize++;

	return TRUE;
}


MibObject *MibListRemove(int index)
{
MibObject *obj;

	obj = MibList[index];

	if (index < MibSize-1)
	{
		memmove(MibList + index, MibList + index + 1, 
				(MibSize - index)*sizeof(MibList[0]));
	}

	MibSize--;

	return obj;
}


VOID MibObjectSetValue(MibObject *obj, SNMP_OBJECT *object)
{
	memcpy(obj->Object.Id, object->Id, (obj->Object.IdLen = object->IdLen)*sizeof(object->Id[0]));
	switch (obj->Object.Type = object->Type)
	{
	case SNMP_INTEGER:
	case SNMP_IPADDR:
	case SNMP_COUNTER:
	case SNMP_GAUGE:
	case SNMP_TIMETICKS:
		obj->Object.Syntax.LngUns = object->Syntax.LngUns;
		obj->Object.SyntaxLen = 0;
		break;
	case SNMP_OCTETSTR:
	/* case SNMP_DISPLAYSTR: */
	case SNMP_OPAQUE:
		memcpy(obj->Object.Syntax.BufChr, object->Syntax.BufChr, obj->Object.SyntaxLen = object->SyntaxLen);
		break;
	case SNMP_OBJECTID:
		memcpy(obj->Object.Syntax.BufInt, object->Syntax.BufInt, (obj->Object.SyntaxLen = object->SyntaxLen)*sizeof(object->Syntax.BufInt[0]));
		break;
	}
}


VOID MibObjectGetValue(MibObject *obj, SNMP_OBJECT *object)
{
	memcpy(object->Id, obj->Object.Id, (object->IdLen = obj->Object.IdLen)*sizeof(object->Id[0]));
	switch (object->Type = obj->Object.Type)
	{
	case SNMP_INTEGER:
	case SNMP_IPADDR:
	case SNMP_COUNTER:
	case SNMP_GAUGE:
	case SNMP_TIMETICKS:
		object->Syntax.LngUns = obj->Object.Syntax.LngUns;
		object->SyntaxLen = 0;
		break;
	case SNMP_OCTETSTR:
	/* case SNMP_DISPLAYSTR: */
	case SNMP_OPAQUE:
		memcpy(object->Syntax.BufChr, obj->Object.Syntax.BufChr, object->SyntaxLen = obj->Object.SyntaxLen);
		break;
	case SNMP_OBJECTID:
		memcpy(object->Syntax.BufChr, obj->Object.Syntax.BufChr, (object->SyntaxLen = obj->Object.SyntaxLen)*sizeof(object->Syntax.BufInt[0]));
		break;
	}
}


WORD MibObjectAdd(SNMP_OBJECT *object)
{
MibObject *obj;
const char *name;
int index, type, support, idlen, compare;
char *script = NULL;
char config[SIZE_NAME];

	if (MibObjectInfo(object, &name, &type, &support, &idlen) == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "unknown MIB object");
		return SNMP_NOSUCHNAME;
	}

	if (object->Type != SNMP_NULL && type != object->Type)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "MIB object type mismatch");
		return SNMP_BADVALUE;
	}

	sprintf(config, "%s.script", name);
	if (object->Type == SNMP_NULL && ConfigGetString(config, &script) == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "MIB object script missing");
		return SNMP_GENERROR;
	}

	if ((index = MibListSearch(object, &compare)) < 0)
		return SNMP_GENERROR;
	if (abs(compare) == 1)		/* no sub or super trees allowed */
		return SNMP_GENERROR;
	if (compare == 2)			/* apparently at the end of the MIB tree */
		index = MibListSize();
	if (compare == 0)
	{
		obj = MibList[index];
	}
	else
	{
		if ((obj = DnpapMalloc(sizeof(MibObject))) == NULL)
		{
			DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "can not allocate MIB object");
			return SNMP_GENERROR;
		}

		if (MibListAdd(index, obj) == FALSE)
		{
			DnpapFree(obj);
			return SNMP_GENERROR;
		}
	}

	MibObjectSetValue(obj, object);

	obj->IdLen = idlen;
	obj->Support = support;
	obj->Script = script;

	return SNMP_NOERROR;
}


WORD MibObjectAddVerbatim(MibObject *object)
{
int index, compare;

	if ((index = MibListSearch(&object->Object, &compare)) < 0)
		return SNMP_GENERROR;
	if (abs(compare) == 1)		/* no sub or super trees allowed */
		return SNMP_GENERROR;
	if (compare == 2)			/* apparently at the end of the MIB tree */
		index = MibListSize();
	if (compare != 0)
	{
		if (MibListAdd(index, object) == FALSE)
		{
			return SNMP_GENERROR;
		}
	}
	return SNMP_NOERROR;
}


WORD MibObjectRemove(SNMP_OBJECT *object)
{
MibObject *obj;
int index, compare;

	if ((index = MibListSearch(object, &compare)) < 0)
		return SNMP_GENERROR;
	if (compare != 0)
		return SNMP_NOSUCHNAME;

	obj = MibListRemove(index);

	DnpapFree(obj);

	return SNMP_NOERROR;
}


WORD MibObjectSet(SNMP_OBJECT *object)
{
MibObject *obj;
int index, compare;

	if ((index = MibListSearch(object, &compare)) < 0)
		return SNMP_GENERROR;
	if (compare != 0 && compare != 1)
		return SNMP_NOSUCHNAME;

	obj = MibList[index];
	
	if (obj->Object.Type != SNMP_NULL && obj->Object.Type != object->Type)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "MIB object type mismatch");
		return SNMP_BADVALUE;
	}

	if (obj->Object.Type == SNMP_NULL && obj->Script != NULL)
		return MibObjectDo(SNMP_PDU_SET, obj->Script, object, obj->IdLen, FALSE);
	else
	{
		if (compare != 0)
			return SNMP_NOSUCHNAME;
		MibObjectSetValue(obj, object);
	}

	return SNMP_NOERROR;
}


WORD MibObjectNext(SNMP_OBJECT *object)
{
MibObject *obj;
int index, compare;
BOOLEAN first;
WORD result;

	if ((index = MibListSearch(object, &compare)) < 0)
		return SNMP_GENERROR;
	if (compare == 2)
		return SNMP_NOSUCHNAME;
	if (compare < 0)
		first = TRUE;
	else
		first = FALSE;
		
	for (result = SNMP_NOSUCHNAME; index < MibListSize(); index++)
	{
		obj = MibList[index];
		
		if (obj->Object.Type == SNMP_NULL && obj->Script != NULL)
		{
			if ((result = MibObjectDo(SNMP_PDU_NEXT, obj->Script, object, 
										obj->IdLen, first)) != SNMP_NOSUCHNAME)
				break;
		}
		else
		{
			if (first == TRUE && 
				obj != &EmulVerbose && obj != &EmulStatus)
			{
				MibObjectGetValue(obj, object);
				result = SNMP_NOERROR;
				break;
			}
		}
		
		compare = 0;
		first = TRUE;
	}

	return result;
}


WORD MibObjectGet(SNMP_OBJECT *object)
{
MibObject *obj;
int index, compare;

	if ((index = MibListSearch(object, &compare)) < 0)
		return SNMP_GENERROR;
	if (compare != 0 && compare != 1)
		return SNMP_NOSUCHNAME;

	obj = MibList[index];
	
	if (obj->Object.Type == SNMP_NULL && obj->Script != NULL)
		return MibObjectDo(SNMP_PDU_GET, obj->Script, object, obj->IdLen, FALSE);
	else
	{
		if (compare != 0)
			return SNMP_NOSUCHNAME;
		MibObjectGetValue(obj, object);
	}

	return SNMP_NOERROR;
}


WORD MibObjectDo(INT request, CHAR *script, SNMP_OBJECT *object, int idlen, BOOLEAN first)
{
WORD result;
FILE *handle;
char command[SIZE_COMMAND], *r, varbind[SIZE_LINE], line[SIZE_LINE];
WORD type, mode[NUMBER_VARBINDS];
SNMP_OBJECT ret, arg[NUMBER_VARBINDS];
int i, n, rc;

	PrintMibVar(object, varbind);
	switch (request)
	{
	case SNMP_PDU_GET:
		r = "get";
		break;
	case SNMP_PDU_SET:
		r = "set";
		break;
	case SNMP_PDU_NEXT:
		r = "next";
		break;
	default:
		r = NULL;
		break;
	}
	sprintf(command, "%s '%s %d %s'", script, r, first == TRUE, varbind);
	if (verbose == EMUL_VERBOSE)
	{
		printf("command: %s\n", command);
	}
	if ((handle = popen(command, "r")) == NULL)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "pipe open to script failed");
		return SNMP_GENERROR;
	}
	line[0] = '\0';
	if (fgets(line, SIZE_LINE, handle) == NULL)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "no result from script");
		return SNMP_GENERROR;
	}
	if (verbose == EMUL_VERBOSE)
	{
		printf("result: %s\n", line);
	}
	if (ParseMibVar(&type, &ret, line) == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "error parsing script result");
		return SNMP_GENERROR;
	}
	if (type != request)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "script returned different request");
		return SNMP_GENERROR;
	}
	if (request == SNMP_PDU_SET)
	{
		if (ObjectCmp(object, &ret) != 0)
		{
			DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "varbind changed by script in set request");
			return SNMP_GENERROR;
		}
	}
	else
	{
		if (request == SNMP_PDU_GET && ObjectCmp(object, &ret) == 1)
		{
			DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "varbind changed by script in get request");
			return SNMP_GENERROR;
		}
	}
	line[0] = '\0';
	for (i = 0; fgets(line, SIZE_LINE, handle) != NULL; i++)
	{
		if (verbose == EMUL_VERBOSE)
		{
			printf("argument: %s\n", line);
		}
		if (ParseMibVar(&mode[i], &arg[i], line) == FALSE)
		{
			DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "error parsing script argument");
			return SNMP_GENERROR;
		}
		line[0] = '\0';
	}
	if ((rc = pclose(handle)) < 0)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "pipe close of script failed");
		return SNMP_GENERROR;
	}
	for (n = i, i = 0, result = SNMP_NOERROR; 
			i < n && result == SNMP_NOERROR; i++)
	{
		switch (mode[i])
		{
		case SNMP_PDU_SET:
			result = MibObjectAdd(&arg[i]);
			break;
		case SNMP_PDU_GET:
			result = MibObjectRemove(&arg[i]);
			break;
		}
	}
	if (result == SNMP_NOERROR)
	{
		memcpy(object, &ret, sizeof(object));
	}

	return result;
}

 
int VerboseCheck(INT type, SNMP_OBJECT *object)
{
	if (IdCmp(object, &EmulVerbose) == 0)
	{
		if (type != SNMP_PDU_SET)
			return SNMP_NOSUCHNAME;
		if (verbose == EMUL_VERBOSE)
			printf("emulator verbose access, ");
		if (object->Syntax.LngInt != EMUL_VERBOSE && object->Syntax.LngInt != EMUL_QUIET)
		{
			if (verbose == EMUL_VERBOSE)
				printf("bad value (%ld)\n", object->Syntax.LngInt);
			return SNMP_BADVALUE;
		}
		EmulVerbose.Object.Syntax.LngInt = object->Syntax.LngInt;
		if (verbose == EMUL_VERBOSE)
			printf("set to %ld\n", EmulVerbose.Object.Syntax.LngInt);
		return SNMP_NOERROR;
	}

	return -1;
}


int StatusCheck(INT type, SNMP_OBJECT *object)
{
	if (IdCmp(object, &EmulStatus) == 0)
	{
		if (type != SNMP_PDU_SET)
			return SNMP_NOSUCHNAME;
		if (verbose == EMUL_VERBOSE)
			printf("emulator status access, ");
		if (object->Syntax.LngInt != EMUL_VALID && object->Syntax.LngInt != EMUL_CONFIG)
		{
			if (verbose == EMUL_VERBOSE)
				printf("bad value (%ld)\n", object->Syntax.LngInt);
			return SNMP_BADVALUE;
		}
		EmulStatus.Object.Syntax.LngInt = object->Syntax.LngInt;
		if (verbose == EMUL_VERBOSE)
			printf("set to %ld\n", EmulStatus.Object.Syntax.LngInt);
		return SNMP_NOERROR;
	}

	return -1;
}


BOOLEAN MibObjectInfo(SNMP_OBJECT *object, const char **name, int *type, int *mode, int *idlen)
{
int n;

	if ((*name = MibDfID2Name2(object->Id, object->IdLen, &n)) == NULL)
		return FALSE;
	*type = MibDfIDType(object->Id, n);
	*mode = MibDfIDMode(object->Id, n);
	*idlen = n;

	return TRUE;
}


INT ToPrint(BYTE *b, INT len)
{
INT i;

	for (i = 0; i < len; i++)
        if (b[i] < 32 || b[i] > 127)
            b[i] = '.';

	return len;
}


INT IsPrint(BYTE *b, INT len)
{
INT i, n = 0;

	for (i = 0; i < len; i++)
        if (b[i] >= 32 && b[i] <= 127)
            n++;

	return n;
}


BOOLEAN ParseMibVar(WORD *type, SNMP_OBJECT *ret, char *line)
{
char *p;

	while (isspace(*line))
		line++;
	p =line;
	while (!isspace(*p))
		p++;
	if (strncmp(line, "get", p-line) == 0)
		*type = SNMP_PDU_GET;
	else
	if (strncmp(line, "set", p-line) == 0)
		*type = SNMP_PDU_SET;
	else
	if (strncmp(line, "next", p-line) == 0)
		*type = SNMP_PDU_NEXT;
	else
		return FALSE;
	line = p;
	while (isspace(*line))
		line++;
	
	return Varbind2Object(line, ret);
}


BOOLEAN Varbind2Object(CHAR *name, SNMP_OBJECT *object)
{
INT l, n, n2;
LONG val;
CHAR *s, *e, *v, tmp[256], *ep, *p;

    s = strchr(name, '[');
    e = strchr(name, ']');
    v = strchr(name, '=');

    l = strlen(name);

    if (s == NULL && e == NULL)
    {
        if (v != NULL)
            l = v - name;
    }
    else
    if (s == NULL || e == NULL)
        return FALSE;

    strncpy(tmp, name, l);
    tmp[l] = '\0';
    
    if ((INT)(object->IdLen = MibDfName2ID(object->Id, tmp)) < 0)
        return FALSE;

    if ((SHORT)(object->Type = MibDfNameType(tmp)) < 0)
        return FALSE;

    if ((l = e - s - 1) > 0)
    {
        strncpy(tmp, s+1, l);
        tmp[l] = '\0';

        object->IdLen += MibString2Id(tmp, object->Id+object->IdLen, 
                                        SNMP_SIZE_OBJECTID-object->IdLen);
    }

    if (v != NULL && (l = strlen(++v)) > 0)
    {
        switch (object->Type)
        {
        case SNMP_INTEGER:
            object->Syntax.LngInt = strtol(v, &ep, 0);
            if (*ep != '\0')
                return FALSE;
            break;
        case SNMP_OCTETSTR:
        /* case SNMP_DISPLAYSTR: */
        case SNMP_OPAQUE:
            if (*v == '\'')
            {
                if (v[l-1] != '\'')
                    return FALSE;
                memcpy(object->Syntax.BufChr, v+1, object->SyntaxLen = l-2);
            }
            else
            {
                strcpy(tmp, v);
                p = strtok(tmp, ":");
                n = 0;
                while (p != NULL)
                {
                    if ((val = strtol(p, &ep, 0)) < 0 || val > UCHAR_MAX || *ep != '\0')
                        return FALSE;
                    object->Syntax.BufChr[n++] = (BYTE)val;
                    p = strtok(NULL, ":");
                    if (p != NULL && n == SNMP_SIZE_BUFCHR)
                        return FALSE;
                }
                object->SyntaxLen = n;
            }
            break;
        case SNMP_OBJECTID:
            strcpy(tmp, v);
            p = strtok(tmp, ".");
            n = 0;
            while (p != NULL)
            {
                if (isdigit(p[0]))
                {
                    if ((val = strtol(p, &ep, 0)) < 0 || *ep != '\0')
                        return FALSE;
                    object->Syntax.BufInt[n++] = val;
                    p = strtok(NULL, ".");
                    if (p != NULL && n == SNMP_SIZE_BUFINT)
                        return FALSE;
                }
                else
                {
                    if ((n2 = MibDfName2ID(object->Syntax.BufInt+n, p)) >= 0)
                        n += n2;
                }
            }
            object->SyntaxLen = n;
            break;
        case SNMP_IPADDR:
            strcpy(tmp, v);
            p = strtok(tmp, ".");
            n = 0;
            while (p != NULL)
            {
                if ((val = strtol(p, &ep, 0)) < 0 || val > UCHAR_MAX || *ep != '\0')
                    return FALSE;
                object->Syntax.LngUns = (object->Syntax.LngUns << 8) | (BYTE)val;
                p = strtok(NULL, ".");
                if (p != NULL && n == SNMP_SIZE_BUFINT)
                    return FALSE;
            }
            if (n != 4)
                return FALSE;
            break;
        case SNMP_COUNTER:
        case SNMP_GAUGE:
        case SNMP_TIMETICKS:
            object->Syntax.LngUns = (ULONG)strtol(v, &ep, 0);
            if (*ep != '\0')
                return FALSE;
            break;
        }
    }

    return TRUE;
}


INT MibString2Id(CHAR *s, LONG *id, INT maxlen)
{  
CHAR tmp[256], *p;
INT i;

	strcpy(tmp, s);
	
	p = strtok(tmp, ".");
	for (i = 0; p != NULL && i < maxlen; p = strtok(NULL, "."), i++)
	{
		if ((id[i] = atoi(p)) == 0)
			return i;
	}				  
	
	return i;
}


INT PrintMibVar(SNMP_OBJECT *object, CHAR *buffer)
{
INT i;
CHAR *type = NULL, value[SIZE_VALUE], var[SIZE_VAR], t[SIZE_TIME], *p;

	PrintMibName(object->Id, object->IdLen, var);

	switch (object->Type)
	{
	case SNMP_NULL:
		type = "NULL";
		strcpy(value, "");
		break;
	case SNMP_INTEGER:
		type = "INTEGER";
		sprintf(value, "%ld", object->Syntax.LngInt);
		break;
	case SNMP_OPAQUE:
	case SNMP_OCTETSTR:
	/* case SNMP_DISPLAYSTR: */
		if (object->Type == SNMP_OCTETSTR)
			type = "OCTETSTRING";
		else
			type = "Opaque";
		if ((FLOAT)IsPrint(object->Syntax.BufChr, object->SyntaxLen)/object->SyntaxLen > 0.75)
		{
			memcpy(value, object->Syntax.BufChr, object->SyntaxLen);
			ToPrint(value, object->SyntaxLen);
			value[object->SyntaxLen] = '\0';
		}
		else
		{
			value[0] = '\0';
			for (i = 0, p = value; i+1 < object->SyntaxLen; i++)
			{
				sprintf(p, "%2.2x:", object->Syntax.BufChr[i]);
				p += strlen(p);
			}
			if (i < object->SyntaxLen)
			{
				sprintf(p, "%2.2x", object->Syntax.BufChr[i]);
			}
		}
		break;
	case SNMP_OBJECTID:
		type = "OBJECTID";
		PrintMibName(object->Syntax.BufInt, object->SyntaxLen, value);
		break;
	case SNMP_IPADDR:
		type = "IpAddress";
		sprintf(value, "%d.%d.%d.%d", object->Syntax.BufChr[0], 
									  object->Syntax.BufChr[1],
									  object->Syntax.BufChr[2],
									  object->Syntax.BufChr[3]);
		break;
	case SNMP_COUNTER:
		type = "Counter";
		sprintf(value, "%lu", object->Syntax.LngUns);
		break;
	case SNMP_GAUGE:
		type = "Gauge";
		sprintf(value, "%lu", object->Syntax.LngUns);
		break;
	case SNMP_TIMETICKS:
		type = "TimeTicks";
		PrintTimeTicks(object->Syntax.LngUns, t);
		sprintf(value, "%s", t);
		break;
	}

	sprintf(buffer, "%s  %s  %s", var, type, value);

	return strlen(buffer);
}


INT PrintMibName(LONG *id, INT len, CHAR *s)
{
CONST CHAR *p;
CHAR *q;
INT i, n;

	p = MibDfID2Name2(id, len, &n);

	if (p == NULL)
		strcpy(s, "");
	else
		strcpy(s, p);

	if (len - n > 0)
	{
		strcat(s, "[");
		for (i = n, q = s+strlen(s); i < len-1; i++)
		{
			sprintf(q, "%ld.", id[i]);
			q += strlen(q);
		}
		if (i < len)
		{
			sprintf(q, "%ld", id[i]);
		}
		strcat(s, "]");
	}

	if (n > 0)
	{
		strcat(s, "(");
		for (i = 0, q = s+strlen(s); i < len-1; i++)
		{
			sprintf(q, "%ld.", id[i]);
			q += strlen(q);
		}
		if (i < len)
		{
			sprintf(q, "%ld", id[i]);
		}
		strcat(s, ")");
	}

	return strlen(s);
}


INT PrintTimeTicks(ULONG timeticks, CHAR *str)
{
INT h, m, s;
ULONG t;
		
	t = timeticks;
	h = t/360000UL;
	t -= h*360000UL;
	m = t/6000UL;
	t -= m*6000UL;
	s = t/100UL;
	t -= s*100UL;

	sprintf(str, "%.02d:%.02d:%02d.%02lu(%lu)", h, m, s, t, timeticks);

	return strlen(str);
}


INT PrintHost(ULONG address, CHAR *host)
{
HOSTENT *ent;
BYTE *b;

	if ((ent = gethostbyaddr((CHAR *)&address, sizeof(address), AF_INET)) == NULL)
		strcpy(host, "");
	else
		strcpy(host, ent->h_name);

	if (ent != NULL)
		strcat(host, "(");

	b = (BYTE *)&address;
	sprintf(host+strlen(host), "%d.%d.%d.%d", b[0], b[1], b[2], b[3]);

	if (ent != NULL)
		strcat(host, ")");

	return strlen(host);
}


INT PrintHost2(CHAR *address, INT port, CHAR *host)
{
	PrintHost(inet_addr(address), host);
	sprintf(host, "%s:%d", host, port);
	return strlen(host);
}


int ObjectCmp(SNMP_OBJECT *obj1, SNMP_OBJECT *obj2)
{
	if (obj1->IdLen != obj2->IdLen)
		return 1;
	if (memcmp(obj1->Id, obj2->Id, obj1->IdLen*sizeof(obj1->Id[0])) != 0)
		return 1;
	if (obj1->Type != obj2->Type)
		return 2;
	switch (obj1->Type)
	{
	case SNMP_NULL:
		break;
	case SNMP_INTEGER:
		if (obj1->Syntax.LngInt != obj2->Syntax.LngInt)
			return 3;
		break;
	case SNMP_OPAQUE:
	case SNMP_OCTETSTR:
		if (memcmp(obj1->Syntax.BufChr, obj2->Syntax.BufChr, obj1->SyntaxLen) != 0)
			return 3;
		break;
	case SNMP_OBJECTID:
		if (memcmp(obj1->Syntax.BufInt, obj2->Syntax.BufInt, 
					obj1->SyntaxLen*sizeof(obj1->Syntax.BufInt[0])) != 0)
			return 3;
		break;
	case SNMP_IPADDR:
	case SNMP_COUNTER:
	case SNMP_GAUGE:
	case SNMP_TIMETICKS:
		if (obj1->Syntax.LngUns != obj2->Syntax.LngUns)
			return 3;
		break;
	}
	
	return 0;
}


int IdCmp(SNMP_OBJECT *obj1, MibObject *obj2)
{
LONG *id1 = obj1->Id, *id2 = obj2->Object.Id;
int len1 = obj1->IdLen, len2 = obj2->Object.IdLen;

    while (len1 > 0 && len2 > 0 && *id1 == *id2)
    {
        len1--;
        len2--;
        id1++;
        id2++;
    }
    if (!len1 && !len2)
        return 0;
    if (!len1)
        return -1;
    if (!len2)
        return 1;
    if (*id1 < *id2)
        return -2;
    else
        return 2;
}
