/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
/************************************************************************
 ** MODULE INFORMATION*
 **********************
 **     FILE    NAME:       stdspook.c
 **     SYSTEM  NAME:       tricklet
 **     ORIGINAL AUTHOR(S): Jan van Oorschot
 **     VERSION NUMBER:
 **     CREATION DATE:
 **
 ** DESCRIPTION: This module implements the stdin/stdout interface
 **              to Spook.
 **
 *************************************************************************
 ** CHANGES INFORMATION **
 *************************
 ** REVISION: $Revision$
 ** AUTHOR:   $Author$
 ** DATE:     $Date$
 ** LOG:      $Log$
 *************************************************************************/
#include <dnpap.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <ctype.h>

#include <message.h>
#include <snmp.h>
#include <mibdf.h>
#include <dstring.h>
#include <mgmt.h>
#include <config.h>
#include "exec.h"
#include "parse.h"
#include "spook.h"



int main(int argc, char *argv[]);

int main(int argc, char *argv[])
{
	SPOOK_OBJ *spook;
	char request[BUFSIZE];
	char answer[BUFSIZE];
	BYTE *bp;
	char *cp;
	USHORT n;
    char stdinbuf[BUFSIZE], stdoutbuf[BUFSIZE];

    /* make sure we are doing line buffering and not a character more */
    setvbuf(stdin, stdinbuf, _IOLBF, BUFSIZE);
    setvbuf(stdout, stdoutbuf, _IOLBF, BUFSIZE);

	/* see to it stdspook keeps running */
	DnpapSignal(SIGINT, SIG_IGN);
	DnpapSignal(SIGTERM, SIG_IGN);

	if (ConfigInit(argc, argv) == FALSE || ConfigLoad() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "ConfigInit failed");
		DnpapExit(1);
	}

	MessageInit("stdio:out", MSGFILE, MSGSTDIO, MSGNULL, NULL);
	MessageConfig(TRICKLET_ERROR, "StdSpook");

	if (MgmtInit() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "MgmtInit failed");
		DnpapExit(1);
	}
	if (MibDfInit() != 0 || MibDfFullMib() != 0)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "MibDfInit failed");
		DnpapExit(1);
	}

	/* create a spool object to store th status in */
	spook = DnpapMalloc(sizeof(SPOOK_OBJ));
	if (spook == NULL)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "Malloc Failed");
		DnpapExit(1);
	}
	/* initialise this spook from the configuration files */
	if (ConfigGetString("mgmt.medium", &cp))
	{
		if (stricmp(cp, "UDP") == 0)
			spook->medium = MGMT_TYPE_UDP;
		else
		{
			if (stricmp(cp, "PIPE") == 0)
				spook->medium = MGMT_TYPE_PIPE;
			else
				spook->medium = MGMT_TYPE_DEFAULT;
		}
	}
	if (ConfigGetString("mgmt.agent", &cp))
	{
		strcpy(spook->host, cp);
	}
	if (ConfigGetBytes("mgmt.community", &bp, &n))
	{
		memcpy(spook->comm, bp, (size_t) n);
		spook->commLen = n;
	}
	if (!ConfigGetLong("mgmt.timeout", &(spook->timeout)))
		spook->timeout = 5000000L;
	if (!ConfigGetShort("mgmt.retries", &(spook->retries)))
		spook->retries = 3;

	/* overwrite this with the command line */
	if (argc >= 2)
		strcpy(spook->host, argv[1]);
	if (argc >= 3)
	{
		if (argv[2][0] == 'U' || argv[2][0] == 'u')
			spook->medium = MGMT_TYPE_UDP;
		else if (argv[2][0] == 'P' || argv[2][0] == 'p')
			spook->medium = MGMT_TYPE_PIPE;
	}
	if (argc >= 4)
	{
		strcpy(spook->comm, argv[3]);
		spook->commLen = strlen(argv[3]);
	}
	if (argc >= 5)
		spook->timeout = atol(argv[4]);
	if (argc >= 6)
		spook->retries = atoi(argv[5]);

	/* try to begin with an open medium */
	spook->handle = NULL;

	/* initialise the state machine */
	spook->have_list = 0;				/* you have no request outstanding */

	/* OK, start receiving requests and send answers back */
	spook->quit = 0;
	while (fgets(request, BUFSIZE, stdin) != NULL)
	{
		if (SpookMessage(spook, request, answer) != 0)
		{
			DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "MibDfInit failed");
			DnpapExit(1);
		}
		puts(answer);
		fflush(stdout);
		if (spook->quit)
			break;
	}
	if (spook->handle != NULL)
		MgmtClose(spook->handle);

	return 0;
}
