/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <dnpap.h>
#include <message.h>
#include <config.h>
#include <mibdf.h>
#include <block.h>
#include <snmp.h>
#include <nw.h>
#include <nwsnmpdef.h>


#define SIZE_HOST		256
#define SIZE_VAR		256
#define SIZE_VALUE		256
#define SIZE_TIME		256


static NW_DG *acceptdg;


static VOID Callback(NW_DG * dg, CHAR *addr, BYTE * frame, INT length, VOID * param);

static INT PrintMibVar(SNMP_OBJECT *object);
static INT PrintMibName(LONG *id, INT idlen, CHAR *var);
static INT PrintTimeTicks(ULONG timeticks, CHAR *str);
static INT PrintHost(ULONG address, CHAR *host);
static INT PrintHost2(CHAR *address, INT port, CHAR *host);

static INT ToPrint(BYTE *b, INT len);


int main(int argc, char **argv)
{
INT port, echo;
CHAR address[NWSNMP_SIZE_ADDRESS];

	echo = (strstr(argv[0], "echo") != NULL);

	if (ConfigInit(argc, argv) == FALSE || ConfigLoad() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "ConfigInit failed");
		DnpapExit(1);
	}

	MessageInit("stdio:out", MSGFILE, MSGSTDIO, NULL);
	if (echo)
		MessageConfig(TRICKLET_ERROR, "Snmp-Echo");
	else
		MessageConfig(TRICKLET_ERROR, "Snmp-Trap");

	if (MibDfInit() != 0 || MibDfFullMib() != 0)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "MibDfInit failed");
		DnpapExit(1);
	}

	if (NwInit() == FALSE)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwInit failed");
		DnpapExit(1);
	}

#ifdef OS2
	if (!BlockInit())
		DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "BlockInit failed");
#else
	if (BlockInit() != 0)
		DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "BlockInit failed");
#endif

	if (echo)
		port = 161;
	else
		port = 162;

	sprintf(address, "snmp:%d", port);
	if ((acceptdg = NwDgAccept(address, Callback, NULL)) == NULL)
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "NwDgAccept failed");
		DnpapExit(1);
	}

#ifdef OS2
	if (!Block())
		DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "Block exit !");
#else
	while (1)
		if (Block() != 0)
			DnpapMessage(DMC_WARNING, TRICKLET_ERROR, "Block exit !");
#endif
}


VOID Callback(NW_DG * dg, CHAR *addr, BYTE * frame, INT length, VOID * param)
{
NwSnmpFrame *pdu = (NwSnmpFrame *)(VOID *)frame;
INT i;
CHAR address[NWSNMP_SIZE_ADDRESS];
CHAR host[NWSNMP_SIZE_HOST];
INT port;
CHAR request[NWSNMP_SIZE_REQUEST];
WORD type;
CHAR community[NWSNMP_SIZE_COMMUNITY];
INT commlen;
CHAR sender[SIZE_HOST];
CHAR timeticks[SIZE_TIME];
CHAR oid[SIZE_VAR];
CHAR *status, *general;

	if (Addr2Snmp(addr, host, &port, request, &type, community, &commlen) == FALSE)
	{
        DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "invalid parameter in SNMP address");
        return;
	}
					  
	PrintHost2(host, port, sender);

	ToPrint(community, commlen);

	printf("%s: %s \"%*s\" ", request, sender, commlen, community);

	if (type == SNMP_PDU_GET || type == SNMP_PDU_SET || type == SNMP_PDU_NEXT)
	{
		printf("%d ", pdu->Pdu.Request.Id);
		sprintf(address, "snmp:%s:%d:%s:%*s", host, port, "response", commlen, community);
		NwDgSendTo(acceptdg, address, frame, length);
	}
	else
	if (type == SNMP_PDU_TRAP)
	{
		PrintMibName(pdu->Pdu.Trap.Id, pdu->Pdu.Trap.IdLen, oid);

		PrintHost(pdu->Pdu.Trap.IpAddress, sender);

		switch (pdu->Pdu.Trap.General)
		{
		case SNMP_TRAP_COLDSTART:
			general = "coldStart";
			break;
		case SNMP_TRAP_WARMSTART:
			general = "warmStart";
			break;
		case SNMP_TRAP_LINKDOWN:
			general = "linkDown";
			break;
		case SNMP_TRAP_LINKUP:
			general = "linkUp";
			break;
		case SNMP_TRAP_AUTFAILURE:
			general = "authenticationFailure";
			break;
		case SNMP_TRAP_EQPNEIGHBORLOSS:
			general = "neighborLoss";
			break;
		case SNMP_TRAP_ENTSPECIFIC:
			general = "specific";
			break;
		default:
			general = "unknown";
			break;
		}

		PrintTimeTicks(pdu->Pdu.Trap.Time, timeticks);

		printf("%s %s %s %s(%d) %d", 
				sender, timeticks, oid, general, pdu->Pdu.Trap.General, pdu->Pdu.Trap.Specific);
	}
	else
	if (type == SNMP_PDU_RESPONSE)
	{
		switch (pdu->Pdu.Request.Status)
		{
		case SNMP_NOERROR:
			status = "noError";
			break;
		case SNMP_TOOBIG:
			status = "tooBig";
			break;
		case SNMP_NOSUCHNAME:
			status = "noSuchName";
			break;
		case SNMP_BADVALUE:
			status = "badValue";
			break;
		case SNMP_READONLY:
			status = "readOnly";
			break;
		case SNMP_GENERROR:
			status = "genericError";
			break;
		default:
			status = "unknown";
			break;
		}

		printf("%d %s(%d) %d", pdu->Pdu.Request.Id, status, pdu->Pdu.Request.Status, pdu->Pdu.Request.Index);
	}
	else
	{
		DnpapMessage(DMC_ERROR, TRICKLET_ERROR, "invalid PDU request type received");
	}

	printf("\n");

	for (i = 0; i < pdu->NrObjects; i++)
		PrintMibVar(&(pdu->Objects[i]));

	printf("end\n");

	fflush(stdout);
}


INT ToPrint(BYTE *b, INT len)
{
INT i;

	for (i = 0; i < len; i++)
        if (b[i] < 32 || b[i] > 127)
            b[i] = '.';

	return len;
}


INT IsPrint(BYTE *b, INT len)
{
INT i, n = 0;

	for (i = 0; i < len; i++)
        if (b[i] >= 32 && b[i] <= 127)
            n++;

	return n;
}


INT PrintMibVar(SNMP_OBJECT *object)
{
INT i;
CHAR *type = NULL, value[SIZE_VALUE], var[SIZE_VAR], t[SIZE_TIME], *p;

	PrintMibName(object->Id, object->IdLen, var);

	switch (object->Type)
	{
	case SNMP_NULL:
		type = "NULL";
		strcpy(value, "");
		break;
	case SNMP_INTEGER:
		type = "INTEGER";
		sprintf(value, "%ld", object->Syntax.LngInt);
		break;
	case SNMP_OPAQUE:
	case SNMP_OCTETSTR:
	/* case SNMP_DISPLAYSTR: */
		if (object->Type == SNMP_OCTETSTR)
			type = "OCTETSTRING";
		else
			type = "Opaque";
		if ((FLOAT)IsPrint(object->Syntax.BufChr, object->SyntaxLen)/object->SyntaxLen > 0.75)
		{
			memcpy(value, object->Syntax.BufChr, object->SyntaxLen);
			ToPrint(value, object->SyntaxLen);
			value[object->SyntaxLen] = '\0';
		}
		else
		{
			value[0] = '\0';
			for (i = 0, p = value; i+1 < object->SyntaxLen; i++)
			{
				sprintf(p, "%2.2x:", object->Syntax.BufChr[i]);
				p += strlen(p);
			}
			if (i < object->SyntaxLen)
			{
				sprintf(p, "%2.2x", object->Syntax.BufChr[i]);
			}
		}
		break;
	case SNMP_OBJECTID:
		type = "OBJECTID";
		PrintMibName(object->Syntax.BufInt, object->SyntaxLen, value);
		break;
	case SNMP_IPADDR:
		type = "IpAddress";
		sprintf(value, "%d.%d.%d.%d", object->Syntax.BufChr[0], 
									  object->Syntax.BufChr[1],
									  object->Syntax.BufChr[2],
									  object->Syntax.BufChr[3]);
		break;
	case SNMP_COUNTER:
		type = "Counter";
		sprintf(value, "%lu", object->Syntax.LngUns);
		break;
	case SNMP_GAUGE:
		type = "Gauge";
		sprintf(value, "%lu", object->Syntax.LngUns);
		break;
	case SNMP_TIMETICKS:
		type = "TimeTicks";
		PrintTimeTicks(object->Syntax.LngUns, t);
		sprintf(value, "%s", t);
		break;
	}

	return printf("%s  %s  %s\n", var, type, value);
}


INT PrintMibName(LONG *id, INT len, CHAR *s)
{
CONST CHAR *p;
CHAR *q;
INT i, n;

	p = MibDfID2Name2(id, len, &n);

	if (p == NULL)
		strcpy(s, "");
	else
		strcpy(s, p);

	if (len - n > 0)
	{
		strcat(s, "[");
		for (i = n, q = s+strlen(s); i < len-1; i++)
		{
			sprintf(q, "%ld.", id[i]);
			q += strlen(q);
		}
		if (i < len)
		{
			sprintf(q, "%ld", id[i]);
		}
		strcat(s, "]");
	}

	if (n > 0)
	{
		strcat(s, "(");
		for (i = 0, q = s+strlen(s); i < len-1; i++)
		{
			sprintf(q, "%ld.", id[i]);
			q += strlen(q);
		}
		if (i < len)
		{
			sprintf(q, "%ld", id[i]);
		}
		strcat(s, ")");
	}

	return strlen(s);
}


INT PrintTimeTicks(ULONG timeticks, CHAR *str)
{
INT h, m, s;
ULONG t;
		
	t = timeticks;
	h = t/360000UL;
	t -= h*360000UL;
	m = t/6000UL;
	t -= m*6000UL;
	s = t/100UL;
	t -= s*100UL;

	sprintf(str, "%.02d:%.02d:%02d.%02lu(%lu)", h, m, s, t, timeticks);

	return strlen(str);
}


INT PrintHost(ULONG address, CHAR *host)
{
HOSTENT *ent;
BYTE *b;

	if ((ent = gethostbyaddr((CHAR *)&address, sizeof(address), AF_INET)) == NULL)
		strcpy(host, "");
	else
		strcpy(host, ent->h_name);

	if (ent != NULL)
		strcat(host, "(");

	b = (BYTE *)&address;
	sprintf(host+strlen(host), "%d.%d.%d.%d", b[0], b[1], b[2], b[3]);

	if (ent != NULL)
		strcat(host, ")");

	return strlen(host);
}


INT PrintHost2(CHAR *address, INT port, CHAR *host)
{
	PrintHost(inet_addr(address), host);
	sprintf(host, "%s:%d", host, port);
	return strlen(host);
}
